/*
 * Service providers for the
 * (Hardware) Address Resolution Protocol, "ARP", (RFC 826) and the
 * Reverse (Hardware) Address Resolution Protocol, "RARP"/"REVARP" (RFC 903)
*/
#define __OSILAYER__ 3 /* actually,... its somewhere between 2 and 3 */

#undef  TRACE_CACHE   /* #define this to trace all cache operations */
#undef  TRACE_NETIO   /* #define this to trace all network operations */
#undef  TRACE_RESOLVE /* #define this to trace _arp_resolve() */

#include "target.h"   /* standard types/defines */
#include "debug.h"    /* KMSG(), DBGPRINTF() */
#include "printk.h"   /* interrupt-time debugging */
#include "sigblk.h"   /* set_cbreak() et al */
#include "timer.h"    /* timer_*(), Random() */
#include "blocker.h"  /* blocker_*() */
#include "ctordtor.h" /* _my_atexit() */
#include "linklayr.h" /* _eth_*() */
#include "icmp.h"     /* icmp_router_solic() */
#include "inaddr.h"   /* is_multicast(), _my_aton(), _my_atoeth() */
#include "loadcfg.h"  /* config things */
#include "ipbase.h"   /* _ip_get_ipaddr(), _ip_get_snmask(), _ip_is_up() */
#include "arp.h"      /* ourselves */

/* ======================================================================= */
/*             ARP (RFC 826) and RARP (RFC 903) spec definitions           */
/* ======================================================================= */

/*
 * opcodes (in net order)
 */
#define ARP_REQUEST    0x0100      /* ARP/RARP op codes, Request   */
#define ARP_REPLY      0x0200      /*                    Reply     */
#define RARP_REQUEST   0x0300
#define RARP_REPLY     0x0400

/*
 * ARP/RARP headers
 */
#pragma pack(1) /* align all structure members on byte boundaries */
typedef struct arp_rarp_Header {
  uint8_t   __hwType_hi;   /* unused - types are always 8 bit */
  uint8_t   hwType;        /* type of hardware address */
  uint16_t   protType;      /* type of protocol address (ETHERTYPE_IP) */
  uint8_t   hwAddrLen;     /* length of hardware address */
  uint8_t   protAddrLen;   /* length of protocol address */
  uint16_t   opcode;        /* {ARP|RARP}_{REQUEST|REPLY} */
#define ARP_RARP_FIXED_HDR_SIZE 8
  struct { /* For info only. structure is never addressed directly */
    uint8_t   srcHWAddr[MAX_LINKADDR_SIZE];  /* hw address of sender (variable len) */
    uint32_t  srcIPAddr;     /* protocol address of sender (4 for IP) */
    uint8_t   dstHWAddr[MAX_LINKADDR_SIZE];  /* "target" hw address (variable len) */
    uint32_t  dstIPAddr;     /* "target" protocol address (4 for IP) */
  } data;
} arp_rarp_Header;
#pragma pack() /* restore default packing */

/* ======================================================================= */

#if defined(TRACE_CACHE) && !defined(NDEBUG)
#undef TRACE_CACHE
#define TRACE_CACHE(x) DBGPRINTF(x)
#else
#undef TRACE_CACHE
#define TRACE_CACHE(x) NO_OPERATION()
#endif
#if defined(TRACE_NETIO) && !defined(NDEBUG)
#undef TRACE_NETIO
#define TRACE_NETIO(x) DBGPRINTF(x)
#else
#undef TRACE_NETIO
#define TRACE_NETIO(x) NO_OPERATION()
#endif
#if defined(TRACE_RESOLVE) && !defined(NDEBUG)
#undef TRACE_RESOLVE
#define TRACE_RESOLVE(x) DBGPRINTF(x)
#else
#undef TRACE_RESOLVE
#define TRACE_RESOLVE(x) NO_OPERATION()
#endif

/* ======================================================================= */

#define ARP_RESOLVE_TIMEOUT          4u /* 4 seconds */
#define ARP_CACHE_FOUND_TIMEOUT   1200u /* 20 minutes */
#define ARP_CACHE_TEMP_TIMEOUT      30u /* 30 seconds */

#define ARP_FLAG_TEMP     1
#define ARP_FLAG_FOUND    2  /* IP-addr has been resolved */
#define ARP_FLAG_FIXED  255  /* IP/hw address added via .cfg */

struct arp_table 
{
  uint32_t       ip;
  uint8_t        hardware[MAX_LINKADDR_SIZE];
  uint8_t        flags;
  uint8_t        bits;      /* bits in network */
  uint32_t       expiry;
};

struct gate_table 
{
  uint32_t  gate_ip;
  uint32_t  subnet;
  uint32_t  mask;
};

struct redirrec_t
{
  uint32_t  dest_ina;
  uint32_t  gate_ina;
};


struct __resolve_block
{
  uint16_t  ethtype;         /* { ETHERTYPE_ARP } or { ETHERTYPE_REVARP } */
  uint32_t *ip_ptr;         /* {   question    } or {      answer      } */
  void *eth_ptr;         /* {    answer     } or {     question     } */
};

/* ====================================================================== */
/*                           private data                                 */
/* ====================================================================== */

static struct
{
  /* the max number of entries in the ARP-cache was 40 but that many is
   * now overkill since arp entries outside our own subnet are no longer
   * cached. In such a case, its the *gateway* thats important. - cyp 5/2001
  */
  struct arp_table  arp_cache_data[10 /* max # of entries in ARP-cache */];
  struct gate_table arp_gate_list [3  /* max # of gateways */];
  struct redirrec_t redir_list[8];    /* LUFO. Does not need to be big */

  unsigned arp_num_gateways;
  int  debug_lev; 
  struct __resolve_block *resolve_block;

  unsigned long searched_count;
  unsigned long hit1_count;
  unsigned long hit2_count;
  unsigned long early_del;
  unsigned long new_entry;

} FAR my_statics;

#define arp_cache_data   (my_statics.arp_cache_data)
#define arp_gate_list    (my_statics.arp_gate_list)
#define resolve_block    (my_statics.resolve_block)
#define redir_list       (my_statics.redir_list)

/* ====================================================================== */
/*                           gateway list                                 */
/* ====================================================================== */

/*
 * _arp_add_gateway - add a gateway to the routing table
 */
void _arp_add_gateway (uint32_t ip, uint32_t subnet, uint32_t mask)
{
  unsigned int i;

  if (ip && !is_local_addr(ip) && ip != ((uint32_t)-1)) /* hrm! */
  {
    for (i = 0; i < my_statics.arp_num_gateways; i++)
    {
      if (arp_gate_list[i].gate_ip == ip)
      {
        /* discard the old entry for replacement with the new */
        my_statics.arp_num_gateways--; /* move the others up */
        FAR_MEMMOVE( &arp_gate_list[i], &arp_gate_list[i+1],
                (my_statics.arp_num_gateways-i) * sizeof(arp_gate_list[i]) );
        break; 
      }
    }

    if (my_statics.arp_num_gateways <
        (sizeof(arp_gate_list)/sizeof(arp_gate_list[0])))
    {
      i = my_statics.arp_num_gateways++;
      arp_gate_list[i].gate_ip = ip;
      arp_gate_list[i].subnet  = subnet;
      arp_gate_list[i].mask    = mask;
      if (my_statics.debug_lev >= 1)
        DBGPRINTF(( "*_arp_add_gateway(%hI,%hI,%hI) [arp_num_gateways=%d]\n",
                    ip,subnet,mask,my_statics.arp_num_gateways ));
    }
  }
  return;
}

/* ======================================================================= */
/*                           cache management                              */
/* ======================================================================= */

static struct arp_table FAR *_arp_search (uint32_t ip, int create)
{
  int freearp = -1, notfound = -1, oldest = -1;
  struct arp_table FAR *arp;
  uint32_t oldest_time = 0;
  unsigned int i; 
  unsigned long now = timer_sec_count();

  my_statics.searched_count++;

  for (i = 0; i < (sizeof(arp_cache_data)/sizeof(arp_cache_data[0])); i++)
  {
    arp = &arp_cache_data[i];
    if (ip == arp->ip)
    {
      if (arp->flags == ARP_FLAG_TEMP)
      {
        arp->flags = ARP_FLAG_FOUND;
        arp->expiry = now + ARP_CACHE_FOUND_TIMEOUT;
      }
      if (arp == &(arp_cache_data[0]) || arp == &(arp_cache_data[1]))
        my_statics.hit1_count++;
      else
        my_statics.hit2_count++;
      if (my_statics.debug_lev >= 2)
        TRACE_CACHE(("*ARP cache hit %d for %hI\n", i, ip ));
      return (arp);
    }
    if (create && freearp == -1)
    {
      if (!arp->ip) /* slot is empty */
      {
        freearp = i;
      }
      else if (arp->flags == 0) /* pending resolve */
      {
        if (notfound == -1 && (now > arp->expiry)) /* expired */
          notfound = i;
      }
      else if (arp->flags != ARP_FLAG_FIXED)
      {
        if (arp->expiry == 0 || (now > arp->expiry)) /* expired */
        {
          oldest = i;
          oldest_time = 0;
        }
        else if (oldest == -1 || arp->expiry < oldest_time)
        {
          oldest = i;
          oldest_time = arp->expiry;
        }
      }
    }
  }
  if (create)
  {
    if (freearp == -1)
      freearp = notfound;
    if (freearp != -1)
    {
      if (my_statics.debug_lev >= 2)
        TRACE_CACHE(("*ARP free cache entry %u for %hI\n", freearp, ip ));
    } 
    else if (oldest != -1)
    {
      freearp = oldest;
      arp = &arp_cache_data[freearp];
      if (my_statics.debug_lev >= 2)
        TRACE_CACHE(("*ARP %sexpired cache entry %u for %hI\n", 
           (oldest_time ? "clobbered un" : ""), freearp, ip ));
      if (oldest_time)
        my_statics.early_del++;
    }
    else /* all entries are fixed! */
    {
      freearp = (sizeof(arp_cache_data)/sizeof(arp_cache_data[0])) - 1;
      if (my_statics.debug_lev >= 2)
        TRACE_CACHE(("*ARP discarded FIXED cache entry %u for %hI\n", 
             freearp, ip ));
      my_statics.early_del++;
    }
    my_statics.new_entry++;
    arp = &arp_cache_data[freearp];
    arp->ip = ip; 
    arp->flags = 0; 
    arp->expiry = 0; 
    return arp;
  }
  if (my_statics.debug_lev >= 2)
    TRACE_CACHE(("*ARP cache miss %hI\n", ip ));
  return (struct arp_table FAR *)0;
}

/* ------------------------------------------------------------------ */

#if 1
#define __speedup_cache(__centry) (__centry = __centry)
#else
/* __speedup_cache() is truly useful only if there 
 * are <= two hosts (including all gateways) being tracked. 
 * ie, it works best in the most common scenario of one 
 * gateway plus one server being connected to.
*/
static struct arp_table FAR *__speedup_cache(struct arp_table FAR *arp)
{
  if (arp != &(arp_cache_data[0]) && arp != &(arp_cache_data[1]))
  {
    unsigned int i;
    for (i = 2; i < (sizeof(arp_cache_data)/sizeof(arp_cache_data[0])); i++)
    {
      if (arp == &(arp_cache_data[i]))
      {
        struct arp_table tmp;
        FAR_MEMCPY( &tmp, arp, sizeof(tmp));
        FAR_MEMMOVE( &(arp_cache_data[1]), &(arp_cache_data[0]), 
                 (i * sizeof(tmp)) );
        arp = &(arp_cache_data[0]);
        FAR_MEMCPY( arp, &tmp, sizeof(tmp) );
        if (my_statics.debug_lev >= 1)
          TRACE_CACHE(("*ARP spedup cache entry (%hI->%M)\n", 
                   tmp.ip, &(tmp.hardware[0]) ));
        break;
      }
    }
  }
  return arp;
}
#endif

/* ------------------------------------------------------------------ */

/*
 * Add given IP/Ether address to ARP-cache.
 * 'ip' is on host order.
 */
void _arp_add_entry (uint32_t ip, const void FAR *hwaddr, int expires)
{
  struct arp_table FAR *arp;
  uint8_t tmp_hwaddr[MAX_LINKADDR_SIZE];
  int addr_len = _eth_get_hwaddrlen();
  int new_entry = 0;

  /* filter out things we don't need to cache */
  if (addr_len == 0)
    return;
  if ((((const uint8_t FAR *)hwaddr)[0] & 1)) /* broad/multicast */
    return;
  if (FAR_MEMCMP(hwaddr, _eth_get_hwaddr(), addr_len)==0)
    return;
  memset( &tmp_hwaddr[0], 0, sizeof(tmp_hwaddr));
  if (FAR_MEMCMP( hwaddr, &tmp_hwaddr[0], addr_len) == 0)
    return;
  if (ip == 0ul || is_local_addr(ip) || is_broadcast(ip) || is_multicast(ip))
    return;
  if (!is_onmysubnet(ip)) /* IP addrs not on our subnet get */
    return;               /* a gateway's ethernet address */

  FAR_MEMCPY( tmp_hwaddr, hwaddr, addr_len );
  blocker_critsect_begin();

  arp = _arp_search (ip, 1);
  if (arp->flags == 0) /* new one was created */
  {
    if (my_statics.debug_lev >= 2)
      TRACE_CACHE(("*ARP created cache entry for (%hI->%M)\n", ip, tmp_hwaddr));
    new_entry = 1;
  }

  if (!expires)
  {
    arp->expiry = 0;
    arp->flags  = ARP_FLAG_FIXED;
    if (my_statics.debug_lev >= 1)
      TRACE_CACHE(("*ARP %s FIXED cache entry for %hI->%M\n",
                   (new_entry ? "created" : "updated"), ip, tmp_hwaddr ));
  }
  else if (new_entry)
  {
    arp->expiry = timer_sec_count() + ARP_CACHE_TEMP_TIMEOUT;
    arp->flags  = ARP_FLAG_TEMP;
    if (my_statics.debug_lev >= 1)
      TRACE_CACHE(("*ARP created TEMP cache entry (%hI->%M)\n", ip, tmp_hwaddr ));
  }
  else if (arp->flags == ARP_FLAG_FOUND)
  {
    arp->expiry = timer_sec_count() + ARP_CACHE_FOUND_TIMEOUT;
    arp->flags  = ARP_FLAG_FOUND;
    if (my_statics.debug_lev >= 2)
      TRACE_CACHE(("*ARP updated DYN cache entry (%hI->%M)\n", ip, tmp_hwaddr ));
  }
  arp->ip = ip;
  FAR_MEMCPY(&arp->hardware[0], tmp_hwaddr, addr_len);

  __speedup_cache( arp );

  blocker_critsect_end();

  return;
}

/* ------------------------------------------------------------------ */

static void __shutdown(void) /* called via an atexit hook() */
{
#if !defined(NDEBUG)
  if (my_statics.debug_lev)
    DBGPRINTF(("* ---------- ARP/RARP shutdown begin\n"));

  if (my_statics.debug_lev)
  {
    uint32_t now = 0L;
    unsigned int i, printed;

    printed = 0;
    for (i = 0; i < my_statics.arp_num_gateways; i++)
    {
      struct gate_table FAR *gw = &arp_gate_list[i];
      if ((++printed) == 1)
      {
        DBGPRINTF(( "*Routing data:\n" ));
        DBGPRINTF(( "*   Network           Router            Mask\n" ));
      }
      DBGPRINTF(("*   %-16hI  %-16hI  %hI\n", 
                        gw->subnet, gw->gate_ip, gw->mask ));
    }
    printed = 0;
    for (i = 0; i < (sizeof(arp_cache_data)/sizeof(arp_cache_data[0])); i++)
    {
      struct arp_table FAR *arp = &arp_cache_data[i];
      if (arp->flags != 0)
      {
        char tmp_hwaddr[sizeof(arp->hardware)]; /* needs to be near */
        if ((++printed) == 1)
        {
          DBGPRINTF(( "*ARP Cache:\n"));
          DBGPRINTF(( "*   IP address        Physical address   Type     Expires\n"));
          now = timer_sec_count();
        }
        FAR_MEMCPY( tmp_hwaddr, arp->hardware, sizeof(arp->hardware) );
        if (arp->flags == ARP_FLAG_FIXED)
        {
          DBGPRINTF(( "*   %-16hI  %M  fixed    ---.--\n",
                       arp->ip, tmp_hwaddr ));
        }
        else if (arp->expiry < now)
        {
          DBGPRINTF(( "*   %-16hI  %M  expired  -%02lu.00s\n",
                       arp->ip, tmp_hwaddr, now - arp->expiry ));
        }
        else
        {
          DBGPRINTF(( "*   %-16hI  %M  dynamic  +%02lu.00s\n",
                       arp->ip, tmp_hwaddr, arp->expiry - now ));
        }
      }
    }
    DBGPRINTF(("*Cache efficacy:\n"));
    DBGPRINTF(("*  Insertions: %lu\n", my_statics.new_entry ));
    DBGPRINTF(("*    premature discards: %lu, timeout or free: %lu\n", 
          my_statics.early_del, my_statics.new_entry-my_statics.early_del ));
    DBGPRINTF(("*  Searches: %lu\n", 
                   my_statics.searched_count - my_statics.new_entry ));
    DBGPRINTF(("*    Hits: %lu  (first/second rec: %lu  3rd-%uth rec: %lu)\n",
     (my_statics.hit1_count+my_statics.hit2_count), my_statics.hit1_count,
     (sizeof(arp_cache_data)/sizeof(arp_cache_data[0])), my_statics.hit2_count));
    DBGPRINTF(("*    Misses: %lu\n", 
                     (my_statics.searched_count - my_statics.new_entry) -
                     (my_statics.hit1_count + my_statics.hit2_count) ));
  }

  if (my_statics.debug_lev)
    DBGPRINTF(("* ---------- ARP/RARP shutdown complete\n"));

#endif /* !NDEBUG */
  return;
}

/* ------------------------------------------------------------------ */

/*
** Find the ether address of 'ip' but DONT do any net I/O.
** The true "inverse" of _arp_add_entry, and the sibling of _arp_resolve().
**
** Set 'expires' if that is not a static entry. 
*/
int _arp_find_entry (uint32_t ip, void FAR *hwaddr, int FAR *expires)
{
  static int shutdown_hook_set = 0;
  struct arp_table FAR *tab;
  int addr_len;

  if (!_ip_is_up())
    return (0);

  /* shutdown hook needs to be set after IP has been initialized
   * so that ARP shutdown occurs before IP is shutdown.
  */ 
  if (!shutdown_hook_set)
  {
    shutdown_hook_set = 1;
    /* KMSG((0,"_my_atexit(__arp_shutdown)\n")); */
    _my_atexit(__shutdown);
  }

  addr_len =  _eth_get_hwaddrlen();
  if (addr_len == 0)
  {
    if (expires)
      *expires = 0;
    return (1);
  }

  if (ip == ((uint32_t)-1) || is_broadcast(ip))
  {
    if (hwaddr)
      FAR_MEMCPY( hwaddr, _eth_get_hwaddrbc(), addr_len);
    if (expires)
      *expires = 0;
    return (1);
  }    

  if (!ip || is_local_addr(ip))
  {
    if (hwaddr)
      FAR_MEMCPY(hwaddr, _eth_get_hwaddr(), addr_len);
    if (expires)
      *expires = 0;
    return (1);
  }

  if (is_multicast(ip))   /* IP addr is class D */
  {
    if (hwaddr)
    {
      uint8_t *eth = (uint8_t *)hwaddr;
      eth[0] = (uint8_t)0x01;
      eth[1] = (uint8_t)0x00;
      eth[2] = (uint8_t)0x5E;
      eth[3] = (uint8_t)((ip >> 16) & 0x7F);
      eth[4] = (uint8_t)((ip >>  8) & 0xFF);
      eth[5] = (uint8_t)((ip >>  0) & 0xFF);
    }
    if (expires)
      *expires = 0;
    return (1);
  }

  blocker_critsect_begin();

  tab = _arp_search (ip, 0);
  if (tab && tab->flags != 0)  /* has been resolved */
  {
    if (hwaddr)
      FAR_MEMCPY(hwaddr, &tab->hardware[0], addr_len);
    if (expires)
      *expires = (tab->flags != ARP_FLAG_FIXED);
  }
  else
  {
    tab = (struct arp_table FAR *)0;
  }

  blocker_critsect_end();
  return (!(!tab));
}

/* ------------------------------------------------------------------ */

int _arp_del_entry(uint32_t ip) /* only used by BSD ioctl() */
{
  struct arp_table FAR *tab;
  unsigned int i;
  for (i = 0; i < my_statics.arp_num_gateways; i++)
  {
    if (ip == arp_gate_list[i].gate_ip)
      return 0;
  }

  blocker_critsect_begin();
  tab = _arp_search (ip, 0);
  if (tab)
    FAR_MEMSET( tab, 0, sizeof(*tab));

  blocker_critsect_end();
  return (!(!tab));
}

/* ======================================================================= */
/*                           ARP/RARP sender                               */
/* ======================================================================= */

static void __get_addr_ptrs( /*non-const*/ arp_rarp_Header /*near*/ *arp,
                             uint8_t **src_hw_addr_ptr,
                             uint32_t **src_ip_addr_ptr,
                             uint8_t **dst_hw_addr_ptr,
                             uint32_t **dst_ip_addr_ptr)
{
  int hw_addr_len = _eth_get_hwaddrlen();
  *src_hw_addr_ptr = (uint8_t *)(((uint8_t *)arp) + ARP_RARP_FIXED_HDR_SIZE);
  *src_ip_addr_ptr = (uint32_t *)(*src_hw_addr_ptr + hw_addr_len);
  *dst_hw_addr_ptr = *src_hw_addr_ptr + hw_addr_len + sizeof(uint32_t);
  *dst_ip_addr_ptr = (uint32_t *)(*dst_hw_addr_ptr + hw_addr_len);
  return;
}  

#define _arp_request(__ip)   _arp_rarp_send( ARP_REQUEST, 0, 0, __ip)
#define _rarp_request(__eth) _arp_rarp_send( RARP_REQUEST, 0, __eth, 0)

static int _arp_rarp_send( uint16_t opcode, uint32_t src_ip /* in host order */,
                           const void FAR *tgt_ethap, 
                           uint32_t tgt_ip /* in host order */)
{
  arp_rarp_Header arp;
  uint16_t ethtype = 0;
  uint8_t *src_hw_addr_ptr; uint32_t *src_ip_addr_ptr;
  uint8_t *dst_hw_addr_ptr; uint32_t *dst_ip_addr_ptr;
  int  arp_pkt_size;

  if (!_ip_is_up())
    return (0);

  arp_pkt_size = (ARP_RARP_FIXED_HDR_SIZE + 
                       2*(sizeof(uint32_t) + _eth_get_hwaddrlen()));

  memset( &arp, 0, sizeof(arp) );
  arp.hwType        = _eth_get_hwtype(); /* uint8_t */
  /* one char pad here */
  arp.protType      = HTONS(ETHERTYPE_IP); /* protocol addr type */
  arp.hwAddrLen     = _eth_get_hwaddrlen(); /* == sizeof(eth_addr) */
  arp.protAddrLen   = (uint8_t)sizeof(uint32_t); /* sizeof(ip address) */
  arp.opcode        = opcode; /* uint16_t in net order */

  __get_addr_ptrs( &arp, &src_hw_addr_ptr, &src_ip_addr_ptr,
                         &dst_hw_addr_ptr, &dst_ip_addr_ptr );

  FAR_MEMCPY(src_hw_addr_ptr, _eth_get_hwaddr(), arp.hwAddrLen);
  *src_ip_addr_ptr = HTONL(_ip_get_ipaddr());

  if (opcode == ARP_REQUEST)
  {
    /* (sender)   src_ethap is our own etheraddr */
    /* (sender)   src_ip    is our own IP address */
    /* (  --  )   tgt_ethap is undefined */
    /* (question) tgt_ip    is the IP address who's etheraddr we want */
    memset (dst_hw_addr_ptr, 0, arp.hwAddrLen);
    *dst_ip_addr_ptr = HTONL(tgt_ip);
    tgt_ethap = _eth_get_hwaddrbc(); /* broadcast address */
    ethtype = ETHERTYPE_ARP;

    if (my_statics.debug_lev >= 1)
      TRACE_NETIO(("*ARP_REQUEST: Qfrom=%M/%I question=%I\n", 
                  src_hw_addr_ptr, *src_ip_addr_ptr,
                  *dst_ip_addr_ptr ));
  }
  else if (opcode == ARP_REPLY)
  {
    /* (answer) src_ethap is our own etheraddr (the answer) */
    /* (question) src_ip is one of our IP addresses (the question) */
    /* (  --  ) tgt_ethap is the questioner's etheraddr */
    /* (  --  ) tgt_ip    is the questioner's IP */
    *src_ip_addr_ptr = HTONL(src_ip); /* ip address that was in the REQUEST.dst */
    FAR_MEMCPY(dst_hw_addr_ptr, tgt_ethap, arp.hwAddrLen);
    *dst_ip_addr_ptr = HTONL(tgt_ip);
    ethtype = ETHERTYPE_ARP;
    /* tgt_ethap is the host who sent the request */    

    if (my_statics.debug_lev >= 1)
      TRACE_NETIO(("*ARP_REPLY: answer=%M question=%I Qfrom=%M/%I\n",
                  src_hw_addr_ptr, *src_ip_addr_ptr,
                  dst_hw_addr_ptr, *dst_ip_addr_ptr  )); 
  }
  else if (opcode == RARP_REQUEST)
  { 
    /* (sender)   src_ethap is our own etheraddr */
    /* (  --  )   src_ip    is undefined */
    /* (question) tgt_ethap is the ether_addr whose IP we want */
    /* (  --  )   tgt_ip    is undefined */
    FAR_MEMCPY(dst_hw_addr_ptr, tgt_ethap, arp.hwAddrLen);
    *dst_ip_addr_ptr = 0ul;
    tgt_ethap = _eth_get_hwaddrbc(); /* broadcast address */
    ethtype = ETHERTYPE_REVARP;
    if (my_statics.debug_lev >= 1)
      TRACE_NETIO(("*RARP_REQUEST: sender=%M question=%M\n",
                  src_hw_addr_ptr, dst_hw_addr_ptr )); 
  }
  else /* RARP_REPLY (mock reply since we are never a RARP server) */
  {
    /* (  --  )   src_ethap is our own etheraddr (we are the sender) */
    /* (  --  )   src_ip    is our IP address (we are the sender) */
    /* (question) tgt_ethap is our ether addr (the question unchanged) */
    /* (answer)   tgt_ip    is our IP addr (the answer) */
    FAR_MEMCPY(dst_hw_addr_ptr, _eth_get_hwaddr() /* was in the REQUEST.dst */,
               arp.hwAddrLen );
    *dst_ip_addr_ptr = HTONL(_ip_get_ipaddr());
    ethtype = ETHERTYPE_REVARP;
    /* tgt_ethap is the host who sent the request */    
    if (my_statics.debug_lev >= 1)
      TRACE_NETIO(("*RARP_REPLY: sender=%M/%I, question=%M, answer=%I\n",
                  src_hw_addr_ptr, *src_ip_addr_ptr,
                  dst_hw_addr_ptr, *dst_ip_addr_ptr  )); 
  }

  return _eth_send ( ethtype,   /* in host order */ 
                     tgt_ethap, /* target ether addr */
                     &arp,      /* packet to send */
                     arp_pkt_size /* length of packet */
                     );
}

/* ======================================================================= */
/*                           ARP/RARP packet handler                       */
/* ======================================================================= */

/*
 * _arp_rarp_handler - handle incoming ARP/RARP packets
 * Called (by trampoline) by stuff in pcsed.
 *
 * Prototype must match that used by _ip_handler().
*/
static void _arp_rarp_handler( const void FAR *packet, unsigned int pkt_len,
                               const void *src_hw_addr, int is_brdcast )
{
  int cache_refresh_mode = 0; /* 0=no refresh, 1=passive refresh, 2=active */
  char ahbuf[ARP_RARP_FIXED_HDR_SIZE + 2*(sizeof(uint32_t)+MAX_LINKADDR_SIZE)];
  uint8_t *src_hw_addr_ptr; uint32_t *src_ip_addr_ptr;
  uint8_t *dst_hw_addr_ptr; uint32_t *dst_ip_addr_ptr;
  const char *ignmsg = "Ignored";
  arp_rarp_Header *ah;
  uint32_t ip;

  if (!_ip_is_up())
    return;

  if (pkt_len < (ARP_RARP_FIXED_HDR_SIZE + 
                    2*(sizeof(uint32_t) + _eth_get_hwaddrlen())))
  {
    if (my_statics.debug_lev >= 1)
      TRACE_NETIO(("*ARP/RARP DROP (too small)\n"));
    return;
  }

  ah = (arp_rarp_Header *)&ahbuf[0];
  FAR_MEMCPY(ah, packet, (pkt_len < sizeof(ahbuf) ? pkt_len : sizeof(ahbuf)));

  if (my_statics.debug_lev >= 1)
    TRACE_NETIO(("*(R)ARP IN from %M, op=%04x, hwType/Len=%02x/%d, etht=%04x\n",
      src_hw_addr, ah->opcode, ah->hwType, ah->hwAddrLen, ah->protType));

  if (ah->hwType    != _eth_get_hwtype() || /* wrong hardware type, */
      ah->hwAddrLen != _eth_get_hwaddrlen() || /* wrong hw addr len */
      NTOHS(ah->protType) != ETHERTYPE_IP)  /* or not IP-protocol   */
  {
    if (my_statics.debug_lev >= 1)
      TRACE_NETIO(("*ARP/RARP DROP (wrong type, bad len)\n"));
    return;
  }

  __get_addr_ptrs( ah, &src_hw_addr_ptr, &src_ip_addr_ptr,
                       &dst_hw_addr_ptr, &dst_ip_addr_ptr );

  if (ah->opcode == ARP_REQUEST) /* someone wants an ether address */
  {
    /* even if doesn't want our address, refresh cache if in there already */
    cache_refresh_mode = 1; /* passive refresh */

    ip = NTOHL(*dst_ip_addr_ptr);
    if ((ip >> 24) == 127 || !is_onmysubnet(ip) || is_broadcast(ip))
    {
      ignmsg = "Illegal";
      cache_refresh_mode = 0;
/* _printk("*%s ARP_REQUEST from %I/%M: (Q=%I)\n", ignmsg,
                  *src_ip_addr_ptr, src_hw_addr_ptr, *dst_ip_addr_ptr );
*/
    }
    else if (is_local_addr(ip))
    {
      /* someone asking for our IP address prolly wants to talk to us */
      /* if not already in cache, create a temporary (fast-expire) entry. */
      cache_refresh_mode = 2; /* active refresh */
      ignmsg = "Handled";

      _arp_rarp_send( ARP_REPLY, ip/*Q*/,
                      src_hw_addr_ptr,NTOHL(*src_ip_addr_ptr));
    }

    if (my_statics.debug_lev >= 1)
      TRACE_NETIO(("*%s ARP_REQUEST from %I/%M: (Q=%I)\n", ignmsg,
                  *src_ip_addr_ptr, src_hw_addr_ptr, *dst_ip_addr_ptr ));
  }
  else if (ah->opcode == RARP_REQUEST)
  {
    if (my_statics.debug_lev >= 1)
      TRACE_NETIO(("*Ignored RARP_REQUEST from %M/??? (Q=%M)\n", 
                            src_hw_addr, dst_hw_addr_ptr ));
  }
  else if (ah->opcode == RARP_REPLY)
  {
    /* did we issue a rarp request and are waiting for a response? */
    if (resolve_block &&
        resolve_block->ethtype == ETHERTYPE_REVARP &&
        memcmp(resolve_block->eth_ptr,dst_hw_addr_ptr,ah->hwAddrLen)==0)
    {
      *(resolve_block->ip_ptr) = NTOHL(*dst_ip_addr_ptr);
      resolve_block->ethtype = 0;
      ignmsg = "Handled";
    }
    if (my_statics.debug_lev >= 1)
      TRACE_NETIO(("*%s RARP_REPLY from %M/%I (Q=%M, A=%I)\n", ignmsg,
                  src_hw_addr_ptr, *src_ip_addr_ptr,
                  dst_hw_addr_ptr, *dst_ip_addr_ptr ));
  }
  else if (ah->opcode == ARP_REPLY)
  {
    /* continuously update the arp tables for any ip addresses in there */
    cache_refresh_mode = 1; /* passive refresh */

    ip = NTOHL(*src_ip_addr_ptr);
    if ((ip >> 24) == 127 || !is_onmysubnet(ip) || is_broadcast(ip))
    {
      ignmsg = "Illegal";
      cache_refresh_mode = 0;
/* _printk("*%s ARP_REPLY (Q=%I,A=%M)\n", ignmsg,
              *src_ip_addr_ptr, src_hw_addr_ptr );
*/
    }
    else if (resolve_block && resolve_block->ethtype == ETHERTYPE_ARP &&
                 *(resolve_block->ip_ptr) == ip )
    {
      memcpy(resolve_block->eth_ptr,src_hw_addr_ptr,ah->hwAddrLen);
      resolve_block->ethtype = 0;
      ignmsg = "Handled";
      cache_refresh_mode = 3; /* active refresh full timeout */
    }
    if (my_statics.debug_lev >= 1)
      TRACE_NETIO(("*%s ARP_REPLY (Q=%I,A=%M)\n", ignmsg,
              *src_ip_addr_ptr, src_hw_addr_ptr ));
  }
  else 
  {
    if (my_statics.debug_lev >= 1)
      TRACE_NETIO(("*Ignored unhandled ARP/RARP opcode %04x\n", ah->opcode ));
  }


  if (cache_refresh_mode > 0)
  {
    struct arp_table FAR *arp;
    unsigned int i; 

    ip = NTOHL(*src_ip_addr_ptr);

    /* we cannot use arp_search because ...
    *  a) an entry may have already expired 
    *  b) we want to be able to add gateway entries.
    *  c) we don't want to influence the cache efficacy counters
    *     since we are searching for *every* ARP_REPLY that comes by.
    */
    for (i = 0; i < (sizeof(arp_cache_data)/sizeof(arp_cache_data[0])); i++)
    {
      arp = &arp_cache_data[i];
      if (ip == arp->ip)
      {
        my_statics.searched_count++;
        if (arp == &(arp_cache_data[0]) || arp == &(arp_cache_data[1]))
          my_statics.hit1_count++;
        else
          my_statics.hit2_count++;
        if (my_statics.debug_lev >= 2)
          TRACE_CACHE(("*ARP refreshed cache entry %d (%I->%M)\n",
                        i, *src_ip_addr_ptr, src_hw_addr_ptr ));
        FAR_MEMCPY(arp->hardware, src_hw_addr_ptr, ah->hwAddrLen);
        if (arp->flags != ARP_FLAG_FIXED)
        {
          arp->flags = ARP_FLAG_FOUND;
          arp->expiry = timer_sec_count() + ARP_CACHE_FOUND_TIMEOUT;
        }
        __speedup_cache(arp);
        cache_refresh_mode = 0;
        break;
      }
    }
    if (cache_refresh_mode > 0)
    {
      for (i = 0; i < my_statics.arp_num_gateways; i++)
      {
        if (ip == arp_gate_list[i].gate_ip)
        {
          cache_refresh_mode = 3;
          break;
        }
      }
    }
    if (cache_refresh_mode > 1)
    {
      blocker_critsect_begin();
      arp = _arp_search(ip,1);
      arp->ip = ip;
      FAR_MEMCPY(arp->hardware, src_hw_addr_ptr, ah->hwAddrLen);

      if (cache_refresh_mode > 2)  
      {
        arp->flags = ARP_FLAG_FOUND;
        arp->expiry = timer_sec_count() + ARP_CACHE_FOUND_TIMEOUT;
      }
      else
      {
        arp->flags = ARP_FLAG_TEMP;
        arp->expiry = timer_sec_count() + ARP_CACHE_TEMP_TIMEOUT;
      }
      blocker_critsect_end();
    }
  }

  is_brdcast = is_brdcast; /* unreferenced */
  src_hw_addr = src_hw_addr; /* possibly unreferenced */
  ignmsg = ignmsg; /* possibly unreferenced */

  return;
}

/* ======================================================================= */
/*                          Reverse-ARP resolver                           */
/* ======================================================================= */

int _rarp_resolve(const void FAR *_ethap, uint32_t FAR *_ip)
{
  char ethap[MAX_LINKADDR_SIZE];
  struct __resolve_block rrb;
  uint32_t  ip;

  if (blocker_get_inisr_flag())
  {
    /* We cannot do a network ARP/RARP resolve while in an ISR because
    * we cannot blocker_yield() while waiting for the response.
    *
    * A RARP lookup _could_ theoretically happen while in an ISR since
    * IP address determination is called periodically until it succeeds.
    * However, that timer callback in the DHCP stuff should be ensuring 
    * that such a condition does not happen.
    *
    * A network-based ARP resolve should NEVER happen because only 
    * passive/background activity is handled from whithin an ISR.
    * Since it is passive, the ARP cache should already have an entry
    * (added when the host about to connect to us got our ethaddr).
    * The need do a network ARP resolve can only happen when initiating
    * a connection for the first time, which is an ACTIVE operation.
    */
    __nonfatal_runtime_error("PANIC! network RARP resolve call while in ISR!\n");
    return 0;
  }

  if (!_ip_is_up())
    return (0);

  FAR_MEMCPY( ethap, _ethap, _eth_get_hwaddrlen());
  ip                 = 0ul;
  rrb.ethtype        = ETHERTYPE_REVARP;
  rrb.eth_ptr        = &ethap[0];
  rrb.ip_ptr         = &ip; /* result will be stored directly in 'ip' */
  resolve_block = &rrb;

  if (_rarp_request(ethap)) /* send request */
  {
    uint32_t timeout   = timer_sec_count() + ARP_RESOLVE_TIMEOUT;
    int oldbrkstate = set_cbreakstate(0);

    /* wait for reply (ip will get filled) */
    while (rrb.ethtype == ETHERTYPE_REVARP)
    {
      if (blocker_yield(YIELDFLAG_HARD|YIELDFLAG_CHKBRK))
        break; /* ^C'd */
      if (timer_sec_count() > timeout)
        break; /* timed out */
    }
    set_cbreakstate(oldbrkstate);
  }

  resolve_block = NULL;

  if (ip)
    *_ip = ip;
  return (ip != 0);
}

/* ====================================================================== */
/*                             redirects                                  */
/* ====================================================================== */

int _arp_record_redirect(uint32_t ina, uint32_t gate_ina, uint32_t sender_ina)
{
  int rc = -1; 
  if (!is_onmysubnet(ina)  /* otherwise its not going through a gateway */
   && is_onmysubnet(gate_ina)) /* can't reach it otherwise */
  {
    unsigned int i;
    /* sender_ina must be a gateway we already know about. Check it */
    for (i = 0; i < my_statics.arp_num_gateways; i++)
    {
      if (arp_gate_list[i].gate_ip == sender_ina)
      {
        unsigned int n = (sizeof(redir_list)/sizeof(redir_list[0]))-1;
        for (i = 0; i < (sizeof(redir_list)/sizeof(redir_list[0])); i++)
        {
          if (redir_list[i].dest_ina == 0)
            n = i;
          else if (redir_list[i].dest_ina == ina)
          {
            n = i;
            break;
          }  
        }
        redir_list[n].gate_ina = gate_ina;
        redir_list[n].dest_ina = ina;
        /* add as gateway in case it isn't there already */
        _arp_add_gateway( gate_ina, 0, 0 );
        rc = 0;
        break;
      }
    }
  }
  return rc;
}

/* --------------------------------------------------------------- */

static uint32_t __redirect_if_needed(uint32_t ina)
{
  unsigned int i;
  for (i = 0; i < (sizeof(redir_list)/sizeof(redir_list[0])); i++)
  {
    if (redir_list[i].dest_ina == ina)
    {
      ina = redir_list[i].gate_ina;
      if (i != 0)
      {
        struct redirrec_t tmp;
        FAR_MEMCPY( &tmp, &redir_list[i], sizeof(tmp) );
        FAR_MEMMOVE( &redir_list[1],&redir_list[0], 
                 (size_t)(&redir_list[i]-&redir_list[0]) );
        FAR_MEMCPY( &redir_list[0], &tmp, sizeof(tmp));
      }
      break;
    }
  }
  return ina;
}

/* ======================================================================= */
/*                               ARP resolver                              */
/* ======================================================================= */

/*
 * Network-only portion of resolver.
*/
static int __arp_net_resolve(uint32_t ip, void *ethap)
{
  struct __resolve_block rrb;

  if (blocker_get_inisr_flag())
  {
    /* We cannot do a network ARP/RARP resolve while in an ISR because
    * we cannot blocker_yield() while waiting for the response.
    *
    * A RARP lookup _could_ theoretically happen while in an ISR since
    * IP address determination is called periodically until it succeeds.
    * However, that timer callback in the DHCP stuff should be ensuring 
    * that such a condition does not happen.
    *
    * A network-based ARP resolve should NEVER happen because only 
    * passive/background activity is handled from whithin an ISR.
    * Since it is passive, the ARP cache should already have an entry
    * (added when the host about to connect to us got our ethaddr).
    * The need do a network ARP resolve can only happen when initiating
    * a connection for the first time, which is an ACTIVE operation.
    */
    __nonfatal_runtime_error("PANIC! network ARP resolve call while in ISR!\n");
    return 0;
  }
  
  rrb.ethtype   = ETHERTYPE_ARP;
  rrb.eth_ptr   = ethap; /* result will be stored directly in 'ethap' */
  rrb.ip_ptr    = &ip;
  resolve_block = &rrb;

  if (_arp_request(ip)) /* send request */
  {
    uint32_t timeout   = timer_sec_count() + ARP_RESOLVE_TIMEOUT;
    int oldbrkstate = set_cbreakstate(0);
      
    /* wait for reply ('ethap' will get filled) */
    while (rrb.ethtype == ETHERTYPE_ARP)
    {
      if (blocker_yield(YIELDFLAG_HARD|YIELDFLAG_CHKBRK))
        break; /* ^C'd */
      if (timer_sec_count() > timeout) 
        break; /* timed out */
    }
    set_cbreakstate(oldbrkstate);
  }

  resolve_block = NULL;

  if (rrb.ethtype == 0) /* success! */
  {
    if (my_statics.debug_lev >= 2)
      TRACE_RESOLVE(("_arp_resolve(%hI) net-resolved to %M\n", ip, ethap));
    return 1;
  }

  if (my_statics.debug_lev >= 1)
    TRACE_RESOLVE(("*_arp_resolve(%hI): ETIMEDOUT\n", ip));
  return 0;
}

/* -------------------------------------------------------------------- */

/*
** _arp_vrfy_addr_change() is called 
** a) by address discovery to ensure that an IP address is not already in use
**    'ina' will then be non-zero. 
** b) by address lock-down to notify arp that a new address is now in use.
**    'ina' will be zero - the addr may be retrieved with _ip_get_ipaddr()
*/
int _arp_vrfy_addr_change(uint32_t ina)
{
  if (_eth_get_hwaddrlen() == 0)
  {
    ; /* nothing to do */
  }
  else if (ina != 0) /* verify that 'ina' is not in use elsewhere */
  {
#if 0
    char ethap[MAX_LINKADDR_SIZE];
    if (__arp_net_resolve( ina, ethap ) != 0)
    {
      _arp_del_entry( ina );
      return -1; /* someone responded to 'ina' */
    }
#endif
  }
#if 0
  else if ((ina = _ip_get_ipaddr()) != 0)
  {
    /* if our IP address has changed (eg DHCP) then localcast it.
     * (may result in a ping to verify) 
    */
    static uint32_t my_last_ip_addr = 0ul; /* a copy of my_ip_addr */

    if (ina != my_last_ip_addr)
    {
      my_last_ip_addr = ina;
      if (_eth_get_hwaddrlen() != 0)
      {    
        _arp_rarp_send( ARP_REPLY, ina, /* "answer" */
                        _eth_get_hwaddrbc(), /* hardware addr of questioner */
                        _ip_get_bcaddr() ); /* IP address of questioner */
      }
    }    

    /* discard any gateways in the gateway table that are not on our subnet */
    if (my_statics.arp_num_gateways > 0 && _ip_get_snmask() != ((uint32_t)-1))
    {
      unsigned int i;
      for (i = 0; i < my_statics.arp_num_gateways; i++)
      {
        if (!is_onmysubnet(arp_gate_list[i].gate_ip))
        {
          memmove( &arp_gate_list[i], &arp_gate_list[i+1],
                   sizeof(arp_gate_list) - (sizeof(arp_gate_list[0])*(i+1)) );
          my_statics.arp_num_gateways--;
        }
      }
      /* if no gateways are defined, ask icmp to get a list */
      if (my_statics.arp_num_gateways == 0)
      {
        int rc;
        if (my_statics.debug_lev >= 1)
          TRACE_RESOLVE(("*_arp_resolve(): asking routers to ident themselves...\n"));
        rc = icmp_router_solic();
        if (my_statics.debug_lev >= 1)
          TRACE_RESOLVE(("*_arp_resolve(): icmp_router_solic() => %rc\n", rc));
      }
    }
  }
#endif
 
  return 0;  
}

/* -------------------------------------------------------------------- */

/*
 * _arp_resolve - resolve IP address to hardware address
*/
int _arp_resolve(uint32_t ip, void FAR *_ethap)
{
  char ethap[MAX_LINKADDR_SIZE];
  unsigned long valid_routes;
  unsigned int i;

  if (!_ip_is_up())
    return (0);

  if (my_statics.debug_lev >= 1)
    TRACE_RESOLVE(("*ARP: resolve req for %hI\n", ip ));

  ip = __redirect_if_needed(ip);

  if (!ip)       /* return if no host */
  {
    if (my_statics.debug_lev >= 1)
      TRACE_RESOLVE(("*_arp_resolve(0.0.0.0): EINVAL\n"));
    return (0);
  }

  /* First see if 'ip' is one of the "synthetic" entries 
   * such as 127.0.0.1 or 255.255.255.255 or mcast addresses.
   *
   * If 'ip' is not a synthetic entry but is on the local subnet
   * and in the cache it will be found too.
  */
  if (_arp_find_entry(ip, ethap, NULL))
  {
    if (my_statics.debug_lev >= 1)
      TRACE_RESOLVE(("*ARP: resolved (in cache) %hI->%M\n", ip, ethap ));
    if (_ethap)
      FAR_MEMCPY(_ethap, ethap, _eth_get_hwaddrlen());
    return 1;
  }

  /* if 'ip' is on our local subnet, we now need  
   * to do a network lookup.
  */
  if (!_ip_get_snmask() || is_onmysubnet(ip)) /* is of this network */
  {
    if (__arp_net_resolve( ip, ethap ))
    {
      if (_ethap)
        FAR_MEMCPY(_ethap, ethap, _eth_get_hwaddrlen());
      return 1;
    }
    return 0; /* appropriate message will have been printed */
  }

  /* 'ip' is NOT on our local subnet - traffic must go through a gateway,
   * so return the etheraddr of the gateway for that route.
  */
  /* if no gateways are defined, ask icmp to get a list */
  if (my_statics.arp_num_gateways == 0)
  {
    if (my_statics.debug_lev >= 1)
      TRACE_RESOLVE(("*_arp_resolve(): asking routers to ident themselves...\n"));
    i = icmp_router_solic();
    if (my_statics.debug_lev >= 1)
      TRACE_RESOLVE(("*_arp_resolve(): icmp_router_solic() => %d\n", i));
  }

  /* determine which gateways are a valid route to 'ip
  */
  valid_routes = 0UL;
  for (i = 0; i < my_statics.arp_num_gateways; i++)
  {
    const struct gate_table FAR *gw = &arp_gate_list[i];
    /*  
     * is the gateway appropriate for 'ip'? (We assume it is if the
     *  mask is FF..FF because we are then on 'no' net). 
    */
    if ( _ip_get_snmask() == ((uint32_t)-1) || 
        ((gw->mask & ip) == gw->subnet && is_onmysubnet(gw->gate_ip)))
    {
      valid_routes |= (1UL << i);
    }
    if (valid_routes == 0UL)
    {
      if (my_statics.debug_lev >= 1)
        TRACE_RESOLVE(("*_arp_resolve(%hI): ENOROUTE\n",ip));
      return 0;
    }
  }

  /* look in the cache to find the etheraddr of 
   * each gateway which is a valid route to 'ip'
  */
  for (i = 0; i < my_statics.arp_num_gateways; i++)
  {
    if (valid_routes & (1UL << i))
    {
      if (_arp_find_entry(arp_gate_list[i].gate_ip, ethap, NULL))
      {
        if (my_statics.debug_lev >= 1)
          TRACE_RESOLVE(("*ARP: resolved (in cache) %hI->%M (GW %hI)\n", 
                      ip, ethap, arp_gate_list[i].gate_ip ));
        if (_ethap)
          FAR_MEMCPY(_ethap, ethap, _eth_get_hwaddrlen());
        return 1;
      }
    }
  }

  /* Nothing found in the cache. 
   * Must do a network resolve.
  */
  for (i = 0; i < my_statics.arp_num_gateways; i++)
  {
    if (valid_routes & (1UL << i))
    {
      if (__arp_net_resolve( arp_gate_list[i].gate_ip, ethap ))
      {
        if (_ethap)
          FAR_MEMCPY(_ethap, ethap, _eth_get_hwaddrlen());
        return 1; /* message already printed */
      }
    }
  }

  return (0);
}

/* ======================================================================= */
/*                          transport initialization                       */
/* ======================================================================= */

static void set_gateway (const char *value) /* "gateip[,subnet[,mask]]" */
{
  if (value)
  {
    unsigned int i;
    uint32_t parts[3]; /* gateway, subnet, mask */
    parts[0] = parts[1] = parts[2] = 0L;

    for (i = 0; i < 3; i++)
    {
      unsigned int n = 0;
      char buf[sizeof("255.255.255.255   ")];

      while (*value == ' ' || *value == '\t')
        value++;
      if (*value == ',')
      {
        if (i == 0) 
          return; /* leading comma is illegal for ip */
        value++;
        while (*value == ' ' || *value == '\t')
          value++;
      }
      while ((*value >= '0' && *value <= '9') || *value == '.')
      {
        if (n >= (sizeof(buf)-1))
          return; /* syntax error */
        buf[n++] = *value++;
      }
      if (*value && *value != ',' && *value != ' ' && *value != ' ')
        return; /* trailing garbage */
      if (n == 0) /* no number there */
        break; 
      buf[n] = '\0';
      if (!_my_aton(buf,&parts[i]))
        return; /* syntax error */
    }   

    if (parts[1] /* subnet */ && !parts[2] /* mask */)
    {
      switch (parts[1] >> 30)
      {
        case 0:
        case 1:  parts[2] = 0xFF000000L;
                 break;
        case 2:  parts[2] = 0xFFFFFE00L;   /* minimal class B */
                 break;
        default: parts[2] = 0xFFFFFF00L;
                 break;
      }
    }
    _arp_add_gateway( parts[0], parts[1], parts[2] );
  }
  return;
}

static void __ctor_arp_rarp(void)
{
  /* 
  ** Whoever added the ETHIP option ought to be shot. Addresses cannot
  ** be added until eth has been initialized, and it can't be initialized
  ** until config is loaded, and setting the addr is called during config load.
  ** Just one more example of how dumb trying to make everything 
  ** configurable can be.
  */ 
  static const struct config_table FAR cfg[] = {
     { "ARP.DEBUG",     ARG_ATOI,     (void FAR *)&(my_statics.debug_lev)   },
     { "GATEWAY",       ARG_FUNC,     (void FAR *)set_gateway               },
     { ((const char FAR *)0), 0, ((void FAR *)0)  }
  };

  FAR_MEMSET(&my_statics, 0, sizeof(my_statics));
  _conf_add_confextn( &cfg[0], NULL );

  /* register "downwards" */
  _eth_register_pkthandler( ETHERTYPE_ARP, _arp_rarp_handler );  
  _eth_register_pkthandler( ETHERTYPE_REVARP, _arp_rarp_handler );  

  return;
}

#define __SETUP_CONSTRUCTOR_FOR__ __ctor_arp_rarp
#include "ctordtor.h"
