/*
 * Yield, interrupt hook, and child-hook dispatch.
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: blocker.c,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
/*
 * yield, including interrupt hook, and child-hook dispatch.
 *
 * Created May 2001 by Cyrus Patel <cyp@fb14.uni-mainz.de>
 *
 * ----------------------------- yielding -----------------------------
 *
 * blocker_cyield()     - (optionally) run idle/periodic callbacks and 
 *                      (optionally) "hard yield" and (optionally) do ^C chks.
 * blocker_cdelay()     - number of millisecs to delay (may be zero).
 *                      Returns non-zero if interrupted (EINTR).
 *                      Runs blocker_cyield() [with a hard yield] while waiting.
 *
 * ----------------------------- callbacks ----------------------------
 *
 * Callbacks come in three categories: 
 *  a) Interrupt time hooks: called directly when processing
 *     the idle hook ISRs. Must live within something like 1K stack and
 *     stack checking must be off for all but watcom.
 *     Interrupts are *enabled* when a hook is called, so the function
 *     being called must be reentrant, or have reentrancy protection.
 *     Is only called if !INDOS, and is thus allowed to make DOS calls.
 *  b) Idle hooks: These are functions that will be called repeatedly
 *     from within blocker_cyield() [blocker_yield() is the tcp_tick() replacement]
 *     An example of an idle hook is _eth_poll().
 *  c) Soft timer tick hooks: these are like idle hooks but are called
 *     at most once per timer tick.
 *     Soft timer tick hooks include the TCP retransmitter, IP frag timeout
 *     checker, IGMP "report due" handler, DHCP renew etc.
 *
 * blocker_tisrhook_{add|del}(void (*fxn)()) - timer ISR callback
 * blocker_idlehook_{add|del}(void (*fxn)()) - idle time callback
 * blocker_perdhook_{add|del}(void (*fxn)()) - periodic callback
 *  
*/
#undef WANT_STACK_CHECKING /* disable stack checking for this module */

#include "target.h"   /* dosish stuff */
#include "debug.h"    /* KMSG() */
#include "sigblk.h"   /* get_cbreakpend() */
#include "dpmi.h"     /* _my_dpmi_*() */
#include "ctordtor.h" /* _my_atterm() */
#include "timer.h"    /* timr_*() */
#include "printk.h"   /* _printk() [interrupt safe debugging] */
#include "blocker.h"  /* ourselves */

/* separate debug clause since almost everything is isr-type debugging */
#define DEBUG_VERBOSITY 0

#if (DEBUG_VERBOSITY >= 1)
  #define TRACE_LEV1(x) _printk x
#else
  #define TRACE_LEV1(x) NO_OPERATION()
#endif
#if (DEBUG_VERBOSITY >= 2)
  #define TRACE_LEV2(x) _printk x
#else
  #define TRACE_LEV2(x) NO_OPERATION()
#endif
#if (DEBUG_VERBOSITY >= 3)
  #define TRACE_LEV3(x) _printk x
#else
  #define TRACE_LEV3(x) NO_OPERATION()
#endif
#if (DEBUG_VERBOSITY >= 4)
  #define TRACE_LEV4(x) _printk x
#else
  #define TRACE_LEV4(x) NO_OPERATION()
#endif
#if (DEBUG_VERBOSITY >= 5)
  #define TRACE_LEV5(x) _printk x
#else
  #define TRACE_LEV5(x) NO_OPERATION()
#endif

/* ===================================================================== */
/*                        static (locked) data                           */
/* ===================================================================== */

static struct
{
  /* the following 3 ints must come first (must be int aligned) */
  volatile int  in_crit_sect; /* increment/decrement */
  volatile int  inside_isr; /* increment on isr entry, decrement on isr exit */
  volatile int  isr_hook_state; /*0=not initialized, 1=doneinit, -1=shutdown*/

  struct 
  {
    void (*vector)(void);
    unsigned char type_bits;
  } hooks[32];
  volatile unsigned long hwork_marker;

  unsigned long calloutcounts[7]; /*[0]=idle 1=1Ch 2=28h 3=31h 4=21h 5=10h 6=2Fh*/
  unsigned long last_idlecall_mark;
  unsigned long yield_time, yield_count;
  volatile int in_blocker_yield;

#if defined(__DPMI32__)
  int int31hpmhook;
  int int21hpmhook;
  unsigned long cmp_count;
  unsigned long indos_flag_ptr;
#else
  volatile unsigned char _far *indos_flag_ptr;
#endif
} FAR my_statics;

/* ===================================================================== */
/*                        initialization                                 */
/* ===================================================================== */

static void __ctor_yield(void)
{
  static int my_bss_count = -1;
  if (my_bss_count < __bss_count) /* always LT compare! */
  {
    FAR_MEMSET( &my_statics, 0, sizeof(my_statics));
    my_bss_count = __bss_count;

    #if defined(__DPMI32__) /* includes pharlap */
    {
      __dpmi_regs regs;
      memset( &regs, 0, sizeof(regs) );
      regs.x.ax = 0x3400;
      __dpmi_int( 0x21, &regs );
      my_statics.indos_flag_ptr = (((unsigned long)regs.x.es)<<4);
      my_statics.indos_flag_ptr += regs.x.bx;

      #if defined(__GNUC__)
        #define READINDOSFLAG(__resptr) \
                 *(__resptr) = _farpeekb(_dos_ds, my_statics.indos_flag_ptr )
      #elif defined(__WATCOMC__) || defined(__BORLANDC__)
        #define READINDOSFLAG(__resptr) \
          if (_my_ds() == _dos_ds) \
            *(__resptr)=*((volatile unsigned char *)(my_statics.indos_flag_ptr)); \
        else \
          dosmemget( my_statics.indos_flag_ptr, 1, __resptr )
      #else /* dosmemget() is assumed to be inline */
        #define READINDOSFLAG(__resptr) \
          dosmemget( my_statics.indos_flag_ptr, 1, __resptr );
      #endif
    }
    #else
    {
      unsigned int indos_o = 0;
      unsigned int indos_s = 0;
      __asm push es
      __asm push bx
      __asm mov ah,34h
      __asm int 21h
      __asm mov indos_o,bx
      __asm mov indos_s,es
      __asm pop bx
      __asm pop es
      my_statics.indos_flag_ptr = 
           (volatile unsigned char _far *)MK_FP(indos_s, indos_o);  
      #define READINDOSFLAG(__resptr) \
              *(__resptr) = *my_statics.indos_flag_ptr
    }
    #endif
  }
  return;
}

#define __init_statics_if_needed  __ctor_yield
#define __SETUP_CONSTRUCTOR_FOR__ __ctor_yield
#include "ctordtor.h"

/* ===================================================================== */
/*                     helper functions (partially public)               */
/* ===================================================================== */

void blocker_critsect_begin(void)
{
  ++my_statics.in_crit_sect;
}

extern void blocker_critsect_end(void)
{
  --my_statics.in_crit_sect;
}

/* --------------------------------------------------------------------- */

int blocker_get_inisr_flag(void) /* this function is NOT locked! */
{                              /* (do not call unless interrupts enabled) */
  return (my_statics.inside_isr != 0);  
}

/* --------------------------------------------------------------------- */

int blocker_get_indos_flag(void) /* this function is NOT locked! */
{                              /* (do not call unless interrupts enabled) */
  unsigned char indos_flag;
  __init_statics_if_needed();
  READINDOSFLAG(&indos_flag);
  return (int)(((unsigned int)(indos_flag)) & 0xffu);
}

/* ===================================================================== */
/*                        hook add/remove/call                           */
/* ===================================================================== */

/* blocker_hook_op() adds/removes functions from the hook list(s). 
 * 'hooktype' is a mask of bits (BLOCKER_HOOKTYPE_*), mode is either 
 * '+' to add, or '-' to remove. Returns -1 on error, 0 on success 
*/
int blocker_hook_op(int hooktypes, void (*fxn)(void), int mode)
{
  int rc = -1;
  __init_statics_if_needed();  
  if (fxn)     
  {
    unsigned int i;
    int empty = -1;
     
    my_statics.in_crit_sect++; /* enter critical section */

    for (i = 0; i < (sizeof(my_statics.hooks)/sizeof(my_statics.hooks[0])); i++)
    {
      if (my_statics.hooks[i].vector == fxn)
      {
        if (mode == '+')
          my_statics.hooks[i].type_bits |= (unsigned char)hooktypes;
        else
        {
          my_statics.hooks[i].type_bits &= ~((unsigned char)hooktypes);
          if (!my_statics.hooks[i].type_bits)
            my_statics.hooks[i].vector = 0;
        }
        rc = 0; /* success */
        break;
      }
      if (empty == -1 && !my_statics.hooks[i].vector)
        empty = i;
    }
    if (rc == -1 && empty != -1 && mode == '+')
    {
      my_statics.hooks[empty].vector = fxn;
      my_statics.hooks[empty].type_bits = (unsigned char)hooktypes;
      rc = 0; /* success */
    }
    my_statics.in_crit_sect--; /* leave critical section */
  }
  return rc;
}

/* --------------------------------------------------------------------- */

/* called from blocker_yield() and the ISR handlers (interrupts enabled) */
static void __call_hooks(int hooktypes)
{
  static fastlock_t FAR in_hook = FASTLOCK_INITIALIZER_UNLOCKED;
  static int FAR add_hooks = 0;

  if (fastlock_trylock(&in_hook))
  {
    /* DBGPRINTF(("call (hooktypes=0x%04x)\n", hooktypes )); */
    do
    {
      static unsigned long prev_periodic = 0;
      unsigned long now;

      now = timer_tick_count();

      hooktypes &= ~BLOCKER_HOOKTYPE_PERIODIC;
      if (prev_periodic != now)
      {
        prev_periodic = now;
        hooktypes |= BLOCKER_HOOKTYPE_PERIODIC;
      }

      if (hooktypes)
      {
        unsigned int i;

        if (hooktypes & BLOCKER_HOOKTYPE_IDLE)
        {
          my_statics.last_idlecall_mark = now;
          my_statics.calloutcounts[0]++;
        }
   
        for (i = 0; i < (sizeof(my_statics.hooks)/sizeof(my_statics.hooks[0])); i++)
        {
          if ((my_statics.hooks[i].type_bits & hooktypes)!=0)
          {
            if (my_statics.hooks[i].vector)
              (*(my_statics.hooks[i].vector))();
          }
        }
      }

      _disable();
      hooktypes = add_hooks;
      add_hooks = 0;
      _enable();

    } while (hooktypes);

    fastlock_unlock(&in_hook);
  } 
  else
  {
    _disable();
    add_hooks |= hooktypes;
    _enable();
  }
  return;
}

/* ==================================================================== */
/*                     common interrupt processor                       */
/* ==================================================================== */

static void __isr_common(int int_num)
{
  register int dos_is_idle = 0;
  register int counter_idx = -1; /* [0]=idle 1=1Ch 2=28h 3=31h 4=21h */

  /* --------- WARNING: do not use switch() here ------------- */

  if (int_num == 0x28)
  {
    dos_is_idle = 1;
    counter_idx = 2;/* int28h index */ 
  }
  else if (int_num == 0x31)
    counter_idx = 3; /* int31h index */ 
  else if (int_num == 0x21)
    counter_idx = 4; /* int21h index */ 
  else if (int_num == 0x10)
    counter_idx = 5; /* int10h index */ 
  else if (int_num == 0x2F)
    counter_idx = 6; /* int2Fh index */ 

  #define SHOW_CALLOUT_COUNTS() \
     TRACE_LEV1(("*Callout counts: idle = %lu  0x1C = %lu  0x28 = %lu  "\
                   "0x31 = %lu  0x21 = %lu  0x10 = %lu  0x2F = %lu\n",   \
              my_statics.calloutcounts[0], my_statics.calloutcounts[1],  \
              my_statics.calloutcounts[2], my_statics.calloutcounts[3],  \
              my_statics.calloutcounts[4], my_statics.calloutcounts[5],  \
              my_statics.calloutcounts[6] ))

  if (counter_idx >= 0 && 
      counter_idx < ((int)(sizeof(my_statics.calloutcounts)/
                         sizeof(my_statics.calloutcounts[0]))) &&
      my_statics.in_crit_sect == 0)
  {
    if ((++my_statics.inside_isr)==1)
    {
      unsigned char indos_flag;
      READINDOSFLAG(&indos_flag);
                                             
      /*_printk("%c", (dos_is_idle ? 'a' : '0') + indos_flag );*/

      if (indos_flag == 0 || (indos_flag == 1 && dos_is_idle))
      { 
        my_statics.calloutcounts[counter_idx]++;

        /* *** --------------------------------------------------- *** */
        /* *** no references to stack variables beyond this point! *** */
        /* *** --------------------------------------------------- *** */
        {
          #if defined(__WATCOMC__) && !defined(__386__)
          static unsigned char locstack[2048]; /* must be near heap */
          /* special case since stack-checking must be handled too */
          stackset(locstack,sizeof(locstack));
          #elif !defined(__DPMI32__)
          static unsigned char locstack[2048]; /* must be near heap */
          unsigned int newsp = FP_OFF(&locstack[sizeof(locstack)-2]);
          __asm  mov  bx,ss
          __asm  mov  cx,sp
          __asm  mov  ax,newsp
          __asm  mov  dx,ds
          __asm  mov  ss,dx
          __asm  mov  sp,ax
          __asm  push bx
          __asm  push cx
          #endif /* !defined(__DPMI32__) */

          _enable();
          TRACE_LEV3(("%c",'A'));
          __call_hooks(BLOCKER_HOOKTYPE_ISR);
          TRACE_LEV3(("%c",'Z'));
          _disable();

          #if defined(__WATCOMC__) && !defined(__386__)
          stackrestore();
          #elif !defined(__DPMI32__)
          __asm  pop cx
          __asm  pop bx
          __asm  mov ss,bx
          __asm  mov sp,cx
          #endif
        }
      }
    }
    --my_statics.inside_isr;
  } /* if (my_statics.in_crit_sect == 0) */

  return;
}

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) /* all DPMI targets including pharlap */

  #undef DONT_WANT_INT28H_ISR
  #undef DONT_WANT_INT31H_ISR
  #undef DONT_WANT_INT21H_ISR

  #if 1
  #define DONT_WANT_INT28H_ISR
  //#define DONT_WANT_INT31H_ISR
  //#define DONT_WANT_INT21H_ISR
  #endif

  #if !defined(DONT_WANT_INT28H_ISR)
  static void __int28h_isr (void) /* this is (must be) always an RMCB */
  {
    __isr_common(0x28);
    /* rmcb chains on return */
  }
  #endif

  /* ---------------------------------------------------------------- */

  #if !defined(DONT_WANT_INT31H_ISR)
  static void __int31h_isr (__dpmi_regs *regs) /* always a pmode int handler */
  {
    /* _printk("Int31h, ax=%04x, bl=%02x\n", regs->x.ax, regs->h.bl ); */
    if (regs->x.ax == 0x300 && 
        regs->x.es == _my_ds() && regs->x.cs == _my_cs())
    {
      unsigned char indosflag;
      READINDOSFLAG(&indosflag);
      if (!indosflag)
      {
        int do_idle = 0;
        __dpmi_regs r;
        movedata( regs->x.es, regs->d.edi, _my_ds(), (unsigned)&r, sizeof(r));

        TRACE_LEV4((
            "\nInt=%02x, ax=%04x bx=%04x cx=%04x dx=%04x ds=%04x es=%04x\n",
             regs->h.bl, r.x.ax, r.x.bx, r.x.cx, r.x.dx, r.x.es, r.x.ds ));

        /* --------- WARNING: do not use switch() here ------------- */
        if (regs->h.bl == 0x15)
        {
          if (r.h.ah == 0x86)
          {
            /* compensate for DJGPP's brain-dead delay() */
            _enable();
            blocker_delay( ((((unsigned int)(r.x.cx))<<16)|(r.x.dx))/1000u );
            _disable();
            return; /* don't chain */
          }
        }
        else if (regs->h.bl == 0x21)
        {
          unsigned char ah = r.h.ah;
          if ( (ah >= 0x01 && ah <= 0x24) ||
               (ah >= 0x27 && ah <= 0x30) ||
               (ah >= 0x37 && ah <= 0x47) ||
               (ah >= 0x4B && ah <= 0x51) )
          {
            __isr_common(0x21);
          }
          if (ah == 0x4B) /* exec */
          {
            r.x.ax = 8; /* ENOMEM */
            r.x.flags |= 1;
            return; 
          }
        }
        else if (regs->h.bl == 0x2F && r.x.ax == 0x1680)
        {
          if (!my_statics.in_blocker_yield)
            do_idle = 0x2f;
        }
        else if (regs->h.bl == 0x10 ||
                 regs->h.bl == 0x11 || 
                 regs->h.bl == 0x16 ||
                 regs->h.bl == 0x1A)
        {
          do_idle = 0x10;
        }
        else if (regs->h.bl == 0x28)
        {
          do_idle = 0x31;
        }
        if (do_idle)
        {
          __isr_common(do_idle);
        }
      }
    }
    __my_dpmi_chain_pmiv(regs);
    return;
  }
  #endif /* #if !defined(DONT_WANT_INT31H_ISR) */

  /* ---------------------------------------------------------------- */

  #if !defined(DONT_WANT_INT21H_ISR)
  static void __int21h_isr (__dpmi_regs *regs) /* always a pmode int handler */
  {
    if (regs->x.cs == _my_cs())
    {
      /* the indosflag check shouldn't be necessary, but doesn't hurt */
      unsigned char indosflag;
      READINDOSFLAG(&indosflag);
      if (!indosflag)
      {
        register unsigned char ah = regs->h.ah;
        if ( (ah >= 0x01 && ah <= 0x24) ||
             (ah >= 0x27 && ah <= 0x30) ||
             (ah >= 0x37 && ah <= 0x47) ||
             (ah >= 0x4B && ah <= 0x51) )
        {
          __isr_common(0x21);
        }
        if (ah == 0x4B) /* exec */
        {
          regs->x.ax = 8; /* ENOMEM */
          regs->x.flags |= 1;
          return; 
        }
      }
    }
    __my_dpmi_chain_pmiv(regs);
  }
  #endif /* #if !defined(DONT_WANT_INT21H_ISR) */

  /* ---------------------------------------------------------------- */

  static int __lockunlock_interrupt_code(int do_lock)
  {
    /* ensure that code locking happens for the correct range. 
    ** The Watcom and GCC compilers are at liberty to rearrange 
    ** the order of [static] functions. (For watcom when compiled 
    ** with -zm and for GCC when not compiled with -mno-move)
    */
    unsigned topf, botf;
    unsigned tmpf = (unsigned)__isr_common;
    topf = botf = tmpf;

    #if !defined(DONT_WANT_INT28H_ISR)
    tmpf = (unsigned)__int28h_isr;
    if (((unsigned)tmpf) < ((unsigned)topf)) topf = tmpf;
    if (((unsigned)tmpf) > ((unsigned)botf)) botf = tmpf;
    #endif
    #if !defined(DONT_WANT_INT31H_ISR)
    tmpf = (unsigned)__int31h_isr;
    if (((unsigned)tmpf) < ((unsigned)topf)) topf = tmpf;
    if (((unsigned)tmpf) > ((unsigned)botf)) botf = tmpf;
    #endif
    #if !defined(DONT_WANT_INT21H_ISR)
    tmpf = (unsigned)__int21h_isr;
    if (((unsigned)tmpf) < ((unsigned)topf)) topf = tmpf;
    if (((unsigned)tmpf) > ((unsigned)botf)) botf = tmpf;
    #endif

    botf += 4096;
    if (!do_lock)
      return __my_dpmi_unlock_code((void *)topf, (botf - topf) );
    return __my_dpmi_lock_code((void *)topf, (botf - topf) );
  }

  /* ---------------------------------------------------------------- */

  #if !defined(DONT_WANT_INT28H_ISR)
  static struct /* does not have to be locked separately */
  {             /* regs will be locked by __my_dpmi_allocate...ichain */
    __dpmi_regs regs; 
    __dpmi_raddr oldvect;
    __dpmi_raddr cbvect;
    int isset;
  } int28h_cbdata;
  #endif

  /* ---------------------------------------------------------------- */

  static void __isr_shutdown (void)
  {
    TRACE_LEV1(("__isr_shutdown called\n"));
    if (my_statics.isr_hook_state > 0)
    {
      my_statics.isr_hook_state = -1;
      #if !defined(DONT_WANT_INT21H_ISR)
      if (my_statics.int21hpmhook)
      {
        __my_dpmi_unhook_pmiv(my_statics.int21hpmhook);
        my_statics.int21hpmhook = 0; 
      }
      #endif
      #if !defined(DONT_WANT_INT31H_ISR)
      if (my_statics.int31hpmhook)
      {
        __my_dpmi_unhook_pmiv(my_statics.int31hpmhook);
        my_statics.int31hpmhook = 0;
      }
      #endif
      #if !defined(DONT_WANT_INT28H_ISR)
      if (int28h_cbdata.isset)
      {
        __dpmi_set_real_mode_interrupt_vector(0x28, &(int28h_cbdata.oldvect));
        __my_dpmi_free_real_mode_callback( &(int28h_cbdata.cbvect) );
        int28h_cbdata.isset = 0;
      }
      #endif
      __lockunlock_interrupt_code(0);
    }
    SHOW_CALLOUT_COUNTS();
    TRACE_LEV1(("*Yield calls: total time=%lu ms, count=%lu, avgms=%lu ms\n",
            my_statics.yield_time, my_statics.yield_count,
            ((my_statics.yield_count == 0) ? 0 :
                (my_statics.yield_time/my_statics.yield_count))
              )); 
    return;
  }

  /* ---------------------------------------------------------------- */

  static void __isr_setup(void)
  {
    __init_statics_if_needed();
    if (my_statics.isr_hook_state == 0)
    {
      TRACE_LEV1(("isr_shutdown = %p\n", __isr_shutdown));
      /* setting at atterm hook is not strictly necessary for DPMI */
      _my_atterm (__isr_shutdown); /* do before isr hooking! */
      my_statics.isr_hook_state = +1;

      if (__lockunlock_interrupt_code(1) != 0)
      {
        __fatal_runtime_error("PANIC! Unable to lock interrupt code\n");
      }

      #if !defined(DONT_WANT_INT28H_ISR)
      {
        /* NTVDM will trash real-mode DS if the CB is called  */
        /* while an Int21 call is in progress */
        const char *env = getenv("OS");
        if (env && strcmp(env, "Windows_NT")==0)
          ; /* don't use int28h callback */
        else if (__my_dpmi_allocate_real_mode_callback_ichain(__int28h_isr,
                     0x28, &(int28h_cbdata.regs), &(int28h_cbdata.cbvect))==0)
        {
          __dpmi_get_real_mode_interrupt_vector(0x28, &(int28h_cbdata.oldvect) );
          __dpmi_set_real_mode_interrupt_vector(0x28, &(int28h_cbdata.cbvect) );
          int28h_cbdata.isset = 1;
        }
      }
      #endif
      #if !defined(DONT_WANT_INT31H_ISR)
      my_statics.int31hpmhook = __my_dpmi_hook_pmiv(0x31, __int31h_isr );
      #endif
      #if !defined(DONT_WANT_INT21H_ISR)
      my_statics.int21hpmhook = __my_dpmi_hook_pmiv(0x21, __int21h_isr );
      #endif

      TRACE_LEV1(("Hook states: 0x28=%s, 0x31=%s, 0x21=%s\n",
                     (int28h_cbdata.isset ? "rmcb" : "none" ),
                     (my_statics.int31hpmhook? "pm": "none"),
                     (my_statics.int21hpmhook? "pm": "none") ));

    }
    return;
  }

#else /* all realmode targets */

  #undef DONT_WANT_INT28h_HANDLER

  #if defined(__TURBOC__) || defined(__BORLANDC__)
    #if defined(__cplusplus)
    #define __ANYARGS ...
    #else
    #define __ANYARGS
    #endif
    #if !defined(DONT_WANT_INT28h_HANDLER)
    static void _interrupt (_far *__old_int28h_isr)(__ANYARGS);
    #endif
  #elif defined(__WATCOMC__)
    #define __ANYARGS
    #if !defined(DONT_WANT_INT28h_HANDLER)
    static void (_interrupt _far *__old_int28h_isr)(__ANYARGS);
    #endif
  #else
    #define __ANYARGS ...
    #if !defined(DONT_WANT_INT28h_HANDLER)
    static void (_interrupt _far *__old_int28h_isr)(__ANYARGS);
    #endif
  #endif

  #if !defined(DONT_WANT_INT28h_HANDLER)
  static void _interrupt _far __new_int28h_isr(__ANYARGS)
  {
    #if defined(__TURBOC__) || defined(__BORLANDC__)
    #pragma warn -pro
    #endif
    if (__old_int28h_isr)
      (*__old_int28h_isr)();
    #if defined(__TURBOC__) || defined(__BORLANDC__)
    #pragma warn .pro
    #endif

    __isr_common(0x28);
  }
  #endif /* !defined(DONT_WANT_INT28h_HANDLER) */

  static void __isr_shutdown (void)
  {
    __init_statics_if_needed();
    if (my_statics.isr_hook_state > 0)
    {
      TRACE_LEV1(("isr_shutdown called\n" ));
      my_statics.isr_hook_state = -1;
      #if !defined(DONT_WANT_INT28h_HANDLER)
      _dos_setvect (0x28, __old_int28h_isr);
      #endif
      SHOW_CALLOUT_COUNTS();
    }
    return;
  }

  static void __isr_setup(void)
  {
    __init_statics_if_needed();
    if (my_statics.isr_hook_state == 0)
    {
      /*_printk("blocker_cdelay = %04x:%04x\n", 
                 FP_SEG(blocker_cdelay), FP_OFF(blocker_cdelay) ); 
      */
      my_statics.isr_hook_state = +1;
      _my_atterm (__isr_shutdown);
      #if !defined(DONT_WANT_INT28h_HANDLER)
      __old_int28h_isr = _dos_getvect(0x28);
      _dos_setvect (0x28, __new_int28h_isr);
      #endif
    }
    return;
  }

#endif

/* ==================================================================== */
/*                            yield functions                           */
/* ==================================================================== */

/* "smart" hard yield: what makes this function smart is that
** it yields using the correct yield method depending on the
** environment its running under. 
**
** DPMI targets _DO_NOT_ automatically use DPMI yield because 
** many extenders (eg PMODE) simply HLT the processor, which
** under plain DOS (where such an extender would be the active
** host) is not a yield.
*/
static void __hard_yield(void) /* smart hard yield */
{
  static int yieldtype = 0;
  #define YIELDTYPE_NONE     0
  #define YIELDTYPE_DESQVIEW 1
  #define YIELDTYPE_DPMI     2
  #define YIELDTYPE_BIOSWAIT 3
  #define YIELDTYPE_DOSIDLE  4

  #if defined(__DPMI32__)
  __dpmi_regs regs;
  #endif

  if (yieldtype == YIELDTYPE_NONE)
  {
    unsigned short sversion = 0;
    char *p;

    ++my_statics.in_crit_sect;

    if ((p = getenv("OS")) != ((char *)0))
    {
      if (strcmp( p, "Windows_NT" )==0)
        yieldtype = YIELDTYPE_DPMI;
    }  
    if (yieldtype == YIELDTYPE_NONE)
    {
      #if defined(__DPMI32__)
      regs.x.ax = 0x3306;
      regs.x.bx = 0xffff;
      __dpmi_int( 0x21, &regs );
      sversion = regs.x.bx;
      #else
      __asm mov ax, 3306h
      __asm mov bx, 0ffffh
      __asm int 21h
      __asm mov sversion, bx
      #endif
      if (((sversion >> 8) & 0xff) <= 100 && 
          (sversion & 0xff) >= 10 && ((sversion & 0xff) % 10) == 0)
        yieldtype = YIELDTYPE_DPMI; /* OS/2 */
    }
    if (yieldtype == YIELDTYPE_NONE)
    {
      #if defined(__DPMI32__)
      regs.x.ax = 0x1600;
      __dpmi_int( 0x2F, &regs );
      sversion = regs.x.ax;
      #else
      __asm mov ax, 1600h
      __asm int 2fh
      __asm mov sversion, ax 
      #endif
      if ((sversion & 0xff) > 1 && (sversion & 0xff) != 0x80)
        yieldtype = YIELDTYPE_DPMI; /* Win16 enhanced mode */
    }
    if (yieldtype == YIELDTYPE_NONE)
    {
      #if defined(__DPMI32__)
      regs.x.ax = 0x2B01;
      regs.x.bx = 0x0;
      regs.x.cx = 0x4445;
      regs.x.dx = 0x5351;
      __dpmi_int( 0x21, &regs );
      if (regs.h.al != 0xff)
        regs.x.bx = 0;
      sversion = regs.x.bx;
      #else
      __asm mov ax,2B01h  /* DESQview installation check */
      __asm mov cx,4445h  /* 'DE' */
      __asm mov dx,5351h  /* 'SQ' */
      __asm mov bx,0      /* clear version # */
      __asm int 21h
      __asm cmp al,0FFh   /* set carry flag if supported (<0xFF) */
      __asm sbb ax,ax     /* make zero if not supported, 0xFFFF otherwise */
      __asm and bx,ax     /* clear version if not supported */
      __asm mov sversion,bx
      #endif

      if (sversion != 0)
        yieldtype = YIELDTYPE_DESQVIEW;
    }
    #if !defined(__DPMI32__)
    if (yieldtype == YIELDTYPE_NONE)
    {
      _asm mov ax,1680h
      _asm int 2fh
      _asm mov sversion,ax
      if ((sversion & 0xff) != 0x80)
        yieldtype = YIELDTYPE_DPMI;
    }
    #endif
    if (yieldtype == YIELDTYPE_NONE)
    {
      yieldtype = YIELDTYPE_DOSIDLE;
    }
    TRACE_LEV1(("yieldtype = 0x%04x\n", yieldtype ));
    --my_statics.in_crit_sect;
  }

  switch (yieldtype)
  {
    case YIELDTYPE_DESQVIEW:
    {
      #if defined(__DPMI32__)
      regs.x.ax = 0x1000;
      __dpmi_int( 0x15, &regs );
      #else
      __asm mov ax,1000h                         /* DesqView idle */
      __asm int 15h
      #endif
      break;
    }
    case YIELDTYPE_DPMI:
    {
      #if defined(__DPMI32__)
      regs.x.ax = 0x1680;
      __dpmi_int( 0x2F, &regs );
      #else
      __asm mov ax,1680h  
      __asm int 2fh
      #endif
      break;
    }
    case YIELDTYPE_BIOSWAIT: /* only 286 and above, not NEC */
    {                        /* hangs under WinNT4 pre-fixpack 3 */
      #if defined(__DPMI32__)
      regs.x.ax = 0x8600;
      regs.x.cx = 0x0000;
      regs.x.dx = 0x0001;
      __dpmi_int( 0x15, &regs );
      #else
      __asm mov ah, 86h                        
      __asm mov cx, 0   /* cx:dx == time to wait in microsecs, but */
      __asm mov dx, 1   /* resolution is 976us */ 
      __asm int 15h
      #endif
      break;
    }
    default:
    case YIELDTYPE_DOSIDLE:
    { 
      #if defined(__DPMI32__)
      /* warning! don't use __dpmi_yield(). That is redirected. */
      __dpmi_int( 0x28, &regs ); 
      #else
      __asm int 28h
      #endif
      break;
    }
  } /* switch */
  return;
}      

/* --------------------------------------------------------------------- */

/*
 * We need a function that 
 * a) ALWAYS checks ^C/^break  
 * b) ALWAYS issues an interrupt (for soft yielding).
 *
 * Most of the libcs have a kbhit() that is entirely unusable for this 
 * purpose - DJGPP [pc.h] is one that uses the BIOS (int16 ick!), and/or
 * does a direct read on the kbqueue. HighC, BorlandC/TurboC, WatcomC and 
 * DJGPP/conio.h have a kbhit() that reflects the state of ungetch().
 * Even when a libc does use DOS, some use the "direct console input" 
 * function (int21h,ah=06h,dl=ffh) which does not check for break.
 *
 * INT 21 - DOS 1+ - GET STDIN STATUS
 *      AH = 0Bh
 * Return: AL = status
 *          00h if no character available
 *          FFh if character is available
 * Notes:  ^C/^Break are checked, and INT 23 is called if either pressed
 *      standard input is always the keyboard under DOS 1.x, but may be
 *      redirected under DOS 2+
*/
#if defined(__WATCOMC__)
extern void __hard_ctrlc_check(void);
#pragma aux __hard_ctrlc_check = 0xb4 0x0b 0xcd 0x21 modify exact [ax];
#elif defined(__DJGPP__)
#define __hard_ctrlc_check() __asm__(".byte 0xb4,0x0b,0xcd,0x21" : : : "eax")
#else
#define __hard_ctrlc_check() __emit__(0xb4,0x0b,0xcd,0x21)
#endif

/* --------------------------------------------------------------------- */

void blocker_inc_hwork(void)
{
  my_statics.hwork_marker = timer_msec_count();
}  

/* --------------------------------------------------------------------- */

int blocker_cyield(int yield_flags, const char *caller, int line)
{
  unsigned long now;
  int sawbrk = 0;

  __isr_setup(); /* setup hooks if they haven't been setup yet */
  ++my_statics.in_crit_sect;

  if ((++my_statics.in_blocker_yield) != 1) 
  {
    if ((yield_flags & YIELDFLAG_NOIHKS) == 0)
    {
      yield_flags |= YIELDFLAG_NOIHKS;
      /*
      _printk("blocker_cyield() called recursively (caller=%s,line=%u)\n",
                                          (caller ? caller : "?"), line );
      */
      /* the app could still hang! */
    }
  }

  TRACE_LEV5(("begin blocker_cyield(%x) %p (caller=%s)\n", 
    yield_flags, blocker_cyield, (caller ? caller : "?"), line ));

  now = timer_msec_count();

  if ((yield_flags & YIELDFLAG_NOIHKS) == 0)
  {
    __call_hooks(BLOCKER_HOOKTYPE_IDLE | BLOCKER_HOOKTYPE_PERIODIC);
  }

  if (yield_flags & YIELDFLAG_HARD)
  {
    /* don't hard yield if hardware work was done in the last 250ms */
    if ((my_statics.hwork_marker>>2) != (now>>2))
      yield_flags &= YIELDFLAG_HARD;
    else
    {
      unsigned long n = now;
      __hard_yield();
      my_statics.yield_time += (timer_msec_count() - n);
      my_statics.yield_count++;
    }
  }

  if (yield_flags & YIELDFLAG_CHKBRK)
  {
    if (get_cbreakstate() == 0) /* ^C is being blocked */
    {
      sawbrk = get_cbreakpend();
      if (!sawbrk)
      {
        static unsigned long kbchk = 0;
        /* don't do break check more often than every 250ms */  
        if ((now>>2) != kbchk)
        {
          int oldbrkpending = get_cbreakpend();
          __hard_ctrlc_check();
          if (oldbrkpending != get_cbreakpend())
          {
            sawbrk = 1; /* ^C seen */
            /* _printk("blocker_cyield() SIGINT seen\n"); */
          }
          kbchk = now>>2;
        }
      }
    }
  }

  if (!sawbrk &&
      (yield_flags & (YIELDFLAG_HARD|YIELDFLAG_NOIHKS)) == YIELDFLAG_HARD)
  {
    __call_hooks(BLOCKER_HOOKTYPE_IDLE | BLOCKER_HOOKTYPE_PERIODIC);
  }

  caller = caller;
  line = line;

  TRACE_LEV5(("end blocker_cyield (sawbrk=%d)\n",sawbrk));

  --my_statics.in_blocker_yield;
  --my_statics.in_crit_sect;

  return sawbrk; /* ^C seen or not */
}

/* --------------------------------------------------------------------- */

int blocker_cdelay(unsigned int millisecs,const char *caller, int line)
{
  if (millisecs == 0)
  {
    /* just let the hooks run */
    blocker_cyield(0,caller,line);
  }
  else
  {
    unsigned long now = 0, elap, start = timer_msec_count();
    int old_brkstate = set_cbreakstate(0);
    while (!blocker_cyield(YIELDFLAG_HARD|YIELDFLAG_CHKBRK,caller,line))
    {
      now = timer_msec_count();
      elap = now - start;
      start = now;
      if (elap > millisecs)
        elap = millisecs;
      millisecs -= (unsigned)elap;
      if (!millisecs)
        break;
    }
    set_cbreakstate(old_brkstate);
    /* if (millisecs) _printk("blocker_cdelay() ^C seen\n"); */
  }
  return millisecs; /* remaining */
}

/* ====================================================================== */
/*                          C-Library replacements                        */
/* ====================================================================== */

#if !defined(__DJGPP__)
/* Watcom and Borland libraries do not actually 
** yield for delay() and sleep(). They just spin.
** Watcom uses DOS time-of-day to get the time!
** TurboC used ioports to read sub-tick time!
**
** DJGPP's delay() uses int15h which we catch
** in the int31h pm ISR.
*/
void delay( unsigned milliseconds )
{
  blocker_delay(milliseconds);
}

void sleep( unsigned secs )
{
  unsigned rem = 0;
  do
  {
    unsigned int s = (UINT_MAX/1000u);
    if (secs < s)
      s = secs;
    secs -= s;
    rem = blocker_delay(s * 1000u);
    secs += (rem+599u)/1000u;
  } while (!rem && secs);
  return /*seconds*/;
}
#endif

/* In order to increase the frequency of calls to blocker_yield(),
** some commonly used libc functions are replaced as follows.
**  - Time/date functions are often called frequently, but DOS
**    does not call int28h. 
**  - kbhit() does call int28, but since it is called often
**    it is more efficient to call blocker_yield() directly.
*/
#if defined(__TURBOC__)
/*
** For turboC/BCC it is far more practical to replace getenv() than all
** functions that use Int21h/ah=2Ah+2Ch because all those functions 
** eventually call getenv() (via tzset()).
*/
char _FAR * __kgetenv(const char _FAR *envvar)
{
  #if (__TURBOC__ <= 0x410)
  #define _environ environ
  #endif

  char *result = (char *)0;
  if (envvar)
  {
    if (*envvar)
    {
      char **envpp = _environ;
      if (envpp)
      {
        for (;;)
        {
          char *envp = *envpp++;
          if (!envp)
            break;
          if (*envp == '\0')
            break;
          if (*envvar == *envp)
          {
            const char _FAR *e = envvar;
            while (*e && *envp && (*e == *envp)) 
            {
              e++;
              envp++;
            }
            if (!*e && *envp == '=')
            {
              result = ++envp;
              break;
            }
          }
        }
      }
    }
  }
  return result;
}
#undef getenv
char _FAR * getenv(const char _FAR *envvar)
{
  blocker_yield(0);
  return __kgetenv(envvar);
}
int kbhit(void)
{
  extern unsigned char _cFlag;
  char hit = 1;
  blocker_yield(0);
  if (!_cFlag)
  {
    blocker_critsect_begin(); /* prevent int28 handler from running */
    __asm mov ah,0Bh
    __asm int 21h
    __asm mov hit,al
    blocker_critsect_end();
  }
  return hit;
}
#endif

/* ---------------------------------------------------------------------- */

#if defined(__WATCOMC__)
/*
** WatcomC's time()/gmtime()/localtime()/ftime() etc all call __getctime()
*/
#include <time.h>
#ifdef __cplusplus
extern "C" int __getctime(struct tm *);
#endif
int __getctime( struct tm *_ti )
{
  #pragma pack(1)
  typedef struct tiny_date_t {
    unsigned char  day_of_month;   /* 1 - 31 */
    unsigned char  month;          /* 1 - 12 */
    unsigned char  year;           /* year minus 1900 */
    unsigned char  day_of_week;    /* 0 - Sun, ..., 6 - Sat */
  } tiny_date_t;
  typedef struct tiny_time_t {
    unsigned char  hundredths;     /* 0 - 99 */
    unsigned char  seconds;        /* 0 - 59 */
    unsigned char  minutes;        /* 0 - 59 */
    unsigned char  hour;           /* 0 - 23 */
  } tiny_time_t;
  #pragma pack()

  extern tiny_date_t _TinyGetDate( void );
  extern tiny_time_t _TinyGetTime( void );
  #if defined(__386__)
  #pragma aux _TinyGetDate = "mov ah,2Ah" "int 21h" "sub cx,1900" "mov ch,al" \
                "shl ecx,16" "mov cx,dx" \
                parm caller [] value [ecx] modify exact [eax ecx edx];
  #pragma aux _TinyGetTime = "mov ah,2Ch" "int 21h" \
                "shl ecx,16" "mov cx,dx" \
                parm caller [] value [ecx] modify exact [eax ecx edx];
  #else
  #pragma aux _TinyGetDate = "mov ah,2Ah" "int 21h" "sub cx,1900" "mov ch,al" \
                parm caller [] value [cx dx] modify exact [ax cx dx];
  #pragma aux _TinyGetTime = "mov ah,2Ch" "int 21h" \
                parm caller [] value [cx dx] modify exact [ax cx dx];
  #endif

  tiny_time_t t, tmp;
  tiny_date_t d;

  blocker_yield(0);
  blocker_critsect_begin(); /* prevent int21 handler from running */
  do
  {
    t = _TinyGetTime();
    d = _TinyGetDate();
    if (t.hour != 23)
      break;
    /* check for midnight rollover */
    tmp = _TinyGetTime();
  } while (tmp.hour != t.hour);
  blocker_critsect_end();

  {
    register struct tm * ti = _ti;
    ti->tm_year = d.year;
    ti->tm_mon  = d.month - 1;
    ti->tm_mday = d.day_of_month;
    ti->tm_wday = d.day_of_week;
    ti->tm_hour = t.hour;
    ti->tm_min  = t.minutes;
    ti->tm_sec  = t.seconds;
    ti->tm_isdst = -1;
  }
  return (10u * ((unsigned int)t.hundredths));
}
_WCRTLINK int kbhit(void)
{
  extern int _cbyte;
  signed char hit = 1;
  blocker_yield(0);
  if (!_cbyte)
  {
    extern signed char my_kbhit();
    #pragma aux my_kbhit = 0xb4 0x0b 0xcd 0x21 value [al] modify exact [ax];
    blocker_critsect_begin(); /* prevent int28 handler from running */
    hit = my_kbhit();    /* mov ah,0bh; int 21h; */
    blocker_critsect_end();
  }
  return hit;
}
#endif /* __WATCOMC__ */

/* ---------------------------------------------------------------------- */

#if defined(__DJGPP__)
static int stub_usleep(unsigned int usecs) 
{
  if (usecs < (UINT_MAX - 599u))
    usecs += 599u;
  return 1000u * blocker_delay(usecs/1000u);
}
static void stub_dpmi_yield(void)
{
  blocker_yield(YIELDFLAG_HARD);
}
static unsigned int stub_delay(unsigned int msecs) 
{
  return blocker_delay(msecs);
}
static int stub_sleep(unsigned int secs) 
{ 
  unsigned rem = 0;
  do
  {
    unsigned int s = (UINT_MAX/1000u);
    if (secs < s)
      s = secs;
    secs -= s;
    rem = blocker_delay(s * 1000u);
    secs += (rem+599u)/1000u;
  } while (!rem && secs);
  return secs;
}
#undef kbhit /* don't let the conio one get in the way */
#if defined(__cplusplus)
extern "C" int kbhit(void);
#else
extern int kbhit(void);
#endif
static int stub_kbhit(void) /* DJGPP's kbhit() truly stinks */
{
  blocker_yield(YIELDFLAG_HARD);
  if (_farpeekw(_dos_ds, 0x41a) != _farpeekw(_dos_ds, 0x41c))
  {
    __dpmi_regs r;
    r.h.ah = 0x11;
    __dpmi_int(0x16, &r);
    if (!(r.x.flags & 0x40)) /* Z */
      return 1;
  }
  return 0;
}
static void __ctor_yieldhelpers(void)
{
  int cs_alias = __my_dpmi_get_csalias();
  if (cs_alias != -1) /* don't stub if failed, dpmi.c will carp (later) */
  {
    static const struct { void *our_func; void *libc_func; } stub_tab[] = {  
      #pragma weak usleep
      { (void *)stub_usleep,     (void *)usleep       },
      #pragma weak __dpmi_yield
      { (void *)stub_dpmi_yield, (void *)__dpmi_yield }, 
      #pragma weak __delay
      { (void *)stub_delay,      (void *)delay        },
      #pragma weak sleep
      { (void *)stub_sleep,      (void *)sleep        },
      #pragma weak kbhit
      { (void *)stub_kbhit,      (void *)kbhit        }
    };
    unsigned int i;
    for (i = 0; i < (sizeof(stub_tab)/sizeof(stub_tab[0])); i++)
    {
      void *libcfunc = stub_tab[i].libc_func;
      if (libcfunc)
      {
        const void *ourfunc = stub_tab[i].our_func;
        unsigned long offset;
        char buf[8];

        memcpy( buf, libcfunc, sizeof(buf) ); /* force page in */
        if (__my_dpmi_lock_code(libcfunc, sizeof(buf)) != 0)
          break; /* stop. out of phys memory. some other init will fail */ 

        offset = ((const char *)ourfunc) - ((const char *)libcfunc);
        buf[0] = 0xe9;    /* "jmp" (relative) */
        *((unsigned long *)(&buf[1])) = offset - 5; /* 5==sizeof('jmp xyz') */
        movedata(_my_ds(), (unsigned)(&buf[0]), cs_alias, (unsigned)libcfunc, 5);
      }
    }
  }  
  return;
}
#define __SETUP_CONSTRUCTOR_FOR__ __ctor_yieldhelpers
#include "ctordtor.h"
#endif /* __DJGPP__ */

/* ====================================================================== */

