/*
** - portable constructor/destructor generation macros for 
**   WatcomC, TurboC/BorlandC and GnuC
** - _my_atexit(): a replacement for atexit() so as to not 
**   consume libc atexit() slots.
** - _my_atterm(): a means to implement shutdown handlers that
**   are called from _exit()/_terminate()
**
** Created May 2001 by Cyrus Patel <cyp@fb14.uni-mainz.de>
** $Id: ctordtor.h,v 1.1 2001/05/07 11:17:21 cyp Exp $
**
** "Constructors" are called before main() is called [directly from crt0 
** generally]. 
**
** "Destructors" are generally called after main() ie, from exit(), 
** BEFORE _exit(), but after atexit() hooks have been run. MSC and WatcomC 
** also have a category of destructor that may be called from _exit() 
** [just before _terminate()].
** 
** Use _my_atterm() [ctordtor.c] to implement a destructor that needs
** to be called from _exit().
**
** - For gcc: construction/destruction order is not specifyable, ergo,
**   contructors/destructors may be called before, after, or any time in-
**   between the libc's own constructors/destructors, so our own ctors/dtors 
**   may NOT depend on any libc subsystem being initialized when they are 
**   called. As a rule-of-thumb: Direct use of DOS or DPMI calls is ok, 
**   everything else (particularly handle i/o or stdio or heap use) is NOT ok.
**   Note that destructors declared here are called from exit() [not _exit()!]
**   but after atexit() hooks.
** - For TurboC:
**   + constructors (declared here) are called after library initialization
**   + destructors declared here are always called FROM exit() [not _exit()!] 
**     ie, after atexit() hooks but before _exit().
**     TURBOC documents that functions qualified with #pragma exit 
**     are called "just before the program terminates through _exit" (not
**     very clear, eh?) and tests show that destructors are *not* called.
** - For MSC and WatcomC: 
**   + constructors declared here are always called after library init, 
**   + destructors declared here are called during _exit() (ie, after 
**     library shutdown)
**
** Declaring constructors/destructors:
**
**   <all other code here>
**   
**   static void my_init_fxn(void) // must take no args
**   {                             // but does not have to be static
**     ...
**   }
**
**  NOTE: constructor function names have to be globally unique even if static.
**  (gcc and watcom need to play anti-optimization tricks by generating
**  public symbols, which they will do based on this name)
**
**   #define __SETUP_CONSTRUCTOR_FOR__ my_init_fxn
**   #include "ctordtor.h" // emits the right code to call my_init_fxn()
**   #define __SETUP_CONSTRUCTOR_FOR__ my_other_init_fxn
**   #include "ctordtor.h" // as you see, can be included more than once
*/
/* There are intentionally no include guards at the top of this file
** to allow it to be included more than once (to declare multiple constructors)
** There is a __CTOR_DTOR_H__ at the BOTTOM of this file.
*/
#undef DEBUG_CTORDTOR

/* ======================================================================== */
#if defined(DEBUG_CTORDTOR) && !defined(__TEXTIFY_DEFINED)
#include "printk.h"
#define __TEXTIFY(x) #x
#define _TEXTIFY(x) __TEXTIFY(x)
#define __TEXTIFY_DEFINED
#endif
#if defined(DEBUG_CTORDTOR) && defined(__SETUP_CONSTRUCTOR_FOR__)
#define CTORB _printk("begin " _TEXTIFY(__SETUP_CONSTRUCTOR_FOR__) "\n");
#define CTORE _printk("end   " _TEXTIFY(__SETUP_CONSTRUCTOR_FOR__) "\n");
#else
#define CTORB
#define CTORE
#endif
#if defined(DEBUG_CTORDTOR) && defined(__SETUP_DESTRUCTOR_FOR__)
#define DTORB _printk("begin " _TEXTIFY(__SETUP_DESTRUCTOR_FOR__) "\n");
#define DTORE _printk("end   " _TEXTIFY(__SETUP_DESTRUCTOR_FOR__) "\n");
#else
#define DTORB
#define DTORE
#endif

/* ======================================================================== */


#if defined(__TURBOC__) /* both turbc/borlandc */
/* TurboC has #pragma startup/shutdown, and we want a common way to do it.
** Tried and failed with ...
** (same structure as used by watcom below, but flagged 'far')
**   #pragma option -zHDGROUP
**   #pragma option -zFEXITDATA            // -zFINITDATA for ctor
**   #pragma option -zE_EXIT_              // -z_INIT_ for ctor
**   AYI( __SETUP_DESTRUCTOR_FOR__ , <n> ) // AXI() for ctor
**   #pragma option -zE*
**   #pragma option -zF*
**   #pragma option -zH*
** but I couldn't get it to work right...
**   1) TC has a bug that only allows either a ctor or a dtor (per module)
**   2) the far segments are always para aligned (not byte aligned), so
**      the structure could end up on a non-packed boundary even when 
**      using pragma pack().
**   The reason why I had to use -zH/-zF/-zE rather than (what should be)
**   -zS/-zT/-zR is because the former can be changed at any time, while
**   the latter can only be changed before any std header is included or
**   code is emitted.
*/
#if defined(__SETUP_CONSTRUCTOR_FOR__)
   #ifndef __DEFINED_CONSTRUCTOR__
   /* we can't use a define inside a #pragma so create a wrapper. */
   static void __ctor_0__(void) { CTORB __SETUP_CONSTRUCTOR_FOR__ (); CTORE }
   #pragma startup __ctor_0__ 200
   #define __DEFINED_CONSTRUCTOR__ 0
   #elif (__DEFINED_CONSTRUCTOR__ == 0)
   static void __ctor_1__(void) { CTORB __SETUP_CONSTRUCTOR_FOR__ (); CTORE }
   #pragma startup __ctor_1__ 200
   #undef __DEFINED_CONSTRUCTOR__
   #define __DEFINED_CONSTRUCTOR__ 1
   #elif (__DEFINED_CONSTRUCTOR__ == 1)   
   static void __ctor_2__(void) { CTORB __SETUP_CONSTRUCTOR_FOR__ (); CTORE }
   #pragma startup __ctor_2__ 200
   #undef __DEFINED_CONSTRUCTOR__
   #define __DEFINED_CONSTRUCTOR__ 2
   #else
   #error Stuff in ctordtor.h only allows up to 3 constructors per file.
   #endif
   #undef __SETUP_CONSTRUCTOR_FOR__ 
#endif
#if defined(__SETUP_DESTRUCTOR_FOR__)
   #if defined(__TURBOC__)
   #  pragma warning "Warning! TC calls destructors from exit(), not from _exit()!"
   #endif
   #ifndef __DEFINED_DESTRUCTOR__
   /* we can't use a define inside a #pragma so create a wrapper. */
   static void __dtor_0__(void) { DTORB __SETUP_DESTRUCTOR_FOR__ (); DTORE }
   #pragma exit __dtor_0__ 200
   #define __DEFINED_DESTRUCTOR__ 0
   #elif (__DEFINED_DESTRUCTOR__ == 0)
   static void __dtor_1__(void) { DTORB __SETUP_DESTRUCTOR_FOR__ (); DTORE }
   #pragma exit __dtor_1__ 200
   #undef __DEFINED_DESTRUCTOR__
   #define __DEFINED_DESTRUCTOR__ 1
   #elif (__DEFINED_DESTRUCTOR__ == 1)   
   static void __dtor_2__(void) { DTORB __SETUP_DESTRUCTOR_FOR__ (); DTORE }
   #pragma exit __dtor_2__ 200
   #undef __DEFINED_DESTRUCTOR__
   #define __DEFINED_DESTRUCTOR__ 2
   #else
   #error Stuff in ctordtor.h only allows up to 3 destructors per file.
   #endif
   #undef __SETUP_DESTRUCTOR_FOR__ 
#endif /* ifdef __SETUP_DESTRUCTOR_FOR__ || ifdef __SETUP_DESTRUCTOR_FOR__ */
#endif /* __TURBOC__ */

/* ======================================================================== */

#if defined(__WATCOMC__) && !defined(__CTOR_DTOR_H__)
#define HAVE_ATTERM_STYLE_DESTRUCTOR

#  if defined(__386__) || defined(__AXP__) || defined(__PPC__)
   typedef void __near (*rtn_fxn_proto)(void);
#  else
   typedef void        (*rtn_fxn_proto)(void);
#  endif

#  pragma pack(1);
   struct rt_init                  // structure placed in XI/YI segment
   {
     unsigned char rtn_type;
     unsigned char priority;
     rtn_fxn_proto rtn_fxn;     
     #if !defined(__386__) && !defined(__MEDIUM__) && \
         !defined(__LARGE__) && !defined(__HUGE__)
     unsigned short pad;
     #endif
   };
#  pragma pack()
#  if defined(__386__) || defined(__AXP__) || defined(__PPC__)
   #define YIXI( seg, label, routine, priority )             \
     struct rt_init __based( __segname( seg ) ) label =      \
     { 0, priority, (rtn_fxn_proto)routine };
#  elif defined(__MEDIUM__) || defined(__LARGE__) || defined(__HUGE__)
   #define YIXI( seg, label, routine, priority )             \
     struct rt_init __based( __segname( seg ) ) label =      \
     { 1, priority, (rtn_fxn_proto)routine };
#  else
   #define YIXI( seg, label, routine, priority )             \
     struct rt_init __based( __segname( seg ) ) label =      \
     { 0, priority, (rtn_fxn_proto)routine, 0 };
#  endif
#  define TS_SEG_XI       "XI"
#  define TS_SEG_YI       "YI"
#  define XI( label, routine, priority ) YIXI( TS_SEG_XI, label, &routine, priority )
#  define YI( label, routine, priority ) YIXI( TS_SEG_YI, label, &routine, priority )
#  define __ANON( x )     __anon_ ## x
#  define ANON( x )       __ANON( x )
#  define AXI( routine, priority ) static XI( ANON( __LINE__ ), routine, priority)
#  define AYI( routine, priority ) static YI( ANON( __LINE__ ), routine, priority)
#  define INIT_PRIORITY_THREAD     1 // priority for thread data init
#  define INIT_PRIORITY_FPU        2 // priority for FPU/EMU init
#  define INIT_PRIORITY_RUNTIME   10 // priority for run/time initialization
#  define INIT_PRIORITY_IOSTREAM  20 // priority for IOSTREAM
#  define INIT_PRIORITY_LIBRARY   32 // default library-initialization priority
#  define INIT_PRIORITY_PROGRAM   64 // default program-initialization priority
#  define FINI_PRIORITY_DTOR      40 // priority for module DTOR
#  define DTOR_PRIORITY           40 // priority for module DTOR
#  define FINI_PRIORITY_EXIT      16 // FINI_PRIORITY_EXIT through 255 are 
                                     // called from exit(), and 
                                     // 0 through FINI_PRIORITY_EXIT-1 
                                     // are called from __exit(),
#define __CONCAT(x,y) x ## y
#define CONCAT(x,y) __CONCAT(x,y)
#define USCORES(x) CONCAT(x,__)
#endif /* defined(__WATCOMC__) && !defined(__CTOR_DTOR_H__) */

#if defined(__WATCOMC__) && defined(__SETUP_CONSTRUCTOR_FOR__)
#  pragma off(unreferenced)
   static void USCORES(__SETUP_CONSTRUCTOR_FOR__) (void)
   { CTORB __SETUP_CONSTRUCTOR_FOR__ (); CTORE }
   static XI( USCORES(USCORES(__SETUP_CONSTRUCTOR_FOR__)) , 
              USCORES(__SETUP_CONSTRUCTOR_FOR__) , INIT_PRIORITY_PROGRAM-1 )
#  pragma on(unreferenced)
#  undef __SETUP_CONSTRUCTOR_FOR__
#endif
#if defined(__WATCOMC__) && defined(__SETUP_DESTRUCTOR_FOR__)
#  pragma off(unreferenced)
   static void USCORES(__SETUP_DESTRUCTOR_FOR__) (void)
   { DTORB __SETUP_DESTRUCTOR_FOR__ (); DTORE }
   static YI( USCORES(USCORES(__SETUP_DESTRUCTOR_FOR__)) , 
              USCORES(__SETUP_DESTRUCTOR_FOR__) , FINI_PRIORITY_EXIT-1 )
#  pragma on(unreferenced)
#  undef __SETUP_DESTRUCTOR_FOR__
#endif

/* ======================================================================== */

#if defined(__GNUC__)
#  if !defined(__CTOR_DTOR_H__)
#    define __CONCAT(x,y) x ## y
#    define CONCAT(x,y) __CONCAT(x,y)
#    define USCORES(x) CONCAT(x,__)
#    if defined(__cplusplus) && !defined(__TEXTIFY_DEFINED)
#      define __TEXTIFY(x) #x
#      define _TEXTIFY(x) __TEXTIFY(x)
#      define __TEXTIFY_DEFINED
#    endif
#  endif /* !defined(__CTOR_DTOR_H__) */
#  if defined(__SETUP_CONSTRUCTOR_FOR__)
#    if defined(__cplusplus)
       /* the c++ frontend doesn't grok __attribute__((__constructor__)) */
       extern "C" void USCORES( __SETUP_CONSTRUCTOR_FOR__) (void);
       void USCORES( __SETUP_CONSTRUCTOR_FOR__) (void) \
       { CTORB __SETUP_CONSTRUCTOR_FOR__ (); CTORE }
       __asm__(".section .ctor; .long _" \
               _TEXTIFY(USCORES(__SETUP_CONSTRUCTOR_FOR__)) ";" \
               ".section .text");
#    else
       static void __attribute__((__constructor__)) \
       USCORES( __SETUP_CONSTRUCTOR_FOR__) (void) \
       { CTORB __SETUP_CONSTRUCTOR_FOR__ (); CTORE }
#    endif
#    undef __SETUP_CONSTRUCTOR_FOR__
#  endif
#  if defined(__SETUP_DESTRUCTOR_FOR__)
#    if defined(__DJGPP__)
#      pragma warning "Warning! DJGPP calls destructors from exit(), not from _exit()!"
#    endif
#    if defined(__cplusplus)
       /* the c++ frontend doesn't grok __attribute__((__destructor__)) */
       extern "C" void USCORES( __SETUP_DESTRUCTOR_FOR__) (void);
       void USCORES( __SETUP_DESTRUCTOR_FOR__) (void) \
       { CTORB __SETUP_DESTRUCTOR_FOR__ (); CTORE }
       __asm__(".section .dtor; .long _" \
               _TEXTIFY(USCORES(__SETUP_DESTRUCTOR_FOR__)) ";" \
               ".section .text");
#    else
       static void __attribute__((destructor)) \
       USCORES(__SETUP_DESTRUCTOR_FOR__) (void) \
       { DTORB __SETUP_DESTRUCTOR_FOR__ ();  DTORE 
#    endif
#    undef __SETUP_DESTRUCTOR_FOR__
#  endif
#endif /* __GNUC__ */

/* ======================================================================== */

#if defined(_MSC_VER)
#  error check this please
/* According to MSC 7.0 (ie 16bit) crt0dat.asm the init/exit segment callout 
** look like this...
** _init:
**       XIF  // init's first
**       XI   // init's
**  exit: 
**       XO   // onexit() handlers
**       XP   // pre-terminators
**  _exit:
**       XC   // terminators
**       XCF  // terminators last
** It appears that the far/non-far names are misleading - the vectors are 
** always called with 'call dword ptr [xyz]', ie both vectors and functions 
** have to be far.
*/
#  if !defined(__CTOR_DTOR_H__)
#    define HAVE_ATTERM_STYLE_DESTRUCTOR
#    define __ANON_CODE( x )     __anon_code ## x
#    define __ANON_DATA( x )     __anon_data ## x
#    define ANON_CODE( x )       __ANON_CODE( x )
#    define ANON_DATA( x )       __ANON_DATA( x )
#  endif /* !defined(__CTOR_DTOR_H__) */
#  if defined(__SETUP_CONSTRUCTOR_FOR__)
#    if _MSC_VER >= 1010 /* 32bit */
     /* C initializers are at .CRT$XIA...XIZ, C++ constructors are XCA..XCZ*/
     /* all libc contructors are at .CRT$XIC */
#    pragma comment( linker, "/Merge:.CRT=.data" )
#    pragma data_seg(".CRT$XIY")
#    define __XFAR /* nothing - flat */
#    else
#    pragma data_seg( "XI" )
#    define __XFAR __far /* links are always far code/far data */
#    endif
     static void __cdecl __XFAR ANON_CODE(__SETUP_CONSTRUCTOR_FOR__) (void) \
     { CTORB  __SETUP_CONSTRUCTOR_FOR__ ();  CTORE }
     static void (__cdecl __XFAR * ANON_DATA(__SETUP_CONSTRUCTOR_FOR__) = \
                                   ANON_CODE(__SETUP_CONSTRUCTOR_FOR__);
#    pragma data_seg()
#    undef __XFAR
#    undef __SETUP_CONSTRUCTOR_FOR__
#  endif
#  if defined(__SETUP_DESTRUCTOR_FOR__)
#    if _MSC_VER >= 1010 /* 32bit */
     /* pre-terminators are .CRT$XPA...XPZ, terminators are .CRT$XTA...XTZ*/
     /* stdio destructor is at .CRT$XPX, handle I/O destructor is at XTX */
#    pragma data_seg( ".CRT$XPC" ) /* before all clib terminators */
#    pragma comment( linker, "/Merge:.CRT=.data" )
#    define __XFAR /* nothing - flat */
#    else
#    pragma data_seg( "XC" )
#    define __XFAR __far /* links are always far code/far data */
#    endif
     static void __cdecl __XFAR ANON_CODE(__SETUP_DESTRUCTOR_FOR__) (void)
     {  DTORB __SETUP_DESTRUCTOR_FOR__ (); DTORE }
     static void (__cdecl __XFAR *ANON_DATA(__SETUP_DESTRUCTOR_FOR__) ) =
                                  ANON_CODE(__SETUP_DESTRUCTOR_FOR__);
#    pragma data_seg()
#    undef __XFAR
#    undef __SETUP_DESTRUCTOR_FOR__
#  endif
#endif

/* ======================================================================== */

#undef CTORB
#undef CTORE
#undef DTORB
#undef DTORE
     
#if defined(__SETUP_DESTRUCTOR_FOR__)
#   error No destructor macro available
#endif
#if defined(__SETUP_CONSTRUCTOR_FOR__)
#   error No contructor macro available
#endif

/* ======================================================================== */

#ifndef __CTOR_DTOR_H__
#define __CTOR_DTOR_H__(x) __tkern_ ## x

#define _my_atexit   __CTOR_DTOR_H__(atexit)
#define _my_atterm   __CTOR_DTOR_H__(atterm)
#define __kern_is_exiting  __CTOR_DTOR_H__(is_exiting)

typedef void (*_my_athook_t)(void);
extern int _my_atexit(_my_athook_t fxn);
extern int _my_atterm(_my_athook_t fxn);

int __kern_is_exiting(void);

#endif /* __CTOR_DTOR_H__ */
