/*
** Debugging message helpers.
** May 2001 <cyp@fb14.uni-mainz.de>
*/
#include <share.h>    /* SH_DENYWR */
#include <fcntl.h>    /* open modes */
#include <errno.h>    /* errno, strerror() */
#include <ctype.h>    /* isprint() */
#include <io.h>       /* read() */

#include "target.h"   /* standard includes */
#include "printk.h"   /* _hprintk() */
#include "loadcfg.h"  /* _conf_*() */
#include "timer.h"    /* timer_*() */
#include "blocker.h"  /* blocker_*() */
#include "debug.h"    /* ourselves */

#undef TRACE_DBGOPT

/* ----------------------------------------------------------------- */

#if !defined(NDEBUG)
static struct
{
  int indent_lev;
  const char *file;
  unsigned int line;
  const char *func;
  int debug_fd;
  char *print_msg;
  int finalized_debug_levels;
  int generic_debug_lev;
} FAR my_statics = { 0, "", 0, "", -1, ((char *)0), 0, 0 };
#endif

/* ----------------------------------------------------------------- */

#if !defined(NDEBUG) && defined(__cplusplus)
static const char *__unmunge(const char *_func, unsigned int *lenP)
{
  register const char *func = _func;
  register unsigned int len = 0; 
  while (func[len] && func[len]!='(')
  {
    if (func[len] != ' ')
      len++;
    else
    {
      func += len+1;
      len = 0;
    }
  }
  *lenP = len;
  return func;
}
#endif

/* ----------------------------------------------------------------- */

#if !defined(NDEBUG)
static int __sel_fd(int fordebugmsg)
{
  int fd = ((fordebugmsg) ? (my_statics.debug_fd) : (1));
  if (fd != -1) /* not invalid handle */
  {
    static unsigned long dosio_squelch_count[2] = {0,0};
    int idx = ((fordebugmsg) ? 1 : 0);

    if (fd != 2 && /* not "stderr" (not direct console output) */
        blocker_get_indos_flag())
    {
      dosio_squelch_count[idx]++;
      fd = -1;
    }
    else 
    {
      blocker_critsect_begin();
      if (fd == 2 || isatty(fd))
      {
        unsigned char column = 0;
#if defined(__WATCOMC__) || defined(__TURBOC__)
        __asm mov ah,0fh
        __asm int 10h
        __asm mov ah,3
        __asm int 10h
        __asm mov column,dl
#endif
        if (column != 0)
          _hprintk( fd, "\n" );
      }
      if (dosio_squelch_count[idx])
      {
        _hprintk( fd, "*** %lu %s messages squelched while in interrupt ***\n", 
                  dosio_squelch_count[idx], (idx ? "DBG" : "KMSG" ) );
        dosio_squelch_count[idx] = 0;
      }
      if (my_statics.print_msg)
      {
        if (!fordebugmsg)
          __sel_fd(1); /* recurse with other fd */
        else
        {
          _hprintk( fd, "%s\n", my_statics.print_msg );
          free(my_statics.print_msg);
          my_statics.print_msg = (char *)0;
        }
      }
      blocker_critsect_end();
    }
  }
  return fd;
}
#endif

/* ----------------------------------------------------------------- */

#if !defined(NDEBUG)
static void __common_msg( int fd, int with_ts, int with_tag, 
                          const char *fmt, va_list args )
{
  blocker_critsect_begin();
  if (fd != -1)
  {
    if (with_ts)
    {
      unsigned long t = timer_msec_count();
      _hprintk( fd, "%lu.%03lu: ", DIV32(t,1000ul), MOD32(t,1000ul) );
    }
    if (with_tag)
    {
      if (my_statics.func)
      {
#if defined(__cplusplus)
        unsigned int len = 0; 
        const char *func = __unmunge(my_statics.func, &len );
        _hprintk( fd, "%.*s(): ", len, func );
#else   
        _hprintk( fd, "%s(): ", my_statics.func );
#endif
      }
      else
      {
        _hprintk( fd, "%s:%u: ", my_statics.file, my_statics.line );
      }
    }
    _vhprintk( fd, fmt, args);
  }
  blocker_critsect_end();
  return;
}
#endif

static void __msg(int msglev, const char *fmt, ...)
{
#if defined(NDEBUG)
  msglev = msglev;
  fmt = fmt;
#else
  va_list args;
  if (msglev > 0)
  {
    #if defined(NDEBUG)
    return;
    #else
    if (!my_statics.finalized_debug_levels)
      _conf_load();
    if (msglev > my_statics.generic_debug_lev)
      return;
    #endif
  }
  va_start (args, fmt);
  __common_msg( __sel_fd(0), (msglev >= 0), (msglev >= 0), fmt, args );
  va_end (args);
  return;
#endif
}

/* ----------------------------------------------------------------- */

void (*__wdl_msg(const char *file, int line, const char *func))
                  (int msglev, const char *fmt, ...)
{
#if defined(NDEBUG)
  file = file;
  line = line;
  func = func;
#else
  my_statics.file = file; 
  my_statics.line = line;
  my_statics.func = func;
#endif
  return __msg;
}

/* ================================================================= */
/*                           DBGPRINTF()                             */
/* ================================================================= */

#ifdef TRACE_DBGOPT
#undef TRACE_DBGOPT
#define TRACE_DBGOPT(x) _printk x
#else
#define TRACE_DBGOPT(x) NO_OPERATION()
#endif

#if !defined(NDEBUG)
static void __open_dbg_fd(const char *fn)
{
  if (!*fn || strcmp(fn,"nul")==0)
  {
    if (my_statics.debug_fd > 2)
      close(my_statics.debug_fd);
    my_statics.debug_fd = -1;
  }
  else if (stricmp(fn,"con") == 0 || stricmp(fn,"stdout")==0)
    my_statics.debug_fd = 1; /* stdout */
  else if (stricmp(fn,"stderr") == 0)
    my_statics.debug_fd = 2 /* stderr */;
  else
  {
#if defined(__DJGPP__) && !defined(sopen)
#define sopen(__fn,__access,__sh,__perm) open(__fn,__access|__sh,__perm)
#endif
    my_statics.debug_fd = sopen(fn, O_BINARY|O_WRONLY|O_CREAT|O_TRUNC,
                                   SH_DENYWR,0666);

    if (my_statics.debug_fd == -1)
    {
      _printk("ERROR: unable to open debug file '%s' (%s)\n"
              "Reverted to using stderr\n", fn, strerror(errno));
      my_statics.debug_fd = 2;    
    }
  }
  return;
}
#endif /* !NDEBUG */

/* ------------------------------------------------------------------ */

#if !defined(NDEBUG)
static struct
{
  struct {const char *name; unsigned char level; } prstates[9];
  int some_changed;

} dbg_statics = 
{
  { { "\4LINK",  0 },
    { "\3ARP" ,  0 },
    { "\2IP"  ,  0 },
    { "\4ICMP",  0 },
    { "\3UDP" ,  0 },
    { "\3TCP" ,  0 },
    { "\4IGMP",  0 },
    { "\3DNS" ,  0 },
    { "\4DHCP",  0 }
  },
  0 /* some_changed */
};
#endif 

/* ---------------------------------------------------------- */

#if !defined(NDEBUG)
static void __finalize_debug_levels(const char *dummy)
{
  dummy = dummy;
  my_statics.finalized_debug_levels = 1;

  if (dbg_statics.some_changed)
  {
    unsigned int i;
    for (i = 0; i < (sizeof(dbg_statics.prstates)/sizeof(dbg_statics.prstates[0])); i++)
    {
      if (dbg_statics.prstates[i].level != 0)
      {
        char levbuf[2];
        char nambuf[16];
        levbuf[0] = (char)('0' + dbg_statics.prstates[i].level);
        levbuf[1] = '\0';
        strcat( strcpy( nambuf, dbg_statics.prstates[i].name+1 ), ".DEBUG" );
        _conf_passopt(nambuf, levbuf);
        TRACE_DBGOPT(("%s=%s\n", nambuf, levbuf ));
      }
    }
  }
  return;
}
#endif

/* ---------------------------------------------------------- */

#if !defined(NDEBUG)
static void __set_proto_debug_lev( const char *value )
{
  char states[sizeof(dbg_statics.prstates)/sizeof(dbg_statics.prstates[0])];
  char buf[10]; /* eg, "ICMP=nn" */
  unsigned int n;  
  int setone = 0;

  for (n = 0; n < sizeof(states); n++)
    states[n] = dbg_statics.prstates[n].level;
  
  while (*value)
  {
    int len = 0;

    while ((*value >= 'a' && *value <= 'z') ||
           (*value >= 'A' && *value <= 'Z'))
    {
      if (len < (int)(sizeof(buf)-1))
        buf[len++] = (char)(*value & ~0x20);  
      value++;
    }
    buf[len] = '\0';
    if (len > 0 && len < 5)
    {
      char lev = 1;

      while (*value == ' ' || *value == '\t')
        value++;

      if (*value == '=')
      {
        value++;
        while (*value == ' ' || *value == '\t')
          value++;
        lev = 0;
        if (*value >= '0' && *value <= '9') 
          lev = (char)(*value - '0');
      }

      if (len == 3 && buf[0] == 'A' && buf[1] == 'L' && buf[2] == 'L')
      {
        setone = 1;
        for (n = 0; n < (sizeof(states)/sizeof(states[0])); n++)
          states[n] = lev;
        /* no break - may be more on line (to change level)*/
      }
      else
      {
        int subs = 0; /* "MAC"/"ETH"="LINK", "RARP"=>"ARP", "BOOT*"=>"DHCP" */
        
        if (len == 3 && ((buf[0] == 'M' && buf[1] == 'A' && buf[2] == 'C') ||
                         (buf[0] == 'E' && buf[1] == 'T' && buf[2] == 'H')))
          { subs=1; len=4; buf[0]='L'; buf[1]='I'; buf[2]='N'; buf[3]='K'; } 
        else if (len == 4 && buf[1] == 'A' && buf[2] == 'R' && buf[3] == 'P')
          { subs = 1; len = 3; buf[0] = 'A'; buf[1] = 'R'; buf[2] = 'P'; } 
        else if (len >= 4 && buf[0] == 'B' && buf[1] == 'O' && buf[2] == 'O')
          { subs=1; len=4; buf[0]='D'; buf[1]='H'; buf[2]='C'; buf[3]='P'; } 

        for (n = 0; n < (sizeof(states)/sizeof(states[0])); n++)
        {
          if (len == dbg_statics.prstates[n].name[0] && 
              memcmp(dbg_statics.prstates[n].name+1, buf, len) == 0)
          {
            if (!subs || lev > states[n])
            {
              states[n] = lev;
              setone = 1;
            }
            break;
          }
        }
      }
    } /* if (len > 0 && len < 5) */

    while (*value && !((*value >= 'a' && *value <= 'z') ||
                       (*value >= 'A' && *value <= 'Z')))
    {
      value++;
    }
  }

  if (setone)
  {
    dbg_statics.some_changed = 1;
    for (n = 0; n < sizeof(states); n++)
      dbg_statics.prstates[n].level = states[n];
    if (my_statics.finalized_debug_levels)
      __finalize_debug_levels(buf /* dummy */);
  }

  return;
}
#endif

/* ---------------------------------------------------------- */

#if !defined(NDEBUG)
static void __set_generic_debug_lev(const char *value)
{
  int lev = atoi(value);
  if (lev >= 0)
  {
    char buf[6]; 
    my_statics.generic_debug_lev = lev;

    buf[0] = 'A'; buf[1] = buf[2] = 'L'; buf[3] = '='; buf[5] = '\0';
    buf[4] = (char)((lev < 5) ? (lev + '0') : '5'); 
    __set_proto_debug_lev(buf);
  }
  return;
}
#endif

/* ------------------------------------------------------------------ */

#if !defined(NDEBUG)
static void __set_print_msg(const char *value)
{
  if (!my_statics.finalized_debug_levels)
  {
    if (my_statics.print_msg)
      free(my_statics.print_msg);
    my_statics.print_msg = (char *)0;
    if (*value)
      my_statics.print_msg = strdup(value);
  }
  return;
}
#endif

/* ------------------------------------------------------------------ */

#if !defined(NDEBUG)
static void __ctor_debug(void)
{
  static const struct config_table FAR my_cfg[] = {
       { "DEBUG",       ARG_FUNC,     (void FAR *)__set_generic_debug_lev },
       { "DEBUG.PROTO", ARG_FUNC,     (void FAR *)__set_proto_debug_lev },
       { "DEBUG.FILE",  ARG_FUNC,     (void FAR *)__open_dbg_fd  },
       { "PRINT",       ARG_FUNC,     (void FAR *)__set_print_msg },
       { "#",           ARG_FUNC,     (void FAR *)__finalize_debug_levels },
       { ((const char FAR *)0), 0,    ((void FAR *)0)  }
  };
  _conf_add_confextn( &my_cfg[0], NULL );
  return;
}

#define __SETUP_CONSTRUCTOR_FOR__ __ctor_debug
#include "ctordtor.h"
#endif /* !NDEBUG */

/* ------------------------------------------------------------------ */

static void __do_dbg_msg(const char *fmt, ...)
{
#if !defined(NDEBUG)
  va_list args;
  int indentlev = 0;
  int with_tag = 1;

  if (*fmt == '*')
  {
    with_tag = 0;
    fmt++;
  }    

  if (*fmt == '-' || *fmt == '+')
    indentlev = *fmt++;

  if (*fmt == '*')
  {
    with_tag = 0;
    fmt++;
  }    


  if (indentlev == '+')
    my_statics.indent_lev++;

  va_start (args, fmt);
  if (*fmt)
    __common_msg( __sel_fd(1), 1, with_tag, fmt, args );
  va_end (args);

  if (indentlev == '-')
    my_statics.indent_lev--;
#else
  fmt = fmt;
#endif /* !NDEBUG */
  return;
}

/* ----------------------------------------------------------------- */
  
void (*__dbg_msg(const char *file,int line,const char *func))
                        ( const char *fmt, ...)
{
#if defined(NDEBUG)
  file = file;
  line = line;
  func = func;
#else
  my_statics.file = file;
  my_statics.line = line;
  my_statics.func = func;
#endif /* NDEBUG */
  return __do_dbg_msg;
}

/* ================================================================= */

static void __do_dbg_dmp(const char *label, unsigned short pktid,
                         const void FAR *_pkt, unsigned int alen )
{
#if !defined(NDEBUG)
  int indentlev = 0;
  int fd, nofln = 0;

  blocker_critsect_begin();

  if (!label)
    label = "";

  if (*label == '*')
    nofln = *label++;

  if (*label == '-' || *label == '+')
    indentlev = *label++;

  if (*label == '*')
    nofln = *label++;

  if (indentlev == '+')
    my_statics.indent_lev++;

  fd = __sel_fd(1);

  if (_pkt && alen && fd != -1)
  {
    const char FAR *pkt = (const char FAR *)_pkt;
    unsigned int i;

    unsigned long t = timer_msec_count();
    _hprintk( fd, "%lu.%03lu: ", DIV32(t,1000ul), MOD32(t,1000ul) );

    if (!nofln)
    {
      if (my_statics.func)
      {
#if defined(__cplusplus)
        unsigned int len = 0; 
        const char *func = __unmunge(my_statics.func, &len );
        _hprintk( fd, "%.*s(): ", len, func );
#else   
        _hprintk( fd, "%s(): ", my_statics.func );
#endif
      }
      else
      {
        _hprintk( fd, "%s:%u: ", my_statics.file, my_statics.line );
      }
    }

    if (!*label)
      label = "???";
    if (pktid)
      _hprintk(fd, "%s-%04x dump - length: 0x%04x (%u)\n", label, pktid, alen, alen );
    else
      _hprintk(fd, "%s dump - length: 0x%04x (%u)\n", label, alen, alen );

    for (i = 0; i < alen; i += 16)
    {
      char buffer[128];
      char *p, *q; unsigned int n;
      buffer[0] = '\0';
      q = 48 + (p = &buffer[strlen(buffer)]);
      for (n = 0; n < 16; n++)
      {
        unsigned int c = ' ';
        p[0] = p[1] = ' ';
        if (( n + i ) < alen )
        {
          static const char *tox="0123456789abcdef";
          c = (((unsigned int)pkt[n+i]) & 0xff);
          p[0] = (char)tox[c>>4];
          p[1] = (char)tox[c&0x0f];
          if (((unsigned char)c) <= 31 /*iscntrl(c)*/ || !isprint(c))
            c = '.';
        }
        p+=2;
        *p++ = ' ';
        *q++ = (char)c;
      }
      *q = '\0';

      _hprintk(fd, "        %04x: %s\n", i, buffer );
    }
  }
  blocker_critsect_end();

  if (indentlev == '-')
    my_statics.indent_lev--;
#else
  label = label; 
  pktid = pktid;
  _pkt = _pkt;
  alen = alen;
#endif /* !NDEBUG */
  return;
}

void (*__dbg_dmp(const char *file,int line,const char *func))
                 (const char *label, unsigned short pktid,
                 const void FAR *pkt, unsigned int len )
{
#if !defined(NDEBUG)
  my_statics.file = file;
  my_statics.line = line;
  my_statics.func = func;
#else
  file = file;
  line = line;
  func = func;
#endif /* NDEBUG */
  return __do_dbg_dmp;
}

/*  =================================================================== */
/*                        time-to-string conversion                     */
/*  =================================================================== */

#if !defined(NDEBUG) /* not used unless debugging */
static char *__n2a( char *p, int numdig, unsigned int num )
{
  while (num)
  {
    *--p = (char)('0' + (num % 10u));
    num /= 10u;
    numdig--;
  }
  while (numdig > 0)
  {
    *--p = '0';
    numdig--;
  }
  return p;
}
#endif

#if !defined(NDEBUG) /* not used unless debugging */
const char *__dbg_t2a( unsigned long t, int scale, 
                       char *buffer, unsigned int buflen )
{
  char buf[sizeof(t)*4];
  char *p = &buf[sizeof(buf)];
  *--p = '\0';
  
  if (scale != 1)
  {
    p = __n2a(p, ((scale == 1000) ? 3 : ((scale == 100) ? 2 : 1)), 
                 (unsigned int)MOD32(t,scale)); 
    t = DIV32(t, scale);
    *--p = '.';
  }
  p = __n2a(p, 2, (unsigned int)MOD32(t, 60));
  t = DIV32(t, 60);
  *--p = ':';    
  p = __n2a(p, 2, (unsigned int)MOD32(t, 60));
  t = DIV32(t, 60);
  *--p = ':';    
  p = __n2a(p, 2, (unsigned int)t);

  strncpy( buffer, p, buflen );
  buffer[buflen-1] = '\0';
  return buffer;
}
#endif

/* =================================================================== */
/*                    Is any module being debugged?                    */
/* =================================================================== */

#if 0 /* not used */
int __ima_debugee(void) 
{ 
  /* we aren't if THIS module was compiled with NDEBUG */
  /* (because no debug I/O would be done) */
  #if defined(NDEBUG)
  return 0;
  #else
  return 1;
  #endif
}
#endif
