/*
 * Domain Name Server protocol
 *
 *  1.0 : May 10, 2001 - Partial rewrite - Cyrus Patel <cyp@fb14.uni-mainz.de>
 *                       Integrated reverse lookup and get{host/dom}stuff;
 *                       Added debug, multi-addr lookup, dynamic get{host|dom},
 *                       and static resolve; general clean up.
 *  0.3 : Jun 16, 1997 - calling usr_yield (i.e. system_yield) while resolving
 *  0.2 : Apr 24, 1991 - use substring portions of domain
 *  0.1 : Mar 18, 1991 - improved the trailing domain list
 *  0.0 : Feb 19, 1991 - pirated by Erick Engelke
 * -1.0 :              - NCSA code
 */

#include "target.h"   /* standard types/includes */
#include "debug.h"    /* KMSG(), DBGPRINTF() */
#include "loadcfg.h"  /* _conf_*() */
#include "inaddr.h"   /* _my_aton() */
#include "ipbase.h"   /* _ip_get_ipaddr() */
#include "userland.h" /* sock_*() */
#include "dns.h"      /* ourselves */

#define DOMSIZE          512  /* maximum domain message size to mess with */
#define DOM_DST_PORT     53   /* destination port number for DNS protocol */

/* The DOMAINSLIST option is screwed up. According to wat-a-poo docs,
 * setdomainname() defines *THE* domain that is searched.
 *
 * The BSD meaning of setdomainname() is simply... set the domain name.
 * (ie the domain part of the host name) and has nothing whatsoever to do 
 * with name resolution.
*/
#undef HAVE_DOMAINSRCHLIST

/* ======================================================================= */
/*                            private config data                          */
/* ======================================================================= */

#define MAXNAMESIZE 64   /* (RFC) maximum host name (including '\0') */
#define MAXDNAME    256  /* see below (including '\0') */
/* I have no idea where the MAXDNAME = 1025 in BIND (nameser.h) comes from.
 * (modifications for IPv6 perhaps?). A comment in resolv.h states that the
 * total size of __res_state is 512 bytes, which is impossible when one of 
 * its members is 1025 bytes in size.
 *
 * RFC833 PAGE 30:
 * <quote>
 * To simplify implementations, the total length of label octets and
 * label length octets that make up a domain name is restricted to
 * 255 octets or less.  Since the trailing root label and its dot are
 * not printed, printed domain names are 254 octets or less.
 * </quote>
*/

static struct
{
  uint32_t  based_ip_addr;    /* dom_name and host_name reflect this address */
  char   FAR *dom_name;
  char   host_name[MAXNAMESIZE]; /* don't put junk here - must be resolvable! */

#ifdef HAVE_DOMAINSRCHLIST
  const char *srchdom_list[4];
  int   srchdom_count;
#endif

  uint32_t  ns_list[6];        /* nameserver_list */
  unsigned  ns_count;          /* count of entries in nameserver_list */ 
  int       res_timeout;
  int       res_try_subdoms; /* previously misnamed 'recurse' */
  int       conf_load_done;
  int       debug_lev;

  char      *db_filename;

} FAR my_statics; /* all variables have a zero default */

/* ----------------------------------------------------------------------- */

/* checkpoint all read accesses from the host/domain name.
 * and ensure that they reflect the correct values for the 
 * current IP address.
*/
static void __chkpt_read_from_hostdomname(void)
{
  uint32_t based_ip_addr = _ip_get_ipaddr();
  if (my_statics.based_ip_addr != based_ip_addr)
  {
    char myname[80]; 
    if (_dns_resbyaddr( based_ip_addr, myname, sizeof(myname) ) > 0)
    {
      if (my_statics.debug_lev)
        DBGPRINTF(("My FQDN: %s\n", myname)); /* informational */
    }
    else /* res failed */
    {
      if (my_statics.debug_lev)
        DBGPRINTF(("Pout. Reverse resolve for myself failed.\n"));
    }
    /* update my_statics.based_ip_addr even on failure */
    my_statics.based_ip_addr = based_ip_addr;
  }
  return;
}

/* ======================================================================= */
/*                            get/set host name                            */
/* ======================================================================= */

#define __is_valid_dnsname_char(__c) \
        ( (__c == '.') || (__c == '-') ||  (__c >= '0' && __c <= '9') || \
          (__c >= 'A' && __c <= 'Z')   ||  (__c >= 'a' && __c <= 'z') )

int _dns_set_hostname( const char *name, int len )
{
  unsigned int pos;

  if (!name || len < 1)
  {
    errno = EINVAL;
    return -1;
  }    
  else
  {
    pos = 0; 
    while (pos < ((unsigned int)len) && name[pos] != '.')
      pos++;
    len = pos;
  }

  if (((unsigned int)len) >= sizeof(my_statics.host_name)) /* MAXNAMESIZE */
  {
    #if defined(EMSGSIZE)
    errno = EMSGSIZE;
    #else
    errno = EINVAL; 
    #endif
    return -1;
  }

  pos = 0;
  while (pos < ((unsigned int)len) )
  {               
    char c = name[pos++];
    if ( c == '.' ||
         !__is_valid_dnsname_char(c) || /* invalid char */
       (c == '-' && (pos == 0 || name[pos] == '.' || name[pos] == '\0')) )
    {
      errno = EINVAL;
      return -1;
    }
  }

  if (my_statics.based_ip_addr != 0) /* already done once */
    my_statics.based_ip_addr = _ip_get_ipaddr(); /* update */
  else /* make sure config is loaded */
    _conf_load();

  FAR_MEMCPY( my_statics.host_name, name, len );
  my_statics.host_name[len] = '\0';

  return 0;
}
  
/* -------------------------------------------------------------------- */
      
int _dns_get_hostname( char *buf, int buflen, int bsd_pure )
{
  if (!buf || buflen < 0)
  {
    errno = EINVAL;
    return -1;
  }
  else if (buflen == 0)
    ; /* nothing */
  else if (buflen == 1 && !bsd_pure)
    buf[0] = '\0';
  else
  {
    unsigned int i;
    __chkpt_read_from_hostdomname();

    i = FAR_STRLEN(my_statics.host_name)+1;
    if (i < ((unsigned int)buflen))
      buflen = i;
    FAR_MEMCPY( buf, my_statics.host_name, buflen );
    if (!bsd_pure)
      buf[buflen-1] = '\0';
  }
  return 0;
}

/* ======================================================================= */
/*                            get/set domain name                          */
/* ======================================================================= */

static int __is_valid_domain_name( const char *name, unsigned int len,
                                   int *errnoP )
{
  unsigned int complen, pos;

  if (len > MAXDNAME)
  {
    if (errnoP)
    {
      #if defined(EMSGSIZE)
      *errnoP = EMSGSIZE;
      #else
      *errnoP = EINVAL; 
      #endif 
    }
    return 0;
  }

  complen = pos = 0;
  while (pos < len)
  {     
    char c = name[pos++];
    if ( complen > 63                   /* segment len > 63 */
      || !__is_valid_dnsname_char(c)    /* invalid char */
      || (c == '-' && (complen == 0 || name[pos] == '.' || name[pos] == '\0'))
      || (c == '.' && complen == 0) )   /* two dots back to back */
    {
      if (errnoP)
        *errnoP = EINVAL;
      return 0;
    }
    if (c == '.')
      complen = 0;
  }

  return 1;
}

/* -------------------------------------------------------------------- */

int _dns_set_domname( const char *name, int len )
{
  if (len < 0)
  {
    errno = EINVAL;
    return -1;
  }
  if (!name)
    len = 0;
  else
  {
    while (len > 0 && *name == '.')
    {
      name++;
      len--;
    }
  }

  if (len > 0)
  {
    if (!__is_valid_domain_name( name, len, &errno ))
      return -1; /* errno is already set */
  }

  if (my_statics.based_ip_addr != 0) /* IP is up */
    my_statics.based_ip_addr = _ip_get_ipaddr();
  else /* make sure config is loaded */
    _conf_load();

  if (len == 0)
  {
    if (my_statics.dom_name)
    {
      FAR_FREE(my_statics.dom_name);
      my_statics.dom_name = NULL;
    }
  }
  else
  {
    char FAR *p = (char FAR *)FAR_REALLOC(my_statics.dom_name, len+1 );
    if (!p) /* ENOMEM */
      return -1;

    my_statics.dom_name = p;
    FAR_MEMCPY( p, name, len );
    p[len] = '\0';
  }

  return (0);
}

/* -------------------------------------------------------------------- */

int _dns_get_domname( char *buf, int buflen, int bsd_pure )
{
  if (!buf || buflen < 0)
  {
    errno = EINVAL;
    return -1;
  }
  if (buflen == 0)
    ; /* nothing */
  else 
  {
    __chkpt_read_from_hostdomname();

    if (!my_statics.dom_name)
      *buf = '\0';
    else
    { 
      char FAR *p = my_statics.dom_name;
      int len = 0;
      while (p[len])
        len++;
      len++; /* include terminating '\0' */
      if (len > buflen)
        len = buflen;
      FAR_MEMCPY( buf, my_statics.dom_name, len );
      if (!bsd_pure)
        buf[buflen-1] = '\0';
    }
  }
  return 0;
}

/* ======================================================================= */
/*                resolution using /etc/hosts (or whatever)                */
/* ======================================================================= */

#include <io.h>       /* read() */
#include <share.h>    /* SH_DENYWR */
#include <fcntl.h>    /* open modes */

static int __open_file(const char *filename)
{
#if defined(__DJGPP__) && !defined(sopen)
  #define sopen(__fn,__access,__sh,__perm) open(__fn,__access|__sh,__perm)
#endif
#ifndef O_NOINHERIT
#define O_NOINHERIT 0
#endif
  return sopen(filename, O_RDONLY|O_BINARY|O_NOINHERIT, SH_DENYWR, 0600);
}

/* ----------------------------------------------------------------------- */

static int __read_tokens( int fd, char *buf, unsigned int buflen,
                          char *tokens[], int numtoks)
{
  int foundtoks = -1; /* assume EOF */
  unsigned int len = read( fd, buf, buflen );  

  if (((int)len) > 0) /* not -1, not 0 */
  {
    unsigned int pos;

    for (pos = 0; pos < len && (buf[pos] != '\r' && buf[pos] != '\n'); pos++)
      pos++;

    if (pos < len)
    {
      unsigned newlen = pos;
      while (pos < len && (buf[pos] == '\r' || buf[pos] == '\n'))
        pos++;
      if (pos == len)
        lseek( fd, -((long)(len-newlen)), SEEK_CUR );
      else
        lseek( fd, -((long)(len-pos)), SEEK_CUR );
      len = newlen;
    }
    else
    {
      for (;;)
      {
        char c;
        if (read( fd, &c, 1 ) < 1)
          break;
        if (c == '\r' || c == '\n')
        {
          lseek( fd, -1L, SEEK_CUR );
          break;
        }
      }
    }
    buf[len-1] = '\0';
  }

  if (((int)len) > 0) /* not -1, not 0 */
  {
    foundtoks = 0;
    while (*buf && *buf != '#' && *buf != '\n')
    {
      char *p;
      while (*buf == ' ' || *buf == '\t' || *buf == ',' || *buf == ';')
        buf++;
      p = buf;
      while (*buf && *buf != '#' && *buf != '\r' && *buf != '\n' && 
             *buf != ' ' && *buf != '\t' && *buf != ',' && *buf != ';')
        buf++;
      if (p != buf)
      {
        char c = *buf;
        *buf++ = '\0';
        tokens[foundtoks++] = p;
        if (foundtoks == numtoks || !c || c == '#' || c == '\r' || c == '\n')
          break;
      }
    }
  }
  return foundtoks;
}

/* ----------------------------------------------------------------------- */

#if defined(__GNUC__)
#define MY_STRCASECMP strcasecmp
#elif (UINT_MAX > USHRT_MAX)
#define MY_STRCASECMP strcmpi
#else
#define MY_STRCASECMP _fstricmp
#endif
#if 0 /* in case this is ever needed */
static int MY_STRCASECMP( register const char FAR *s1, 
                          register const char FAR *s2 )
{
  unsigned char c1, c2;
  for (;;)
  {
    c1 = (unsigned char)*s1;
    c2 = (unsigned char)*s2;
    if (c1 >= 'A' && c1 <= 'Z' && c2 >= 'A' && c2 <= 'Z')
    {
      c1 += (unsigned char)('a' - 'A');
      c2 += (unsigned char)('a' - 'A');
    }
    if (!c1 || c1 != c2)
      break;
    s1++;
    s2++;
  }
  return (c1 - c2);
}
#endif

/* ----------------------------------------------------------------------- */
  
static int __static_resolve(const char *hostname, /* resolve by name */
                            uint32_t hostaddr,       /* resolve by addr */
                            void *result_buf, unsigned int result_buf_len )
{                      /* addresses are received and returned IN HOST ORDER */
  char buf [128];
  int found = 0, fd = -1;
  unsigned int i;

  if (my_statics.db_filename)
    fd = __open_file(my_statics.db_filename);

  if (fd == -1)
  {
    const char *p = getenv("OS");
    if (p && strcmp(p, "Windows_NT")==0)
      p = getenv("SystemRoot");
    if (!p)
      p = getenv("windir");
    if (p)
    {
      /* On Win95/98 use `hosts = $(WINDIR)\system\hosts' */
      /* On WinNT/2K use `hosts = $(WINDIR)\system32\drivers\etc\hosts' */
      const char *places[] = { "\\system\\hosts", 
                               "\\system32\\drivers\\etc\\hosts" };
      for (i = 0; i < (sizeof(places)/sizeof(places[0])); i++)
      {
        if ((strlen(p) + strlen(places[i])) < sizeof(buf))
        {
          fd = __open_file( strcat( strcpy( buf, p ), places[i] ));
          if (fd != -1)
            break;
        }
      }
    }
  }

  if (fd != -1)
  {
    int numtoks = 0;
    while (numtoks >= 0)
    {
      uint32_t ina;
      char *tokens[16]; 

      numtoks = (sizeof(tokens)/sizeof(tokens[0]));
      numtoks = __read_tokens( fd, buf, sizeof(buf), &tokens[0], numtoks );

      if (numtoks >= 1 && _my_aton(tokens[0], &ina))
      {
        if (hostname) /* resolve by name */
        {
          for (i = 1; i < ((unsigned int)numtoks); i++)
          {
            /* there is no domain-ification done for names
             * from /etc/hosts, so this is really trivial.
            */
            if (MY_STRCASECMP( tokens[i], hostname ) == 0)
            {
              *((uint32_t *)result_buf) = ina;
              found = 1;
              break;
            }
          }
        }
        else if (hostaddr == ina) /* resolving by addr */
        {
          if (ina == _ip_get_ipaddr())
          {
            char *p = tokens[1];
            i = 0;
            my_statics.based_ip_addr = ina;
            while (p[i] && p[i] != '.')
              i++;
            if (p[i] == '.')
              _dns_set_domname(&(p[i+1]), strlen(&(p[i+1])));
            p[i] = '\0'; 
            _dns_set_hostname( p, strlen(p) );
          }  
          for (i = 1; i < ((unsigned int)numtoks); i++)
          {
            if (strlen(tokens[i]) < result_buf_len)
            {
              strcpy( (char *)result_buf, tokens[i] );
              found = 1;
              break;
            }
          }
          if (!found)
          {
            strncpy( (char *)result_buf, tokens[1], result_buf_len );
            *(((char *)result_buf)+(result_buf_len-1)) = '\0';
            found = 1;
          }
          break;
        }
      } /* if (numtoks >= 1 && _my_aton(tokens[0], &ina)) */
    } /* while (numtoks >= 0) */

    close(fd);
  } /* if (fd != -1) */

  return found;
}    

/* ======================================================================= */
/*                          subsystem configuration                        */
/* ======================================================================= */

static void __conf_opt_NAMESERVER(const char *value)
{
  uint32_t ina;
  if (_my_aton( value, &ina ))
  {
    if (ina && ina != 0xfffffffful)
    {
      unsigned int i;
      for ( i = 0; i < my_statics.ns_count; i++)
      {
        if (my_statics.ns_list[i] == ina)
          return; /* already there */
      }
      if (my_statics.conf_load_done == 1) /* conf_load has been done */
      {   
        /* discard all addresses that came from the .cfg. We now have
         * addresses from an authenticated source.
        */
        my_statics.conf_load_done = 2;
        my_statics.ns_count = 0;
      }
      if (my_statics.ns_count < 
           (sizeof(my_statics.ns_list)/sizeof(my_statics.ns_list[0])) )
      {
        my_statics.ns_list[my_statics.ns_count++] = ina;
      }
    }
  }
  return;
}

/* ----------------------------------------------------------------------- */

static void __conf_opt_HOSTNAME(const char *value)
{
  unsigned int hlen = 0;
  const char *hname = value;
  
  while (*value && *value != '.')
  {
    hlen++;
    value++;
  }
  if (_dns_set_hostname(hname, hlen) == 0)
  {
    while (*value == '.')
      value++;
    if (*value)
      _dns_set_domname(value, strlen(value));
  }
  return;
}  

/* ----------------------------------------------------------------------- */

static void __conf_opt_DOMAINNAME(const char *value)
{
  _dns_set_domname(value, strlen(value));
}

/* ----------------------------------------------------------------------- */

static void __conf_opt_DOMAINSLIST(const char *value)
{
#ifndef HAVE_DOMAINSRCHLIST
  __conf_opt_DOMAINNAME(value);
#else
  if (!my_statics.dom_name)
    __conf_opt_DOMAINNAME(value);
  else if (my_statics.srchdom_count < 
          (sizeof(my_statics.srchdom_list)/sizeof(my_statics.srchdom_list[0])))
  {
    char *p, *v;
    unsigned int i, l1 = strlen(value);
    if (!l1)
      return;
    if (!__is_valid_domain_name( value, l1, (int *)0 ))
      return;
    v = strdup(value);
    if (!v)
      return; 

    for (p = v; *p; p++)
      *p = (char)tolower(*p);

    p = (char *)0;
    for (i = 0; !p && i < my_statics.srchdom_count; i++)
    {      
      unsigned int l2 = strlen(my_statics.srchdom_list[i]);
      if (l1 == l2)
      {
        if ( strcmp( my_statics.srchdom_list[i], v ) == 0 )
          p = v;
      }
      else 
      {    
        const char *q = v;
        if (l1 < l2)
        {
          q = my_statics.srchdom_list[i];
          p = strstr( my_statics.srchdom_list[i], v );
        }
        else
        {
          q = v;
          p = strstr( v, my_statics.srchdom_list[i] );
        }
        if (p)
        {
          if (!((p == q || *(p-1) == '.') && (p[l1] == '\0' || p[l1] == '.')))
            p = (char *)0;
          else if (l1 > l2)
          {
            p = my_statics.srchdom_list[i];
            my_statics.srchdom_list[i] = v;
            v = p;
          }
        }
      }
    }

    if (p)
      free(v);
    else
      my_statics.srchdom_list[my_statics.srchdom_count++] = v;
  }
#endif
  return;
}

/* ----------------------------------------------------------------------- */

static void __set_etc_fn(const char *value)
{
  int fd;
  if (my_statics.db_filename)
  {
    free(my_statics.db_filename);
    my_statics.db_filename = (char *)0;
  }
  fd = __open_file(value);
  if (fd != -1)
  {
    close(fd); 
    my_statics.db_filename = strdup(value);
  }
  return;
}

/* ----------------------------------------------------------------------- */

static void __set_conf_load_done(const char *unused)
{
  unused = unused;
  my_statics.conf_load_done = 1;
}

static void __ctor_dns_res(void)
{
  static const struct config_table FAR my_cfg[] = {
     { "HOSTNAME",      ARG_FUNC,            (void FAR *)__conf_opt_HOSTNAME },
     { "DOMAINNAME",    ARG_FUNC|ARG_MULTI,  (void FAR *)__conf_opt_DOMAINNAME },
     { "NAMESERVER",    ARG_FUNC|ARG_MULTI,  (void FAR *)__conf_opt_NAMESERVER },
     { "DOMAINSLIST",   ARG_FUNC,            (void FAR *)__conf_opt_DOMAINSLIST },
     { "DOMAIN_LIST",   ARG_FUNC,            (void FAR *)__conf_opt_DOMAINSLIST },
     { "DOMAINTO",      ARG_ATOI,            (void FAR *)&(my_statics.res_timeout) },
     { "DOMAIN_TO",     ARG_ATOI,            (void FAR *)&(my_statics.res_timeout) },
     { "DOMAIN_RECURSE",ARG_ATOI,            (void FAR *)&(my_statics.res_try_subdoms) },
     { "DNS.DEBUG",     ARG_ATOI,            (void FAR *)&(my_statics.debug_lev) },
     { "DEBUG.DNS",     ARG_ATOI,            (void FAR *)&(my_statics.debug_lev) },
     { "HOSTS",         ARG_FUNC,            (void FAR *)__set_etc_fn },
     { "#",             ARG_FUNC,            (void FAR *)__set_conf_load_done },
     { ((const char FAR *)0), 0,    ((void FAR *)0)  }
  };
  FAR_MEMSET( &my_statics, 0, sizeof(my_statics));
  _conf_add_confextn( &my_cfg[0], NULL );
  return;
}

#define __SETUP_CONSTRUCTOR_FOR__ __ctor_dns_res
#include "ctordtor.h"

/* ======================================================================= */
/*                         resolution using DNS                            */
/* ======================================================================= */

/*
 *  flag masks for the flags field of the DOMAIN header
*/
#define DQR       0x8000    /* query = 0, response = 1  */
#define DOPCODE   0x7800    /* opcode, see below        */
#define DAA       0x0400    /* Authoritative answer     */
#define DTC       0x0200    /* Truncation, response was cut off at 512 */
#define DRD       0x0100    /* Recursion desired        */
#define DRA       0x0080    /* Recursion available      */
#define DRCODE    0x000F    /* response code, see below */

/*
 * opcode possible values:
 */
#define DOPQUERY  0         /* a standard query */
#define DOPIQ     1         /* an inverse query */
#define DOPCQM    2         /* a completion query, multiple reply */
#define DOPCQU    3         /* a completion query, single reply   */
/* the rest reserved for future */

/*
 * legal response codes:
 */
#define DROK      0         /* okay response */
#define DRFORM    1         /* format error  */
#define DRFAIL    2         /* their problem, server failed           */
#define DRNAME    3         /* name error, we know name doesn't exist */
#define DRNOPE    4         /* no can do request                      */
#define DRNOWAY   5         /* name server refusing to do request     */

#define DTYPE_A      1      /* host address resource record (A) */
#define DTYPE_NS     2      /* authoritative server */
#define DTYPE_CNAME  5      /* canonical name */
#define DTYPE_PTR   12      /* domain name ptr (PTR)            */
#define DTYPE_HINFO 13      /* host information */
#define DTYPE_MX    15      /* mail exchange */
#define DTYPE_AXFR  252     /* transfer zone of authority */
#define DTYPE_ANY   255     /* wildcard match */

#define DCLASS_IN    1      /* ARPA internet class                         */
#define DCLASS_ANY 255      /* wildcard for several of the classifications */

/*
 *  Header for the DOMAIN queries
 *  ALL OF THESE ARE BYTE SWAPPED QUANTITIES!
 *  We are the poor slobs who are incompatible with the world's byte order
 *                                  1  1  1  1  1  1 
 *    0  1  2  3  4  5  6  7  8  9  0  1  2  3  4  5 
 *  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *  |                      ID                       |
 *  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *  |QR|   Opcode  |AA|TC|RD|RA|        |   RCODE   |
 *  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *  |                    QDCOUNT                    |
 *  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *  |                    ANCOUNT                    |
 *  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *  |                    NSCOUNT                    |
 *  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 *  |                    ARCOUNT                    |
 *  +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
 */
#pragma pack(1) /* pack struct members on byte boundaries */
struct dhead {
       uint16_t  ident;         /* unique identifier              */
       uint16_t  flags;
       uint16_t  qdcount;       /* question section, # of entries */
       uint16_t  ancount;       /* answers, how many              */
       uint16_t  nscount;       /* count of name server RRs       */
       uint16_t  arcount;       /* number of "additional" records */
     };
#pragma pack() /* restore default packing */

/*
 *  a resource record is made up of a compressed domain name followed by
 *  this structure.  All of these words need to be byteswapped before use.
 */
#pragma pack(1) /* pack struct members on byte boundaries */
struct rrpart {
       uint16_t   rtype;          /* resource record type = DTYPE_A   */
       uint16_t   rclass;         /* RR class = DCLASS_IN             */
       uint32_t  ttl;            /* time-to-live, changed to 32 bits */
       uint16_t   rdlength;       /* length of next field             */
       uint8_t   rdata[DOMSIZE]; /* data field                       */
     };
#pragma pack() /* restore default packing */

/*
 *  data for domain name lookup
 */
#pragma pack(1) /* pack struct members on byte boundaries */
struct useek {
       struct dhead h;
       uint8_t   x[DOMSIZE];
     };
#pragma pack() /* restore default packing */

/* ======================================================================= */
/*                              DNS packet debug                           */
/*   based on code originally by Mike Borella <mike_borella@mw.3com.com>   */
/* ======================================================================= */

#if !defined(NDEBUG)
static const char *__q_type(uint16_t type)
{
  switch (type)
  {
    case DTYPE_A:     return "A";
    case DTYPE_NS:    return "NS";
    case DTYPE_CNAME: return "CNAME";
    case DTYPE_PTR:   return "PTR";
    case DTYPE_HINFO: return "HINFO"; /* host info */
    case DTYPE_MX:    return "MX";    /* mail exchange */
    case DTYPE_AXFR:  return "AXFR";  /* transfer zone of authority */
    case DTYPE_ANY:   return "ANY";   /* wildcard match */
    default:          type = 0;
  }
  return "??";
}  
#endif

/* ------------------------------------------------------------------- */

#if defined(NDEBUG)
#define __debug_dump_pkt(__qa, __l) NO_OPERATION()
#else
static unsigned int unpackdom (uint8_t *dst, const uint8_t *src, const uint8_t *buf);
static void __debug_dump_pkt(const struct useek *questans, unsigned length)
{
  static const char *resp_names[] = {
            "no error",   "format error",    "server error",
            "name error", "not implemented", "service refused"
          };
  static char *op_names[] = {
            "standard", "inverse", "server status"
          };
  uint16_t   flags, part, recno, reccount, qtype, qclass, opcode;
  uint8_t   buf[MAXDNAME+4];
  const  uint8_t *p;

  if (length < sizeof(questans->h))
    return;

  p      = questans->x;
  flags  = NTOHS(questans->h.flags);
  opcode = (uint16_t)((flags & DOPCODE) >> 11);

  DBGPRINTF(("*DNS debug dump: (debug_lev=%d)\n", my_statics.debug_lev));
  DBGPRINTF(("*    ident %u, type = %s, opcode %u (%s)\n", 
              NTOHS(questans->h.ident),
              (flags & DQR) ? "Response" : "Query",
              opcode, (opcode < (sizeof(op_names)/sizeof(op_names[0])) 
                                       ? op_names[opcode] : "??") ));
  DBGPRINTF(("*    auth answer %d, truncated %d, rec-req %d, rec-avail %d\n",
              ((flags & DAA)!=0), ((flags & DTC)!=0), 
              ((flags & DRD)!=0), ((flags & DRA)!=0) ));
  DBGPRINTF(("*    response code: %u (%s)\n",
              (flags & DRCODE), 
              ((flags & DRCODE) < (int)(sizeof(resp_names)/sizeof(resp_names[0])) ?
                                   resp_names[(flags & DRCODE)] : "??")  ));

  /* Do the question part of the packet. */
  reccount = NTOHS(questans->h.qdcount);
  for (recno = 0; recno < reccount; recno++ )
  {
    DBGPRINTF(("*Question:\n"));
    p += unpackdom( buf, p, (const uint8_t *)questans );
    DBGPRINTF(("*    query name: '%s'\n", buf ));

    qtype  = NTOHS(*(const uint16_t *)p);  p += sizeof(qtype);
    qclass = NTOHS(*(const uint16_t *)p);  p += sizeof(qclass);

    DBGPRINTF(("*    query type %d (%s), query class %d\n",
                     qtype, __q_type(qtype), qclass));
  }

  for (part = 0; part < 3; part++)
  {
    const char *partname = (const char *)0;
    reccount = 0;
    switch (part)
    {
      default:/* dummy */ 
      case 0: /* dump the resource records for the answers */
              partname="Answer"; reccount = NTOHS(questans->h.ancount); 
              break;
      case 1: /* dump the resource records for the name servers */
              partname="Authority"; reccount = NTOHS(questans->h.nscount); 
              break;
      case 2: /* dump the resource records for the additional info */
              partname="Additional"; reccount = NTOHS(questans->h.arcount); 
              break;
    }
    DBGPRINTF(("*%s record count: %u\n", partname, reccount ));
    for (recno = 0; recno < reccount; recno++ )
    {
      uint32_t ttl;
      uint16_t  reslen;

      DBGPRINTF(("*%s #%u:\n", partname, recno+1 ));

      p += unpackdom( buf, p, (const uint8_t *)questans );
      DBGPRINTF(("*    name: '%s'\n", buf ));

      /* Do query type, class, ttl and resource length */
      qtype  = NTOHS(*(const uint16_t *)p);  p += sizeof(qtype);
      qclass = NTOHS(*(const uint16_t *)p);  p += sizeof(qclass);
      ttl    = NTOHL(*(const uint32_t *)p);  p += sizeof(ttl);
      reslen = NTOHS(*(const uint16_t *)p);  p += sizeof(reslen);

      DBGPRINTF(("*    type %u (%s), class %u, ttl %s, rdlen %u\n",
                 qtype, __q_type(qtype), qclass, 
                 DBG_T2A( ttl, 0, ((char *)&buf[0]), sizeof(buf) ), reslen ));

      /* Print the contents of a resource record */
      switch (qtype)
      {
        case DTYPE_A:     /* A record; ip address(es) */
             {
               uint16_t offs;
               for (offs = 1; offs <= reslen; offs += sizeof(uint32_t))
               {
                 DBGPRINTF(("*    DTYPE_A: %I\n", *(const uint32_t *)p ));
                 p += sizeof(uint32_t);
               }
             }
             break;

        case DTYPE_NS:    /* NS record; Name Server */
             p += unpackdom( buf, p, (const uint8_t *)questans );
             DBGPRINTF(("*    DTYPE_NS: '%s'\n", buf ));
             break;

        case DTYPE_CNAME: /* CNAME record; canonical name */
             p += unpackdom( buf, p, (const uint8_t *)questans );
             DBGPRINTF(("*    DTYPE_CNAME: '%s'\n", buf ));
             break;

        case DTYPE_PTR:   /* PTR record; hostname for IP */
             p += unpackdom( buf, p, (const uint8_t *)questans );
             DBGPRINTF(("*    DTYPE_PTR: '%s'\n", buf ));
             break;

        default:
             p += reslen;

      } /* switch(qtype) */
    } /* for (recno = reccount; recno > 0; recno--) */
  } /* for (part = 0; part < 3; part++) */

  return;
}
#endif  /* !NDEBUG */

/* ======================================================================= */
/*                          generic net I/O                                */
/* ======================================================================= */

/*
 * __send_req_recv_reply() - Send pre-formatted request, get reply.
 * Returns length of answer, or...
 *  0  authority says name doesn't exist 
 * -1  format error
 * -2  server failure or no timely response from server
 * -3  request cancelled (EINTR)
 * -4  network error
 * -5  request refused
*/
static int __send_req_recv_reply(uint32_t nameserver, struct useek *questans,
                                 int questlen )
{
  int rc = -4; /* assume network error */

  if (my_statics.debug_lev >= 2)
    DBGDUMP(("Question", 0, questans, questlen));

  if (nameserver)
  {
    udp_Socket dom_sock;

    if (sock_wattopen(__IPPROTO_UDP__,&dom_sock,0,nameserver,DOM_DST_PORT,NULL))
    {
      static unsigned short id = 0;
      questans->h.ident = HTONS(++id);
      /* the id in the response should always be correct because we
       * use a new port number for every transaction. 
      */

      if (sock_send_to( &dom_sock, questans, questlen, WSO_MSG_WAITALL, 0,0))
      {         
        unsigned long timeout;
        /* my_statics.res_timeout is from the config file */
        if (my_statics.res_timeout < 1)
          my_statics.res_timeout = 30;
        if (my_statics.res_timeout < 5)
          my_statics.res_timeout = 5;
        timeout = MUL32(1000ul, ((unsigned)my_statics.res_timeout));

        rc = sock_poll( &dom_sock, WSO_ASYSEL_READ, timeout );
        if (rc == -1)  /* error - must be an EINTR */
          rc = -3;
        else if ((rc & WSO_ASYSEL_READ)==0) /* timeout */
          rc = -2;
        else 
        {
          rc = sock_fastread (&dom_sock, questans, sizeof(*questans));
          if (rc < ((int)sizeof(questans->h)))
            rc = -4;
          else 
          {
            uint8_t drc;
            if (my_statics.debug_lev >= 2)
              __debug_dump_pkt( questans, rc );
            drc = (uint8_t)(DRCODE & NTOHS(questans->h.flags));
            if (drc == 3) 
              rc = 0;
            else if (drc != 0)
              rc = -((int)drc);
          }
        } 
      }
      sock_destroy(&dom_sock);

    }
  }

  if (my_statics.debug_lev >= 2)
    DBGPRINTF(("*DNS __send_req_recv_reply() => %d\n", rc));

  return (rc);
}

/*------------------------------------------------------------------*/

static uint32_t __pick_ns(unsigned int idx)
{
  uint32_t ns = 0UL;
  _ip_get_ipaddr(); /* make sure bootp/dhcp are done */
  if (idx < my_statics.ns_count)
    ns = my_statics.ns_list[idx];
  if (!ns && idx == 0 && my_statics.debug_lev)
    DBGPRINTF(("Resolve failed. No nameserver defined.\n"));
  return ns;
}

/* ======================================================================= */
/*                             resolve by name                             */
/* ======================================================================= */

/*
 * packdom()
 *   pack a regular text string into a packed domain name, suitable
 *   for the name server.
 *
 *   returns length
 */
static unsigned int packdom (register char *dst, 
                    register const char FAR *hname, const char FAR *dname)
{
  char *savedst = dst;
  int  dotflag = 0;

  while (*hname && (dst - savedst) < (DOMSIZE-200))
  {
    char len = 0;
    char *lenp = dst++;

    while (*hname && *hname != '.')
    {
      if (len < 0x3F)
      {
        *dst++ = *hname;
        len++;
      }
      hname++;
    }
    *lenp = len;

    if (*hname)                   /* next segment */
    {
      hname++;
      dotflag = 1;
    }
    else if (!dotflag && dname)
    {
      hname = dname;         /* continue packing with default */
      dname = NULL;          /* don't do it twice */
    }
  }

  *dst++ = 0;
  return (unsigned int)(dst - savedst);       /* length of packed string */
}

/* --------------------------------------------------------------------- */ 

/*
 * unpackdom()
 *  Unpack a compressed domain name that we have received from another
 *  host.  Handles pointers to continuation domain names -- buf is used
 *  as the base for the offset of any pointer which is present.
 *  returns the number of bytes at src which should be skipped over.
 *  Includes the NULL terminator in its length count.
 */
static unsigned int unpackdom (uint8_t *dst, const uint8_t *src, const uint8_t *buf)
{
  unsigned int retval  = 0;
  const uint8_t *p = src;

  while (*src)
  {
    int i,j = *src;

    while ((j & 0xC0) == 0xC0)
    {
      if (!retval)
        retval = (int)(src - p + 2);
      src++;
      src = &buf[(j & 0x3F)*256 + *src];  /* pointer dereference */
      j = *src;
    }

    src++;
    for (i = 0; i < (j & 0x3F); i++)
      *dst++ = *src++;

    *dst++ = '.';
  }

  *(--dst) = 0;              /* add terminator */
  src++;                     /* account for terminator on src */

  if (!retval)
    retval = (unsigned int)(src - p);

  return (retval);
}


/* --------------------------------------------------------------------- */ 

/*
 * ddextract()
 *   extract the ip address(es) from a response message.
 *   returns DNS resonse codes as negative numbers,
 *    -1 format error,
 *    -2 server failure or no response from server
 *    -3 (never - returned as zero found)
 *    -4 no can do request
 *    -5 request refused
*/
static int __dd_extract (const struct useek *qp, uint32_t *mips, int nummips)
{
  int numstored = 0;
  uint16_t nans  = NTOHS(qp->h.ancount);
  uint8_t rcode = (uint8_t)(DRCODE & NTOHS(qp->h.flags));

  if (rcode != 0) /* this should have been done already */
  {
    if (rcode == 3) /* name does not exist */
      return 0;     /* zero found */
    return -((int)rcode);
  }

  if (nans > 0 &&                              /* at least one answer   */
      (NTOHS(qp->h.flags) & DQR))            /* response flag is set  */
  {
    uint8_t *space = (uint8_t *)malloc(260);
    const uint8_t *p;
    unsigned int i;

    if (!space)
      return -1;

    p = (const uint8_t *) &qp->x[0];
    i = unpackdom (space, p, (const uint8_t *)qp);

   /* spec defines name then QTYPE + QCLASS = 4 bytes
    */
    p += i + 4;

   /* at this point, there may be several answers.  We will take the first
    * one which has an IP number.  There may be other types of answers that
    * we want to support later.
    */
    while (nans-- > 0)                     /* look at each answer   */
    {
      const struct rrpart *rrp;

      i   = unpackdom (space, p, (const uint8_t *)qp);
      p  += i;                             /* account for string    */
      rrp = (const struct rrpart*) p;      /* resource record here  */

      if (!*p && *(p+1) == DTYPE_A &&       /* correct type and class */
          !*(p+2) && *(p+3) == DCLASS_IN)
      {
        numstored++;
        *mips++ = NTOHL(*((const uint32_t *)(&rrp->rdata[0]))); /* save IP addr */
        nummips--; 
        if (my_statics.debug_lev)
          DBGPRINTF(("answer:'%hI'\n",*(mips-1) ));

        if (nummips == 0) /* done all */
          break;
      }
      p += NTOHS(*((const uint16_t *)&rrp->rdlength));
      p += 10;               /* length of rest of RR */
    }
    free(space);
  }

  if (my_statics.debug_lev)
    DBGPRINTF(("num stored = %d\n", numstored ));

  return (numstored);
}

/* --------------------------------------------------------------------- */ 

/* put together a domain lookup question packet  */
static int __dd_qinit (struct useek *question, const char *hname, 
                     const char FAR *dname)
{
  uint8_t *p, *start;
  unsigned int  i;

  question->h.flags   = HTONS(DRD);
  question->h.qdcount = HTONS(1);
  question->h.ancount = 0;
  question->h.nscount = 0;
  question->h.arcount = 0;
  question->h.ident   = 0; /* filled in by send */

  start = &(question->x[0]);
  i = packdom ((char*)start, hname, dname);
  p = &(question->x[i]);

  *p++ = 0;               /* high byte of qtype */
  *p++ = DTYPE_A;         /* number is < 256, so we know high byte=0 */
  *p++ = 0;               /* high byte of qclass */
  *p++ = DCLASS_IN;       /* qtype is < 256 */

  return sizeof(struct dhead) + ((unsigned int)(p - start));
}

/* --------------------------------------------------------------------- */ 

/*
 * resolve-by-name with
 *   - one particular default domain name (may be NULL).
 *   - one particular nameserver.
 *
 * Returns:
 * >= 0 number of addresses found/stored in 'addrlist' (in net order)
 *   -1 format error,
 *   -2 server failure or no response from server
 *   -3 request cancelled (EINTR)
 *   -4 network error
 *   -5 request refused
 * All return codes except -2 and -5 should cause further attempts
 * to resolve 'hname' to STOP.
*/
static int __dd_lookup (const char *hname, const char FAR *dname,
                        uint32_t nameserver, uint32_t *mips, int nummips )
{
  struct useek *question = (struct useek *)malloc(sizeof(struct useek));
  int rc = -1;
  if (question)
  {
    int questlen = __dd_qinit( question, hname, dname);
    rc = __send_req_recv_reply( nameserver, question, questlen );

    if (rc > 0)
      rc = __dd_extract ( question, mips, nummips);
 
    free(question);
  }
  return (rc);
}

/* --------------------------------------------------------------------- */ 

/* resolve-by-name with -
 *   - one particular default domain name (may be NULL).
 *   - query all nameservers.
 *
 * Returns:
 * >= 0 number of addresses found/stored in 'addrlist' (in net order)
 *   -1 format error,
 *   -2 server failure or no response from server
 *   -3 request cancelled (EINTR)
 *   -4 network error
 *   -5 request refused
 * All return codes except -2 and -5 should cause further attempts
 * to resolve 'hname' to STOP.
*/
static int __resbyname_1( const char *name, const char FAR *dname,
                          uint32_t *addrlist, int numaddrs )
{
  int rc = -2;

  for (;;)
  {
    unsigned int i;
    for (i = 0; (rc == -2 || rc == -5); i++)
    {
      uint32_t ns = __pick_ns(i);
      if (!ns)
      {
        if (i == 0) /* no first nameserver */
          rc = 0;   /* thus no nameservers */
        break;
      }
      rc = __dd_lookup (name, dname, ns, addrlist, numaddrs );
    } 
    if (rc != -2 && rc != -5)  /* if (!serv_fail && !request_refused) */
      break; 
    if (!dname)
      break;
    while (*dname && *dname != '.')
      dname++;
    if (*dname)
      dname++;
    if (!*dname)
      break; /* end of name */
  }  
  return rc;
}

/* --------------------------------------------------------------------- */ 

int _dns_resbyname (const char *hname, uint32_t *addrlist, int numaddrs )
{
  uint32_t tmp;
  int rc, hn_has_dots;

  if (!hname || !*hname || !addrlist || numaddrs <= 0)
    return 0;

  if (_my_aton(hname, &tmp))
  {
    *addrlist = tmp;
    return 1;
  }

  if (MY_STRCASECMP( hname, "localhost" ) == 0)
  {
    *addrlist = 0x7f000001ul;
    return 1;
  }

  if (__static_resolve(hname, 0, addrlist, sizeof(*addrlist) ))
  {
    return 1;
  }

  hn_has_dots = (!(!strchr(hname, '.')));

  __chkpt_read_from_hostdomname();
  if (!my_statics.dom_name || !hn_has_dots)
  {
    if (MY_STRCASECMP( my_statics.host_name, hname) == 0)
    {
      *addrlist = _ip_get_ipaddr();
      return 1;
    }
  }

  rc = __resbyname_1( hname, (const char FAR *)0, addrlist, numaddrs );

  if ((rc == -2 || rc == -5) && (my_statics.res_try_subdoms && !hn_has_dots))
  {
#ifdef HAVE_DOMAINSRCHLIST
    int i;
    for (i = 0; (rc == -2 || rc == -5) && i < my_statics.srchdom_count; i++)
    {
      rc = __resbyname_1( hname, my_statics.srchdom_list[i], 
                          addrlist, numaddrs );
    }
#endif
    if ((rc == -2 || rc == -5) && my_statics.dom_name)
    {
      rc = __resbyname_1( hname, my_statics.dom_name, 
                          addrlist, numaddrs );
    }
  }

  return (rc);
}

/* ===================================================================== */
/*                           reverse lookup                              */
/* ===================================================================== */

/*
 * Fill in the reverse lookup question packet
 * IP addr is assumed to be in host order.
 */
static int __rr_qinit (struct useek *q, uint32_t ipaddr /* in host order */)
{
  uint8_t *c;
  uint8_t  i;

  q->h.ident   = 0; /* filled in by __send */
  q->h.flags   = HTONS(DRD);   /* recursion desired */
  q->h.qdcount = HTONS(1);
  q->h.ancount = 0;
  q->h.nscount = 0;
  q->h.arcount = 0;

  c  = q->x;
  ipaddr = HTONL(ipaddr); /* host->net order */
  for (i = 0; i < 4; ++i)
  {
    uint8_t x = (uint8_t)(ipaddr >> 24);
    ipaddr <<= 8;
    *c++ = (uint8_t)((x < 10) ? 1 : (x < 100) ? 2 : 3);
    if (x >= 100)
      *c++ = (uint8_t)('0'+(x / 100));
    if (x >= 10)
      *c++ = (uint8_t)('0'+((x % 100) / 10));
    *c++ = (uint8_t)('0'+(x % 10));
  }
  strcpy((char *)c, "\7in-addr\4arpa");
  c += 14;
  *((uint16_t *)c) = HTONS(DTYPE_PTR);
  c += sizeof(uint16_t);
  *((uint16_t *)c) = HTONS(DCLASS_IN);
  c += sizeof(uint16_t);

  return (int)(c - ((uint8_t *)q)); /* questlen */
}

/* ----------------------------------------------------------------------- */

/*
 * rrextract() - read answer and extract host name
 *               return 0 on error, length of resname on success.
 */
static int __rr_extract(const struct useek *qp, char FAR *name, int namelen)
{
  uint8_t rcode = (uint8_t)(DRCODE & NTOHS(qp->h.flags));
  const struct rrpart *rr;
  int donelen;
  const uint8_t *c;    

  namelen--; /* for terminating '\0' */
  donelen = 0;

  if (rcode != 0)
  {
    if (rcode == 3) /* name does not exist */
      return 0;     /* zero found */
    return -((int)rcode);
  }

  if (NTOHS(qp->h.qdcount) != 1 || qp->h.ancount == 0)
    return (0);

  /* Skip question */
  c = &qp->x[0];
  while (*c)
    ++c;
  c += 5;

  /* Skip name */
  while (*c)
  {
    if ((*c & 0xC0) == 0xC0)
    {
      c += 2;
      break;
    }
    else
      ++c;
  }
  rr = (const struct rrpart *) c;
  if (NTOHS(rr->rtype) != DTYPE_PTR)
    return (0);

  c = (const uint8_t *)&(rr->rdata[0]);
  while (*c && donelen < namelen)
  {
    unsigned int i, n;
    const uint8_t *pp;
    if ((*c & 0xC0) == 0xC0)
      c = &qp->x[0] + ((((unsigned int)(*c & 0x3F)) << 8) + c[1]);
    n = namelen - donelen;
    if (n > *c)
      n = *c;
    pp = (const uint8_t *)(c+1);
    c += *c + 1;
    for (i = 0; i < n; i++) 
      name[donelen++] = *pp++;
    if (donelen == namelen)
      break;

    if (*c)
      name[donelen++] = '.';
  }

  name[donelen] = 0;
  return (donelen);
}

/*------------------------------------------------------------------*/

static void __fill_in_own_if_needed(const struct useek *qp, uint32_t ipaddr)
{
  if (my_statics.based_ip_addr != ipaddr && ipaddr == _ip_get_ipaddr())
  {
    char FAR *ans = (char FAR *)FAR_MALLOC(sizeof(*qp));
    if (ans)
    {
      int rc = __rr_extract(qp, ans, sizeof(*qp) );
      if (rc > 0 && ((unsigned int)rc) < (sizeof(*qp)-1))
      {
        unsigned int n = 0;
        char FAR *tmp = ans;

        while (*tmp && *tmp != '.' && n < sizeof(my_statics.host_name)-1)
          my_statics.host_name[n++] = *tmp++;
        my_statics.host_name[n] = '\0';

        n = 0;
        while (*tmp && *tmp != '.') 
          tmp++;
        if (*tmp == '.')
          tmp++;
        while (*tmp)
          ans[n++] = *tmp++;
        ans[n] = '\0';

        my_statics.based_ip_addr = ipaddr;

        tmp = my_statics.dom_name;
        my_statics.dom_name = ans;
        ans = tmp;
        /* wotchit! 'ans' may be NULL now */
      }
    } 
    if (ans)
      FAR_FREE(ans);
  }
  return;
}

/*------------------------------------------------------------------*/

/*
 * reverse_lookup() - Translate an IP into a host name.
 * Returns >0 on success, 0 if no result, else
 * -1  format error
 * -2  server failure or no timely response from server
 * -3  request cancelled (EINTR)
 * -4  network error
 * -5  request refused
 */
static int __rr_lookup (uint32_t ipaddr /* in host order */, 
                        char *name, int namelen,
                        uint32_t nameserver /* in host order */)
{
  int rc = -1;
  struct useek *question = (struct useek *)malloc(sizeof(struct useek));
  if (question)
  {
    int questlen = __rr_qinit (question, ipaddr);
    rc = __send_req_recv_reply( nameserver, question, questlen );

    if (rc > 0) /* length of answer */
    {
      rc = __rr_extract( question, name, namelen );
      __fill_in_own_if_needed( question, ipaddr );
    }
    free(question);
  }
  return (rc);
}

/*------------------------------------------------------------------*/

/*
 * _dns_resolvebyaddr() - Translate an IP into a host name.
 * Returns > 0 on success, 0 if no result, else
 * -1  format error
 * -2  server failure or no timely response from server
 * -3  request cancelled (EINTR)
 * -4  network error
 * -5  request refused
 */
int _dns_resbyaddr( uint32_t ipaddr, char *hnamebuf, unsigned int buflen )
{
  unsigned int n;
  int rc = -2;

  if (buflen < 1)
    return -1;
  if (!ipaddr || buflen < 2)
    rc = 0;
  else if ((ipaddr >> 24) == 127)
  {
    strncpy( hnamebuf, "localhost", buflen );
    hnamebuf[buflen-1] = '\0';
    rc = strlen(hnamebuf);
  }
  else if (my_statics.host_name[0] &&
           my_statics.based_ip_addr == ipaddr && 
           my_statics.based_ip_addr == _ip_get_ipaddr())
  {
    buflen--;
    n = 0; rc = 0;
    while (((unsigned int)rc) < buflen && my_statics.host_name[n])
      hnamebuf[rc++] = my_statics.host_name[n++];
    if (my_statics.dom_name && *my_statics.dom_name)
    {
      if (((unsigned int)rc) < buflen)
        hnamebuf[rc++] = '.'; 
      n = 0;
      while (((unsigned int)rc) < buflen && my_statics.dom_name[n])
        hnamebuf[rc++] = my_statics.dom_name[n++];
    }
    hnamebuf[rc] = '\0';
  }
  else if (__static_resolve((const char *)0, ipaddr, hnamebuf, buflen ))
  {
    rc = strlen(hnamebuf);
  }
  else
  {
    for (n = 0; (rc == -2 || rc == -5); ++n)
    {
      uint32_t ns = __pick_ns(n);
      if (!ns)
      {
        if (n == 0) /* no nameserver in first try */
          rc = 0;
        break;
      }
      rc = __rr_lookup( ipaddr, hnamebuf, buflen, ns );
    }
  }

  if (rc == 0)
    hnamebuf[0] = '\0';

  return rc;
}

