/* 
 * "uniform" C-callable interface to the DPMI API
 *
 * Created May 2001 by Cyrus Patel <cyp@fb14.uni-mainz.de>
 *
 * For simplicity/consistancy, the __dpmi_*() functions used/prototyped 
 * adhere to the same argument/return conventions as __dpmi_*() functions
 * prototyped in DJGPP's dpmi.h. 
 * 
 * The DJGPP GO32 extensions to DPMI are _not_ emulated. Functionality of
 * the _go32_*() functions may be achieved with the _my_dpmi_*() calls 
 * which do not suffer from _go32_*() hackiness.
 *
 * Note: __DPMI32__ will always be defined by dpmi.h for 32bit targets.
 *  - if a 386 extender is not DPMI compliant (for example PMODE
 *    without its Int31h interface), then the application that links
 *    this library will never go anywhere. :)
 *  - __DPMI32__ is defined even when pharlap.h has been included.
 *    (Some people think that an app running under a pharlap extender 
 *    cannot make DPMI calls. Let those poor, huddled, unwashed masses 
 *    be herewith corrected).
 *
 * Created May 2001 by Cyrus Patel <cyp@fb14.uni-mainz.de>
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *    $Id: dpmi.h,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/

#ifndef __KERN_DPMI_H 
#define __KERN_DPMI_H

#if !defined(__DPMI32__) /* some compilers already define this */
  #include <limits.h>
  #if (UINT_MAX > USHRT_MAX) /* 32bit build */
    #define __DPMI32__
  #endif
#endif

/* ===================================================================== */
/*                   the DPMI API (emulating DJGPP's one)                */
/* ===================================================================== */

#if defined(__DJGPP__)
  #include <sys/movedata.h>  /* dosmem*(), movedata() */
  #include <sys/segments.h>  /* _my_cs(), _my_ds() */
  #include <sys/exceptn.h>   /* __djgpp_ds_alias */
  #include <libc/farptrgs.h> /* _farpeek*()/_farpoke*() */
  #include <go32.h>          /* _dos_ds define */
  #include_next <dpmi.h> 
  #define __DPMI32_DOSDS_INTRINSIC__ /* _dos_ds is not a function */
  #define __DPMI32_PEEKPOKE_DEFINED__

  static __inline__ 
  void __inline_movedata(unsigned _src_sel, unsigned _src_off,
                         unsigned _dst_sel, unsigned _dst_off,
                         size_t _length)
  {
    int dummy1, dummy2;
    /* _printk("movedata: %02x:%04x => %02x:%04x (len=%u)\n",
                 _src_sel, _src_off, _dst_sel, _dst_off, _length );
    */
    /* stolen from src\libc\pc_hw\mem\djmd.s */
    __asm__ __volatile__(
        "pushl   %%ds         \n\t"
        "pushl   %%es         \n\t"
        "pushl   %%esi        \n\t"
        "pushl   %%edi        \n\t"
        "                     \n\t"
        "movw    %%dx,%%ds    \n\t"
        "movw    %%ax,%%es    \n\t"
        "                     \n\t"
        "cld                  \n\t"
        "cmpl    $15,%%ecx    \n\t"
        "jbe     3f           \n\t"
        "jmp     1f           \n\t"
        "                     \n\t"
        "0:                   \n\t"
        "movsb                \n\t"
        "decl    %%ecx        \n\t"
        "1:                   \n\t"
        "testl   $3,%%esi     \n\t"
        "jz      2f           \n\t"
        "testl   $3,%%edi     \n\t"
        "jnz     0b           \n\t"
        "                     \n\t"
        "2:                   \n\t"
        "movb    %%cl,%%al    \n\t"
        "shrl    $2,%%ecx     \n\t"
        "andb    $3,%%al      \n\t"
        "rep                  \n\t"
        "movsl                \n\t"
        "                     \n\t"
        "movb    %%al,%%cl    \n\t"
        "3:                   \n\t"
        "rep                  \n\t"
        "movsb                \n\t"
        "                     \n\t"
        "popl    %%edi        \n\t"
        "popl    %%esi        \n\t"
        "popl    %%es         \n\t"
        "popl    %%ds         \n\t"
        : "=a"(dummy1), "=c"(dummy2)
        : "d"(_src_sel), "S"(_src_off), 
          "a"(_dst_sel), "D"(_dst_off), "c"(_length) 
        );
    /* _printk("end movedata\n" ); */
  }
  static __inline__ 
  void __inline_dosmemget(unsigned long _offset, 
                          size_t _length, void *_buffer)
  {
    __inline_movedata( _dos_ds, _offset, _my_ds(), 
                       ((unsigned)(_buffer)), _length );
  }
  static __inline__ 
  void __inline_dosmemput(const void *_buffer, size_t _length, 
                          unsigned long _offset)
  {
    __inline_movedata( _my_ds(), ((unsigned)(_buffer)), 
                       _dos_ds, _offset, _length );
  }
  #undef  dosmemget
  #undef  dosmemput
  #undef  movedata
  #define dosmemget __inline_dosmemget
  #define dosmemput __inline_dosmemput
  #define movedata  __inline_movedata

#elif defined(__DPMI32__)

  #ifdef __cplusplus
  extern "C" {
  #endif

  #if (defined(_MSC_VER) && (_MSC_VER >= 700)) || \
   (defined(__BORLANDC__) && defined(__FLAT__))
  #  pragma warning(disable:4103)
  #  pragma pack(1)
  #elif defined(__CCDL__)
  #  pragma packed
  #elif defined(__HIGHC__)
  #  pragma Off(Align_members)
  #elif defined(__WATCOMC__) && (__WATCOMC__ >= 1000)
  #  pragma pack(__push,1); /* #pragma pack(1) works too */
  #elif defined(__TURBOC__)
  #  pragma option -a-
  #elif defined(__GNUC__)
  #  pragma pack(1)
  #else
  #  error huh?
  #endif

  typedef struct {
    unsigned short offset16;
    unsigned short segment;
  } __dpmi_raddr;

  typedef struct {
    unsigned long offset32;
    unsigned short selector;
  } __dpmi_paddr;

  typedef union {
    struct {
      unsigned long edi;
      unsigned long esi;
      unsigned long ebp;
      unsigned long res;
      unsigned long ebx;
      unsigned long edx;
      unsigned long ecx;
      unsigned long eax;
    } d;
    struct {
      unsigned short di, di_hi;    /* 0x00,  0 */
      unsigned short si, si_hi;    /* 0x04,  4 */
      unsigned short bp, bp_hi;    /* 0x08,  8 */
      unsigned short res, res_hi;  /* 0x0C, 12 */
      unsigned short bx, bx_hi;    /* 0x10, 16 */
      unsigned short dx, dx_hi;    /* 0x14, 20 */
      unsigned short cx, cx_hi;    /* 0x18, 24 */
      unsigned short ax, ax_hi;    /* 0x1C, 28 */
      unsigned short flags;        /* 0x20, 32 */
      unsigned short es;           /* 0x22, 34 */
      unsigned short ds;           /* 0x24, 36 */
      unsigned short fs;           /* 0x26, 38 */
      unsigned short gs;           /* 0x28, 40 */
      unsigned short ip;           /* 0x2a, 42 */
      unsigned short cs;           /* 0x2c, 44 */
      unsigned short sp;           /* 0x2e, 46 */
      unsigned short ss;           /* 0x30, 48 */
    } x;                    /* size = 0x32, 50 */
    struct {
      unsigned char edi[4];
      unsigned char esi[4];
      unsigned char ebp[4];
      unsigned char res[4];
      unsigned char bl, bh, ebx_b2, ebx_b3;
      unsigned char dl, dh, edx_b2, edx_b3;
      unsigned char cl, ch, ecx_b2, ecx_b3;
      unsigned char al, ah, eax_b2, eax_b3;
    } h;
  } __dpmi_regs;

  #if (defined(_MSC_VER) && (_MSC_VER >= 700)) || \
   (defined(__BORLANDC__) && defined(__FLAT__))
  #  pragma warning(enable:4103)
  #  pragma pack()
  #elif defined(__CCDL__)
  #  pragma unpacked
  #elif defined(__HIGHC__)
  #  pragma Pop(Align_members)
  #elif defined(__WATCOMC__) && (__WATCOMC__ >= 1000)
  #  pragma pack(__pop); /* #pragma pack() works too */
  #elif defined(__TURBOC__)
  #  pragma option -a.
  #elif defined(__GNUC__)
  #  pragma pack()
  #endif


  /* --- RAW dpmi functions - DJGPP compatible --- */

  /* LSL isn */ extern unsigned long __dpmi_get_segment_limit ( unsigned int _sel );
  /* 2f:1680 */ extern void __dpmi_yield (void);
  /* 32:0000 */ extern int __dpmi_allocate_ldt_descriptors(int count);
  /* 31:0001 */ extern int __dpmi_free_ldt_descriptor (int _descriptor);
  /* 31:0006 */ extern int __dpmi_get_segment_base_address( int seg, unsigned long *base );
  /* 31:0007 */ extern int __dpmi_set_segment_base_address(int sel, unsigned long base);
  /* 31:0008 */ extern int __dpmi_set_segment_limit(int sel, unsigned long limit);
  /* 31:000a */ extern int __dpmi_create_alias_descriptor (int sel);
  /* 31:0100 */ extern int __dpmi_allocate_dos_memory(int paras, int *sel_or_max);
  /* 31:0101 */ extern int __dpmi_free_dos_memory(int sel);
  /* 31:0102 */ extern int __dpmi_resize_dos_memory(int sel, int newparas, int *ret_max);
  /* 31:0200 */ extern int __dpmi_get_real_mode_interrupt_vector(int iv, __dpmi_raddr *);
  /* 31:0201 */ extern int __dpmi_set_real_mode_interrupt_vector(int iv, __dpmi_raddr *rmaddr);
  /* 31:0204 */ extern int __dpmi_get_protected_mode_interrupt_vector(int iv, __dpmi_paddr *);
  /* 31:0205 */ extern int __dpmi_set_protected_mode_interrupt_vector(int iv, __dpmi_paddr *);
  /* 31:0300 */ extern int __dpmi_simulate_real_mode_interrupt(int iv, __dpmi_regs *regs);
  /* 31:0300 */ extern int __dpmi_int(int iv, __dpmi_regs *regs);
  /* 31:0301 */ extern int __dpmi_simulate_real_mode_procedure_retf(__dpmi_regs *regs);
  /* 31:0301 */ extern int __dpmi_simulate_real_mode_procedure_iret(__dpmi_regs *regs);
  /* 31:0303 */ extern int __dpmi_allocate_real_mode_callback(
                                       auto void (*_handler)(void), 
                                       __dpmi_regs *regs,
                                       __dpmi_raddr *rmaddr ); 
  /* 31:0304 */ extern int __dpmi_free_real_mode_callback(__dpmi_raddr *rmaddr ); 
  /* 31:09xx */ extern int __dpmi_get_and_set_virtual_interrupt_state(int newon);
  /* 31:0900 */ extern int __dpmi_get_and_disable_virtual_interrupt_state(void);
  /* 31:0901 */ extern int __dpmi_get_and_enable_virtual_interrupt_state(void);
  /* 31:0902 */ extern int __dpmi_get_virtual_interrupt_state(void);

  /* --- commonly used selectors - DJGPP compat -- */

  #if defined(__TURBOC__) || defined(__BORLANDC__)
    #define _my_cs()  _CS
    #define _my_ds()  _DS
    #define _dos_ds __my_dpmi_get_dosds()
  #elif defined(__WATCOMC__)
    extern unsigned short _my_cs(void);
    extern unsigned short _my_ds(void);
    #pragma aux _my_cs = "mov ax,cs" value [ax] modify exact [ax];
    #pragma aux _my_ds = "mov ax,ds" value [ax] modify exact [ax];
    /* See <libsource>\clib\startup\c\inirmsel.c */
    extern unsigned short _ExtenderRealModeSelector;
    #define _dos_ds _ExtenderRealModeSelector
    #define __DPMI32_DOSDS_INTRINSIC__ /* _dos_ds is not a function */
  #elif defined(__HIGHC__)
    #define _my_cs() _mwgetcs()   /* SS_CODE|SEL_LDT|attributes */
    #define _my_ds() _mwgetcs()   /* SS_DATA|SEL_LDT|attributes */
    #error please confirm cs == ds
    /* *WARNING* do not use the SS_CODE/SS_DATA constants as these   */
    /* change between extender versions. Only the SS_DOSMEM selector */
    /* is truly constant (since the C libraries depend on it).       */
    #if defined(PHARLAP_H)
    /* ((SS_DOSMEM | SEL_LDT) | read/write-attributes-of-_my_ds) */
    #define _dos_ds (0x34 | (_my_ds() & 0x03))
    #define __DPMI32_DOSDS_INTRINSIC__ /* _dos_ds is not a function */
    #else
    #define _dos_ds __my_dpmi_get_dosds()
    #endif
  #else
    #error "I don't know how to do this"
    #error "(whatever the solution, it must be ISR safe)"
    #define _dos_ds __my_dpmi_get_dosds()
  #endif

  #ifdef __cplusplus
  }
  #endif

  /* --- dosmemget/dosmemput - DJGPP compatible --- */
  /* movedata is assumed to be inline */
  #define dosmemget(__offset, __length, __buffer) \
     movedata( __my_dpmi_get_dosds(), __offset, \
               __my_dpmi_get_ds(), ((unsigned)(__buffer)), __length )
  #define dosmemput(__buffer, __length, __offset) \
     movedata( __my_dpmi_get_ds(), ((unsigned)(__buffer)),  \
               __my_dpmi_get_dosds(), __offset, __length )
  #define _dosmemgetb(__offt, __xfers, __buffer) \
     dosmemget( __offset, __xfers, __buffer ) 
  #define _dosmemgetw(__offt, __xfers, __buffer) \
     dosmemget( __offset, ((size_t)(__xfers))<<1, __buffer )
  #define _dosmemgetl(__offt, __xfers, __buffer) \
     dosmemget( __offset, ((size_t)(__xfers))<<2, __buffer )
  #define _dosmemputb(__buffer, __xfers, __offt) \
     dosmemput( __buffer, __xfers, __offt )
  #define _dosmemputw(__buffer, __xfers, __offt) \
     dosmemput( __buffer, ((size_t)(__xfers))<<1, __offt )
  #define _dosmemputl(__buffer, __xfers, __offt) \
     dosmemput( __buffer, ((size_t)(__xfers))<<2, __offt )

  #if defined(MK_FP) && !defined(__DPMI32_PEEKPOKE_DEFINED__)
  #define __mk_peek( __type, __sel, __off ) \
                   *((__type _far *)MK_FP((__sel),(__off)))
  #define _farpeekb( __s, __o ) __mk_peek( unsigned char, __s, __o )
  #define _farpeekw( __s, __o ) __mk_peek( unsigned short, __s, __o )
  #define _farpeekl( __s, __o ) __mk_peek( unsigned long, __s, __o )
  #define __mk_poke( __type, __sel, __off, __val ) \
          (void)(__mk_peek( __type, __sel, __off ) = ((__type)(__val)))
  #define _farpokeb( __s, __o, __v ) __mk_poke( unsigned char, __s, __o, __v )
  #define _farpokew( __s, __o, __v ) __mk_poke( unsigned short, __s, __o, __v )
  #define _farpokel( __s, __o, __v ) __mk_poke( unsigned long, __s, __o, __v )
  #define __DPMI32_PEEKPOKE_DEFINED__
  #endif

#endif /* if defined(__DPMI32__) && !defined(__DJGPP__) */

/* ===================================================================== */
/*               my dpmi extensions using/of the DPMI API                */
/* ===================================================================== */

#if defined(__DPMI32__) /* all */

  #ifdef __cplusplus
  extern "C" {
  #endif

  #define __my_dpmi_get_cs() _my_cs()
  #define __my_dpmi_get_ds() _my_ds()
  #if defined(__DPMI32_DOSDS_INTRINSIC__) /* _dos_ds is not a function */
  #define __my_dpmi_get_dosds() _dos_ds
  #else
  extern int __my_dpmi_get_dosds(void);
  #define _dos_ds __my_dpmi_get_dosds()
  #endif

  extern int __my_dpmi_get_csalias(void);

  /* wrappers around lock/unlock region */
  extern int __my_dpmi_unlock_data (const void *address, unsigned length);
  extern int __my_dpmi_lock_data (const void *address, unsigned length);
  extern int __my_dpmi_unlock_code (const void *address, unsigned length);
  extern int __my_dpmi_lock_code (const void *address, unsigned length);

  typedef void (*__dpmi_rmcb_handler_t)(void);

  extern int __my_dpmi_allocate_real_mode_callback_iret ( 
                                           __dpmi_rmcb_handler_t _handler, 
                                           __dpmi_regs *regs,
                                           __dpmi_raddr *rmaddr ); 
  extern int __my_dpmi_allocate_real_mode_callback_ichain ( 
                                           __dpmi_rmcb_handler_t _handler, 
                                           int vector,
                                           __dpmi_regs *regs,
                                           __dpmi_raddr *rmaddr ); 
  extern int __my_dpmi_allocate_real_mode_callback_retf ( 
                                           __dpmi_rmcb_handler_t _handler, 
                                           __dpmi_regs *regs,
                                           __dpmi_raddr *rmaddr );
  extern int __my_dpmi_free_real_mode_callback ( __dpmi_raddr *rmaddr );

  extern void __my_dpmi_chain_pmiv(__dpmi_regs *regs);
  extern int  __my_dpmi_unhook_pmiv( int hook );
  extern int  __my_dpmi_hook_pmiv( int vector, 
                                    void (*pm_entry)(__dpmi_regs *regs) );

  #ifdef __cplusplus
  }
  #endif

#endif /* __DPMI32__ */

/* ====================================================================== */

#endif /* !__KERN_DPMI_H */
