/*
 * Service provider for the 
 * Internet Control Message Protocol, "ICMP", (RFC 792)
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: icmp.c,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
#define __OSILAYER__ 3 /* actually, ... both 3 and 4 */

#include "target.h"   /* standard types/prototypes */
#include "inchksum.h" /* checksum() */
#include "inaddr.h"   /* is_broadcast(), is_local_addr() */
#include "loadcfg.h"  /* _conf_add_confextn() */
#include "sigblk.h"   /* set_breakstate() */
#include "timer.h"    /* wintr_*(), Random() */
#include "blocker.h"  /* blocker_*() */
#include "linklayr.h" /* _eth_send() for icmp_retour() */
#include "arp.h"      /* _arp_add_gateway() */
#include "debug.h"    /* DBGPRINTF() */
#include "ipbase.h"   /* _ip_output() etc */
#include "icmp.h"     /* ourselves */

#define INCL_ICMP_MASKREQ_STUFF    /* yields! - used by IPBOOT */
#define INCL_ICMP_ROUTERDISC_STUFF /* yields! - used by ARP resolver */
#define INCL_ICMP_REDIRECT_STUFF   /* passive - used by ARP cache */

/*
 * These are the ICMP messages. Should be same as in <netinet/ip_icmp.h>
 */
#define ICMP_ECHOREPLY     0     /* echo reply */
#define ICMP_UNREACH       3     /* dest unreachable */
#define ICMP_SOURCEQUENCH  4     /* packet lost, slow down */
#define ICMP_REDIRECT      5     /* shorter route        */
#define ICMP_ECHO          8     /* echo service         */
#define ICMP_ROUTERADVERT  9     /* router advertisement */
#define ICMP_ROUTERSOLICIT 10    /* router solicitation  */
#define ICMP_TIMXCEED      11    /* time exceeded        */
#define ICMP_PARAMPROB     12    /* ip header bad        */
#define ICMP_TSTAMP        13    /* timestamp request    */
#define ICMP_TSTAMPREPLY   14    /* timestamp reply      */
#define ICMP_IREQ          15    /* information request  */
#define ICMP_IREQREPLY     16    /* information reply    */
#define ICMP_MASKREQ       17    /* address mask request */
#define ICMP_MASKREPLY     18    /* address mask reply   */
#define ICMP_MAXTYPE       18

/* ------------------------------------------------------------ */

#pragma pack(1)
struct icmp_header {
       uint8_t      type;
       uint8_t      code;
       uint16_t      checksum;
       /* followed by op data */
};      
#pragma pack()

/* ======================================================================= */
/*                            ICMP config stuff                            */
/* ======================================================================= */

static struct 
{
  int debug_lev;
} FAR my_statics;

/* ======================================================================= */
/*                            ICMP sender                                  */
/* ======================================================================= */

/*
 * __icmp_send - format and send a ICMP packet
 *
 * source IP address MUST be appropriate.
 * ip_output() will not insert _ip_get_ipaddr() by default
 * since some messages need to come from 0.0.0.0
*/
static int __icmp_send( uint32_t src, uint32_t dest, uint8_t ttl, uint8_t tos,
                        const void FAR *icmp_pkt, unsigned int icmp_pkt_len, 
                        const void FAR *extra_data, unsigned int extra_data_len)
{
  char buf[32]; /*the largest 'icmp_header' any caller of icmp_send may have*/
  struct icmp_header *icmp_hdr = (struct icmp_header *)&buf[0];
  int so_errno;

  if (!extra_data || !extra_data_len)
  {
    extra_data = (const void FAR *)0;
    extra_data_len = 0;
  }

  if (icmp_pkt_len < 1 || icmp_pkt_len > sizeof(buf))
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*__icmp_send() fail: icmp_pkt_len < 1 or > %u\n",
                    (sizeof(buf)-sizeof(in_Header)) ));
    return -1;
  }

  FAR_MEMCPY( icmp_hdr, icmp_pkt, icmp_pkt_len );

  icmp_hdr->checksum = 0;
  icmp_hdr->checksum = ~checksum2( icmp_hdr, icmp_pkt_len,  
                                   extra_data, extra_data_len );

  if (_ip_output_2( src, dest, IPPROTO_ICMP, ttl, tos, 0,
                    icmp_hdr, icmp_pkt_len, extra_data, extra_data_len,
                    &so_errno ) == -1 )
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*__icmp_send() (type=%u,code=%u) failed (err=%d)\n",
            icmp_hdr->type, icmp_hdr->code, so_errno ));
    return -1;
  }
  if (my_statics.debug_lev)
    DBGPRINTF(("*__icmp_send() (type=%u,code=%u) OK\n",
             icmp_hdr->type, icmp_hdr->code ));
  return 0;
}                    

/* ================================================================= */
/*                              PING                                 */
/* ================================================================= */

#pragma pack(1)
struct icmp_echo_reqrep 
{
  struct icmp_header hdr;
  uint16_t  identifier;
  uint16_t  sequence;
  /* -- user data begins here -- */
};
#pragma pack()

/* ----------------------------------------------------------------- */

static struct
{
  icmp_echo_reply_hook_t hk;
} ping_ctrl;

int icmp_echo_reply_hook_add(icmp_echo_reply_hook_t hk)
{
  ping_ctrl.hk = hk;
  return 0;
}

int icmp_echo_reply_hook_del(icmp_echo_reply_hook_t hk)
{
  if (hk != ping_ctrl.hk)
    return -1;
  ping_ctrl.hk = NULL;
  return 0;
}

static void __handle_echo_reply( const in_Header FAR *ipdgram, uint16_t pktid )
{
  if (!is_local_addr(NTOHL(ipdgram->destination)))
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*ICMP-%04x Echo Reply DROP %I->%I (!forme)\n", pktid,
                                      ipdgram->source, ipdgram->destination ));
  }
  else if (!ping_ctrl.hk)
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*ICMP-%04x Echo Reply DROP %I->%I (no handler)\n", pktid, 
                                        ipdgram->source, ipdgram->destination ));
  }
  else
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*ICMP-%04x Echo Reply OK %I->%I\n", pktid, 
                                        ipdgram->source, ipdgram->destination ));
    (*(ping_ctrl.hk))(ipdgram);
  }

  pktid = pktid; /* shaddup compiler */
  return;
}
  
/* ----------------------------------------------------------------- */

static void __handle_echo( const in_Header FAR *ipdgram, 
                           uint16_t pktid, int is_bcast )
{
  uint32_t dst_addr = NTOHL(ipdgram->source);
  uint32_t src_addr = NTOHL(ipdgram->destination);

  if (my_statics.debug_lev)
    DBGPRINTF(("*ICMP-%04x PING from %I\n", pktid, ipdgram->source ));

  if (is_bcast || !dst_addr || is_broadcast(dst_addr) || 
      !is_local_addr(src_addr))
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*ICMP-%04x PING-DROP (bcast or not for me)\n", pktid));
  }
  else
  {
    struct icmp_header ihdr;
    const uint8_t FAR *dataptr;
    uint16_t datalen; 
    int rc;

    dataptr = (((const uint8_t FAR *)ipdgram) + in_GetHdrLen(ipdgram));
    FAR_MEMCPY(&ihdr, dataptr, sizeof(ihdr));
    dataptr += sizeof(ihdr);

    datalen = (uint16_t)(NTOHS(ipdgram->length) - 
                     (in_GetHdrLen(ipdgram) + sizeof(ihdr)));

    ihdr.type     = ICMP_ECHOREPLY;
    ihdr.checksum = 0;

    rc = __icmp_send( src_addr, dst_addr, ((uint8_t)0xff), ipdgram->tos,
                      &ihdr, sizeof(ihdr), dataptr, datalen );

    if (my_statics.debug_lev)
      DBGPRINTF(("*ICMP-%04x echo reply %I->%I (payload len=%u)... %s\n",
         pktid, src_addr, dst_addr, datalen, ((rc==0) ? "OK" : "FAILED") ));

    rc = rc; /* not used unless debugging */
  }
  pktid = pktid; /* not used unless debugging */
  return;
}

/* ================================================================= */
/*                           ICMP mask req/reply                     */
/* ================================================================= */

/*
 *  Send an ICMP Address Mask Request 
 *  icmp_mask_req() is called during stack init
 */
#pragma pack(1)
struct icmp_mask_reqrep 
{
  struct icmp_header hdr;
  uint16_t  identifier;
  uint16_t  sequence;
  uint32_t mask;
};
#pragma pack()

#if defined(INCL_ICMP_MASKREQ_STUFF)
static struct icmp_mask_reqrep *pending_mask_req = NULL;
#endif

static void __handle_mask_reply( const struct icmp_header FAR *_pkt, 
                                 unsigned int pkt_len, uint16_t pktid )
{
#if !defined(INCL_ICMP_MASKREQ_STUFF)
  _pkt = _pkt; pkt_len = pkt_len;
#else
  const struct icmp_mask_reqrep FAR *pkt;
  pkt = (const struct icmp_mask_reqrep FAR *)_pkt;

  if (pkt_len < sizeof(struct icmp_mask_reqrep))
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*ICMP-%04x MASKREPLY too short\n", pktid ));
    return;
  }

  if (my_statics.debug_lev)
    DBGPRINTF(("*ICMP-%04x MASKREPLY. (mask=%I)\n", pktid, pkt->mask ));

  if (pending_mask_req)
  {
    if (pkt->identifier == pending_mask_req->identifier
     && pkt->sequence   == pending_mask_req->sequence)
    {
      pending_mask_req->mask = pkt->mask;
    }
  }
#endif
  pktid = pktid; /* unused unless debugging */
  return;  
}

/* ----------------------------------------------------------------- */

/*
** icmp_mask_req() needs to yield and MAY ONLY BE CALLED BY IPBOOT,
** (IPBOOT knows when it is safe to yield).
*/
uint32_t icmp_mask_req (int timeout_secs)
{
#if !defined(INCL_ICMP_MASKREQ_STUFF)
  timeout_secs = timeout_secs;
  return 0;
#else
  /* mask request/reply packet looks like this:
  * struct 
  * {
  *   icmp_header hdr;
  *   uint16_t identifier;
  *   uint16_t sequence;
  *   uint32_t mask;
  * }
  */
  struct icmp_mask_reqrep req;

  req.hdr.type               = ICMP_MASKREQ;
  req.hdr.code               = 0;
  req.hdr.checksum           = 0;
  req.identifier             = (uint16_t)Random(0,0xffff); /* random ID */;
  req.sequence               = 1; /* duzznt matter */
  req.mask                   = 0;
  pending_mask_req           = &req;  

  if (my_statics.debug_lev)
    DBGPRINTF(("*ICMP sending ICMP_MASKREQ (broadcast)\n"));

  /* destination is always 0xFFFFFFFF (broadcast) 
  ** and source is always 0.0.0.0, even if we know our own address.
  */
  if (__icmp_send( 0, ~0ul, (uint8_t)0xff, 0, &req, sizeof(req), NULL, 0)
                   < ((int)sizeof(req)))
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*ICMP_MASKREQ failed\n"));
  }
  else
  {
    uint32_t start;
    int oldbrkstate = set_cbreakstate(0);
    if (my_statics.debug_lev)
      DBGPRINTF(("*ICMP_MASKREQ sent\n"));

    if (timeout_secs <= 0)
      timeout_secs = 2;

    start = timer_sec_count(); 
    while (!req.mask && (timer_sec_count() - start) <= ((unsigned)timeout_secs))
    {
      if (blocker_yield(YIELDFLAG_HARD|YIELDFLAG_CHKBRK))
        break;
    }
    if (!req.mask)
    {
      if (my_statics.debug_lev)
        DBGPRINTF(("*ICMP_MASKREQ ETIMEDOUT or EINTR.\n"));
    }
    set_cbreakstate(oldbrkstate);
  }

  pending_mask_req = NULL;  
  return NTOHL(req.mask);
#endif
}

/* ======================================================================= */
/*                      ROUTER solicitation/advertisement                  */  
/* ======================================================================= */

#if defined(INCL_ICMP_ROUTERDISC_STUFF)
static int __router_solic_ack_wait = 0;
#endif

/*
** icmp_router_solic() needs to yield and MAY ONLY BE CALLED BY ARP!
** (ARP knows when it is safe to yield).
*/
int icmp_router_solic(void)
{
#if !defined(INCL_ICMP_ROUTERDISC_STUFF)
  return 0;
#else
  /* router solicitation packet looks like this:
  * struct 
  * {
  *   icmp_header hdr;
  *   uint32_t unused;
  * }
  */
  char buf[sizeof(struct icmp_header)+sizeof(uint32_t)];
  struct icmp_header *icmp_hdr = (struct icmp_header *)&buf[0];

  memset( buf, 0, sizeof(buf) );
  icmp_hdr->type = ICMP_ROUTERSOLICIT;

  if (my_statics.debug_lev)
    DBGPRINTF(("*ICMP sending ROUTERSOLICIT (broadcast)\n" ));

  __router_solic_ack_wait = 0;
  if (__icmp_send( _ip_get_ipaddr(), ((uint32_t)-1) /* bcast */, 
            5 /* TTL */, 0 /* TOS */, buf, sizeof(buf), NULL, 0 ) != 0)
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*ICMP ROUTERSOLICIT tx failed\n" ));
  }
  else
  {
    unsigned long start = timer_msec_count();
    while (__router_solic_ack_wait < 1 &&
           (timer_msec_count() - start) < 1000ul)
    {
      blocker_yield(YIELDFLAG_HARD);
    }
  }
  return __router_solic_ack_wait;
#endif
}       

static void __handle_router_advert( const struct icmp_header FAR *pkt, 
                                    unsigned int pkt_len, uint16_t pktid )
{
#if !defined(INCL_ICMP_ROUTERDISC_STUFF)
  pkt = pkt; pkt_len = pkt_len; 
#else
  /* router advert packet looks like this:
  * struct 
  * {
  *   icmp_header hdr;
  *   uint8_t  num_entries;       
  *   uint8_t  entry_size;   // size in 32bit words (2 for IPv4)
  *   uint16_t  lifetime;     // in seconds
  *   struct {
  *     uint32_t address;
  *     uint32_t preference;
  *   } entry[0..0xff];
  * }
  */
  const uint8_t FAR *pos    = ((const uint8_t FAR *)pkt)+sizeof(struct icmp_header);
  const uint32_t FAR *entry = ((const uint32_t FAR *)&pos[4]);
  unsigned int num_entries, entry_size, num, num_added = 0;
  uint16_t lifetime;

  if (pkt_len < (sizeof(struct icmp_header)+4))
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*ICMP-%04x ROUTERADVERT too short\n", pktid ));
    return;
  }

  num_entries   = pos[0];
  entry_size    = pos[1]; /* size in 32bit words (>=2) */
  lifetime      = NTOHS(*((const uint16_t *)&pos[2]));

  if (my_statics.debug_lev)
    DBGPRINTF(("*ICMP-%04x Router Advert. num_entries=%u, entry_size=%u\n",
                          pktid, num_entries, entry_size ));                

  if (pkt_len < (sizeof(struct icmp_header)+4+(num_entries * (entry_size*4))))
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*ICMP-%04x ROUTERADVERT too short\n", pktid ));
    return;
  }

  if (entry_size >= 2)
  {
    for (num = 0; num < num_entries; num++)
    {
      if (my_statics.debug_lev)
        DBGPRINTF(("*ICMP-%04x ROUTERADVERT %I pref=%ld (lifetime=%u)\n",
                         pktid, entry[0], entry[1], lifetime ));
      _arp_add_gateway( NTOHL(entry[0]), 0, 0 );
      entry += entry_size;
      num_added++;
    }
  }
  __router_solic_ack_wait += num_added;

  lifetime = lifetime; /* unused unless debugging */
#endif
  pktid = pktid; /* unused unless debugging */
  return;
}

/* ======================================================================= */
/*                         ICMP_REDIRECT                                   */  
/* ======================================================================= */

/* __handle_redirect() is called by the icmp_handler when it gets an 
** ICMP_REDIRECT message. It forwards the info to ARP.
*/
static void __handle_redirect(const in_Header FAR *ipdgram,
                              const struct icmp_header FAR *pkt, 
                              unsigned int pkt_len, uint16_t pktid )
{
#if !defined(INCL_ICMP_REDIRECT_STUFF)
  ipdgram = ipdgram; pkt = pkt; pkt_len = pkt_len;
#else
  /* redir packet looks like this:
  * struct 
  * {
  *   icmp_header hdr;
  *   uint32_t ipaddr;         // first_hop address to send to in the future.
  *   in_Header ipdgram;    // original ip-header
  * }
  */
  uint32_t redir_to, orig_dest;
  const in_Header FAR *orig_ipdgram;

  if (pkt_len < (sizeof(struct icmp_header)+sizeof(uint32_t)+sizeof(in_Header)))
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*ICMP-%04x Redir too short\n", pktid ));
    return;
  }

  redir_to     = NTOHL(*((const uint32_t FAR *)
                       (((const uint8_t FAR *)pkt)+sizeof(*pkt))));
  orig_ipdgram = (const in_Header FAR *)( ((const uint8_t FAR *)pkt) + 
                                   sizeof(*pkt)+sizeof(uint32_t) );
  orig_dest    = NTOHL(orig_ipdgram->destination);

  if (pkt->code > 3) /* 00000011b */
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*ICMP-%04x Redir has unknown code (%u)\n", pktid, 
                                                    pkt->code ));
    return; /* bad code */
  }

  if (my_statics.debug_lev)
    DBGPRINTF(("*+ICMP-%04x Redir (%s%s) for %hI->%hI\n", pktid, 
                  ((pkt->code & 2) ? ("TOS and ") : ("")),
                  ((pkt->code & 1) ? ("Host") : ("Network")),
                           orig_dest, redir_to ));

  if (!is_local_addr(NTOHL(orig_ipdgram->source)))
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*-ICMP-%04x Fake Redir (orig not from us)\n", pktid )); 
    return;
  }

  if (_arp_record_redirect(orig_dest, redir_to, NTOHL(ipdgram->source))!=0)
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*-ICMP-%04x ARP sez fake Redir\n", pktid )); 
  }
  else 
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*-ICMP-%04x Redir recorded %hI=>%hI\n", 
                                  pktid, orig_dest, redir_to ));
  }
  
#endif
  pktid = pktid; /* unused unless debugging */
  return;
}

/* ======================================================================= */

#if !defined(NDEBUG)
static void __code_to_name(uint8_t val, char *buf)
                              
{
  if (val >= 100)
    *buf++ = (char)('0' + (val/100));
  if (val >= 10)
    *buf++ = (char)('0' + ((val/10)%10));
  *buf++ = (char)('0' + (val%10));
  *buf++ = ' ';
  *buf++ = '(';
  *buf++ = '?';
  *buf++ = '?';
  *buf++ = ')';
  *buf++ = '\0';
  return;
}
#endif

#if !defined(NDEBUG)
static const char *__get_icmp_type_name(uint8_t icmp_type)
{
  static const char *names[] = 
  {
    "0 (Echo Reply)",  
    ((const char *)0), 
    ((const char *)0),
    "3 (Destination Unreachable)",
    "4 (Source Quench)",
    "5 (Redirect)",    
    ((const char *)0), 
    ((const char *)0),
    "6 (Echo Request)",
    "7 (Router Advert)",
    "8 (Router Solic)",
    "9 (TTL exceeded)",
    "10 (Param Problem)",
    "11 (Tstamp Message)",
    "12 (Tstamp Reply)",
    "13 (Info Request)",
    "14 (Info Reply)",
    "15 (Mask Request)",
    "16 (Mask Reply)"
  };
  const char *p = (const char *)0;
  if (icmp_type < (sizeof(names)/sizeof(names[0])))
    p = names[icmp_type];
  if (!p)
  {
    static char buf[sizeof("nnn (unrecognized msg_type)")];
    __code_to_name( icmp_type, buf );
    p = buf;
  }
  return p;
}
#endif /* if !defined(NDEBUG) */

#if !defined(NDEBUG)
static const char *__get_type_code_name(uint8_t icmp_type, uint8_t type_code)
{
  const char *p = (const char *)0;
  if (icmp_type == ICMP_TIMXCEED)
  {
    static const char *names[] = {
      "0 (TTL exceeded in transit)",
      "1 (Frag ReAsm time exceeded)"
    };
    if (type_code < (sizeof(names)/sizeof(names[0])))
      p = names[type_code];
  }
  else if (icmp_type == ICMP_UNREACH)
  {
    static const char *names[] = {
      "0 (Network Unreachable)",
      "1 (Host Unreachable)",
      "2 (Protocol Unreachable)",
      "3 (Port Unreachable)",
      "4 (Fragmentation needed and DF set)",
      "5 (Source Route Failed)",
      "6 (Network unknown)",
      "7 (Host unknown)",
      "8 (Source host isolated)",
      "9 (Net-access denied)",
      "10 (Host-access denied)",
      "11 (Bad TOS for net)",
      "12 (Bad TOS for host)",
      "13 (Admin prohibited)",
      "14 (Host precedence violation)",
      "15 (Precedence cutoff)"
    };
    if (type_code < (sizeof(names)/sizeof(names[0])))
      p = names[type_code];
  }
  if (!p)
  {
    static char buf[sizeof("nnn (unrecognized type_code)")];
    __code_to_name( type_code, buf );
    p = buf;
  }
  return p;
}
#endif /* if !defined(NDEBUG) */

/* ======================================================================= */

void icmp_retour( int soicmp_err, const void FAR *orig_ipdgram,
                  const void FAR *hwaddr/* optional */, int callers_debug_lev)
{
  /* retour packet looks like this:
  * struct 
  * {
  *   icmp_header hdr;
  *   uint32_t unused;
  *    followed by original IP hdr (including options)
  *    plus 64 bits of whatever follows that.
  * }
  */
  const in_Header FAR *ipdgram = (const in_Header FAR *)orig_ipdgram;
  uint32_t dest_addr = NTOHL(ipdgram->source);
  if (/* ipdgram->proto != IPPROTO_ICMP && */ /* ICMPs are not sent for ICMPs */
      dest_addr && !is_broadcast(dest_addr) && !is_multicast(dest_addr)
      && is_local_addr(NTOHL(ipdgram->destination)) )
  {
    /*  0                   1                   2                   3
     *  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
     * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     * |     Type      |     Code      |          Checksum             |
     * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     * |                             unused                            |
     * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
     * |      Internet Header + 64 bits of Original Data Datagram      |
     * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    */
    uint8_t icmp_type = 0, type_code = 0;
    char buf[sizeof(in_Header)+1+1+2+4+64+(64/8)]; 
    unsigned int len;

    switch (soicmp_err)
    {
      case SOICMPERR_UNREACH_HOST:  icmp_type=ICMP_UNREACH; type_code = 1; 
                                    break;
      case SOICMPERR_UNREACH_PROTO: icmp_type=ICMP_UNREACH; type_code = 2; 
                                    break;
      case SOICMPERR_UNREACH_PORT:  icmp_type=ICMP_UNREACH; type_code = 3; 
                                    break;
      case SOICMPERR_CANTFRAG:      icmp_type=ICMP_UNREACH; type_code = 4; 
                                    break;
      case SOICMPERR_TIMXCEED_FRAG: icmp_type=ICMP_TIMXCEED; type_code = 1; 
                                    break;
      case SOICMPERR_TIMXCEED_TTL:  icmp_type=ICMP_TIMXCEED; type_code = 0; 
                                    break;
      case SOICMPERR_SOURCEQUENCH:  icmp_type=ICMP_SOURCEQUENCH; type_code=0;
                                    break;
      case SOICMPERR_PARAMPROB:     /* fallthrough */
      default:                       
      {
        if (callers_debug_lev) 
          DBGPRINTF(("*ICMP-%04x: unknown/unhandled soicmp_err %d...\n", 
                   NTOHS(ipdgram->identification), soicmp_err ));
        return;
      }
    } /* switch() */

    if (hwaddr) /* direct tx. Used from IP input handlers. */
    {
      in_Header *in_hdr;

      /* build icmp data part of packet */
      len = NTOHS(ipdgram->length) - in_GetHdrLen(ipdgram);
      if (len > (64/8))
        len = (64/8);
      len += in_GetHdrLen(ipdgram);
      FAR_MEMCPY( &buf[sizeof(in_Header)+1+1+2+4], ipdgram, len );

      /* build icmp header part of packet */
      buf[sizeof(in_Header)+0] = icmp_type;
      buf[sizeof(in_Header)+1] = type_code;
      *((uint16_t  *)&buf[sizeof(in_Header)+2]) = 0; /* checksum field */
      *((uint32_t *)&buf[sizeof(in_Header)+4]) = 0; /* unused field */
      len += 1+1+2+4;
      *((uint16_t *)&buf[sizeof(in_Header)+2]) = 
                     ~checksum( &buf[sizeof(in_Header)+0], len );

      /* build IP header part of packet */
      in_hdr = (in_Header *)&buf[0];
      len += sizeof(in_Header);
      in_hdr->hdrlenver      = (uint8_t)((0x04u<<4) | (sizeof(in_Header)>>2));
      in_hdr->tos            = ipdgram->tos;
      in_hdr->length         = HTONS((uint16_t)len);
      in_hdr->identification = ipdgram->identification; /* same ID */
      in_hdr->frag_ofs       = 0;
      in_hdr->ttl            = 0xff;
      in_hdr->proto          = IPPROTO_ICMP;
      in_hdr->checksum       = 0;
      in_hdr->source         = ipdgram->destination;
      in_hdr->destination    = ipdgram->source;
      in_hdr->checksum       = ~checksum(in_hdr, sizeof(in_Header) );

      _eth_send( ETHERTYPE_IP, hwaddr, in_hdr, len );
      if (icmp_type == ICMP_TIMXCEED && type_code == 0 &&
          !is_local_addr(dest_addr))
      { 
        /* Send 3 more times. Yes, I know that the most probable cause for a 
         * timeout is congestion, and that sending the packet 4 times makes
         * the congestion worse.
        */
        int i;
        for (i = 0; i < 3; i++)
          _eth_send( ETHERTYPE_IP, hwaddr, in_hdr, len );
      }
    }
    else /* no hardware address available. Use regular indirect tx */
    {
      memset( buf, 0, sizeof(buf) );
      ((struct icmp_header *)&buf[0])->type = icmp_type;
      ((struct icmp_header *)&buf[0])->code = type_code;

      len = in_GetHdrLen(ipdgram);
      len += 64 / 8; /* followed by 64bits of whatever follows hdr */
      if (len > NTOHS(ipdgram->length))
        len = NTOHS(ipdgram->length);

      __icmp_send( NTOHL(ipdgram->destination), NTOHL(ipdgram->source), 
                   (uint8_t)0xff, ipdgram->tos, buf, sizeof(buf), ipdgram, len );
    }

    if (callers_debug_lev) 
      DBGPRINTF(("*ICMP-%04x: sent (type %s, code %s) to %hI...\n", 
                   NTOHS(ipdgram->identification), 
                   __get_icmp_type_name(icmp_type),
                   __get_type_code_name(icmp_type,type_code), ipdgram->source ));
  }
  return;
}

/* ---------------------------------------------------------------- */

static void __handle_retour( const in_Header FAR *ipdgram, 
                             const struct icmp_header FAR *pkt, 
                             unsigned int pkt_len, uint16_t pktid )
{
  const in_Header FAR *orig_ipdgram;

  if (pkt_len < (sizeof(struct icmp_header)+sizeof(uint32_t)+sizeof(in_Header)))
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*ICMP-%04x Retour too short\n", pktid ));
    return;
  }

  orig_ipdgram = (const in_Header FAR *)( ((const uint8_t FAR *)pkt) + 
                                    sizeof(*pkt)+sizeof(uint32_t) );

  if (pkt_len < (sizeof(struct icmp_header)+sizeof(uint32_t)+
                in_GetHdrLen(orig_ipdgram)+(64/8)))
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*ICMP-%04x Retour too short\n", pktid ));
    return;
  }

  if (is_local_addr(NTOHL(orig_ipdgram->source)))
  {
    uint8_t icmp_type = pkt->type;
    uint8_t type_code = pkt->code;
    int soicmperr_code = 0;

    switch (icmp_type)
    {
      /* These are the only messages we handle here. 
      ** (we shouldn't be getting any others anyway).
      */
      case ICMP_UNREACH:
         if (my_statics.debug_lev)
           DBGPRINTF(("*ICMP IN-%04x: UNREACH: type %s\n", pktid, 
                    __get_type_code_name(icmp_type, type_code) ));
         if (type_code == 2) /* proto unreachable */
           soicmperr_code = SOICMPERR_UNREACH_PROTO;
         else if (type_code == 3) /* port unreachable */
           soicmperr_code = SOICMPERR_UNREACH_PORT;
         else if (type_code == 4) /* Fragmentation needed and DF set */
           soicmperr_code = SOICMPERR_CANTFRAG;
         else if (type_code >= 9) /* various access denied/TOS messages */
           soicmperr_code = SOICMPERR_UNREACH_HOST;
         else /* host/net unreachable, host/net isolated, host unknown */
           soicmperr_code = SOICMPERR_UNREACH_HOST;
         break;
      case ICMP_TIMXCEED:
         if (my_statics.debug_lev)
           DBGPRINTF(("*ICMP IN-%04x TIMXCEED: type %s\n", pktid, 
                    __get_type_code_name(icmp_type, type_code) ));
         if (type_code == 0)
           soicmperr_code = SOICMPERR_TIMXCEED_TTL;
         else if (type_code == 1)
           soicmperr_code = SOICMPERR_TIMXCEED_FRAG;
         else
           soicmperr_code = 0; /* huh? */
         break;
      case ICMP_SOURCEQUENCH:
         if (my_statics.debug_lev)
           DBGPRINTF(("*ICMP IN-%04x SOURCE QUENCH\n", pktid ));
         soicmperr_code = SOICMPERR_SOURCEQUENCH;
         break;
      case ICMP_PARAMPROB:
         if (my_statics.debug_lev)
         {
           /* paramprob packet looks like this:
            * struct 
            * {
            *   icmp_header hdr;
            *   uint8_t pointer;
            *   uint8_t unused [3];
            *   in_Header orig_ipdgram;
            * };
           */
           if (type_code)
             DBGPRINTF(("*ICMP IN-%04x PARAMPROB code %d\n", pktid, 
                            type_code));
           else 
              DBGPRINTF(("*ICMP IN-%04x PARAMPROB at offset %u\n", pktid,
                    *(((const uint8_t FAR *)pkt)+sizeof(struct icmp_header)) ));
         }
         soicmperr_code = SOICMPERR_PARAMPROB;
         break;
      default:
         soicmperr_code = 0;
         break;
    }
    if (orig_ipdgram->proto == IPPROTO_ICMP)
    {
      if (ping_ctrl.hk)
        (*(ping_ctrl.hk))(ipdgram);
    }
    else if (soicmperr_code != 0) /* not all msgs get passed */
    {       
      _ip_icmp_callback( orig_ipdgram, soicmperr_code );
    }
  }
  pktid = pktid; /* unused unless debugging */
  return;
}

/* ======================================================================= */

/*
 * Handler for incoming ICMP packets.  Called from IP handler.
*/
static void __icmp_handler(const in_Header FAR *ipdgram, int protohndlr_flags)
{
  int  is_bcast = ((protohndlr_flags & PROTOHNDLR_FLAGS_LLBCAST) != 0);
  uint16_t pktid    = NTOHS(ipdgram->identification);
  const struct icmp_header FAR *pkt;
  unsigned int pkt_len;

  if (my_statics.debug_lev)
    DBGPRINTF(("*+ICMP-%04x RECV %I->%I\n", pktid, 
                        ipdgram->source, ipdgram->destination ));

  pkt_len = in_GetHdrLen(ipdgram);
  pkt = (const struct icmp_header FAR *)
        (((const uint8_t FAR *)ipdgram) + pkt_len);
  pkt_len = NTOHS(ipdgram->length) - pkt_len;

  if (pkt_len < sizeof(struct icmp_header))
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*-ICMP-%04x: DROP (short header)\n", pktid ));
    return;
  }

  if (checksum(pkt, pkt_len) != 0xFFFF)
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*-ICMP-%04x: DROP (bad checksum=%04x)\n", pktid,
                                 checksum(pkt, pkt_len) ));
    return;
  }
   
  switch (pkt->type)
  {
    case ICMP_ECHOREPLY: /* must be unicast */
         __handle_echo_reply( ipdgram, pktid );
         break;

    case ICMP_ECHO:/* must be unicast */
         __handle_echo( ipdgram, pktid, is_bcast );
         break;

    case ICMP_REDIRECT:/* must be unicast */
         __handle_redirect( ipdgram, pkt, pkt_len, pktid );
         break;

    case ICMP_UNREACH:      /* must be unicast */
    case ICMP_SOURCEQUENCH: /* must be unicast */
    case ICMP_TIMXCEED:     /* must be unicast */
    case ICMP_PARAMPROB:    /* must be unicast */
         __handle_retour( ipdgram, pkt, pkt_len, pktid );
         break;

    case ICMP_ROUTERADVERT:/* may be unicast or broadcast */
         __handle_router_advert( pkt, pkt_len, pktid );
         break;

    case ICMP_MASKREPLY:/* will be broadcast */
         __handle_mask_reply( pkt, pkt_len, pktid );
         break;

    default:
         if (my_statics.debug_lev)  
           DBGPRINTF(( "*ICMP-%04x Unhandled type %s, code %s\n", pktid, 
                 __get_icmp_type_name(pkt->type),
                 __get_type_code_name(pkt->type,pkt->code) ));
         break;
  }

  if (my_statics.debug_lev >= 3)
    DBGDUMP(("*-ICMP", pktid, pkt, pkt_len));
  else if (my_statics.debug_lev)
    DBGPRINTF(( "*-" ));

  return;
}

/* ======================================================================= */
/*                          protocol initialization                        */
/* ======================================================================= */

static void __stack_shutdown(void)
{
  if (my_statics.debug_lev)
    DBGPRINTF(("*ICMP shutdown done.\n"));
  /* no shutdown necessary */
}

static void __ctor_icmp(void)
{
  static const struct config_table FAR my_cfg[] = {
     { "ICMP.DEBUG",    ARG_ATOI,   (void FAR *)&my_statics.debug_lev },
     { ((const char FAR *)0), 0,    ((void FAR *)0)  }
   };
  _conf_add_confextn( &my_cfg[0], NULL );

  /* register "downwards" */
  _ip_proto_register(IPPROTO_ICMP, __icmp_handler, __stack_shutdown, 
                                      &my_statics.debug_lev );

  return;
}

#define __SETUP_CONSTRUCTOR_FOR__ __ctor_icmp
#include "ctordtor.h"

