/*
 * 1) Internet Group Management Protocol (IGMP) as per RFC 1112.
 *    The IGMP protocol is to allow multicast routers to establish group
 *    membership on attached networks. Thus allowing the router to not send
 *    data that is not needed.
 *
 * 2) IP Multicasting extensions as per RFC 1112.
 *    These extensions include routines to detect multicast addresses,
 *    transform Multicast IP addresses to Multicast Ethernet addresses, as
 *    well as a mechanism to join and leave multicast groups.
 *
 * Jim Martin
 * Rutgers University - RUCS-TD/NS
 * jim@noc.rutgers.edu
 * 6/6/93
 *
 *
 * The only global made visible by this module is:
 * int __IPPROTO_IGMP__ = IPPROTO_IGMP;
 * This may be used to force linkage.
*/
#define __COMPILING_IGMP__
#define __OSILAYER__ 3 /* well, half way between 3 and 4 */

#include "target.h"   /* standard types/includes */
#include "blocker.h"  /* blocker_perdhook_add() */
#include "timer.h"    /* timer_sec_count() */
#include "debug.h"    /* DBGPRINTF() */
#include "linklayr.h" /* _eth_mcast_*() */
#include "inchksum.h" /* checksum() */
#include "inaddr.h"   /* is_local_addr()/is_multicast() */
#include "printk.h"   /* _printk() */
#include "loadcfg.h"  /* config_table(), _conf_add_confextn() */
#include "ipbase.h"   /* _ip_output_x(), _ip_proto_register(), IPPROTO_IGMP */
#include "ipmcast.h"  /* _ip_mcast_set_grpopfn() */

const int __IPPROTO_IGMP__ = IPPROTO_IGMP; /* the only public in this file */

static int __mcast_is_in_group(uint32_t ina);
static int __mcast_join_group(uint32_t ina);
static int __mcast_leave_group(uint32_t ina);

/* ==================================================================== */
/*                 structures/constants used internally                 */
/* ==================================================================== */

/*
 * Internet Group Management Protocol Packet
 */
#pragma pack(1) /* align all structure members on byte boundaries */
typedef struct IGMP_packet {
#if defined(__BIG_ENDIAN__)
        unsigned char version : 4;
        unsigned char type    : 4;
#elif defined(__TURBOC__)
        unsigned int  type    : 4;
        unsigned int  version : 4;
#else
        unsigned char type    : 4;
        unsigned char version : 4;
#endif
        uint8_t   mbz;
        uint16_t   checksum;
        uint32_t  address;
      } IGMP_packet;
#pragma pack()

#define IGMP_VERSION  1
#define IGMP_QUERY    1
#define IGMP_REPORT   2


/*
 * Stuff for Multicast Support - JRM 6/7/93
 */
#define CLASS_D_MASK  0xE0000000UL  /* the mask that defines IP Class D  */
#define IPMULTI_MASK  0x007FFFFFUL  /* to get the low-order 23 bits      */
#define ETH_MULTI     0x01005EUL    /* high order bits of multi eth addr */
#define ALL_SYSTEMS   0xE0000001UL  /* the default mcast addr 224.0.0.1  */
#define IS_MULTICAST(__ina) ((((uint32_t)__ina) & 0xF0000000UL) == CLASS_D_MASK)

struct mcast_group_entry 
{
  uint32_t ina;            /* IP address of group                 */
  uint32_t lastquerytime;  /* time (secs) last IGMP query came in */
  unsigned processes;      /* number of interested processes      */
};

/* ===================================================================== */
/*                           private data                                */
/* ===================================================================== */

static struct 
{
  struct mcast_group_entry __mgroups[20 /*number of groups addable*/];
  int debug_lev;
} FAR my_statics;

#define __mgroups (my_statics.__mgroups)

/* ===================================================================== */
/*                               debug helper                            */
/* ===================================================================== */

#if defined(NDEBUG)
#define __dbg_dump_header(__i, __p, __h) NO_OPERATION()
#else
static void __dbg_dump_header( int is_output, uint16_t pktid, 
                               const IGMP_packet FAR *hdr )
{
  static const char *igmp_types[] = { "0",  "Query", "Report", "3" };

  if (is_output)
    DBGPRINTF(("*IGMP OUT-xxxx: header dump (size=%u)\n", sizeof(*hdr) ));
  else
    DBGPRINTF(("*IGMP IN-%04x: header dump (size=%u)\n", pktid, sizeof(*hdr) ));

  DBGPRINTF(("*  -> version   = %u\n", hdr->version ));
  DBGPRINTF(("*  -> type      = %u (%s)\n", hdr->type,
      hdr->type < (sizeof(igmp_types)/sizeof(igmp_types[0])) ? 
                         igmp_types[hdr->type] : "unknown" ));
  DBGPRINTF(("*  -> mbz       = %u\n",  hdr->mbz ));
  DBGPRINTF(("*  -> checksum  = 0x%04x\n", hdr->mbz ));
  DBGPRINTF(("*  -> address   = %I\n", hdr->address ));
  return;                                                  
}
#endif

/* ===================================================================== */
/*   PART 1: Internet Group Management Protocol (IGMP) as per RFC 1112.  */
/* ===================================================================== */

/* 
 * igmp_handler - handles the incoming IGMP packets
 * void igmp_handler (in_Header *ip) 
 * Where:
 *      ip    is the IP packet in question
 *
 * Returns: None
 *
 */
static void __igmp_handler (const in_Header FAR *ip, int protohndlr_flags)
{
  /*
   * Note that even when there is no igmp support, 
   * the ip_layer does not need to send an 'ICMP unknown proto'
   * because the IGMP handler is purely passive and not 
   * supporting it is equivalent to supporting it without any groups.
  */
  const IGMP_packet FAR *igmp = 
   (const IGMP_packet FAR *)(((const uint8_t FAR *)ip) + in_GetHdrLen(ip));
  uint16_t igmplen = (uint16_t)(NTOHS(ip->length) - in_GetHdrLen(ip));
  uint16_t pktid   = NTOHS(ip->identification);
  uint32_t host;
  unsigned int i;
  int found = 0;

  protohndlr_flags = protohndlr_flags; /* not referenced */
  pktid = pktid; /* not referenced unless debugging */

  if (igmplen < sizeof(*igmp))
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*IGMP IN-%04x: DROP (too short)\n", pktid));
    return;
  }

  if (checksum(igmp, igmplen) != 0xFFFF)
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*IGMP IN-%04x: DROP (bad chksum)\n", pktid));
    return;
  }

  if (my_statics.debug_lev >= 2)
    __dbg_dump_header( 0, pktid, igmp );
  if (my_statics.debug_lev >= 3)
    DBGDUMP(("*IGMP", pktid, igmp+1, NTOHS(ip->length)-sizeof(*igmp) ));

  host = NTOHL(igmp->address);

  /* Determine whether this is a report or a query
   */
  if (igmp->type == IGMP_QUERY)
  {
    found = 0;
    if (my_statics.debug_lev)
      DBGPRINTF(("*IGMP IN-%04x: IGMP_QUERY for %I\n", pktid, igmp->address));
    for (i = 0; i < (sizeof(__mgroups)/sizeof(__mgroups[0])); i++)
    {
      if (__mgroups[i].processes > 0      &&
          __mgroups[i].ina != ALL_SYSTEMS &&
          __mgroups[i].lastquerytime == 0)
      {
        __mgroups[i].lastquerytime = timer_sec_count();
        if (__mgroups[i].lastquerytime == 0)
          __mgroups[i].lastquerytime--;
        found = 1;
      }
    }
    if (!found)
    {
      if (my_statics.debug_lev)
        DBGPRINTF(("*IGMP IN: DROP IGMP_QUERY (none of ours)\n"));
    }
  }
  else if (igmp->type == IGMP_REPORT)
  {
    found = 0;
    if (my_statics.debug_lev)
      DBGPRINTF(("*IGMP IN-%04x: DROP IGMP_REPORT for %I\n", pktid, igmp->address));

    if (host == ALL_SYSTEMS)
    {
      if (my_statics.debug_lev)
        DBGPRINTF(("*IGMP IN-%04x: DROP IGMP_REPORT (ALL_SYSTEMS rep)\n", pktid));
    }
    else
    {
      for (i = 0; i < (sizeof(__mgroups)/sizeof(__mgroups[0])); i++)
      {
        if (__mgroups[i].processes > 0  &&
            __mgroups[i].ina == host)
        {
          __mgroups[i].lastquerytime = 0;
          found = 1;
          break;
        }
      }
      if (!found && my_statics.debug_lev)
        DBGPRINTF(("*IGMP IN-%04x: DROP IGMP_REPORT (none of ours)\n",pktid));
    }
  }

  return;
}

/* ===================================================================== */
/*                                   xxx                                 */
/* ===================================================================== */

/* 
 * __igmp_report - send a IGMP Report packet
 *
 * int __igmp_report (uint32_t ina)
 * Where:
 *      ina is the IP address to report.
 *
 * Returns:
 *     -1   if unable to send report
 *    else  report was sent successfully
 */
static int __igmp_report (uint32_t ina)
{
  IGMP_packet igmp;

  /* fill in the igmp packet
   */
  igmp.version  = IGMP_VERSION;
  igmp.type     = IGMP_REPORT;
  igmp.mbz      = 0;
  igmp.address  = HTONL(ina);
  igmp.checksum = 0;
  igmp.checksum = ~checksum (&igmp,sizeof(igmp));
  
  if (my_statics.debug_lev)
    DBGPRINTF(("*IGMP-tx IGMP_REPORT for %I\n", igmp.address ));

  if (my_statics.debug_lev >= 2)
    __dbg_dump_header( 1, 0, &igmp );

  return _ip_output_2(_ip_get_ipaddr(), (uint32_t)ALL_SYSTEMS, IPPROTO_IGMP,
                      _ip_get_default_ttl(), _ip_get_default_tos(), 0,
                      &igmp, sizeof(igmp), NULL, 0, NULL);
}

/* -------------------------------------------------------------------- */

/* 
 * __igmp_check_reports - check all groups
 * to see if we owe an IGMP report. 
 *
 * Called as a periodic callback registered with blocker_perdhook_add()
*/
static void __igmp_check_reports(void)
{
  unsigned int i;
  unsigned long now = timer_sec_count();
  for (i = 0; i < (sizeof(__mgroups)/sizeof(__mgroups[0])); i++)
  {
    if (__mgroups[i].processes > 0 &&
        __mgroups[i].ina != ALL_SYSTEMS &&
        __mgroups[i].lastquerytime != 0 &&
        (now != __mgroups[i].lastquerytime)) /* 1-1000 ms random interval */
    {
      __igmp_report (__mgroups[i].ina);
      __mgroups[i].lastquerytime = 0;
    }
  }
  return;
}

/* ===================================================================== */
/*           PART 2: IP Multicasting extensions as per RFC 1112.         */
/* ===================================================================== */

/*
 * __initialize_mcast_if_needed() is called by all the extension
 * functions to ensure that mcast is supportable, and if so, to 
 * ensure that the ALL_SYSTEMS group is joined at the first 
 * possible opportunity. 
*/
static int __initialize_mcast_if_needed(void)
{
  static int cond_mcast_on = 0; /* 0=unknown, -1=error, +1=ok */
  if (cond_mcast_on == 0)
  {
    cond_mcast_on = -1;
    _ip_get_ipaddr(); /* force IP to initialize */

    if (_eth_get_hwaddrlen() == 6 /* the only hwaddr size we know */
      && _eth_mcast_get_max_groups() > 0) /* hardware supports mcast */
    {
      if (blocker_perdhook_add(__igmp_check_reports)==0) /* periodic callback */
      {
        cond_mcast_on = +1;
        /* all multicast level 2 systems must join at startup */
        /* might get called recursively now */
        if (!__mcast_join_group(ALL_SYSTEMS))
          cond_mcast_on = -1;
      }
    }
  }
  return (cond_mcast_on > 0);
}

/* ===================================================================== */

/* 
 * __multi_to_eth - calculates the proper ethernet address for a given
 *                  IP Multicast address. Used internally by
 *                  _mcast_join_group() and _mcast_leave_group().
 *
 * int __multi_to_eth (uint32_t ip, void *eth)
 * Where:
 *       ip:  IP address to be converted
 *       eth: pointer to ethernet MAC address
 */
static void __multi_to_eth (uint32_t ip, void *_eth)
{
  unsigned char *e = (unsigned char *)_eth;
  uint32_t top = ETH_MULTI;

  ip &= IPMULTI_MASK;
  e[0] = (unsigned char)(top >> 16);
  e[1] = (unsigned char)((top >> 8) & 0xFF);
  e[2] = (unsigned char)(top & 0xFF);
  e[3] = (unsigned char)(ip >> 16);
  e[4] = (unsigned char)((ip >> 8) & 0xFF);
  e[5] = (unsigned char)(ip & 0xFF);
}

/* ===================================================================== */
/*                            GROUP OPERATIONS                           */
/* ===================================================================== */

/* 
 * mcast_leave_group - leaves a multicast group
 *
 * int igmp_leave_group( uint32_t ina )
 * Where:
 *       ina IP address of the group to be left
 * Returns:
 *       1   if the group was left successfully
 *       0   if attempt failed
 */
static int __mcast_leave_group(uint32_t ina)
{
  char list[MAX_LINKADDR_SIZE * (sizeof(__mgroups)/sizeof(__mgroups[0]))];
  unsigned int i, addr_len, listlen, numused = 0;
  int entry = -1;

  if (!__initialize_mcast_if_needed())
    return (0);

  /* first verify that it's a valid mcast address
   */
  if (!IS_MULTICAST(ina))
     return (0);

  /* determine if the group has more than one interested
   * process. if so, then just decrement .processes and
   * return
   */
  for (i = 0; i < (sizeof(__mgroups)/sizeof(__mgroups[0])); i++)
  {
    if (__mgroups[i].processes > 0 && ina == __mgroups[i].ina)
    {
      __mgroups[i].processes--;
      if (__mgroups[i].processes > 0)
        return (1);
      entry = i;
      break;
    }
  }

  /* was it one of the groups we know about?
  */
  if (entry == -1) /* its not in the list! */
    return (0);

  /* alas, we need to leave it
  */
  addr_len = _eth_get_hwaddrlen();
  listlen  = 0;
  numused  = 0;
  for (i = 0; i < (sizeof(__mgroups)/sizeof(__mgroups[0])); i++)
  {
    if (__mgroups[i].processes > 0)
    {
      __multi_to_eth(__mgroups[i].ina, &list[listlen] );
      listlen += addr_len;
      numused++;
    }
  }

  i = _eth_mcast_get_max_groups();
  if (numused >= i)
    listlen = (addr_len * i);
  else if ((numused+1) > i)
    _eth_mcast_set_mode(ETH_MCAST_MODE_LIMITED);
  else if (numused == 0)
    _eth_mcast_set_mode(ETH_MCAST_MODE_OFF);

  _eth_mcast_set_list( &list[0], listlen );
  return (1);
}

/* --------------------------------------------------------------------- */

/* 
 * mcast_join_group - join a multicast group
 *
 * int igmp_join_group( uint32_t ina )
 * Where:
 *       ina IP address of the group to be joined
 * Returns:
 *       1   if the group was joined successfully
 *       0   if attempt failed
 */
static int __mcast_join_group(uint32_t ina)
{
  unsigned int i, numused = 0;
  int entry = -1;

  if (!__initialize_mcast_if_needed())
    return (0);

  /* first verify that it's a valid mcast address
   */
  if (!IS_MULTICAST(ina))
     return (0);

  /* Determine if the group has already been joined.
   * As well as a) what the first free slot is, b) how many are used
   */
  for (i = 0; i < (sizeof(__mgroups)/sizeof(__mgroups[0])); i++)
  {
    if (__mgroups[entry].processes == 0)
    {
      if (entry == -1)
        entry = i;
    }
    else if (ina == __mgroups[i].ina)
    {
      __mgroups[i].processes++;
      return (1);
    }
    else 
    {
      numused++;
    }
  }

  /* alas, no...we need to join it
   */
  if (entry == -1)        /* out of slots! */
     return (0);

  /* initialize a few things
  */
  __mgroups[entry].processes = 0;
  __mgroups[entry].lastquerytime = 1; /* immediate "reply" */
  __mgroups[entry].ina       = ina;

  if ((numused+1) > ((unsigned int)_eth_mcast_get_max_groups()))
  {
    if (_eth_mcast_set_mode(ETH_MCAST_MODE_UNLIMITED) != 0)
      return (0); /* couldn't switch to unlimited mcast */
  }
  else /* room for more addresses in the hardware adapter */
  {
    /* try to stay with limited multicast as long as possible */
    unsigned int addr_len = _eth_get_hwaddrlen();
    char list[MAX_LINKADDR_SIZE * (sizeof(__mgroups)/sizeof(__mgroups[0]))];
    unsigned int listlen = 0;
    for (i = 0; i < (sizeof(__mgroups)/sizeof(__mgroups[0])); i++)
    {
      if (i == ((unsigned)entry) || __mgroups[i].processes)
      { 
        __multi_to_eth(__mgroups[i].ina, &list[listlen] );
        listlen += addr_len;
      }
    } 
    if (numused == 0)
    {
      if (_eth_mcast_set_mode(ETH_MCAST_MODE_LIMITED) != 0)
        return (0); /* couldn't switch to limited mcast */
    }
    if (_eth_mcast_set_list( &list[0], listlen ) != 0)
      return (0);
  }

  __mgroups[entry].processes++;
  return (1);
}

/* --------------------------------------------------------------------- */

/* Determine if an IP address is in one of the groups that we are in.
** Called from the IP handler to filter out packets that are known
** to be UDP+multicast in order to reduce the amount of work needed
** to be done at the higher layers.
*/
static int __mcast_is_in_group(uint32_t ip)
{
  if (__initialize_mcast_if_needed())
  {
    if (IS_MULTICAST(ip))
    {
      unsigned int i;
      for (i = 0; i < (sizeof(__mgroups)/sizeof(__mgroups[0])); i++)
      {
        if (__mgroups[i].processes > 0 &&
            __mgroups[i].ina == ip)
          return (1);
      }
    }  
  }
  return (0);
}

/* --------------------------------------------------------------------- */

/* __mcast_group_op() is what we export using 
 * the _ip_mcast_setgopfn() trampoline 
*/
static int __mcast_group_op(uint32_t ina, int op)
{
  if (op > 0)
    return __mcast_join_group(ina);
  if (op < 0)
    return __mcast_leave_group(ina);
  return __mcast_is_in_group(ina);
}

/* ==================================================================== */
/*                  transport initialization/shutdown                   */
/* ==================================================================== */

/* 
 * called when the IP stack is about to shut down.
*/
static void __igmp_shutdown(void)
{
  unsigned int i;
  if (my_statics.debug_lev)
    DBGPRINTF(("*IGMP shutdown begin\n"));
  for (i = 0; i < (sizeof(__mgroups)/sizeof(__mgroups[0])); i++)
  {
    if (__mgroups[i].processes > 0)
    {
      __mgroups[i].processes = 1;
      __mcast_leave_group(__mgroups[i].ina);
    }
  }
  if (my_statics.debug_lev)
    DBGPRINTF(("*IGMP shutdown complete\n"));
  return;
}

/* --------------------------------------------------------------------- */

static void __ctor_igmp(void)
{
  static const struct config_table FAR my_cfg[] = 
  {
    { "IGMP.DEBUG",  ARG_ATOI,     (void FAR *)&my_statics.debug_lev },
    { ((const char FAR *)0), 0,    ((void FAR *)0)  }
  };
  #if !defined(NDEBUG)
  unsigned int sz = sizeof(IGMP_packet);
  if (sz != 8)
  {
    _printk("PANIC: misaligned fields in struct IGMP_packet\n");
    exit(255);
  }
  #endif
  _conf_add_confextn( &my_cfg[0], NULL );
  FAR_MEMSET( &my_statics, 0, sizeof(my_statics) );

  /* register the protocol with ipbase.c */
  _ip_proto_register(IPPROTO_IGMP, __igmp_handler, __igmp_shutdown,
                                    &my_statics.debug_lev );
  /* tell ipmcast.c how to do group operations */
  _ip_mcast_setgopfn(0,__mcast_group_op);
}

#define __SETUP_CONSTRUCTOR_FOR__ __ctor_igmp
#include "ctordtor.h"

