/*
 * Network address conversion routines. 
 * don't use lib functions here!
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: inaddr.c,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
#include "target.h"  /* standard types/includes */
#include "ipbase.h"  /* _ip_get_ipaddr(), _ip_get_snmask() */
#include "inaddr.h"  /* ourselves */

/* ======================================================================= */
/*                         conversion primitives                           */
/* ======================================================================= */

/*
 * __inet_atox() is the base of all ASCII->bin addr conversion.
 *
 * The address is returned in *NET_ORDER* for IP addresses (inet_aton()
 * format), and in *HOST_ORDER* for net addresses (inet_network() format)
 *
 * For backwards compatibility with quirks/oddities/features in Eric's stuff,
 * it will handle square brackets ('[' and ']') around a *host* address,
 * but will not do so for a net address. If an open bracket is present, it
 * must be matched by a closing one.
 *
 * __inet_atox() returns non-zero if 'cp' was ok, otherwise it returns zero.
 * (return code compatible with inet_aton())
*/
static int __inet_atox( const char *cp, /* stringified address (source) */
                        uint32_t *destp,   /* destination buffer */
                        int minparts,   /* must have at least this many parts */
                        int addrtype )  /* 'h':inet_aton(), 'n':inet_network()*/
{
  int err = 1;

  if (cp)
  {
    int parts = 0, bracket = 0;
    unsigned long maxval = 0xfffffffful;
    unsigned long buf[4];

    if ( *cp == '[' && addrtype != 'n')
    {
      bracket = 1;
      cp++;
    }

    buf[0] = 0; err = 0;
    while (*cp && !err)
    {
      register unsigned long val = 0;
      unsigned int radix = 10, len = 0;

#if 0 /* the isc code implies this, but manpage says the last part may be >0xff */
      if (addrtype == 'n') /* if doing inet_network() all parts are <=0xff */
        maxval = 0xff;
#endif

      if ( *cp == '0' )
      {
        radix = 8;
        cp++;
        len = 1;
      }
      if ( *cp == 'X' || *cp == 'x')
      {
        radix = 16;
        cp++;
        len = 0;
      }
      while (!err)
      {
        register int c = (int)*cp;
        if (c >= '0' && c <= '9') /* ebcdic safe */
          c -= '0';
        else if (radix == 16 && c >= 'A' && c <= 'F') /* ebcdic safe */
          c = 10 + (c - 'A');
        else if (radix == 16 && c >= 'a' && c <= 'f') /* ebcdic safe */
          c = 10 + (c - 'a');
        else
          break; /* dot or bad char */
        if ( c < 0 || ((unsigned int)c) >= radix )
          err = 1;
        else
        {
          unsigned long newval;
          if (radix == 10)
            newval = ((val+(val<<2))<<1) + c;
          else if (radix == 16)
            newval = (val << 4) + c;
          else /* if (radix == 8) */
            newval = (val << 3) + c;
          if (newval < val) /* wrapped */
            err = 1;
          else if (newval > maxval)
            err = 1;
          else
          {
            val = newval;
            cp++;
            len++;
          }
        }
      } /* while (!err) */

      if (!err)
      {
        if (!( parts < 4 )) /* already did 4 */
          err = 1;
        else if (parts && buf[parts-1]>0xff) /* completed parts must be <=0xff */
          err = 1;
        else
        {
          buf[parts++] = val;
          maxval >>= 8; /* for next round */
          if ( *cp != '.' )
            break;
          cp++;
          if (!*cp)
            err = 1;
        }
      }
    } /* while (*cp && !err) */

    if (err || parts == 0 || parts < minparts)
      err = 1;
    else if ( bracket && *cp !=']' ) /* unmatched bracket */
      err = 1;
    else if (*cp && *cp!='\n' && *cp!='\r' && *cp!='\t' && *cp!=' ')
      err = 1;
    else if (destp)
    {
      register int n;
      *destp = 0;
      if (addrtype == 'n')                   /* inet_network() */
      {
        for ( n = 0; n < parts; n++ )
        {
          *destp <<= 8;
          *destp |= buf[n];
        }
      }
      else                                   /* inet_aton() or inet_addr() */
      {
        char *p;
        if (parts == 1)
          maxval = buf[0];
        else if (parts == 2)
          maxval = (buf[0]<<24) | buf[1];
        else if (parts == 3)
          maxval = (buf[0]<<24) | (buf[1]<<16) | buf[2];
        else if (parts == 4)
          maxval = (buf[0]<<24) | (buf[1]<<16) | (buf[2]<<8) | buf[3];
        p = (char *)destp;
        *destp = 0;
        for ( n= 0; n < 4; n++ )
          *p++ = (char)((maxval >> ((3-n)<<3)) & 0xff);
      }
    }
  } /* if (cp) */

  if (err)
    return 0;
  return 1;
}

/* ----------------------------------------------------------------------- */

/*
 * __inet_ntoa() is the base of all bin->ASCII addr conversion.
 *
 * The source is assumed to be in NET_ORDER.
 * If 'buffer' is NULL, a local static buffer is used.
 *
 * __inet_ntoa() returns a pointer to the result, or NULL if 'bufflen'
 * is too small.
*/
static char *__inet_ntoa( uint32_t in, char *buffer, int buflen, int *reslen )
{
  char scratch[sizeof("255.255.255.255  ")];
  const char *p = (const char *)&in;
  int i, len; 

  if (!buffer)
  {
    static int idx = -1;
    static char buf[4 * sizeof(scratch)];
    buffer = &(buf[((++idx)&3) * sizeof(scratch)]);
    buflen = sizeof(scratch);
  } 
  len = 0;
  for (i = 0; i < 4; i++)
  {
    register unsigned int r, c;
    c = (((unsigned int)(*p++)) & 255 );
    if ( i )
      scratch[len++] = '.';
    r = (c / 100);
    if ( r != 0 )
      scratch[len++] = (char)( r + '0' );
    r <<= 8; r |= ((c % 100)/10);
    if ( r != 0)
      scratch[len++] = (char)( (r & 0xff) + '0' );
    scratch[len++] = (char)((c%10) + '0');
  }
  scratch[len++] = '\0';
  if (buflen < len) /* len includes terminating '\0' */
    buffer = (char *)0;
  else
  {
    for (i = 0; i < len; i++)
      buffer[i] = scratch[i];
    if (reslen)
      *reslen = len - 1;
  }
  return buffer;
}

/* ======================================================================= */
/*                    host-order based conversion                          */
/* ======================================================================= */

/*
 * _my_aton() - 
 *      - converts [a.b.c.d] or a.b.c.d to 32 bit long (host order)
 *        (address must have 4 parts)
 *      - returns 0 on error and !0 on success (stores result in 'result')
 */
int _my_aton(const char *str, uint32_t *result)
{
  if (!__inet_atox(str, result, 4, 'h'))
    return (0);
  if (result)
    *result = NTOHL(*result);
  return 1;
}


/*
 * Convert an IP address to a dotted quad.
 * 'ina' is in host order.
 */
const char *_my_ntoa (uint32_t ina, char *buf, unsigned int buflen)
{
  char tmp[sizeof("255.255.255.255")];
  if (buf && buflen < sizeof(tmp))
  {
    unsigned int i = 0;
    __inet_ntoa( HTONL(ina), tmp, sizeof(tmp), (int *)0 );
    buflen--;
    while(tmp[i] && i < buflen)
    {
      buf[i] = tmp[i];
      i++;
    }
    buf[i] = '\0';
    return buf;
  }
  return __inet_ntoa( HTONL(ina), buf, buflen, (int *)0 );
}  

/* ======================================================================= */
/*                             address scope                               */
/* ======================================================================= */

int is_onmysubnet(uint32_t ina)
{
  return (ina && _ip_get_snmask() && 
          ((ina ^ _ip_get_ipaddr()) & _ip_get_snmask()) == 0);
}

/* ---------------------------------------------------------------------- */

/*
 * Return !0 if ip address is a local address or on Loopback network. 
 * 'ina' is in host order.
 */
int is_local_addr (uint32_t ina)
{
  return ((ina >> 24) == 127 || ina == _ip_get_ipaddr());
  /* (my_ip_addr <= ip <= my_ip_addr+multihomes) or 127.x.x.x */
}

/* ---------------------------------------------------------------------- */

/*
 * return !0 if an address is a broadcast (directed or otherwise) address.
 * 'ina' is in host order.
*/
int is_broadcast(uint32_t ina)
{
  uint32_t sin_mask, my_addr;
  if (!ina)
    return 0;
  if (ina == ((uint32_t)-1))
    return 1;
  my_addr = _ip_get_ipaddr();
  if (ina == my_addr)
    return 0;
  sin_mask = _ip_get_snmask();
  if ((ina & sin_mask) != (my_addr & sin_mask))
    return 0;
  return ((~ina & ~sin_mask) == 0);
}

/* ---------------------------------------------------------------------- */

/*
 * is_multicast - return !0 if the given IP addr is Class D (Multicast)
 * 'ina' is in host order.
*/
int is_multicast (uint32_t ina)
{
  return ((ina & 0xF0000000ul) == 0xE0000000UL);  /* or IN_CLASSD(ina) */
}

