/*
 * Internet Protocol (IP) layer base functionality.
 * (RFC 791/RFC 760)
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: ipbase.c,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
/* 
 * Internet Protocol (IP) layer.
 *
 * Written May 2001 by Cyrus Patel <cyp@fb14.uni-mainz.de>
 *
 *
 *
 * The seven layers of the OSI Basic Reference Model and how it is mapped
 * in this library:
 *  ________________  ________________________  ________________________
 * |  Application   || (mailer, browser, etc) ||                        |
 * |----------------||------------------------||------------------------|
 * |  Presentation  ||     (SMTP/HTTP etc)    ||                        |
 * |----------------||------------------------||------------------------|
 * |     Session    ||        sock_*()        ||       sock_io.c        |
 * |----------------||------------------------||------------------------|
 * |   Transport    ||    TCP, UDP   ,------, || pctcp, pcudp           |
 * |----------------||---------------| ICMP |-||----------------pcicmp--|
 * |    Network     || ,-----,  IP   '------' ||        ip_out          |
 * |----------------||-| ARP |----------------||--pcarp-----------------|
 * |   Data Link    || '-----'    pkt driver  ||         pcsed, pcpkt32 |
 * |----------------||------------------------||------------------------|
 * |    Physical    ||        (cable)         ||                        |
 * '----------------''------------------------''------------------------'
 *
 * Each OSI layer (and the files on each layer) communicates with only its 
 * immediate layer neighbours and knows nothing about how the data is 
 * formatted or handled once it has been passed up or down.
 * For example, a transport knows nothing about Ethernet packets, framing
 * or addressing, and vice-versa; ethernet knows nothing about IP protos.
 *
 *            ,------- tcp_*() --------,     
 *            |        |                \    
 *            |        |_               |   arp_*() --    loopback
 *            /          \              |  /          \  /
 *  sock_*()==          icmp_*() ---- ip_*()         eth_*()
 *            \         _/             /|  \          /  \
 *            |        |    igmp_*() -' |   ----------     packet drvr
 *            |        |                / 
 *            `------- udp_*() --------' 
 *
 * The following two schematics describe how sent data flows down 
 * the OSI layers, and received data flows up through the OSI layers. 
 *
 *                                             -<--> arp_resolve
 *                                            /
 *   sock_send_* ->--> xxx_send_to() ->-> ip_output
 *                                                \__>-- eth_send
 *
 *                     xxx_handler() <-<- ip_handler <-<- eth_recv
 *                            ()
 *   sock_recv_* -<--< xxx_recv_from()
 *
 * (In case you are wondering why ICMP and ARP hang in the middle like that:
 * The IP protocol suite pre-dates OSI. The ARPA protocols are four-level: 
 *  _______________________  ________________________ 
 * |                       || (mailer, browser, etc) |
 * |  Process/Application  ||      userland API      |
 * |-----------------------||------------------------|
 * |       Host-Host       ||        TCP, UDP        |
 * |-----------------------||------------------------|
 * |       Internet        ||      IP, ICMP, IGMP    |
 * |-----------------------||------------------------|
 * |    Network access     ||       ARP, ether       |
 * '-----------------------''------------------------'
 * which in turn are based on the 3-layer DOD Protocol Model (Application, 
 * Host-to-Host, and Network Access) 
*/
#define __OSILAYER__ 3 /* Network Layer */

#include "target.h"   /* standard types/includes */
#include "inchksum.h" /* checksum() */
#include "linklayr.h" /* _eth_send, _eth_get_mtu, _eth_register_pkthandler */
#include "timer.h"    /* Random() */
#include "blocker.h"  /* blocker_critsect_*() */
#include "inaddr.h"   /* _my_aton(), is_local_addr() */
#include "arp.h"      /* _arp_resolve(), _arp_find_entry(), _arp_add_entry() */
#include "loadcfg.h"  /* _conf_add_confextn() */
#include "debug.h"    /* KMSG(), DBGPRINTF() */
#include "ctordtor.h" /* _my_atexit() */
#include "icmp.h"     /* icmp_retour() */
#include "userland.h" /* WSO_xxx error numbers */
#include "ipbase.h"   /* ourselves */
#include "ipboot.h"   /* _ip_addr_boot() */
#include "ipfrag.h"   /* _ip_frag_handler(), _ip_frag_free() */
#include "ipmcast.h"  /* _mcast_is_in_group() */

/* ====================================================================== */
/*                             private data                               */
/* ====================================================================== */

static struct /* all defaults are zero */
{
  uint32_t  prim_ip_addr;   /* our IP address */
  uint32_t  prim_net_mask;  /* our net-mask */

  int    default_ttl;       /* default (128) is set at calltime */
  int    default_tos;       /* default is 0 */

  unsigned bcopen_count;    /* count of open (UDP) broadcast PCBs */

  int ip_id_increment;      /* IP ID increment is in range 1...this */  
  int used_dhcp_bootp;      /* IPaddr was obtained dynamically (BOOTP/DHCP) */
  int domask_req;           /* do an "ICMP Mask Request" when configured */
  int debug_lev;            /* debug level */
  int init_done;            /* done __check_guard_marker() once */
  int exit_seen;            /* done atexit() stuff */
  long ctors_are_working;   /* check that constructors are being called ok */
  #define CTOR_CALLED_MAGIC 0x12345678L

  int (*ip_raw_hook)(const in_Header FAR *);

} FAR my_statics; /* all defaults are zero */

/* ======================================================================= */

int _ip_is_up(void) /* this is only for ARP's benefit. */
{
  return (my_statics.init_done && !my_statics.exit_seen);
}

/* ----------------------------------------------------------------------- */

static void __set_my_ip (const char *value)
{
  uint32_t addr;
  if (_my_aton(value, &addr)) /* not "bootp"/"dhcp"/"rarp" */
  {
    /* not 0.0.0.0, not 255.255.255.255, not 127.x.x.x, not class D/multicast */
    if (addr && addr!=0xfffffffful && (addr>>24)!=127 && (addr>>28)!=0xE &&
      ((addr & (~my_statics.prim_net_mask)) != (~my_statics.prim_net_mask)) )
    {                         /* also ok prim_net_mask hasn't been set yet */
      _ip_set_ipaddr(addr);
    }
  }
  return;
}

/* ----------------------------------------------------------------------- */

static void __set_my_netmask(const char *value)
{
  uint32_t addr;
  if (_my_aton(value, &addr))
  {
    uint32_t tmp = my_statics.prim_ip_addr;
    if (!tmp || ((tmp & ~addr) != ~addr))
    {
      unsigned int n = 0;
      tmp = addr;
      while ((tmp>>31) == 1)
      {
        n++;
        tmp<<=1;
      }
      if (n < 32 && tmp == 0UL)
        my_statics.prim_net_mask = addr;
    }
  }
  return;
}

/* ----------------------------------------------------------------------- */

static void __init_private_data(void)
{
  static const struct config_table FAR my_cfg[] = {
       { "MY_IP",         ARG_FUNC,  (void FAR *)__set_my_ip          },
       { "NETMASK",       ARG_FUNC,  (void FAR *)__set_my_netmask     },
       { "IP.DEF_TTL",    ARG_ATOI,  (void FAR *)&my_statics.default_ttl   },
       { "IP.DEF_TOS",    ARG_ATOI,  (void FAR *)&my_statics.default_tos   },
       { "IP.ID_INCR",    ARG_ATOI,  (void FAR *)&my_statics.ip_id_increment},
  /*   { "MTU",           ARG_ATOI,  (void FAR *)&mtu                      }, */
  /* --  'MTU' is obsolete - _eth_get_mtu() never fails -- */
       { "ICMP_MASK_REQ", ARG_ATOI,  (void FAR *)&my_statics.domask_req    },
       { "IP.DEBUG",      ARG_ATOI,  (void FAR *)&my_statics.debug_lev     },
       { ((const char FAR *)0), 0, ((void FAR *)0)  }
  };
  _conf_add_confextn( &my_cfg[0], NULL );

  FAR_MEMSET( &my_statics, 0, sizeof(my_statics) );
}

/* ====================================================================== */
/*            data that not even IP is allowed to touch directly          */
/* ====================================================================== */

#define __load_cfg() _conf_load()

uint8_t _ip_get_default_ttl(void)
{
  int rc;
  __load_cfg();
  rc = my_statics.default_ttl;
  if (rc == 0)
    rc = 128; 
  else if (rc < 0)
    rc = 254;
  else if (rc > 255)
    rc = 255;
  return (uint8_t)rc;
}  
#define default_ttl direct_access_not_allowed


uint8_t _ip_get_default_tos(void)
{
  int rc;
  __load_cfg();
  rc = my_statics.default_tos;
  if (rc < 0 || rc > 255)
    rc = 0;
  return (uint8_t)rc;
}  
#define default_tos direct_access_not_allowed


static uint16_t _get_ip_id (void)
{
  static int ip_id = 0;
  int increment = 1;
  if (my_statics.ip_id_increment < 2) /* user didn't set it */
    increment = (int)Random(1, 32);
  else if (my_statics.ip_id_increment > 1)
    increment = (int)Random(1, my_statics.ip_id_increment);
  ip_id += increment;
  if (ip_id == 0) /* zero value is reserved to mean "put an id in for me" */
    ip_id++;
  return HTONS((uint16_t)ip_id);
}
#define ip_id_increment direct_access_not_allowed


uint16_t _ip_get_mtu(void)
{
  static uint16_t real_mtu = 0;
  if (real_mtu == 0)
  {
    /* note that the number _eth_get_mtu() returns is the OSI 3 layer MTU,
    ** not the MTU returned by the packet driver (which includes MAC hdr size).
    ** Note also that _eth_get_mtu() never fails. If it can't get the MTU from
    ** the packet driver then it determines MTU based on driver class. (RFC791)
    **
    ** _eth_get_mtu() returns zero if it can't initialize the driver, 
    ** so fake it, otherwise _ip_init() (and all the protos) will go poof.
    */
    unsigned int newmtu = _eth_get_mtu();
    if (newmtu < 1006) /* 1006 is the official MTU of SLIP */
      newmtu = 1006; /* FYI: the smallest MTU according to RFC791 is 68 */

    if (my_statics.debug_lev)
      DBGPRINTF(("*IP mtu is %u\n", newmtu ));
    real_mtu = (uint16_t)newmtu;
  }
  return real_mtu;
}

/* ====================================================================== */
/*                     sub-protocol registration                          */
/* ====================================================================== */

static struct
{
  int ip_proto;
  void (*pkt_handler)(const in_Header FAR *ippkt, int protohndlr_flags);
  void (*stack_shutdown)(void);  
  int FAR *proto_debug_lev;
} regprotos[5]; /* ICMP+TCP+UDP+IGMP+spare */

/* ----------------------------------------------------------------------- */

void _ip_proto_register(int ip_proto, 
       void (*pkt_handler)(const in_Header FAR *ippkt, int protohndlr_flags),
       void (*stack_shutdown)(void), int FAR *proto_debug_lev )
{
  unsigned int i;

  /* KMSG((-1,"_ip_proto_register(%d)\n", ip_proto)); */
  for (i = 0; i < (sizeof(regprotos)/sizeof(regprotos[0])); i++)
  {
    if (regprotos[i].ip_proto == ip_proto)
      return;
    if (regprotos[i].ip_proto == 0)
      break;
  }
  for (i = 0; i < (sizeof(regprotos)/sizeof(regprotos[0])); i++)
  {
    if (regprotos[i].ip_proto == 0)
    {
      regprotos[i].ip_proto = (uint8_t)ip_proto;
      regprotos[i].pkt_handler = pkt_handler;
      regprotos[i].stack_shutdown = stack_shutdown;
      regprotos[i].proto_debug_lev = proto_debug_lev;
      return;
    }
  }
  __fatal_runtime_error("PANIC: IP proto table is too small!\n");
}

/* ----------------------------------------------------------------------- */

static int __find_proto_debug_lev(int ip_proto)
{       /* returns min( *regprotos[i].proto_debug_lev, my_statics.debug_lev) */
  unsigned int i;
  for (i = 0; i < (sizeof(regprotos)/sizeof(regprotos[0])); i++)
  {
    if (regprotos[i].ip_proto == ip_proto)
    {
      if (*regprotos[i].proto_debug_lev < my_statics.debug_lev)  
        return *regprotos[i].proto_debug_lev;
      break;
    }
    if (regprotos[i].ip_proto == 0)
      break;
  }  
  return my_statics.debug_lev;
}

static void (*__find_proto_handler(int ip_proto))( const in_Header FAR *, int )
{
  unsigned int i;
  for (i = 0; i < (sizeof(regprotos)/sizeof(regprotos[0])); i++)
  {
    if (regprotos[i].ip_proto == ip_proto)
      return regprotos[i].pkt_handler;
    if (regprotos[i].ip_proto == 0)
      break;
  }
  return NULL;
}

/* ----------------------------------------------------------------------- */

void _ip_icmp_callback( const in_Header FAR *orig_ip, int soicmperr_code )
{
  auto void (*hdlr)( const in_Header FAR *, int );
  hdlr = __find_proto_handler(orig_ip->proto );
  if (hdlr)
    (*hdlr)( orig_ip, (soicmperr_code | PROTOHNDLR_FLAGS_ICMPCB) );
  return;
}    

/* ====================================================================== */
/*                      IP packet header dump                             */
/* ====================================================================== */

#if defined(NDEBUG)
#define __dbg_dump_header( __prefix, __ipdgram ) NO_OPERATION()
#else
static void __dbg_dump_header( const char *prefix, const in_Header FAR *ipdgram )
{
  const uint8_t FAR *p; 
  const char *np;
  uint16_t tmp;

  DBGPRINTF(("*IP-%04x %s: header dump:\n", 
                 NTOHS(ipdgram->identification), prefix ));
  DBGPRINTF(("*  ->hdrlenver   = 0x%02x (len=%d, ver=%d)\n",
          ipdgram->hdrlenver, in_GetHdrLen(ipdgram), in_GetHdrVer(ipdgram)));

  DBGPRINTF(("*  ->tos         = 0x%02x\n", ipdgram->tos ));
  tmp = NTOHS(ipdgram->length);
  DBGPRINTF(("*  ->length      = 0x%04x (%u)\n", tmp, tmp ));
  tmp = NTOHS(ipdgram->identification);
  DBGPRINTF(("*  ->id          = 0x%04x (%u)\n", tmp, tmp ));

  tmp = NTOHS(ipdgram->frag_ofs);
  DBGPRINTF(("*  ->frag_ofs    = 0x%04x\n", tmp ));
  DBGPRINTF(("*         offset = %ld\n", ((uint32_t)(tmp & IP_OFFMASK)) << 3 ));
  DBGPRINTF(("*         flags  = %s%s%s%s\n", 
                                 ((tmp & IP_CE) ? "CE " : ""),
                                 ((tmp & IP_DF) ? "DF " : ""),
                                 ((tmp & IP_MF) ? "MF"  : ""),
                                 ((tmp & ~IP_OFFMASK)==0) ? "(none)" : "" ));

  DBGPRINTF(("*  ->ttl         = 0x%02x (%d)\n", ipdgram->ttl, ipdgram->ttl ));
  np = "";
  if (ipdgram->proto == IPPROTO_ICMP)
    np = "= (ICMP)";
  else if (ipdgram->proto == IPPROTO_IGMP)
    np = "= (IGMP)";
  else if (ipdgram->proto == IPPROTO_TCP)
    np = "= (TCP)";
  else if (ipdgram->proto == IPPROTO_UDP)
    np = "= (UDP)";

  DBGPRINTF(("*  ->proto       = 0x%02x (%d) %s\n", ipdgram->proto, ipdgram->proto, np ));
  DBGPRINTF(("*  ->checksum    = 0x%04x\n", ipdgram->checksum ));
  DBGPRINTF(("*  ->source      = %I\n", ipdgram->source ));
  DBGPRINTF(("*  ->destination = %I\n", ipdgram->destination ));
  DBGPRINTF(("*  ->options%s\n", 
   ((in_GetHdrLen(ipdgram) == sizeof(in_Header)) ? "     = (none)" : ":" ) ));
  
  p = (const uint8_t FAR *)(ipdgram + 1);
  tmp = (uint16_t)(in_GetHdrLen(ipdgram) - sizeof(in_Header));
  while (tmp)
  {
    static const char *opclasses[] = {"control", "reserved", 
                                      "debug"  , "reserved" };
    uint8_t opcopied = ((*p & 0x80) != 0);
    uint8_t opclass  = (uint8_t)((*p >> 5) & 3);
    uint8_t opnum    = (uint8_t)((*p & 0x1F));
    uint8_t oplen    = (uint8_t)((*p == 0 || *p == 1) ? (1) : (p[1]));
    int  do_dump  = 0;
    const char *fmt;
     
    DBGPRINTF(("*   Op 0x%02x (%u)\n", *p, *p));
    DBGPRINTF(("*       length: %d\n", oplen ));
    DBGPRINTF(("*       copied: %s\n", opcopied ? "yes" : "no" ));
    DBGPRINTF(("*       class : %d (%s)\n", opclass, opclasses[opclass] ));
    DBGPRINTF(("*       number: %d\n", opnum ));
    fmt =      "*       name  : '%s'\n";

    switch (*p)
    {
      case   0: DBGPRINTF((fmt, "end-of-options")); break;
      case   1: DBGPRINTF((fmt, "NOP/padding")); break;
      case  11: /* fallthrough */
      case  12: DBGPRINTF((fmt, (*p == 11 ? "probe-MTU" : "reply-MTU") ));
                DBGPRINTF(( "*       => lowest MTU: %u\n",
                            NTOHS(*((const uint16_t FAR *)&p[2])) ));
                break;
      case 130: DBGPRINTF((fmt, "Security" )); 
                do_dump = 1; break;
      case 131: DBGPRINTF((fmt, "Loose Source and Record Route"));
                do_dump = 1; break;
      case   7: DBGPRINTF((fmt, "Record Route"));
                do_dump = 1; break;
      case 136: DBGPRINTF((fmt, "Stream Identifier"));
                do_dump = 1; break;
      case  68: DBGPRINTF((fmt, "Time Stamp"));
                DBGPRINTF(( "*       => first entry: %I %lu\n",
                *((const uint32_t FAR *)&p[4]), *((const uint32_t FAR *)&p[8]) ));
                break;
      default:  DBGPRINTF((fmt, "(unrecognized)"));  
                do_dump = 1; break;
    }

    if (do_dump && oplen > 2)
       DBGDUMP(( "Raw option dump (data area)", 0, &p[2], oplen-2 ));

    tmp -= oplen;
    p += oplen;
  }
  return;
}
#endif /* NDEBUG */

/* ====================================================================== */
/*        IP packet sender (with on-demand initialization)                */
/* ====================================================================== */

static void __ip_atexit_hook(void)
{
  /* Do sub-protocol shutdown first - some protocols
   * may need to flush their queues. This also follows
   * the top-down deinitialization order as outlined
   * below.
  */
  unsigned int i;

  blocker_critsect_begin();

  for (i = 0; i < (sizeof(regprotos)/sizeof(regprotos[0])); i++)
  {
    if (regprotos[i].ip_proto == 0)
      break;
    if (regprotos[i].stack_shutdown)
      (*(regprotos[i].stack_shutdown))();
  }
  if (my_statics.debug_lev)
    DBGPRINTF(("*IP shutdown begin\n"));
  if (my_statics.used_dhcp_bootp)
  {
    my_statics.prim_ip_addr = 0ul;
    my_statics.prim_net_mask = 0ul;
    my_statics.used_dhcp_bootp = 0;
  }
  my_statics.exit_seen = 1;
  if (my_statics.debug_lev)
    DBGPRINTF(("*IP shutdown complete\n"));

  blocker_critsect_end();

  return;
}

/* ---------------------------------------------------------------------- */

/*
** __check_guard_marker() with integrated initialization engine.
** Never fails unless it runs out of memory or the heap is trashed, 
** in which case (either case) it has no choice but to exit().
**
** There are a number of "hidden" initialization calls that
** may be made by the subsystems addressed from here. For 
** example, _ip_get_mtu() calls _eth_get_mtu() which in turn
** initializes the _eth_*() stuff (if needed). The _eth_*() 
** stuff initialization may in turn cause a config load, etc.
**
** Use of any subsystem thus causes it and all other subsystems
** that it depends on to become initialized. This reinforces 
** my (cyp) design choice that all subsystems be autonomous from 
** each other. 
**
** It also has the following (desirable) side-effect:
** Since initialization is effectively bottom-up (OSI layer), 
** deinitialization using atexit() hooks will always be 
** in order (top-down). 
**
** If you look at the following code, you'll notice an eth call
** being made first, then sock_atext() to register the IP shutdown,
** then DHCP being done. The resultant atexit() call sequence will 
** then be... dhcp_down()... udp_down()... ip_down()... eth_down()
*/
static void __check_guard_marker(void)
{
  if (!my_statics.init_done)
  {
    if (my_statics.exit_seen)
    {
      /* nonfatal because can't exit()/_exit() because exit() already done. */
      __nonfatal_runtime_error("PANIC: destructors already called!\n");
      return; 
    }

    if (my_statics.ctors_are_working != CTOR_CALLED_MAGIC)
    {
      __fatal_runtime_error("PANIC: constructors were not called.\n");
    }    

    if (my_statics.debug_lev)
      DBGPRINTF(("*begin IP stack init\n"));

    my_statics.init_done = 1;

    blocker_critsect_begin();

    _ip_get_mtu(); /* initialize ether if not already initialized */
    _my_atexit(__ip_atexit_hook);         /* IP shutdown before ETH shutdown */
    _arp_resolve( ((uint32_t)-1), NULL ); /* ARP shutdown before IP shutdown */

    if (!my_statics.prim_ip_addr)
    {
      my_statics.used_dhcp_bootp = 1;
    }

    /* Do address discovery if needed, and 
     * finalize sin_mask. If address discovery failed, 
     * a retry hook will be placed.
    */
    _ip_address_boot( &my_statics.prim_ip_addr, &my_statics.prim_net_mask,
                      my_statics.debug_lev );

    blocker_critsect_end();
  }

  return;
}

/* ---------------------------------------------------------------------- */

#define TEST_PROTO_CHECKSUM

#if defined(TEST_PROTO_CHECKSUM) && !defined(NDEBUG)
static int test_proto_chksum (const in_Header FAR *in_hdr)
{
  if ((NTOHS(in_hdr->frag_ofs) & (IP_OFFMASK|IP_MF)) == 0) 
  {
    const uint8_t FAR *prdata = ((const uint8_t FAR *)in_hdr) + in_GetHdrLen(in_hdr);
    uint16_t prdatalen = (uint16_t)(NTOHS(in_hdr->length) - in_GetHdrLen(in_hdr));

    switch (in_hdr->proto)
    {
      case IPPROTO_ICMP:
      case IPPROTO_IGMP:
      {
        return (checksum(prdata, prdatalen) == 0xFFFF);
      }
      case IPPROTO_UDP:
      {
        if (*((const uint16_t FAR *)(prdata+6)) == 0)
          break; /* not using checksums */
        /* fallthrough */
      }
      case IPPROTO_TCP:
      {
        tcp_PseudoHeader ph;
        ph.src      = in_hdr->source;
        ph.dst      = in_hdr->destination;
        ph.mbz      = 0;
        ph.protocol = in_hdr->proto;
        if (in_hdr->proto == IPPROTO_UDP)
          ph.length = *((const uint16_t FAR *)(prdata+4));
        else 
          ph.length = HTONS(prdatalen);
        ph.checksum = checksum(prdata, NTOHS(ph.length));
        return (checksum(&ph,sizeof(ph)) == 0xFFFF);
      }
      default:
        break;
    }
  }
  return 1;
}
#endif

/* ---------------------------------------------------------------------- */

/* expect large datagrams if sending large datagrams. Growing the
** queue is not necessary when _sending_ to local_addr, but doesn't hurt.
*/
static void __raise_ethnumbufs_if_needed( const in_Header FAR *in_hdr )
{
  uint16_t offs = (uint16_t)(NTOHS(in_hdr->frag_ofs) & (IP_OFFMASK));
  if (offs) /* div 8 */
  {
    signed int data_mtu = _ip_get_mtu() - 64; /* 64=max IP header size */
    if (data_mtu > 8)
    {
      static unsigned int _numbufs = 0;
      unsigned int n = 1+(offs / ((unsigned)data_mtu)); /* times eight */
      if (n > _numbufs) 
      {  
        _numbufs = n;
        n = _eth_numbufs_set(n<<3); /* 8, 16, 24, 32, 40, 48 */
        if (n == 0 && my_statics.debug_lev)
          DBGPRINTF(("*IP: increased numbufs to %u (used data_mtu=%u)\n", 
                                          _numbufs<<3, data_mtu ));
      }
    }
  }
  return;
}

/* ---------------------------------------------------------------------- */

/* 
 * All fields in the ip header (except ID and checksum, which are optional)
 * must have been filled in. 
 *
 * _ip_output_raw() is reentry-safe.
*/
int _ip_output_raw( const void FAR *ipseg, /* required */
                    unsigned int ipseglen, /* required */
                    uint32_t first_hop_addr,  /* optional */
                    int FAR *so_errno )    /* optional */
{
  const in_Header FAR *in_hdr = (const in_Header FAR *)ipseg;
  uint8_t dest_hw_addr[MAX_LINKADDR_SIZE];
  struct _ip_ifinfo ifinfo;
  int dummy_so_errno;
  int proto_debug_lev = __find_proto_debug_lev(in_hdr->proto);

  /* KMSG((0,"ipseglen=%u\n", ipseglen)); */

  if (!so_errno)
    so_errno = &dummy_so_errno;

  if (my_statics.exit_seen)
  {
    if (proto_debug_lev)
      DBGPRINTF(("*IP OUT: DROP (exit_seen)\n"));
    *so_errno = WSO_ENETDOWN;
    return -1;
  }

  if (ipseglen < (sizeof(in_Header)+6) || 
      ipseglen > _ip_get_mtu() )
  {
    /* if something got here its probably a signedness bug */
    if (proto_debug_lev)
      DBGPRINTF(("bad ipseg_len (%d) in _ip_output_raw!\n", ipseglen ));
    *so_errno = WSO_EMSGSIZE;
    return -1;
  }

  if (in_GetHdrVer(in_hdr) != 4                ||
      in_GetHdrLen(in_hdr) < sizeof(*in_hdr)   ||
      NTOHS(in_hdr->length) != ipseglen        ||
      in_hdr->destination == 0)
  {
    if (proto_debug_lev)
      DBGPRINTF(("*IP OUT: DROP (bad header)\n"));
    *so_errno = WSO_EINVAL;
    return -1;
  }

  if (in_hdr->source == 0)
  {
    /* refuse to work with hacks. Caller _must_ provide a source IP address */
    if (in_hdr->proto != IPPROTO_ICMP && my_statics.prim_ip_addr != 0)
    {
      if (proto_debug_lev)
        DBGPRINTF(("*IP OUT: DROP (missing source addr)\n"));
      *so_errno = WSO_EINVAL;
      return -1;
    }
  }
  else if (!is_local_addr(NTOHL(in_hdr->source)))
  {
    if (proto_debug_lev)
      DBGPRINTF(("*IP OUT: DROP (bad source addr)\n"));
    *so_errno = WSO_EINVAL;
    return -1;
  }

  if (first_hop_addr == 0)
  {
    first_hop_addr = NTOHL(in_hdr->destination);
  }

  if (first_hop_addr == 0)
  {
    if (proto_debug_lev)
      DBGPRINTF(("*IP OUT: DROP (missing dest addr)\n"));
    *so_errno = WSO_EDESTADDRREQ;
    return -1;
  }

  /* prim_ip_addr will be 0 if a DHCP renew failed or is still in progress */
  if (in_hdr->proto != IPPROTO_ICMP)
  {
    if (my_statics.prim_ip_addr == 0 && 
        (first_hop_addr >> 24) != 127 &&
        (in_hdr->proto != IPPROTO_UDP ||
         NTOHS(*((const uint16_t FAR *)(((const uint8_t FAR *)in_hdr) + 
                                       in_GetHdrLen(in_hdr)+0))) != 68 ||
         NTOHS(*((const uint16_t FAR *)(((const uint8_t FAR *)in_hdr) + 
                                       in_GetHdrLen(in_hdr)+2))) != 67))
    {
      if (proto_debug_lev)
        DBGPRINTF(("*IP OUT: DROP (IP is down)\n"));
      *so_errno = WSO_ENETDOWN;
      return -1;
    }
  }

  if (_ip_select_if( &first_hop_addr, &ifinfo ) != 0)
  {
    if (proto_debug_lev)
      DBGPRINTF(("*IP OUT: DROP (bad dest addr)\n"));
    *so_errno = WSO_EHOSTUNREACH;
    return -1;
  }    

  if (ipseglen > ifinfo.ipif_mtu)
  {
    if (proto_debug_lev)
      DBGPRINTF(("*IP OUT: DROP (pkt too large)\n"));
    *so_errno = WSO_EMSGSIZE;
    return -1;
  }

  if (!_arp_resolve( first_hop_addr, &dest_hw_addr[0]))
  {
    if (proto_debug_lev)
      DBGPRINTF(("*IP: OUT: DROP (arp resolve failed)\n"));
    *so_errno = WSO_EHOSTUNREACH;
    return -1;
  }

  /* --- no return from here until end! --- */
  {
    uint32_t old_src_addr = in_hdr->source;
    uint16_t  old_identification = in_hdr->identification;
    uint16_t  old_checksum = in_hdr->checksum;
    int   is_local_dest = (is_local_addr(first_hop_addr));
    in_Header FAR *ncin_hdr; /* non-const copy of in_hdr */

    {
      union { in_Header FAR *notc; const in_Header FAR *isc; } uncu;
      uncu.isc = in_hdr;
      ncin_hdr = uncu.notc;
    }

    blocker_critsect_begin();

    if ((NTOHL(in_hdr->source) >> 24) == 127 && !is_local_dest) /* route! :) */
    {
      ncin_hdr->source = HTONL(my_statics.prim_ip_addr);
    }

    if (in_hdr->identification == 0)
    {
      ncin_hdr->identification = _get_ip_id();
    }

    ncin_hdr->checksum = 0;
    ncin_hdr->checksum = ~checksum(in_hdr, in_GetHdrLen(in_hdr));

    #if defined(TEST_PROTO_CHECKSUM) && !defined(NDEBUG)
    if (proto_debug_lev)
    {
      if (!test_proto_chksum(in_hdr))
        DBGPRINTF(("*IP OUT-%04x: _ip_output_raw() BAD PROTO CHECKSUM!\n",
                       NTOHS(in_hdr->identification) ));
    }
    #endif

    if (proto_debug_lev >= 3)
      __dbg_dump_header( "OUT", in_hdr );

    if (in_hdr->ttl == 0 && !is_local_dest)
    {
      if (proto_debug_lev)
        DBGPRINTF(("*IP OUT-%04x: DROP (TIMXCEED TTL)\n",
                    NTOHS(in_hdr->identification) ));
      /* enqueue a timexceed message to the caller */
      icmp_retour( SOICMPERR_TIMXCEED_TTL, in_hdr, 
                   ((const void FAR *)0), proto_debug_lev );
      /* fallthrough - pretend it was sent */ 
    }  
#if 0 /* stress the loopback queue */
    else if (is_local_dest && (NTOHS(in_hdr->frag_ofs) & IP_MF))
    {
      /* avoid uneccessarily stressing the loopback queue with 
       * packets that will need to be requeued in the defragger anyway.
       * Note that the last fragment (the one without IP_MF set) 
       * is still spooled via _eth_send() so that defrag completion 
       * occurs asynchronously.
      */
      _ip_frag_handler(in_hdr, _eth_get_hwaddr(), proto_debug_lev);
    }
#endif 
    else 
    {
      /* expect large datagrams if sending large datagrams */
      __raise_ethnumbufs_if_needed(in_hdr);

      if (_eth_send( ETHERTYPE_IP,      /* packet type */
                     &dest_hw_addr[0],  /* destinatation address */
                     ipseg,             /* packet to send */
                     ipseglen           /* packet length */   
                   ) != ((int)ipseglen))
      {
        if (proto_debug_lev)
          DBGPRINTF(("*IP OUT-%04x: DROP (eth fail)\n", 
                                   NTOHS(in_hdr->identification)));
        *so_errno = WSO_ENETDOWN;
        ipseglen = ((unsigned int)-1);
      }
    }
    if (ipseglen != ((unsigned int)-1)) /* xmit ok */
    {
      if (proto_debug_lev)
      {
        DBGPRINTF(("*IP OUT-%04x %I->%I (ttl=%d, local?=%d, loopback?=%d)\n",
                      NTOHS(in_hdr->identification), 
                    in_hdr->source, in_hdr->destination,
                    in_hdr->ttl, is_local_dest, ifinfo.ipif_flags.lo ));
        DBGPRINTF(("*            (tlen=%u, dlen=%u, offs=%u, mf=%d)\n", 
                      ipseglen, ipseglen - in_GetHdrLen(in_hdr),
                      ((NTOHS(in_hdr->frag_ofs) & IP_OFFMASK) << 3),
                      ((NTOHS(in_hdr->frag_ofs) & IP_MF)!=0) ));
      }        
    }

    ncin_hdr->source = old_src_addr;
    ncin_hdr->identification = old_identification;
    ncin_hdr->checksum = old_checksum;

    blocker_critsect_end();
  }

  return ipseglen;
}

/* ---------------------------------------------------------------------- */

/*
** Remove the options that are not copied during fragmentation.
*/
static void __strip_uncarried_options( in_Header *in_hdr )
{
  if (in_GetHdrLen(in_hdr) > sizeof(in_Header)) /* have opts */
  {
    int optlen; 
    uint8_t optbuf[48];
    uint8_t *r = ((uint8_t *)in_hdr) + sizeof(in_Header);
    uint8_t *q = &optbuf[0];

    while (*r != 0) /* while !end-of-options */
    {
      if (*r == 1) /* nop - keep for lsrr alignment */
      {
        *q++ = *r++; /* nop (padding) */
      }
      else /* apparently all but IPOPT_EOL and IPOPT_NOP have a length */
      {
        uint8_t i = r[1]; /* length */
        if (*r & 0x80) /* copied option */
        {
          while (i) {
            *q++ = *r++;
            i--;
          }
        }
        r += i;
      }
    }
    optlen = (uint8_t)(q - &optbuf[0]); 
    if (optlen)
    {
      q[optlen++] = 0; /* end of list */
      while (optlen & 3)
        q[optlen++] = 0;
    }

    if ((optlen + sizeof(in_Header)) <= in_GetHdrLen(in_hdr)) /* else uh,oh */
    {
      memcpy( ((uint8_t *)in_hdr) + sizeof(in_Header), optbuf, optlen );
      in_hdr->hdrlenver = (uint8_t)((in_GetHdrVer(in_hdr)<<4) + 
                                ((sizeof(in_Header) + optlen) >> 2));
    }
  }
  return;
}

/* ---------------------------------------------------------------------- */

int _ip_output_1 ( const in_Header *hdrP,
                   uint32_t dest_addr, /* if different from hdrp->destination */
                   const void FAR *data_1,  /* protocol header ... */
                   unsigned int data_1_len, /* ... with or without data */
                   const void FAR *data_2,  /* more (optional) data */
                   unsigned int data_2_len,
                   int FAR *so_errno )
{
  uint16_t datalen, hdrlen, mtu, offset, id, org_flags;
  int sent; char carried_hdr[64];
  in_Header FAR *in_hdr;
  int proto_debug_lev = __find_proto_debug_lev(hdrP->proto);

  if (my_statics.exit_seen)
  {
    if (so_errno)
      *so_errno = WSO_ENETDOWN;
    return -1;
  }

  hdrlen = (uint16_t)in_GetHdrLen(hdrP);

  if (hdrlen < sizeof(*hdrP))
  {
    if (so_errno)
      *so_errno = WSO_EINVAL;
    return -1;
  }    

  datalen = (uint16_t)(data_1_len + data_2_len);
  if (datalen > (NTOHS(hdrP->length) - hdrlen))
    datalen = (uint16_t)(NTOHS(hdrP->length) - hdrlen);

#define IP_MAXPACKET 0xffffu 

  if (datalen < 8 || /* no message is less than 8 bytes (empty ping is 8) */
      datalen > (IP_MAXPACKET - hdrlen) ||
      datalen < data_1_len || /* wrap check for 32bit target */
      datalen < data_2_len)   /* wrap check for 32bit target */
  {
    if (so_errno)
      *so_errno = WSO_EMSGSIZE;
    return -1;
  }

  if ((NTOHS(hdrP->frag_ofs) & (IP_OFFMASK|IP_MF)) != 0)
  {                                 /* can't start in mid-space */
    if (so_errno)
      *so_errno = WSO_EINVAL;
    return -1;
  }    

  if (!dest_addr)
    dest_addr = NTOHL(hdrP->destination);
  if (!dest_addr)
  {
    if (so_errno)
      *so_errno = WSO_EDESTADDRREQ;
    return -1;
  }    

  mtu    = _ip_get_mtu();
  in_hdr = (in_Header FAR *)FAR_MALLOC( mtu + 100 );
  if (!in_hdr)
  {
    if (proto_debug_lev)
      DBGPRINTF(("*_ip_output_1(): NO MEMORY!!\n"));
    if (so_errno)
      *so_errno = WSO_ENOBUFS;
    return -1;
  }

  offset    = 0;
  sent      = -1;
  id        = (hdrP->identification ? hdrP->identification : _get_ip_id());
  org_flags = (uint16_t)(NTOHS(hdrP->frag_ofs) & ~(IP_OFFMASK|IP_MF));

  while (datalen)
  {
    uint16_t availlen, payloadlen, flags;

    /* copy the IP header into the buffer.
    */
    if (offset == 0)
    {
      hdrlen = (uint16_t)in_GetHdrLen(hdrP);
      FAR_MEMCPY( in_hdr, hdrP, hdrlen );
      FAR_MEMCPY( carried_hdr, hdrP, hdrlen );
      __strip_uncarried_options( (in_Header *)(&carried_hdr[0]) );
    }
    else
    {
      hdrlen = (uint16_t)in_GetHdrLen( ((in_Header *)&carried_hdr[0]) );
      FAR_MEMCPY( in_hdr, carried_hdr, hdrlen );
    }

    /* determine flags and how much data to 
     * copy behind the IP header 
    */
    flags = 0;
    availlen = (uint16_t)(mtu - hdrlen);
    if (availlen >= datalen) /* last fragment */
      availlen = datalen; 
    else
    {
      availlen &= ~7;
      flags = IP_MF;
    }

    /* copy the data from data_{1|2} behind the IP header.
    */
    payloadlen = 0;
    if (data_1_len && availlen)
    {
      uint16_t copylen = (uint16_t)data_1_len;
      if (copylen > availlen)
        copylen = availlen;
      FAR_MEMCPY( ((uint8_t FAR *)in_hdr)+hdrlen+payloadlen, data_1, copylen );

      data_1_len -= copylen;
      data_1 = (const void FAR *)(((const uint8_t FAR *)data_1)+copylen);
      availlen -= copylen;
      payloadlen += copylen;

    }
    if (data_2_len && availlen)
    {
      uint16_t copylen = (uint16_t)data_2_len;
      if (copylen > availlen)
        copylen = availlen;
      FAR_MEMCPY( ((uint8_t FAR *)in_hdr)+hdrlen+payloadlen, data_2, copylen );

      data_2_len -= copylen;
      data_2 = (const void FAR *)(((const uint8_t FAR *)data_2)+copylen);
      availlen -= copylen;
      payloadlen += copylen;

    }
    if (!data_1_len && !data_2_len) /* everything copied over? */
    {
      /* simply make this the last fragment */
      datalen = payloadlen; 
      flags &= (uint16_t)(~IP_MF);        
    }        

    /* make any necessary adjustments
     * to the header.
    */
    in_hdr->length         = HTONS((uint16_t)(hdrlen+payloadlen));
    in_hdr->identification = id; 
    in_hdr->frag_ofs       = HTONS((uint16_t)((offset>>3)|flags|org_flags));
    in_hdr->checksum       = 0; /* done by _ip_output_raw() */

    if (_ip_output_raw(in_hdr, hdrlen+payloadlen, dest_addr, so_errno) == -1)
    {
      sent = -1;
      break; 
    }
    
    datalen -= payloadlen;
    offset  += payloadlen;
    if (sent == -1) /* first frag */
      sent = hdrlen;
    sent += payloadlen;  
  }

  FAR_FREE(in_hdr);

  return sent;
}

/* ---------------------------------------------------------------------- */

int _ip_output_2 ( uint32_t src_addr,   /* IP hdr from address (host order) */
                   uint32_t dst_addr,   /* IP hdr dest address (host order) */
                   uint8_t  protocol,   /* IP-protocol number */
                   uint8_t  ttl,        /* Time To Live */
                   uint8_t  tos,        /* Type Of Service, 0 unspecified */
                   uint16_t  flags,      /* eg, IP_DF */
                   const void FAR *data_1, /* protocol header and/or data */
                   unsigned int data_1_len,
                   const void FAR *data_2, /* more data (optional) */
                   unsigned int data_2_len,
                   int FAR *so_errno )
{
  in_Header in_hdr;
  uint16_t len;

  len = (uint16_t)(data_1_len + data_2_len);

  if (len < 8 || /* no message is less than 8 bytes (empty ping is 8) */
      len > (IP_MAXPACKET - sizeof(in_hdr)) ||
      len < data_1_len || /* wrap check for 32bit target */
      len < data_2_len)   /* wrap check for 32bit target */
  {
    if (so_errno)
      *so_errno = WSO_EMSGSIZE;
    return -1;
  }

  in_hdr.hdrlenver      = (uint8_t)((0x04u<<4) | (sizeof(in_hdr)>>2));
  in_hdr.tos            = tos;      
  in_hdr.length         = HTONS((uint16_t)(len+sizeof(in_hdr)));
  in_hdr.identification = 0;  /* done by _ip_output_X() */
  in_hdr.frag_ofs       = HTONS((uint16_t)(flags & ~(IP_OFFMASK|IP_MF)));
  in_hdr.ttl            = ttl;
  in_hdr.proto          = protocol;
  in_hdr.checksum       = 0; /* done by _ip_output_X() */
  in_hdr.source         = HTONL(src_addr); 
  in_hdr.destination    = HTONL(dst_addr);
  
  if (_ip_output_1( &in_hdr, 0, data_1, data_1_len, data_2, data_2_len, 
                    so_errno ) == -1)
    return -1;
  return (((int)len) & 0xffffu);
}  

/* ====================================================================== */
/*                         pseudo-IF selection                            */
/* ====================================================================== */

int _ip_select_if(const uint32_t FAR *destaddr, /* NULL destaddr means "default IF" */
                  struct _ip_ifinfo FAR *ifinfoP)
{
  if (destaddr && (*destaddr >> 24) == 127)
  {
    /* Doing an _ip_select_if() on 127.0.0.1 is a simple (and 
     * intentionally undocumented) way to tell if the IP stack is 
     * up-and-running or not - ifinfoP->flags.up will be zero if the 
     * stack is not running yet.
    */
    FAR_MEMSET( ifinfoP, 0, sizeof(*ifinfoP) );
    ifinfoP->ipif_flags.up   = (my_statics.prim_ip_addr != 0); /* maybe! ;) */
    ifinfoP->ipif_flags.lo   = 1;
    ifinfoP->ipif_mtu        = 1500; /* anything really */
    ifinfoP->ipif_host_addr  = 0x7f000001ul;
    ifinfoP->ipif_net_mask   = 0xffffff00ul;
    ifinfoP->ipif_lcast_addr = 0x7f0000fful; /* not really */
    ifinfoP->ipif_name       = "lo";
  }
  else
  {
    int ifflags;

    if (my_statics.exit_seen)
      return -1;

    /* careful! - order is important here */
    __check_guard_marker(); /* IP/ether may get initialized NOW */

    ifflags = _eth_get_ifflags();
    if (destaddr)
    {
      if (!(ifflags & ETH_IFFLAGS_MCAST) && is_multicast(*destaddr))
        return -1;
    }

    FAR_MEMSET( ifinfoP, 0, sizeof(*ifinfoP) );
    if (!(ifflags & ETH_IFFLAGS_P2P)) 
    {
      ifinfoP->ipif_flags.bcast = 1;
      ifinfoP->ipif_flags.mcast = ((ifflags & ETH_IFFLAGS_MCAST)!=0);
    }
    ifinfoP->ipif_flags.up      = 1; /* always ;) */
    ifinfoP->ipif_flags.defif   = 1;
    ifinfoP->ipif_mtu           = _ip_get_mtu(); 
    ifinfoP->ipif_host_addr     = my_statics.prim_ip_addr;
    ifinfoP->ipif_net_mask      = my_statics.prim_net_mask;
    ifinfoP->ipif_lcast_addr    = my_statics.prim_ip_addr|~my_statics.prim_net_mask;
    ifinfoP->ipif_name          = "pkt0"; /* blah */
  }
  return 0;
}

/* ====================================================================== */
/*                  controlled access to internal vars                    */
/* ====================================================================== */

int _ip_destaddr_open(uint32_t ina)
{
  if (is_broadcast(ina))
    my_statics.bcopen_count++; 
  return 0;
}

int _ip_destaddr_close(uint32_t ina)
{
  if (is_broadcast(ina))
  {
    if (my_statics.bcopen_count == 0)
      DBGPRINTF(("AIII! Mismatched _ip_destaddr_{open|close}() calls\n"));
    else
      my_statics.bcopen_count--;
  }
  return 0;
}

uint32_t _ip_get_ipaddr(void)
{
  if (my_statics.exit_seen)
    return 0ul;
  __check_guard_marker();
  /* KMSG((-1,"_ip_get_ipaddr = %hI\n",my_statics.prim_ip_addr)); */
  return my_statics.prim_ip_addr;
}

uint32_t _ip_get_snmask(void)
{
  if (my_statics.exit_seen)
    return 0ul;
  __check_guard_marker();
  return my_statics.prim_net_mask;
}

uint32_t _ip_get_bcaddr(void)
{
  if (my_statics.exit_seen)
    return 0ul;
  __check_guard_marker();
  return my_statics.prim_ip_addr | ~my_statics.prim_net_mask;
}

int _ip_set_ipaddr(uint32_t ipaddr)
{
  if (my_statics.exit_seen)
    return -1;
  if (!_ip_is_up()) /* can only be done if IP is not up yet */
  {
    __load_cfg();
    if ((ipaddr & 0xff) != 0xff && 
        (ipaddr & 0xff) != 0    && 
        (ipaddr >>  24) != 127 )
    { 
      if (my_statics.prim_ip_addr == 0)
      { 
        if (my_statics.debug_lev)
          DBGPRINTF(("Fixed IP configuration %hI.\n", ipaddr));
      }
      my_statics.prim_ip_addr = ipaddr;
      return 0;
    }
  }
  return -1;
}

/* ====================================================================== */
/*                         IP packet receiver                             */
/* ====================================================================== */

/*
 *  Check for correct IP-version, header length and header checksum.
 *  If failed, discard the packet. ref. RFC1122: section 3.1.2.2
 */
static const in_Header FAR *__chk_ip_header(const void FAR *ip_pkt, 
                                            unsigned int ip_pkt_len)
{
  const in_Header FAR *ip = (const in_Header FAR *)ip_pkt;
  unsigned ip_hlen;

  /* if (ip->proto == IPPROTO_ICMP && ip->frag_ofs == 0) 
   *   __asm int 3 
  */

  if (ip_pkt_len < sizeof(*ip))
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*IP-???? IN: DROP (pkt too small)\n"));
    return (NULL);
  }

  if (in_GetHdrVer(ip) != 4)    /* We only speak IPv4 */
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*IP-???? IN: DROP (bad hdr ver %d)\n", in_GetHdrVer(ip)));
    return (NULL);
  }          

  ip_hlen = in_GetHdrLen (ip);
  if (ip_hlen < sizeof(*ip))
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*IP-???? IN: DROP (bad hdr len %d)\n", in_GetHdrLen(ip)));
    return (NULL);
  }          

  if (ip_pkt_len < NTOHS(ip->length)) /* may be more though */
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*IP IN-%04x: DROP (bad len)\n", NTOHS(ip->identification) ));
    return (NULL);
  }

  if (checksum(ip,ip_hlen) != 0xFFFF)
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("*IP IN-%04x: DROP (bad chksum, expected 0xFFFF, got 0x%04x)\n", 
           NTOHS(ip->identification), checksum(ip,ip_hlen) ));
    return (NULL);
  }

  return (ip);
}


/* -------------------------------------------------------------- */

/*
 * Do some basic filtering to reduce the amount of work needed
 * to be done at the higher layers. We want to filter *before*
 * it gets into the defragger.
 *
 * Note that the incoming packet must be assumed to be fragmented
 * and therefore does not contain proto-specific headers.
*/
static const in_Header FAR *__ip_filter(const in_Header FAR *ipdgram, 
                                        const void *src_hw_addr, 
                                        int is_llbcast, int proto_debug_lev)
{
  int unhandled_proto = 0;
  uint32_t dstip = NTOHL(ipdgram->destination);
  uint32_t srcip = NTOHL(ipdgram->source);
  int bcast = (is_llbcast ||            /* a link-layer broadcast */
                is_broadcast(dstip)); /* or a (directed) broadcast */

#if defined(NDEBUG)
  #define DROPMSGSET(x) /* nothing */
#else
  const char *dropmsg = "filtered";
  #define DROPMSGSET(x) dropmsg = x
#endif

  if (my_statics.exit_seen)
  {
    DROPMSGSET("done exit");
    proto_debug_lev = my_statics.debug_lev;
  }
  else if (!__find_proto_handler(ipdgram->proto))
  {
    DROPMSGSET("unhandled proto");
    unhandled_proto = 1;
    proto_debug_lev = my_statics.debug_lev;
  }
  else 
  {
    switch (ipdgram->proto)
    {
      case IPPROTO_TCP:
           {
             /* tcp source may only be a unicast IP address.
              * tcp dest may only be OUR address.
             */
             if (!bcast && !is_multicast(srcip) && is_local_addr(dstip))
               return ipdgram;
             DROPMSGSET("bcast/mcast/!forme");
           }
           break; /* go return NULL */
      case IPPROTO_UDP:  
           {
             /* udp source may only be 0.0.0.0 or a unicast IP address.
              * udp dest may only be bcast/mcast or OUR address.
             */
             DROPMSGSET("not for me");
             if (is_multicast(dstip))
             {
               if (_ip_mcast_is_in_group(dstip))
                 return ipdgram;
               DROPMSGSET("!in mcast grp");
             }
             else if (is_broadcast(srcip))
             {
               DROPMSGSET("bcast src");
             }
             else if (!my_statics.prim_ip_addr)
               return ipdgram;
             else if (is_broadcast(dstip))
             {
               if (my_statics.bcopen_count)
                 return ipdgram;
               DROPMSGSET("!conn bcast");
               /* don't show UDP bcast drops unless UDP debug_lev >= 3 */
               /* on an ether net with winstations there are many of these */
               if (proto_debug_lev < 3)
                 proto_debug_lev = 0;
             }
             else if (is_local_addr(dstip))
               return ipdgram;
           }
           break; /* go return NULL */
      case IPPROTO_ICMP:
           {
             /* icmp source may only be 0.0.0.0 or a unicast IP address.
              * icmp dest may only be bcast (router advert) or OUR address.
             */
             DROPMSGSET("bad source or mcast dest");
             if (is_broadcast(srcip) || is_multicast(dstip))
               ; /* fall through and drop */
             else if (bcast || is_local_addr(dstip))
               return ipdgram;
           }
           break; /* go return NULL */
      case IPPROTO_IGMP:
           {
             return ipdgram;
           }
      default:
           proto_debug_lev = my_statics.debug_lev;
           unhandled_proto = 1;
           break;
    }
  }

  if (unhandled_proto && my_statics.ip_raw_hook)
  {
    /* defrag of all non-standard protocols is the raw hook's problem */
    if ( (*(my_statics.ip_raw_hook))(ipdgram) )
    {
      if (my_statics.debug_lev >= 1)
        DBGPRINTF(("*IP IN-%04x: <= raw_handler() [handled?=1]\n", 
                     NTOHS(ipdgram->identification) ));
      return (const in_Header FAR *)0;
    }
  }

  if (unhandled_proto && !bcast /*hwbc that is*/ && is_local_addr(dstip))
  {
    icmp_retour( SOICMPERR_UNREACH_PROTO,
                 ipdgram, src_hw_addr, proto_debug_lev ); 
  }

  if (proto_debug_lev)
    DBGPRINTF(("*IP-%04x DROP %I->%I proto %u (%s)\n", 
               NTOHS(ipdgram->identification), ipdgram->source, 
               ipdgram->destination, ipdgram->proto, dropmsg ));

  return (const in_Header FAR *)0;
}

/* -------------------------------------------------------------- */

/*
 * _handler - handle incoming IP packets
 * Registered at startup with pcsed.
*/
static void __pkt_handler( const void FAR *ip_pkt, unsigned int ip_pkt_len, 
                           const void *src_hw_addr, int is_brdcast )
{
  const in_Header FAR *ipdgram;
  int protohndlr_flags;
  int proto_debug_lev;
  unsigned short pktid;

  /* check IP header - return 'ip_pkt' recast as ip packet */
  ipdgram = __chk_ip_header( ip_pkt, ip_pkt_len ); 
  if (!ipdgram)  /* not a valid IP packet */
  {
    /* debug message printed */
    return;
  }

  proto_debug_lev = __find_proto_debug_lev(ipdgram->proto);

  /* to-do: check for LSRR option and replace ip->source
   *        with actual source-address buried in option
   *        (ref. RFC1122)
  */

  /* do some basic filtering to reduce the amount of work done by
  ** the higher layers. Filter *before* we do any defragging.
  */
  ipdgram = __ip_filter( ipdgram, src_hw_addr, is_brdcast, proto_debug_lev );
  if (!ipdgram) /* not for us */
  {
    /* debug message printed */
    return;
  }

  pktid = NTOHS(ipdgram->identification);

  if (proto_debug_lev >= 1)
    DBGPRINTF(("*IP IN-%04x: %I->%I proto %u\n", pktid,
                ipdgram->source,ipdgram->destination, ipdgram->proto));

  if (proto_debug_lev >= 3)
    __dbg_dump_header("IN", ipdgram);

  /* we'll probably want to talk ipdgram->source, so add his ether addr 
  ** to the arp cache so that it doesn't need to be resolved.
  ** Note that _arp_add_entry() does not consume cache space for
  ** addresses that can be resolved without ARP lookups.
  */
  if (!is_brdcast)
    _arp_add_entry( NTOHL(ipdgram->source), src_hw_addr, 1 /*expires*/ );  

  /* increase the size of the recv queue if appropriate */
  __raise_ethnumbufs_if_needed(ipdgram);

  /* Push the packet through the defragger. 
  ** If defragmentation isn't required ...
  **       it returns 'ip' as is.
  ** Else ...
  **       If reassembly is now complete ...
  **         it returns a new 'ip'. 
  **       Else if error 
  **         it returns NULL
  **       Else (reassembly is not complete but no error)
  **         it returns NULL - packet will be completed later
  ** Note that even if _ip_frag_handler() returns a new pointer, the
  ** packet ID and other characteristics will be the same.
  */
  ipdgram = _ip_frag_handler( ipdgram, src_hw_addr, proto_debug_lev );
  if (!ipdgram)  /* reassembly not complete or error */
  {
    /* debug message printed */
    return; 
  }
  protohndlr_flags = 0;
  if (is_brdcast)
    protohndlr_flags |= PROTOHNDLR_FLAGS_LLBCAST;

  /* Match 'ipdgram' against SOCK_RAW sockets before regular multiplexing. */
  if (my_statics.ip_raw_hook)
  {
    if (proto_debug_lev >= 1)
      DBGPRINTF(("*IP IN-%04x: => raw_handler()\n", pktid ));
    if ( (*(my_statics.ip_raw_hook))(ipdgram) )
      protohndlr_flags |= PROTOHNDLR_FLAGS_RAWDONE;
    if (proto_debug_lev >= 1)
      DBGPRINTF(("*IP IN-%04x: <= raw_handler() [handled?=%d]\n", pktid,
                 (protohndlr_flags & PROTOHNDLR_FLAGS_RAWDONE)  ));
  }

  /* Reassembly complete or no reassembly necessary.
  ** now send the packet up to the higher layer by
  ** demultiplexing the packet to correct protocol handler.
  **
  ** Anything for a protocol that does not have a handler
  ** registered will already have been filtered out.
  */
  #if 0
  if (proto_debug_lev >= 1)
    DBGPRINTF(("*IP IN-%04x: passed up %I->%I, len=%u\n", pktid,
          ipdgram->source, ipdgram->destination, NTOHS(ipdgram->length) ));
  #endif
  (*(__find_proto_handler(ipdgram->proto)))(ipdgram, protohndlr_flags );

  /* We no longer need the packet. 
  ** If the packet was from the defragger
  ** then give it back for reuse.
  */
  _ip_frag_free(ipdgram); 

  pktid = pktid; /* unused unless debugging */
  return;
}

/* ======================================================================= */
/*                          transport initialization                       */
/* ======================================================================= */

static void __ctor_inet(void)
{
  __init_private_data(); /* set config hook etc */
  my_statics.ctors_are_working = CTOR_CALLED_MAGIC;

  /* register "downwards" */
  _eth_register_pkthandler( ETHERTYPE_IP, __pkt_handler );

  return;
}

#define __SETUP_CONSTRUCTOR_FOR__ __ctor_inet
#include "ctordtor.h"

