/*
 * Internet Protocol (IP) layer base functionality.
 * (RFC 791/RFC 760)
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * Portions Copyright (C) 1991, 1997 Geof Cooper and Rick Rodman (LGPL)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: ipbase.h,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
#ifndef __KERN_IP_BASE_H
#define __KERN_IP_BASE_H(x) __tkern_ ## x

#if defined(__WATCOMC__) && !defined(_ENABLE_AUTODEPEND)
#pragma read_only_file;
#endif

/* ================================================================== */
/*                    manifest constants used everywhere              */
/* ================================================================== */

/*
 * These are IP protocol numbers (see RFC-1700)
 */
#if !defined(IPPROTO_TCP) /* IPPROTO_* also defined in netinet/in.h */
#define IPPROTO_IP              0    /* pseudo protocol number for (raw) IP */
#define IPPROTO_ICMP            1    /* control message protocol */
#define IPPROTO_IGMP            2    /* group management protocol */
#define IPPROTO_TCP             6    /* transport core protocol */
#define IPPROTO_UDP             17   /* user datagram protocol */
#endif

/* ================================================================== */
/*               controlled access to IP kernel internals             */
/* ================================================================== */

#define _ip_get_ipaddr      __KERN_IP_BASE_H(ip_get_ipaddr)
#define _ip_set_ipaddr      __KERN_IP_BASE_H(ip_set_ipaddr)
#define _ip_get_snmask      __KERN_IP_BASE_H(ip_get_snmask)
#define _ip_get_bcaddr      __KERN_IP_BASE_H(ip_get_bcaddr)
#define _ip_get_default_ttl __KERN_IP_BASE_H(ip_get_default_ttl)
#define _ip_get_default_tos __KERN_IP_BASE_H(ip_get_default_tos)

extern uint32_t _ip_get_ipaddr(void);
extern int      _ip_set_ipaddr(uint32_t ipaddr);
extern uint32_t _ip_get_snmask(void);
extern uint32_t _ip_get_bcaddr(void);
extern uint8_t  _ip_get_default_ttl(void);
extern uint8_t  _ip_get_default_tos(void);

/* ================================================================== */
/*             functions and structures needed by protocols           */
/* ================================================================== */

#if (defined(__OSILAYER__) && (__OSILAYER__ == 3 || __OSILAYER__ == 4))

/*
 * These are the IP options
 */
#if !defined(IPOPT_EOL)  /* IPOPT_* is defined in netinet/ip.h */
#define IPOPT_EOL        0       /* end-of-option list */
#define IPOPT_NOP        1       /* no-operation */
#define IPOPT_RR         7       /* record packet route */
#define IPOPT_TS         68      /* timestamp */
#define IPOPT_SECURITY   130     /* provide s,c,h,tcc */
#define IPOPT_LSRR       131     /* loose source route */
#define IPOPT_SATID      136     /* satnet id */
#define IPOPT_SSRR       137     /* strict source route */
#define IPOPT_RA         148     /* router alert */
#endif

/* ----------------------------------------------------------------- */

/*
 * The Internet (ip) Header
 */
#pragma pack(1) /* align struct members on byte boundaries */
typedef struct in_Header {
        uint8_t   hdrlenver; /* low 4 bits = hdrlen, high 4 bits = ver */
                          /* standard hdr with no options == 0x45 */
        uint8_t   tos;
        uint16_t  length;
        uint16_t  identification;
        uint16_t  frag_ofs;
        uint8_t   ttl;
        uint8_t   proto;
        uint16_t  checksum;
        uint32_t  source;
        uint32_t  destination;
      } in_Header;
#pragma pack() /* restore default packing */

/* 'in_Header.frag_ofs' masks (apply to host-order value) */
#define IP_CE      0x8000u    /* Congestion Experienced */
#define IP_DF      0x4000u    /* Don't Fragment */
#define IP_MF      0x2000u    /* More Fragments */
#define IP_OFFMASK 0x1fffu    /* offset (left-shift by 3 to get offset) */

/* macro to get # of bytes in IP-header */
#define in_GetHdrLen(ip) ((((ip)->hdrlenver) & 0x0fu)<<2u) 
#define in_GetHdrVer(ip) ((((ip)->hdrlenver) & 0xf0u)>>4u) 

/* ----------------------------------------------------------------- */

/* flags used by ip_handler() calling the {tcp|udp|icmp|igmp}_handler()
** PROTOHNDLR_FLAGS_ICMPCB is set when icmp_handler() forward a message
** to the {udp|tcp}_handler, otherwise it is clear. If PROTOHNDLR_FLAGS_ICMPCB
** is set, then the low 8 bits of the protohndlr_flags is the icmp_type.
*/
#define PROTOHNDLR_FLAGS_ICMPCB  0x0100 /* handler is being called w/ icmp msg */
#define PROTOHNDLR_FLAGS_RAWDONE 0x0200 /* _ip_raw_hook has also processed */
#define PROTOHNDLR_FLAGS_LLBCAST 0x0400 /* this is a link layer broadcast */

/* --------------------------------------------------------------------- */

struct _ip_ifinfo 
{
  struct 
  {
    unsigned int up    : 1; /* interface is up */
    unsigned int lo    : 1; /* interface is loopback */
    unsigned int defif : 1; /* interface is default IF */
    unsigned int mcast : 1; /* interface can do mcast */
    unsigned int bcast : 1; /* interface can do bcast */
  } ipif_flags;
  unsigned short ipif_mtu;  /* interface's IP mtu */
  uint32_t ipif_host_addr;     /* interface's IP addr */
  uint32_t ipif_net_mask;      /* interface's net mask */
  uint32_t ipif_lcast_addr;    /* interface's broadcast addr */
  const char *ipif_name;    /* interface's name */
  char ipif_hwaddr[16];     /* interface's hardware address */
};

#define _ip_select_if __KERN_IP_BASE_H(ip_select_if)
extern int _ip_select_if (const uint32_t FAR *destaddr, 
                          struct _ip_ifinfo FAR *info);


/* --------------------------------------------------------------------- */

#define _ip_is_up           __KERN_IP_BASE_H(ip_is_up)
#define _ip_icmp_callback   __KERN_IP_BASE_H(ip_icmp_callback)
#define _ip_destaddr_open   __KERN_IP_BASE_H(ip_destaddr_open)
#define _ip_destaddr_close  __KERN_IP_BASE_H(ip_destaddr_close)
#define _ip_proto_register  __KERN_IP_BASE_H(ip_proto_register)
#define _ip_get_mtu         __KERN_IP_BASE_H(ip_get_mtu)

extern int   _ip_is_up          (void);
extern void  _ip_icmp_callback  (const in_Header FAR *orig_ip, int soicmperr_code );
extern int   _ip_destaddr_open  (uint32_t ina);
extern int   _ip_destaddr_close (uint32_t ina);
extern void  _ip_proto_register (int ip_proto, 
                                 void (*pkt_handler)(const in_Header FAR *ippkt, 
                                                    int protohndlr_flags),
                                 void (*stack_shutdown)(void),
                                 int FAR *debug_lev );
extern uint16_t  _ip_get_mtu        (void);

/* ------------------------------------------------------------------ */

#define _ip_output_raw      __KERN_IP_BASE_H(ip_output_raw)
#define _ip_output_1        __KERN_IP_BASE_H(ip_output_1)
#define _ip_output_2        __KERN_IP_BASE_H(ip_output_2)

extern int _ip_output_raw ( const void FAR *ipseg, 
                   unsigned int ipseglen, 
                   uint32_t dest_addr, 
                   int FAR *so_errno );
extern int _ip_output_1 ( const in_Header *hdrP,
                   uint32_t dest_addr, /* if different from hdrp->destination */
                   const void FAR *data_1,  /* protocol header ... */
                   unsigned int data_1_len, /* ... with or without data */
                   const void FAR *data_2,  /* more (optional) data */
                   unsigned int data_2_len,
                   int FAR *so_errno );
extern int _ip_output_2 ( uint32_t src_addr,   /* IP hdr from address (host order) */
                   uint32_t dst_addr,   /* IP hdr dest address (host order) */
                   uint8_t  protocol,   /* IP-protocol number */
                   uint8_t  ttl,        /* Time To Live */
                   uint8_t  tos,        /* Type Of Service, 0 unspecified */
                   uint16_t  flags,      /* eg IP_DF */
                   const void FAR *data_1, /* protocol header and/or data */
                   unsigned int data_1_len,
                   const void FAR *data_2, /* more data (optional) */
                   unsigned int data_2_len,
                   int FAR *so_errno );

/* ------------------------------------------------------------------ */
/*
 * tcp_PseudoHeader doesn't belong here, unless its renamed (as it should be).
 */
#pragma pack(1) /* align struct members on byte boundaries */
typedef struct tcp_PseudoHeader {
        uint32_t  src;
        uint32_t  dst;
        uint8_t   mbz;
        uint8_t   protocol;
        uint16_t  length;
        uint16_t  checksum;
      } tcp_PseudoHeader;
#pragma pack() /* restore default packing */

#endif /* __COMPILING_TRANSPORT_MODULE__ */
#endif /* __KERN_IP_BASE_H */

