/*
 * IP dynamic boot/configuration 
 * using RARP (RFC 903) 
 * and/or DHCP (RFC 1541/2131/2132)
 * and/or BOOTP (RFC 854/951/1048)
 *
 * Version
 *   0.5 : Oct 28, 1996 : Gisle Vanem - implemented RFC1541
 *   0.6 : May 18, 1997 : Gisle Vanem - added RFC2131 DHCPINFORM message
 *   1.0 : May 15, 2001 : Cyrus Patel - partial rewrite, support BOOTP,
 *                        integrated into IP kernel, Moved all statics 
 *                        to a static structure. Made dhcp_open() more 
 *                        general purpose. removed 'sock' from the statics 
 *                        list (always pass as arg). use _conf_add_confextn() 
 *                        instead of parse_args thing. Removed dhcp_init() 
 *                        [confextn hook is done by a constructor]. 
 *                        DHCP_send_release() can be static - called as 
 *                        _my_atexit() hook. Proper incremental delay.
 *                        Added break blocking/checking. Removed "HOST" 
 *                        config option (was being incorrectly used as 
 *                        the host to udp_open() to). 
 *                        _do_dhcp was incorrectly changing the mtu to 1500 - 
 *                        Fixed htype field (Was sending _pktdevclass - 
 *                        which is crap. The packet driver class has nothing 
 *                        to do with hardware type (that DIX is type 1 in 
 *                        both is a pure coincidence).
*/
/* - HELP - HELP - HELP - HELP - HELP - HELP - HELP - HELP -
 *
 * This DHCP client stuff requires a massive HACK to the stack to work.
 *
 * Usually (that is, on proper implementations), DHCP is tightly
 * integrated into the IP stack, such that the DHCP subsystem is always 
 * "live" and can be sent messages at any time (to force renew for example).
 *
 * THIS DHCP client however is simply tacked onto the IP stack, and ...
 * - actually uses the UDP transport for communications with the DHCP server. 
 *   This means that the whole UDP code and overhead is yanked in irrespective 
 *   of whether UDP is actually needed for anything else.
 * - allocates memory for communicating with the server, which is
 *   a possible point of failure.
 * - worst of all, it is completely blocking, a) the comunications to/from 
 *   the server are processed synchronously before returning control to IP.
 *   b) Since the RENEW/REBIND happens on a background process handler, this
 *   DHCP code is not actually allowed to yield.
 *
 * What needs to be done is rewrite the whole DHCP client as a state 
 * machine. All IP packets that are UDP and destined for port 68 
 * should then be passed directly to a receiver in the DHCP client code.
 * That receiver then processes the packet based on a 'state'.
 *
 * The DHCP state diagram is as follows:
 * ('states' are the labels inside boxes. For simplicity, I've left out
 * the flow lines required when no response is received. In such a case,
 * simple go back to the point of the previous 'Send ...').
 *
 *                       +------+
 *    +----------------->+ Init +
 *    |                  +------+
 *    |                      |
 *    |                      v
 *    |                 Send DHCPDISCOVER
 *    |                   (broadcast)
 *    |                      |
 *    |                      v  
 *    |                 +----+--+---+
 *    |                 | Selecting |<-----<-----+
 *    |                 +----+------+            |
 *    |                      |             Send DHCPDECLINE
 *    |                 got DHCPOFFER        (unicast)
 *    |                   |       |              |
 *    |                   v       +-->------>----+
 *    |                   |
 *    |            Send DHCPREQUEST                            
 *    |              (broadcast)                               
 *    |                   |     +---<-----<--+---<---<--------+
 *    |                   |     |            ^                ^    
 *    ^                   v     v            |                |
 *    |              +----+-------+   Send DHCPREQUEST  Send DHCPREQUEST
 *    |              + Requesting |     (broadcast)        (unicast)
 *    |              +----+-------+          |                |
 *    |               |   |                  |                |
 *    +--Got DHCPNAK--+   |                  |                |
 *                        |              +---+----+       +--------+
 *                    Got DHCPACK        | Rebind |       + Renew  +
 *                    - Record Lease     +--------+       +--------+
 *                    - Set T1 & T2          ^                ^ 
 *                        |                  |                |     
 *                     +-------+             |                |     
 *             +------>| Bound |             |                |
 *             |       +-------+             |                |
 *             |           |                 |                |
 *             ^       T2 expired? --> yes --+                |
 *             |           |                                  |
 *             |          (no)                                |
 *             |           |                                  |
 *             ^   got DHCPFORCERENEW? --> yes ---->------>---+
 *             |           |                                  |
 *             |          (no)                                |
 *             ^           |                                  |
 *             |       T1 expired? --> yes ------>-------->---+
 *             |           |
 *             +--<---<-- (no)
 *
*/
#define WANT_DHCPBOOTP_SUPPORT /* want DHCP (with inherant BOOTP) support */
#define WANT_ONLYBOOTP_SUPPORT /* want _only_ BOOTP support (requires less code) */
                               /* #undef both if you want neither */

#define __OSILAYER__ 3 /* network layer */

#include "target.h"   /* standard types/includes */
#include "ctordtor.h" /* my_atexit() */
#include "timer.h"    /* wintr_*(), Random() */
#include "blocker.h"  /* blocker_*() */
#include "debug.h"    /* DBGPRINTF() */
#include "userland.h" /* sock_*() */
#include "inaddr.h"   /* _my_ntoa() */
#include "linklayr.h" /* _eth_get_hwtype(), _eth_get_hwaddrlen() */
#include "arp.h"      /* _arp_vrfy_addr_change() */
#include "loadcfg.h"  /* _conf_add_confextn() */
#include "sigblk.h"   /* set_breakstate() */
#include "icmp.h"     /* icmp_mask_req */
#include "ipboot.h"   /* ourselves */

#define VM_RFC1048   0x63825363UL  /* magic cookie (in host order) */

enum { 
  DHCP_DISCOVER = 1,
  DHCP_OFFER,
  DHCP_REQUEST,
  DHCP_DECLINE,
  DHCP_ACK,
  DHCP_NAK ,
  DHCP_RELEASE,
  DHCP_INFORM,
  DHCP_FORCERENEW
};

enum { 
  DHCP_RENEWAL_TO = 1,
  DHCP_REBIND_TO
};

/*
 * DHCP options and values (ref. RFC1533)
 * Most are also BOOTP options.
 */
enum { 
  DHCP_OPT_PAD                      =  0,
  DHCP_OPT_SUBNET_MASK              =  1,
  DHCP_OPT_TIME_OFFSET              =  2,
  DHCP_OPT_ROUTERS_ON_SNET          =  3,
  DHCP_OPT_TIME_SRV                 =  4,
  DHCP_OPT_NAME_SRV                 =  5,
  DHCP_OPT_DNS_SRV                  =  6,
  DHCP_OPT_LOG_SRV                  =  7,
  DHCP_OPT_COOKIE_SRV               =  8,
  DHCP_OPT_LPR_SRV                  =  9,
  DHCP_OPT_IMPRESS_SRV              = 10,
  DHCP_OPT_RES_LOCATION_SRV         = 11,
  DHCP_OPT_HOST_NAME                = 12,
  DHCP_OPT_BOOT_FSIZE               = 13,
  DHCP_OPT_MERIT_DUMPFILE           = 14,
  DHCP_OPT_DOMAIN_NAME              = 15,
  DHCP_OPT_SWAP_SRV                 = 16,
  DHCP_OPT_ROOT_PATH                = 17,
  DHCP_OPT_EXTENTIONS_PATH          = 18,
  DHCP_OPT_IP_FORWARDING            = 19,
  DHCP_OPT_NON_LOCAL_SRC_ROUTE      = 20,
  DHCP_OPT_POLICY_FILTER            = 21,
  DHCP_OPT_MAX_DGRAM_REASM_SIZE     = 22,
  DHCP_OPT_IP_DEFAULT_TTL           = 23,
  DHCP_OPT_PATH_MTU_AGING_TIMEOUT   = 24,
  DHCP_OPT_PATH_MTU_PLATEAU_TABLE   = 25,
  DHCP_OPT_IF_MTU                   = 26,
  DHCP_OPT_ALL_SUBNETS_LOCAL        = 27,
  DHCP_OPT_BROADCAST_ADDR           = 28,
  DHCP_OPT_PERFORM_MASK_DISCOVERY   = 29,
  DHCP_OPT_MASK_SUPPLIER            = 30,
  DHCP_OPT_PERFORM_ROUTER_DISCOVERY = 31,
  DHCP_OPT_ROUTER_SOLICITATION_ADDR = 32,
  DHCP_OPT_STATIC_ROUTE             = 33,
  DHCP_OPT_TRAILER_ENCAPSULATION    = 34,
  DHCP_OPT_ARP_CACHE_TIMEOUT        = 35,
  DHCP_OPT_ETHERNET_ENCAPSULATION   = 36,
  DHCP_OPT_TCP_DEFAULT_TTL          = 37,
  DHCP_OPT_TCP_KEEPALIVE_INTERVAL   = 38,
  DHCP_OPT_TCP_KEEPALIVE_GARBAGE    = 39,
  DHCP_OPT_NIS_DOMAIN_NAME          = 40,
  DHCP_OPT_NIS_SRVS                 = 41,
  DHCP_OPT_NTP_SRVS                 = 42,
  DHCP_OPT_VENDOR_SPECIFIC_INFO     = 43,
  DHCP_OPT_NBIOS_NAME_SRV           = 44,
  DHCP_OPT_NBIOS_DGRAM_DIST_SRV     = 45,
  DHCP_OPT_NBIOS_NODE_TYPE          = 46,
  DHCP_OPT_NBIOS_SCOPE              = 47,
  DHCP_OPT_XFONT_SRV                = 48,
  DHCP_OPT_XDISPLAY_MANAGER         = 49,
  DHCP_OPT_REQUESTED_IP_ADDR        = 50,
  DHCP_OPT_IP_ADDR_LEASE_TIME       = 51,
  DHCP_OPT_OPT_OVERLOAD             = 52,
  DHCP_OPT_MSG_TYPE                 = 53,
  DHCP_OPT_SRV_IDENTIFIER           = 54,
  DHCP_OPT_PARAM_REQUEST            = 55,
  DHCP_OPT_MSG                      = 56,
  DHCP_OPT_MAX_MSG_SIZE             = 57,
  DHCP_OPT_T1_VALUE                 = 58,
  DHCP_OPT_T2_VALUE                 = 59,
  DHCP_OPT_CLASS_ID                 = 60,
  DHCP_OPT_CLIENT_ID                = 61,
  DHCP_OPT_NIS_DOMAIN_OPT           = 64,
  DHCP_OPT_NIS_SRV_OPT              = 65,
  DHCP_OPT_TFTP_SERVER              = 66,
  DHCP_OPT_BOOT_FILENAME            = 67,
  DHCP_OPT_MOBIP_HOME_AGENTS        = 68,
  DHCP_OPT_SMTP_SRVS                = 69,
  DHCP_OPT_POP3_SRVS                = 70,
  DHCP_OPT_NNTP_SRVS                = 71,
  DHCP_OPT_WWW_SRVS                 = 72,
  DHCP_OPT_FINGER_SRVS              = 73,
  DHCP_OPT_IRC_SRVS                 = 74,
  DHCP_OPT_STREET_TALK_SRVS         = 75,
  DHCP_OPT_STDA_SRVS                = 76,
  DHCP_OPT_POSIX_TZ_SPEC            = 88,
  DHCP_OPT_END                      = 255
};

#pragma pack(1) /* pack struct members on byte boundaries */
struct dhcp  {
       uint8_t  dh_op;      /* packet op code / message type.            */
       uint8_t  dh_htype;   /* RFC826 address type, 1 = DIX, 20 = slip   */
       uint8_t  dh_hlen;    /* hardware address len, eg '6' for 10mb eth */
       uint8_t  dh_hops;    /* client sets to zero, optionally used by   */
                            /* gateways in cross-gateway booting.        */
       uint32_t dh_xid;     /* transaction ID, a random number           */
       uint16_t dh_secs;    /* filled in by client, seconds elapsed      */
                            /* since client started trying to boot.      */
       uint16_t dh_flags;
       uint32_t dh_ciaddr;  /* client IP address filled in by client if known*/
       uint32_t dh_yiaddr;  /* 'your' (client) IP address                */
                            /* filled by server if client doesn't know   */
       uint32_t dh_siaddr;  /* server IP address returned in bootreply   */
       uint32_t dh_giaddr;  /* gateway IP address,                       */
                            /* used in optional cross-gateway booting.   */
       uint8_t  dh_chaddr[16]; /* client hardware address, filled by client */
       uint8_t  dh_sname[64];  /* optional server host name, null terminated*/

       uint8_t  dh_file[128];  /* boot file name, null terminated string    */
                            /* 'generic' name or null in bootrequest,    */
                            /* fully qualified directory-path            */
                            /* name in bootreply.                        */
       /* size upto here == 236 */

       uint8_t  dh_opt[(576-20)-236]; /* DHCP options area (minimum 308 bytes)  */
     };
#pragma pack() /* restore default packing */

#define BROADCAST_FLAG  1
#define BOOTREQUEST     1       /* bootp.bp_op */
#define BOOTREPLY       2
#define IPPORT_BOOTPS   (uint16_t)67
#define IPPORT_BOOTPC   (uint16_t)68

/* ================================================================ */

static struct
{
  uint32_t gateway       ; /* state */
#if defined(WANT_DHCPBOOTP_SUPPORT)
  uint32_t dhcp_server   ; /* state */
  uint32_t dhcp_renewal  ; /* state */
  uint32_t dhcp_rebind   ; /* state */
  uint32_t dhcp_iplease  ; /* state */
#endif
  int   force_bcast   ; /* state */
  uint32_t renewal_time  ; /* state */
  uint32_t rebind_time   ; /* state */

  uint32_t FAR *my_ip_addrP  ; /* passed by IP */
  uint32_t FAR *sin_maskP    ; /* passed by IP */
  int   ip_debug_lev  ; /* passed by IP */

  int   req_timeout   ; /* local conf */
  int   debug_lev     ; /* local conf */
#if defined(WANT_DHCPBOOTP_SUPPORT)   /* local conf */
  int   do_inform     ; /* local conf */
  struct {  /* DHCP.REQ_LIST= option */
    const uint8_t FAR *data;
    unsigned int size;
  } request_list;
#endif

} FAR my_statics;


/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT) || defined(WANT_ONLYBOOTP_SUPPORT)
static void *dhcp_open (udp_Socket *skt, uint32_t host)
{
  if (!host)
  {
#if defined(WANT_DHCPBOOTP_SUPPORT)
    host = my_statics.dhcp_server;
    if (!host || my_statics.force_bcast)
#endif
      host = (uint32_t)-1; /* broadcast */  
  }
  if (!sock_wattopen(__IPPROTO_UDP__, skt, IPPORT_BOOTPC, host, IPPORT_BOOTPS, NULL))
    return (void *)0;
  return (void *)skt;
}
#endif /* defined(WANT_DHCPBOOTP_SUPPORT) || defined(WANT_ONLYBOOTP_SUPPORT) */

/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT) || defined(WANT_ONLYBOOTP_SUPPORT)
static void __make_boot_header (struct dhcp *out, uint8_t op, int newid)
{
  static uint32_t xchg_id = 0;
  if (newid)
    xchg_id++;

  memset (out, 0, sizeof(*out));
  out->dh_op     = op;
  out->dh_htype  = _eth_get_hwtype();
  out->dh_hlen   = _eth_get_hwaddrlen();
  out->dh_xid    = HTONL(xchg_id);
  out->dh_secs   = HTONS((uint16_t)(timer_sec_count()));
  /* WinNT DHCP server _requires_ BROADCAST_FLAG if sending to bcast */
  out->dh_flags  = (my_statics.force_bcast) ? BROADCAST_FLAG : 0;
  out->dh_yiaddr = HTONL((*my_statics.my_ip_addrP));
  out->dh_ciaddr = out->dh_yiaddr;
  out->dh_giaddr = HTONL(my_statics.gateway);
  FAR_MEMCPY(&out->dh_chaddr[0], _eth_get_hwaddr(), _eth_get_hwaddrlen());
  *((uint32_t *)&(out->dh_opt[0])) = HTONL(VM_RFC1048);
}
#endif /* defined(WANT_DHCPBOOTP_SUPPORT) || defined(WANT_ONLYBOOTP_SUPPORT) */

/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT)
static uint8_t *__put_request_list (uint8_t *opt, unsigned int maxlen )
{
  static const uint8_t default_request_list[] = {
            DHCP_OPT_ROUTERS_ON_SNET,
            DHCP_OPT_DNS_SRV,
/*          DHCP_OPT_HOST_NAME, */
/*          DHCP_OPT_DOMAIN_NAME, */
            DHCP_OPT_POSIX_TZ_SPEC,
            DHCP_OPT_TIME_OFFSET,

            DHCP_OPT_LOG_SRV,
            DHCP_OPT_COOKIE_SRV,
            DHCP_OPT_LPR_SRV,
            DHCP_OPT_IP_DEFAULT_TTL
          };
  const uint8_t FAR *data = my_statics.request_list.data;
  unsigned int size = my_statics.request_list.size;
  unsigned int len;

  if (data && size && *data == 0)
  {
    /* If request list start with Pad option ("DHCP.REQ_LIST=0"),
     * disable options all-together.
    */
    return opt;
  } 

  if (maxlen > 0xffu) 
    maxlen = 0xffu;
  if (maxlen <= 2) /* DHCP_PARAM_REQUEST, size */
    return opt;
  maxlen -= 2;

  len = sizeof(default_request_list);
  if (len > maxlen)
    len = maxlen;
  memcpy( &opt[2], &default_request_list[0], len );

  if (data && size)
  {
    while (size && len < maxlen)
    {
      unsigned int n = 0;
      while (n < sizeof(default_request_list))
      {
        if (default_request_list[n] == *data)
          break;
        n++;
      } 
      if (n == sizeof(default_request_list))
        opt[2+(len++)] = *data;
      data++;
      size--;
    }
  }

  *opt++ = DHCP_OPT_PARAM_REQUEST;
  *opt++ = (uint8_t)len;
  opt += len;
  return (opt);
}
#endif /* DHCP */

/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT)
static int DHCP_send_inform (struct dhcp *out, void *skt)
{
  uint8_t *opt   = (uint8_t*) &out->dh_opt[4];
  uint8_t *start = opt;
  int len;

  __make_boot_header (out, BOOTREQUEST,1);
  *opt++ = DHCP_OPT_MSG_TYPE;
  *opt++ = 1;
  *opt++ = DHCP_INFORM;

  *opt++ = DHCP_OPT_CLIENT_ID;           /* Client ID option */
  *opt++ = (uint8_t)(_eth_get_hwaddrlen() + 1);
  *opt++ = _eth_get_hwtype();
  FAR_MEMCPY(opt, _eth_get_hwaddr(), _eth_get_hwaddrlen());
  opt += _eth_get_hwaddrlen();

  opt = __put_request_list (opt, (size_t)(sizeof(out->dh_opt) - ((opt-start)+1)) );
  if (((opt - ((uint8_t*)out)) & 1) == 0)
    *opt++ = DHCP_OPT_PAD; 
  *opt++ = DHCP_OPT_END;

  if (my_statics.debug_lev)
    DBGPRINTF(("*DHCP Sending inform.. (len=%u)\n", opt - (uint8_t*)out));
  len = sock_send_to(skt, (uint8_t*)out, (size_t)(opt - (uint8_t*)out), 
                     WSO_MSG_DONTWAIT, NULL, NULL );
  if (my_statics.debug_lev)
    DBGPRINTF(("*DHCP sent inform. sock_write() => %d\n", len ));
  return (len == (opt - (uint8_t*)out));
}
#endif /* DHCP */

/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT) || defined(WANT_ONLYBOOTP_SUPPORT)
static int DHCP_send_discover(struct dhcp *out, void *skt)
{
  uint8_t *opt   = (uint8_t*) &out->dh_opt[4];
  int len;

#if defined(WANT_DHCPBOOTP_SUPPORT)
  uint8_t *start = opt;
#endif

  __make_boot_header (out, BOOTREQUEST,1);
#if defined(WANT_DHCPBOOTP_SUPPORT)
  *opt++ = DHCP_OPT_MSG_TYPE;
  *opt++ = 1;
  *opt++ = DHCP_DISCOVER;

  *opt++ = DHCP_OPT_CLIENT_ID;            /* Client ID option */
  *opt++ = (uint8_t)(_eth_get_hwaddrlen() + 1);
  *opt++ = (uint8_t)_eth_get_hwtype();
  FAR_MEMCPY(opt, _eth_get_hwaddr(), _eth_get_hwaddrlen());
  opt +=  _eth_get_hwaddrlen();

  *opt++ = DHCP_OPT_MAX_MSG_SIZE;         /* Maximum DHCP message size */
  *opt++ = 2;
  *(uint16_t*)opt = HTONS(sizeof(struct dhcp));
  opt += 2;

#if 0 /* suggest a lease time */
  *opt++ = DHCP_IP_ADDR_LEASE_TIME;
  *opt++ = 4;
  *(uint32_t*)opt = HTONL(7 * 24UL * 60UL * 60UL); /* one week */
  opt += 4;
#endif

  opt = __put_request_list (opt, (size_t)(sizeof(out->dh_opt) - ((opt-start)+1)) );
  if (((opt - ((uint8_t*)out)) & 1) == 0)
    *opt++ = DHCP_OPT_PAD; 
  *opt++ = DHCP_OPT_END;
#endif /* defined(WANT_DHCPBOOTP_SUPPORT) */

  if (my_statics.debug_lev)
    DBGPRINTF(("*DHCP: Sending discover. (len=%u)...\n", opt - (uint8_t*)out ));
  len = sock_send_to(skt, (uint8_t*)out, (size_t)(opt - (uint8_t*)out), 
                     WSO_MSG_DONTWAIT, NULL, NULL );
  if (my_statics.debug_lev)
    DBGPRINTF(("*DHCP: sent discover. sock_write() => %d\n", len ));
  return (len == (opt - (uint8_t*)out));
}
#endif

/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT)
static int DHCP_renew_or_bind(struct dhcp *out, int is_renew_call, 
                              uint32_t ciaddr, void *skt)
{                                       /* send a renew or bind request */
  uint8_t *opt = (uint8_t*) &out->dh_opt[4];
  int len;

  __make_boot_header (out, BOOTREQUEST, 0);

  *opt++ = DHCP_OPT_MSG_TYPE;
  *opt++ = 1;
  *opt++ = DHCP_REQUEST;

  if (!is_renew_call) /* BIND request */
  {
    *opt++ = DHCP_OPT_SRV_IDENTIFIER;
    *opt++ = sizeof (my_statics.dhcp_server);
    *(uint32_t*)opt = HTONL(my_statics.dhcp_server);
    opt += sizeof (my_statics.dhcp_server);
    *opt++ = DHCP_OPT_REQUESTED_IP_ADDR;
    *opt++ = sizeof (ciaddr);
    *(uint32_t*)opt = ciaddr;
    opt += sizeof (ciaddr);
  }

  if (my_statics.dhcp_iplease)
  {
    *opt++ = DHCP_OPT_IP_ADDR_LEASE_TIME;
    *opt++ = sizeof (my_statics.dhcp_iplease);
    *(uint32_t*)opt = HTONL(my_statics.dhcp_iplease);
    opt += sizeof (my_statics.dhcp_iplease);
  }
  if (((opt - ((uint8_t*)out)) & 1) == 0)
    *opt++ = DHCP_OPT_PAD; 
  *opt++ = DHCP_OPT_END;

  if (my_statics.debug_lev)
    DBGPRINTF(("*DHCP Sending %s request... (len=%u)\n", 
          ((is_renew_call) ? ("renew") : ("bind")), opt - (uint8_t*)out ));
  len = sock_send_to(skt, (uint8_t*)out, (size_t)(opt - (uint8_t*)out), 
                      WSO_MSG_DONTWAIT, NULL, NULL );
  if (my_statics.debug_lev)
    DBGPRINTF(("*DHCP sent renew/bind request. sock_write() => %d\n", len ));
  return (len == (opt - (uint8_t*)out));
}
#endif /* DHCP */

/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT)
static int DHCP_send_release_decline (struct dhcp *out, 
                                 uint8_t msg_type, const char *msg, void *skt)
{                                           /* send a release or decline */
  uint8_t *opt = (uint8_t*) &out->dh_opt[4];
  int len;

  if (my_statics.debug_lev && msg_type == DHCP_RELEASE)
    DBGPRINTF(("*DHCP release begin.\n"));

  __make_boot_header (out, BOOTREQUEST, 1);
  out->dh_secs = 0;

  *opt++ = DHCP_OPT_MSG_TYPE;
  *opt++ = 1;
  *opt++ = msg_type;
  *opt++ = DHCP_OPT_SRV_IDENTIFIER;
  *opt++ = sizeof (my_statics.dhcp_server);
  *(uint32_t*)opt = HTONL(my_statics.dhcp_server);
  opt += sizeof (my_statics.dhcp_server);

  if (msg)
  {
    len = strlen(msg);
    if (len > 0)
    {
      if (len > 64)
        len = 64;
      *opt++ = DHCP_OPT_MSG;
      *opt++ = (uint8_t)len;
      memcpy(opt, msg, len);
      opt += len;
    }
  }
  if (((opt - ((uint8_t*)out)) & 1) == 0)
    *opt++ = DHCP_OPT_PAD; 
  *opt++ = DHCP_OPT_END;

  if (my_statics.debug_lev)
    DBGPRINTF(("*DHCP Sending %s... (len=%u)\n", 
                ((msg_type == DHCP_RELEASE) ? "release" : "decline" ),
                opt - (uint8_t*)out ));
  len = sock_send_to(skt, (uint8_t*)out, (size_t)(opt - (uint8_t*)out), 
                     WSO_MSG_DONTWAIT, NULL, NULL );
  if (my_statics.debug_lev)
    DBGPRINTF(("*DHCP sent release/decline. sock_write() => %d\n", len ));

  if (my_statics.debug_lev && msg_type == DHCP_RELEASE)
    DBGPRINTF(("*DHCP release end.\n" ));

  return (len == (opt - (uint8_t*)out));
}
#endif /* DHCP */

/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT)
static void DHCP_send_release (void)
{
  if (my_statics.debug_lev)
    DBGPRINTF(("* ---------- DHCP release begin\n"));

  if (my_statics.dhcp_server)
  {
    struct dhcp *dhcp_pkt;
    dhcp_pkt = (struct dhcp *)malloc(sizeof(*dhcp_pkt)+sizeof(udp_Socket));
    if (!dhcp_pkt)
    {
      if (my_statics.debug_lev)
        DBGPRINTF(("*DHCP release malloc() failed.\n"));
    }
    else
    {
      void *skt = dhcp_open ((udp_Socket *)(dhcp_pkt + 1), 
                             my_statics.dhcp_server); /* release is unicast */
      if (skt)
      {
        DHCP_send_release_decline (dhcp_pkt, DHCP_RELEASE, NULL, skt);
        sock_destroy(skt);
      }
      free(dhcp_pkt);  
    }
  }

  if (my_statics.debug_lev)
    DBGPRINTF(("* ---------- DHCP release complete\n"));
}
#endif /* DHCP */

/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT)
static int DHCP_is_ack (const struct dhcp *in)
{
  const uint8_t *opt = (const uint8_t*) &in->dh_opt[4];
  int rc = (opt[0] == DHCP_OPT_MSG_TYPE && opt[1] == 1 && opt[2] == DHCP_ACK);
  if (rc && my_statics.debug_lev)
    DBGPRINTF(("*DHCP got ack\n" ));
  return rc;
}
#endif /* DHCP */

/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT)
static int DHCP_is_nack (const struct dhcp *in)
{
  const uint8_t *opt = (const uint8_t *)&in->dh_opt[4];
  int rc = (opt[0] == DHCP_OPT_MSG_TYPE && opt[1] == 1 && opt[2] == DHCP_NAK);
  if (rc && my_statics.debug_lev)
    DBGPRINTF(("*DHCP got nack\n" ));
  return rc;
}
#endif /* DHCP */

/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT) || defined(WANT_ONLYBOOTP_SUPPORT)
static unsigned long __compute_timeout(int num_try,   /* >= 1 */
                                       unsigned long timeout) 
                                       /* total timeout (remaining) */
{
  unsigned int n = ((num_try > 4) ? 4 : num_try);
  unsigned long to = MUL32(1000ul , (2u << n)); 
  to += Random(0,(1000u * n));
  if (to > 30000ul)
    to = 30000ul;
  if (to > timeout)
    to = timeout;
  /*
  if (my_statics.debug_lev)
    DBGPRINTF(("recv try = %u, timeout = %lu.%03lu\n", 
                        num_try, to / 1000ul, to % 1000ul));
  */
  return to;
}
#endif

/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT)
/*
** MASSIVE HACK ALERT:
** 
** __skt_wait_readable() is functionally equivalent to 
**         sock_poll(skt, WSO_ASYSEL_READ, timeout ); 
** But because the address discovery/renew process is 
** is running as a periodic process (which means it 
** cannot yield), we cannot use sock_poll() (sock_poll() yields)
** Therefore, we must a) get the sock state by other means, 
** and b) call _eth_poll() manually.
**
** What we really, really need to do is make the discovery/renew
** process a state machine, with ip delivering the packet to
** us via a hook. That way, we would not need the UDP TPI,
** nor need the above mentioned hackery, nor need to allocate 
** buffers for reading the packet (which could fail).
*/
static int __skt_wait_readable(void *skt, unsigned long timeout)
{
  int oldbrkstate = set_cbreakstate(0);
  unsigned long start = timer_msec_count();
  int events = 0;

  blocker_critsect_begin();

  if (my_statics.debug_lev)
    DBGPRINTF(("*DHCP: begin __skt_wait_readable(to=%lu)\n", timeout ));
  
  while (!events)
  {
    events = WSO_ASYSEL_READ;
    if (sock_opt_getset(skt, WSO_LF_ASYNC_POLL, NULL, &events ) != 0)
      events = WSO_ASYSEL_ERROR;
    if (events != 0)
      break;
    if ((timer_msec_count() - start) > timeout)
      break; /* fallthrough and return 0 */
    _eth_poll();
  }

  if (my_statics.debug_lev)
    DBGPRINTF(("*DHCP: end __skt_wait_readable() => 0x%04x\n", events ));

  blocker_critsect_end();
  set_cbreakstate(oldbrkstate);

  return events;
}
#endif

/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT)
static int __do_renew(struct dhcp *dhcp_pkt, void *skt, int is_renewing)
{                                          /* renewing or rebinding state */
  int num_try, stop_now = 0;
  uint32_t timeout = MUL32(1000ul, my_statics.req_timeout);

  if (my_statics.req_timeout < 1) /* nothing set in config */
    timeout = 30000ul;            /* so default to 30 secs */
  if (timeout < 4000ul)
    timeout = 4000ul;

  for (num_try = 1;timeout > 0 && !stop_now && num_try <= 32; num_try++)
  {
    uint32_t xid;

    if (!DHCP_renew_or_bind(dhcp_pkt, is_renewing, HTONL((*my_statics.my_ip_addrP)), skt))
      break; /* sock_err */

    xid = dhcp_pkt->dh_xid;

    while (!stop_now)
    {
      unsigned long to = __compute_timeout(num_try, timeout);
      int ev = __skt_wait_readable(skt, to);
      if (ev == -1 || (ev & WSO_ASYSEL_READ) == 0) /* error or timeout */
      {             
        if (ev == -1)                     
          stop_now = 1; 
        else
          timeout -= to;
        break; 
      }
      memset (dhcp_pkt, 0, sizeof(*dhcp_pkt));
      sock_fastread (skt, dhcp_pkt, sizeof(*dhcp_pkt));

      if (dhcp_pkt->dh_op  != BOOTREPLY ||      /* not a BOOT reply    */
          dhcp_pkt->dh_xid != xid)  /* not our exchange ID */
         continue;
     
      my_statics.gateway = NTOHL(dhcp_pkt->dh_giaddr);
      if (DHCP_is_ack(dhcp_pkt))
      {
        /* RFC 2131, Section 4.4.1: 
         * The client SHOULD perform a check on the suggested address 
         * to ensure that the address is not already in use. For example, 
         * if the client is on a network that supports ARP, the 
         * client may issue an ARP request for the suggested request. 
         * When broadcasting an ARP request for the suggested address, 
         * the client must fill in its own hardware address as the sender's 
         * hardware address, and 0 as the sender's IP address, to avoid 
         * confusing ARP caches in other hosts on the same subnet. 
         * If the network address appears to be in use, the client MUST 
         * send a DHCPDECLINE message to the server.
        */
#if 0
        if ((*my_statics.my_ip_addrP) != 0)
        { 
          if (_arp_vrfy_addr_change((*my_statics.my_ip_addrP)) != 0)
          {
            if (my_statics.debug_lev)
              DBGPRINTF(("*DHCP ARP sez %hI is in use\n", 
                      (*my_statics.my_ip_addrP) ));
            (*my_statics.my_ip_addrP) = 0;
          }
        }
        if ( (*my_statics.my_ip_addrP) == 0 )
        {
          (*my_statics.my_ip_addrP) = 0;     /* decline from 0.0.0.0 */
          DHCP_send_release_decline (dhcp_pkt, DHCP_DECLINE, 
                                "IP address is not free",skt);
          blocker_delay(Random(4000, 6000));
          break; /* send another req. */
        }
#endif
        if (my_statics.debug_lev)
          DBGPRINTF(("*DHCP ARP sez %hI is not in use\n", 
                          (*my_statics.my_ip_addrP) ));
        stop_now = 1;
        break; /* OK return */
      }
      else if (DHCP_is_nack(dhcp_pkt))
      {
        (*my_statics.my_ip_addrP) = 0;
        stop_now = 1; /* sock_err */
        break;
      }

    } /* while (!stop_now) */
  } /* for (num_try,...) */

  return ((*my_statics.my_ip_addrP) != 0);
}
#endif /* DHCP */

/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT)
static void __do_inform(struct dhcp *dhcp_pkt, void *skt)
{
  if ((*my_statics.my_ip_addrP) != 0 && my_statics.do_inform && my_statics.dhcp_server)
  {
    if (DHCP_send_inform (dhcp_pkt, skt) )
    {
      int got_ack = 0;
      uint32_t xid = dhcp_pkt->dh_xid;
      int ev = __skt_wait_readable(skt, 4000ul );

      if (ev != -1 && (ev & WSO_ASYSEL_READ) != 0) 
      {
        sock_fastread( skt, dhcp_pkt, sizeof(*dhcp_pkt));

        if (dhcp_pkt->dh_op  == BOOTREPLY &&      /* a BOOT reply    */
            dhcp_pkt->dh_xid == xid)              /* our exchange ID */
        {
          if (DHCP_is_ack(dhcp_pkt))
            got_ack = 1;
        }
      }
      if (!got_ack && my_statics.debug_lev)
        DBGPRINTF(("*DHCP inform couldn't get DHCP ack\n" ));
    }
  } /* if (my_statics.do_inform) */
  return;
}
#endif /* DHCP */

/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT) || defined(WANT_ONLYBOOTP_SUPPORT)
static const char *__my_uitoa10( unsigned int val, char *buffer, 
                                 unsigned int bufsz )
{
  buffer[--bufsz] = '\0';
  if (!val)
    buffer[--bufsz] = '0';
  else
  {
    while (val && bufsz)
    {
      buffer[--bufsz] = (char)('0' + (val % 10u));
      val /= 10u;
    }   
  }
  return &buffer[bufsz];
}
#endif /* DHCP || BOOTP */

/*-------------------------------------------------------------------*/

/* handle all options - called recursively for DHCP_OPT_OVERLOAD 
*/
#if defined(WANT_DHCPBOOTP_SUPPORT) || defined(WANT_ONLYBOOTP_SUPPORT)
static int __do_options( const uint8_t *opt /* start */, 
                         unsigned int maxlen /* optlen */, 
                         const struct dhcp *base )
{
  int    got_offer = 0;
  char   scratch[260];
  unsigned int len;
  const uint8_t   *opt_end;
  uint8_t   c;
  uint32_t  u32;

  if (!opt && base)
  {
    if (*((const uint32_t *)&(base->dh_opt[0])) != HTONL(VM_RFC1048))
      return 0;
    opt = &base->dh_opt[4];
    maxlen = sizeof(base->dh_opt)-4;
  }
  if (!opt) /* should never happen */
    return 0;

  opt_end = opt + maxlen;

  while (opt < opt_end)
  {
    if (*opt == DHCP_OPT_END)
    { 
      if (my_statics.debug_lev)
        DBGPRINTF(("*  DHCP_OPT_END.\n"));
      break;
    }

    if (*opt == DHCP_OPT_PAD)
    { 
      if (my_statics.debug_lev)
        DBGPRINTF(("*  DHCP_OPT_PAD.\n"));
      opt++;
      continue;
    }

    switch (*opt)
    {
      case DHCP_OPT_SUBNET_MASK:
      {
        if (opt[1] % 4 == 0)           /* length = n * 4 */
        {
          *(my_statics.sin_maskP) = NTOHL(*(const uint32_t*)(opt+2));
          if (my_statics.debug_lev)
            DBGPRINTF(("*  SUBNET_MASK: %hI\n", *(my_statics.sin_maskP) ));
        }
        break;
      }     
      case DHCP_OPT_TIME_OFFSET:
      {
        unsigned int n;
        /* range is 0xFFFF6550 (GMT-11) ... 0x0000A8CD (GMT+12) */
        /* PST (GMT-8) == FFFF8F80 (-28800). CET (GMT+1) == 00000E10 */
        c = '+';
        u32 = NTOHL(*(const uint32_t*)(opt+2)); /* seconds east */
        if ((u32 >> 16) != 0)
        {
          c = '-';
          u32 = -u32;
        }  
        len = sizeof(scratch);
        scratch[--len] = '\0';
        n = (unsigned)DIV32(MOD32(u32,3600UL),60UL);
        scratch[--len] = (char)('0'+(n%10));
        scratch[--len] = (char)('0'+(n/10));
        n = (unsigned)DIV32(u32,3600UL);
        scratch[--len] = (char)('0'+(n%10)); 
        scratch[--len] = (char)('0'+(n/10));
        scratch[--len] = (char)c; 
        scratch[--len] = 'x';
        scratch[--len] = 'x';
        scratch[--len] = 'x';
        scratch[--len] = '=';
        scratch[--len] = 'Z';
        scratch[--len] = 'T';
        putenv(&scratch[len]);
        /* don't actually do a tzset() though. */
        if (my_statics.debug_lev)
          DBGPRINTF(("*  TIME_OFFSET: %s\n", &scratch[len+6] ));
        break;
      }
      case DHCP_OPT_ROUTERS_ON_SNET:
      {
        if (opt[1] % 4 == 0)           /* length = n * 4 */
        {
          for (len = 0; len < *(opt+1); len += sizeof(uint32_t))
          {
            u32 = NTOHL(*(const uint32_t*)(opt+2));
            _arp_add_gateway (u32, 0UL /* subnet */, 0UL /*mask*/);
            if (my_statics.debug_lev)
              DBGPRINTF(("*  Gateway: %hI\n", u32 ));
          }
        }
        break;
      }
      case DHCP_OPT_DNS_SRV:
      {
        if (opt[1] % 4 == 0)           /* length = n * 4 */
        {
          for (len = 0; len < *(opt+1); len += sizeof(uint32_t))
          {
            u32 = NTOHL(*(const uint32_t*)(opt+2+len));
            _conf_passopt("NAMESERVER", _my_ntoa(u32, NULL, 0));
            if (my_statics.debug_lev)
              DBGPRINTF(("*  DNS_SRV: %hI\n", u32 ));
          }
        }
        break;
      }
      case DHCP_OPT_LOG_SRV:
      {
        if (opt[1] % 4 == 0)           /* length = n * 4 */
        {
          u32 = NTOHL(*(const uint32_t*)(opt+2)); /* only the first one */
          _conf_passopt("SYSLOG.HOST", _my_ntoa(u32, NULL, 0));
          if (my_statics.debug_lev)
            DBGPRINTF(("*  LOG_SRV: %I\n", u32 ));
        }
        break;
      }
      case DHCP_OPT_HOST_NAME:
      {
        memcpy( scratch, &opt[2], opt[1] );
        scratch[opt[1]] = '\0';
        _conf_passopt("HOSTNAME", scratch );
        if (my_statics.debug_lev)
          DBGPRINTF(("*  HOST_NAME: '%s'\n", scratch ));
        break;
      }
      case DHCP_OPT_DOMAIN_NAME:
      {
        memcpy( scratch, &opt[2], opt[1] );
        scratch[opt[1]] = '\0';
        _conf_passopt("DOMAINNAME", scratch );
        if (my_statics.debug_lev)
          DBGPRINTF(("*  DOMAIN_NAME: '%s'\n", scratch ));
        break;
      }
      case DHCP_OPT_IP_DEFAULT_TTL:
      {
        _conf_passopt("IP.DEF_TTL", 
             __my_uitoa10( opt[2], scratch, sizeof(scratch) ));
        if (my_statics.debug_lev)
          DBGPRINTF(("*  DEFAULT_IP_TTL: %d\n", opt[2] ));
        break;
      }
      case DHCP_OPT_PERFORM_ROUTER_DISCOVERY:
      {
        if (my_statics.debug_lev)
          DBGPRINTF(("*  DO_ROUTER_DISC: Yes\n" ));
        break;
      }
      case DHCP_OPT_ROUTER_SOLICITATION_ADDR:
      {
        if (opt[1] % 4 == 0)           /* length = n * 4 */
        {
          u32 = NTOHL(*(const uint32_t*)(opt+2));
          if (my_statics.debug_lev)
            DBGPRINTF(("*  ROUTER_DISC_SERV: %hI\n", u32 ));
        }
        break;
      }
      case DHCP_OPT_MSG_TYPE:
      {
        if (opt[2] == DHCP_OFFER)
          got_offer = 1;
        break;
      } 
      case DHCP_OPT_MSG:
      {
        if (my_statics.debug_lev)
        {
          len = 0;
          while (len < opt[1])
          {
            c = opt[2+len];
            if (c == '\r' || c == '\n' || len == 70)
              break;
            if ((c & 0x80)!=0 || (c < 26 && c!='\t'))
              break; /* contains binary data */
          }
          if (len > 0)
            DBGPRINTF(("*  DHCP_MSG: %.*s\n", len, opt+2 ));
        }
        break;
      }
#if defined(WANT_DHCPBOOTP_SUPPORT)
      case DHCP_OPT_SRV_IDENTIFIER:
      {
        my_statics.dhcp_server = NTOHL(*(const uint32_t*)(opt+2));
        if (my_statics.debug_lev)
          DBGPRINTF(("*  DHCP_SRV: %I\n", *(const uint32_t*)(opt+2) ));
        break;
      }
      case DHCP_OPT_IP_ADDR_LEASE_TIME:
      {
        my_statics.dhcp_iplease = u32 = NTOHL(*(const uint32_t*)(opt+2));
        if (my_statics.debug_lev)
          DBGPRINTF(("*  IP_ADDR_LEASE: %s\n", 
                   DBG_T2A(u32, 1, scratch, sizeof(scratch)) ));
        break;
      }
      case DHCP_OPT_T1_VALUE:
      {
        my_statics.dhcp_renewal = u32 = NTOHL(*(const uint32_t*)(opt+2));
        if (my_statics.debug_lev)
          DBGPRINTF(("*  DHCP_T1 (renewal): %s\n", 
                DBG_T2A(u32, 1, scratch, sizeof(scratch)) ));
        break;
      }
      case DHCP_OPT_T2_VALUE:
      {
        my_statics.dhcp_rebind = u32 = NTOHL(*(const uint32_t*)(opt+2));
        if (my_statics.debug_lev)
          DBGPRINTF(("*  DHCP_T2 (rebind): %s\n", 
               DBG_T2A(u32, 1, scratch, sizeof(scratch)) ));
        break;
      }
#endif
      case DHCP_OPT_TCP_KEEPALIVE_INTERVAL: /* seconds */
      {
        u32 = NTOHL(*(const uint32_t*)(opt+2));
        if (my_statics.debug_lev)
          DBGPRINTF(("*  DHCP_KEEPALIVE_INTVL: %s\n", 
                      DBG_T2A(u32, 1, scratch, sizeof(scratch)) ));
        if (u32 > INT_MAX)
          u32 = INT_MAX;
        _conf_passopt("TCP.KEEPALIVE.INTVL", 
           __my_uitoa10( (unsigned int)u32, scratch, sizeof(scratch) ));
        break;
      }
      case DHCP_OPT_OPT_OVERLOAD:
      {
        if (base && opt[2] >= 1 && opt[2] <= 3)
        {
          if (opt[2] & 0x01) /* 1 or 3 */
          {
            if (my_statics.debug_lev)
              DBGPRINTF(("*  Overload: 'dh_file' options\n"));
            if (__do_options(&base->dh_file[0], 
                                 sizeof(base->dh_file), NULL))
              got_offer = 1;
          }
          if (opt[2] & 0x02) /* 2 or 3 */
          {
            if (my_statics.debug_lev)
              DBGPRINTF(("*  Overload: 'dh_sname' options\n"));
            if (__do_options( &base->dh_sname[0], 
                                  sizeof(base->dh_sname), NULL))
              got_offer = 1;
          }
        }
        break;
      }
      case DHCP_OPT_TFTP_SERVER:
      {
        if (opt[1] % 4 == 0)           /* length = n * 4 */
        {
          u32 = NTOHL(*(const uint32_t*)(opt+2));
          _conf_passopt("TFTP.SERVER", _my_ntoa(u32,NULL,0) );
          if (my_statics.debug_lev)
            DBGPRINTF(("*  TFTP-SRV: '%hI'\n", u32 ));
        }
        break;
      }
      case DHCP_OPT_BOOT_FILENAME:
      {
        memcpy( scratch, &opt[2], opt[1] );
        scratch[opt[1]] = '\0';
        _conf_passopt("TFTP.BOOT_FILE", scratch );
        if (my_statics.debug_lev)
          DBGPRINTF(("*  DHCP_BOOT_FILENAME: '%s'\n", scratch ));
        break;
      }
      case DHCP_OPT_POSIX_TZ_SPEC: 
      {
        scratch[0] = 'T'; scratch[1] = 'Z'; scratch[2] = '=';
        memcpy( &scratch[3], &opt[2], opt[1] );
        scratch[3+opt[1]] = '\0';
#if defined(__WATCOMC__) || defined(__GNUC__) /* others can't do POSIX TZ */
        putenv( scratch );
        /* don't actually do a tzset() though. */
#endif
        if (my_statics.debug_lev)
          DBGPRINTF(("*  DHCP_POSIX_TZ_SPEC: '%s'\n", &scratch[3] ));
        break;
      }
      default:
      {
        if (my_statics.debug_lev)
          DBGPRINTF(("*  ignorning option %d\n", *opt));
        break;
      }
    } /* switch() */
    opt += *(opt+1) + 2;
  }

  return (got_offer);
} 
#endif /* DHCP || BOOTP */

/*-------------------------------------------------------------------*/

#if defined(WANT_DHCPBOOTP_SUPPORT) || defined(WANT_ONLYBOOTP_SUPPORT)
static int __do_get_offer( struct dhcp *dhcp_pkt, void *skt )
{
  int try_num, stopnow = 0, got_dhcp_offer = 0;
  uint32_t timeout = 10000ul; /* default to 10 secs */

  if (my_statics.req_timeout >= 1) /* was set in config */
  {
    timeout = 4000ul; /* minimum */
    if (my_statics.req_timeout > 4)
      timeout = MUL32(1000ul, my_statics.req_timeout);
  }

  for (try_num = 1; timeout > 0 && !stopnow && try_num <= 32; try_num++)
  {
    uint32_t xid;

    if (!DHCP_send_discover(dhcp_pkt, skt ))
      break; /* sock_err */
    xid = dhcp_pkt->dh_xid;

    while (!stopnow)
    {
      unsigned long to = __compute_timeout( try_num, timeout );
      int ev = __skt_wait_readable(skt, to );

      if (ev == -1 || (ev & WSO_ASYSEL_READ) == 0) /* error or timeout */
      {                                  
        if (ev == -1)
          stopnow = 1; 
        else
          timeout -= to;
        break; 
      }

      if (my_statics.debug_lev)
        DBGPRINTF(("DHCP: Begin sock_read(len=%u)\n", sizeof(*dhcp_pkt) ));

      memset( dhcp_pkt, 0, sizeof(*dhcp_pkt) );

      if (sock_fastread (skt, dhcp_pkt, sizeof(*dhcp_pkt)) < 44)
      {
        /* not enough data for important parts of header */
        break; /* resend in next try_num loop */
      }

      if (my_statics.debug_lev)
        DBGPRINTF(("DHCP: End sock_read()\n", sizeof(*dhcp_pkt) ));

      if (dhcp_pkt->dh_op  == BOOTREPLY &&      /* a BOOT reply    */
          dhcp_pkt->dh_xid == xid)              /* our exchange ID */
      {
        my_statics.gateway = NTOHL(dhcp_pkt->dh_giaddr);
        (*my_statics.my_ip_addrP) = NTOHL(dhcp_pkt->dh_yiaddr);

        if (my_statics.debug_lev)
          DBGPRINTF(("DHCP: Begin __do_options()\n" ));

        got_dhcp_offer = __do_options(NULL, 0, dhcp_pkt);

        if (my_statics.debug_lev)
          DBGPRINTF(("*DHCP Got %s ((*my_statics.my_ip_addrP) = %hI)\n", 
            (got_dhcp_offer ? "DHCP offer" : "BOOTP reply"), (*my_statics.my_ip_addrP) ));

#if defined(WANT_DHCPBOOTP_SUPPORT)
        /* fixup some DHCP defaults */
        if (my_statics.dhcp_iplease == 0L) /* cannot happen ? */
          my_statics.dhcp_iplease = 3600L; /* default */
        if (my_statics.dhcp_renewal == 0L) /* default the T1 time */
          my_statics.dhcp_renewal = my_statics.dhcp_iplease >> 1;     
        if (my_statics.dhcp_rebind == 0) /* default the T2 time */
          my_statics.dhcp_rebind = MUL32(my_statics.dhcp_iplease,7)>>4;  
#endif

        stopnow = 1;
        break; /* OK */
      }
    } /* while (!stopnow) */
  } /* for (try_num = 1; ... ) */

  return got_dhcp_offer; /* its dhcp, not bootp */
}
#endif /* DHCP || BOOTP */

/*-------------------------------------------------------------------*/

/*
 * __idle_handler() - called via a blocker_idlehook_add()
 *
 * This MUST be an idle callback and not periodic or isr
 * because we need to wait when doing net I/O
*/
static void __idle_callback(void) 
{
  uint32_t now;
  int t1, t2;

  now = timer_sec_count();
  t1 = (now >= my_statics.renewal_time);
  t2 = (now >= my_statics.rebind_time);

  if (!t1 && !t2)
     return;

  my_statics.force_bcast = 0;           /* can Unicast now       */
  if (t1) my_statics.gateway  = 0;      /* don't use relay agent */
  if (t2) my_statics.force_bcast = 1;   /* must use Broadcast    */

  _ip_address_boot(NULL, NULL, 0); /* renew or retry */
}

/*-------------------------------------------------------------------*/

static void __fixup_netmask_if_needed(void)
{
  if (*my_statics.my_ip_addrP && !*my_statics.sin_maskP)
  {
    /* send an ICMP_MASK_REQ */
    uint32_t new_mask = icmp_mask_req(5/*secs timeout*/); 
    if (new_mask != 0)
      *my_statics.sin_maskP = new_mask;
    else
    {
      *my_statics.sin_maskP = 0xFFFFFF00L; /* assume class C */
      if (((*my_statics.my_ip_addrP) & 0x80000000ul) == 0) 
        *my_statics.sin_maskP = 0xFF000000L;  /* IN_CLASSA */
      else if (((*my_statics.my_ip_addrP) & 0xC0000000ul) == 0x80000000ul)
        *my_statics.sin_maskP = 0xFFFFFE00L; /* minimal IN_CLASSB */
    }
  }
  return;
}

/* this is the DHCP/BOOTP/RARP frontend called 
** only from _ip_address_fixup() 
*/
static void __do_dhcp_bootp_rarp( int is_first_time, int is_dhcp_renewing )
{
  int is_dhcp = 0;

  if (my_statics.debug_lev)
    DBGPRINTF(("*DHCP: __do_dhcp_bootp_rarp(first_time=%d,dhcp_renewing=%d)\n",
                                      is_first_time, is_dhcp_renewing ));

  blocker_idlehook_del(__idle_callback);  /* don't run callback now */
  is_dhcp_renewing = is_dhcp_renewing; /* possibly unused */

  #if defined(WANT_DHCPBOOTP_SUPPORT) || defined(WANT_ONLYBOOTP_SUPPORT)
  {
    struct dhcp *dhcp_pkt;

    dhcp_pkt = (struct dhcp *)malloc(sizeof(*dhcp_pkt)+sizeof(udp_Socket));
    if (!dhcp_pkt)
    {
      if (my_statics.ip_debug_lev || my_statics.debug_lev)
        DBGPRINTF(("*IP: DHCP malloc() failed\n" ));
    }
    else
    {
      void *skt = dhcp_open ((udp_Socket *)(dhcp_pkt + 1), 0);
      if (skt)
      {
        int oldbrkstate = set_cbreakstate(0);

        if (my_statics.ip_debug_lev || my_statics.debug_lev)
        {
          if (is_dhcp_renewing)
            DBGPRINTF(("*IP: reconfiguring via DHCP renew/rebind...\n"));
          else
            DBGPRINTF(("*IP: configuring via DHCP/BOOTP...\n"));
        }

        if (!is_dhcp_renewing)
        {
          is_dhcp = __do_get_offer( dhcp_pkt, skt );
          __fixup_netmask_if_needed();
        }
#if defined(WANT_DHCPBOOTP_SUPPORT)
        if (is_dhcp && (*my_statics.my_ip_addrP)) 
        {
          if (!__do_renew( dhcp_pkt, skt, is_dhcp_renewing ))
          {
            (*my_statics.my_ip_addrP) = 0;
          }
          else if (!is_dhcp_renewing) /* first time boot */
          {
            /* KMSG((0,"DHCP: _my_atexit(DHCP_send_release)\n")); */
            _my_atexit(DHCP_send_release); /* be nice and release our leased address */
  
            __do_inform( dhcp_pkt, skt );
          }
        }
#endif /* DHCP */

        if (*(my_statics.my_ip_addrP) == 0)
        {
          if (my_statics.ip_debug_lev || my_statics.debug_lev)
          {
            if (is_dhcp_renewing)
              DBGPRINTF(("*IP: DHCP renew failed (will retry)\n" ));
            else
              DBGPRINTF(("*IP: DHCP/BOOTP failed (will retry if RARP fails too)\n"));
          }
        }

        set_cbreakstate(oldbrkstate);
        sock_destroy(skt);

      } /* if (skt) */
      free(dhcp_pkt);
    } /* if (dhcp_pkt) */
  }
  #endif /* DHCP || BOOTP */

  /*
  ** Try RARP (only once)
  */
  if (*(my_statics.my_ip_addrP) == 0 && is_first_time)
  {                                    /* first call failed */
    if (my_statics.ip_debug_lev)
      DBGPRINTF(("*IP: Configuring via RARP...\n"));
    if (!_rarp_resolve( _eth_get_hwaddr(), my_statics.my_ip_addrP ))
    {
      if (my_statics.ip_debug_lev)
        DBGPRINTF(("*IP: RARP failed\n"));
    }
  }

  /* Set up the renew/retry hook. */
  if (!(*my_statics.my_ip_addrP) || is_dhcp)
  {
    static unsigned int retry_delay = 0;
    uint32_t now = timer_sec_count();
#if defined(WANT_DHCPBOOTP_SUPPORT)
    if (*my_statics.my_ip_addrP) /* RENEW/BIND succeeded */
    {
      retry_delay = 0;
      my_statics.renewal_time  = now + my_statics.dhcp_renewal;
      my_statics.rebind_time   = now + my_statics.dhcp_rebind;
      if (my_statics.rebind_time == my_statics.renewal_time)
        my_statics.rebind_time += 10;             /* add 10 seconds */
    }
    else /* retry getting an IP address in N secs */
#endif
    {
      /* 4, 6, 9, 13, 19, 28, 42, 63, 94, 94, 94, 94,... */
      retry_delay += ((retry_delay == 0) ? 4 : (retry_delay>>1));
      if (retry_delay > 94)
        retry_delay = 94;
      my_statics.renewal_time = now + retry_delay;
      my_statics.rebind_time  = now + retry_delay;
    }
    blocker_idlehook_add (__idle_callback); 
  }

  return;
}

/* ================================================================== */
/*                       public entry point                           */
/* ================================================================== */

/*
 * _ip_address_boot() - public entry point.
 *
 * The first time this is called, 'my_ip_addrP' and 'sin_maskP'
 * will point to IP's private values. These will be saved
 * for later use. This first call is a call to DISCOVER/BIND.
 *
 * Subsequent calls will be with my_ip_addrP/sin_maskP set to NULL. 
 * If the original DISCOVER/BIND had failed (my_ip_addr is still zero), 
 * then these calls will be calls to DISCOVER/BIND, otherwise they will 
 * be calls to RENEW/REBIND.
*/
int _ip_address_boot( uint32_t FAR *my_ip_addrP, uint32_t FAR *sin_maskP, 
                      int ip_debug_lev )
{
  int is_first_time = 0;
  int is_dhcp_renewing = 0;

  if (blocker_get_inisr_flag())
  {
    /* 
     * This should never happen since we are only called directly
     * during stack initialization (which is an ACTIVE operation),
     * or indirectly from an IDLE callback (which is only called
     * when the application yields explicitely).
    */
    __fatal_runtime_error("PANIC! _ip_address_boot() called from inside ISR\n");
  }

  if (my_ip_addrP) /* first call */
  {
    is_first_time = 1;
    _conf_load(); /* load config if not already loaded */

    my_statics.my_ip_addrP  = my_ip_addrP;
    my_statics.sin_maskP    = sin_maskP;
    my_statics.ip_debug_lev = ip_debug_lev;

    if (_eth_get_ifflags() & ETH_IFFLAGS_LBONLY)
    {
      *my_statics.my_ip_addrP = 0xAC1FFEFEul; /* 172.31.254.254 */
      *my_statics.sin_maskP = 0xffffff00ul; /* 255.255.255.0 */
    }
  }
  else if (!my_statics.my_ip_addrP || !my_statics.sin_maskP)
  {                  /* should never happen */
    return 0;
  }
  else if (*my_statics.my_ip_addrP)
  {
    is_dhcp_renewing = 1;
  }

  if (!(*my_statics.my_ip_addrP) || is_dhcp_renewing)
  {
    __do_dhcp_bootp_rarp( is_first_time, is_dhcp_renewing );
  } 

  if ((*my_statics.my_ip_addrP) != 0)
  {
    __fixup_netmask_if_needed();

    /* tell arp that _ip_get_ipaddr() has a new value.
     * if no gateways are on defined (on our subnet) ARP
     * will ask ICMP to solicit a gateway list.
    */
    _arp_vrfy_addr_change(0);

    if (!is_dhcp_renewing && my_statics.ip_debug_lev)
      DBGPRINTF(( "*IP: my IP addr=%hI, netmask=%hI\n", 
                *my_statics.my_ip_addrP, *my_statics.sin_maskP ));
  }

  return (*my_statics.my_ip_addrP != 0);
}

/* ================================================================== */
/*                          configuration silliness                   */
/* ================================================================== */

/*
 * Parse a list of DHCP-request options from config-file.
 * e.g DHCP_REQ_LIST = 1,23,24,28,36
 */
#if defined(WANT_DHCPBOOTP_SUPPORT)
static void __set_request_list (const char *opts)
{
  if (!my_statics.request_list.data)
  {
    uint8_t FAR *data = (uint8_t FAR *)FAR_MALLOC(0xffu);
    if (data)
    {
      unsigned int size = 0;
      while (*opts && size < 0xffu)
      {
        unsigned int val = 0, radix = 10;
        const char *start;

        while (strchr(" \t/\\,", *opts))
          opts++;
        if (*opts == '0')
        {
          radix = 8;
          opts++;
          if (*opts == 'x' || *opts == 'X')
          {
            radix = 16;
            opts++;
          }
        }
        start = opts;
        for (;;)
        { 
          unsigned int newval;
          uint8_t c = (uint8_t)0xff;
          if (*opts >= '0' && *opts <= '9')
            c = (uint8_t)(*opts++ - '0');
          else if (*opts >= 'A' && *opts <= 'F')
            c = (uint8_t)((*opts++ - 'A') + 10);
          else if (*opts >= 'a' && *opts <= 'f')
            c = (uint8_t)((*opts++ - 'a') + 10);
          else
            break;
          newval = (val * radix) + c;
          if (c >= radix || newval < val)
          {
            start = opts;
            break;
          }
          val = newval;
        } 
        if (start == opts)
          break;
        if (val == DHCP_OPT_MSG_TYPE           ||   
            val == DHCP_OPT_CLIENT_ID          ||
            val == DHCP_OPT_MAX_MSG_SIZE       ||
            val == DHCP_OPT_PARAM_REQUEST      ||
            val == DHCP_OPT_SRV_IDENTIFIER     ||
            val == DHCP_OPT_REQUESTED_IP_ADDR  ||
            val == DHCP_OPT_IP_ADDR_LEASE_TIME)
          break;
        if (val != DHCP_OPT_PAD)
        {
          if (val < 0xffu && (opts - start) < 4)
          {
            data[size++] = (uint8_t)val;
            if (val == DHCP_OPT_END)
              break;
          }
        }
      }
      if (size)
      {
        my_statics.request_list.data = data;
        my_statics.request_list.size = size;
        return;
      }  
      FAR_FREE(data);
    }
  }
  return;
}
#endif

/*-------------------------------------------------------------------*/

static void __set_op_timeout(const char *v) /* take highest */
{
  int i = atoi(v);
  if (i > my_statics.req_timeout)
    my_statics.req_timeout = i;
} 

static void __ctor_ina_boot(void)
{
  static const struct config_table FAR my_cfg[] = {
#if defined(WANT_DHCPBOOTP_SUPPORT)
    { "DHCP.REQ_LIST", ARG_FUNC,   (void FAR *)__set_request_list      },
    { "DHCP.INFORM",   ARG_ATOI,   (void FAR *)&(my_statics.do_inform) },
#endif
    { "DHCP.DEBUG",    ARG_ATOI,   (void FAR *)&my_statics.debug_lev   },
    { "DHCP.TIMEOUT",  ARG_FUNC,   (void FAR *)__set_op_timeout        },
    { "BOOTPTO",       ARG_FUNC,   (void FAR *)__set_op_timeout        },
    { "BOOTP_TO",      ARG_FUNC,   (void FAR *)__set_op_timeout        },
    { ((const char FAR *)0), 0,    ((void FAR *)0)  }
  };
  FAR_MEMSET( &my_statics, 0, sizeof(my_statics));

  _conf_add_confextn( &my_cfg[0], NULL );
  return;
}

#define __SETUP_CONSTRUCTOR_FOR__ __ctor_ina_boot
#include "ctordtor.h"

