/*
 * Packet defragmentation support for the Internet Protocol (IP).
 * (RFC 791/RFC 760)
 * 
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: ipfrag.c,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
#define WANT_IPFRAG_SUPPORT /* whether we do anything here or not */

#undef  FORCE_DEBUG  /* causes debug_level to be max(1,proto_debug_lev) */
#undef  TRACE_INSERT
#undef  TRACE_BKTOPS

#define __OSILAYER__ 3
#include "target.h"   /* standard types/includes */
#include "debug.h"    /* DBGPRINTF() */
#include "timer.h"    /* timer_sec_count() */
#include "blocker.h"  /* blocker_perdhook_add() */
#include "inaddr.h"   /* is_broadcast(), is_multicast(), is_local_addr() */
#include "inchksum.h" /* checksum() */
#include "icmp.h"     /* icmp_retour(), SOICMPERR_xxx */
#include "linklayr.h" /* MAX_LINKADDR_SIZE */
#include "ipbase.h"   /* in_Header */
#include "ipfrag.h"   /* ourselves */

/* ====================================================================== */

#pragma pack(1)
typedef struct 
{
  uint8_t   in_use;
  uint8_t   proto;
  uint16_t  ident;
  uint32_t  src_addr;
  uint32_t  dst_addr;
} frag_key;
#pragma pack()

#define MAX_BUCKETS (sizeof(int)<<1) /* 8 for 32bit, 4 for 16bit */
#define IPPACKET_MAX 0xffffu  /* max size of a reassembled IP packet */

#pragma pack(1)
typedef struct frag_bucket /* member order is important */
{
  uint8_t     map[(((unsigned long)IPPACKET_MAX)+7)/8];  /* one bit per octet */
  uint16_t    set_count;  /* count of bits that are set (no overlap) */
  uint16_t    total_len;  /* if known (known once !MF) */
  uint8_t     in_hdr_buf[64+(64/8)]; /* copy of in_Header + 64bits */
  uint8_t     FAR *buf;
  uint8_t     mru_hwaddr[MAX_LINKADDR_SIZE];
  frag_key    key;
  struct frag_bucket FAR *next;
  uint32_t    expires;
  int         debug_lev;  /* original debug_lev, used when releasing pkt */
} frag_bucket;
#pragma pack()

/*----------------------------------------------------------------------*/

#if defined(WANT_IPFRAG_SUPPORT)
/* update the map, and return count of non-overlapped bits. */
/* this is optimized code. */
static uint16_t __update_map(uint8_t FAR *map, unsigned int offset, unsigned int data_len)
{
#if (defined(__TURBOC__) && !defined(__FLAT__))
  /* very poor code generated - use inline asm instead */
  uint8_t FAR *pos = &map[offset>>3];
  unsigned int olap = 0;

  #define ASM_LABEL(x) } _##x __asm {
  __asm 
  {
    push es
    push di
    les  di, pos
    mov  dx, data_len

    mov  ah, dl
    and  ah, 7
    mov  cl, 3
    shr  dx, cl
    mov  cx, dx
    xor  dx, dx

    cmp  cx, 1
    jz   _byteadd
    jb   _fini
    test di,1
    jz   _byteadd

    ASM_LABEL(again:)

    cmp  cx, 1
    jz   _byteadd
    jb   _fini

    mov  bx,0ffffh
    xchg es:[di],bx    
    inc  di
    dec  cx
    inc  di
    dec  cx
    or   bx,bx    
    jz   _again
    add  dx,16
    cmp  bx,0ffffh
    jz   _again
    sub  dx,16

    ASM_LABEL(wordadd:)

    shr  bx,1
    adc  dx,0
    or   bx,bx
    jnz  _wordadd
    jmp  short _again

    ASM_LABEL(byteadd:)

    mov  al,0ffh
    xchg es:[di],al
    inc  di
    dec  cx    

    or   al,al
    jz   _again
    add  dx,8
    cmp  al,0ffh
    jz   _again
    sub  dx,8

    ASM_LABEL(bitadd:)

    shr  al,1
    adc  dx,0
    or   al,al
    jnz  _bitadd
    jmp  short _again

    ASM_LABEL(postadd:)

    mov  cl,ah
    mov  al,1
    shl  al,cl
    dec  al
    mov  ah,es:[di]
    and  ah,al
    or   es:[di],al
    mov  al,ah
    xor  ah,ah
    xor  cl,cl
    jmp  short _bitadd 

    ASM_LABEL(fini:)

    or   ah,ah
    jnz  _postadd
    mov  olap, dx
    pop  di
    pop  es
  }
#else
  register unsigned int len = (data_len >> 3);
  uint8_t FAR *pos = &map[offset>>3];
  unsigned int olap = 0;
  unsigned int c;

  while (len)
  {
    if ( (len >= sizeof(int)) && (((int)pos) & (sizeof(int)-1)) == 0 )
    {
      while (len >= sizeof(int))
      {
        c = *((unsigned int FAR *)pos);
        if (c)
        {
          olap += (sizeof(int)*8);
          if (c != ~0u)
          {
            olap -= (sizeof(int)*8);
            while (c)
            {
              olap += (c & 1);
              c >>= 1;
            }
          }
        }
        *((int FAR *)pos) = -1;
        pos += sizeof(int);
        len -= sizeof(int);
      }
    } /* if ((len >= sizeof(int)) && !(((int)pos) & (sizeof(int)-1))) */
    else
    {
      c = *pos;
      if (c)
      {    
        olap += 8;
        if (c != 0xff)
        {
          olap -= 8;
          while (c)
          {
            olap += (c & 1);
            c >>= 1;
          }
        }
      }
      *pos++ = 0xff;
      len--;
    }
  } /* while (len) */

  len = (data_len & 7); /* only happens if last fragment */
  if (len)
  {
    len = ((1u<<len)-1);
    c = *pos & len;
    *pos |= (uint8_t)(len);
    while (c)
    {
      olap += (c & 1);
      c >>= 1;
    }
  }
#endif

  return (uint16_t)(data_len - olap);
}
#endif /* WANT_IPFRAG_SUPPORT */

/*----------------------------------------------------------------------*/

#if defined(WANT_IPFRAG_SUPPORT)
static const in_Header FAR *__insert_frag( frag_bucket FAR *bkt, 
                                           const in_Header FAR *ippkt, 
                       uint16_t pktid, uint16_t flags, uint16_t offset, int debug_lev)
{
  uint16_t  data_len = (uint16_t)(NTOHS(ippkt->length) - in_GetHdrLen(ippkt));
  const uint8_t FAR *data_pos  = ((const uint8_t FAR *)ippkt)  + in_GetHdrLen(ippkt);
  unsigned int len;

#if defined(TRACE_INSERT)
#undef TRACE_INSERT
#define TRACE_INSERT(x) DBGPRINTF(x)
#else
#define TRACE_INSERT(x) NO_OPERATION()
#endif

  if (bkt->map[0] == 0) /* most recent in_hdr until we get the first frag */
  {
    /* full header plus at least 64bits */
    len = NTOHS(ippkt->length);
    if (len > sizeof(bkt->in_hdr_buf))
      len = sizeof(bkt->in_hdr_buf);
    FAR_MEMCPY( &(bkt->in_hdr_buf[0]), ippkt, len );
  }

  if (debug_lev)
    TRACE_INSERT(("offset = %u, data_len=%u IP_MF=%d\n", 
                    offset, data_len, ((flags & IP_MF)!=0) ));

  /* copy the data over */
  FAR_MEMCPY( &(bkt->buf[offset]), data_pos, data_len );

  /* set the total length (known once we get the last fragment) */
  if ((flags & IP_MF) == 0)
    bkt->total_len = (uint16_t)(offset + data_len);

  /* update the # of bits set (update map while counting overlapped bits) */
  bkt->set_count += __update_map( &bkt->map[0], offset, data_len );

  if (debug_lev)
    TRACE_INSERT(("set_count = %u, total_len = %u\n", bkt->set_count, bkt->total_len));

  ippkt = (const in_Header FAR *)0;
  if (bkt->total_len && bkt->set_count == bkt->total_len) /* got everything */
  {
    in_Header FAR *hdr = (in_Header FAR *)&(bkt->in_hdr_buf[0]);
    len = (uint16_t)in_GetHdrLen(hdr);

    if (bkt->total_len > (IPPACKET_MAX - len) || 
        bkt->total_len > (0xffffu - len) )
    {
      bkt->key.in_use = 0;
     
      if (debug_lev)
        TRACE_INSERT(("*IP-%04x defrag: killer pkt (hdrlen+datalen > %u) from %I\n", 
                   pktid, IPPACKET_MAX, ippkt->source ));
 
      #if 0
      if (IPPACKET_MAX < 0xffffu)
      {
        /* would sending a source quench help? */
      }
      #endif
      /* fall through and return NULL */
    }
    else
    {
      hdr->length    = HTONS((uint16_t)(len + bkt->total_len));
      hdr->frag_ofs &= HTONS(IP_CE);
      hdr->checksum  = 0;
      hdr->checksum  = ~checksum (hdr, len );

      FAR_MEMMOVE( bkt->buf + len, bkt->buf, (unsigned int)(bkt->total_len) );
      FAR_MEMCPY( bkt->buf, hdr, len );

      ippkt = (const in_Header FAR *)(&(bkt->buf[0]));
      bkt->expires = ((uint32_t)-1);

      if (debug_lev)
        TRACE_INSERT(("*IP-%04x defrag: reasm complete\n", pktid ));
    }
  }
  else 
  {
    if (debug_lev)
    {
      TRACE_INSERT(("*IP-%04x defrag: reasm incomplete - packet requeued\n", pktid));
      if (bkt->total_len || (flags & IP_MF) == 0)
        TRACE_INSERT(("*                (total_len=%u, got_so_far=%u)\n",
                                   bkt->total_len, bkt->set_count ));
    }
  }

  pktid = pktid; /* possibly unused */
  return ippkt;
}
#endif

/*----------------------------------------------------------------------*/

#if defined(WANT_IPFRAG_SUPPORT)
static int __badpkt_history_op(int op, const frag_key FAR *key)
{
  static frag_key FAR bad_packets[16];
  int free_slot = -1, match_slot = -1;
  unsigned int i;
  
  free_slot = -1;
  for (i = 0; i < (sizeof(bad_packets)/sizeof(bad_packets[0])); i++)
  {
    if ( !bad_packets[i].in_use )
    {
      if (free_slot == -1)
        free_slot = i;
    }
    else if ( FAR_MEMCMP( &bad_packets[i], key, sizeof(*key) ) == 0)
    {
      match_slot = i;
      break;
    }
  }

  if (op == '+') /* add to history */
  {
    if (match_slot == -1)
      match_slot = free_slot;
    if (match_slot == -1)
    {
      FAR_MEMMOVE( &bad_packets[0], &bad_packets[1], sizeof(bad_packets)-
                                                 sizeof(bad_packets[0]) );
      match_slot = (sizeof(bad_packets)/sizeof(bad_packets[0])) - 1;
    }
    FAR_MEMCPY( &bad_packets[match_slot], key, sizeof(*key) );
    bad_packets[match_slot].in_use = 1;
  }

  return (match_slot >= 0);
}
#endif /* WANT_IPFRAG_SUPPORT */

/*----------------------------------------------------------------------*/

#if defined(WANT_IPFRAG_SUPPORT)
static int __check_and_do_timeout(frag_bucket FAR *bkt, int fullchain, 
                                  unsigned long now)
{
  int rc = 0;
  while (bkt)
  {
    if (bkt->key.in_use && bkt->expires != ((uint32_t)-1)) /* work not completed */
    {
      if (now <= bkt->expires) /* not timed out yet */
      {
        blocker_inc_hwork();
      }
      else
      {
        bkt->key.in_use = 0;
        __badpkt_history_op('+', &(bkt->key));

        /* in_hdr_buf contains the full IP header plus at least 
         * 64bits (8 octets) of whatever follows the header.
        */
        icmp_retour( SOICMPERR_TIMXCEED_FRAG, &(bkt->in_hdr_buf[0]),
                     &bkt->mru_hwaddr[0], bkt->debug_lev );

        if (bkt->debug_lev)
        {
          DBGPRINTF(("*IP-%04x defrag: bucket=%p timed out\n", 
                              NTOHS(bkt->key.ident), bkt ));
        }
        rc++;
      }
    }
    if (!fullchain)
      break;
    bkt = bkt->next;
  }
  return rc; 
}
static void __check_timeouts(void); /* forward reference */
#endif

/*----------------------------------------------------------------------*/

#if defined(WANT_IPFRAG_SUPPORT)
/* __bucket_op does all bucket operations:
 * timeout check    : key == NULL, ippkt == NULL       
 * release by key   : key != NULL, ippkt == NULL       
 * release fragchain: key == NULL, ippkt != NULL
 * alloc/find bucket: key != NULL, ippkt != NULL
*/
static frag_bucket FAR *__bucket_op( const frag_key *key, 
                                     const in_Header FAR *ippkt,
                                     uint16_t pktid, const void FAR *srchwaddr, 
                                     int debug_lev)
{
  static frag_bucket FAR *bucket_head = NULL;
  frag_bucket FAR *spare, FAR *tail, FAR *bkt = bucket_head;
  unsigned long now = timer_sec_count();
  unsigned int num_buckets;


#if defined(TRACE_BKTOPS)
#undef TRACE_BKTOPS
#define TRACE_BKTOPS(x) DBGPRINTF(x)
#else
#define TRACE_BKTOPS(x) NO_OPERATION()
#endif

  /* ----------------------------------------------- */

  if (!key && !ippkt) /* timeout check (called once per tick via perdhook) */
  {
    __check_and_do_timeout( bucket_head, 1 /* follow */, now );
    return (frag_bucket FAR *)0;
  }

  /* ----------------------------------------------- */

  if (key && !ippkt) /* release request by key */
  {
    while (bkt)
    {
      if (FAR_MEMCMP(key, &bkt->key, sizeof(*key)) == 0)
      {
        bkt->key.in_use = 0;
        if (bkt->debug_lev)
          TRACE_BKTOPS(("free_fragment(%p) => bucket=%p\n", ippkt, bkt ));
        break;
      }
      bkt = bkt->next;
    }
    return bkt; /* NULL or released bkt */
  }

  /* ----------------------------------------------- */

  if (!key && ippkt) /* release request by ippkt */
  {
    while (bkt)
    {
      if (bkt->expires == ((uint32_t)-1) && /* work completed */
          ippkt == ((const in_Header FAR *)&(bkt->buf[0])) )
      {
        bkt->key.in_use = 0;
        if (bkt->debug_lev)
          TRACE_BKTOPS(("free_fragment(%p) => bucket=%p\n", ippkt, bkt ));
        break;
      }
      bkt = bkt->next;
    }
    return bkt; /* NULL or released bkt */
  } /* if (!key) */

  /* ----------------------------------------------- */

  if (debug_lev)
    TRACE_BKTOPS(("*IP-%04x begin bucket search\n", pktid ));

  num_buckets  = 0;
  spare = tail = (frag_bucket FAR *)0;
  while (bkt)
  {
    num_buckets++;
    if (!bkt->key.in_use)
    {
      if (!spare)
        spare = bkt;
    }
    else if (FAR_MEMCMP(&(bkt->key), key, sizeof(*key))==0)
    {
      if (bkt->expires == ((uint32_t)-1))
      {
        if (debug_lev)
          DBGPRINTF(("*IP-%04x defrag: reasm already completed!\n", pktid ));
        return (frag_bucket FAR *)0;
      }
      break;
    }
    else if (__check_and_do_timeout(bkt, 0, now)) /* timed out */
    {
      if (!spare)
        spare = bkt;
    }
    tail = bkt;
    bkt = bkt->next;
  }
  
  if (!bkt)
    bkt = spare;

  if (debug_lev)
    TRACE_BKTOPS(("*IP-%04x end bucket search =>%p\n", pktid, bkt ));
   
  if (!bkt)
  {
    if (num_buckets == MAX_BUCKETS)
    {
      if (debug_lev)
        DBGPRINTF(("*IP-%04x defrag: Can't requeue. "
             "Already have %u buckets open\n", pktid, num_buckets));
    }      
    else
    {
      bkt = (frag_bucket FAR *)FAR_MALLOC(sizeof(*bkt));
      if (!bkt)
      {
        if (debug_lev)
          DBGPRINTF(("*IP-%04x defrag: Unable to alloc bkt (size=%u)\n",
                      pktid, sizeof(*bkt)));
      }
      else
      {
        FAR_MEMSET(bkt, 0, sizeof(*bkt));
        #if (UINT_MAX != USHRT_MAX) || (IPPACKET_MAX <= (0xffffu - 128u))
        bkt->buf = (uint8_t FAR *)FAR_MALLOC(IPPACKET_MAX);
        #else
        {
          #if defined(__TURBOC__)
          void far *hp = farmalloc( (((unsigned long)IPPACKET_MAX)+31ul) );
          #elif defined(__WATCOMC__)
          void huge *hp = halloc((((unsigned long)IPPACKET_MAX)+31ul)>>4, 16);
          if (!hp)
          {
            _heapshrink();
            _fheapshrink();
            hp = halloc( (((unsigned long)IPPACKET_MAX)+31ul)>>4, 16);
          }
          #endif
          bkt->buf = (uint8_t FAR *)0;
          if (hp)
          {
            unsigned short seg = FP_SEG(hp);
            unsigned short off = FP_OFF(hp);
            if (off != 0)
            {
              seg += (off >> 4);
              if (off & 0x0f)
                seg++;
              off = 0;
            }
            bkt->buf = (uint8_t FAR *)MK_FP(seg, off);
          }
          if (debug_lev)
            DBGPRINTF(("*IP-%04x defrag: halloc(%u) => %04x:%04x, "
              "bkt->buf => %04x:%04x\n", pktid, IPPACKET_MAX, 
              FP_SEG(hp), FP_OFF(hp), FP_SEG(bkt->buf), FP_OFF(bkt->buf) ));
        }
        #endif
        if (!bkt->buf)
        {
          FAR_FREE(bkt);
          bkt = (frag_bucket FAR *)0;
          if (debug_lev)
            DBGPRINTF(("*IP-%04x defrag: Unable to alloc bkt->buf (size=%u)\n",
                      pktid, IPPACKET_MAX));
        }  
        else 
        {
          if (debug_lev)
            TRACE_BKTOPS(("*IP-%04x defrag: end bucket alloc =>%lx\n", pktid, bkt ));
          if (tail)
            tail->next = bkt;
          else
          {
            bucket_head = bkt;
            if (blocker_perdhook_add(__check_timeouts) != 0)
            {
              if (debug_lev)
                DBGPRINTF(("*IP defrag: blocker_perdhook_add(__check_timeouts) failed\n"));
              /* non-fatal - timeouts will be checked when new pkt come in */
            }
          }
        } /* bkt->buf allocation */
      } /* bkt allocation */
    } /* MAX_BUCKETS check */
  } /* if (!bkt) */

  if (!bkt)
  {
    __badpkt_history_op('+', key); /* add to bad packet history */
    icmp_retour( SOICMPERR_SOURCEQUENCH, ippkt, srchwaddr, debug_lev );
    return bkt; /* NULL */
  }

  if (!bkt->key.in_use)
  { 
    FAR_MEMSET( &(bkt->map[0]), 0, sizeof(bkt->map) );
    bkt->set_count = 0;
    bkt->total_len = 0;
    FAR_MEMSET( &(bkt->in_hdr_buf[0]), 0, sizeof(bkt->in_hdr_buf) );
    FAR_MEMCPY( &(bkt->key), key, sizeof(*key) ); /* sets key.in_use */
  }
  else if (((in_Header FAR *)&(bkt->in_hdr_buf[0]))->tos != ippkt->tos)
  {
    if (debug_lev)
      DBGPRINTF(("*IP-%04x defrag: TOS changed!\n", pktid )); 
    /* otherwise ignore the change */
  }

  FAR_MEMCPY( &(bkt->mru_hwaddr[0]), srchwaddr, sizeof(bkt->mru_hwaddr));
  bkt->expires = now + 1 + (ippkt->ttl > 15 ? 15 : ippkt->ttl);
  bkt->debug_lev = debug_lev;

  pktid = pktid; /* possibly unused */
  return bkt;
}


/* periodic handler registered with blocker_perdhook_add(). */
static void __check_timeouts(void)
{
  __bucket_op( (const frag_key *)0, (const in_Header FAR *)0, 0, 
                                    (const void FAR *)0, 0 );
}
#endif

/*----------------------------------------------------------------------*/

/*
 *  Free/release (what might be) a reassembled IP-packet.
*/
int _ip_frag_free(const in_Header FAR *ippkt)
{
  if (ippkt) /* consume var */
  {
#if defined(WANT_IPFRAG_SUPPORT)
    if (__bucket_op( ((const frag_key *)0), ippkt, 0, (const void FAR *)0, 0 ))
      return 1;
#endif
  }
  return 0;
}

/*----------------------------------------------------------------------*/

#if defined(WANT_IPFRAG_SUPPORT)
static int __check_pkt(const frag_key *key, const in_Header FAR *ippkt, uint16_t pktid,
            const void FAR *srchwaddr, uint16_t flags, uint16_t offset, int debug_lev)
{
  uint16_t  frag_data_len = (uint16_t)(NTOHS(ippkt->length) - in_GetHdrLen(ippkt));
  uint32_t offs_plus_len = ((uint32_t)offset) + ((uint32_t)frag_data_len);
  int rc = +1; /* >0=ok, <=0=fail : 0=don't add to history, -1=add to hist */

  if (debug_lev)
    DBGPRINTF(("*IP-%04x defrag newfrag offs=%u, MF?=%d, data_len=%u\n",
                    pktid, offset, (flags & IP_MF)!=0, frag_data_len ));

  if (__badpkt_history_op('?', key))
  {
    if (debug_lev)
      DBGPRINTF(("*IP-%04x defrag: DROP (on bad packet history)\n", pktid ));
    rc = 0;
  }
  else if (frag_data_len == 0 && (flags & IP_MF) && offset != 0)
  { 
    /* only first and last frag need to be handled if datalen == 0 */
    if (debug_lev)
      DBGPRINTF(("*IP-%04x IN: defrag: ignored (no data)\n", pktid ));
    rc = 0; /* no data, ignore it */
  }
  else if ((flags & IP_MF) && (frag_data_len & 7) != 0)
  {
    /* the next offset will not be evenly divisible by 8 */
    if (debug_lev)
      DBGPRINTF(("*IP-%04x IN: defrag: DROP (bad data length %u (0x%x))\n", pktid,
                                            frag_data_len, frag_data_len ));
    rc = -1; /* fallthrough, fail, and add to history */
  }
  else if (offs_plus_len >= (0xffffu - sizeof(in_Header)) )
  {
    if (debug_lev)
      DBGPRINTF(("*IP-%04x IN: defrag: DROP (off + len >= 0xffeb)\n", pktid ));
    rc = -1; /* fallthrough, fail, and add to history */
  }
  else if (offs_plus_len >= (IPPACKET_MAX-sizeof(in_Header)))
  {
    /* the actual size of the header is not known yet, so this
     * test assumes a minimal 20 byte header. If during final
     * assembly __insert_frag() finds that the hdrlen+datalen
     * exceeds IPPACKET_MAX, it will drop the packet. 
    */
    if (debug_lev)
      DBGPRINTF(("*IP-%04x IN: defrag: SQUENCH (%u + %u) >= %u\n", pktid,
          offset, frag_data_len, (IPPACKET_MAX-sizeof(in_Header)) ));
    icmp_retour( SOICMPERR_SOURCEQUENCH, ippkt, srchwaddr, debug_lev );
    rc = -1; /* fallthrough, fail, and add to history */
  }

  if (rc < 0)
  {
    /* drop any existing packets matching 'key' */
    __bucket_op( key, (const in_Header FAR *)0, pktid, srchwaddr, debug_lev);
    /* add to bad packet history */
    __badpkt_history_op('+', key); 
  }

  return (rc > 0);
}
#endif

/*----------------------------------------------------------------------*/

/* Push the packet through the defragger. 
** If defragmentation isn't required ...
**       it returns 'ippkt' as is.
** Else ...
**       If reassembly is now complete ...
**         it returns a new 'ippkt'. 
**       Else if error 
**         it returns NULL
**       Else (reassembly is not complete but no error)
**         it returns NULL - packet will be completed later
**
** srcethaddr is only needed in case we have to send an ICMP_TIMXCEED
** message somewhere down the line.
*/
const in_Header FAR * _ip_frag_handler(const in_Header FAR *ippkt,
                                       const void FAR *srchwaddr, 
                                       int debug_lev )
{
  uint16_t  pktid, offset, flags;

  pktid  = NTOHS(ippkt->identification);
  offset = NTOHS(ippkt->frag_ofs);
  flags  = (uint16_t)(offset & ~IP_OFFMASK);
  offset = (uint16_t)((offset & IP_OFFMASK) << 3);  /* 0 <= ip_ofs <= 65536-8 */

#ifdef FORCE_DEBUG
  if (debug_lev < 1)
    debug_lev = 1;
#endif

  /* if offset is 0 and !IP_MF then the packet is not fragmented.
   * Fallthrough and return ippkt as is
  */
  if (!(offset == 0 /* first fragment */ &&
       (flags & IP_MF) == 0 /* last fragment */))
  {
#if !defined(WANT_IPFRAG_SUPPORT)

    /* No defragmentation support. Treat as error. 
    */
    if (debug_lev)
      DBGPRINTF(("*IP-%04x IN: defrag: DROP (reasm not supported)\n", pktid ));
    icmp_retour( SOICMPERR_SOURCEQUENCH, ippkt, srchwaddr, debug_lev );

    ippkt = (const in_Header FAR *)0;
    /* fallthrough and return ippkt == NULL */

#else

    /* IP-packet is part of a fragment chain. Match with previous
    * fragments and return reassembled IP-packet.
    *
    * If defragmentation is successful, return the reassembled packet. 
    * On error or reassembly failure return NULL.
    */
    frag_key key;

    key.in_use   = 1;
    key.proto    = ippkt->proto;
    key.ident    = ippkt->identification;
    key.src_addr = ippkt->source; 
    key.dst_addr = ippkt->destination;

    if (!__check_pkt(&key, ippkt, pktid, srchwaddr, flags, offset, debug_lev))
    {
      /* bad offset or length, drop it */
      ippkt = (const in_Header FAR *)0; 
      /* message already printed */
    }
    else
    {
      frag_bucket FAR *bkt = __bucket_op( &key, ippkt, 
                                          pktid, srchwaddr,debug_lev );
      if (bkt)
        ippkt = __insert_frag( bkt, ippkt, pktid, flags, offset, debug_lev );
      else /* debug message printed, and maybe source quench sent */
      {
        ippkt = (const in_Header FAR *)0; /* alloc failed */
      }
    }
#endif
  }

  pktid = pktid; /* possibly unused */
  return ippkt;
}

/* ======================================================================= */

