/*
 * Link-layer (OSI layer 2) interface service stub provider.
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: linklayr.c,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
#define __OSILAYER__ 2 /* Data Link Layer  */

#include "target.h"   /* standard includes */
#include "ctordtor.h" /* _my_atterm() */
#include "inaddr.h"   /* is_local_addr() [for SLIP loopback probing] */
#include "debug.h"    /* KMSG(), DBGPRINTF() */
#include "printk.h"   /* printk() */
#include "arp.h"      /* _rarp_resolve() */
#include "blocker.h"  /* blocker_* */
#include "loadcfg.h"  /* _conf_add_confextn(), _conf_load() */
#include "pktdrvr.h"  /* pdapi_*() */ 
#include "linklayr.h" /* ourselves */

#define MAX_HWPKTHDR_SIZE 75u /* the largest possible size of a pkt header */
/* SLIP/PPP = 0, DIX = 14, IEEE802.3 = <= 22, IEEE802.5= <= 40, X.25 <=72 */

#define IEEE8025_SUPPORT_CHECKED_OK /*TR support was only tested with loopback*/

/* ======================================================================== */
/*                            state data                                    */
/* ======================================================================== */

static struct
{     
  int pdintr;                      /* non-zero only when up-and-running */
  int shutting_down;               /* shutdown in progress */
  int debug_lev;                   /* guess */
  int conf_pdintr;                 /* interrupt # from .cfg */

  unsigned short handle_tab[3];    /* packet driver handle slots */
  unsigned short FAR *ip_handleP;  /* pointer to slot (NULL if not valid) */
  unsigned short FAR *arp_handleP; /* pointer to slot (NULL if not valid) */
  unsigned short FAR *rarp_handleP;/* pointer to slot (NULL if not valid) */

  uint8_t FAR *send_buf;   /* of size osi3_tx_mtu + MAX_HWPKTHDR_SIZE */
  uint16_t osi3_tx_mtu;    /* OSI layer 3 (network layer, ie IP) MTU.
                            * this is only needed for transmission, received
                            * packets can be of any size 
                           */

  uint8_t my_hwaddr[MAX_LINKADDR_SIZE];   /* local link-layer source addr*/
  uint8_t my_bcaddr[MAX_LINKADDR_SIZE];   /* Link-layer broadcast addr   */
  int  mcast_curr_mode; 
  uint8_t hw_type;        /* see comments at _eth_get_hwtype()*/
  uint8_t hw_addrlen;
  int  hw_mcast_max_groups;
  int  pd_is_nullpkt;
  int  pd_is_p2p;
  int  pd_class;

} FAR my_statics; /* all zero */

#define CALLTYPE_INIT   0
#define CALLTYPE_ATEXIT 1
#define CALLTYPE_ATTERM 2

static int __init_release(int call_type); /* forward reference */

/* ======================================================================== */
/*                         Config worthlessness                             */
/* ======================================================================== */

/*
** convert a string in the form 0x?? or ??h or ?? to an interrupt vector,
** and ensure that the vector is in the range 0x60...0x80 (inclusive).
*/
static int __str_to_vector(const char *p) /* p may be null (from getenv()) */
{
  int vec = 0;
  if (p)
  {
    int i, isx = 0;
    if (*p == '0')
      p++;
    if (*p == 'x' || *p == 'X')
      isx = *p++;
    else
    {
      i = 0;
      while ((*p>='0' && *p<='9') || 
             (*p>='a' && *p<='f') || 
             (*p>='A' && *p<='F'))
      { 
        i++;
        p++;
      }
      if (*p == 'h' || *p == 'H')
        isx = 'H';
      p-=i;
    }
    while ((*p>='0' && *p<='9') || 
           (*p>='a' && *p<='f') || 
           (*p>='A' && *p<='F'))
    {
      i = 0;
      if (*p >= '0' && *p <= '9')
        i = *p++ - '0';
      else if (!isx)
        return 0;
      else if (*p >= 'a' && *p <= 'f')
        i = (*p++ - 'a')+10;
      else if (*p >= 'A' && *p <= 'F')
        i = (*p++ - 'A')+10;
      if (isx)
        vec *= 16;
      else
        vec *= 10;
      vec += i;
      if (vec > 255)
        return 0;
    } 
    if (isx == 'H')
      p++;
    if (*p && *p != ' ' && *p != '\t' && *p != ',' && *p != ';')
      vec = 0;
    else if (vec < 0x60 || vec > 0x80)
      vec = 0;  /* discard illegal value */
  }
  return vec;
}

/* -------------------------------------------------------------------- */

static void __set_pdintr (const char *value) /* set interrupt # */
{
  int intr = __str_to_vector(value);
  if (intr != 0)
    my_statics.conf_pdintr = intr;
}

/* -------------------------------------------------------------------- */

static void __ctor_eth(void)
{
  static const struct config_table FAR cfg[] = {
     { "PKT_INTR",      ARG_FUNC,     (void FAR *)__set_pdintr },
     { "LINK.DEBUG",    ARG_ATOI,     (void FAR *)&my_statics.debug_lev },
     { ((const char FAR *)0), 0, ((void FAR *)0) }
  };
  _conf_add_confextn( &cfg[0], NULL );
}
#define __SETUP_CONSTRUCTOR_FOR__ __ctor_eth
#include "ctordtor.h"

/* =================================================================== */
/*                           encapsulation                             */
/* =================================================================== */

/* the following structures are all 'worst-case', ie have the 
** largest possible size and all possible fields.
*/
#pragma pack(1)  /* pack struct members on byte boundaries */
typedef struct bluebook_hdr /* bluebook header is fixed size */
{
  unsigned char destination[6];
  unsigned char source[6];
  unsigned short type;
} bluebook_hdr;
#pragma pack()

#pragma pack(1)
typedef struct ieee8025_hdr
{
  struct {
    unsigned priority:3; /* Bit 0,1 and 2: 0=lowest, 7=highest, default=3 */
    unsigned token:1;    /* Bit 3:         0=token, 1=frame */  
    unsigned monitor:1;  /* Bit 4:         next host on ring sets this */ 
    unsigned reservation:3; /* Bit 5,6,7:  000=low, 111=high */
  } access_control;
  struct {
    unsigned frame_type:2; /* Bit 0,1: 00 = MAC Frame, 01 = LLC Frame. */
    unsigned control:6;    /* Bits 2-7: 0 = Normal Bufferd, 3 = Claim Token*/
  } frame_control;
  uint8_t dest_source[6*2]; /* (6 or 2 octets) * 2 */
  /* source follows immediately after destination */
  /* (source[0] & 1) == RIF is present */
  /* (source[0] & 2) == U[0]/L[1] bit (Univers/Locally administered)*/
} ieee8025_hdr;
#pragma pack()

#pragma pack(1)
typedef struct ieee8025_rif /* (only present if have_rif bit is set) */
{                           /* 2-18 octets */
  struct {
    unsigned routing_type:3; /*0xx=source,10x=allroute bcast,11x=limitedbcast*/
    unsigned rif_length:5;   /* total length of RIF in octets */
    unsigned direction:1;    /* 0=forward,1=reverse */
    unsigned max_frame_size:3; /*0=516,1=1500,2=2042,3=4472,4=8144,5=11407,...*/
    unsigned reserved:3;     /* all zero */
  } routing_control;
  uint16_t routing_designators[8]; /* 0-8 (one for every bridge traversed) */
} ieee8025_rif;
#pragma pack()

#pragma pack(1)
typedef struct ieee802x_llc
{
  uint8_t dsap; /* 0xAA (170 dec) - the assigned global SAP value for SNAP */
             /* or 0x06 (IP if no snap) or 0xE0 (IPX if no snap) */
  uint8_t ssap; /* 170 (decimal) - the assigned global SAP value for SNAP */
             /* or 0x06 (IP if no snap) or 0xE0 (IPX if no snap) */
  uint8_t control;/* 3 (decimal) - "Unnumbered Information" if have SNAP */
             /* else 0xxxxxxx = transmitter seq # */ 
             /* else 1xxxxxxx = supervisory info */
} ieee802x_llc;
#pragma pack()

#pragma pack(1)
typedef struct ieee80x_snap
{
  uint8_t proto_id_or_org_code[3]; /*all zero to ignore */
  uint16_t ether_type; /* in bit transmission order (ie:little-endian) */
} ieee802x_snap;
#pragma pack()

/* ==================================================================== */
/*        convert an network-layer (eg IP) packet to media format       */
/* ==================================================================== */

static const void FAR *__encapsulate_pkt( const void FAR *ep, 
                                          unsigned int *ep_len,
                                          unsigned short ethtype, 
                                          const void FAR *dest_hw_addr, 
                                          int *is_loopback )
{
  switch (my_statics.pd_class)
  {
    case PDCLASS_DIX:
    {
      bluebook_hdr FAR *eth = (bluebook_hdr FAR *)0;
      if (*ep_len <= _eth_get_mtu() /* net layer mtu */)
      {
        eth = (bluebook_hdr FAR *)my_statics.send_buf;

        /* build the DIX packet */
        FAR_MEMCPY( &(eth->destination[0]), dest_hw_addr, 6);
        FAR_MEMCPY( &(eth->source[0]), _eth_get_hwaddr(), 6);
        eth->type = HTONS(ethtype);
        FAR_MEMCPY( eth + 1, ep, *ep_len ); 

        *ep_len += sizeof(bluebook_hdr);
        if (*ep_len < 60) 
          *ep_len = 60;   /* DIX packets must be at least 60 octets long */

        /* _arp_resolve() returns our own ether-addr for is_local_addr() */
        *is_loopback = (FAR_MEMCMP(dest_hw_addr,&(eth->source[0]),6)==0);
      }
      return (void FAR *)eth;
    }
    case PDCLASS_SLIP:
    {
      if (ethtype != ETHERTYPE_IP)
        ep = (const void FAR *)0;
      else
      {
        uint32_t ina = *((const uint32_t FAR *)(((const uint8_t FAR *)ep)+16));
        *is_loopback = is_local_addr(NTOHL(ina));
      }
      return ep;
    }
#if defined(IEEE8025_SUPPORT_CHECKED_OK)
    case PDCLASS_IEEE802_5:
    {     
      if (*ep_len <= _eth_get_mtu()) /* net layer MTU */
      {
        int hwaddrlen = _eth_get_hwaddrlen();
        ieee8025_hdr FAR *hdr;
        uint8_t FAR *payload;

        payload = (uint8_t FAR *)my_statics.send_buf;
        hdr = (ieee8025_hdr FAR *)payload;
        payload += (sizeof(ieee8025_hdr)-(sizeof(hdr->dest_source)))+(hwaddrlen*2);
      
        hdr->access_control.priority = 3;
        hdr->access_control.token = 1; /* its a frame */
        FAR_MEMCPY( &hdr->dest_source[0], dest_hw_addr, hwaddrlen);
        FAR_MEMCPY( &hdr->dest_source[hwaddrlen], _eth_get_hwaddr(), hwaddrlen);
        /* source is immediately after destination */

        /* _arp_resolve() returns our own ether-addr for is_local_addr() */
        *is_loopback = (FAR_MEMCMP(dest_hw_addr,_eth_get_hwaddr(),hwaddrlen)==0);
      
        if (ethtype == ETHERTYPE_ARP || ethtype == ETHERTYPE_REVARP ||
            (*((uint8_t FAR *)&hdr->dest_source[hwaddrlen]) & 1) != 0) /* broadcast */
        { 
          ieee8025_rif FAR *rif = (ieee8025_rif FAR *)payload;
          payload += sizeof(rif->routing_control);

          /* leave the bit set - it now means 'have_rif' */
          *((uint8_t FAR *)&hdr->dest_source[hwaddrlen]) |= 1;
     
          rif->routing_control.rif_length = sizeof(rif->routing_control);
          rif->routing_control.routing_type = 0x06; /* limited bcast */

          if (ethtype == ETHERTYPE_ARP || ethtype == ETHERTYPE_REVARP)
          {
            rif->routing_control.routing_type = 0x04; /* all ring bcast */
            /* HACK! we need to do all-ring right away for ARP/RARP because 
            ** _arp_resolve() doesn't know anything about rings. Oh well.
            */
          }
        }

        {
          ieee802x_llc FAR *llc = (ieee802x_llc FAR *)payload;
          payload += sizeof(*llc);
        
          llc->dsap = 0xAA;
          llc->ssap = 0xAA;
          llc->control = 0x03;
        }

        {
          ieee802x_snap FAR *snap = (ieee802x_snap FAR *)payload;
          payload += sizeof(*snap);

          snap->ether_type = ethtype; /* stored in bit order */
        }

        *ep_len += (size_t)(payload - ((uint8_t FAR *)my_statics.send_buf));
        FAR_MEMMOVE( payload, ep, *ep_len );
        return my_statics.send_buf;
      }
      break;
    }
#endif /* IEEE8025_SUPPORT_CHECKED_OK */
    default:
    {
      *ep_len = 0;
      break;
    }
  }

  *ep_len = 0;
  return (void FAR *)0;
}

/* ==================================================================== */
/*        convert an media-layer packet to network-layer (eg IP)        */
/* ==================================================================== */

static const void FAR *__decapsulate_pkt( const void FAR *ep, 
                                          unsigned int *ep_len,
                                          unsigned short *ethtype,
                                          void FAR *src_hw_addr,
                                          int *is_bcast )
{
  switch (my_statics.pd_class)
  {
    case PDCLASS_DIX:
    {
      const bluebook_hdr FAR *eth = (const bluebook_hdr FAR *)0;
      if (*ep_len >= sizeof(*eth))
      {
        eth = (const bluebook_hdr FAR *)ep;
        FAR_MEMCPY( src_hw_addr, &(eth->source[0]), 6);
        *is_bcast = (eth->destination[0] & 1);
        *ethtype = NTOHS(eth->type);

        *ep_len -= sizeof(eth);
        eth++;
      }
      return eth;
    }
    case PDCLASS_SLIP:
    {
      *ethtype = ETHERTYPE_IP;
      *is_bcast = 0;
      return ep;
    }
#if defined(IEEE8025_SUPPORT_CHECKED_OK)
    case PDCLASS_IEEE802_5: /* IEEE 802.5 */
    {
      const ieee8025_hdr  FAR *hdr;
      const ieee8025_rif  FAR *rif;
      const ieee802x_llc  FAR *llc;
      const ieee802x_snap FAR *snap;
      const uint8_t FAR *payload;
      int hwaddrlen = _eth_get_hwaddrlen();

      struct bluebook_hdr ethhdr;
      int have_rif = 0;

      payload = (const uint8_t FAR *)ep;
      hdr = (const ieee8025_hdr FAR *)ep;
      payload += (sizeof(ieee8025_hdr)-(sizeof(hdr->dest_source)))+(hwaddrlen*2);

      ethhdr.type = 0x0000;
      FAR_MEMCPY( &ethhdr.destination[0], &hdr->dest_source[0], hwaddrlen);
      FAR_MEMCPY( &ethhdr.source[0], &hdr->dest_source[hwaddrlen], hwaddrlen);
      /* source address is immediately after destination */
      have_rif = (ethhdr.source[0] & 1);
      ethhdr.source[0] &= 0xfe; /* assume non-broadcast */

      if (hdr->access_control.token == 1) /* should always be true */
      {
        if (have_rif) /* have RIF */
        {
          rif = (const ieee8025_rif FAR *)payload;
          payload += rif->routing_control.rif_length; /* variable size */

          if ((rif->routing_control.routing_type & 0x04) != 0)
          {
            /* limited or all-ring bcast */
            ethhdr.source[0] |= 1; /* make ether bcast */
          }
        }

        llc = (const ieee802x_llc FAR *)payload;
        payload += sizeof(*llc); /* fixed size */
      
        if (llc->dsap == 0xAA && llc->ssap == 0xAA && llc->control == 0x03) 
        {                                                    /* have SNAP */
          snap = (const ieee802x_snap FAR *)payload;
          payload += sizeof(*snap); /* fixed size */
          /* snap ether_type is in bit order */
          *(((uint8_t *)&ethhdr.type)+0) = (uint8_t)(snap->ether_type>>8);
          *(((uint8_t *)&ethhdr.type)+1) = (uint8_t)(snap->ether_type&0xff);
        }
        else 
        {  
          if (llc->dsap == 0x06)    
            ethhdr.type = HTONS(ETHERTYPE_IP);          
          else if (llc->dsap == 0xE0)
            ethhdr.type = HTONS(0x8137); /* DIX version of IPX */
          /* Appletalk       = 0x809B, */
          /* Loopback        = 0x9000 */
        }

        FAR_MEMCPY( src_hw_addr, &(ethhdr.source[0]), hwaddrlen);
        *is_bcast = (ethhdr.destination[0] & 1);
        *ethtype = NTOHS(ethhdr.type);

        *ep_len -= (size_t)(payload - ((const uint8_t FAR *)ep)); 
        return (const void FAR *)payload;

      } /* is it a frame? */ 
      break;

    } /* case RFC826_HWTYPE_IEEE802 */
#endif /* IEEE8025_SUPPORT_CHECKED_OK */
    default:
    {
      ep_len = ep_len;
      break;
    }
  } /* switch() */

  return (void FAR *)0;
}  

/* ===================================================================== */
/*                          Debug helpers                                */
/* ===================================================================== */

/*
 * Return textual error representing error-code.
 */
#if !defined(NDEBUG)
static const char *pd_strerror(int code)
{
  static const char *err[] = {
              "error 0",
              "Invalid handle number",
              "No interfaces of specified class found",
              "No interfaces of specified type found",
              "No interfaces of specified number found",
              "Bad packet type specified",
              "This interface does not support multicast",
              "This packet driver cannot terminate",
              "An invalid receiver mode was specified",
              "Operation failed because of insufficient space",
              "Type previously accessed, and not released",
              "The command was out of range, or not implemented",
              "Cannot send packet (hardware error)",
              "Cannot change hardware address",
              "Hardware address has bad length/format",
              "Could not reset interface",
              "Extended driver needed"
            };
  if (code == 0xff)
  {
    return "Bad vector";
  }
  if (code < 0 || (((unsigned int)code) >= sizeof(err)/sizeof(err[0])) )
  {
    static char errany[] = "Unknown driver error (0xNN)";
    char *p = &errany[sizeof(errany)-3];
    p[0] = (char)(code   >> 4);
    p[0] += (char)('0'+((p[0] > 9) ? ('a'-('9'+1)) : 0));
    p[1] = (char)(code & 0x0f); 
    p[1] += (char)('0'+((p[1] > 9) ? ('a'-('9'+1)) : 0));
    return (const char *)errany;
  }
  return (err[code]);
}
#endif

/* ==================================================================== */
/*                            pkt receiver                              */
/* ==================================================================== */

static struct 
{
  uint16_t ethtype; /* ETHERTYPE_* in host order */
  void (*handler)( const void FAR *netpkt, unsigned int netpkt_len,
                   const void *src_hw_addr, int is_brdcast );
                   
} pkthandlers[4] = 
{
  { 0, NULL },
  { 0, NULL },
  { 0, NULL },
  { 0, NULL }
};

/* -------------------------------------------------------------------- */

/* Packet driver upcall */
static int pkt_enqueue( int intno, unsigned short handle, 
                        unsigned int rxLen, const void FAR *rxBuf)
{
  uint16_t ethtype;
  unsigned int i;

  intno = intno; /* unused - we only service one packet driver */

  if (my_statics.pdintr == 0 || my_statics.shutting_down) /* oops! */
    return 1; /* pretend we queued it so that it can be purged */

  if (rxLen > (_eth_get_mtu()+MAX_HWPKTHDR_SIZE) )
    return 1; /* pretend we queued it so that it can be purged */

  ethtype = 0;
  if (my_statics.ip_handleP        && handle == *my_statics.ip_handleP)
    ethtype = ETHERTYPE_IP;
  else if (my_statics.arp_handleP  && handle == *my_statics.arp_handleP)
    ethtype = ETHERTYPE_ARP;
  else if (my_statics.rarp_handleP && handle == *my_statics.rarp_handleP)
    ethtype = ETHERTYPE_REVARP;
  else  /* unknown handle */
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("unknown handle 0x%04x\n", handle ));
    return 1; /* pretend we queued it so that it can be purged */
  }

  if (my_statics.debug_lev >= 2)
    DBGPRINTF(("*ETH: Upcall ... (type=0x%04x, len=%u)\n", ethtype, rxLen ));

  for (i = 0; i < (sizeof(pkthandlers)/sizeof(pkthandlers[0])); i++)
  {
    if (pkthandlers[i].ethtype == 0)
      break;
    if (pkthandlers[i].ethtype == ethtype)
    {
      unsigned short netpkt_ethtype;
      uint8_t srchwaddr[MAX_LINKADDR_SIZE];
      int is_bcast = 0;

      blocker_critsect_begin();
      rxBuf = __decapsulate_pkt( rxBuf, &rxLen, 
                                 &netpkt_ethtype, srchwaddr, &is_bcast );
      if (!rxBuf)
      {
        if (my_statics.debug_lev)
           DBGPRINTF(("Decapsulation failed\n" ));
      }
      else if (netpkt_ethtype != ethtype)
      {
        if (my_statics.debug_lev)
           DBGPRINTF(("Unexpected packet of type 0x%04x (expected 0x%04x)\n",
                      netpkt_ethtype, ethtype ));
      }
      else
      { 
        (*(pkthandlers[i].handler))( rxBuf, rxLen, srchwaddr, is_bcast );
      }
      blocker_critsect_end();

      return 1; /* handled - can be purged now */
    }
  }

  if (my_statics.debug_lev)
    DBGPRINTF(("No packet handler for ethtype=0x%04x!\n", ethtype ));

  return 1; /* pretend it was handled so that it can be purged */
}  

/* -------------------------------------------------------------------- */

int _eth_register_pkthandler( uint16_t ethtype, 
                              void (*handler)( const void FAR *netpkt,
                              unsigned int netpkt_len,
                              const void *src_hw_addr,
                              int is_brdcast ) )
{
  unsigned int i;
  for (i = 0; i < (sizeof(pkthandlers)/sizeof(pkthandlers[0])); i++)
  {
    if (pkthandlers[i].ethtype == 0)
      break;
    if (pkthandlers[i].ethtype == ethtype)
      return 1;
  }
  for (i = 0; i < (sizeof(pkthandlers)/sizeof(pkthandlers[0])); i++)
  {
    if (pkthandlers[i].ethtype == 0)
    {
      pkthandlers[i].ethtype = ethtype;
      pkthandlers[i].handler = handler;
      return 1;
    }
  }   
  return 0;
}  

/* -------------------------------------------------------------------- */

/*
 * _eth_poll() - poll the receiver queues.
 *
 * MASSIVE HACK ALERT: _eth_poll SHOULD be static, but isn't because 
 * it needs to called directly to support a massive hack in ipboot.c. 
 * See comments there for more information.
*/
void _eth_poll(void)
{
  if (my_statics.pdintr)
  {
    /* _eth_poll() could be called from both an ISR and from
     * the idle loop, so make sure that a firing ISR doesn't cause 
     * us to run a poll when we're already running one.
    */
    static fastlock_t isr_guard = FASTLOCK_INITIALIZER_UNLOCKED;

    if (fastlock_trylock(&isr_guard) > 0)
    {
      int got_some = 0;
      while (pdapi_poll_receiver(-1 /* any */))
        got_some = 1;
      if (got_some)
        blocker_inc_hwork();
      fastlock_unlock(&isr_guard);
    }
  }
  return;
}     

/* ===================================================================== */
/*                           send a packet                               */
/* ===================================================================== */

/*
 *  _eth_send() does the actual transmission.
 *  Do any last minute patches here, like fix the size. 
 *
 *  Returns (original) 'pkt_len' on success, 0 on error.
 */
int _eth_send ( uint16_t ethtype, /* ETHERTYPE_* in host order */
                const void FAR *dest_hw_addr,
                const void FAR *pkt, unsigned int pkt_len)
{
  unsigned short FAR *lb_handleP = (unsigned short FAR *)0;
  unsigned int added_len = 0;

  if (!pkt)
  {
    pkt_len = 0;
  }
  else if (__init_release(CALLTYPE_INIT) != 0)
  {
    pkt_len = 0;
  }
  else if (pkt_len)
  {
    int dest_is_local_addr = 0; 

    added_len = pkt_len;
    pkt = __encapsulate_pkt( pkt, &pkt_len, ethtype, dest_hw_addr,
                             &dest_is_local_addr );
    if (!pkt)
    {
      if (my_statics.debug_lev)
        DBGPRINTF(("*ETH: __encap failed\n"));
      pkt_len = 0;
    }
    else
    {
      added_len = pkt_len - added_len;

      if (ethtype == ETHERTYPE_IP)
        lb_handleP = my_statics.ip_handleP;
      else if (ethtype == ETHERTYPE_ARP)
        lb_handleP = my_statics.arp_handleP;
      else if (ethtype == ETHERTYPE_REVARP)
        lb_handleP = my_statics.rarp_handleP;
      else
      {
        if (my_statics.debug_lev)
          DBGPRINTF(("*ETH: unhandled proto %04x\n", ethtype ));
        pkt_len = 0;
      }
      if (!dest_is_local_addr)
        lb_handleP = (unsigned short FAR *)0;
    }
  }

  if (pkt_len)
  {
    int pderrno;

    if (my_statics.debug_lev)
      DBGPRINTF(("*ETH: pkt_send(intno=0x%02x, local?=%d, len=%u)\n", 
                     my_statics.pdintr, (lb_handleP != NULL), pkt_len ));

    pderrno = pdapi_send_pkt( my_statics.pdintr, lb_handleP, pkt, pkt_len);

    if (my_statics.debug_lev)
      DBGPRINTF(("*ETH: pkt_send() => %s\n", 
                     (pderrno ? pd_strerror(pderrno) : "OK") ));

    if (pderrno == 0)
      pkt_len -= added_len;
    else
      pkt_len = 0;
  }

  return pkt_len; /* net layer length, 0 if error */
}  

/* ======================================================================= */
/*                        Set the number of recv bufs                      */
/* ======================================================================= */

int _eth_numbufs_set(unsigned int count)
{
  if (__init_release(CALLTYPE_INIT) != 0)
    return -1;
  if (pdapi_set_recvbuf_count(my_statics.pdintr, count) == 0)
    return 0;
  return -1;
}

/* ======================================================================= */
/*                   Jim Martin's Multicast extensions                     */
/* ======================================================================= */

int _eth_mcast_set_mode(int mode)
{
  int newmode, oldmode, pderrno;

  if (__init_release(CALLTYPE_INIT) != 0)
    return -1;

  if (_eth_mcast_get_max_groups() < 1) /* not initialized or not supported */
    return -1;

  newmode = -1;
  if (mode == ETH_MCAST_MODE_OFF)
    newmode = PDMODE_BROADCAST; /* the normal state when not mcast/promiscuous */
  else if (mode == ETH_MCAST_MODE_LIMITED)
    newmode = PDMODE_MCAST_ASLISTED;
  else if (mode == ETH_MCAST_MODE_UNLIMITED)
    newmode = PDMODE_MCAST_ALL;
  else
    return -1;

  pderrno = pdapi_get_recv_mode( my_statics.pdintr, 
                                 *my_statics.ip_handleP, &oldmode );
  if (pderrno != 0)               /* something seems to be broken */
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("Error getting recv mode: %s\n",pd_strerror(pderrno)));
    return -1;
  }

  if (oldmode != newmode)
  {
    pderrno = pdapi_set_recv_mode( my_statics.pdintr, 
                                   *my_statics.ip_handleP, newmode);
    if (pderrno != 0)               /* something seems to be broken */
    {
      if (my_statics.debug_lev)
        DBGPRINTF(("Error setting recv mode: %s\n",pd_strerror(pderrno)));
      return -1;
    }
    my_statics.mcast_curr_mode = newmode;
  }

  return (0);
}

/* -------------------------------------------------------------------- */

int _eth_mcast_set_list(const void *hwaddrlist, unsigned int hwlistlen)
{
  int pderrno;

  if (__init_release(CALLTYPE_INIT) != 0)
    return -1;

  if (_eth_mcast_get_max_groups() < 1) /* not initialized or not supported */
    return (-1);

  if ((hwlistlen % _eth_get_hwaddrlen()) != 0)
    return (-1);  /* bad length */

  pderrno = pdapi_set_multicast_list(my_statics.pdintr, hwaddrlist, hwlistlen);
  if (pderrno != 0)
  {
    if (my_statics.debug_lev)
      DBGPRINTF(("Error setting mcast list: %s\n",pd_strerror(pderrno)));
    return (-1);
  }
  return (0); /* everything ok */
}

/* ===================================================================== */
/*                   Commonly referred to globals                        */
/* ===================================================================== */

const uint8_t FAR *_eth_get_hwaddr(void)
{
  __init_release(CALLTYPE_INIT);
  return &(my_statics.my_hwaddr[0]); /* {0,0,0,0,0,0} if __init_release(CALLTYPE_INIT) fails */
}

/* -------------------------------------------------------------------- */

const uint8_t FAR *_eth_get_hwaddrbc(void)
{
  __init_release(CALLTYPE_INIT);
  return &(my_statics.my_bcaddr[0]); /* {0,0,0,0,0,0} if __init_release(CALLTYPE_INIT) fails */
}

/* -------------------------------------------------------------------- */

uint8_t _eth_get_hwaddrlen(void)
{
  __init_release(CALLTYPE_INIT);
  return my_statics.hw_addrlen; /* 0 if __init_release(CALLTYPE_INIT) fails */
}

/* -------------------------------------------------------------------- */

int _eth_mcast_get_max_groups(void)
{
  __init_release(CALLTYPE_INIT);
  return my_statics.hw_mcast_max_groups; /* 0 if __init_release(CALLTYPE_INIT) fails */
}

/* -------------------------------------------------------------------- */

uint16_t _eth_get_mtu(void)
{
  /* Note that the value we return here is that needed on the network layer
  ** (OSI layer 3, eg IP), and not that returned by the packet driver 
  ** (which includes MAC header).
  */
  __init_release(CALLTYPE_INIT);
  return my_statics.osi3_tx_mtu; /* 0 if __init_release(CALLTYPE_INIT) fails */
}

/* -------------------------------------------------------------------- */

uint8_t _eth_get_hwtype(void) /* => ARP "Assigned Numbers" RFC */
{
  /* See http://www.iana.org/assignments/arp-parameters 
  ** These numbers are used in arp, bootp, dhcp and anywhere else
  ** where a 'hwtype' field is needed.
  **
  1 Ethernet. 
  2 Experimental Ethernet. 
  3 Amateur Radio AX.25. 
  4 Proteon ProNET Token Ring. 
  5 Chaos. 
  6 IEEE 802. 
  7 ARCNET. 
  8 Hyperchannel. 
  9 Lanstar. 
  10 Autonet Short Address. 
  11 LocalTalk. 
  12 LocalNet (IBM PCNet or SYTEK LocalNET). 
  13 Ultra link. 
  14 SMDS. 
  15 Frame Relay. 
  16 ATM, Asynchronous Transmission Mode. 
  17 HDLC. 
  18 Fibre Channel. 
  19 ATM, Asynchronous Transmission Mode. 
  20 Serial Line. 
  21 ATM, Asynchronous Transmission Mode. 
  22 MIL-STD-188-220. 
  23 Metricom. (STRIP: Richochet Starmode Radio)
  24 IEEE 1394.1995. 
  25 MAPOS. 
  26 Twinaxial. 
  27 EUI-64. 
  28 HIPARP. 
  29 IP and ARP over ISO 7816-3. 
  30 ARPSec. 
  */
  __init_release(CALLTYPE_INIT);
  return my_statics.hw_type; /* 0 if _eth_init fails */
}

/* -------------------------------------------------------------------- */

int _eth_get_ifflags(void)
{
  int flags = 0;
  if (__init_release(CALLTYPE_INIT) == 0)
  {
    flags |= ETH_IFFLAGS_UP;
    if (my_statics.pd_is_nullpkt)
      flags |= ETH_IFFLAGS_LBONLY;
    if (my_statics.pd_is_p2p)
      flags |= ETH_IFFLAGS_P2P;
    /* should the mcast test be mcast_curr_mode >= limited multicast */
    else if (_eth_mcast_get_max_groups() > 0)
      flags |= ETH_IFFLAGS_MCAST; 
  }
  return flags;
}

/* ===================================================================== */
/*                               Init/Release                            */
/* ===================================================================== */

#if defined(NDEBUG)
#define __init_release_msg(__ct, __msg) NO_OPERATION()
#else
static void __init_release_msg(int call_type, const char *msg)
{
  if (my_statics.debug_lev >= 2)
  {
    if (call_type == CALLTYPE_ATTERM)
    {
      __runtime_notice("ETH: ");
      __runtime_notice(msg);
      __runtime_notice("\n");
    }
    else
      DBGPRINTF(("*ETH: %s\n", msg ));
  }
}
#endif

/* -------------------------------------------------------------------- */

/*
** __init_pktintr() depends on nothing.
*/
static int __init_release_pktintr(int call_type)
{
  my_statics.pdintr = 0;

  if (call_type == CALLTYPE_INIT)
  {
    my_statics.pdintr = my_statics.conf_pdintr;

    /*
    * Determine vector to use;
    * if env. vars TCP_PKTINT or WHAT?TCP.VEC is specified, use that vector,
    * else search for PKTDRVR between 0x60 - 0x80 as per pktdriver spec.
    */
    if (my_statics.pdintr == 0)  /* "TCP_PKTINT" is for tcplib compatibility */
      my_statics.pdintr = __str_to_vector((const char *)getenv("TCP_PKTINT"));
    if (my_statics.pdintr == 0)  /* user didn't specify int # in *.cfg */
    {
      /* what-a-poo tcp compatibility */
      char e[12]; e[0]='W'; e[1]='A'; e[2]=e[3]='T'; e[4]=e[9]='C';
                  e[5]='P'; e[6]='.'; e[7]='V'; e[8]='E'; e[10]='\0';
      my_statics.pdintr = __str_to_vector((const char *)getenv(e));
    }

    /* If interrupt specified in environment or in *.cfg, then
     *  check single a single vector at that interrupt
     */
    if (my_statics.pdintr)
    {
      if (pdapi_find_driver(my_statics.pdintr) != my_statics.pdintr)
      {
        if (my_statics.debug_lev)
        {
          DBGPRINTF(("No packet driver found @ preferred int 0x%02x.\n",
                      my_statics.pdintr));
          DBGPRINTF(("Defaulting to scan for drivers...\n"));
        }
        my_statics.pdintr = 0;
      }
    }

    if (!my_statics.pdintr)
    {
      if (my_statics.debug_lev >= 2)
        DBGPRINTF(("*ETH: Begin pd scan...\n"));
      my_statics.pdintr = pdapi_find_driver(0);
      if (my_statics.debug_lev >= 2)
        DBGPRINTF(("*ETH: End pd scan. pdintr=0x%02X\n", my_statics.pdintr));
      if (my_statics.pdintr == -1) 
      {
        /* should never happen since there is a nullpkt "driver" at 80h */
        if (my_statics.debug_lev)
          DBGPRINTF(("*ETH: hmm! No packet driver found in range Int 0x60...0x80\n"));
        my_statics.pdintr = 0;  /* must be out of (DOS) memory */
        return -1;
      }
    }

    if (my_statics.debug_lev)
      DBGPRINTF(("*ETH: selected pdintr=%02x\n", my_statics.pdintr));
  }

  return 0;
}

/* -------------------------------------------------------------------- */

/*
 * __init_handles depends on a successful __init_pktintr
*/
static int __init_release_handles(int call_type)
{
  uint16_t ethtypes[3];
  unsigned int handnum, numhandles;
  int pderrno = 0;

  my_statics.hw_mcast_max_groups = 0;
  my_statics.pd_is_nullpkt = 0;
  my_statics.pd_is_p2p = 0;
  my_statics.pd_class = 0;
  my_statics.osi3_tx_mtu = 0;
  my_statics.pd_is_nullpkt = 0;
  my_statics.hw_addrlen = 0;
  my_statics.hw_type = 0;
  FAR_MEMSET( &(my_statics.my_bcaddr[0]), 0xFF, sizeof(my_statics.my_bcaddr));
  FAR_MEMSET( &(my_statics.my_hwaddr[0]), 0, sizeof(my_statics.my_hwaddr));

  if (my_statics.pdintr == 0)
  {
    /* hmm! */
  }
  else if (call_type == CALLTYPE_INIT)
  {
    if (my_statics.debug_lev >= 2)
      DBGPRINTF(("*ETH: begin init_handles\n" ));

    /* ethertypes are always in host order */
    ethtypes[0] = ETHERTYPE_IP;
    ethtypes[1] = ETHERTYPE_ARP;
    ethtypes[2] = ETHERTYPE_REVARP;
    numhandles = 0;

    for (handnum = 0; handnum < 3; handnum++)
    {
      pderrno = pdapi_open_driver( my_statics.pdintr, ethtypes[handnum], 
                          pkt_enqueue, &(my_statics.handle_tab[handnum]) );
      if (pderrno != 0)
      {  
        if (my_statics.debug_lev >= ((handnum == 0)?(1):(2)) )
          DBGPRINTF(("Couldn't allocate handle for 0x%04x: %s\n",
                              ethtypes[handnum], pd_strerror(pderrno) ));
        if (handnum != 0)
          pderrno = 0;
        break;
      }
      
      if (handnum == 0)
      {
        struct pdapi_driver_info info;
        const char *pdclassname = "";
        int frame_overhead = 0;
        memset( &info, 0, sizeof(info));
        pdclassname = pdclassname; /* shaddup compiler */

        /* --------------------- begin getinfo/part 1 --------------- */

        if (pderrno == 0)
        {
          if (my_statics.debug_lev >= 2)
            DBGPRINTF(("*ETH: begin get_driver_info\n" ));

          pderrno = pdapi_get_driver_info( my_statics.pdintr, 
                              my_statics.handle_tab[handnum], &info );
          if (pderrno != 0)
          {
            if (my_statics.debug_lev)
              DBGPRINTF(("Unable to get_driver_info: err=%s\n",pd_strerror(pderrno)));
          }
          else       
          {
            const char *msg = (const char *)0;
            if (stricmp(info.drvr_name,"nullpkt")==0)
              my_statics.pd_is_nullpkt = 1;

            switch (info.if_class)
            {
              case PDCLASS_DIX:
              {
                pdclassname = "EtherDIX";
                my_statics.pd_class = info.if_class; 
                my_statics.pd_is_p2p = 0;
                my_statics.osi3_tx_mtu = 1500; /* RFC 894 */
                my_statics.hw_type = RFC826_HWTYPE_BLUEBOOK;
                my_statics.hw_addrlen = 6;
                frame_overhead = sizeof(bluebook_hdr);
                break;
              }
              case PDCLASS_IEEE802_5: /* 4Mbit Token Ring */
              {
                #if !defined(IEEE8025_SUPPORT_CHECKED_OK)
                msg = "*ETH: Unsupported PD class (Token Ring: class %d). " \
                      "Use the IBMTOKEN.COM shim\n";
                pderrno = -1;
                #else
                my_statics.pd_class = info.if_class;
                pdclassname = "Token Ring";
                my_statics.pd_is_p2p = 0;
                my_statics.hw_type = RFC826_HWTYPE_IEEE802;
                my_statics.hw_addrlen = 6; /* expected hw address length */
                my_statics.osi3_tx_mtu = 2002; /* RFC 1042 [4Mb recommended] */
                                            /* or = 4464; */ /* RFC 1042 [4Mb max] */
                                            /* or = 17914; */ /* 16Mb */
                frame_overhead = 18+18+3+5; /* hdr+RIF+LLC+SNAP */
                #endif
                break;
              }
              case PDCLASS_SLIP:
              {
                pdclassname = "SLIP";
                my_statics.pd_class = info.if_class;
                my_statics.pd_is_p2p = 1;
                my_statics.osi3_tx_mtu = 1006; /* RFC 1055 */
                my_statics.hw_type = RFC826_HWTYPE_SERIALLINE;
                my_statics.hw_addrlen = 0;
                frame_overhead = 0;
                break;
              }
              default:
              {
                msg = "*ETH: Unsupported PD class (%d) @int 0x%02x ('%s')\n";
                pderrno = -1;
                break;
              }
            } /* switch() */

            if (pderrno != 0 && msg)
            {
              if (my_statics.debug_lev)
                DBGPRINTF((msg,info.if_class,my_statics.pdintr,info.drvr_name));
            }
            if (my_statics.debug_lev >= 2)
              DBGPRINTF(("*ETH: end get_driver_info\n" ));

          } /* pdapi_get_driver_info ok */
        } /* if (pderrno == 0) */

        /* --------------------- end getinfo/part 1 --------------- */

        if (pderrno == 0)
        {
          struct pdapi_driver_params params; 

          if (my_statics.debug_lev >= 2)
            DBGPRINTF(("*ETH: begin get_driver_params\n" ));
   
          /* NOTE that pdapi_get_driver_params() is technically a 
           * an "advanced" driver function. If not supported by
           * the packet driver, pdapi_get_driver_params() will
           * synthesize the params.
          */
          if (pdapi_get_driver_params(my_statics.pdintr, &params ) == 0)
          {
            /* the packet driver returns an MTU that includes MAC headers. Resize
            ** it so that it reflects the size available for the network (ie IP)
            ** layer. (eg 1500 for RFC894 DIX ether, 1006 for SLIP)
            ** IEEE802.5 frame link-layer overhead is variable (depending on 
            ** whether RIF and SNAP are present or not) in which case 
            ** 'frame_overhead' reflects the 'worst case' scenario.
            */
            my_statics.osi3_tx_mtu = (uint16_t)(params.ll_mtu - frame_overhead); 
        
            if (my_statics.hw_addrlen != params.addr_len)
            {
              if (my_statics.debug_lev)
                DBGPRINTF((
                "Unexpected hardware address len %u for PD @ int 0x%02x ('%s')\n",
                params.addr_len, my_statics.pdintr, info.drvr_name ));
              pderrno = -1;
            }
            else
            {
              if (info.drvr_functionality >= 2 /* requires extended driver */
               && my_statics.hw_addrlen != 0 && params.multicast_aval > 0)
              {
                if ((params.multicast_aval % my_statics.hw_addrlen) == 0)
                  my_statics.hw_mcast_max_groups = 
                          params.multicast_aval / my_statics.hw_addrlen;
              }
            }
          }
          if (my_statics.debug_lev >= 2)
            DBGPRINTF(("*ETH: end get_driver_params\n" ));

        } /* if (pderrno == 0) */

        /* --------------------- begin getparams/part 2-------------- */

        if (pderrno == 0 && my_statics.hw_addrlen != 0)
        {
          char tmp_hwaddr[sizeof(my_statics.my_hwaddr)];
          unsigned int len = sizeof(my_statics.my_hwaddr);
          pderrno = pdapi_get_address(my_statics.pdintr, 
                                      my_statics.handle_tab[handnum],
                                      &(tmp_hwaddr[0]), &len );
          if (pderrno != 0)
          {
            if (my_statics.debug_lev)
              DBGPRINTF(("Cannot determine ethernet address: %s\n",
                                          pd_strerror(pderrno)));
          }
          else if (my_statics.hw_addrlen != len ||
                   len > sizeof(my_statics.my_hwaddr) )
          {
            if (my_statics.debug_lev)
              DBGPRINTF(("AIII! unexpected hardware address len " \
                "(my_statics.hw_addrlen = %u, len = %u, max = %u)\n", 
              my_statics.hw_addrlen, len, sizeof(my_statics.my_hwaddr) ));
            pderrno = -1;
          }
          else
          {
            FAR_MEMCPY( my_statics.my_hwaddr, tmp_hwaddr, len );
          }
        }

        /* --------------------- end getaddr/part 3---------- */

        if (pderrno != 0)
        {
          pdapi_close_driver( my_statics.pdintr, 
                              my_statics.handle_tab[handnum] );
          break;
        }
        if (my_statics.debug_lev)
        {
          DBGPRINTF(("*ETH: Using '%s', Int=0x%02x, LL=%s, OSI3-tx-mtu=%u\n",
            info.drvr_name, my_statics.pdintr, pdclassname, 
            my_statics.osi3_tx_mtu));
        }
      } /* if (handnum == 0) */

      numhandles++;

    } /* for (handnum = 0; i < 3; i++) */

    /* ------------------------------------------------------------ */

    if (pderrno == 0)
    {  
      if (my_statics.debug_lev >= 2)
        DBGPRINTF(("*ETH: end init_handles OK. %d handles allocated)\n",
                                numhandles ));  
      if (numhandles >= 1)
        my_statics.ip_handleP   = &(my_statics.handle_tab[0]);
      if (numhandles >= 2)
        my_statics.arp_handleP  = &(my_statics.handle_tab[1]);
      if (numhandles >= 3)
        my_statics.rarp_handleP = &(my_statics.handle_tab[2]);
    }
    else
    {
      my_statics.hw_mcast_max_groups = 0;
      my_statics.pd_is_nullpkt = 0;
      my_statics.pd_is_p2p = 0;
      my_statics.pd_class = 0;
      my_statics.osi3_tx_mtu = 0;
      my_statics.pd_is_nullpkt = 0;
      my_statics.hw_addrlen = 0;
      my_statics.hw_type = 0;
      FAR_MEMSET( &(my_statics.my_bcaddr[0]), 0xFF, sizeof(my_statics.my_bcaddr));
      FAR_MEMSET( &(my_statics.my_hwaddr[0]), 0, sizeof(my_statics.my_hwaddr));

      my_statics.ip_handleP = (unsigned short *)0;
      my_statics.arp_handleP = (unsigned short *)0;
      my_statics.rarp_handleP = (unsigned short *)0;
    }
  }
  else /* atexit or atterm */
  {    
    unsigned short handles[3];
    numhandles = 0;

    if (my_statics.ip_handleP)
      handles[numhandles++] = *my_statics.ip_handleP;
    if (my_statics.arp_handleP)
      handles[numhandles++] = *my_statics.arp_handleP;
    if (my_statics.rarp_handleP)
      handles[numhandles++] = *my_statics.rarp_handleP;

    my_statics.ip_handleP = (unsigned short *)0;
    my_statics.arp_handleP = (unsigned short *)0;
    my_statics.rarp_handleP = (unsigned short *)0;

    for (handnum = 0; handnum < numhandles; handnum++)
    {
      __init_release_msg(call_type, "beg release handle");
      pderrno = pdapi_close_driver(my_statics.pdintr, handles[handnum]);
      __init_release_msg(call_type, "end release handle");
      if (pderrno != 0)
      {
        __init_release_msg(call_type, "AIII!! Error releasing PD handle");
        if (call_type != CALLTYPE_ATTERM && my_statics.debug_lev)
          DBGPRINTF(("Error releasing PD handle %d: (error=%d)\n", 
                                          handnum, pderrno ));
      }
    }
  }
  return pderrno;
}

/* -------------------------------------------------------------------- */

/*
 * __init_buffers depends on a successful __init_release_handles.
*/
static int __init_release_buffers(int call_type)
{
  int pderrno = 0;

  if (call_type != CALLTYPE_ATTERM)
  {
    __init_release_msg(call_type, "begin release_buffers");
    if (my_statics.send_buf)
    {
     FAR_FREE(my_statics.send_buf);
      my_statics.send_buf = (uint8_t FAR *)0;
    }
    __init_release_msg(call_type, "end release_buffers");
  }
  if (call_type == CALLTYPE_INIT)
  {
    unsigned int buf_lengths = _eth_get_mtu() + MAX_HWPKTHDR_SIZE;

    if (my_statics.debug_lev >= 2)
      DBGPRINTF(("*ETH: begin init_buffers (buf_lengths=%u)\n",
                                            buf_lengths ));

    my_statics.send_buf = (uint8_t FAR *)FAR_MALLOC(buf_lengths);

    if (!my_statics.send_buf)
    {
      pderrno = PDERR_NO_SPACE; 
      if (my_statics.debug_lev >= 2)
        DBGPRINTF(("*ETH: end init_buffers (FAILED)\n" ));
    }
    else
    {
      if (my_statics.debug_lev >= 2)
        DBGPRINTF(("*ETH: end init_buffers (OK)\n" ));
    }
  }
  return pderrno;
}    
    
/* -------------------------------------------------------------------- */

/*
 *  initialize/deinitialize the network driver interface
 *  Return -1 on error or 0 if okay.
 *
 *  The only possible reason for a failure is an inability 
 *  to alloc resources (memory, atterm() hook, idle cb, etc) 
*/
static void _atexit_release(void) { __init_release(CALLTYPE_ATEXIT); }
static void _atterm_release(void) { __init_release(CALLTYPE_ATTERM); }

static int __init_release(int call_type)
{
  if (call_type != CALLTYPE_INIT)
  {
    if (my_statics.pdintr)
    {
      blocker_critsect_begin();
      my_statics.shutting_down = 1;

      __init_release_msg(call_type,"shutdown begin");

      /* deinitialization order MUST be 
      * the reverse of initialization order 
      */ 
      __init_release_buffers(call_type);
      __init_release_handles(call_type);
      __init_release_pktintr(call_type);

      my_statics.shutting_down = 0;
      blocker_critsect_end();

      __init_release_msg(call_type,"shutdown complete");
    }
    return 0;
  }

  if (!my_statics.pdintr)
  {
    int no_rsrcs = 0;
    _conf_load(); /* load config if not already loaded */

    /* _my_atterm()/_my_atexit() do an exit(255) on failure, 
     * so call them _before_ we do any hardware diddling.
    */
    _my_atexit(_atexit_release);
    _my_atterm(_atterm_release);

    blocker_critsect_begin();

    if (my_statics.debug_lev >= 2)
      DBGPRINTF(("*ETH: begin hook add.\n"));
    if (blocker_idlehook_add(_eth_poll) != 0 ||  /* efficient */
        blocker_tisrhook_add(_eth_poll) != 0)    /* fallback */
    {
      /* should never happen (my error in timer stuff) */
      __fatal_runtime_error("PANIC! Unable to add idle hooks\n");
    }
    if (my_statics.debug_lev >= 2)
      DBGPRINTF(("*ETH: end hook add (OK)\n"));

    /* CAREFUL! correct initialization order is critical.
    */ 
    if (__init_release_pktintr(CALLTYPE_INIT) == 0)
    {
      int pass;
      for (pass = 0; pass < 2; pass++)
      {
        /* init_handles depends on a successful init_pktintr */
        if (__init_release_handles(CALLTYPE_INIT) == 0)
        {
          /* init_buffers depends on a successful init_handles */
          if (__init_release_buffers(CALLTYPE_INIT) == 0)
          {  
            if (my_statics.debug_lev >= 2)
              DBGPRINTF(("*ETH: init complete.\n"));
            blocker_critsect_end();
            return (0); /* success */
          }
          no_rsrcs = 1; /* buffer or hook set failed */
          __init_release_handles(CALLTYPE_ATEXIT);
        }
        if (no_rsrcs || my_statics.pdintr == 0x80)
          break;
        my_statics.pdintr = 0x80; 
      } /* for (pass = 0; pass < 2; pass++) */
      __init_release_pktintr(CALLTYPE_ATEXIT);
    }

    blocker_critsect_end();

    /* Update August 2001 (cyp):
     * 
     * Failure to alloc resources (mem) is now FATAL (since the app would
     * die soon anyway), but strictly speaking, the rest of the
     * lib should be able to cope fine, since the failure will
     * then merely causes all the other _eth*() functions to fail.
     *
     * BUT, some things can have bad side effects when init fails
     * eg, _eth_get_mtu() returns zero, and although that _should_ 
     * be the callers problem to deal with, its often forgotten, 
     * with terrible consequences.
    */
    if (!my_statics.pdintr) /* dummy if (always true) */
    {
      if (no_rsrcs)
        __fatal_runtime_error("PANIC! Out of memory or other critical resource\n");
      exit(255);
    }
    return -1; /* failed */

  } /* if (!my_statics.pdintr) */
    
  return 0; /* success */
}

/* ======================================================================= */
/*                         48 bit ethernet-address conversion              */
/* ======================================================================= */

#if 0 /* not used */
const char *_my_ethtoa (const void *_eth, char *buffer)
{
  const char *eth = (const char *)_eth;
  int i; char *tmp = buffer;
  for (i = 0; i < 6; i++)
  {
    char c;
    if (i) *tmp++ = ':';
    c = (char)(eth[i] >> 4);
    *tmp++ = (char)((c <= 9) ? ('0'+c) : ('a'+(c - 10)));
    c = (char)(eth[i] & 0xff);
    *tmp++ = (char)((c <= 9) ? ('0'+c) : ('a'+(c - 10)));
  }
  *tmp = '\0';
  return buffer;
}
#endif

#if 0 /* not used */
const void *_my_atoeth(const char *str, void *eth_addr)
{
  unsigned int pos = 0;
  uint8_t buff[6];
  while (*str)
  {    
    int val = 0;
    if (*str >= '0' && *str <= '9')
      val = (*str++ - '0');
    else if (*str >= 'A' && *str <= 'F')
      val = (*str++ - 'A')+10; 
    else  if (*str >= 'a' && *str <= 'f') 
      val = (*str++ - 'a')+10; 
    else 
      return (const void *)0;
    if (*str >= '0' && *str <= '9')
      val = (val << 4)+(*str++ - '0');
    else if (*str >= 'A' && *str <= 'F')
      val = (val << 4)+(*str++ - 'A')+10; 
    else  if (*str >= 'a' && *str <= 'f') 
      val = (val << 4)+(*str++ - 'a')+10; 
    if (*str != ':' && pos < 5)
      return (const void *)0;
    if (*str != '\0' && pos == 5)
      return (const void *)0;
    if (pos == 6)
      return (const void *)0;
    buff[pos++] = (unsigned char)val;
    if (*str)
      str++;
  } 
  if (pos != 6)
    return (const void *)0;
  for (pos = 0; pos < sizeof(buff); pos++)
     ((uint8_t *)eth_addr)[pos] = buff[pos];
  return eth_addr;
}
#endif
