/*
 *  Configuration/Settings stuff including
 *  - Whut-a-Poo config file compatible reader/parser.
 *  - Support for G. Vanem's "extensions" to Whut-a-Poo.cfg
*/
#define WANT_WHATAPOOCFG_SUPPORT /* yes, we want to support whutapoo.cfg */

#include <stdio.h>
#include <stdlib.h>   /* getenv() */
#include <string.h>   /* str*() */
#include <io.h>       /* read() */
#include <share.h>    /* SH_DENYWR */
#include <fcntl.h>    /* open modes */
#include <ctype.h>    /* toupper() */

#include "target.h"   /* standard includes */
#include "debug.h"    /* KMSG(), __wkern_is_debugee() */
#include "inaddr.h"   /* _my_aton() */
#include "printk.h"   /* _printk() */
#include "loadcfg.h"  /* ourselves */

#undef TRACE_LOADCFG

#if defined(TRACE_LOADCFG)
#undef TRACE_LOADCFG
#define TRACE_LOADCFG(x) _printk x
#else
#define TRACE_LOADCFG(x) NO_OPERATION()
#endif

/* ------------------------------------------------------------------ */

static struct
{
  struct { 
     const struct config_table FAR *tbl; 
     const char FAR *lbl; 
  } confextn[16];
  int done_config;
} FAR my_statics;

/* ------------------------------------------------------------------ */

static void __init_statics_if_needed(void)
{
  static int my_bss_count = -1;
  if (my_bss_count < __bss_count) /* always LT compare! */
  {
    FAR_MEMSET( &my_statics, 0, sizeof(my_statics));
    my_bss_count = __bss_count;
  }
  return;
}

/* ------------------------------------------------------------------ */
    
void _conf_add_confextn(const struct config_table FAR *tbl, 
                        const char FAR *optprefix)
{
  if (tbl)
  {
    unsigned int i, n = 0;
    __init_statics_if_needed();    
    for (i = 0;
         i < (sizeof(my_statics.confextn)/sizeof(my_statics.confextn[0]));
         i++)
    {
      if (tbl == my_statics.confextn[i].tbl)
        return;
      if (!n && !my_statics.confextn[i].tbl)
        n = i+1;
    }
    if (n)
    {
      my_statics.confextn[n-1].tbl = tbl;
      my_statics.confextn[n-1].lbl = optprefix;  
    }
    else
    {
      __fatal_runtime_error("AIIIEE! The confextn table is too small!\n");
    }
  }
  return;
}

/* ------------------------------------------------------------------ */
    
/*
 * Parse the config-table and if match is found for
 * ('section'+'.'+)'name' either store variable to 'value' or
 * call function with 'value'.
 *
 * Returns the config-entry-type if found, else 0.
 */
static int parse_config_table (const struct config_table FAR *tab,
                               const char FAR *section,
                               const char *name,
                               const char *value)
{
  /* don't bother scanning this tab if name does not begin with 'section' */
  if (section)
  {
    int namelen = strlen(name);
    int sectlen = FAR_STRLEN(section);
    if (namelen <= sectlen)
      return 0;
    if (FAR_MEMCMP(name, section, sectlen)!=0)
      return 0;
    name += sectlen;
  }

  TRACE_LOADCFG(("parse_config_table sect='%s',name='%s',val='%s'\n", 
                        (section?section:""), name, value ));

  for ( ; tab && tab->keyword; tab++)
  {
    if (FAR_STRCMP(tab->keyword,name)==0)
    {
      void FAR *arg = tab->arg_func;

      /* TRACE_LOADCFG(("ARG_xxx=%c, table=%s matched '%s%s'='%s'\n",
                        tab->type,  (section?section:"(null)"), 
                            (section?section:""), name, value));
      */

      if (tab->type & ARG_ATOI)
      {
        *(int FAR *)arg = atoi (value);
      }
      else if (tab->type & ARG_INADDR)
      {
        uint32_t tmp;
        if (_my_aton(value, &tmp))
          *(uint32_t FAR *)arg = tmp;
      }
      else if (tab->type & ARG_FUNC)
      {
        typedef void (*arg_func_handler_t)(const char *);
        (*((arg_func_handler_t)arg))(value);
      }
      return (tab->type);
    }
  }
  return (0);
}

/* ------------------------------------------------------------------ */

static int __internal_pass_opt(const char *name, const char *value)
{
  int found = 0, post_load_callout = 0; 
  unsigned int q;

  if (*name == '#' && name[1] == '\0')
    post_load_callout = 1;

  for (q=0;q<(sizeof(my_statics.confextn)/sizeof(my_statics.confextn[0]));q++)
  { 
    if (my_statics.confextn[q].tbl)
    {
      int rc;
      const char FAR *section = (const char FAR *)0;
      if (!post_load_callout)
        section = my_statics.confextn[q].lbl;
      rc = parse_config_table(my_statics.confextn[q].tbl,
                             section, name, value);
      if (rc != 0) 
      {
        found = 1;
        if (((rc & ARG_MULTI) == 0) && (!post_load_callout))
          break;
      }
    }
  }
  return found;
}

int _conf_passopt( const char *name, const char *value)
{
  if (name && *name && value && *value)
  {
    _conf_load(); /* load config if not already loaded */
    return __internal_pass_opt(name, value);
  }
  return 0;
}

/* ================================================================== */
/*                    What-a-poo.cfg file open/read/parse             */
/* ================================================================== */

#if defined(WANT_WHATAPOOCFG_SUPPORT)
static int __read_line( int fd, char *buf, unsigned int buflen)
{
  unsigned int readlen = (unsigned int)read(fd, buf, buflen);
  if (readlen == ((unsigned int)-1) )
    readlen = 0;
  if (readlen > 0)
  {
    unsigned int retlen, pos = 0;
    while (pos < readlen && (buf[pos] == '\t' || buf[pos] > 26))
      pos++;
    retlen = pos;
    while (pos < readlen && buf[pos] <= 26) 
      pos++;
    if (pos < readlen)
      lseek( fd, -((long)(readlen - pos)), SEEK_CUR);
    else if (readlen == buflen)
    {
      char c = 0x7f;
      while (c > 26)
      {
        if (read(fd, &c, 1) != 1)
          c = 0;
      }
      while (c && c <= 26)
      {        
        if (read(fd, &c, 1) != 1)
          c = 0;
      }      
    }
    if (retlen == buflen)
      retlen--;
    buf[retlen] = '\0';
    readlen = retlen;
  }
  return readlen;
}
#endif

/* ------------------------------------------------------------------ */

#if defined(WANT_WHATAPOOCFG_SUPPORT)
static int __open_file(const char *filename)
{
#if defined(__DJGPP__) && !defined(sopen)
  #define sopen(__fn,__access,__sh,__perm) \
           open(__fn,((__access)|(__sh)),__perm)
#endif
#ifndef O_NOINHERIT
#define O_NOINHERIT 0
#endif
  return sopen(filename, O_RDONLY|O_BINARY|O_NOINHERIT, SH_DENYWR, 0600);
}                                 
#endif
                                 
/* ------------------------------------------------------------------ */

/* 
 * Expand one or more $(ENVVAR)s embedded in a string, may be in 
 * the form $(ENVVAR) or ${ENVVAR}, but bracket style must match.
 *
 * Return a string with environment variables expanded.
 *   E.g. if environment variable ETC is "c:\network\bin",
 *    "$(ETC)\hosts" becomes "c:\network\bin\hosts"
 * Returns NULL if the expanded string length would be > destlen.
 *
 * Nesting is illegal (recursion not supported) and renders the whole 
 * nest invalid (just as if the envvar within the outermost bracket pair 
 * didn't exist in the environment).
*/
#if defined(WANT_WHATAPOOCFG_SUPPORT)
static const char *__expand_envvars (register const char *str,
                                     char *destbuf, unsigned int destlen)
{
  register unsigned int len = 0;

  len = strlen(str);
  if (len > destlen)
    return (const char *)0;
  destlen--; /* leave space for '\0' */
  memmove(&destbuf[destlen-len],str,len);
  destbuf[destlen] = '\0';

  str = &destbuf[destlen-len];
  len = 0;

  while (*str)
  {
    if (*str != '$' || (str[1] != '(' && str[1] != '{'))
    {
      if (destlen < 1)
        return (const char *)0;
      destbuf[len++] = *str++;  
      destlen--;
    }
    else
    {
      char brkt = (str[1] == '(' ? ')' : '}');
      int i = 0, nest_count = 0, was_nested = 0; 
      char env[30];
      
      str += 2;
      while (*str)
      {
        if (i == 0 && (*str == ' ' || *str == '\t'))
        {
          while (*str == ' ' || *str == '\t')
            str++; 
        }
        if (nest_count == 0 && (*str == ')' || *str == '}'))
        {
          if (*str++ != brkt) /* wrong bracket kind */
            i = 0;
          if (!was_nested && i > 0 && i < ((int)(sizeof(env)-1)))
          {
            const char *p = (const char *)0;
            while (i > 0 && (env[i-1] == ' ' || env[i-1] == '\t'))
              i--;
            env[i] = '\0';
            if (i > 0)
              p = getenv(env);
            if (p)
            {
              i = strlen(p);
              if ((len + i + strlen(str)) >= destlen)
                return (const char *)0;
              memcpy( &destbuf[len], p, i );
              len += i;
              destlen -= i;
            }
          }
          break;
        }
        if (*str == '(' || *str == '{')
        {
          nest_count++;
          was_nested = 1;
          str++;
        }
        else if (*str == ')' || *str == '}')
        {
          --nest_count;
          str++;
        }
        else if (nest_count > 0)
        {
          str++;
        }
        else if (*str && i < ((int)(sizeof(env)-1)))
          env[i++] = (char)toupper(*str++);
      }
    }
  }
  destbuf[len] = '\0';
  return destbuf;
}
#endif

/* ------------------------------------------------------------------ */

#if defined(WANT_WHATAPOOCFG_SUPPORT)
static void __parse_config(int fd, char *buf, unsigned int buflen)
{
  while (__read_line(fd, buf, buflen) > 0)
  {
    char *value, *tmp, *name;
    char c[4];

    name = buf;
    while (*name == ' ' || *name == '\t')
      name++;
    if (*name == '#' || *name == ';')
      continue;

    tmp = buf;
    while (*name && *name != '=')
      *tmp++ = (char)toupper(*name++);
    if (*name != '=') /* missing '=' */
      continue;
    if (tmp == buf) /* zero length name */
      continue;

    value = name+1;
    tmp--;
    while (*tmp == ' ' || *tmp == '\t')
      tmp--;
    tmp++;
    *tmp++ = '\0';

    while (*value == ' ' || *value == '\t')
      value++;
    c[0] = '\t';
    c[1] = ' ';
    c[2] = '#';
    c[3] = ';';
    if (*value == '\'' || *value == '\"')
      c[0] = c[1] = c[2] = c[3] = *value++;
      
    name = tmp;
    while (*value && 
      *value!=c[0] && *value!=c[1] && *value!=c[2] && *value!=c[3])
      *tmp++ = *value++;
    if (name == tmp) /* zero length value */
      continue;
    *tmp = '\0'; 
    value = name;
    name = buf; 

    if (__expand_envvars (value, value, (size_t)(&buf[buflen-1] - value) ))
    {
      /* INCLUDE= is never handled in the dispatcher */
      if (strcmp(name, "INCLUDE")==0)
      {
        if (value[0] != '?' || value[1] != '\0')
        {
          /* stack usage per invocation:
          **   (5 * sizeof(void *)) + 4 + (3 * sizeof(int)) => 40 (pmode app).
          */
          static int include_level = 0;
          if ((++include_level) <= 5)
          {
            int newfd = __open_file(value + (*value == '?' ? 1 : 0) );
            if (newfd != -1)
            {
              /* Recursion, but we're reentrant. */
              __parse_config(newfd, buf, buflen); 
              close(newfd); 
            }
          }
          --include_level;
        }
      }
      else 
      {
        __internal_pass_opt(name, value);
      }
    } /* if (__expand_envvars,... */
  } /* while (__read_line(fd, buf, buflen) > 0) */
  return;
}
#endif

/* ------------------------------------------------------------------ */

#if defined(WANT_WHATAPOOCFG_SUPPORT)
#ifdef __cplusplus
extern "C" {
#endif
#if defined(__HIGHC__)
  extern char *GetArg0 (USHORT sel);
#elif defined (__DJGPP__)
  extern char **__crt0_argv;
#elif defined (_MSC_VER)
  extern char **__argv;
#elif defined(__WATCOMC__) || defined(__BORLANDC__) || defined(__TURBOC__)
  extern char **_argv;
#else
  #error "What's up doc?"
#endif 
#ifdef __cplusplus
}
#endif
#endif 

/* ------------------------------------------------------------------ */

#if defined(WANT_WHATAPOOCFG_SUPPORT)
static const char *__get_argv0(void)
{
  /* This function is duplicated in pcconfig.c, printk.c and syslog.c */
#if defined(__HIGHC__)
  if (_osmajor >= 3)
  {
    /* extern char *GetArg0 (USHORT sel); */
    CONFIG_INF   cnf;
    _dx_config_inf (&cnf, (UCHAR*)&cnf);  /* get config block */
    return GetArg0 (cnf.c_env_sel);       /* in exc_??.lib    */
  }
#elif defined (__DJGPP__)
  _get_dos_version (0); /* 2.02 bug; _osmajor/_osminor not set in crt0.o */
  if (_osmajor >= 3)
  {
    /* extern char **__crt0_argv; */
    return __crt0_argv[0];
  }
#elif defined (_MSC_VER)
  if (_osmajor >= 3)
  {
    /* extern char **__argv; */
    return __argv[0];
  }
#elif defined(__WATCOMC__) || defined(__BORLANDC__) || defined(__TURBOC__)
  if (_osmajor >= 3)
  {
    /* extern char **_argv; */
    return _argv[0];
  }
#else
  #error "What's up doc?"
#endif 
  return (const char *)0;
}
#endif

/* ------------------------------------------------------------------ */

#if defined(WANT_WHATAPOOCFG_SUPPORT)
static void __internal_conf_load(const char *config_name, 
                                 const char *environ_name)
{
  const char *base_name, *path = (const char *)0;
  char  buf[132]; /* both linebuf and filename buf */
  int len;

  base_name = config_name;
  if (!base_name)
    base_name = "";

  /* if 'config_name' is canonical, then try that first.
  */
  if (!path && *base_name)
  {
    path = base_name;
    len = strlen(base_name);
    while (len > 0 && base_name[len-1] != '\\' && base_name[len-1] != '/')
      len--;
    base_name += len;
    if (len == 0 || access(path, 0)!=0) /* not canonical or doesn't exist */
      path = (const char *)0;
  }    

  /* Look for an envvar 'environ_name' which may be a filename or
  ** a directory. If a dir, try it with the base_name  from 'config_name'
  */
  if (!path)
  {
    path = environ_name;
    if (path)
      path = getenv (path);
    if (path)
    {
      len = strlen( path );
      if (len == 0)
        path = (const char *)0;
      else if (access(path, 0) == 0) /* its a filename */
        ;
      else if (!*base_name)
        path = (const char *)0;
      else if ((len + 3 + strlen(base_name)) <= sizeof(buf))
        path = (const char *)0;
      else          
      {
        strcpy(buf, path);
        path = (const char *)0; 
        if (buf[len-1] != '\\' && buf[len-1] != '/')
          buf[len++] = '\\';
        buf[len] = '\0';
        path = strcat( buf, base_name);
        if (access(path, 0) != 0)
          path = (const char *)0;
      }
    }
  }

  /* Try using the directory in argv[0] with 
  ** the base_name from 'config_name';
  */
  if (!path && *base_name) 
  {
    path = __get_argv0();
    if (path)
    { 
      buf[0] = '\0';
      len = strlen(path);
      while (len > 0 && path[len-1] != '/' && path[len-1] != '\\')
        len--;
      if (len > 0 && (len + strlen(base_name) + 1) < sizeof(buf))
      {
        memcpy( buf, path, len);
        strcpy( &buf[len], base_name );
      }
      path = buf;
      if (!buf[0] || access(buf,0)!=0)
        path = (const char *)0;
    }
  }

  /* Try the current directory with
   * the base_name from 'config_name'
  */
  if (!path && *base_name)
  {
    if (access(base_name, 0) == 0)
      path = base_name;
  }

  /* Search the search path with
   * the base_name from 'config_name'
  */
  if (!path)
  {
    path = getenv("PATH");
    if (path)
    {
      while (*path)
      {
        char quot = ';';
        while (*path == ' ' || *path == '\t')
          path++; 
        if (*path == '"' || *path == '\'')
          quot = *path++;
        len = 0;
        while (*path && *path != quot)
        {
          if (len < (int)(sizeof(buf)-1))
            buf[len++] = *path;
          path++;
        }
        if (len < (int)(sizeof(buf)-1))
        { 
          while (len > 0 && (buf[len-1] == ' ' || buf[len-1] == '\t'))
            len--;
          if (len > 0)
          {
            if (buf[len-1] != '\\' && buf[len-1] != '/' && buf[len-1] != ':')
              buf[len++] = '\\';
            if ((len + strlen(base_name)) < sizeof(buf))
            {
              strcpy( &buf[len], base_name);
              if (access(buf, 0) == 0)
              {
                path = buf;
                break;
              }
            }
          }
        }
        while (*path && *path != ';')
          path++;
        if (*path == ';')
          path++;
      }
      if (!*path)
        path = (const char *)0;
    }
  }

  if (path)
  {
    int fd = __open_file(path);
    if (fd != -1)
    {
      __parse_config(fd, buf, sizeof(buf));
      close(fd);    
    }
  }
  return; 
}
#endif

/* ------------------------------------------------------------------ */

void _conf_load(void)
{
  __init_statics_if_needed();
  if (!my_statics.done_config)
  {
    my_statics.done_config = 1;
    #if defined(WANT_WHATAPOOCFG_SUPPORT)
    {
      char e[12]; e[0]='W'; e[1]='A'; e[2]=e[3]='T'; e[4]=e[7]='C';
                  e[5]='P'; e[6]='.'; e[8]='F'; e[9]='G'; e[10]='\0';
      __internal_conf_load(e, e);
    }
    #endif
    __internal_pass_opt("#", ""); /* post-load callout */
    TRACE_LOADCFG(("load config complete\n"));
  }
  return; /* always ok */
}

