/*
 * Portable (model and real/protected mode independant) 
 * Packet Driver Interface.
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: pktdrvr.c,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
/*
** Portable (model and real/protected mode independant) 
** Packet Driver Interface.
**
** Written May 2001, Cyrus Patel <cyp@fb14.uni-mainz.de>.  
**
** - Supports as many packet drivers as available memory
**   will allow.
** - Provides a loopback interface, irrespective of whether 
**   the packet driver supports (and correctly deals with) 
**   ethertype 0x9000 or not.
** - Has a built-in 'nullpkt' driver should no real packet driver 
**   be available. This 'nullpkt' driver behaves just as a 
**   real driver would, but returns PDERR_CANT_SEND (if the 
**   destination is not loopback) for the send_pkt() call.
**   The interrupt number and driver class are configurable
**   via #defines and default to 0x80 and 1 respectively.
** - Opening the packet driver for access does not require
**   knowing the class of packet driver in advance.
** - Some packet driver functions that are only available
**   in non-basic drivers are emulated if necessary.
**   For example, the 'get driver params' call.
** - Synchronous spool backend which allows the "packet driver" 
**   to be polled for available packets. The spool is also
**   interrupt safe, so polling may (also) be done from an 
**   interrupt handler if desired.
** - Spool can grow dynamically, so packet loss is minimized
**   even when the queue is not polled frequently.
** - Spool supports packets of (theoretically) any size.
**   (I have however implemented an (arbitrary) cap of 32K 
**   in order to not have queues that are uneccessarily large).
** - Does not use real-mode callbacks for the pm interface,
**   the spool is entirely in real mode, which avoids the
**   necessity of switching processor modes (twice!) on 
**   each upcall.
**
** Based originally on ...
** "Packet Driver Specification Revision 1.09" 
** http://www.crynwr.com/packet_driver.html
**
** Updated for ...
** "Packet Driver Specification Revision 1.11" 
** http://195.148.172.4/~aw/net/pds111.html
**
** If you have questions/flames about the packet-driver specification,
** send them to the author, John Romkey <romkey@apocalypse.org>
*/
#undef WANT_STACK_CHECKING /* disable stack checking for this module */
#include "target.h"  /* standard includes */
#include "dpmi.h"    /* __dpmi_*()  */
#include "debug.h"   /* DBGPRINTF() */
#include "printk.h"  /* printk() */
#include "pktdrvr.h" /* ourselves */

#undef  SHOW_PACKET_POP /* _printk()s a 'P' on pop */
#undef  SHOW_POLL_CALL  /* _printk()s a 'p' on pop */
#undef  SHOW_QUEUE_GROW /* _printk()s a '+G...G-' on grow */

/* DO NOT PUT NULL_DRIVER_* constants in a header. These are absolutely
 * private. Any application that depends on them having a specific
 * value is herewith considered broken.
*/
#define NULL_DRIVER_CLASS  1    /* 1...11 (any spec 1.08 class) except 4 */
#define NULL_DRIVER_INTNO  0x80 /* intno can be 0-INT_MAX but should be */
/* 0x60..0x80 for compatibility. A real pktdrvr at that # has precedence. */

/* ---------------------------------------------------------------------- */

/* Disable stack-checking here
 */
#if defined(__HIGHC__)
#pragma off(check_stack)
#pragma off(call_trace)
#pragma off(prolog_trace)
#pragma off(epilog_trace)
#elif defined(__WATCOMC__)
#pragma off(check_stack)
#elif defined(_MSC_VER)
#pragma off(check_stack)
#elif defined(__TURBOC__) && (__TURBOC__ > 0x301)
#pragma option -N-
#elif defined(__BORLANDC__)
#pragma option -N-
#endif

/* ---------------------------------------------------------------------- */

#if 0 /* already defined in target.h */
#if (UINT_MAX == USHRT_MAX) /* 16bit code */ 
/* no reason why buffers need to be on the local heap */
#define FAR _far
#define FAR_MALLOC  _fmalloc
#define FAR_REALLOC _frealloc
#define FAR_FREE    _ffree
#define FAR_MEMSET  _fmemset
#define FAR_MEMCPY  _fmemcpy
#else
#define FAR /*nothing*/
#define FAR_MALLOC  malloc
#define FAR_REALLOC realloc
#define FAR_FREE    free
#define FAR_MEMSET  memset
#define FAR_MEMCPY  memcpy
#endif
#endif

/* ===================================================================== */
/*                          Packet API entry                             */
/* ===================================================================== */

/*
 * IREGS structure for PKT_API()
*/
#if defined(PHARLAP_H)
  typedef struct IREGS {
          unsigned short r_ax, r_bx, r_cx, r_dx, r_bp;
          unsigned short  r_si, r_di, r_ds, r_es, r_flags;
        } IREGS;
#elif defined(__DPMI32__) /* all DPMI targets - includes pharlap */
  #define IREGS      __dpmi_regs
  #define r_flags    x.flags
  #define r_ax       x.ax
  #define r_bx       x.bx
  #define r_dx       x.dx
  #define r_cx       x.cx
  #define r_si       x.si
  #define r_di       x.di
  #define r_ds       x.ds
  #define r_es       x.es
#elif defined(__WATCOMC__)
  typedef union REGPACK IREGS;
  #define r_ax w.ax
  #define r_bx w.bx
  #define r_cx w.cx
  #define r_dx w.dx
  #define r_bp w.bp
  #define r_di w.di
  #define r_si w.si
  #define r_ds w.ds
  #define r_es w.es
  #define r_flags w.flags
#elif defined(__TURBOC__)
  typedef struct REGPACK IREGS;
#elif defined(_MSC_VER) || defined(__WATCOMC__)
  typedef struct IREGS {
    union REGS _irr;
    #define r_ax    _irr.w.ax
    #define r_bx    _irr.w.bx
    #define r_cx    _irr.w.cx
    #define r_dx    _irr.w.dx
    #define r_si    _irr.w.si
    #define r_di    _irr.w.di
    #define r_flags _irr.w.cflag
    struct SREGS _irs;
    #define r_es    _irs.es
    #define r_ds    _irs.ds
  } IREGS;
#endif

/*
 *  Return pderrno (0 on success)
 */
static int PKT_API(int intno, IREGS *reg)
{
  int pderrno = 0;
#if defined(PHARLAP_H)
  SWI_REGS r;
  r.eax        = reg->r_ax;
  r.ebx        = reg->r_bx;
  r.ecx        = reg->r_cx;
  r.edx        = reg->r_dx;
  r.esi        = reg->r_si;
  r.edi        = reg->r_di;
  r.es         = reg->r_es;
  r.ds         = reg->r_ds;
  r.flags      = reg->r_flags;
  _dx_real_int(((UINT)intno), &r);
  reg->r_ax    = (unsigned short)r.eax;
  reg->r_bx    = (unsigned short)r.ebx;
  reg->r_cx    = (unsigned short)r.ecx;
  reg->r_dx    = (unsigned short)r.edx;
  reg->r_si    = (unsigned short)r.esi;
  reg->r_di    = (unsigned short)r.edi;
  reg->r_es    = (unsigned short)r.es;
  reg->r_ds    = (unsigned short)r.ds;
  reg->r_flags = (unsigned short)r.flags;
#elif defined(__DPMI32__) /* dpmi targets (which includes pharlap) */
  __dpmi_int((int)intno, reg);
#elif defined(__WATCOMC__)
  intr( intno, (union REGPACK *)reg);
#elif defined(__TURBOC__)
  intr( intno, (struct REGPACK *)reg);
#elif defined(_MSC_VER) || defined(__WATCOMC__)
  int86x( intno, &(reg->_irr), &(reg->_irr), &(reg->_irs));
#endif
  if (reg->r_flags & 1)   
  {
    pderrno = (unsigned short)(0xff & (reg->r_dx >> 8));
    if (pderrno == 0) 
      pderrno = 0xff;
  }
  return pderrno;
}

/* ==================================================================== */
/*                         real mem helper                              */
/* ==================================================================== */

static void __my_dosmem_free(unsigned long segoff)
{
  segoff = segoff; /* we never free. Let DOS do it when the app exits */
  return;
}

static unsigned long __my_dosmem_alloc(size_t numoctets) /* returns segoff */
{
  /* start at 1514 since we'll probably need
  ** a block of (at least) that size eventually.
  */
  if (numoctets < (2048-16))
    numoctets = (2048-16);

  if (numoctets <= 0xfff0) /* although possible we don't allow it */
#if defined(PHARLAP_H)  /* pharlap target with pharlap API */
  {
    static struct
    {
      unsigned short seg;
      unsigned short curr_size; /* in paras */
    } rmem = {0,0};
    unsigned short paras = (unsigned short)((numoctets + 15)>>4);

    if (rmem.seg && paras > rmem.curr_size)
    {
      _dx_real_free(rmem.seg);
      rmem.curr_size = 0;
      rmem.seg = 0;
    }
    if (!rmem.seg)
    {
      USHORT rm_seg, rm_size;
      if (_dx_real_above(paras,&rm_seg,&rm_size)!=0)
      {
        ULONG temp;
        rm_seg = 0;
        if (_dx_cmem_usage(0, FALSE, &temp, &temp) == 0)
        {
          if (_dx_real_above(paras, &rm_seg, &rm_size) != 0)
            rm_seg = 0;
          _dx_cmem_usage(0, TRUE, &temp, &temp);
        }
      }
      if (rm_seg)
      {
        rmem.curr_size = paras;
        rmem.seg = rm_seg;
      }
    }
    return ((unsigned long)rmem.seg)<<16;
  }
#elif defined(__DPMI32__) /* all pmode targets (includes pharlap) */
  {
    /* we don't really need the selector, and therefore we ought to be
     * able to use a plain int21/ah=48, but for some reason that call
     * crashes (under DOS4GW anyway).
    */
    static struct
    {
      unsigned short sel;
      unsigned short seg;
      unsigned short curr_size; /* in paras */
    } rmem = {0,0,0};

    unsigned short paras = (unsigned short)((numoctets + 15)>>4);

    if (rmem.seg && paras > rmem.curr_size)
    {
      __dpmi_free_dos_memory( rmem.sel );
      rmem.curr_size = 0;
      rmem.seg = 0;
      rmem.sel = 0;
    }
    if (!rmem.seg)
    {
      int seg, sel;
      seg = __dpmi_allocate_dos_memory( paras, &sel );
      if (seg != -1)
      {
        rmem.seg = (unsigned short)seg;
        rmem.sel = (unsigned short)sel;
        rmem.curr_size = paras;
      }
    }
    return ((unsigned long)rmem.seg)<<16;
  }
#else /* real mode targets */
  {
    static struct {
      void FAR *segoff;
      unsigned int curr_size; /* in bytes */
    } rmem = {0,0};

    if (rmem.segoff && numoctets > rmem.curr_size)
    {
      void FAR *segoff = FAR_REALLOC( rmem.segoff, numoctets );
      if (segoff)
      {
        rmem.segoff = segoff;
        rmem.curr_size = numoctets;
      }
      else
      {
        FAR_FREE( rmem.segoff );
        rmem.segoff = (void FAR *)0;
        rmem.curr_size = 0;
      }
    }
    if (!rmem.segoff)
    {
      rmem.segoff = FAR_MALLOC( numoctets );
      if (rmem.segoff)
        rmem.curr_size = numoctets;
    }
    return (unsigned long)(rmem.segoff);
  }
#endif       
  return 0ul;
}

#if defined(__DPMI32__)  /* all DPMI targets (includes pharlap) */
  /* DPMI DOSmem read/write must be inline because functions are not locked */
  #define __my_dosmem_read( __buf, __segoff, __len) \
    dosmemget( (((__segoff)>>12)&0xffff0u)+((__segoff) & 0xffffu), (__len), (__buf) )
  #define __my_dosmem_write( __segoff, __buf, __len) \
    dosmemput( (__buf) , (__len), (((__segoff)>>12)&0xffff0u)+((__segoff) & 0xffffu) )
#else /* real mode target */
  static void __my_dosmem_read(void FAR *buf,unsigned long segoff,size_t len)
  {  FAR_MEMCPY( buf, (void FAR *)segoff, len ); }
  static void __my_dosmem_write(unsigned long segoff,const void FAR *buf,size_t len)
  {  FAR_MEMCPY( (void FAR *)segoff, buf, len ); }
#endif

/* ==================================================================== */
/*                    PKT API: is driver there?                         */
/* ==================================================================== */

static int __is_driver_there( int intno )
{
  static char is_x80 = ((char)0xff);
  static unsigned long is_there = 0ul, not_there = 0ul;
  unsigned long rmaddr = 0ul;

  /* DBGPRINTF(("*__is_driver_there(0x%x)\n",intno)); */

  if (intno != 0x80)
  {
    if (intno < 0x60 || intno > 0x80)
      return 0;
    if (is_there  & (1ul << (intno - 0x60)))
      return 1;
    if (not_there & (1ul << (intno - 0x60)))
      return 0;
  }
  else if (is_x80 != ((char)0xff))
    return is_x80;

  #if defined(PHARLAP_H) /* pharlap target with pharlap API */
  {
    _dx_rmiv_get(intno, (REALPTR *)&rmaddr );
  }
  #elif defined(__DPMI32__) /* all pmode targets (including pharlap) */
  {
    __dpmi_raddr rmiv;
    if (__dpmi_get_real_mode_interrupt_vector(intno, &rmiv)==0)
      rmaddr = (((unsigned long)rmiv.segment)<<16)+rmiv.offset16;
  }
  #else
  rmaddr = (unsigned long)_dos_getvect(intno);
  #endif

  if (rmaddr)
  {
    char sigbuffer[12]; /* "\0\0\0\PKT DRVR" */
    __my_dosmem_read( sigbuffer, rmaddr, sizeof(sigbuffer) );

    if (memcmp( &sigbuffer[3], "PKT DRVR", 8 ) != 0)
      rmaddr = 0ul;
  }

  if (intno == 0x80)
    is_x80 = (rmaddr ? +1 : 0);
  else if (rmaddr)
    is_there  |= (1ul << (intno - 0x60));
  else
    not_there |= (1ul << (intno - 0x60));

  /* DBGPRINTF(("*<= __is_driver_there()=>%d\n",intno, rmaddr!=0)); */
  return (rmaddr != 0);
}

/* ==================================================================== */
/*                       construct a 'type'                             */
/* ==================================================================== */

/* fill a buffer (must be >= 8 octets long) with a 'type'
** (a 'type' is something that a packet driver demultiplexes on -
** For DIX this is the ether_type, for IEEE 802.x its the LLC+SNAP.)
**
** The PD API requires the 'type' to be in the endian-order and format
** appropriate to each packet-driver class. This is a catch-22 situation
** since the class can only be determined with the driver_info() call, 
** which (prior to 1.08) requires a handle, which can only be obtained 
** with the access_type() call, which requires a type/typelen. *sigh*
** To get around that idiocy, the pdapi_get_driver_info() tries each 
** packet-driver class in turn, with a 'type' formatted for that class.
**
** __fill_proto_type() is also used by pdapi_open_driver() in order
** to not have a dependancy on 'type' endian-order/formatting before 
** the driver is open. Ergo, the 'proto_type' argument to pdapi_open_driver()
** is expected to be a valid _DIX_/_BLUEBOOK_ protocol number in _HOST_ order.
*/
static signed char __fill_proto_type( unsigned short if_class,
                                      unsigned short proto_type,
                                      void FAR *_typebuf )
{
  unsigned short FAR *typebuf = ((unsigned short FAR *)_typebuf); 
  unsigned short be_proto_type = proto_type;
  unsigned short le_proto_type = proto_type;

  /* 1, <2?>, 3, <4>, 5, 6, 7, 8, 9, 10, 11, 12, <13>, 14, 15, 16, 17, 18 */

  if ( (proto_type & 0xff) == *((unsigned char *)&proto_type) )
    le_proto_type = (unsigned short)( ((proto_type & 0xff00)>>8)|
                                      ((proto_type & 0x00ff)<<8) );
  else /* little-endian (ie motorola) */
    be_proto_type = (unsigned short)( ((proto_type & 0xff00)>>8)|
                                      ((proto_type & 0x00ff)<<8) );

  if (if_class == 1 ||  /* DIX ("Bluebook") */ 
      if_class == 14 || /* LANSTAR encapsulating DIX */
      if_class == 7)    /* old starlan (subsumed by DIX) */
  {
    *typebuf  = le_proto_type; /* ETHERTYPE_* in network order */
    return 2; /* two octet type length */
  }
  if (if_class == 2) /* Proteon's ProNET-10 */
  {
    /* http://www.iana.org/assignments/pronet80-type-numbers
    */
    /* IEEE 802.2 on ProNET-10 is Type 23 (decimal) and would 
     * warrant a separate packet driver class, so I assume the
     * driver expects one of the other one-octet types here. 
    */
    if (proto_type == 0x0800)
      proto_type = 1;
    else if (proto_type == 0x0806)
      proto_type = 3;
    else if (proto_type == 0x8137)
      proto_type = 34; /* new (no trailer) type. old (3 byte tr) is type 10 */
    else
      return -1;
    *((unsigned char FAR *)typebuf) = (unsigned char)proto_type;  
    return 1; /* one octet typelen */ 
  }
  if (if_class == 5  || /* AppleTalk - localtalk.asm driver only supports IP */
      if_class == 6)    /* SLIP - IP only (duh) */
  {
    if (proto_type != 0x0800)
      return -1;
    return 0;/* no type length (only handles IP) */
  }
  if (if_class == 9)    /* AX.25 */
  {
    #if 1
    return 0;/* no type length (no multiplexing done by driver) */
    #else
    if (proto_type == 0x0800)
      proto_type = 0xCC;
    else if (proto_type == 0x0806)
      proto_type = 0xCD;
    else  /* all other proto_types are not supported by packet drivers */
      return ((signed char)-1);
    *((unsigned char FAR *)typebuf) = (unsigned char)proto_type;  
    return 1; /* one octet typelen */ 
    #endif
  }
  if (if_class == 10 || /* KISS */
      if_class == 15)   /* SLFP (MIT serial spec) */
  {
    return 0;/* no type length (no multiplexing done by driver) */
  }
  if (if_class == 8) /* ARCnet */
  {
    /* See RFC 1201 (IP over ARCNET) and Crynwr's ARCETHER.ASM - These
     * protocol numbers support variable MTUs upto 60,480 octets.
     * RFC 1051 (which predates RFC1201), uses 240 for IP and 241 for ARP
     * but only permits 256 octet MTU packets.
     * Both RFCs define broadcast address = 0 and ARP hardware type = 7.
     *
     * The ARCNET drivers for Linux adhere to RFC1201. The ARCNET 
     * drivers for Win31 adhere to RFC 1051.
    */
    if (proto_type == 0x0800)
      proto_type = 212; /* IP on ARCnet */
    else if (proto_type == 0x0806)
      proto_type = 213; /* ARP on ARCnet */
    else if (proto_type == 0x0835)
      proto_type = 214; /* RARP on ARCnet */
    else if (proto_type == 0x8137) /* ipx on DIX */
      proto_type = 250; /* IPX on ARCnet */
    else if (proto_type == 0xACCA) /* LANSoft */
      proto_type = 251; /* LANSoft on ARCnet */
    else  /* all other proto_types are not supported by packet drivers */
      return ((signed char)-1);
    *((unsigned char FAR *)typebuf) = (unsigned char)proto_type;  
    return 1; /* one octet typelen */
  }
  if (if_class == 16 || /* Point to Point Protocol (no LCP) */
      if_class == 18)   /* Point to Point Protocol (with LCP) */
  {
    *typebuf  = le_proto_type; /* ETHERTYPE_* in network order */
    return 2; /* two octet typelen - is this correct? */
  }
  if (if_class == 3  || /* 802.5 (Token Ring with verbatim RIF) */
      if_class == 11 || /* 802.3 Ethernet with 802.2 hdrs */
      if_class == 12 || /* FDDI with 802.2 headers */
      if_class == 17)   /* Token Ring with expanded/fixedlen RIF */
  {
    /* IPX note: 
     * #define ETHERTYPE_IPX_8022 0x00e0 - 802.2 DSAP number
     * #define ETHERTYPE_IPX_8023 0x0000 - raw frame (0xffff at top)
     * #define ETHERTYPE_IPX_II	  0x8137 - Ethernet_II (aka DIX)
     * #define ETHERTYPE_IPX_SNAP 0x8137 - Ethernet_SNAP (as below)
    */
    *typebuf++ = 0xAAAA; /* LLC: {{0xAA},{0xAA},{0x03}} */
    *typebuf++ = 0x0003; /* SNAP {{0x00,0x00,0x00},{ethtype}} */
    *typebuf++ = 0x0000;
    *typebuf++ = be_proto_type; /* ETHERTYPE_* in *bit* order */
    return 8; /* eight octet typelen - LLC and SNAP together */
  }
  if (if_class == 0  ||  /* not defined */
      if_class == 4  ||  /* Omninet */
      if_class == 13 ||  /* X.25 */
      if_class > 18)     /* not defined */
  {
    return ((signed char)-1);
  }
  /* WHOA! HOW DID YOU GET HERE? */
  return ((signed char)-1);
}  

/* ==================================================================== */
/*                    PKT API: get driver info                          */
/* ==================================================================== */

#if 0
struct pdapi_driver_info
{
  unsigned short version;   /* bx */
  unsigned char  if_class;  /* ch */
  unsigned short if_type;   /* dx */
  unsigned char  if_number; /* cl */
  unsigned char  if_functionality; /* al */
  char if_name[64];         /* ds:si */
};
#endif

int pdapi_get_driver_info( int intno, unsigned short handle,
                           struct pdapi_driver_info FAR *info )
{
  int pderrno = 0xff; /* not there */

  if (__is_driver_there(intno))
  {
    IREGS regs;
    regs.r_ax = 0x01FF; /* AH == 01h, AL == 255. driver_info*/
    regs.r_bx = handle; /* "optional" for drivers >= 1.07 */

    pderrno = PKT_API(intno, &regs);
    if (pderrno != 0)
    {
      static struct {
        int intno; 
        unsigned short info[6];
      } cache[2]; /* = {{0,0,0,0,0,0,0},{0,0,0,0,0,0,0}}; */
      unsigned int i;

      if (pderrno == PDERR_BAD_HANDLE && (!handle || handle == 0xffff)) /* old PD */ 
      {
        for (i = 0; i < (sizeof(cache)/sizeof(cache[0])); i++)
        {
          if (cache[i].intno == intno)
          {
            regs.r_ax = cache[i].info[0];
            regs.r_bx = cache[i].info[1];
            regs.r_cx = cache[i].info[2];
            regs.r_dx = cache[i].info[3];
            regs.r_ds = cache[i].info[4];
            regs.r_si = cache[i].info[5];
            pderrno = 0;
            break;
          }
        }
      }
      if (pderrno == PDERR_BAD_HANDLE && (!handle || handle == 0xffff)) /* old PD */ 
      {
        unsigned short cmddata[8+5]; /* max 16 bytes 'type' + 10 bytes code */
        unsigned long rmbuf = __my_dosmem_alloc(sizeof(cmddata));
        pderrno = PDERR_NO_SPACE;

        if (rmbuf)
        { 
          unsigned char nclass;

          cmddata[0] = 0xc00b; /* or ax,ax */
          cmddata[1] = 0x0475; /* jnz +4 */
          cmddata[2] = 0xff33; /* xor di,di */
          cmddata[3] = 0xc78e; /* mov es,di */
          cmddata[4] = 0x90cb; /* retf; nop */

          /* pre-1.07 specification only defined classes 1 through 8 */
          for (nclass = 1; nclass <= 8; nclass++)
          {
            /* first try without a typelength, then with.
             * A zero-length typlen implies "match all requests", which
             * may be refused by packet drivers developed to conform 
             * to versions of this spec previous to 1.07
            */
            unsigned char ntype;
            for (ntype = 0; ntype < 3; ntype++)
            {
              signed char typelen = 0;

              if (ntype != 0) /* try specific type */
              {
                if (ntype == 1) /* try proto_type == ETHERTYPE_IP */
                  typelen = __fill_proto_type(nclass, 0x0800, &cmddata[5]);
                if (ntype == 2) /* try proto_type == ETHERTYPE_LOOPBACK */
                  typelen = __fill_proto_type(nclass, 0x9000, &cmddata[5]);
                if (typelen < 1) /* (-1/0) already done this */
                  break;
              }

              __my_dosmem_write( rmbuf, cmddata, 10 + typelen);

              regs.r_ax = 0x200; /* access_type() */
              regs.r_ax |= nclass;
              regs.r_bx = 0xffff; /* any if_type */ 
              regs.r_dx = 0;      /* first interface */
              regs.r_cx = typelen;
              /* es:di = segoff to recvr code */
              regs.r_es = (unsigned short)(rmbuf >> 16);    
              regs.r_di = (unsigned short)(rmbuf & 0xffff);
              /* ds:si = segoff to filled eth_type */
              regs.r_ds = regs.r_es;
              regs.r_si = (unsigned short)(regs.r_di + 10);
     
              pderrno = PKT_API(intno, &regs);

              if (pderrno == 0) /* opened driver ok */      
              {   
                IREGS regs2;
                regs2.r_ax = 0x0300; /* release_type() */
                regs2.r_bx = regs.r_ax; /* handle */

                regs.r_bx = regs.r_ax;
                regs.r_ax = 0x01FF; /* AH == 01h, AL == 255. driver_info*/
                pderrno = PKT_API(intno, &regs);

                PKT_API(intno, &regs2); /* release the handle */

                if (pderrno == 0)
                {
                  for (i = 0; i < (sizeof(cache)/sizeof(cache[0])); i++)
                  {
                    if (cache[i].intno == 0)
                    {
                      cache[i].intno   = intno;
                      cache[i].info[0] = regs.r_ax;
                      cache[i].info[1] = regs.r_bx;
                      cache[i].info[2] = regs.r_cx;
                      cache[i].info[3] = regs.r_dx;
                      cache[i].info[4] = regs.r_ds;
                      cache[i].info[5] = regs.r_si;
                      break;
                    }
                  }
                } /* if (got info) */
                break; /* for for (ntype = 0; ... ) */
              } /* if (opened driver ok) */

              if (pderrno != PDERR_TYPE_INUSE && 
                  pderrno != PDERR_BAD_TYPE && pderrno != PDERR_NO_TYPE)
                break; /* for (ntype = 0; ... )  */

            } /* for (ntype = 0; ntype < 3; ntype++) */

            if (pderrno != PDERR_NO_CLASS) /* 0 or error eg, PDERR_TYPE_INUSE */
              break; /* for (nclass = 0; ... )  */
          } /* for (nclass = 0; nclass <= 18; nclass++) */

          __my_dosmem_free(rmbuf);
        } /* if (rmbuf) */
      } /* if (pderrno == PDERR_BAD_HANDLE && (!handle) */ /* old PD */         
      if (pderrno != 0)
        pderrno = PDERR_BAD_COMMAND;
    } /* if (pderrno != 0) */

    if (pderrno == 0)
    {
      unsigned long segoff;
      info->drvr_functionality = (unsigned char)(regs.r_ax & 0xff);
      info->drvr_version = regs.r_bx;
      info->if_class     = (unsigned char)(regs.r_cx >> 8);
      info->if_typ       = regs.r_dx;
      info->if_number    = (unsigned char)(regs.r_cx & 0xff);
      segoff = ((((unsigned long)(regs.r_ds))<<16)+regs.r_si);
      __my_dosmem_read( &info->drvr_name[0], segoff, sizeof(info->drvr_name));
      info->drvr_name[sizeof(info->drvr_name)-1] = '\0';
    }
  }
  else if (intno == NULL_DRIVER_INTNO) /* nullpkt driver */
  {
    info->drvr_version = 1;
    info->if_class  = NULL_DRIVER_CLASS; /* 1,2,3,6 or 8 */
    info->if_typ    = 0;
    info->if_number = 0;
    info->drvr_name[0] = 'n';
    info->drvr_name[1] = 'u';
    info->drvr_name[2] = 'l';
    info->drvr_name[3] = 'l';
    info->drvr_name[4] = 'p';
    info->drvr_name[5] = 'k';
    info->drvr_name[6] = 't';
    info->drvr_name[7] = '\0';
    info->drvr_functionality = 2; /* basic and extended functions present */
    pderrno = 0;
  }

  return pderrno;
}
  
/* ==================================================================== */
/*                    PKT API: get/set MAC address                      */
/* ==================================================================== */

int pdapi_get_address( int intno, unsigned short handle, 
                       void FAR *addrbuf, unsigned int FAR *addrlen )
{
  int pderrno = 0xff; /* not there */

  if (__is_driver_there(intno))
  {
    unsigned long rmbuf;
    IREGS regs;

    pderrno = PDERR_NO_SPACE;
    rmbuf = __my_dosmem_alloc(*addrlen);
    if (rmbuf)
    {
      regs.r_ax = 0x600; /* AH == 6, getaddress */
      regs.r_bx = handle;
      regs.r_es = (unsigned short)(rmbuf >> 16);
      regs.r_di = (unsigned short)(rmbuf & 0xffff);
      regs.r_cx = (unsigned short)(*addrlen);
      if (*addrlen & ~0xffffu)
        regs.r_cx = 0xffffu;
      pderrno = PKT_API(intno, &regs);
      if (pderrno == 0)
      {
        if ((regs.r_cx & 0xffffu) <= *addrlen)
        {
          *addrlen = (regs.r_cx & 0xffffu); 
          __my_dosmem_read( addrbuf, rmbuf, *addrlen );
        }
      }
      __my_dosmem_free(rmbuf);
    }
  }
  else if (intno == NULL_DRIVER_INTNO) /* null pkt driver */
  {
    /* 0x00005E is IANA non-public reserved. */
    /* Similar to 0x444553 for Windows95 internal "adapters" */
    static unsigned char FAR dummy[6] = {0x00,0x00,0x00,0x00,0x00,0x00};
    struct pdapi_driver_params params;
    if (!dummy[2])
    {
      char junk[sizeof(dummy)]; /* pick up junk from the stack */
      junk[0] = 0x00;
      junk[1] = 0x00;
      junk[2] = 0x5e;  
      FAR_MEMCPY( &dummy[0], &junk[0], sizeof(junk) );
    }
    pderrno = pdapi_get_driver_params( intno, &params );
    if (pderrno == 0)
    {
      pderrno = PDERR_NO_SPACE;
      if (*addrlen >= params.addr_len && params.addr_len <= sizeof(dummy))
      {
        *addrlen = params.addr_len;
        FAR_MEMCPY( addrbuf, &dummy[sizeof(dummy) - *addrlen], *addrlen );
        pderrno = 0;
      }
    }
  }
  return pderrno;
}

int pdapi_set_address( int intno, const void FAR *addrbuf, 
                                           unsigned int addrlen )
{                                         /* extended driver function */
  int pderrno = 0xff; /* not there */

  if (addrlen & ~0xffffu)
  {
    pderrno = PDERR_CANT_SET;
  }
  else if (__is_driver_there(intno))
  {
    IREGS regs;
    unsigned long rmbuf;
    pderrno = PDERR_CANT_SET;

    rmbuf = __my_dosmem_alloc(addrlen);
    if (rmbuf)
    {
      __my_dosmem_write( rmbuf, addrbuf, addrlen );
      regs.r_ax = 0x2500; /* AH == 25, setaddress */
      regs.r_es = (unsigned short)(rmbuf >> 16);
      regs.r_di = (unsigned short)(rmbuf & 0xffff);
      regs.r_cx = (unsigned short)addrlen;

      pderrno = PKT_API(intno, &regs);
      if (pderrno == PDERR_BAD_COMMAND)
        pderrno = PDERR_CANT_SET;

      __my_dosmem_free(rmbuf);
    }
  }
  else if (intno == NULL_DRIVER_INTNO) /* null pkt driver */
  {
    pderrno = PDERR_CANT_SET;
  }
  return pderrno;
}

/* ==================================================================== */
/*                    PKT API: set/get recv mode                        */
/* ==================================================================== */

/*
 * Where 'mode' is one of the following:
 *            1       turn off receiver
 *            2       receive only packets sent to this interface
 *            3       mode 2 plus broadcast packets <default>
 *            4       mode 3 plus limited multicast packets
 *            5       mode 3 plus all multicast packets
 *            6       all packets ("promiscuous mode")
*/
int pdapi_get_recv_mode( int intno, unsigned short handle, 
                                             int FAR *mode )
{                                       /* extended driver function */
  int pderrno = 0xff; /* not there */

  if (__is_driver_there(intno))
  {  
    IREGS regs;
    regs.r_ax = 0x2100; /* AH == 21h, get_rcv_mode */
    regs.r_bx = handle;

    pderrno = PKT_API(intno, &regs);
    if (pderrno == 0)
    {
      *mode = regs.r_ax;
    }
  }
  else if (intno == NULL_DRIVER_INTNO) /* null pkt driver */
  {
    *mode = 1; /* receive only packets sent to this interface */
    pderrno = 0;
  }
  return pderrno;
} 

int pdapi_set_recv_mode( int intno, unsigned short handle, int mode )
{                                       /* extended driver function */
  int pderrno = 0xff; /* not there */

  if (__is_driver_there(intno))
  {  
    pderrno = PDERR_BAD_MODE;
    if (mode && (mode & ~0xffffu) == 0)
    {
      IREGS regs;
      regs.r_ax = 0x2000; /* AH == 20h, set_rcv_mode */
      regs.r_bx = handle;
      regs.r_cx = (unsigned short)mode;
      pderrno = PKT_API(intno, &regs);
    }
  }
  else if (intno == NULL_DRIVER_INTNO) /* null pkt driver */
  {
    pderrno = PDERR_BAD_MODE;
    if (mode == 0 || mode == 1)
      pderrno = 0;
  }
  return pderrno;
} 

/* ==================================================================== */
/*                      PKT API: get/set multicast                      */
/* ==================================================================== */

int pdapi_get_multicast_list(int intno, void FAR *addrlstbuf, 
                                    unsigned int FAR *buflen )
{                                       /* extended driver function */
  int pderrno = 0xff; /* not there */

  if (__is_driver_there(intno))
  {  
    IREGS regs;
    regs.r_ax = 0x2300; /* AH == 23, get_multicast_list */

    pderrno = PKT_API(intno, &regs);
    if (pderrno == 0)
    {
      pderrno = PDERR_NO_SPACE;
      if (*buflen >= regs.r_cx)
      {
        unsigned long segoff = (((unsigned long)regs.r_es)<<16)+regs.r_di;
        *buflen = regs.r_cx;
        __my_dosmem_read( addrlstbuf, segoff, *buflen );
        pderrno = 0;
      }
    }
  }
  else if (intno == NULL_DRIVER_INTNO) /* nullpkt driver */
  {
    pderrno = PDERR_NO_MULTICAST;
  }
  return pderrno;
}

int pdapi_set_multicast_list(int intno, const void FAR *addrlstbuf, 
                                                 unsigned int buflen )
{                                       /* extended driver function */
  int pderrno = 0xff; /* not there */

  if (buflen & ~0xffffu)
  {
    pderrno = PDERR_NO_SPACE;
  }
  else if (__is_driver_there(intno))
  {  
    unsigned long rmbuf;
    IREGS regs;

    pderrno = PDERR_NO_SPACE;
    rmbuf = __my_dosmem_alloc(buflen);
    if (rmbuf)
    {
      __my_dosmem_write( rmbuf, addrlstbuf, buflen );
      regs.r_ax = 0x2200; /* AH == 22, set_multicast_list */
      regs.r_es = (unsigned short)(rmbuf >> 16);
      regs.r_di = (unsigned short)(rmbuf & 0xffff);
      regs.r_cx = (unsigned short)(buflen);
      pderrno = PKT_API(intno, &regs);

      __my_dosmem_free(rmbuf);
    }
  }  
  else if (intno == NULL_DRIVER_INTNO) /* nullpkt driver */
  {
    pderrno = PDERR_NO_MULTICAST;
  }
  return pderrno;
}

/* ==================================================================== */
/*                      packet driver upcall spool                      */
/* ==================================================================== */

/* Don't use the DPMI versions - they don't work as expected,
 * and since the interrupt safe stuff will only ever get called
 * under real DOS, we don't have to diddle with the "virtual" funcs.
*/
#if 0 /* defined(__DPMI32__) */
#define __are_interrupts_enabled __dpmi_get_virtual_interrupt_state
#define __cli __dpmi_get_and_disable_virtual_interrupt_state
#define __sti __dpmi_get_and_enable_virtual_interrupt_state
#elif defined(__GNUC__)
static __inline int __are_interrupts_enabled(void)
{
  int state;
  __asm__ __volatile__("pushfl; popl %0; andl $0x200,%0;" : "=r"(state));
  return state;
}
#define __sti() __asm__ __volatile__("sti")
#define __cli() __asm__ __volatile__("cli")
#elif defined(__WATCOMC__)
extern unsigned short __are_interrupts_enabled(void);
extern void __cli(void);
extern void __sti(void);
#pragma aux __are_interrupts_enabled = "pushf" "pop ax" "and ax,200h" \
                                       parm [] value [ax] \
                                       modify exact [ax];
#pragma aux __cli = "cli" parm [] modify exact [];
#pragma aux __sti = "sti" parm [] modify exact [];
#elif defined(__TURBOC__)
static int __are_interrupts_enabled(void)
{
  __asm pushf
  __asm pop ax
  __asm and ax,200h
  return _AX;
}
#define __cli() __emit__(0xFA)
#define __sti() __emit__(0xFB)
#endif
#define __hlt() for (;;) __cli() /* when cannot continue safely */

/* ------------------------------------------------------------------ */

struct recvr_info_t
{
  struct recvr_info_t FAR *next;
  int intno;

  struct
  {
    unsigned long code_segoff;
    unsigned long no_buffs_ctr;

    int stop_queue_growth;
    unsigned short numqueues;
    struct 
    {
      unsigned short sel;
      unsigned long segoff;
      unsigned short size;
    } queues[16];

    #if defined(__DPMI32__) /* only needed when _dos_ds != _my_ds() */
    void *tmp_buf;          /* temporary buffer for manipulating rm queue. */
    unsigned short tmp_buf_size;/*size of buff (>=size of 1st/largest queue)*/
    #endif
#undef FORCE_COMMON_PMCODE /* pretend _dos_ds is always != _my_ds() */

  } cb;

  struct
  {
    unsigned short handle;
    unsigned short proto;
    int (*rcvr)( int intno, unsigned short handle, 
                 unsigned int len, const void FAR *buf );
  } handle_tab[4];
  unsigned int handle_count;  
};

static struct 
{
  unsigned long magiczero;
  struct recvr_info_t FAR *head;
} recvrlist = {0UL, (struct recvr_info_t FAR *)0 };

#define __get_receiver_head() recvrlist.head

/* ------------------------------------------------------------------ */

#pragma pack(1)
struct queue_hdr_t 
{
  unsigned long  next; /* real mode address, ie 'struct queue_hdr_t FAR *' */
  unsigned short size; /* size includes sizeof(struct queue_hdr_t) */
  unsigned short used; /* used does not include sizeof(header) */
};
#pragma pack()

/* ------------------------------------------------------------------ */

/* returns difference between old numqueues and new, ergo 0 if failed. */
static int __grow_queues( struct recvr_info_t FAR *rit )
{
  /*
  ** WARNING, WARNING, WARNING. Do not modify the wrapper without
  ** reading the notes for __rit_reserve_pktbuf(). 
  */
  static const unsigned char wrapper[] = {
  /*0000*/ 0x9C,                       /* pushf                       */
  /*0001*/ 0xFA,                       /* cli                         */
  /*0002*/ 0x55,                       /* push      bp                */
  /*0003*/ 0x89,0xD2,                  /* mov       dx,dx             */
  /*0005*/ 0xE8,0x10,0x00,             /* call      X$191             */
  /* --------- the order/placement of data fields is critical ------- */
  /*0008*/ 0x00,0x00,0x00,0x00,        /* queue_head: dd 0            */
  /*000C*/ 0x00,0x00,0x00,0x00,        /* count_no_buffs: dd 0        */
  /*0010*/ 0x00,0x00,0x00,0x00,        /* count_bad_len: dd 0         */
  /*0014*/ 0x00,0x00,0x00,0x00,        /* count_added_ok: dd 0        */
  /* --------- ---------------------------------------------- ------- */
  /*0018*/                             /* X$191:                      */
  /*0018*/ 0x5D,                       /* pop       bp                */
  /*0019*/ 0x09,0xC0,                  /* or        ax,ax             */
  /*001B*/ 0x75,0x61,                  /* jne       X$194             */
  /*001D*/ 0x8C,0xC8,                  /* mov       ax,cs             */
  /*001F*/ 0x8E,0xC0,                  /* mov       es,ax             */
  /*0021*/ 0x2E,0x8D,0x7E,0x00,        /* lea       di,[bp]           */
  /*0025*/ 0x83,0xC1,0x04,             /* add       cx,0x0004         */
  /*0028*/ 0x83,0xF9,0x18,             /* cmp       cx,0x0018         */
  /*002B*/ 0x77,0x09,                  /* ja        X$192             */
  /*002D*/ 0x31,0xFF,                  /* xor       di,di             */
  /*002F*/ 0x8E,0xC7,                  /* mov       es,di             */
  /*0031*/ 0xB8,0x01,0x00,             /* mov       ax,0x0001         */
  /*0034*/ 0xEB,0x32,                  /* jmp       X$193             */
  /*0036*/                             /* X$192:                      */
  /*0036*/ 0x26,0xC4,0x3D,             /* les       di,es:[di]        */
  /*0039*/ 0x8C,0xC0,                  /* mov       ax,es             */
  /*003B*/ 0x09,0xF8,                  /* or        ax,di             */
  /*003D*/ 0x74,0x29,                  /* je        X$193             */
  /*003F*/ 0x26,0x8B,0x45,0x04,        /* mov       ax,es:[di+0x4]    */
  /*0043*/ 0x83,0xE8,0x08,             /* sub       ax,0x0008         */
  /*0046*/ 0x26,0x2B,0x45,0x06,        /* sub       ax,es:[di+0x6]    */
  /*004A*/ 0x39,0xC8,                  /* cmp       ax,cx             */
  /*004C*/ 0x72,0xE8,                  /* jb        X$192             */
  /*004E*/ 0x26,0x8B,0x45,0x06,        /* mov       ax,es:[di+0x6]    */
  /*0052*/ 0x26,0x01,0x4D,0x06,        /* add       es:[di+0x6],cx    */
  /*0056*/ 0x83,0xC7,0x08,             /* add       di,0x0008         */
  /*0059*/ 0x01,0xC7,                  /* add       di,ax             */
  /*005B*/ 0x26,0x89,0x0D,             /* mov       es:[di],cx        */
  /*005E*/ 0x26,0x89,0x5D,0x02,        /* mov       es:[di+0x2],bx    */
  /*0062*/ 0x83,0xC7,0x04,             /* add       di,0x0004         */
  /*0065*/ 0xB8,0x02,0x00,             /* mov       ax,0x0002         */
  /*0068*/                             /* X$193:                      */
  /*0068*/ 0x40,                       /* inc       ax                */
  /*0069*/ 0xD1,0xE0,                  /* shl       ax,0x0001         */
  /*006B*/ 0xD1,0xE0,                  /* shl       ax,0x0001         */
  /*006D*/ 0x01,0xC5,                  /* add       bp,ax             */
  /*006F*/ 0x2E,0x83,0x46,0x00,0x01,   /* add       word ptr cs:[bp],0x0001*/
  /*0074*/ 0x2E,0x83,0x56,0x02,0x00,   /* adc       word ptr cs:[bp+0x2],0x0000*/
  /*0079*/ 0x83,0xE9,0x04,             /* sub       cx,0x0004         */
  /*007C*/ 0x31,0xC0,                  /* xor       ax,ax             */
  /*007E*/                             /* X$194:                      */
  /*007E*/ 0x5D,                       /* pop       bp                */
  /*007F*/ 0x9D,                       /* popf                        */
  /*0080*/ 0xCB                        /* retf                        */
  };
  unsigned short left_to_alloc, min_queue_size;
  unsigned int i, prev_numqueues = rit->cb.numqueues;
  int is_sti = 0;

  /* The initial size of the queue(s) is 32K(+1K). If 32K proves to be 
  ** insufficient, __rit_poll() will call __grow_queues() again,
  ** in which case the grow-increment will be USHRT_MAX (or as
  ** much of that as possible).
  **
  ** The logic here is that 32K(+1K) will generally be sufficient, but
  ** should that not be the case, then we need to get >64K quickly.
  */ 
  left_to_alloc = (33u * 1024u);
  min_queue_size = (4u * 1024u);

  if (rit->intno == NULL_DRIVER_INTNO && !__is_driver_there(rit->intno))
  {
    /* The params stuff here is only required for the nullpkt driver in order
     * to have the most efficient possible queues for full-MTU-sized packets 
     * on the loopback. For a physical adapter, such optimization would be
     * meaningless since the packet sizes are varied (and generally small).
    */
    struct pdapi_driver_params params;
    if (pdapi_get_driver_params(rit->intno, &params) != 0)
      params.ll_mtu = 1514u; /* get_driver_params shouldn't fail for nullpkt */
    else if (params.ll_mtu > ((5u*1024u)-4u))
      params.ll_mtu = ((5u*1024u)-4u); /* should never happen for nullpkt */
    min_queue_size = (unsigned short)(8u + ((4u + params.ll_mtu)<<1));
    left_to_alloc  = (unsigned short)(8u + ((4u + params.ll_mtu)<<3));
  }
  else if (prev_numqueues != 0) /* not first queue allocation */
  {
    /* interrupt en/dis is not required for the initial allocation */
    is_sti = __are_interrupts_enabled();
  }

#if defined(SHOW_QUEUE_GROW)
  _printk("G+ (%u => ", prev_numqueues );
#endif

  while (left_to_alloc > 0 &&
    rit->cb.numqueues < (sizeof(rit->cb.queues)/sizeof(rit->cb.queues[0])) )
  {
    struct queue_hdr_t qhdr;
    unsigned short this_qsize;
    unsigned long segoff;
    int sel = 0;

    this_qsize = left_to_alloc;
    if (rit->cb.numqueues > prev_numqueues)
      this_qsize = rit->cb.queues[prev_numqueues].size;
    #if defined(__DPMI32__) /* all queues must be <= sizeof(first queue) */
    if (rit->cb.tmp_buf && this_qsize > rit->cb.tmp_buf_size)
      this_qsize = rit->cb.tmp_buf_size;
    #endif

    segoff = 0UL;
    while (this_qsize > min_queue_size)
    {
      unsigned short to_alloc = this_qsize;
      if (rit->cb.numqueues == 0) /* first queue includes wrapper */
        to_alloc += (unsigned short)((sizeof(wrapper)+0x0f) & 0xfff0);

      #if !defined(__DPMI32__)
      segoff = (unsigned long)FAR_MALLOC(to_alloc);
      #else
      {
        int seg = __dpmi_allocate_dos_memory( (to_alloc+15) >> 4, &sel );
        if (seg != -1)
          segoff = ((unsigned long)seg)<<16;
      }
      #endif

      if (segoff != 0)
      {
        if (rit->cb.numqueues == 0)
          rit->cb.code_segoff = segoff + this_qsize;
        break;
      }

      this_qsize -= 1024;
    } 

    if (!segoff)
      break;

    left_to_alloc -= this_qsize;

    i = rit->cb.numqueues++;
    rit->cb.queues[i].sel = (unsigned short)sel;
    rit->cb.queues[i].segoff = segoff;
    rit->cb.queues[i].size = this_qsize;

    qhdr.next = 0UL;
    qhdr.size = this_qsize;
    qhdr.used = 0;
    __my_dosmem_write( segoff, &qhdr, sizeof(qhdr) );

    if (i != 0)
    {
      if (is_sti)
        __cli();
      __my_dosmem_write(rit->cb.queues[i-1].segoff, &segoff, sizeof(segoff));
      if (is_sti)
        __sti();
    }
    else
    {
      __my_dosmem_write( rit->cb.code_segoff, wrapper, sizeof(wrapper));
      __my_dosmem_write( rit->cb.code_segoff + 8, &segoff, sizeof(segoff));
    }

  } /* while (left_to_alloc > 0) */

  #if defined(__DPMI32__)
  if (prev_numqueues == 0 &&    /* was first queue allocation */
      #if !defined(FORCE_COMMON_PMCODE)
      _dos_ds != _my_ds() &&    /* not zero-based addressing */
      #endif
      rit->cb.numqueues != 0)   /* first queue alloc did not fail */
  {
    rit->cb.tmp_buf_size = rit->cb.queues[0].size;
    rit->cb.tmp_buf = malloc(rit->cb.tmp_buf_size);
    if (rit->cb.tmp_buf)
    {
      if (__my_dpmi_lock_data(rit->cb.tmp_buf, rit->cb.tmp_buf_size) != 0)
      {
        free(rit->cb.tmp_buf);
        rit->cb.tmp_buf = (void *)0;
      }
    }
    if (!rit->cb.tmp_buf)
    {       
      for (i = 0; i < rit->cb.numqueues; i++)
        __dpmi_free_dos_memory( rit->cb.queues[i].sel );          

      rit->cb.numqueues = 0;
      rit->cb.code_segoff = 0;
      rit->cb.tmp_buf_size = 0;
    }
  }
  #endif

#if defined(SHOW_QUEUE_GROW)
  _printk("%u)", rit->cb.numqueues );
  for (i = prev_numqueues; i < rit->cb.numqueues; i++)
     _printk(", 0x%08lx (%u)", rit->cb.queues[i].segoff, 
                               rit->cb.queues[i].size);
  _printk(" G-\n" );
#endif

  return (rit->cb.numqueues - prev_numqueues);
}

/* ------------------------------------------------------------------ */

/* "Bind" (automatically alloc callback if none existing yet).
 * If 'handleP' is not NULL, bind '*handleP' to the callback.
 *
 * On success: Returns the code address of the callback.
 * On failure: Returns 0UL - either no mem, or no more room for handles.
 *             (either case is a PDERR_NO_SPACE)
*/
static unsigned long __rit_bind(int intno, 
                                const unsigned short FAR *handleP,
                                unsigned short proto,
                                int (*rcvr)( int intno, 
                                             unsigned short handle, 
                                   unsigned int len, const void FAR *buf ) )
{
  struct recvr_info_t FAR *rit;
  unsigned int i;

  rit = __get_receiver_head();
  if (!rit && recvrlist.head)
    return 0UL;

  while (rit)
  {  
    if (rit->intno == intno)
      break;
    rit = rit->next;
  }

  if (!rit)
  {
    rit = (struct recvr_info_t FAR *)FAR_MALLOC(sizeof(*rit));

    #if defined(__DPMI32__)
    if (rit)
    {
      if (__my_dpmi_lock_data(rit, sizeof(*rit)) != 0)
      {
        FAR_FREE(rit);
        rit = (struct recvr_info_t FAR *)0;
      }
      else if (!recvrlist.head) /* will be the first? */
      {
        if (__my_dpmi_lock_data(&recvrlist, sizeof(recvrlist)) != 0)
        {
          __my_dpmi_unlock_data(rit, sizeof(*rit));
          FAR_FREE(rit);
          rit = (struct recvr_info_t FAR *)0;
        }
      }
    }
    #endif

    if (rit)
    {
      FAR_MEMSET(rit, 0, sizeof(*rit));
      rit->intno = intno; /* intno is needed by __grow_queues() */

      if (__grow_queues( rit ) > 0) /* everything ok */
      {
        rit->next = recvrlist.head;
        recvrlist.head = rit;
      }
      else /* grow failed */
      {
        #if defined(__DPMI32__)
        __my_dpmi_unlock_data(rit, sizeof(*rit));
        if (!recvrlist.head) /* would have been the first? */
          __my_dpmi_unlock_data(&recvrlist, sizeof(recvrlist));
        #endif

        FAR_MEMSET(rit, 0, sizeof(*rit));
        FAR_FREE(rit); 
        rit = (struct recvr_info_t FAR *)0;
      }

    } /* if (rit allocated ok) */
  } /* if (!rit) */

  if (rit)
  {
    /* Note that we fail even if not adding a handle NOW. */
    if (rit->handle_count >=
        (sizeof(rit->handle_tab)/sizeof(rit->handle_tab[0])))
    {
      rit = (struct recvr_info_t FAR *)0;
    }
    else if (handleP)
    {
      i = rit->handle_count++;
      rit->handle_tab[i].rcvr = rcvr;
      rit->handle_tab[i].handle = *handleP;
      rit->handle_tab[i].proto = proto;
    }
  }

  return (rit ? rit->cb.code_segoff : 0UL);
}

/* ------------------------------------------------------------------ */

/* "Unbind" (automatically free callback when no handles remaining).
 *
 * If 'handleP' is not NULL, unbind '*handleP' from the callback.
 * If no handles remain bound, free the callback,
 * 
 * Return 0 on success, else pderrno (PDERR_BAD_HANDLE).
*/
static int __rit_unbind( int intno, const unsigned short *handleP )
{
  struct recvr_info_t FAR *rit_prev = (struct recvr_info_t FAR *)0;
  struct recvr_info_t FAR *rit = __get_receiver_head();
  unsigned int i;

  while (rit)
  {  
    if (rit->intno == intno)
      break;
    rit_prev = rit;
    rit = rit->next;
  }

  if (handleP)
  {         
    unsigned int n = 0; i = 0; 
    if (rit)
    {    
      n = rit->handle_count;
      for (i = 0; i < n; i++)
      {
        if (rit->handle_tab[i].handle == *handleP)
        {
          if (__is_driver_there(intno))
          {
            IREGS regs;
            regs.r_ax = 0x2000; /* AH == 20h, set_rcv_mode */
            regs.r_bx = *handleP;
            regs.r_cx = 1; /* turn off receiver */
            PKT_API(intno, &regs);
          }
          if (n != 1)
            FAR_MEMMOVE( &(rit->handle_tab[i]), &(rit->handle_tab[i+1]),
                         (n - (i+1)) * sizeof(rit->handle_tab[i]) );
          rit->handle_count--;
          break;
        }
      }
      i = rit->handle_count;
    }
    if (n == i)
      return PDERR_BAD_HANDLE;
  }

  if (rit)
  {
    if (rit->handle_count == 0)
    {
      int is_sti = __are_interrupts_enabled();

      if (is_sti)
        __cli();
      if (rit->cb.code_segoff)
      {
        unsigned short rcvr_nop[5];
        rcvr_nop[0] = 0xc00b; /* or ax,ax */
        rcvr_nop[1] = 0x0475; /* jnz +4 */
        rcvr_nop[2] = 0xff33; /* xor di,di */
        rcvr_nop[3] = 0xc78e; /* mov es,di */
        rcvr_nop[4] = 0x90cb; /* retf; nop */
        __my_dosmem_write( rit->cb.code_segoff, rcvr_nop, sizeof(rcvr_nop));
      }
      if (rit_prev)
        rit_prev->next = rit->next;
      else
        recvrlist.head = rit->next; 
      if (is_sti)
        __sti();
   
      #if defined(__DPMI32__)
      {  
        if (!recvrlist.head) /* all removed? */
          __my_dpmi_unlock_data(&recvrlist, sizeof(recvrlist.head));

        for (i = 0; i < rit->cb.numqueues; i++)
          __dpmi_free_dos_memory( rit->cb.queues[i].sel );          

        if (rit->cb.tmp_buf) /* not zero-base addressing */
        {
          __my_dpmi_unlock_data( rit->cb.tmp_buf, rit->cb.tmp_buf_size );
          free( rit->cb.tmp_buf );  
          rit->cb.tmp_buf = (unsigned char *)0;
          rit->cb.tmp_buf_size = 0;
        }

        __my_dpmi_unlock_data(rit, sizeof(*rit));
      }
      #else
      {
        for (i = 0; i < rit->cb.numqueues; i++)
          FAR_FREE((void FAR *)rit->cb.queues[i].segoff);
      }
      #endif

      FAR_MEMSET(rit, 0, sizeof(*rit));
      FAR_FREE(rit);
    }
  }

  return 0;
}

/* ------------------------------------------------------------------ */

/* __rit_reserve_pktbuf(): Reserve space in a receiver's queue (for loopback)
** Returns segoff, or 0xfffffffful (PDERR_CANT_SEND) or 0x0 (PDERR_NO_SPACE)
**
** WARNING!: This function assumes that the receiver has certain 
** characteristics. a) It does not deal with the ax=1 call, b) it disables
** interrupts while looking for room, c) it preserves all regs except es:di
** All these apply to the wrapper created in __grow_queues(), but that
** may not be so for other upcall handlers.
*/
static unsigned long __rit_reserve_pktbuf( struct recvr_info_t FAR *rit,
                                           unsigned short handle,
                                           unsigned int buflen )
{
  unsigned long bufsegoff = 0xfffffffful;
  if (rit->cb.code_segoff)
  {
#if defined(__DPMI32__) /* dpmi targets (which includes pharlap) */
    __dpmi_regs regs;
    memset( &regs, 0, sizeof(regs) );

    regs.x.cx = (unsigned short)(buflen);
    regs.x.bx = handle;
    regs.x.ax = 0;
    regs.x.cs = (unsigned short)(rit->cb.code_segoff >> 16);
    regs.x.ip = (unsigned short)(rit->cb.code_segoff & 0xffff);

    if (__dpmi_simulate_real_mode_procedure_retf(&regs) == 0) /* else -1UL */
      bufsegoff = (((unsigned long)regs.x.es)<<16)+regs.x.di;
#else
    unsigned short x_seg = (unsigned short)(rit->cb.code_segoff >> 16);
    unsigned short x_off = (unsigned short)(rit->cb.code_segoff & 0xffff);

    __asm mov  dx, x_seg
    __asm mov  ax, x_off
    __asm push es
    __asm push di
    __asm push bp
    __asm push dx
    __asm push ax
    __asm mov  bx, handle
    __asm mov  cx, buflen
    __asm xor  ax, ax
    __asm mov  bp, sp
    __asm call dword ptr [bp]
    __asm mov  dx,es
    __asm mov  ax,di
    __asm add  sp,4
    __asm pop  bp
    __asm pop  di
    __asm pop  es
    __asm mov  x_off, ax
    __asm mov  x_seg, dx
    bufsegoff = (((unsigned long)x_seg)<<16)+x_off;
#endif
  }
  return bufsegoff;
}

/* ------------------------------------------------------------------ */

/* __rit_do_downcall(): pass a packet from the application to the queue 
** returns PDERR_BAD_HANDLE, PDERR_CANT_SEND, PDERR_NO_SPACE or zero.
*/
static int __rit_do_downcall( struct recvr_info_t FAR *rit, 
                              unsigned short handle,
                              const void FAR *buffer, unsigned int buflen )
{
  int pderrno = PDERR_BAD_HANDLE; /* not matched */
  unsigned int i, n = rit->handle_count;

  for (i = 0; i < n; i++)
  {
    if (rit->handle_tab[i].handle == handle)
    {
      unsigned long bufsegoff;

      bufsegoff = __rit_reserve_pktbuf(rit, handle, buflen );
      if (bufsegoff == 0UL)
      {
        if (__grow_queues(rit) > 0)
          bufsegoff = __rit_reserve_pktbuf(rit, handle, buflen);
      }
      if (bufsegoff == 0xfffffffful) /* communication error */
        pderrno = PDERR_CANT_SEND;
      else if (bufsegoff == 0ul)
        pderrno = PDERR_NO_SPACE;
      else
      {
        __my_dosmem_write( bufsegoff, buffer, buflen ); 
        pderrno = 0;
      }
      break;
    }
  }
  return pderrno;
}

/* ------------------------------------------------------------------ */

/* __rit_do_upcall(): pass a packet from the queue to the application.
** returns PDERR_BAD_HANDLE, PDERR_CANT_SEND, or zero.
*/ 
static int __rit_do_upcall(struct recvr_info_t FAR *rit, 
                            unsigned short handle, 
                            const void FAR *buffer, unsigned int buflen)
{                        
  int pderrno = PDERR_BAD_HANDLE; /* not matched */
  unsigned int i, n = rit->handle_count;

  for (i = 0; i < n; i++)
  {
    if (rit->handle_tab[i].handle == handle)
    {
      int purge;
      purge = (*(rit->handle_tab[i].rcvr))(rit->intno, handle, buflen, buffer);

      pderrno = PDERR_CANT_SEND; 
      if (purge)
        pderrno = 0;
      break;
    }
  }
  return pderrno;
}

/* ==================================================================== */
/*                      PKT API: send packet                            */
/* ==================================================================== */

int pdapi_send_pkt( int intno, const unsigned short FAR *lb_handle,
                    const void FAR *buffer, unsigned int buflen)
{
  int pderrno = 0xff; /* not there */

  if (buflen > 0x7fffu) /* no packet driver has an MTU >= 32768 */ 
  {                     /* its very likely a signededness error */
    pderrno = PDERR_NO_SPACE;
  }
  else if (lb_handle) /* do loopback */
  {
    struct recvr_info_t FAR *rit = __get_receiver_head();
    pderrno = PDERR_BAD_HANDLE;
    while (rit)
    {
      if (rit->intno == intno)
      {
        pderrno = __rit_do_downcall(rit, *lb_handle, buffer, buflen);
        break;
      }
      rit = rit->next;
    }
  }
  else if (__is_driver_there(intno))
  {  
    unsigned long rmbuf;

    /* must use a second buf because some drivers modify the source buffer. */
    /* (usually those that convert frame format or shim between drvr types) */
    rmbuf = __my_dosmem_alloc(buflen);
    pderrno = PDERR_NO_SPACE;

    if (rmbuf)
    {
      unsigned int tries;
      __my_dosmem_write( rmbuf, buffer, buflen ); 

      for (tries=0;tries<3;tries++)
      {
        IREGS regs;
        regs.r_ax = 0x0400; /* AH == 4, send_pkt() */
        regs.r_ds = (unsigned short)(rmbuf >> 16);
        regs.r_si = (unsigned short)(rmbuf & 0xffff);
        regs.r_es = regs.r_ds; /* for buggy univation driver */
        regs.r_di = regs.r_si;
        regs.r_cx = (unsigned short)buflen;

        pderrno = PKT_API(intno, &regs);

        if (pderrno == 0)
          break;
      }
      __my_dosmem_free(rmbuf);
    } /* if (xb) */
  } /* (pd_is_driver_there(intno)) */
  else if (intno == NULL_DRIVER_INTNO)
  {
    pderrno = PDERR_CANT_SEND;
  }

  return pderrno;
}

/* ------------------------------------------------------------------ */

/* Pop packets from the queue(s). Grow the queues if needed.
** Return number of packets popped.
*/
static int __rit_poll(struct recvr_info_t FAR *rit)
{
  unsigned int Q, numqueues = rit->cb.numqueues;
  unsigned long total_size = 0UL, total_used = 0UL;
  int is_sti = __are_interrupts_enabled();
  int popped = 0;

  for (Q = 0; Q < numqueues; Q++)
  {
    struct queue_hdr_t FAR *qhdrp;
    unsigned short qused;

    #if !defined(__DPMI32__)
    qhdrp = (struct queue_hdr_t FAR *)(rit->cb.queues[Q].segoff);
    #else
    qhdrp = (struct queue_hdr_t *)rit->cb.tmp_buf;
    if (qhdrp) /* not zero-base addressing */
      __my_dosmem_read(qhdrp,rit->cb.queues[Q].segoff,rit->cb.queues[Q].size);
    else
      qhdrp = (struct queue_hdr_t FAR *)((rit->cb.queues[Q].segoff & 0xffff)+
                                 ((rit->cb.queues[Q].segoff >> 12) & 0xffff0));
    #endif

    qused = qhdrp->used;
    total_size += qhdrp->size - sizeof(qhdrp);
    total_used += qused;

    if (qused)
    {
      unsigned char FAR *qbufp = ((unsigned char FAR *)qhdrp) + sizeof(*qhdrp);
      unsigned short qoffs = 0;

      while (qoffs < qused)
      {
        unsigned short plen = *((unsigned short FAR *)&qbufp[0]);
        unsigned short phdl = *((unsigned short FAR *)&qbufp[2]);

#if defined(SHOW_PACKET_POP)
_printk("P");
//_printk("(Q=%u, offs=%u, used=%u plen=%u, phdl=%04x)\n", Q, qoffs, qused, plen, phdl);
#endif
        if (__rit_do_upcall(rit, phdl, &qbufp[4], plen-4) == PDERR_NO_SPACE)
        {
          Q = numqueues;
          break;        
        }

        qoffs += plen;
        qbufp += plen;
        popped++;

      } /* while (qused) */

      qbufp = ((unsigned char FAR *)qhdrp) + sizeof(*qhdrp);
      if (is_sti) 
        __cli();
      #if defined(__DPMI32__)
      if (rit->cb.tmp_buf) /* not zero-base addressing */
        __my_dosmem_read(qhdrp,rit->cb.queues[Q].segoff,rit->cb.queues[Q].size);
      #endif
      qhdrp->used -= qoffs;
      qused = qhdrp->used;
      FAR_MEMMOVE( qbufp, qbufp + qoffs, qused );
      #if defined(__DPMI32__)
      if (rit->cb.tmp_buf) /* not zero-base addressing */
        __my_dosmem_write(rit->cb.queues[Q].segoff,qhdrp,rit->cb.queues[Q].size);
      #endif
      if (is_sti) 
        __sti();
 
    } /* while (qused) */
  } /* for (Q = 0; Q < numqueues; Q++) */   

  if (!rit->cb.stop_queue_growth &&
      numqueues < (sizeof(rit->cb.queues)/sizeof(rit->cb.queues[0])) )
  {
    unsigned long counters[6];

    __my_dosmem_read( &counters[0], rit->cb.code_segoff, sizeof(counters) );

    if ((rit->cb.no_buffs_ctr != counters[3]) ||
         total_used >= (total_size-(total_size>>3)) ) /* >= 7/8ths used */
    {
      /* counter[0] == internal (its code) */
      /* counter[1] == internal (its code) */
      /* counter[2] == reserved (its the queue head) */
      /* counter[3] == "no buffs error" counter */
      /* counter[4] == "bad length" counter */
      /* counter[5] == "added ok" counter */

#if defined(SHOW_QUEUE_GROW)
      _printk("G* [%lu => %lu] (4=%lu 5=%lu)\n", 
              rit->cb.no_buffs_ctr, counters[3], counters[4], counters[5] );
#endif
      rit->cb.no_buffs_ctr = counters[3];
      if (__grow_queues(rit) < 1) /* failed (no mem) */
        rit->cb.stop_queue_growth = 1;
    }
  }  

  return popped;
}

/* ==================================================================== */
/*                        pseudo-API: poll receiver                     */
/* ==================================================================== */

int pdapi_poll_receiver(int intno)
{
  struct recvr_info_t FAR *rit = __get_receiver_head();
  int popped = 0;

  while (rit)
  {
    if (intno == -1 /* all */ || rit->intno == intno)
    {
#if defined(SHOW_POLL_CALL)
_printk("p");
#endif
      popped += __rit_poll(rit);
      if (intno != -1)
        break;
    }
    rit = rit->next;
  }

  return popped;
}

/* ==================================================================== */
/*                    PKT API: get driver params                        */
/* ==================================================================== */

#if 0
struct pdapi_driver_params
{
  unsigned char  major_rev;      /* Revision of Packet Driver spec */
  unsigned char  minor_rev;      /*  this driver conforms to. */
  unsigned char  addr_len;       /* Length of a MAC-layer address */
  unsigned short ll_mtu;         /* MTU, including MAC headers */
  unsigned short multicast_aval; /* Buffer size for multicast addr */
  unsigned short rcv_bufs;       /* (# of back-to-back MTU rcvs) - 1 */
  unsigned short xmit_bufs;      /* (# of successive xmits) - 1 */
  unsigned short int_num;        /* interrupt for post-EOI processing */
};
#endif

int pdapi_get_driver_params( int intno, 
                             struct pdapi_driver_params FAR *params )
{
  int pderrno = 0xff; /* not there */

  if (__is_driver_there(intno))
  {
    IREGS regs; 
    regs.r_ax = 0x1000; /* ah = 10h, get driver parameters */
                        /* Added in v1.09 (Sep-14-1989) of the spec */

    pderrno = PKT_API(intno, &regs);
    if (pderrno == 0)
    {
      unsigned long segoff = (((unsigned long)regs.r_es)<<16)+regs.r_di;
      char parambuf[14];

      __my_dosmem_read( parambuf, segoff, sizeof(parambuf));
      if ( *((unsigned short *)(&parambuf[4])) < 256) /* broken MTU check */
        pderrno = -1;
      else
      {
        params->major_rev      = *((unsigned char  *)(&parambuf[0]));
        params->minor_rev      = *((unsigned char  *)(&parambuf[1]));
        params->addr_len       = *((unsigned char  *)(&parambuf[3]));
        params->ll_mtu         = *((unsigned short *)(&parambuf[4]));
        params->multicast_aval = *((unsigned short *)(&parambuf[6]));
        params->rcv_bufs       = *((unsigned short *)(&parambuf[8]));
        params->xmit_bufs      = *((unsigned short *)(&parambuf[10]));
        params->int_num        = *((unsigned short *)(&parambuf[12]));
        pderrno = 0;
      }
      /* DBGPRINTF(("*get_driver_params() mtu is %u\n",params->ll_mtu)); */
    }
  }  
  else if (intno == NULL_DRIVER_INTNO) /* null pkt driver */
  {
    pderrno = PDERR_BAD_COMMAND;
  }

  /* Commonly used MTUs: (ref. [x] = see MTU discovery [rfc1191] footnotes)
  *
  ** NOTE: the packet driver spec says the "mtu" obtained from the 
  ** 'driver params' INCLUDES MAC HEADERS. Ergo, the MAC header size 
  ** must be added to values in this list to be packet-driver-spec compliant.
  *
  *  Plateau    MTU    Comments                      Reference
  *  ------     ---    --------                      ---------
  *             65535  Official maximum MTU          RFC 791
  *             65535  Hyperchannel                  RFC 1044
  *  65535
  *  32000             Just in case
  *             17914  16Mb IBM Token Ring           ref. [6]     class ??
  *  17914
  *             8166   IEEE 802.4                    RFC 1042
  *  8166
  *             4464   IEEE 802.5 (4Mb max)          RFC 1042     class 3
  *             4352   FDDI (Revised)                RFC 1188     class 12
  *  4352 (1%)
  *             2048   Wideband Network              RFC 907
  *             2002   IEEE 802.5 (4Mb recommended)  RFC 1042     class 3
  *  2002 (2%)
  *             1536   Exp. Ethernet Nets (3Mb)      RFC 895
  *             1500   Ethernet Networks  (10Mb)     RFC 894      class 1
  *             1500   Point-to-Point (default)      RFC 1134     class 16/18
  *             1492   IEEE 802.3                    RFC 1042     class 11
  *  1492 (3%)
  *             1006   SLIP                          RFC 1055     class 6
  *             1006   ARPANET                       BBN 1822
  *  1006
  *             576    X.25 Networks                 RFC 877      class 5
  *             544    DEC IP Portal                 ref. [10]
  *             512    NETBIOS                       RFC 1088
  *             508    IEEE 802/Source-Rt Bridge     RFC 1042
  *             508    ARCNET                        RFC 1051     class 8
  *  508 (13%)
  *             296    Point-to-Point (low delay)    RFC 1144
  *  296
  *  68                Official minimum MTU          RFC 791
  */

  if (pderrno == PDERR_BAD_COMMAND)  /* <=1.08 driver */
  {
    struct pdapi_driver_info info;
    if (pdapi_get_driver_info(intno, 0xffff, &info) == 0)
    {
      unsigned short mtu = 0;
      unsigned char alen = 0;
      /* 
       * alen == maximum (worst case) hardware address len.
       * mtu  == mtu at the packet driver (includes MAC header len).
       *         if header length is not fixed, use largest possible.
       *         if variable MTU sizes are possible, use smallest.
      */
      /* the specifications <=1.08 only define classes 1..11  */
      switch (info.if_class) 
      {
        case 1:  /* DIX/BLUEBOOK */
             mtu = (1500 + 14); alen = 6;
             break;
        case 2: /* Proteon Token Ring 10Mbit */
             /* (MAC hdr+18 octet RIF)+(LLC+SNAP) => (18+18)+(3+5) */
             /* mtu = (4464+18+18+3+5); alen = 6; */
             mtu = (2002+18+18+3+5); alen = 6;
             break;
        case 3: /* Token Ring (variable RIF, 4Mbit) */
             /* (MAC hdr+18 octet RIF)+(LLC+SNAP) => (18+18)+(3+5) */
             mtu = (2002+18+18+3+5); alen = 6; 
             break;
        case 4: /* Corvus Omninet (4Mbit) */
             mtu = 0; alen = 0;                   /* UNKNOWN! */
             break;
        case 5: /* IP/DDP (Localtalk) */
             mtu = 586; alen = 4;
             break;
        case 6: /* SLIP */
             mtu = 1006; alen = 0;
             break;
        case 7:  /* old starlan */
             mtu = (1500 + 14); alen = 6;
             break;
        case 8: /* ARCNET */
             /* mtu = (508 + 4); alen = 1; */
             mtu = (252 + 4); alen = 1;
             break;
        case 9: /* AX.25 */
        case 10: /* KISS */
             mtu = 2048; alen = 7; /* includes SSID */
             break;
        case 11: /* 802.3 w/802.2 ("8"==LLC+SNAP) */ 
             mtu = (1500-8)+(14+8); alen = 6;
             break;
        case 12: /* FDDI (802.2 headers) */
             mtu = (4352)+(18+18+3+5); alen = 6;
             break;
        default:
             mtu = 0;
             break;
      }
      if (alen)
      {
        params->major_rev = 1;
        params->minor_rev = 8;
        params->addr_len  = alen;
        params->ll_mtu    = mtu;
        params->multicast_aval = 0;
        params->rcv_bufs  = 0;
        params->xmit_bufs = 0;  
        params->int_num   = 0;
        pderrno = 0;
      }
    }
  }  

  if (pderrno == 0 && (params->rcv_bufs+1u) != 0u)
  {
    struct recvr_info_t FAR *rit = __get_receiver_head();
    while (rit)
    {
      if (rit->intno == intno)
        break;
      rit = rit->next;
    }
    if (rit)
    {
      unsigned int ll_mtu = params->ll_mtu + 4; /*4=perpkt storage overhead*/
      unsigned int nbufs, Q, numqueues = rit->cb.numqueues;
      for (nbufs = Q = 0; Q < numqueues; Q++)
        nbufs += ((rit->cb.queues[Q].size-sizeof(struct queue_hdr_t))/ll_mtu);
      if (nbufs > (params->rcv_bufs+1u))
        params->rcv_bufs = (unsigned short)(nbufs-1);
    }
  }

  return pderrno;
}

/* ==================================================================== */
/*                    pseudo-API: set # of recv buffs                   */
/* ==================================================================== */

int pdapi_set_recvbuf_count( int intno, unsigned int newcount )
{
  int pderrno = 0;
  if (newcount > 0)
  {
    struct recvr_info_t FAR *rit = __get_receiver_head();
    pderrno = PDERR_NO_SPACE;

    while (rit)
    {
      if (rit->intno == intno)
        break;
      rit = rit->next;
    }

    if (rit)
    {
      struct pdapi_driver_params params;
      pderrno = pdapi_get_driver_params( intno, &params );
      if (pderrno == 0)
      {
        if (params.rcv_bufs < (newcount - 1))
        {
          unsigned int llmtu = params.ll_mtu + 4;
          unsigned int nbufs = params.rcv_bufs + 1;
          unsigned int numqueues = rit->cb.numqueues;
          #if defined(SHOW_QUEUE_GROW)
          _printk("start nbufs = %u (numqueues=%u), to get=%u\n", 
                                      nbufs, numqueues, newcount );
          #endif
          while (nbufs < newcount &&
             numqueues < (sizeof(rit->cb.queues)/sizeof(rit->cb.queues[0])) )
          {
            unsigned int Q = numqueues;
            if (__grow_queues(rit) < 1) 
              break; /* failed to grow */
            numqueues = rit->cb.numqueues;
            for (; Q < numqueues; Q++)
              nbufs += ((rit->cb.queues[Q].size-sizeof(struct queue_hdr_t))/llmtu);
            #if defined(SHOW_QUEUE_GROW)
            _printk("next nbufs = %u (numqueues=%u)\n", nbufs, numqueues );
            #endif
          }
          #if defined(SHOW_QUEUE_GROW)
          _printk("end nbufs = %u (numqueues=%u)\n", nbufs, numqueues );
          #endif
          if (nbufs < newcount) 
            pderrno = PDERR_NO_SPACE;
        }
      } /* if (rit) */
    } /* if (got params ok) */
  } /* if (newcount > 0) */

  return pderrno;
}

/* ==================================================================== */
/*                        PKT API: open driver                          */
/* ==================================================================== */

/*
** 'proto_type' must be a DIX/BLUEBOOK ETHERTYPE_* in _HOST_ order!!
** Use 0xffff for all protocols (== "all/unspecified").
*/
int pdapi_open_driver(int intno, unsigned short proto_type,
                      int (*rcvr)( int intno, unsigned short handle, 
                                   unsigned int len, const void FAR *buf ), 
                      unsigned short FAR *handle )
{
  int pderrno = 0xff;

  if (__is_driver_there(intno))
  {
    struct pdapi_driver_info info;
    pderrno = pdapi_get_driver_info( intno, 0xffff/*handle*/, &info );

    if (pderrno == 0)
    {
      unsigned short typebuf[8]; /* max 16bytes type */
      unsigned long rmbuf = __my_dosmem_alloc(sizeof(typebuf));
      pderrno = PDERR_NO_SPACE;
      if (rmbuf)
      {
        /* allocate a receiver callback and return the seg:off of the callback */
        unsigned long rcvr_code;
        rcvr_code = __rit_bind( intno, ((const unsigned short *)0), 0, 0);

        if (rcvr_code)
        {
          IREGS regs;
          signed char typelen = 0;

          pderrno = 0;
          if (proto_type != 0xffffu) /* not all ETHERTYPE_* */
          {
            typelen = __fill_proto_type( info.if_class, proto_type, typebuf );
            if (typelen < 0) /* don't know how to deal with this class/proto */
              pderrno = PDERR_BAD_TYPE;
          }
          if (pderrno == 0)
          {
            __my_dosmem_write( rmbuf, typebuf, typelen );

            regs.r_ax = 0x200; /* access_type() AH == 2 */
            regs.r_ax += (short)(info.if_class & 0xff); /* AL = if_class */
            regs.r_bx = info.if_typ;             /* BX = if_type */
            regs.r_dx = info.if_number;          /* DL = if_number */
            regs.r_cx = typelen;
            /* ds:si = segoff to proto_type */
            regs.r_ds = (unsigned short)(rmbuf >> 16);    
            regs.r_si = (unsigned short)(rmbuf & 0xffff);
            /* es:di = segoff to recvr code */
            regs.r_es = (unsigned short)(rcvr_code >> 16);      
            regs.r_di = (unsigned short)(rcvr_code & 0xffff); 

            pderrno = PKT_API(intno, &regs);
          }
          if (pderrno != 0)
          {
            __rit_unbind( intno, ((const unsigned short *)0) );
          }
          else 
          {
            unsigned short tmp_handle = regs.r_ax;
            if (__rit_bind( intno, &tmp_handle, proto_type, rcvr ))
            {
              *handle = regs.r_ax; 
              pderrno = 0;  /* success */
            }
            else /* couldn't add receiver - out of handle slots */
            {  
              __rit_unbind( intno, ((const unsigned short *)0) );
              regs.r_bx = regs.r_ax;  /* handle */
              regs.r_ax = 0x0300; /* release_type() */
              PKT_API(intno, &regs);
              pderrno = PDERR_NO_SPACE;
            }
          } /* opened ok */
        } /* if (rcvr_code) */
        __my_dosmem_free(rmbuf);
      } /* if (rmbuf) */
    } /* if (gotinfo) */
  } /* if (__is_driver_there(intno)) */
  else if (intno == NULL_DRIVER_INTNO) /* nullpkt driver */
  {
    struct recvr_info_t FAR *rit = __get_receiver_head();
    pderrno = 0;

    while (rit)
    {
      if (rit->intno == intno)
      {
        unsigned int i, n  = rit->handle_count;
        for (i = 0; i < n; i++)
        {
          if (rit->handle_tab[i].proto == proto_type)
          {
            pderrno = PDERR_TYPE_INUSE;
            break;
          }
        }
        break;
      }
      rit = rit->next;
    }

    if (pderrno == 0)
    {
      *handle = proto_type; /* dummy */
      __rit_bind( intno, handle, proto_type, rcvr );
      /* a failed __rit_bind() (ENOMEM) is not an error */
    }
  } /* nullpkt driver */

  return pderrno;
}                 

/* ==================================================================== */
/*                        PKT API: close driver                         */
/* ==================================================================== */

int pdapi_close_driver(int intno, unsigned short handle)
{
  int pderrno = PDERR_BAD_HANDLE; /* not there */

  /* DBGPRINTF(("+ pdapi_close_driver(x%02x,x%04x)\n",intno,handle)); */

  if (__is_driver_there(intno)) /* not a nullpkt driver */
  {
    if (__rit_unbind( intno, &handle ) == 0)
    {
      IREGS regs;
      regs.r_bx = handle;
      regs.r_ax = 0x0300; /* release_type() */

      /* DBGPRINTF(("*begin 0x0300/%04x\n",handle));  */
      pderrno = PKT_API(intno, &regs);
      /* DBGPRINTF(("*end 0x0300/%04x pderrno=%d\n",pderrno)); */
    }
  }
  else if (intno == NULL_DRIVER_INTNO) /* nullpkt driver */
  {
    __rit_unbind( intno, &handle );
    pderrno = 0;
  }

  /* DBGPRINTF(("- pdapi_close_driver()=>%d\n",pderrno)); */
  return pderrno;
}      


/* ==================================================================== */
/*                        PKT API: find driver                          */
/* ==================================================================== */

int pdapi_find_driver(int first_intno_to_try)
{
  int intno = first_intno_to_try;

  if (intno > 0x80)
    return -1;

  if (intno < 0x60)
    intno = 0x60;

  while (intno <= 0x80)
  {
    if (__is_driver_there(intno))
      return intno;
    intno++;
  }
  return NULL_DRIVER_INTNO; /* nullpkt driver */
}

/* ==================================================================== */

/*
** The following is the source to the packet receiver.
*/
#if 0 /* defined(__WATCOMC__) && defined(__LARGE__) */
extern void pkt_recvr(void);
#pragma aux pkt_recvr =                                         \
              "pushf                     " /*   0, 0x00 */      \
              "cli                       " /*   1, 0x00 */      \
              "push  bp                  " /*   2, 0x02 */      \
              "mov   dx,dx               " /*   3, 0x03 */      \
              "call  receiver_entry      " /*   5, 0x05 */      \
              "                          "                      \
              ""/* the order/placement of data fields is critical */ \
              "queue_head:      dd 0     " /*   8, 0x08 */      \
              "count_no_buffs:  dd 0     " /*  12, 0x0C */      \
              "count_bad_len:   dd 0     " /*  16, 0x10 */      \
              "count_added_ok:  dd 0     " /*  20, 0x14 */      \
              "                          "                      \
              "receiver_entry:           "                      \
              "pop   bp                  " /* data pointer */   \
              "                          "                      \
              "or    ax,ax               "                      \
              "jnz   receiver_end        "                      \
              "                          "                      \
              "mov   ax, cs              "                      \
              "mov   es, ax              " /* q_head ** seg */  \
              "lea   di, cs:[bp+0]       " /* q_head ** off */  \
              "add   cx, 4               " /* wrap chk and... */\
              "cmp   cx, 24              " /* len>=20 check */  \
              "ja    call0_start         "                      \
              "xor   di, di              "                      \
              "mov   es, di              "                      \
              "mov   ax, 1               " /* second counter */ \
              "jmp   short call0_end     " /* fini with ax=1 */ \
              "                          "                      \
              "call0_start:              "                      \
              "les   di, es:[di+0]       "                      \
              "mov   ax, es              "                      \
              "or    ax, di              " /* ax == first ctr */\
              "jz    call0_end           " /* fini with ax=0 */ \
              "mov   ax, es:[di+4]       " /* queue size */     \
              "sub   ax, 8               " /* -= sizeof(qhdr) */\
              "sub   ax, es:[di+6]       " /* -= queue used */  \
              "cmp   ax, cx              "                      \
              "jb    call0_start         "                      \
              "mov   ax, es:[di+6]       " /* queue used */     \
              "add   es:[di+6], cx       " /* add new len */    \
              "add   di, 8               " /* skip to q buf */  \
              "add   di, ax              " /* add old q used */ \
              "mov   es:[di+0], cx       " /* this block len */ \
              "mov   es:[di+2], bx       " /* this block hdl */ \
              "add   di, 4               " /* => buff space */  \
              "mov   ax, 2               " /* third counter */  \
              "                          "                      \
              "call0_end:                "                      \
              "inc   ax                  " /* skip past qptr */ \
              "shl   ax, 1               " /* x2             */ \
              "shl   ax, 1               " /* x4             */ \
              "add   bp, ax              " /* add to origin */  \
              "add   word ptr cs:[bp+0],1"                      \
              "adc   word ptr cs:[bp+2],0"                      \
              "                          "                      \
              "sub   cx, 4               " /* restore cx */     \
              "xor   ax, ax              " /* restore ax */     \
              "                          "                      \
              "receiver_end:             "                      \
              "pop   bp                  "                      \
              "popf                      "                      \
              "retf                      "                      \
              modify exact [];           
void dummy(void)
{
  pkt_recvr();
}
#endif

/* ======================================================================== */
