/*
 * printk.c - Formatted printf style routines.
 * Used by syslog() et al and also supercedes outs() et al, that
 * are now simply macros that use _printk().
 *
 * All functions are inherantly interrupt safe. (but fd ops==DOS ops)
 * Non-standard formats that are supported are...
 *   - %m (strerror() for current errno), 
 *   - %t (current time) 
 *   - %I (IP address) 
 *   - %M (MAC address)
 * Doesn't do floating-point formats.
 * Doesn't use any library functions other than strerror() and 
 * those that will be inlined or are always linked anyway, eg 
 * strlen()/write().
 *
 * _[v]snprintk() is like [v]sprintf(), except we may also specify 
 * the length of the output buffer. 
 * _[v]hprintk() is like [v]fprintf(), except that it takes a file handle
 * (file descriptor), and if the fd is 2 ('stderr'), then the output goes
 * directly to video (direct video buffer write or video-BIOS call)
 * _printk() is equivalent to fprintf(stderr) but writes directly to video.
 *
 * Inspired by Linux's printk().
 *
 * Written by Gisle Vanem <giva@bgnett.no>  Jan-97
 * Extensive modifications by Cyrus Patel <cyp@fb14.uni-mainz.de>, May 2001
 * to include: 
 * - Added an _hprintk(fd) to write to an arbitrary file descriptor.
 * - _printk(), _sprintk() and _hprintk(2,...) (2 = 'stderr') are
 *   all completely interrupt safe. Uses direct video output when the 
 *   destination is 'stderr'. Uses cmos time rather than time()/DOS; 
 * - Made reentrant (no static buffers); 
 * - Removed stuff used only by sylog() [syslog is completely standalone now]; 
 * - Removed _printk_safe() - no longer needed since everything is intr safe; 
 * - Fixed call to strerror() stomping on returned *; 
 * - Removed dependancy on all non-standard-lib-externals (htonl() etc); 
 * - Added long form (%ld etc) support for number formats; '%S' and '%C' 
 *   (uppercase %s/%c), %M (MAC addr), 'h' modifier for %I (host order addr), 
 *   'f' and 'n' modifier for %p and %s (far/near pointer).
 * - 16bit code uses inline asm for 32bit divide by 10 (to support non-_TEXT 
 *   segment based small-model)
 */
#include <stdarg.h>  /* va_* */
#include <string.h>  /* strlen(), memcpy(), strncpy() */
#include <ctype.h>   /* isdigit() */
#include <errno.h>   /* strerror() ["%m" fmt] */
#include <limits.h>  /* INT_MAX/UINT_MAX */
#include <stdlib.h>  /* _os_major is here */
#include <signal.h>  /* SIGINT blocking */
#if defined(__DJGPP__)
#include <pc.h>      /* for inp()/outp() in cmos_ctime_r() ["%t" fmt] */
#include <unistd.h>  /* write() */
#endif
#include <conio.h>   /* putch() and maybe inp()/outp() */
#include <io.h>      /* write() */
#include <dos.h>     /* bdos() for _printk() */

#include "printk.h"

/*
** "%t" format support - similar as ctime(&time(NULL)) [which is again
** the same as asctime( localtime( &time(NULL) ) )].
**
** Unlike time() based functions, this is interrupt safe.
*/ 
static int __cmos_ctime_r(char *buf, unsigned int buflen) 
{
  char resbuf[sizeof("Sat Mar 21 15:58:27 1987\r\0")+5];
  const char *names;
  unsigned char status;
  unsigned char parts[7];
  unsigned int i, n;

  #if (defined(__TURBOC__) || defined(__TURBOC__) && !defined(__DPMI32__))
  #undef outp                     /* use calls instead of intrinsics */
  #undef inp                      /* the intrinics make the compiler puke */
  #elif defined(__DJGPP__)
  #define inp inportb
  #define outp outportb
  #endif

  /* cmos port numbers for wday, month, mday, hour, min, sec, year*/
  parts[0] = 0x06; parts[1] = 0x08; parts[2] = 0x07; parts[3] = 0x04;
  parts[4] = 0x02; parts[5] = 0x00; parts[6] = 0x09;

  do
  {
    (void)outp ( 0x70, 0x0a ); /* CMOS 0Ah - RTC - STATUS REGISTER A */
    status = (unsigned char)(0xff & inp( 0x71 ));
  } while ((status & 0x80)!=0); /* while (update in progress) */

  for (n = 0; n <= 6; n++)
  {
    (void)outp ( 0x70, parts[n] ); 
    parts[n] = (unsigned char)inp (0x71);     
  }

  (void)outp ( 0x70, 0x0b ); /* CMOS 0Ah - RTC - STATUS REGISTER B */
  status = (unsigned char)(0xff & inp( 0x71 ));

  /*
  ** printf("in hex?=%d  in ampm?=%d\n", ((status & (1<<2))!=0),
  **                                  ((status & (1<<1))==0) );
  */
  if ((status & (1<<2))==0) /* values in bcd - convert to binary */
  {
    for (n = 0; n <= 7; n++) 
    {
      if (n != 3 || (status & (1<<1))!=0) /* not hours or not in am/pm */
        parts[n] = (unsigned char)(((parts[n]>>4)*10)+(parts[n] & 0x0f));
      else if (parts[3] == 0x92)           /* 12:nn pm */
        parts[3] = 0;                      /* ... is 0:nn */
      else
      {
        i = (parts[3] & 0x7F);             /* 80-91 => 00-11 */
        i = (((i>>4)*10)+(i & 0x0f));      /* 00-11 => 0x00-0x0b */
        if (parts[3] & 0x80) i += 12;      /* 80-91 => 0x0c-0x17 */
        parts[3] = (unsigned char)i;
      }
    }
  }
  else  /* already binary */
  {
    /* note that a number of bioses have a problem with dealing with
    ** cmos in hex (its very, very uncommon anyway)
    */
    if ((status & (1<<1))==0) /* in am/pm form */
    {
      if (parts[3] == 0x8c)                /* 12pm => 0 am */
        parts[3] = 0;
      else if ((parts[3] & 0x80)!=0)       /* "pm" (0x80-0x8B) */
        parts[3] = (unsigned char)(12 + (parts[3] - 0x80)); /* => 0x0C-0x17 */
    }
  }

  if ((parts[0] < 1 || parts[0] > 7) /* many bios' have a broken weekday */
  && (parts[1] >= 1 && parts[1] <= 12))
  {
    #define dow(y,m,d) \
        ( ( ( 3*(y) - (7*((y)+((m)+9)/12))/4 + (23*(m))/9 + (d) + 2    \
        + (((y)-((m)<3))/100+1) * 3 / 4 - 15 ) % 7 ) )
    n = (parts[6] + ((parts[6] < 80)?(1900):(2000))); /* 1980-2080 */
    i = (parts[1] - 1); /* 0-11 */
    parts[0] = (unsigned char)(1+(dow( n, i, ((int)parts[2])))); /* 1-7 */
    #undef dow         
  }

  i = 0;
  n = parts[0]; /* CMOS 06h - RTC - DAY OF WEEK (0x1-0x7 1=sunday) */
  if (n > 7) 
    n = 0;
  names = "???SunMonTueWedThuFriSat";
  names += (n * 3);
  resbuf[i++] = *names++;
  resbuf[i++] = *names++;
  resbuf[i++] = *names++;
  resbuf[i++] = ' ';
  n = parts[1]; /* CMOS 08h - RTC - MONTH (now in hex: 0x01-0x0C) */
  if (n > 12) 
    n = 0;
  names = "???JanFebMarAprMayJunJulAugSepOctNovDec";
  names += (n * 3);
  resbuf[i++] = *names++;
  resbuf[i++] = *names++;
  resbuf[i++] = *names++;
  for (n = 2; n <= 5; n++)
  {
    resbuf[i++] = ((n > 3) ? (':') : (' '));
    resbuf[i++] = (char)('0'+(parts[n] / 10));
    resbuf[i++] = (char)('0'+(parts[n] % 10));
  }
  n = parts[6];     /* CMOS 09h - RTC - YEAR (now in hex: 0x00-0x63) */
  resbuf[i++] = ' ';
  resbuf[i++] = (char)((n < 80)?('2'):('0'));
  resbuf[i++] = (char)((n < 80)?('0'):('9'));
  resbuf[i++] = (char)('0'+(n / 10));
  resbuf[i++] = (char)('0'+(n % 10));
  resbuf[i++] = '\n';
  resbuf[i++] = '\0';  /* length includes '\0' and is always 26 */ 

  if (buflen > i)
    buflen = i; 
  strncpy(buf, resbuf, buflen);
  buf[buflen-1] = '\0';
  return buflen;
}

/* ------------------------------------------------------------------ */

#ifdef __cplusplus
extern "C" {
#endif
#if defined(__HIGHC__)
  extern char *GetArg0 (USHORT sel);
#elif defined (__DJGPP__)
  extern char **__crt0_argv;
  #if defined(__cplusplus) 
  /* emit this into every object file that is compiled with c++ */
  /* we don't actually use c++ elements, and don't want lib???cx */ 
  __asm__(".weak ___gxx_personality_v0\n ___gxx_personality_v0 = _errno\n");
  #endif
#elif defined (_MSC_VER)
  extern char **__argv;
#elif defined(__WATCOMC__) || defined(__BORLANDC__) || defined(__TURBOC__)
  extern char **_argv;
#else
  #error "What's up doc?"
#endif 
#ifdef __cplusplus
}
#endif

static /*const*/ char * __get_argv0(void)
{
  /* This function is duplicated in pcconfig.c, printk.c and syslog.c */
#if defined(__HIGHC__)
  if (_osmajor >= 3)
  {
    /* extern char *GetArg0 (USHORT sel); */
    CONFIG_INF   cnf;
    _dx_config_inf (&cnf, (UCHAR*)&cnf);  /* get config block */
    return GetArg0 (cnf.c_env_sel);       /* in exc_??.lib    */
  }
#elif defined (__DJGPP__)
  _get_dos_version (0); /* 2.02 bug; _osmajor/_osminor not set in crt0.o */
  if (_osmajor >= 3)
  {
    /* extern char **__crt0_argv; */
    return __crt0_argv[0];
  }
#elif defined (_MSC_VER)
  if (_osmajor >= 3)
  {
    /* extern char **__argv; */
    return __argv[0];
  }
#elif defined(__WATCOMC__) || defined(__BORLANDC__) || defined(__TURBOC__)
  if (_osmajor >= 3)
  {
    /* extern char **_argv; */
    return _argv[0];
  }
#else
  #error "What's up doc?"
#endif 
  return (/*const*/ char *)0;
}

/* ===================================================================== */

#if defined(__WATCOMC__)
extern void __dos_putch(unsigned int c);
#pragma aux __dos_putch="mov ah,0x06""int 0x21" parm [dx] modify exact [dx ax];
#elif defined(__TURBOC__) || defined(__BORLANDC__)
#define __dos_putch(__c) do { _AX=0x600; _DX=__c; __asm int 21h; } while(0)
#elif defined(__GNUC__)
#define __dos_putch(__c) __asm__("movb $0x06,%%al\n\tint $0x21"::"d"(__c))
#else
#define __dos_putch(__c) bdos(0x06, __c, 0)
#endif

/* ----------------------------------------------------------------------- */

#if defined(__GNUC__)
static __inline void __bios_putch(unsigned int c)
{
  __asm__ __volatile__("movb $0x0e,%%ah\n\t"
                       "int $0x10\n\t"
                       : "=a"(c /* dummy */) : "a"(c) );
}
#elif defined(__TURBOC__)
#define __bios_putch(__c)                    \
{                                            \
  _AL = (unsigned char)__c;                  \
  _AH = 0x0e;                                \
  __emit__(0xCD,0x10);                       \
}
#elif defined(__WATCOMC__)
extern void __bios_putch(unsigned int c);
#pragma aux __bios_putch="mov ah,0x0e""int 0x10" parm [ax] modify exact [ax];
#endif

/* ----------------------------------------------------------------------- */

#if defined(__DJGPP__) 
/* cant use conio (putch) because _printk() is called from constructors */
#define __vid_putch(__c) __bios_putch(__c)
#define __vid_sync(__c)  /* nothing */
#elif defined(__HIGHC__) || defined(PHARLAP_H)
static void __vid_putch(unsigned c) 
{
  REGS regs;
  regs.w.ax = 0x0e00 | ((unsigned char)c);
  int386( 0x10, &regs, &regs );
}
#define __vid_sync(__c)  /* nothing */
#else /* watcomc/borlandc/turboc/msc - rmode/pmode */
static void __vid_putch(unsigned int c)
{
  #if defined(__TURBOC__)
  #pragma option -a-
  #else
  #pragma pack(1)
  #endif
  struct vidinfo {
    unsigned char mode;
    unsigned short cols;
    unsigned short pagesize;
    unsigned short pageoff;
    struct {unsigned char col,row;} xy[8];
    unsigned short curstype;
    unsigned char pagenum;
    unsigned short crtbase; /* 0x03D4 for text mode */
  };
  #pragma pack() 

  static int force_bios = -1;
  #if (UINT_MAX == USHRT_MAX) /* real mode target */
  #define myFAR _far
  #define myMEMMOVE _fmemmove
  #define dosMK_FP(__seg,__off) ((((unsigned long)(__seg))<<16)+(__off))
  unsigned short save_es = 0;
  unsigned short save_ds = 0;
    #if defined(__TURBOC__)
    #define SAVE_ES_DS()     save_es = _ES; save_ds = _DS;
    #define RESTORE_ES_DS()  _ES = save_es; _DS = save_ds;
    #else /* watcomc */
    #define SAVE_ES_DS()     __asm mov save_es, es \
                             __asm mov save_ds, ds
    #define RESTORE_ES_DS()  __asm mov es, save_es \
                             __asm mov ds, save_ds
    #endif
  #else
  #define myFAR 
  #define myMEMMOVE memmove
  #define dosMK_FP(__seg,__off) ((((unsigned long)(__seg))<<4)+(__off))
  #define SAVE_ES_DS()    /* nothing */
  #define RESTORE_ES_DS() /* nothing */
  #endif
  struct vidinfo myFAR *vi;
  unsigned char myFAR *scr;
  unsigned int cols, rows;
  unsigned int col, row;
  unsigned char pagenum;
  unsigned short myFAR *pos;
  unsigned short cell;

  SAVE_ES_DS();

  if (force_bios == -1)
  {
    #if (UINT_MAX > USHRT_MAX) /* pmode target */
    /* zero-base addressing is required, otherwise use BIOS */
    char is_z_base = 0;
    __asm push ebx
    __asm mov eax,6
    __asm mov bx,ds
    __asm int 31h
    __asm pop ebx
    __asm sbb eax,eax    /* if cflag == 0 && dx == 0 && cx == 0 */
    __asm or  dx,cx
    __asm or  ax,dx
    __asm setz al
    __asm mov is_z_base,al
    force_bios = (is_z_base ? 0 : 1);
    #else
    /* the direct screen mem stuff is extremely erratic (works 
     * sometimes and not othertimes) for real mode under NT/2K. 
     * No idea why. Pmode is fine though.
    */
    const char *p = getenv("OS");
    force_bios = ((p && strcmp(p,"Windows_NT")==0) ? 1 : 0);
    #endif
  }
  if (force_bios)
  {
    __bios_putch(c);
    RESTORE_ES_DS();
    return;
  }

  vi = (struct vidinfo myFAR *)dosMK_FP(0x40,0x49);

  if (vi->crtbase != 0x03D4)
  {
    __bios_putch((unsigned char)c);
    RESTORE_ES_DS();
    return;
  }

  scr = (unsigned char myFAR *)dosMK_FP(0xffff,0x0003);
  if (*((unsigned short myFAR  *)scr) == 0xfd80) /* PC98 */
    scr = (unsigned char myFAR *)dosMK_FP(0xa000,0);
  else if (vi->mode == 0x07)
    scr = (unsigned char myFAR *)dosMK_FP(0xb000,0);
  else
    scr = (unsigned char myFAR *)dosMK_FP(0xb800,0);
  scr += vi->pageoff;

  cols = vi->cols;
  rows = vi->pagesize/(((unsigned int)vi->cols) << 1);
  pagenum = vi->pagenum;

  if (c == '\r')
  {
    if (vi->xy[pagenum].col == 0)
    {
      RESTORE_ES_DS();
      return; /* no position change */
    }
    vi->xy[pagenum].col = 0;
  }
  else if (c == '\a')
  {
    __bios_putch(c);
    col = vi->curstype; /* consume reference to keep watcom happy */
    RESTORE_ES_DS();
    return; /* no position change */
  }
  else if (c == '\b')
  {
    if (vi->xy[pagenum].col == 0)
    {
      RESTORE_ES_DS();
      return; /* no position change */
    }
    vi->xy[pagenum].col--;
  }
  else if (c == '\f')
  {
    col = vi->xy[pagenum].col;
    row = vi->xy[pagenum].row;
    pos = (unsigned short myFAR *)(scr + (((cols<<1)*row) + (col<<1)) );
    cell = (unsigned short)(((*(pos-1)) & 0xff00) | ' ');
    pos = (unsigned short myFAR *)(scr) ;
    row = (cols * rows);
    for (col = 0; col < row; col++)
      *pos++ = cell;
    vi->xy[pagenum].col = 0;
    vi->xy[pagenum].row = 0;
  }
  else if (c == '\v')
  {
    if ((vi->xy[pagenum].row + 1) >= rows)
    {
      RESTORE_ES_DS();
      return; /* no position change */
    }
    vi->xy[pagenum].row++;
  }
  else if (c && c != '\n')
  {
    unsigned int cnt = 1;
    col = vi->xy[pagenum].col;
    row = vi->xy[pagenum].row;

    if (c == '\t')
    {
      cnt = (col % 8);
      if (!cnt)
        cnt = 8;
      c = ' ';
    }

    pos = (unsigned short myFAR *)(scr);
    cell = (unsigned short)(0x0700 | c); /* attrib + '\0' */
    if (col || row)
    {
      pos = (unsigned short myFAR *)(scr + (((cols<<1)*row) + (col<<1)) );
      cell = (unsigned short)(((*(pos-1)) & 0xff00) | c);
    }
    for (;cnt && col < cols;cnt--)
    {
      *pos++ = cell;
      col++;
    }
    if (col >= cols)
    {
      col = 0;
      c = '\n';
    }
    vi->xy[pagenum].col = (unsigned char)col;
  }
  if (c == '\n')
  {
    vi->xy[pagenum].col = 0;
    if ((vi->xy[pagenum].row + 1) < rows)
      vi->xy[pagenum].row++;
    else
    {
      myMEMMOVE( scr, scr + (cols << 1), ((rows-1)*(cols<<1)) );
      pos = (unsigned short myFAR *)(scr + ((rows-1)*(cols<<1)) );
      cell = (unsigned short)((*pos & 0xff00) + 0x20); /* attrib + ' ' */
      for (col = 0; col < cols; col++)
        *pos++ = cell;
    }
  }

              /* "sync" - some bios' do not use update the cursor position */
  {           /*          automatically when bios data is changed. */
    cell = (short)(((unsigned)(vi->xy[pagenum].row))<<8)|(vi->xy[pagenum].col);
    __asm mov ah,2
    __asm mov bh,pagenum
    __asm mov dx,cell
    __asm int 10h
  }

  RESTORE_ES_DS();
  return;
}
#define __vid_sync() /* __vid_putch(0) */
#endif

/* ------------------------------------------------------------------- */

static int TTYOUTCHAR(int fd, unsigned int c)
{
  if (c == 0)
    return 0;
  if (((unsigned int)c) >= 0x80u)
    c = '.';
  else if (c == '\a') /* bell */
  {
    __bios_putch(0x07);
    return 1;
  }
  else if (c == '\n')
  {
    if (fd == 2)
      __vid_putch('\r');
    else
     __dos_putch('\r');
  }
  else if (c <= 26 && c != '\r' && c != '\n' && c != '\t' && c != '\b')
  {
    if (fd == 2)
      __vid_putch('^');
    else
     __dos_putch('^');
    c += 'A';
  }
  if (fd == 2)
    __vid_putch(c);
  else
   __dos_putch(c);
  return 1;
}

struct _kfmt
{
  char *buf;
  size_t buflen;
  size_t donelen;
  int fd;
};

/* write one char 'c' to file/tty/buffer. Stop if '\0'.
 * return 1 (continue) if 'c' was written, else return 0 (stop).
*/
static int OUTCHAR(struct _kfmt *kfmt, unsigned char c)
{
  if (!c)
    return 0;
  if (kfmt->buf) /* buffer operation */
  {
    if (kfmt->buflen < 1)
      return 0;
    kfmt->buflen--;
    kfmt->buf[kfmt->donelen++] = c;
    kfmt->buf++;
  }
  else if (kfmt->fd == 1 || kfmt->fd == 2) 
    TTYOUTCHAR(kfmt->fd, c);
  else if (write(kfmt->fd, &c, 1) != 1)
    return 0;
  kfmt->donelen++;
  return 1;
}


/* write 'len' bytes from 'p' to file/tty/buffer. Stop at '\0'.
 * return 1 (continue) if all 'len' bytes were written, else return 0 (stop).
*/
static int OUTBUF(struct _kfmt *kfmt,const char *p, int len, int uplow)
{
  int i, n, rlen, olen = len;
  len = 0;

  if (*p)
  {
    const char *q = (const char *)memchr(p, 0, olen);
    len = olen;
    if (q)
      len = (int)(q - p);
  }

  if (len < 1)
  {
    ;/* fallthrough */
  }
  else if (kfmt->buf) /* buffer operation */
  {
    char *q = kfmt->buf;
    if (kfmt->buflen < ((unsigned int)len))
      len = kfmt->buflen;

    if (uplow == 'u')
    {
      for (i = 0; i < len; i++)
        *q++ = (char)toupper(*p++);
    }
    else if (uplow == 'l')
    {
      for (i = 0; i < len; i++)
        *q++ = (char)tolower(*p++);
    }
    else 
    {
      memcpy( q, p, len );
    }
    kfmt->buflen -= len;
    kfmt->buf += len;
    kfmt->donelen += len;
  }
  else if (kfmt->fd == 1 || kfmt->fd == 2)
  {
    rlen = len;
    len = 0;
    n = kfmt->fd;
    while (rlen > 0)
    {
      if (uplow == 'u')
        i = TTYOUTCHAR(n, toupper(*p++));  
      else if (uplow == 'l')
        i = TTYOUTCHAR(n, tolower(*p++));  
      else
        i = TTYOUTCHAR(n, *p++);  
      if (!i)
        break;
      rlen--;
      len++;
    }
    kfmt->donelen += len;
  }
  else if (uplow == 'u' || uplow == 'l')
  {
    rlen = len;
    len = 0;
    while (rlen > 0)
    {
      char temp[32];
      n = rlen;
      if (n > (int)sizeof(temp))
        n = sizeof(temp);
      for (i = 0; i < n; i++)
        temp[i] = (char)((uplow == 'u')?(toupper(*p++)):(tolower(*p++)));
      i = write(kfmt->fd, temp, n);
      if (i > 0)
        len += i;
      if (i < n)
        break; 
      rlen -= n;
    }
    kfmt->donelen += len;
  }
  else
  {
    len = write(kfmt->fd, p, len);
    if (len > 0)
      kfmt->donelen += len;
  }
  return (len == olen);
}


static int __do_pad(struct _kfmt *kfmt, int width, int len, int fillch )
{
  if (width > len)
  {
    unsigned int n;
    for (n = (width - len); n; n--)
    {
      if (!OUTCHAR (kfmt, ((char)fillch) ))
        return 0;
    }  
  }
  return 1;
}

static int __do_k (struct _kfmt *kfmt, const char *fmt, va_list args)
{
  const char *f, *_fmt = fmt;
  int      saved_errno = errno;  /* for use in "%m" */
  int      c, i, n, len, width, prec, fmtctrl, fillch, caseconv;
  long     l;
  unsigned int radix;
  unsigned long val = 0L;
  char     *str;
  unsigned char *p;
  char     fmtmod;
  char     num[32];

  if (!fmt)
    return 0;

  if (kfmt->buf)
  {
    if (kfmt->buflen == 0)
      return 0;
    kfmt->buflen--;
  }
  kfmt->donelen = 0;

  for (;;)
  {
    for (f = _fmt; *f != '%' && *f != 0; ++f)
        ;
    if (f > _fmt)
    {
      len = (int)(f - _fmt);
      if (!OUTBUF(kfmt, _fmt, len, 0))
        break;
      _fmt = f;
    }
    if (*_fmt == 0)
       break;
    c = *++_fmt;

    width = prec = fmtctrl = caseconv = 0;
    fillch = ' ';

    #define FMTCTRL_LJUST 0x01 /* - */
    #define FMTCTRL_PLUS  0x02 /* + */
    #define FMTCTRL_SPACE 0x04 /*   */
    #define FMTCTRL_POUND 0x08 /* # */
    #define FMTCTRL_NEG   0x10 /* if value is negative */
    #define FMTCTRL_QUOTE 0x20 /* if format is quoted */
    #define FMTCTRL_ERRNO 0x40 /* special for "%m" */

    for (;;)
    {
      if (c == '-')      fmtctrl |= FMTCTRL_LJUST;
      else if (c == '+') fmtctrl |= FMTCTRL_PLUS;
      else if (c == ' ') fmtctrl |= FMTCTRL_SPACE;
      else if (c == '#') fmtctrl |= FMTCTRL_POUND;
      else break;
      c = *++_fmt;
      /* repeat */
    }
    if (c == '0')
    {
      fillch = '0';
      c = *++_fmt;
    }
    if (c == '*')
    {
      width = va_arg (args, int);
      c = *++_fmt;
    }
    else
    {
      while (isdigit(c))
      {
        width = width * 10 + c - '0';
        c = *++_fmt;
      }
    }
    if (width < 0) /* A negative field width value is interpreted as a left-*/
    {              /* justification flag, followed by a positive field width.*/
      fmtctrl |= FMTCTRL_LJUST;
      width = -width;
    }

    if (c == '.')
    {
      c = *++_fmt;
      if (c == '*')
      {
        prec = va_arg (args, int);
        c = *++_fmt;
      }
      else
      {
        while (isdigit(c))
        {
          prec = prec * 10 + c - '0';
          c = *++_fmt;
        }
      }
    }

    num[0]= '\0';
    str   = NULL;
    radix  = 0;
    fmtmod= 0;
    len  = -1;
    ++_fmt;

again:
    switch (c)
    {
      case 'd':
      case 'i':
           if (fmtmod == 'l')
             l = va_arg (args, long); 
           else
             l = va_arg (args, int);
           val = l;
           if (l < 0)
           {
             fmtctrl |= FMTCTRL_NEG;
             val = -l;
           }
           radix = 10;
           break;

      case 'u':
           if (fmtmod == 'l')
             val = va_arg (args, unsigned long); 
           else
             val = va_arg (args, unsigned int);
           radix = 10;
           break;

      case 'o':
           if (fmtmod == 'l')
             val = va_arg (args, unsigned long); 
           else
             val = va_arg (args, unsigned int);
           radix = 8;
           break;

      case 'X':
           caseconv = 'u';
           /* fall through */
      case 'x':
           if (fmtmod == 'l')
             val = va_arg (args, unsigned long); 
           else
             val = va_arg (args, unsigned int);
           radix = 16;
           break;

      case 'p':
           {
             typedef void (*codeptr_t)(void);
             val = (unsigned long)va_arg (args, codeptr_t );
             len = (sizeof(codeptr_t)<<1); 
             str = &num[sizeof(num)];
             *--str = '\0';
             for (i=0; i < len; i++)
             {
               #if (UINT_MAX == USHRT_MAX)
               if (len == 8 && i == 4)
                 *--str = ':';
               #endif
               c = (char)(val & 0x0f);
               val >>= 4;
               if (c > 9)
                 c += (char)('A'-('9'+1));
               *--str = (char)('0' + c);
             }  
             /*
             *--str = 'x';
             *--str = '0';
             */
             len = (int)((&num[sizeof(num)] - str) - 1);
           }
           break;

      case 'S':
           caseconv = 'u';
           /* fallthrough */
      case 's':
           str = va_arg (args, char*);
           if (!str) 
             str = "(null)";
           break;

      case 'C':
           caseconv = 'u';
           /* fallthrough */
      case 'c':
           num[0] = (char)va_arg (args, int);
           num[1] = 0;
           str = num;
           if (num[0] == '\0') /* got a terminating '\0' */
           {
             _fmt = num;
             continue; /* break out of topmost for() loop */
           }
           break;

      case 'a': 
           caseconv = 'l';  
           /* fallthrough */
      case 'A': /* basename.ext of appname ('l' modifier == full path) */
           str = __get_argv0();
           if (!str)
             str = ""; /* don't do "(null)" etc here. may be _sprintk() */
           else if (fmtmod != 'l') /* not full path */
           {
             n = strlen(str);
             while (n > 0 && str[n-1] != '\\' && str[n-1] != '/')
               n--;
             str += n;
           }
           break;

      case 'M':                         /* ether (MAC) address */
           str = va_arg (args, char*);
           if (!str) 
             str = "(null)";
           else 
           {
             len = 0;
             for (i = 0;i < 6; i++)
             {
               if (i > 0)
                 num[len++] = ':';
               c = ((str[i] >> 4) & 0x0f);
               if (c > 9) 
                 c += ('a'-('9'+1));
               num[len++] = (char)('0' + c);
               c = (str[i] & 0x0f);
               if (c > 9) 
                 c += ('a'-('9'+1));
               num[len++] = (char)('0' + c);
             } 
             str = num;
           }
           break;

      case 'm':                         /* strerror(errno) */
           #if defined(__SMALL__) || defined(__TINY__) || defined(__COMPACT__)
           val = saved_errno;
           radix = 10;
           fmtctrl |= FMTCTRL_ERRNO;
           #else
           str = strerror(saved_errno);
           len = 0;
           while (str[len] && str[len] != '\r' && str[len] != '\n')
             len++;
           #endif
           break;

      case 'I':                         /* IP address (net order) */
           val = va_arg(args,unsigned long);
           if (fmtmod == 'h')           /* IP address (host order) */  
           {
             i = 1;
             if (*((char *)&i) == 1) /* its a little endian machine */
               val = (val<<24)|((val&0xff00)<<8)|((val>>8)&0xff00)|(val>>24);
           }
           p = (unsigned char *)&val;
           str = num;
           for (i = 0; i < 4; i++)
           {
             if (i) 
               *str++ = '.';
             if (*p > 100) 
               *str++ = (char)('0'+(*p / 100));
             if (*p > 10)  
               *str++ = (char)('0'+((*p % 100)/10));
             *str++ = (char)('0'+(*p % 10)); 
             p++;
           }
           *str = '\0';
           str = num;
           break;

      case 't':                               /* current time */
           __cmos_ctime_r(num, sizeof(num));
           str = num;
           str += 4;           /* chop off the day name */
           len = 15;           /* chop off year and newline */
           break;

      case 'v':                /* "visible" string */
      case 'q':                /* quoted string */
           {
             int quoted = c == 'q';
             p = va_arg (args, unsigned char *);
             if (fillch == '0' && prec > 0)
                n = prec;
             else
             {
               n = strlen ((char*)p);
               if (prec > 0 && prec < n)
                  n = prec;
             }
             len = 0;
             while (n > 0)
             {
               if (len >= (int)(sizeof(num)-7))
               {
                 if (!OUTBUF(kfmt, num, len, 0))
                   break;
                 len = 0;
               }
               c = (*p++ & 0xff);
               --n;
               if (!quoted && c >= 0x80)
               {
                 num[len++] = 'M';
                 num[len++] = '-';
                 c -= 0x80;
               }
               if (quoted && (c == '"' || c == '\\'))
                 num[len++] = '\\';
               if (c >= 0x20 && !(0x7F <= c && c < 0xA0))
                 num[len++] = (char)c;
               else if (quoted)
               {
                 num[len++] = '\\';
                 switch (c)
                 {
                   case '\r': num[len++] = 'r'; break;
                   case '\t': num[len++] = 't'; break;
                   case '\n': num[len++] = 'n'; break;
                   case '\b': num[len++] = 'b'; break;
                   case '\f': num[len++] = 'f'; break;
                   default  : 
                   {
                     num[len++] = 'x';
                     i = (c >> 4);
                     num[len++] = (char)((i <= 9) ? ('0'+i) : ('a'+(i-10)));
                     i = c & 0x0f;
                     num[len++] = (char)((i <= 9) ? ('0'+i) : ('a'+(i-10)));
                     break;
                   }
                 }
               }
               else
               {
                 if (c == '\t')
                   num[len++] = (char)c;
                 else
                 {
                   num[len++] = '^';
                   num[len++] = (char)(c ^ 0x40);
                 }
               }
             } /* while (n > 0) */
             if (n > 0) /* out of space */
               break;
             if (len)
             {
               if (!OUTBUF(kfmt, num, len, 0))
                 break;
               len = 0;
             }
           }
           continue;

      case 'l':
      case 'h':
      case 'n':
      case 'f':
           if (fmtmod == 0 && *_fmt)
           {
             fmtmod = (char)c;
             c = *_fmt++;
             goto again;
           }
           /* fallthrough */ 

      default:
           if (!OUTCHAR(kfmt,'%'))
             break;
           if (c != '%')
              --_fmt;        /* so %z outputs %z etc. */
           continue;
    }

    if (radix) /* its numeric */
    {
      str = num + sizeof(num);
      *--str = 0;

      if (val == 0)
      {
        *--str = '0'; 
        /* "0" cancels non-decimal fmtctrl */
        fmtctrl &= ~FMTCTRL_POUND; 
      } 
      else if (radix == 10)
      {
        #if (UINT_MAX == USHRT_MAX) /* 16 bit target */
        unsigned short vhi = *(((unsigned short *)&val)+1);
        unsigned short vlo = *(((unsigned short *)&val)+0);
        while (vlo || vhi)
        {
          if (vhi < 10)
          {
            __asm mov  dx,vhi
            __asm mov  ax,vlo
            __asm mov  cx,10
            __asm div  cx
            __asm xor  cx,cx
            __asm mov  vhi,cx
            __asm mov  vlo,ax
            __asm mov  i,dx
          }
          else
          {
            __asm mov  ax,vhi
            __asm xor  dx,dx
            __asm mov  cx,10
            __asm div  cx
            __asm mov  vhi,ax
            __asm mov  ax,vlo
            __asm div  cx
            __asm mov  vlo,ax
            __asm mov  i,dx
          }
          *--str = (char)('0' + i);
        }
        #else  
        while (val)
        {
          c = (char)(val % 10ul);
          val /= 10ul;
          *--str = (char)('0' + c);
        }
        #endif
      }
      else if (radix == 16) 
      {
        while (val)
        {
          c = (unsigned char)(val & 15);
          val >>= 4; 
          if (c > 9)
            c += 7;
          *--str = (char)('0' + c);
        } 
      }
      else if (radix == 8) 
      {
        while (val)
        {
          c = (unsigned char)(val & 7);
          val >>= 3; 
          *--str = (char)('0' + c);
        }
      }
      n = (int)(num + sizeof(num) - 1 - str); /* sizeof(number) without '-' etc */

      if (fmtctrl & FMTCTRL_POUND)
      {
        if (radix == 16)
          *--str = 'x';
        *--str = '0';
      }
      if (fmtctrl & FMTCTRL_NEG) /* "%d" format and value was signed */
        *--str = '-';
      else if (fmtctrl & FMTCTRL_PLUS) /* '+' format control flag */
        *--str = '+';
      else if (fmtctrl & FMTCTRL_SPACE) /* ' ' format control flag */
        *--str = ' ';
      if (fmtctrl & FMTCTRL_ERRNO)
      {
        *--str = '#'; *--str = ' '; *--str = 'r'; *--str = 'r'; *--str = 'E';
      }
      len = (int)(num + sizeof(num) - 1 - str);

      if (prec > 0 && len > prec)
        len = prec;

      /* is this correct? 
      if (width > 0 && len < width && (fmtctrl & FMTCTRL_LJUST)==0)
      {
        if (!OUTBUF(kfmt,s,len-n, caseconv))
          break;
        s += (len-n);
        len = n;
      }
      */
    }
    else /* radix == 0, ie its not a number */
    {
      if (len == -1) /* no override */
      {
        if (prec < 1)
          len = strlen (str);
        else /* might not be '\0' terminated */
        {
          len = 0;
          while (len < prec && str[len])
            len++;
        }
      }
      if (prec > 0 && len > prec)
        len = prec;
      if (width > 0 && width < len)
        width = len;
    }
    if (width > 0 && (fmtctrl & FMTCTRL_LJUST)==0)
    {
      /* _printk("%5s", 'foo') => '  foo' */
      if (!__do_pad( kfmt, width, len, fillch ))
        break;
    }
    if (!OUTBUF(kfmt,str,len, caseconv))
      break;
    if (width > 0 && (fmtctrl & FMTCTRL_LJUST)!=0)
    {
      /* _printk("%-5s", 'foo') => 'foo  ' */
      if (!__do_pad( kfmt, width, len, fillch ))
        break;
    }
  }
  if (kfmt->buf)
  {
    *(kfmt->buf) = 0;
  }
  else if (kfmt->fd == 2 && /* direct video, aka "stderr" */
           kfmt->donelen)   /* wrote something to video */
  {
    __vid_sync();
  }
  return kfmt->donelen;
}

/*
 * _vsnprintk - like vsprintf, except we also specify the length of the 
 * output buffer, and we handle %m (error message), %t (current time) 
 * and %I (IP address) formats. Doesn't do floating-point formats.
 * Returns the number of chars put into buf.
 */
int _vsnprintk (char *buf, int buflen, const char *fmt, va_list args)
{
  struct _kfmt kfmt;
  kfmt.buf = buf;
  kfmt.buflen = buflen;
  kfmt.fd = -1;
  return __do_k(&kfmt, fmt, args );
}

/*
 *  _snprintk - like sprintf() except we also specify the length of the 
 *  output buffer, and we handle %m (error message), %t (current time) 
 *  and %I (IP address) formats. Doesn't do floating-point formats.
 *  Returns the number of chars put into buf.
 */
int _snprintk (char *buf, int buflen, const char *fmt, ...)
{
  int     len;
  va_list args;
  va_start (args, fmt);
  len = _vsnprintk (buf, buflen, fmt, args);
  va_end (args);
  return (len);
}


int _vsprintk (char *buf, const char *fmt, va_list args)
{
  return _vsnprintk (buf, INT_MAX, fmt, args);
}  

int _sprintk (char *buf, const char *fmt, ...)
{
  int     len;
  va_list args;
  va_start (args, fmt);
  len = _vsnprintk (buf, INT_MAX, fmt, args);
  va_end (args);
  return (len);
}  

static int my_outs(int fd, int flags, int len, const char *str)
{
  int wlen = 0;
  if (len == -1)
    len = strlen(str);
  while (len > 0)
  {
    if (str[0] != '\r' || str[1] != '\n') /* '\n' would be expanded */
      wlen += TTYOUTCHAR(fd, *str);  
    str++;
    len--;
  }
  if (flags & 0x01)
    wlen += TTYOUTCHAR(fd, '\n');
  return wlen;  
}

/*
** Interrupt safe and reentrant version of [v]printf()
*/
int _vhprintk( int fd, const char *fmt, va_list args)
{
  auto void (*sighandler)(int);
  struct _kfmt kfmt;
  int len = -1;

  if (fd <= 0 || fd > 255) /* stdin or no such descriptor */
    return 0;

  sighandler = 0;
  if (fd != 2)  
    sighandler = signal(SIGINT, SIG_IGN);

  if (fd == 1 || fd == 2) /* some outsxx optimization */
  {
    if (fmt[0] != '%')
    {
      if (fmt[0] == '\0')
        len = 0;
      if (fmt[1] == '\0')
        len = TTYOUTCHAR(fd, fmt[0] );          
    }
    else if (fmt[1] == 's')
    {
      if (fmt[2] == '\0')                                        /* outs */
        len = my_outs(fd, 0x00, -1, va_arg (args, const char*));
      else if (fmt[2] == '\n' && fmt[3] == '\0')                 /* outsnl */
        len = my_outs(fd, 0x01, -1, va_arg (args, const char*));
    }
    else if (fmt[1] == 'c' && fmt[2] == '\0')                    /* outch */
      len = TTYOUTCHAR(fd, (char)va_arg (args, int));  
    else if (fmt[2] == '.' && fmt[3] == '*' && fmt[4] == 's'     /* outsn */
             && fmt[5] == '\0')
    {
      len = va_arg (args, int);
      len = my_outs(fd, 0x00, len, va_arg (args, const char*));
    }
    else if (!strchr(fmt,'%'))
    {
      len = 0;
      while (*fmt)
        len += TTYOUTCHAR(fd, *fmt++ );
    }
  }

  if (len == -1)
  {
    kfmt.buf = NULL;
    kfmt.buflen = 0;
    kfmt.fd = fd;
    len = __do_k ( &kfmt, fmt, args);
  }

  if (fd != 2)
    signal(SIGINT, sighandler);

  return len;
}
  
int _hprintk( int fd, const char *fmt, ...)
{
  int len;
  va_list args;
  va_start (args, fmt);
  len = _vhprintk( fd, fmt, args);
  va_end (args); 
  return len;
}

int _printk (const char *fmt, ...)
{
  int len;
  va_list args;
  va_start (args, fmt);
  len = _vhprintk( 2, fmt, args);
  va_end (args); 
  return len;
}
