/*
** Temporary signal blocking/unblocking for kernel functions that wait.
**
** Written May 2001 by Cyrus Patel <cyp@fb14.uni-mainz.de>
** 
** signal blocking/unblocking for all signals that cause a socket operation
** to be (partially-) cancelled [aka EINTR]. For most libcs this is limited 
** to SIGINT/SIGBREAK. Some (DJGPP only?) also require SIGALRM handling.
**
** Usage:
**
** - To disable default signal handling use set_cbreakstate(0). 
**   signals will then be "blocked" (but flagged as pending).
** - To re-enable default signal handling use set_cbreakstate(1). 
**   A signal that may have occurred while blocked will then
**   get the default signal handling (which could well be an exit()).
** - To determine if one of the signals is pending (only while they 
**   are blocked of course) use get_cbreakpend(). 
**
** Readers familiar with DJGPP's _go32_want_ctrl_break() and/or
** _go32_was_ctrl_break_hit() will recognize that set_cbreakstate()
** and get_cbreakpend() are much like them. The differences are:
** - get_cbreakpend() and set_cbreakstate() refer to all the necessary sigs. 
**   (Note: If the user had __djgpp_set_ctrl_c(0) [ie, ^C being read as 
**   normal keyboard input], then ^Cs will not be seen as SIGINT. Fine.)
** - get_cbreakpend() will not automatically call set_cbreakstate().
**
** Implementation notes:
**
** SIGBREAK handling may not be necessary for all libraries, since the
** default handler for 0x1B simply causes dos to treat it as a ^C.
** However, most (if not all) DOS extenders hook this vector in anticipation
** that the application will, and the default handling in those cases is
** to simply terminate the application directly with Int21h/ah=4ch.
*/
#undef  TRACE_SIGHANDLER /* CAREFUL!! Has Stack-overflow issues */
#undef  TRACE_SETSTATE

#define WANT_SIGBLK_SUPPORT

#undef WANT_STACK_CHECKING /* disable stack checking for this module */

#include "target.h"   /* standard includes */
#include "ctordtor.h" /* _my_atexit(), __kern_is_exiting() */
#include "dpmi.h"     /* __dpmi_*() */
#include "timer.h"    /* timer_*() */
#include "blocker.h"  /* blocker_*() */
#include "debug.h"    /* DBGPRINTF() */
#include "printk.h"   /* signal safe debugging */
#include "sigblk.h"   /* ourselves */

#include <setjmp.h>
#include <signal.h>

/* --------------------------------------------------------------------- */

typedef void (*sighandler_t)(int);

#if defined(__TURBOC__)
#undef  SIG_IGN  /* originally defined as ((void _Cdecl (_FARFUNC *)())1) */
#define SIG_IGN ((void _Cdecl (_FARFUNC *)(int))1)
#undef  SIG_DFL
#define SIG_DFL ((void _Cdecl (_FARFUNC *)(int))0)
#undef  SIG_ERR
#define SIG_ERR ((void _Cdecl (_FARFUNC *)(int))-1)
#endif

#if defined(__DJGPP__) 
#define SIGBREAK SIGQUIT /* doesn't have SIGBREAK, but has discrete SIGQUIT */
#include <go32.h> /* want to get __djgpp_cbrk_count without the side */
                  /* effects of _go32_was_ctrl_break_hit() */
#endif

/* --------------------------------------------------------------------- */

#ifdef TRACE_SIGHANDLER
#undef TRACE_SIGHANDLER
#define TRACE_SIGHANDLER(x) _printk x
#else
#define TRACE_SIGHANDLER(x) NO_OPERATION()
#endif 

#ifdef TRACE_SETSTATE
#undef TRACE_SETSTATE
#define TRACE_SETSTATE(x) DBGPRINTF(x)
#else
#define TRACE_SETSTATE(x) NO_OPERATION()
#endif 

/* --------------------------------------------------------------------- */

#if defined(WANT_SIGBLK_SUPPORT)
static struct 
{
  volatile unsigned int brkseencount; /* increment on sig catch */
  #define GOT_INT   (1)
  #define GOT_BREAK (0)
  #define GOT_ALRM  (0)
  #if defined(SIGBREAK)
    #undef GOT_BREAK
    #define GOT_BREAK (1)
  #endif
  #if defined(SIGALRM)
    #undef GOT_ALRM
    #define GOT_ALRM (1)
  #endif
  struct {
    int signo;
    sighandler_t oldhandler;
    int pending;
  } sigstates[GOT_INT+GOT_BREAK+GOT_ALRM];
  #undef GOT_ALRM
  #undef GOT_BREAK
  #undef GOT_ALRM 
  int  exit_allowed;
  int  origbrkmode; /* as determined during initialization. Restored on exit() */ 
  int  exit_seen;   /* ignore all further sigints */
  jmp_buf sig_jmp;
#if defined(__DJGPP__) 
  int old__djgpp_cbrk_count; /* these two are saved when beginning blocking */
  int old_ctrl_break_counting; /* and restored when ending blocking */
#endif
} brkstatics;
#endif

/* --------------------------------------------------------------------- */

#if defined(WANT_SIGBLK_SUPPORT)
static int __find_sig_state(int sig)
{
  unsigned int i = 0;
  while (i < (sizeof(brkstatics.sigstates)/sizeof(brkstatics.sigstates[0])))
  {
    if (brkstatics.sigstates[i].signo == sig)
      return i;
    i++;
  }
  return 0; /* should never happen */
}
#endif

/* --------------------------------------------------------------------- */

/* clear the keyboard queue. must be done without using DOS */  
#if defined(WANT_SIGBLK_SUPPORT)
static void __kbclear(void)
{                          
  /* clearing the keyboard queue is necessary to remove the ^C */
#if defined(__HIGHC__) && defined(PHARLAP_H) 
  /* only highc has PokeRealWord() */
  REALPTR headp, tailp;
  RP_SET(headp, 0x0040, 0x001A);
  RP_SET(tailp, 0x0040, 0x001C);
  PokeRealWord(tailp,PeekRealWord(headp));
#elif defined(__DPMI32__) /* all DPMI targets (including pharlap) */
  unsigned short headtail[2];
  dosmemget( ((0x0040ul<<4)|0x001Aul), sizeof(headtail), &headtail[0] );
  headtail[0] = headtail[1];
  dosmemput( &headtail[0], sizeof(headtail), ((0x0040ul<<4)|0x001Aul) );
  /* a rare instance of DJGPP's _farpeek/poke being useful in this lib: */
  /* _farpokew( _dos_ds, 0x41a, _farpeekw(_dos_ds, 0x41c) ); */
#else /* real mode */
  short _far *headp = (short _far *)MK_FP(0x0040, 0x001A);
  short _far *tailp = (short _far *)MK_FP(0x0040, 0x001C);
  *headp = *tailp;
#endif
}
#endif

/* --------------------------------------------------------------------- */

#if defined(WANT_SIGBLK_SUPPORT)
static void __process_sig( int sig, int from_raise );
static void __sig_handler( int sig ) { __process_sig( sig, 0 ); }
#endif

/* --------------------------------------------------------------------- */

#if defined(WANT_SIGBLK_SUPPORT)
static void __process_sig( int sig, int from_raise )
{
  /* CAUTION: don't do any I/O (including con) from here. The signal
  ** handler may be called for an Int 1Bh, and DOS may be active.
  */
  int sig_is_brk_type = 1;
  sighandler_t oldhandler = (sighandler_t)0;
  int *pending = (int *)0;
  unsigned int sigidx;

  #if defined(SIGALRM)
  if (sig == SIGALRM) 
  {
    /* sig_is_brk_type = 0; */    
    TRACE_SIGHANDLER(("\ncaught SIGALRM\n"));
  }
  #endif
  #if defined(SIGINT)
  if (sig == SIGINT)
  {
    TRACE_SIGHANDLER(("\ncaught SIGINT\n"));
  }
  #endif
  #if defined(SIGBREAK)
  if (sig == SIGBREAK)
  {
    TRACE_SIGHANDLER(("\ncaught SIGBREAK\n"));
  }
  #endif

  if (__kern_is_exiting())
  {
    TRACE_SIGHANDLER(("simple return since __kern_is_exiting()\n"));
    return;
  }

  sigidx = __find_sig_state(sig);
  oldhandler = brkstatics.sigstates[sigidx].oldhandler;
  pending = &(brkstatics.sigstates[sigidx].pending);

  if (!from_raise)
  {
    TRACE_SIGHANDLER(("calling __kbclear()\n"));
    __kbclear(); /* clear the keyboard without using DOS */
    TRACE_SIGHANDLER(("end __kbclear()\n"));
    if (sig_is_brk_type)
    {
      if ((++brkstatics.brkseencount)==0)
        ++brkstatics.brkseencount;
      TRACE_SIGHANDLER(("new brkstatics.brkseencount=%u\n",
                         brkstatics.brkseencount ));
    }
  }

  if (!brkstatics.exit_allowed || /* break "handling" disabled */
      brkstatics.exit_seen || /* already long jumped */
      oldhandler == SIG_IGN )
  {
    if (!from_raise)
      *pending = 1;
    signal( sig, ((brkstatics.exit_seen) ? SIG_IGN : __sig_handler) );
    /* as far as 'oldhandler' is concerned, no signal arrived */
    return;
  }

  if (oldhandler == SIG_DFL )
  {                                   
    /* the default for all signals we catch (ALRM/INT/BREAK) is to die. We do
    ** that, but we longjmp first so that all routines called via atexit() 
    ** hooks will have enough stack. - Note that any DOS call may be safely
    ** made from an INT 23h (SIGINT) handler, so for SIGINT we exit(). For 
    ** an Int 1Bh (SIGBREAK) call, we need to _exit(). 
    ** NB: All DJGPP signals are DOS safe.
    */
    #if defined(__DJGPP__)
    exit(sig); /* note 'exit()', not '_exit()' (see libc/go32/dpmiexcp.c) */
    #else
    TRACE_SIGHANDLER(("doing brkstatics.sig_jmp (from_raise=%u)\n",from_raise));
    longjmp(brkstatics.sig_jmp, ((from_raise)?(-sig):(+sig)) );
    #endif
  }
  else
  {
    /* not SIG_DFL, so call the old handler,
    ** which may or may not return.
    */
    signal(sig, __sig_handler);
    (*(oldhandler))(sig);
    /* handler did return, so point sighandler back to ourselves */ 
    brkstatics.sigstates[sigidx].oldhandler = 
            (sighandler_t)signal(sig, __sig_handler);
  }
  return;
}
#endif

/* --------------------------------------------------------------------- */

#if defined(WANT_SIGBLK_SUPPORT)
static int __extbreakchk_onoff(int on)
{
  /* INT 21 - DOS 2+ - EXTENDED BREAK CHECKING
  **   AH = 33h
  **   AL = subfunction
  **      00h get current extended break state
  **          Return: DL = current state, 00h = off, 01h = on
  **      01h set state of extended ^C/^Break checking
  **          DL = new state
  **              00h off, check only on character I/O functions
  **              01h on, check on all DOS functions
  **          Return: (Novell DOS 7) DL = old state of extended Break checks
  */
  int old_mode;
#if defined(__DPMI32__) && !defined(__HIGHC__)
  /* 
  ** TCC/BCC: have a working setcbrk() but don't bother using it.
  ** DJGPP: has setcbrk() but is buggy (assumes value of DL on return
  **        is the old break state, which is incorrect).
  ** Watcom C: has break_on()/break_off(), but those return void.
  */
  __dpmi_regs reg;
  reg.d.eax = 0x3300;
  TRACE_SETSTATE(("__extbreakchk_onoff 30\n"));
  __dpmi_int (0x21, &reg);
  TRACE_SETSTATE(("__extbreakchk_onoff 31\n"));
  old_mode = reg.h.dl;
  if (on) on = 1;
  if (on != old_mode)
  {  
    reg.d.eax = 0x3301;
    reg.d.edx = on; /* 0=^C/^Break only on con I/O, 1=on any DOS call */
    TRACE_SETSTATE(("__extbreakchk_onoff 32\n"));
    __dpmi_int (0x21, &reg);
    TRACE_SETSTATE(("__extbreakchk_onoff 33\n"));
  }
#else /* __HIGHC__ && real mode */
  union REGS reg;
  reg.h.ah = 0x33; /* reg.w.ax = 0x3300; but */
  reg.h.al = 0x00; /* BCC/TCC <= 3.1 has no .w. member */
  TRACE_SETSTATE(("__extbreakchk_onoff 34\n"));
  int86 (0x21, &reg, &reg);
  TRACE_SETSTATE(("__extbreakchk_onoff 35\n"));
  old_mode = reg.h.dl;
  if (on) on = 1;
  if (on != old_mode)
  {
    reg.h.ah = 0x33; /* reg.w.ax = 0x3301; but */
    reg.h.al = 0x01; /* BCC/TCC <= 3.1 has no .w. member */
    reg.h.dl = (char)on; /* 0=^C/^Break only on con I/O, 1=on any DOS call */
    TRACE_SETSTATE(("__extbreakchk_onoff 36\n"));
    int86 (0x21, &reg, &reg);
    TRACE_SETSTATE(("__extbreakchk_onoff 37\n"));
  }
#endif
  return old_mode;
}
#endif

/* --------------------------------------------------------------------- */

#if defined(WANT_SIGBLK_SUPPORT)
static void __atexit_handler(void)
{
  __extbreakchk_onoff(brkstatics.origbrkmode - 1);
}
#endif

/* --------------------------------------------------------------------- */

/*
 * new_mode = 0: begin blocking: ^C/^Break does NOT cause exit
 * new_mode = 1: end blocking: ^C/^Break may cause exit
*/
#if defined(WANT_SIGBLK_SUPPORT)
static int __set_brk_mode(int new_state)
{
  unsigned int i;
  int old_state;

  old_state = brkstatics.exit_allowed;
  brkstatics.exit_allowed = new_state;

  TRACE_SETSTATE(("__set_brk_mode 14\n"));
  if (!new_state && old_state)     /* enter nest */
  {
    if (brkstatics.origbrkmode == 0) /* first time */
    {
      TRACE_SETSTATE(("__set_brk_mode 15\n"));
      brkstatics.origbrkmode = 1 + __extbreakchk_onoff(1);
      TRACE_SETSTATE(("__set_brk_mode 16\n"));
 
      /* KMSG((0,"_my_atexit(__cbrk_shutdown)\n")); */
      _my_atexit(__atexit_handler);
      TRACE_SETSTATE(("__set_brk_mode 17\n"));
    }
    i = 0;
    while (i < sizeof(brkstatics.sigstates)/sizeof(brkstatics.sigstates[0]))
      brkstatics.sigstates[i++].pending = 0;
    brkstatics.brkseencount = 0;
    TRACE_SETSTATE(("__set_brk_mode 18\n"));
    timer_syst_enter();
    TRACE_SETSTATE(("__set_brk_mode 19\n"));

#if defined (__DJGPP__) /* don't let the ^Break kill the app */
    /* We should probably have a signal handler for signal(0x1B)
    ** that simply converts the ^Break to SIGQUIT
    */
    brkstatics.old__djgpp_cbrk_count = __djgpp_cbrk_count;
    brkstatics.old_ctrl_break_counting = (__djgpp_hwint_flags & 2);
    _go32_want_ctrl_break(1); /* _might_ clear the count */
    __djgpp_cbrk_count = 0; /* make sure */
    /* note that we do not tinker with __djgpp_set_ctrl_c() for DJGPP - 
    ** if the user had __djgpp_set_ctrl_c(0) [ie, ^C being read as normal
    ** keyboard input], then the signal handler won't get called. Fine.
    ** If the user had __djgpp_set_ctrl_c(non-zero) then OUR signal handler
    ** will be the one being called. Alls well that ends well.
    */
#endif
  }

  if (new_state && !old_state)  /* exit nest */
  {
#if defined(__DJGPP__)
    int cnt = __djgpp_cbrk_count; /* pick up any not seen by get_cbrkpend() */
    if (cnt)
    {
      i = __find_sig_state(SIGBREAK);
      brkstatics.sigstates[i].pending = 1;
      brkstatics.brkseencount += cnt;
      __djgpp_cbrk_count = 0;
    }
    _go32_want_ctrl_break(brkstatics.old_ctrl_break_counting);
    __djgpp_cbrk_count = brkstatics.old__djgpp_cbrk_count;
#endif
    timer_syst_exit();

    i = 0;
    while (i < sizeof(brkstatics.sigstates)/sizeof(brkstatics.sigstates[0])) 
    {
      if (brkstatics.sigstates[i].signo > 0)
      {
        if (brkstatics.sigstates[i].pending)
          __process_sig( brkstatics.sigstates[i].signo, 1 /* fake raise */ );
        brkstatics.sigstates[i].pending = 0;
        if (brkstatics.sigstates[i].oldhandler != SIG_DFL)
          signal( brkstatics.sigstates[i].signo, 
                  brkstatics.sigstates[i].oldhandler );
      }
      i++;
    }
  }

  return old_state;
}
#endif

/* --------------------------------------------------------------------- */

#if defined(WANT_SIGBLK_SUPPORT)
static void __set_sig_handler_if_needed(void)
{
  if (!brkstatics.exit_seen)
  {
    unsigned int i;
    for (i = 0; 
         i < sizeof(brkstatics.sigstates)/sizeof(brkstatics.sigstates[0]); 
         i++)
    {
      int sig = brkstatics.sigstates[i].signo;
      if (sig > 0)
      {
        sighandler_t oldhandler = signal(sig, (sighandler_t)__sig_handler);
        if (oldhandler != ((sighandler_t)__sig_handler) )
          brkstatics.sigstates[i].oldhandler = oldhandler;
      }
    }
  }
  return;
}
#endif

/* --------------------------------------------------------------------- */

#if defined(WANT_SIGBLK_SUPPORT)
static void __init_statics_if_needed(void)
{
  static int my_bss_count = -1;

  TRACE_SETSTATE(("__init_statics_if_needed 3\n"));
  if (my_bss_count < __bss_count) /* always LT compare! */
  {
    int sig;
    /* initialization order is important here */
    memset(&brkstatics, 0, sizeof(brkstatics));

    if ((sig = setjmp(brkstatics.sig_jmp))!=0)
    {
      int exit_is_safe = 0; /* assume came here from an Int1Bh handler */

      if (sig < 0) /* raise() was used */
      {
        exit_is_safe = 1;
        sig = -sig;
      }
      #if defined(__DJGPP__)
      exit_is_safe = 1; /* all djgpp signals are DOS safe */
      #else
      if (sig == SIGINT) /* any dos call may be made from an int23h handler */
        exit_is_safe = 1; 
      else if (blocker_get_indos_flag() == 0)
        exit_is_safe = 1;
      #endif

      brkstatics.exit_seen = sig;
      signal (sig, SIG_IGN); /* don't want any handling any more */

      /*_printk("doing %sexit()\n", ((exit_is_safe) ? "" : "_") );*/

      if (exit_is_safe)
        exit(sig); 
      _exit(sig); 
    }

    TRACE_SETSTATE(("__init_statics_if_needed 4\n"));
    sig = 0;
    brkstatics.sigstates[sig++].signo = SIGINT;
    #if defined(SIGBREAK)
    brkstatics.sigstates[sig++].signo = SIGBREAK;
    #endif
    #if defined(SIGALRM)
    brkstatics.sigstates[sig++].signo = SIGALRM;
    #endif
    while (sig < 
      (int)(sizeof(brkstatics.sigstates)/sizeof(brkstatics.sigstates[0])))
    {
      brkstatics.sigstates[sig++].signo = -1;
    }

    TRACE_SETSTATE(("__init_statics_if_needed 5\n"));
    __set_sig_handler_if_needed();
    TRACE_SETSTATE(("__init_statics_if_needed 6\n"));

    brkstatics.exit_allowed = 1;
    my_bss_count = __bss_count;
  }
  TRACE_SETSTATE(("__init_statics_if_needed 7\n"));
  return;
}
#endif

/* ===================================================================== */
/*                        initialization                                 */
/* ===================================================================== */

#if defined(WANT_SIGBLK_SUPPORT)
static void __ctor_cbrk(void)
{
  /* setup default handlers in case the user doesn't */
  __init_statics_if_needed();
  return;
}

#define __SETUP_CONSTRUCTOR_FOR__ __ctor_cbrk
#include "ctordtor.h"
#endif

/* ===================================================================== */
/*                               PUBLICS                                 */
/* ===================================================================== */

/*
 * new_state=0: block ^C/^Break 
 *              (^C/^Break are counted, but ignored)
 * new_state=1: unblock ^C/^Break 
 *              (the default if our code weren't there is to [_]exit())
 * returns previous mode as 1/0 (and only 1/0 - never anything else -
 * this is important to allow the calling code to assert unitialized brkstate).
*/
int set_cbreakstate (int new_state)
{
  int rc = -1;
#if !defined(WANT_SIGBLK_SUPPORT)
  new_state = new_state;
#else
  if (!blocker_get_indos_flag())
  {
    TRACE_SETSTATE(("set_cbreakstate(%d)\n", new_state));
    if (new_state == 0 || new_state == 1)
    {
      TRACE_SETSTATE(("set_cbreakstate 10\n"));
      __init_statics_if_needed();
      TRACE_SETSTATE(("set_cbreakstate 11\n"));
      __set_sig_handler_if_needed();
      TRACE_SETSTATE(("set_cbreakstate 12\n"));
      rc = __set_brk_mode(new_state);
      TRACE_SETSTATE(("set_cbreakstate 13\n"));
    }
    TRACE_SETSTATE(("set_cbreakstate(%d) => %d\n", new_state, rc));
  }
#endif
  return rc;
}

/* --------------------------------------------------------------------- */

/* get_cbreakstate() gets the current state. (true if ^C is not blocked).
*/
int get_cbreakstate (void)
{
#if !defined(WANT_SIGBLK_SUPPORT)
  return -1;
#else
  if (brkstatics.origbrkmode == 0) /* not initialized yet */
    return -1;                     /* (non-zero/^C is not being blocked) */
  return brkstatics.exit_allowed;  /* 1/0 */
#endif
}

/* --------------------------------------------------------------------- */

int get_cbreakpend (void)
{
#if !defined(WANT_SIGBLK_SUPPORT)
  return 0;
#else
  if (brkstatics.origbrkmode != 0 && /* is initialized */
      brkstatics.exit_allowed == 0)  /* being blocked */
  {
    unsigned int i;
#if defined(__DJGPP__)
    /* the following is only valid when brkstatics.exit_allowed == 0 */
    int cnt = __djgpp_cbrk_count;
    if (cnt)
    {
      i = __find_sig_state(SIGBREAK);
      brkstatics.sigstates[i].pending = 1;
      brkstatics.brkseencount += cnt;
      __djgpp_cbrk_count = 0;
    }
#endif
    i = 0; 
    while (i < sizeof(brkstatics.sigstates)/sizeof(brkstatics.sigstates[0]))
    {
      if (brkstatics.sigstates[i++].pending)
      {
        TRACE_SIGHANDLER(("get_cbreakpend() returning %d\n", 
                           brkstatics.brkseencount));
        return brkstatics.brkseencount;
      }
    }
  }
  return 0;     
#endif
}

/* ======================================================================= */

