/*
 * Monotonic time source using TSC or Int 1Ch interrupt.
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: timer.c,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
/*
 * Monotonic time source using TSC or Int 1Ch interrupt.
 *
 * Created May 2001 by Cyrus Patel <cyp@fb14.uni-mainz.de>
 *
 * --------------------------------------------------------------------
 *
 * timer_xxx_count() return the number of XXXs since the first time 
 * any timer_xxx() function was called.
 * 
 * timer_msec_count() - number of millisecs (wraps at ~49 days)
 * timer_hsec_count() - number of hundreths of seconds (wraps at ~492 days)
 * timer_tick_count() - number of (54ms) timer ticks (wraps at ~4 years)
 * timer_sec_count()  - number of seconds (wraps at ~136 years)
 *
 * timer_syst_enter() - begin tracking "system time"
 * timer_syst_exit()  - end tracking "system time"
 * timer_syst_read()  - get elapsed "system time". 
 * timer_usrt_read()  - get elapsed "user time".
 *                      (the difference between system and monotonic time)
 *
*/
#undef WANT_STACK_CHECKING /* disable stack checking for this module */

#include "target.h"   /* dosish stuff */
#include "dpmi.h"     /* _my_dpmi_*() */
#include "ctordtor.h" /* _my_atterm() */
#include "printk.h"   /* _printk() [interrupt safe debugging] */
#include "timer.h"    /* ourselves */

#undef  TRACE_TSC
#undef  TRACE_ISR

/* ===================================================================== */
/*                        static (locked) data                           */
/* ===================================================================== */

static struct
{
  volatile int  isr_hook_state; /*0=not initialized, 1=doneinit, -1=shutdown*/

  volatile struct { unsigned int wraps; unsigned long ctr; } tick_count;
  volatile struct { unsigned long secs; unsigned msecs, carry; } millisec_count;
  volatile unsigned int timer_res_msecs; /* timer resolution in msecs */

  volatile           int  syst_enter_count;   /* enter increments, exit dec */
  volatile unsigned long  syst_mark;          /* time at top level enter */
  volatile unsigned long  syst_elapsed_secs;  /* \_ "system time" in full */
  volatile unsigned short syst_elapsed_msecs; /* /  and fractional secs   */

#if defined(__DPMI32__)
  int int1Chpmhook; /* handle if using pmiv, else 0 (using rmcb) */
#endif
} FAR my_statics;

/* ===================================================================== */
/*                        initialization                                 */
/* ===================================================================== */

static void __ctor_timer(void)
{
  static int my_bss_count = -1;
  if (my_bss_count < __bss_count) /* always LT compare! */
  {
    FAR_MEMSET( &my_statics, 0, sizeof(my_statics));
    my_bss_count = __bss_count;

    my_statics.tick_count.ctr = 4; /* start at something > 0 */
    /* the relationship between the millisec_count fields for 
    ** tickcount == 4 is: (54925 * 4) => 219700 usecs
    */
    my_statics.millisec_count.secs  = 0;   /* usecs / 1000000) */
    my_statics.millisec_count.msecs = 219; /* (usecs / 1000) % 1000 */
    my_statics.millisec_count.carry = 700; /* usecs % 1000 */

    my_statics.timer_res_msecs = 54; /* timer resolution in msecs */
  }
  return;
}

#define __init_statics_if_needed  __ctor_timer
#define __SETUP_CONSTRUCTOR_FOR__ __ctor_timer
#include "ctordtor.h"

/* ===================================================================== */
/*                          read BIOS counter                            */
/* ===================================================================== */

static unsigned long __read_bios_ticks(void)
{
  unsigned long ticker1;          /* one of the two 'tickerX' has to be ... */
  volatile unsigned long ticker2; /* ... volatile so that the compiler ... */
  do                              /* ... doesn't optimize away the loop. */
  {
    #if !defined(__DPMI32__)
    ticker1 = *((unsigned long _far *)MK_FP(0x0040, 0x006C));
    ticker2 = *((unsigned long _far *)MK_FP(0x0040, 0x006C));
    #elif defined(__DJGPP__)
    ticker1 = _farpeekl(_dos_ds, 0x46c);
    ticker2 = _farpeekl(_dos_ds, 0x46c);
    #else
    if (_my_ds() == _dos_ds)
    {
      ticker1 = *((unsigned long *)(0x46Cu));
      ticker2 = *((volatile unsigned long *)(0x46Cu));
    }
    else
    {
      dosmemget( 0x46c,  sizeof(ticker1), &ticker1 );
      dosmemget( 0x46c,  sizeof(ticker1 /* non-volatile one */), &ticker2 );
    }
    #endif
  } while (ticker1 != ticker2);
  return ticker1;
}

/* ===================================================================== */
/*                          Random number generator                      */
/* ===================================================================== */

/*
 * returns a random 32bit value in range [a..b]
 */
unsigned long Random (unsigned long a, unsigned long b)
{
  unsigned long r = a;
  if (a != b)
  {
    static unsigned long FAR seed = 0UL;

    b = ((b>a) ? (b-a) : (a-b)) + 1;
    if (b == 0) /* a was 0 and b was 0xffffffff */
      b--;

    /* separately so that the stdlib rand()/srand() aren't affected */
    if (seed == 0UL)
    {
      seed = (unsigned long)__read_bios_ticks();
      if (seed == 0ul)
        seed++;
    }
    seed = MUL32(seed, 1103515245ul) + 12345;
    r = (unsigned long)(seed >> 16);
    if (b >= (1ul << (sizeof(unsigned short)*8)))
    {
      seed = MUL32(seed, 1103515245ul) + 12345;
      r |= (unsigned long)(seed & 0xffff0000ul);
    }
    if (seed == 0ul)
      seed++;
    /* 'r' is a full 16/32 bit precision random number */

    if (b < (1ul << (sizeof(unsigned short)*8)))
      r = a + (unsigned long)(((unsigned short)r) % ((unsigned short)b));
    else if (b != 0xfffffffful)
      r = a + MOD32(r, b);
  }
  return r;
} 

/* ==================================================================== */
/*                       TSC-based monotonic time source                */
/* ==================================================================== */

#if defined(__WATCOMC__) 
#define HAVE_TSC_SUPPORT /* both real and pmode */
#elif (defined(__TURBOC__) && defined(__FLAT__))
#define HAVE_TSC_SUPPORT
#elif 0 /* (defined(__GNUC__) && defined(__i386__)) */
#define HAVE_TSC_SUPPORT
#elif defined(HAVE_TSC_SUPPORT)
#undef HAVE_TSC_SUPPORT
#endif

#if defined(TRACE_TSC)
  #undef TRACE_TSC
  #define TRACE_TSC(x) _printk x
#else
  #define TRACE_TSC(x) NO_OPERATION()
#endif

#if defined(HAVE_TSC_SUPPORT)
static unsigned long __get_cpu_and_flags(void)
{
   /* [0..3]=step,[4..7]=model,[8..11]=family,[12..15]=0 */
   /* [16..32]=low 16 bits of cpuid feature flags */

#if defined(__WATCOMC__) || defined(__TURBOC__)
   unsigned short cpu_id = 0x0300;
   unsigned short cpu_flags = 0;
   int old_vmmgr = 0;

   #if (UINT_MAX == USHRT_MAX)
   {
     __asm mov     al,1         /* 808x will shift using full cl */
     __asm mov     cl,32        /* while 8018x+ use only low 4 bits */
     __asm shr     al,cl        /* 808x clears ax, 8018x+ will leave untouched*/
     __asm cmp     al,1         /* set carry if 808x, clear carry if 8018x+ */
     __asm cmc                  /* toggle carry (now set if 8018x+) */
     __asm sbb     ax,ax        /* make 0xffff if 8018x+, else zero */
     __asm and     ax,100h      /* make 0x0100 if 8018x+, else zero */

     __asm push    sp           /* 286+ push value first then, then dec SP */
     __asm pop     cx           /* pre-286 dec SP, then push value */
     __asm cmp     cx,sp        /* if cx < sp, its 808x/8018x/V20/V30 */
     __asm cmc                  /* set carry if its a 286+ */
     __asm sbb     cx,cx        /* cx=0xffff if 286+, else 0 */
     __asm and     cl,1         /* cx=1 if 286+, else 0 */
     __asm shl     ax,cl        /* ax=200h if 286+,100h if 8018x,0 if 808x */

     __asm and     ch,0f0h      /* control mask (was 0xff if 286+) */
     __asm pushf                /* save flags from damage */
     __asm pushf                /* copy flags ... */
     __asm pop     dx           /* ... to dx */
     __asm or      dh,ch        /* set top 4 bits */
     __asm push    dx           /* copy modified flags ... */
     __asm popf                 /* ... to flags */
     __asm pushf                /* copy flags ... */
     __asm pop     dx           /* ... to dx */
     __asm popf                 /* restore flags saved from damage */
     __asm and     dh,ch        /* its not a 386+ if zero */
     __asm cmp     dh,1         /* set carry if not a 386+ */
     __asm cmc                  /* toggle carry (set if a 386+) */
     __asm sbb     dh,dh        /* make 0xffff if its a 386+ */
     __asm and     dh,1         /* isolate add mask */
     __asm add     ah,dh        /* ax=300h=386+,200h=286+,100h=8018x,0=808x */

     __asm xor     dx,dx        /* no feature bits here */
     __asm mov     cpu_id,ax
     __asm mov     cpu_flags,dx
     TRACE_TSC(("cpu detect stage 1: %04x:%04x\n",cpu_flags,cpu_id));
   }
   #endif
   if ((cpu_id >> 8) >= 3)
   {      
     /* an interrupt could change the AC bit, so loop 'n'
      * times, incrementing a state counter for each state
     */
     __asm 
     {
       #if (UINT_MAX == USHRT_MAX)
       .386
       #endif
       pushfd               /* save flags from damage */
       xor     edx,edx      /* clear counters */
       #if defined(__TURBOC__)
       } _386next: __asm {
       #else
       _386next:
       #endif
       pushfd               /* copy flags ... */
       pop     ecx          /* ... to ecx */
       mov     eax, ecx     /* copy original EFLAGS */
       xor     eax, 40000h  /* toggle AC bit (1<<18) in eflags */
       push    eax          /* copy modified eflags */
       popfd                /* ... to EFLAGS */
       pushfd               /* copy (possible modified) EFLAGS */
       pop     eax          /* ... back to eax */
       and     eax, 40000h  /* isolate AC flag */
       and     ecx, 40000h  /* isolate AC flag */
       cmp     eax, ecx     /* will be 386 if no change */
       setz    al           /* set to one if 386, else zero */
       setnz   ah           /* set to zero if 386, else one */
       add     dx,ax        /* add to totals */
       mov     al,dl        /* copy is386 count */
       add     al,dh        /* add not486 count */
       cmp     al,31        /* done 31 loops? */
       jb      _386next     /* continue if not */
       popfd                /* restore saved flags */
       cmp     dl,dh        /* is386 < not386? (set carry if so) */
       mov     ax,300h      /* assume its a 386 */
       adc     ah,al        /* make it a 486 if carry is set */
       xor     dx,dx        /* no feature bits here */
       mov     cpu_id,ax
       mov     cpu_flags,dx
       #if (UINT_MAX == USHRT_MAX)
       .8086
       #endif
     }
   }
   if ((cpu_id >> 8) >= 4)
   {      
     #if (UINT_MAX == USHRT_MAX)
     __asm .8086
     #endif
     __asm pushfd                  /* save flags from damage */
     __asm mov     edx, 200000h    /* ID flag pos (low 16bits/featflags clear) */
     __asm pushfd                  /* copy EFLAGS */
     __asm pop     ecx             /* ... into ecx */
     __asm mov     eax, ecx        /* copy original flags */
     __asm xor     eax, edx        /* try to toggle ID bit */
     __asm push    eax             /* copy modified eflags */
     __asm popfd                   /* ... to EFLAGS */
     __asm pushfd                  /* push back flags */
     __asm pop     eax             /* ... into e_ax */
     __asm and     ecx, edx        /* clear all but the ID bit */
     __asm and     eax, edx        /* clear all but the ID bit */
     __asm popfd                   /* restore saved flags */
     __asm cmp     eax, ecx        /* was ID flag changed? */
     __asm setnz   ah              /* set to one if changed, else zero */
     __asm add     ah, 4           /* assume no change */
     __asm mov     cpu_id,ax
     __asm mov     cpu_flags,dx
     #if (UINT_MAX == USHRT_MAX)
     __asm .8086
     #endif
     TRACE_TSC(("cpu detect stage 4: %04x:%04x\n",cpu_flags,cpu_id));
   }
   #if (UINT_MAX == USHRT_MAX)
   if ((cpu_id >> 8) >= 5)
   {
     /* the 586 code will hang or gpf on the rdtsc insn under a virtualizer 
      * that is not pentium-aware eg emm386 or win16 DOS box (NTVDM is ok 
      * though). Attempting to load cr4 to test the TSC bit will hang/gpf too.
      * So, we have to detect if running in virtual-8086 and not do 
      * cpuid/rdtsc if such a condition is detected.
     */
     unsigned short x_msw = 0;
     __asm .286p
     __asm smsw ax
     __asm .8086
     __asm mov x_msw, ax
     /* _printk("x_msw = %04x\n", x_msw ); */
     cpu_flags |= ((x_msw >> 1) & 1);
     if (x_msw & 1) /* is virtual-8086 */
     {
       const char *p = getenv("OS");
       old_vmmgr = (!(p && strcmp(p, "Windows_NT")==0));
     }
   }
   #endif
   if (!old_vmmgr && (cpu_id >> 8) >= 5)
   {      
     #if (UINT_MAX == USHRT_MAX)
     __asm .386
     #endif
     __asm push    ebx             /* save ebx (cpuid modifies it) */
     __asm mov     eax,1           /* cpuid level 1 */
     __asm db      0fh,0a2h        /* "cpuid" (no 0x66 opcode in real mode) */
     __asm pop     ebx
     /* family/model/stepping is in (e)ax; feature flags are in (e)dx */
     __asm and     eax,0fffh       /* keep only family/model/stepping */
     __asm mov     cpu_id,ax
     __asm mov     cpu_flags,dx
     #if (UINT_MAX == USHRT_MAX)
     __asm .8086
     #endif
     TRACE_TSC(("cpu detect stage 5: %04x:%04x\n",cpu_flags,cpu_id));
   }
   return (((unsigned long)cpu_flags)<<16)+(((unsigned long)cpu_id));
#elif defined(__GNUC__)
   unsigned long result;
   __asm__ __volatile__(
     "pushl    %%edx           \n\t"
     "pushl    %%ecx           \n\t"
     "xorl     %%edx, %%edx    \n\t"
     "0:                       \n\t"
     "pushfl                   \n\t"
     "pushfl                   \n\t"
     "popl     %%ecx           \n\t"
     "movl     %%ecx,%%eax     \n\t"
     "xorl     $0x40000,%%eax  \n\t"
     "pushl    %%eax           \n\t"
     "popfl                    \n\t"
     "pushfl                   \n\t"
     "popl     %%eax           \n\t"
     "popfl                    \n\t"
     "andl     $0x40000,%%eax  \n\t"
     "andl     $0x40000,%%ecx  \n\t"
     "cmpl     %%ecx,%%eax     \n\t"
     "setz     %%al            \n\t"
     "setnz    %%ah            \n\t"
     "addw     %%ax,%%dx       \n\t"
     "movb     %%dl,%%al       \n\t"
     "addb     %%dh,%%al       \n\t"
     "cmpb     $31,%%al        \n\t"
     "jb       0b              \n\t"
     "cmpb     %%dh,%%dl       \n\t"
     "movl     $0x0300,%%eax   \n\t"
     "adc      %%al,%%ah       \n\t"
     "cmp      $4,%%ah         \n\t"
     "jb       1f              \n\t"
     "                         \n\t"
     "pushfl                   \n\t"
     "movl     $0x200000,%%edx \n\t"
     "pushfl                   \n\t"
     "popl     %%ecx           \n\t"
     "movl     %%ecx,%%eax     \n\t"
     "xorl     %%edx,%%eax     \n\t"
     "pushl    %%eax           \n\t"
     "popfl                    \n\t"
     "pushfl                   \n\t"
     "popl     %%eax           \n\t"
     "andl     %%edx,%%ecx     \n\t"
     "andl     %%edx,%%eax     \n\t"
     "popfl                    \n\t"
     "cmp      %%ecx,%%eax     \n\t"
     "movl     $0,%%eax        \n\t"
     "setnz    %%ah            \n\t"
     "addb     $4,%%ah         \n\t"
     "cmp      $5,%%ah         \n\t"
     "jb       1f              \n\t"
     "                         \n\t"
     "pushl    %%ebx           \n\t"
     "movl     $1,%%eax        \n\t"
     ".byte    0x0f, 0xa2      \n\t"
     "popl     %%ebx           \n\t"
     "andl     $0x0fff,%%eax   \n\t"
     "shl      $16,%%edx       \n\t"
     "orl      %%edx,%%eax     \n\t"
     "1:                       \n\t"
     "popl     %%ecx           \n\t"
     "popl     %%edx           \n\t"
     : "=a"(result) );
   TRACE_TSC(("cpu detect stage: %08x\n", result ));
   return result;
#else /* !(defined(__WATCOMC__) || defined(__TURBOC__)) */
   return 0ul;
#endif
}
#define __have_tsc() ((__get_cpu_and_flags() & (1ul<<(4+16)))!=0)
#endif /* defined(HAVE_TSC_SUPPORT) */

/* ------------------------------------------------------------------- */

#if !defined(HAVE_TSC_SUPPORT)
/* nothing */
#elif defined(__WATCOMC__) && defined(__386__)
extern void rdtsc(unsigned long *tsc); /* two long array */
#pragma aux rdtsc = "db 0Fh,31h" "mov [ebx],eax" "mov [ebx+4],edx" \
                    parm [ebx] modify exact [eax edx];
#elif defined(__WATCOMC__) && !defined(__386__)
extern void rdtsc(unsigned long _far *tsc); /* two long array */
#pragma aux rdtsc = "push ds" "mov ds,dx" "mov bx,ax" "db 0Fh,31h" \
                    ".386" "mov [bx],eax" "mov [bx+4],edx" ".8086" \
                    "pop ds" parm [ax dx] modify exact [ax bx dx];
#elif defined(__TURBOC__) && defined(__FLAT__)
#define rdtsc(__tscptr) {                             \
           __emit__(0x0f,0x31);         /* rdtsc */   \
           *(((unsigned long *)(__tscptr))+0) = _EAX; \
           *(((unsigned long *)(__tscptr))+1) = _EDX; \
           }
#elif defined(__TURBOC__) && !defined(__FLAT__)
#define rdtsc(__tscptr) {                              \
            __emit__(0x0f,0x31);        /* rdtsc */    \
            *(((unsigned short *)(__tscptr))+0) = _AX; \
            *(((unsigned short *)(__tscptr))+2) = _DX; \
            __emit__(0x66,0xC1,0xE8,0x16); /* shr eax,16 */ \
            __emit__(0x66,0xC1,0xEA,0x16); /* shr edx,16 */ \
            *(((unsigned short *)(__tscptr))+1) = _AX; \
            *(((unsigned short *)(__tscptr))+3) = _DX; \
            }
#elif defined(__GNUC__)
static __inline__ void rdtsc(unsigned long *tsc)
{
  unsigned int lo, hi;
  __asm__ __volatile__ (".byte 0x0f, 0x31" : "=a" (lo), "=d" (hi) : );
  tsc[0] = lo; tsc[1] = hi;
}
#endif

/* ------------------------------------------------------------------- */

#if !defined(HAVE_TSC_SUPPORT)
/* nothing */
#elif defined(__WATCOMC__)
extern unsigned char __get_cmos_secs(void);
#pragma aux __get_cmos_secs = "retry:" "mov al,0ah" "out 70h,al" \
 "db 0EBh,0" "in al,71h" "test al,80h" "jnz retry"  "mov al,0" \
 "out 70h,al" "db 0EBh,0" "in al,71h" value [al] modify exact [ax];
#elif defined(__TURBOC__)
static unsigned char __get_cmos_secs(void)
{
  retry:
  __asm mov  al,0Ah
  __asm out  70h,al
  __asm dw   00ebh   /* jmp $+2 */
  __asm in   al,71h
  __asm test al,80h
  __asm jnz  retry
  __asm mov  al,0
  __asm out  70h,al
  __asm dw   00ebh   /* jmp $+2 */
  __asm in   al,71h
  return _AL;
}
#elif defined(__GNUC__)
static __inline__ unsigned char __get_cmos_secs(void)
{
#if defined(__DJGPP__)
  unsigned int secs;
  __asm__ __volatile__ ( "0: movb  $0x0a,%%al  \n\t"
                         "   outb  %%al,$0x70  \n\t"
                         "   .byte 0xeb, 0x00  \n\t"
                         "   inb   $0x71,%%al  \n\t"
                         "   testb $0x80,%%al  \n\t"
                         "   jnz   0b          \n\t"
                         "   xorl  %%eax,%%eax \n\t"
                         "   outb  %%al,$0x70  \n\t"
                         "   .byte 0xeb, 0x00  \n\t"
                         "   inb   $0x71,%%al  \n\t"
                         : "=a"(secs) : /* no inputs */ );
  TRACE_TSC(("__get_cmos_secs: %u\n", secs ));
  return (unsigned char)secs;
#else
  struct timeval tv;
  gettimeofday(&tv,0);
  return (unsigned char)(tv.tv_sec % 60u);
#endif
}
#endif

/* ------------------------------------------------------------------- */

#if defined(HAVE_TSC_SUPPORT)
static unsigned long __tsc_get_rate(void) 
{
  static unsigned long tsc_rate = 1; /* unknown (impossible) */
  if (tsc_rate == 1)
  {
    if (!__have_tsc())
      tsc_rate = 0;
    else
    {
      unsigned long tsc1[2], tsc2[2];
      unsigned long rhi, rlo;
      unsigned int t2, t1;

      t2 = __get_cmos_secs();
      do { rdtsc(&(tsc1[0])); } while ((t1 = __get_cmos_secs()) == t2);
      do { rdtsc(&(tsc2[0])); } while ((t2 = __get_cmos_secs()) == t1);

      rlo = tsc2[0] - tsc1[0];
      rhi = tsc2[1] - tsc1[1];
      if (rlo > tsc2[0])
        rhi--;
      
      TRACE_TSC(("tsc get rate: 1: %u (%lu:%lu), " \
                 "2: %u (%lu:%lu), rate=%lu:%lu\n",
                  t1,tsc1[1],tsc1[0],t2,tsc2[1],tsc2[0], rhi,rlo));

      if (rhi == 0 && rlo >= 1000000ul) 
        tsc_rate = rlo;
      else /* rate exceeds 4.2Ghz or rate is less than 1Mhz */
        tsc_rate = 0;
    }
    TRACE_TSC(("tsc rate = %lu\n", tsc_rate ));
  }
  return tsc_rate;
}
#endif /* defined(HAVE_CPUIDENT_CODE) */

/* ------------------------------------------------------------------- */

#if defined(HAVE_TSC_SUPPORT)
static int __tsc_get_smst( unsigned long *secsP,
                           unsigned int *msecsP,
                           unsigned long *ticksP )
{
  unsigned long rate = __tsc_get_rate();
  if (rate)
  {
    unsigned long secs, msecs;
    unsigned long tsc[2];
    rdtsc( &tsc[0] );

    #if defined(__WATCOMC__) || defined(__TURBOC__)
    {
      secs = tsc[1]; /* tsc hi */
      msecs = tsc[0]; /* tsc lo */
      #if (UINT_MAX == USHRT_MAX)
      __asm .386
      #endif      
      /* secs = (tsc_hi:tsc_lo)/rate */
      __asm mov  eax,secs       /* load tsc_hi */ 
      __asm xor  edx,edx
      __asm mov  ecx,rate     
      __asm div  ecx            /* => eax=secshi (discarded) */
      __asm mov  eax,msecs      /* load tsc_lo */
      __asm div  ecx
      __asm mov  secs,eax       /* save quotient (secs) */

      /* msecs = (remainder * 1000)/rate */
      __asm mov  eax,edx
      __asm xor  edx,edx
      __asm push ecx
      __asm mov  ecx,1000
      __asm mul  ecx
      __asm pop  ecx

      __asm push eax
      __asm mov  eax,edx
      __asm xor  edx,edx
      __asm div  ecx
      __asm pop  eax
      __asm div  ecx
      __asm mov  msecs,eax    /* save quotient (msecs) */
      #if (UINT_MAX == USHRT_MAX)
      __asm .8086
      #endif
    }
    #elif defined(__GNUC__)
    {
      secs = tsc[1]; /* tsc hi */
      msecs = tsc[0]; /* tsc lo */

      __asm__ __volatile__("    \n\t"
       "pushl    %%eax          \n\t"
       "movl     %%edx,%%eax    \n\t"
       "xorl     %%edx,%%edx    \n\t"
       "divl     %%ecx          \n\t"
       "popl     %%eax          \n\t"
       "divl     %%ecx          \n\t"
       "pushl    %%eax          \n\t"
       "                        \n\t"
       "movl     %%edx,%%eax    \n\t"
       "xorl     %%edx,%%edx    \n\t"
       "pushl    %%ecx          \n\t"
       "movl     $1000,%%ecx    \n\t"
       "mull     %%ecx          \n\t"
       "popl     %%ecx          \n\t"
       "                        \n\t"
       "pushl    %%eax          \n\t"
       "movl     %%edx,%%eax    \n\t"
       "xorl     %%edx,%%edx    \n\t"
       "divl     %%ecx          \n\t"
       "popl     %%eax          \n\t"
       "divl     %%ecx          \n\t"
       "                        \n\t"
       "popl     %%edx          \n\t"
       : "=a"(msecs), "=d"(secs) : "a"(msecs), "d"(secs), "c"(rate) );
    }
    #endif
    {
      static unsigned long base_secs = 0;
      static unsigned long base_msecs = 0;
      if ((base_secs | base_msecs) == 0ul)
      {
        my_statics.timer_res_msecs = 1;
        do {
          base_msecs = my_statics.millisec_count.msecs; 
          base_secs = my_statics.millisec_count.secs;
        } while (base_msecs != my_statics.millisec_count.msecs);
        base_secs = secs - base_secs;
        if (msecs >= base_msecs)
          base_msecs = msecs - base_msecs;  
        else
        {
          base_secs--;
          base_msecs = (msecs + 1000u) - base_msecs;
        }
      }
      if (msecs < base_msecs)
      {
        msecs += 1000;
        secs--;
      }
      secs -= base_secs;
      msecs -= base_msecs;
    }

    TRACE_TSC(("tsc %lu:%lu/%lu\n       => time = %lu.%03lu\n",
                  tsc[1], tsc[0], rate, secs, msecs ));

    if (msecsP)
      *msecsP = (unsigned)msecs;
    if (secsP) 
      *secsP = secs;
    if (ticksP)
    {
      /* either ((((secs * 1000)+msecs) * 119318) / 6553600)
      **        => 18.20648193359375
      **    or  (((secs * 1000000)+(msecs * 1000)) / 54925);
      **        => 18.206645425580336822940373236231
      */
      unsigned long ticks;

      #if defined(__WATCOMC__) || defined(__TURBOC__)
      #if (UINT_MAX == USHRT_MAX)
      __asm .386
      #endif
      __asm mov  eax,msecs
      __asm mov  edx,secs

      __asm push edx
      __asm mov  ecx,0x1D216   /* 119318 */
      __asm xor  edx,edx
      __asm mul  ecx
      __asm pop  edx

      __asm push eax
      __asm mov  eax,edx
      __asm mov  ecx,0x71CA5F0  /* 119318000 */
      __asm xor  edx,edx
      __asm mul  ecx
      __asm pop  ecx
      __asm add  eax,ecx
      __asm adc  edx,0

      __asm push eax
      __asm mov  eax,edx
      __asm xor  edx,edx
      __asm mov  ecx,0x640000  /* 6553600 */
      __asm div  ecx
      __asm pop  eax
      __asm div  ecx
      __asm mov  ticks, eax
      #if (UINT_MAX == USHRT_MAX)
      __asm .8086
      #endif

      #elif defined(__GNUC__)
      __asm__ __volatile__ (
        "pushl %%edx           \n\t"
        "movl  $0x1d216,%%ecx  \n\t"
        "xorl  %%edx,%%edx     \n\t"
        "mull  %%ecx           \n\t"
        "popl  %%edx           \n\t"
        "                      \n\t"
        "pushl %%eax           \n\t"
        "movl  %%edx,%%eax     \n\t"
        "movl  $0x71CA5F0,%%ecx\n\t"
        "xorl  %%edx,%%edx     \n\t"
        "mull  %%ecx           \n\t"
        "popl  %%ecx           \n\t"
        "addl  %%ecx,%%eax     \n\t"
        "adcl  $0,%%edx        \n\t"
        "                      \n\t"
        "pushl %%eax           \n\t"
        "movl  %%edx,%%eax     \n\t"
        "xorl  %%edx,%%edx     \n\t"
        "movl  $0x640000,%%ecx \n\t"
        "divl  %%ecx           \n\t"
        "popl  %%eax           \n\t"
        "divl  %%ecx           \n\t"
        : "=a"(ticks) : "a"(msecs), "d"(secs) : "ecx" );
      #endif

      TRACE_TSC(("ms=%lu%03lu => ticks=%lu\n", secs, msecs, ticks ));

      *ticksP = ticks;
    }

    return 0;
  }
  return -1;
}  
#endif /* defined(HAVE_TSC_SUPPORT) */

/* ===================================================================== */
/*                   Interrupt-driven timer (common part)                */
/* ===================================================================== */

#if defined(TRACE_ISR)
  #undef TRACE_ISR
  #define TRACE_ISR(x) _printk x
#else
  #define TRACE_ISR(x) NO_OPERATION()
#endif

#define TIMER    0x1C /* 0x1C *NOT* 0x08!!! */
                      /* also do not use the PIT */

/* --------------------------------------------------------------------- */

static void __increment_timer_counters(void) 
{                                            
  /* called from ISR with interrupts disabled. */
  /* DO NOT INLINE. DO NOT MOVE - IS IN A LOCKED PAGE */

  unsigned carry, msecs, incsecs;                         
                                                          
  if ((++my_statics.tick_count.ctr) == 0)                 
    ++my_statics.tick_count.wraps;                        
                                                          
  msecs   = my_statics.millisec_count.msecs + 54;         
  carry   = my_statics.millisec_count.carry + 925;        
  incsecs = 0;                                            
  if (carry >= 1000)                                      
  {                                                       
    msecs++;                                              
    carry -= 1000;                                        
  }                                                       
  if (msecs >= 1000)                                      
  {                                                       
    incsecs++;                                            
    msecs -= 1000;                                        
  }                                                       
  my_statics.millisec_count.carry = carry;                
  my_statics.millisec_count.secs += incsecs;              
  my_statics.millisec_count.msecs = msecs;                
                                                          
  return;                                                 
}                                                         

/* ==================================================================== */
/*              timer interrupt handler for PMODE targets               */
/* ==================================================================== */

#if defined(__DPMI32__) /* all DPMI targets including pharlap */

/*             PMODE/W/DJ      DOS4G(W)      CWSDPMI/Causeway/DOSX32
** 0x1C-pmisr   crash(1)    not called(3)             ok
** 0x1C-rmcb    crash(2)        n/a                   ok
** 0x08-pmisr     ok        not called(3)             ok
** 0x08-rmcb    crash(2)        n/a                   ok
** 
** Causeway, DOSX32, CSDPMI were ok for all four.
** 
** (1) [PMODE/W]: Abort 03
**     No RM stack for INT redirection!
** (2) PMODE/W over NTVDM bombs if a DOS call is in progress when the 
**     cb is called. I'm not convinced this is a bug in NT since
**     cwsdpmi and Causway appear to work fine.
** (3) When the application is larger than a certain size
**     the pm isrs are hooked ok, but never get called.
**  
** (a) 0x1C is preferred over 0x08 since only the 0x1C interrupt
**     is guaranteed to be called @ 55ms.
**     Also, 0x1C is in the auto-callup range while 0x08 is not.
**     (0x1C will be signalled up to pmode even when the cpu is 
**     in real mode).
** (b) RMCB is preferable over PMISR because it is called
**     irrespective of which mode the processor is in.
**
** Solution: Under NTVDM use 0x1C-pmisr, otherwise use 0x1C-rmcb.
** DOS4GW cannot be supported in either mode.
*/

/* ---------------------------------------------------------------- */

static void __int1Ch_isr(__dpmi_regs *regs)
{
  /* TRACE_ISR((".")); */
  __increment_timer_counters();
  if (my_statics.int1Chpmhook) /* regs is invalid if !pmhook */
    __my_dpmi_chain_pmiv(regs);
  /* rmcb chains on return */
}

/* ---------------------------------------------------------------- */

static int __lockunlock_interrupt_code(int do_lock)
{
  /* ensure that code locking happens for the correct range. 
  ** The Watcom and GCC compilers are at liberty to rearrange 
  ** the order of [static] functions. (For watcom when compiled 
  ** with -zm and for GCC when not compiled with -mno-move)
  */
  unsigned topf, botf, tmpf;

  topf = botf = ((unsigned)__increment_timer_counters);
  tmpf = ((unsigned)__int1Ch_isr);  
  if (tmpf < topf) topf = tmpf;
  if (tmpf > botf) botf = tmpf;
  botf += 4096;
  if (!do_lock)
    return __my_dpmi_unlock_code((void *)topf, (botf - topf) );
  return __my_dpmi_lock_code((void *)topf, (botf - topf) );
}

/* ---------------------------------------------------------------- */

static struct /* does not have to be locked separately */
{             /* regs will be locked by __my_dpmi_allocate...ichain */
  __dpmi_regs regs; 
  __dpmi_raddr oldvect;
  __dpmi_raddr cbvect;
  int isset;
} int1Ch_cbdata;

/* ---------------------------------------------------------------- */

static void __isr_shutdown(void)
{
  if (my_statics.isr_hook_state > 0)
  {
    my_statics.isr_hook_state = -1;
    TRACE_ISR(("begin __isr_shutdown\n"));
    if (my_statics.int1Chpmhook)
    {
      __my_dpmi_unhook_pmiv(my_statics.int1Chpmhook);
      my_statics.int1Chpmhook = 0;
    }
    else if (int1Ch_cbdata.isset)
    {
      __dpmi_set_real_mode_interrupt_vector(TIMER, &(int1Ch_cbdata.oldvect));
      __my_dpmi_free_real_mode_callback( &(int1Ch_cbdata.cbvect) );
      int1Ch_cbdata.isset = 0;
    }
    __lockunlock_interrupt_code(0);
    TRACE_ISR(("end __isr_shutdown\n"));
  }
  return;
}

/* ---------------------------------------------------------------- */

static int __check_if_ticking(void)
{
  unsigned int elapsed = 0;
  unsigned long start_tics = my_statics.tick_count.ctr;
  while (elapsed < 3 && start_tics == my_statics.tick_count.ctr)
  {
    unsigned long start = __read_bios_ticks();
    while (start_tics == my_statics.tick_count.ctr && 
           start == __read_bios_ticks())
    {
      __dpmi_yield();
    }
    elapsed++;
  }
  return (start_tics != my_statics.tick_count.ctr);
}

/* ---------------------------------------------------------------- */

static void __isr_setup(void)
{
  __init_statics_if_needed();
  if (my_statics.isr_hook_state == 0)
  {
    #if defined(HAVE_TSC_SUPPORT)
    if (__tsc_get_smst( 0, 0, 0 )==0)
      my_statics.isr_hook_state = +1;
    else
    #endif
    {
      int use_pmisr = 0;

      TRACE_ISR(("begin __isr_setup (refaddr=timer_tick_count=%p)\n",
                                      timer_tick_count ));

      /* setting at atterm hook is not strictly necessary for DPMI */
      _my_atterm (__isr_shutdown); /* do before isr hooking! */
      my_statics.isr_hook_state = +1;

      if (__lockunlock_interrupt_code(1) != 0)
      {
        __fatal_runtime_error("PANIC! Unable to lock interrupt code\n");
      }

      {
        /* PMODE/W over NTVDM will trash real-mode DS if any CB is called  */
        /* while an Int21 call is in progress */
        const char *env = getenv("OS");
        use_pmisr = (env && strcmp(env, "Windows_NT")==0);
      }
      if (!use_pmisr) 
      {
        typedef void (*dummy_cast)(void); /* shaddup watcom */
        use_pmisr = 1;   
        if (__my_dpmi_allocate_real_mode_callback_ichain((dummy_cast)__int1Ch_isr,
                 TIMER, &(int1Ch_cbdata.regs), &(int1Ch_cbdata.cbvect))==0)
        {
          use_pmisr = 0;   
          __dpmi_get_real_mode_interrupt_vector(TIMER, &(int1Ch_cbdata.oldvect) );
          __dpmi_set_real_mode_interrupt_vector(TIMER, &(int1Ch_cbdata.cbvect) );
          int1Ch_cbdata.isset = 1;
        }
      }
      if (use_pmisr)
      {
        my_statics.int1Chpmhook = __my_dpmi_hook_pmiv(TIMER, __int1Ch_isr );
        if (!my_statics.int1Chpmhook)
        {
          __fatal_runtime_error("PANIC: Unable to hook timer interrupt\n");
        }
      }
      TRACE_ISR(("Timer Hook state: 0x1C=%s\n",
                     (my_statics.int1Chpmhook? "pm": "rmcb") ));
      if (!__check_if_ticking())
      {
        __isr_shutdown();
        /* only DOS4GW appears to have this bug */
        __fatal_runtime_error("PANIC: Timer interrupt handler is not being called."
                              " Try using a different DOS extender.\n");
      }
    } /* no TSC support */
  }
  return;
}
#endif /* defined(__DPMI32__) */

/* ==================================================================== */
/*              timer interrupt handler for real-mode targets           */
/* ==================================================================== */

#if !defined(__DPMI32__)

#if defined(__TURBOC__) || defined(__BORLANDC__)
  #if defined(__cplusplus)
  #define __ANYARGS ...
  #else
  #define __ANYARGS
  #endif
  static void _interrupt (_far *__old_timer_isr)(__ANYARGS);
#elif defined(__WATCOMC__)
  #define __ANYARGS
  static void (_interrupt _far *__old_timer_isr)(__ANYARGS);
#else
  #define __ANYARGS ...
  static void (_interrupt _far *__old_timer_isr)(__ANYARGS);
#endif

/* ---------------------------------------------------------------- */

static void _interrupt _far __new_timer_isr(__ANYARGS)
{
  #if defined(__TURBOC__) || defined(__BORLANDC__)
  #pragma warn -pro
  #endif
  (*__old_timer_isr)();    /* chain now */
  #if defined(__TURBOC__) || defined(__BORLANDC__)
  #pragma warn .pro
  #endif

  __increment_timer_counters();
}

/* ---------------------------------------------------------------- */

static void __isr_shutdown(void)
{
  __init_statics_if_needed();
  if (my_statics.isr_hook_state == +1) /* +2 is have TSC */
  {
    my_statics.isr_hook_state = -1;
    _dos_setvect (TIMER, __old_timer_isr);
  }
  return;
}

/* ---------------------------------------------------------------- */

static void __isr_setup(void)
{
  __init_statics_if_needed();
  if (my_statics.isr_hook_state == 0)
  {
    #if defined(HAVE_TSC_SUPPORT)
    if (__tsc_get_smst(0, 0, 0)==0)
      my_statics.isr_hook_state = +1;
    else
    #endif
    {
      my_statics.isr_hook_state = +1;
      _my_atterm (__isr_shutdown);
      __old_timer_isr = _dos_getvect(TIMER);
      _dos_setvect (TIMER, __new_timer_isr);
    }
  }
  return;
}
#endif /* !defined(__DPMI32__) */

/* ==================================================================== */
/*                     get monotonic time functions                     */
/* ==================================================================== */

static unsigned long __get_secs_and_millisecs(unsigned int *msecsP /*0..999*/)
{
  unsigned long secs;
  __isr_setup(); /* setup interrupt handler if necessary */
  #if defined(HAVE_TSC_SUPPORT)
  if (__tsc_get_smst(&secs, msecsP, ((unsigned long *)0) ) != 0)
  #endif
  {
    volatile unsigned int msecs;
    do
    {
      msecs = my_statics.millisec_count.msecs;
      secs  = my_statics.millisec_count.secs;
    } while (msecs != my_statics.millisec_count.msecs);
    *msecsP = msecs;
    if (my_statics.isr_hook_state < 0) /* already deinitialized */
      __increment_timer_counters();
  }
  return secs;
}

/* ---------------------------------------------------------------- */

unsigned long timer_tick_count(void) /* max 4 years */
{
  unsigned long t;
  __isr_setup(); /* setup interrupt handler if necessary */
  #if defined(HAVE_TSC_SUPPORT)
  if (__tsc_get_smst(0, 0, &t) != 0)
  #endif
  {
    /* make sure an interrupt doesn't occur during a read */
    do {
      t = my_statics.tick_count.ctr;
    } while (t != my_statics.tick_count.ctr); 
    if (my_statics.isr_hook_state < 0) /* already deinitialized */
      __increment_timer_counters();
  }
  return t;
}

/* ---------------------------------------------------------------- */

unsigned long timer_sec_count(void) /* max ~49710.2 days */
{
  unsigned int dummy;
  return __get_secs_and_millisecs(&dummy);
}

/* ---------------------------------------------------------------- */

unsigned long timer_hsec_count(void) /* max ~497.1 days */
{
  unsigned int msecs;
  unsigned long secs = __get_secs_and_millisecs(&msecs);
  return MUL32(secs, 100ul) + (msecs/10u);
}

/* ---------------------------------------------------------------- */

unsigned long timer_msec_count(void) /* max ~49.7 days */
{
  unsigned int msecs;
  unsigned long secs = __get_secs_and_millisecs(&msecs);
  return MUL32(secs, 1000ul) + (msecs);
}

/* ---------------------------------------------------------------- */

unsigned int timer_msec_tres(void) /* timer resolution in msecs */
{
  return my_statics.timer_res_msecs;
}

/* ==================================================================== */
/*                         system time thing                            */
/* ==================================================================== */

void timer_syst_enter (void)
{
  if (my_statics.isr_hook_state <= 0) /* not init, or already deinitialized */
    ++my_statics.syst_enter_count;
  else if ((++my_statics.syst_enter_count) == 1)
    my_statics.syst_mark = timer_msec_count();
  return;
}

/* ---------------------------------------------------------------- */

void timer_syst_exit (void)
{
  if (my_statics.isr_hook_state <= 0) /* not init, or already deinitialized */
    --my_statics.syst_enter_count;
  else if ((--my_statics.syst_enter_count) == 0)  
  {
    unsigned long prev = my_statics.syst_mark;
    unsigned long now = timer_msec_count();
    if (now < prev)
      now += 1 + (0xfffffffful - prev);
    else
      now = (prev - now);
    prev = DIV32(now, 1000ul) + my_statics.syst_elapsed_secs;
    now  = MOD32(now, 1000ul) + my_statics.syst_elapsed_msecs;
    if (now > 1000)
    {
      prev++;
      now -= 1000;
    }
    my_statics.syst_elapsed_secs = prev;
    my_statics.syst_elapsed_msecs = (short)now;
  }
  return;
}

/* ---------------------------------------------------------------- */

unsigned long timer_syst_read (unsigned long *secsP, unsigned int *msecsP)
{
  unsigned long s = 0;
  unsigned short m = 0;
  if (my_statics.isr_hook_state > 0) /* is initialized */
  {
    s = my_statics.syst_elapsed_secs;
    m = my_statics.syst_elapsed_msecs;
  }
  if (secsP)  *secsP = s;
  if (msecsP) *msecsP = m;
  return MUL32(s, 1000ul) + m;
}

/* ---------------------------------------------------------------- */

unsigned long timer_usrt_read (unsigned long *secsP, unsigned int *msecsP)
{
  unsigned long  ss, us = 0; 
  unsigned int   sm, um = 0;
  if (my_statics.isr_hook_state > 0) /* is initialized */
  {
    us = __get_secs_and_millisecs(&um);
    timer_syst_read (&ss, &sm);
    if (um < sm)
    {
      us --;
      um += 1000;
    }
    um -= sm;
    us -= ss;
  }
  if (secsP)  *secsP = us;
  if (msecsP) *msecsP = um;
  return MUL32(us, 1000ul) + um;
}

/* ==================================================================== */
