/*
 * Service Provider for the
 * User Datagram Protocol, "UDP", (RFC 768)
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: udp.c,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
/* 
** UDP transport provider
** Rewritten from scratch May 2001 by Cyrus Patel <cyp@fb14.uni-mainz.de>
**
** Open issues:
** a) Review multicast logic. 
** b) Write code to support a hostcache.
**
** *all* functions in the protocol layers (which this module is) are static.
** entry points are exported with a dispatch table that is
** registered from a constructor.
**
** The only globals made visible by this module are:
** int __IPPROTO_UDP__ = IPPROTO_UDP;
** int __SOCK_DGRAM__  = 2; // same value as BSD's SOCK_DGRAM
** These may be used to force linkage.
*/
#define __OSILAYER__ 4
#define WANT_STACK_CHECKING

#include "target.h"   /* standard types/includes */
#include "ipbase.h"   /* _ip_*() */
#include "inchksum.h" /* checksum() */
#include "debug.h"    /* KMSG(), DBGPRINTF() */
#include "icmp.h"     /* _icmp_*() */
#include "timer.h"    /* wintr_*() */
#include "blocker.h"  /* blocker_*() */
#include "loadcfg.h"  /* _conf_*() */
#include "inaddr.h"   /* is_broadcast() etc */
#include "sigblk.h"   /* set_cbreak_state() */
#include "printk.h"   /* printk() */
#include "utports.h"  /* lport_{reserve|release}() */
#include "utsio.h"    /* dispatch table */
#include "userland.h" /* WSO_* constants */

#define EXPORTED static
#undef IPPROTO_TCP /* marks things that are TCP specific */

/* ====================================================================== */

typedef struct udp_pcb /* UDP protocol control block */
{
  uint32_t        pcb_marker_top;
  struct udp_pcb FAR *next;          /* link to next socket */
  struct sockio_info *sio_link;      /* userland token */

  uint32_t        locflags;          /* local option flags */
  int             so_errno;          /* socket error code */          
  uint16_t        state;             /* tcp_StateXXX (both UDP/TCP) */

  uint8_t         xmit_timeout_s;    /* per-socket copy of ... */     
  uint8_t         recv_timeout_s;    /* ... global 'sock_delay'. */   
  uint8_t         ttl;               /* Time To Live  */              
  uint8_t         tos;               /* Terms of service (priority) */

  struct 
  {                                                                             
    void (*yield_fn)(void);          /* yield while waiting */             
    uint32_t      usertimer_s;       /* ip_timer_init()/ip_timer_expired() */
  } bc_idiocy;
                                                                             
  uint32_t        myaddr;            /* my ip-address */                   
  uint32_t        hisaddr;           /* peer's internet address */         
  uint16_t         hisport;          /* peer's source port */              
  uint16_t         myport;           /* my source port */                  

  struct
  {
    unsigned int usedlen;            /* Rx length */
    unsigned int maxlen;             /* Rx buffer size */
    uint8_t FAR *buf;                /* Rx buffer ptr */
  } rcv_buf;

  uint32_t        pcb_marker_bottom;
} udp_pcb;

/* ----------------------------------------------------------------------- */

/*
 * The UDP header
 */

#pragma pack(1)
typedef struct udp_Header 
{
  uint16_t   srcPort;
  uint16_t   dstPort;
  uint16_t   length;
  uint16_t   checksum;
} udp_Header;
#pragma pack()

/* ==================================================================== */
/*                             static data                              */
/* ==================================================================== */

/* globals that may be referenced to force linkage */
int __IPPROTO_UDP__ = IPPROTO_UDP;
int __SOCK_DGRAM__  = 2; /* same value as BSD's SOCK_DGRAM */

static struct 
{
  int udp_rw_timeout;       /* "SOCKDELAY" opt. (0 = use builtin val) */

  int debug_lev; /* 0=none, 1=msgs, 2=1+hdrdump, 3=2+datadump */
  udp_pcb FAR *_udp_allsocs;  /* list of udp-sockets */

} FAR my_statics;

#define udp_rw_timeout     (my_statics.udp_rw_timeout)
#define debug_lev          (my_statics.debug_lev)
#define _udp_allsocs       (my_statics._udp_allsocs)

static void __init_private_data(void)
{
  static const struct config_table FAR my_cfg[] = {
     { "UDP.DEBUG",            ARG_ATOI,     (void FAR *)&debug_lev  },
     { "SOCKDELAY",  ARG_MULTI|ARG_ATOI,     (void FAR *)&udp_rw_timeout },
     { ((const char FAR *)0), 0,    ((void FAR *)0)  }
   };
  _conf_add_confextn( &my_cfg[0], NULL );
  FAR_MEMSET( &my_statics, 0, sizeof(my_statics) );
}

static udp_pcb FAR *__validate_allsocs(struct sockio_info *__s); /* fwd ref */
static const struct wso_usrreq_tab FAR *__get_dispatch_tab(void); /* fwd ref */

/* ======================================================================= */
/*                               udp write                                 */
/* ======================================================================= */

#if defined(NDEBUG)
#define __dbg_dump_header(__a, __b, __c) NO_OPERATION()
#else
static void __dbg_dump_header(int is_output, uint16_t pktid, 
                              const udp_Header FAR *hdr)
{
  if (is_output)
    DBGPRINTF(("*UDP OUT-xxxx: header dump\n"));
  else 
    DBGPRINTF(("*UDP-IN %04x: header dump\n", pktid));
  DBGPRINTF(("*  -> srcPort   = %u\n", NTOHS(hdr->srcPort) ));
  DBGPRINTF(("*  -> dstPort   = %u\n", NTOHS(hdr->dstPort) ));
  DBGPRINTF(("*  -> length    = %u\n", NTOHS(hdr->length) ));
  DBGPRINTF(("*  -> checksum  = 0x%04x\n", hdr->checksum ));
  return;
}
#endif

/* ======================================================================= */
/*                               udp write                                 */
/* ======================================================================= */

/* max segment size (we don't want ip to do fragmenting) 
** 
** minimum  =>  576 - (20+8) => 548 
** DIX      => 1500 - (20+8) => 1472 
** SLIP     => 1006 - (20+8) => 978
*/
#define UDPSEGSIZE_MAX(__mtu) \
        ((unsigned int)((__mtu) - (sizeof(in_Header)+sizeof(udp_Header))))
/* is it ok to use '__mtu' instead of 576 (IPMTU_MIN)? */

/* ----------------------------------------------------------------------- */

static int _udp_write_segment( udp_pcb FAR *s, const void FAR *data, 
                               unsigned int datalen,
                               uint32_t dest_ip, uint16_t dest_port,
                               uint32_t local_ip, uint16_t local_port)
                              
{
  int  dontchksum = 0;    
  udp_Header hdr;
  uint8_t  ttl = 0; /* yet unknown */
  uint8_t  tos = (uint8_t)_ip_get_default_tos();
  struct _ip_ifinfo ifinfo;
  int dummy_so_errno;
  int FAR *so_errnoP = &dummy_so_errno;

  if (debug_lev)
     DBGPRINTF(("+*UDP OUT: dest=%hI:%u, src=%hI:%u, datalen=%u\n",
                  dest_ip, dest_port, local_ip, local_port, datalen ));

  /* ------------------------------------------------ */

  /* transmits to broadcast (whether directed or otherwise) addresses 
   * is only permitted when associated with a PCB, and that PCB is "connected".
   * This is because IP won't pass up bcast unless we have told it to 
   * do so with _ip_destaddr_{open|close}(). This is akin to the BSD concept
   * and handling of "privileged" sockets.
   * Yes, this implies that we can never be serve for a proto that uses 
   * broadcast (eg bootp). This is a tradeoff I think we can live with.
   *
   * This does not really apply to multicast. IP gets that info from IGMP.
  */
  if (is_broadcast(dest_ip) || is_multicast(dest_ip))
  {
    if (!s)
      return -1;
    if (s->hisaddr != dest_ip)
    {
      s->so_errno = WSO_EACCES;
      return -1;
    }
  }

  /* ------------------------------------------------ */

  /* do NOT subsititute dest_ip/dest_port with those from
  ** the socket when dest_ip == 0 or dest_port == 0.
  ** The caller MUST provide these because there 
  ** would otherwise be a logical race condition in the 
  ** following 'if' clause when the "connection" has been 
  ** closed due to an icmp message.
  */
  if (s)
  {
    if (s->hisaddr != 0 && /* "connected" */
        dest_ip == s->hisaddr)
    {
      ttl = s->ttl;
      tos = s->tos;
      dontchksum = ((s->locflags & WSO_LF_NOCHECKSUM) != 0);
    }
    /* don't auto-bind here */
    if (!local_port)
      local_port = s->myport;
    if (!local_ip)
      local_ip = s->myaddr; /* checked again later */
    so_errnoP = &s->so_errno;
  }

  /* ------------------------------------------------ */

  if (!local_port) /* should never happen */
  {
    *so_errnoP = WSO_EINVAL;
    if (debug_lev)
      DBGPRINTF(("*-UDP OUT: DROP: no local port\n"));
    return (-1);
  }
  /* do NOT subsititute dest_ip/dest_port with those from
  ** the socket when dest_ip == 0 or dest_port == 0.
  ** See comment at top of function.
  */
  if (dest_ip == 0 || dest_port == 0) 
  {
    /* can't send it nowhere */
    *so_errnoP = WSO_EDESTADDRREQ;
    if (debug_lev)
       DBGPRINTF(("*-UDP OUT: DROP: no dest addr/port\n"));
    return (-1);
  }

  if (!data) /* flush request */
  {
    if (debug_lev)
      DBGPRINTF(("*-UDP OUT: done: no data\n"));
    return (0); /* done all */
  }

  /* ------------------------------------------------ */

  /* if IP/ether have not been initialized yet, that will happen now. 
  */
  if (_ip_select_if(&dest_ip, &ifinfo)!=0) 
  {            
    /* oops! link-layer doesn't like 'dest_ip' - probably because
     * its mcast and IF is slip or something. 
    */
    *so_errnoP = WSO_EHOSTUNREACH;
    if (debug_lev)
      DBGPRINTF(("*-UDP OUT: DROP: unsupported dest addr\n"));
    return (-1);
  }

  /* ------------------------------------------------ */

  if (local_ip == 0) /* own addr still unknown? */
    local_ip = ifinfo.ipif_host_addr;

  if (ttl == 0) /* TTL still unknown? */
  {
    if (is_multicast(dest_ip)) 
      ttl = 1; /* don't send worldwide by default */
    else
      ttl = (uint8_t)_ip_get_default_ttl();
  } 

  /* ------------------------------------------------ */

  if (datalen > UDPSEGSIZE_MAX(ifinfo.ipif_mtu))
    datalen = UDPSEGSIZE_MAX(ifinfo.ipif_mtu); 

  /* ------------------------------------------------ */

  /* build udp header
   */
  hdr.srcPort  = HTONS(local_port);
  hdr.dstPort  = HTONS(dest_port);
  hdr.length   = HTONS((uint16_t)(sizeof(hdr) + datalen));
  hdr.checksum = 0;

  if (!dontchksum)
  {
    tcp_PseudoHeader ph;
    memset (&ph, 0, sizeof(ph));
    ph.src = HTONL(local_ip);
    ph.dst = HTONL(dest_ip);
    ph.protocol = IPPROTO_UDP;
    ph.length   = hdr.length;
    ph.checksum = checksum2( &hdr, sizeof(hdr), data, datalen );
    hdr.checksum = ~checksum (&ph, sizeof(ph));
  }

  if (_ip_output_2( local_ip, dest_ip, IPPROTO_UDP, ttl, tos, 0,
                    &hdr, sizeof(hdr), data, datalen, so_errnoP ) == -1)
  {
    if (debug_lev)
      DBGPRINTF(("*-UDP OUT: DROP: _ip_output failed\n"));
    return -1; /* s->so_errno will be set */
  }

  if (debug_lev >= 2)
    __dbg_dump_header(1, 0, &hdr);
  if (debug_lev >= 3)
    DBGDUMP(("*UDP OUT-xxxx", 0, data, datalen ));
  if (debug_lev)
    DBGPRINTF(("*-UDP OUT-xxxx: done. sent %u\n", sizeof(hdr)+datalen));

  return (datalen);
}

/* -------------------------------------------------------------------- */

EXPORTED int udp_send_to(struct sockio_info *__s, const void *buf, 
                         unsigned int buflen, int flags,
                         const uint32_t *_peer_ip, const uint16_t *_peer_port )
{
  udp_pcb FAR *s    = __validate_allsocs(__s);
  int   written     = -1;
  uint32_t timeout     = 0ul;
  int   oldbrkstate = -1;
  uint32_t peer_ip;
  uint16_t  peer_port;

  if (!s)
  {
    if (debug_lev)
      DBGPRINTF(("*udp_send_to() huh? s == NULL.\n"));
    return -1;
  }

  peer_ip   = s->hisaddr;
  if (_peer_ip)
    peer_ip = *_peer_ip;
  peer_port = s->hisport;
  if (_peer_port)
    peer_port = *_peer_port;

  /*
  if (debug_lev)
    DBGPRINTF(("*udp_send_to(%p, %u, %0x, %hI, %u)\n", buf, buflen, flags, 
                        peer_ip, peer_port ));
  */

  if (s->locflags & WSO_LF_CANTSENDMORE)
  {
    s->so_errno = WSO_ESHUTDOWN;
    if (debug_lev)
      DBGPRINTF(("*udp_send_to() WSO_LF_CANTSENDMORE\n"));
    return -1;
  }

  if (buflen == ~0u) /* ... otherwise we won't be able to return -1 */
  {
    s->so_errno = WSO_EINVAL;
    return -1;
  }

  if ((flags & (WSO_MSG_WAITALL|WSO_MSG_DONTWAIT)) == (WSO_MSG_WAITALL|WSO_MSG_DONTWAIT))
  {
    s->so_errno = WSO_EINVAL;
    if (debug_lev)
      DBGPRINTF(("*udp_send_to() both MSG_WAITALL and MSG_DONTWAIT specified\n"));
    return -1;
  }
  else if ((flags & (WSO_MSG_WAITALL|WSO_MSG_DONTWAIT)) == 0)
  {
    if (s->locflags & WSO_LF_BLOCKING)
      flags |= WSO_MSG_WAITALL;
  }

  if (flags & WSO_MSG_WAITALL)
  {
    if (s->xmit_timeout_s) /* otherwise infinite */
      timeout = timer_sec_count() + s->xmit_timeout_s;
    if (debug_lev)
      DBGPRINTF(("*udp_send_to() using waitall - timeout=%lu\n", timeout));
    blocker_yield(0);
  }

  for (;;) /* not 'while (buflen > 0)' - buflen may be zero */
  {
    /* buflen will be capped as needed  by _udp_write_segment(),
    ** zero length send is legal.
    */
    int rc = _udp_write_segment( s, buf, buflen, peer_ip, peer_port,
                                 s->myaddr, s->myport );
    if (rc == -1)
      break;

    if (written == -1)
      written = 0;
    written += rc;
    if (buf)
      buf = ((const uint8_t *)buf) + rc;
    buflen -= rc;  

    if (buflen == 0 || !(flags & WSO_MSG_WAITALL))
      break;

    if (timeout && timer_sec_count() > timeout)
    {
      s->so_errno = WSO_ETIMEDOUT;
      break;
    }

    if (oldbrkstate == -1)
    { 
      if ((oldbrkstate = set_cbreakstate(0)) != 0)
        oldbrkstate = 1;
    }

    /* don't need to hard yield here (it'll just force uneccessary delays) */
    if (blocker_yield(YIELDFLAG_CHKBRK))
    {
      s->so_errno = WSO_EINTR;
      break;
    }
  }

  if (oldbrkstate != -1)
    set_cbreakstate(oldbrkstate);

  if (debug_lev)
    DBGPRINTF(("*udp_send_to() => written = %d\n", written));

  return written;
}

/* ====================================================================== */
/*                             shutdown/destroy                           */
/* ====================================================================== */

static int __pcb_shutdown(udp_pcb FAR *s, int how)
{
  if (how == 0 || how == 2)
  {
    s->rcv_buf.usedlen = 0;
    s->locflags |= WSO_LF_CANTRECVMORE;
  }
  if (how == 1 || how == 2)
  {
    s->locflags |= WSO_LF_CANTSENDMORE;
  }
  return (0);
}

EXPORTED int udp_shutdown (struct sockio_info *__s, int how)
{
  udp_pcb FAR *s = __validate_allsocs(__s);
  if (s)
    return __pcb_shutdown(s, how);
  return -1;
}

/* ----------------------------------------------------------------------- */

/*
 *  Since UDP is stateless, simply reclaim the local-port and
 *  unthread the socket from the list.
 */
static int __pcb_destroy(udp_pcb FAR *s)
{
  udp_pcb FAR *prev = NULL;
  udp_pcb FAR *ds;

  ds = _udp_allsocs;
  while (ds)
  {
    if (ds == s)
      break;
    prev = ds;
    ds = ds->next;
  }

  if (!ds)
    return -1;

  /* must unthread before releasing port */
  /* (could also set myport to zero) */
  if (!prev)
    _udp_allsocs = s->next;
  else
    prev->next   = s->next;

  if (s->myport != 0)
    localport_release (s->myport, IPPROTO_UDP, 0 /* no quiet time */);
  s->myport = 0;

  if (s->hisaddr) /* was "connected" */
  {
    _ip_destaddr_close(s->hisaddr);
  }

  /* free recv buffer */
  FAR_FREE(s->rcv_buf.buf);
  s->rcv_buf.buf = (uint8_t FAR *)0;
  s->rcv_buf.maxlen = 0;
  s->rcv_buf.usedlen = 0;

  /* unlink from user space */
  if (s->sio_link)
    FAR_MEMSET( s->sio_link, 0, sizeof(*(s->sio_link)));

  /* free the pcb itself */
  FAR_MEMSET(s, 0, sizeof(*s));
  FAR_FREE(s);

  return 0;
}

EXPORTED int udp_destroy (struct sockio_info *__s)
{
  udp_pcb FAR *s = __validate_allsocs(__s);
  if (s)
    return __pcb_destroy(s);
  return -1;
}

static void __stack_shutdown(void) 
{                            /* called when the lib is being deinitialized */
  udp_pcb FAR *s = _udp_allsocs;
  if (debug_lev)
    DBGPRINTF(("*UDP shutdown begin\n"));
  while (s)
  {
    udp_pcb FAR *ds = s->next;
    __pcb_destroy(s);
    s = ds;
  }
  if (debug_lev)
    DBGPRINTF(("*UDP shutdown complete\n"));
  return;
}

/* ====================================================================== */
/*                             create                                     */
/* ====================================================================== */

EXPORTED int udp_create(struct sockio_info *__s)
{
  /* BE VERY, VERY CAREFUL HERE! udp_create() should under no
   * circumstances cause the IP stack to boot (be initialized).
   * See also WSO_LF_IFCOUNT and WSO_LF_IFCONF
  */ 
  udp_pcb FAR *s = __validate_allsocs(__s);
  uint8_t FAR *recvbuf = (uint8_t FAR *)0;
  #define DEFAULT_RECVBUF_SIZE 8192u

  if (s)
  {
    __pcb_destroy (s);
  }
  else if (!__s)
  {
    errno = EINVAL;
    return -1;
  }

  s = (udp_pcb FAR *)FAR_MALLOC(sizeof(udp_pcb));
  if (s)
  {
    recvbuf = (uint8_t FAR *)FAR_MALLOC(DEFAULT_RECVBUF_SIZE);
    if (!recvbuf)  
    {
      FAR_FREE(s);
      s = (udp_pcb FAR *)0;
    }
  }
  if (!s)
    return -1; /* errno is already set to ENOMEM */
  

  FAR_MEMSET(s, 0, sizeof(*s));

  s->ttl          = (uint8_t)_ip_get_default_ttl();
  s->tos          = (uint8_t)_ip_get_default_tos();
  s->state        = tcp_StateCLOSED;
  s->locflags     = WSO_LF_BLOCKING; /* note! */

  s->xmit_timeout_s = 128; /* secs - 0 (forever) ... 255 (4.25min) */
  s->recv_timeout_s = 128; /* secs - 0 (forever) ... 255 (4.25min) */
  if (udp_rw_timeout > 0)
  {
    if (udp_rw_timeout > 255)
      udp_rw_timeout = 255;
    s->xmit_timeout_s = s->recv_timeout_s = (uint8_t)udp_rw_timeout;
  }

  s->myaddr       = 0;
  s->hisaddr      = 0;
  s->myport       = 0;
  s->hisport      = 0;

  s->rcv_buf.buf  = recvbuf;
  s->rcv_buf.maxlen  = DEFAULT_RECVBUF_SIZE;
  s->rcv_buf.usedlen = 0;

  s->pcb_marker_top = s->pcb_marker_bottom = (uint32_t)s;

  s->sio_link     = __s;
  __s->dispatch   = __get_dispatch_tab();
  __s->pcb        = (void FAR *)s;

  s->next         = _udp_allsocs;
  _udp_allsocs    = s;

  #undef DEFAULT_RECVBUF_SIZE

  return (0);
}

/* ====================================================================== */
/*                             bind                                       */
/* ====================================================================== */

EXPORTED int udp_bind(struct sockio_info *__s, const uint32_t *laddr, const uint16_t *lport)
{
  udp_pcb FAR *s = __validate_allsocs(__s);

  if (!s)
    return -1;

  if (s->state != tcp_StateCLOSED || /* already listening/"connected" */
      s->myport != 0 ||              /* closed but already bound */
      !laddr || !lport )             /* bad args */
  {
    s->so_errno = WSO_EINVAL; /* TOUTSTATE */
    return -1;
  }

  /* If the following validation of *laddr checks are
   * modified then the incoming packet handler loops 
   * will need modification too.
  */
  if (*laddr && *laddr != ((uint32_t)-1) && !is_local_addr(*laddr))
  {
    s->so_errno = WSO_EINVAL; /* bad address */
    return -1;
  }

  s->myport = localport_reserve(*lport, IPPROTO_UDP, s->locflags );

  if (debug_lev)
    DBGPRINTF(("*udp_bind() lport = %u, s->myport = %u\n", *lport, s->myport ));

  if (s->myport == 0)
  {
    s->so_errno = WSO_EACCES; /* port in use */
    return -1; 
  }

  s->myaddr = *laddr;
  s->state  = tcp_StateLISTEN;
  return (0);
}

/* ====================================================================== */
/*                             connect                                    */
/* ====================================================================== */

EXPORTED int udp_connect(struct sockio_info *__s, const uint32_t *raddr, const uint16_t *rport)
{
  udp_pcb FAR *s = __validate_allsocs(__s);

  if (!s)
    return -1;

  if (!raddr || !rport)
  {
    s->so_errno = WSO_EINVAL; /*EDESTADDRREQ;*/
    return -1;
  }

  if ((s->locflags & WSO_LF_CANTRECVMORE) || s->state != tcp_StateESTAB)
  {
    s->rcv_buf.usedlen = 0;
  }

  /* already "connected" and changing hisaddr/hisport? */
  if ((*raddr != 0 || *rport != 0) &&
     (*raddr != s->hisaddr || *rport != s->hisport))
  {
    _ip_destaddr_close(s->hisaddr);

    if (s->rcv_buf.usedlen != 0) /* clear any inappropriate data from the rx queue */
    {
      uint8_t FAR *pos = s->rcv_buf.buf;
      unsigned int remlen = s->rcv_buf.usedlen;
  
      while (remlen > 0)
      {
        /* pktlen includes size of len field/src ip/src port */
        uint16_t pktlen = *((uint16_t FAR *)&(pos[0]));
        uint32_t pkt_ip = *((uint32_t FAR *)&(pos[0+2]));
        uint16_t pkt_port = *((uint16_t FAR *)&(pos[0+2+4]));
  
        if ((*raddr != 0 && *raddr != pkt_ip) ||
            (*rport != 0 && *rport != pkt_port))
        {
          s->rcv_buf.usedlen -= pktlen;
          remlen -= pktlen;
          FAR_MEMMOVE( pos, pos+pktlen, remlen );
        }
        else
        {
          pos += pktlen;
          remlen -= pktlen;
        }
      }
    }
  }  

  if (s->state == tcp_StateCLOSED && s->myport == 0) /* not bound yet */
  {
    uint32_t baddr = 0; uint16_t bport = 0;
    if (udp_bind(__s, &baddr, &bport) != 0)
      return -1;
  }

  s->hisaddr   = *raddr;
  s->hisport   = *rport;
  s->state     = tcp_StateESTAB;
  s->locflags &= ~(WSO_LF_CANTRECVMORE|WSO_LF_CANTSENDMORE|WSO_LF_ASYNC_ERROR);

  if (s->hisaddr && is_multicast(s->hisaddr))   /* check for multicast */
  {
    if (s->ttl == (uint8_t)_ip_get_default_ttl()) /* user hasn't changed default */
      s->ttl = 1;     /* so we don't send worldwide as default */
  }

  _ip_destaddr_open(s->hisaddr);

  return (0);
}

/* ====================================================================== */
/*                             accept/listen                              */
/* ====================================================================== */

EXPORTED int udp_accept(struct sockio_info *__s, 
                        struct sockio_info *__news, 
                        uint32_t *raddr, uint16_t *rport)
{
  udp_pcb FAR *s = __validate_allsocs(__s);
  if (s)
  {
    __news = __news; /* shaddup compiler */
    raddr = raddr;   /* shaddup compiler */
    rport = rport;   /* shaddup compiler */
    s->so_errno = WSO_EOPNOTSUPP;
  }
  return -1;
}

EXPORTED int udp_listen_bsd(struct sockio_info *__s, int qlen)
{
  udp_pcb FAR *s = __validate_allsocs(__s);
  if (s)
  {
    qlen = qlen; /* shaddup compiler */
    s->so_errno = WSO_EOPNOTSUPP;
  }
  return -1;
}

EXPORTED int udp_listen_wcompat(struct sockio_info *__s, uint16_t lport, uint32_t raddr, uint16_t rport)
{
  udp_pcb FAR *s = __validate_allsocs(__s);

  if (!s)
    return -1;

  if (s->state != tcp_StateCLOSED || /* already doing something */
      s->myport != 0) /* already bound */
  {
    s->so_errno = WSO_EINVAL; /* TOUTSTATE */
    return -1;
  }

  if ((s->locflags & WSO_LF_WATTCOMPAT) == 0) /* must use udp_listen() */
  {
    s->so_errno = WSO_EINVAL;
    return -1;
  }
  
  s->locflags |= WSO_LF_REUSEADDR; /* share allowed if in other in timewait */
  s->locflags |= WSO_LF_REUSEPORT; /* share allowed if all are reuseport */

  lport = localport_reserve(lport, IPPROTO_UDP, s->locflags );
  if (lport == 0)
  {
    s->so_errno = WSO_EACCES; /* port in use (shouldn't happen here) */
    return -1;    
  }

  s->state     = tcp_StateLISTEN;
  s->myport    = lport;
  s->hisport   = rport;
  s->hisaddr   = raddr;

  return (0);
}

/* ====================================================================== */
/*                             UDP/IP handler                             */
/* ====================================================================== */

static void __chargen_reply( uint32_t dest_ip, uint16_t dest_port,
                             uint32_t local_ip, uint16_t local_port, uint16_t pktid )
{
  unsigned int maxlen = UDPSEGSIZE_MAX(_ip_get_mtu() - sizeof(udp_Header));
  uint8_t FAR *d, FAR *dataptr = (uint8_t FAR *)FAR_MALLOC(maxlen);
  int r, rows;

  if (!dataptr)
  {
    if (debug_lev)
      DBGPRINTF(("*-UDP-%04x: chargen failed - no mem\n",pktid));
    return;
  }

  d = dataptr;
  rows = (int)Random(2, maxlen/74 ); /* 2*74 = 148, 7*74 = 518 */
  for (r = 0; r < rows; r++)
  {
    static uint8_t startr = 0;
    int c; uint8_t cc = startr;

    for (c = 0; c < 72; c++)
    {
      *d++ = (uint8_t)('!' + cc);
      if ((++cc) > 94)
      {
        *(d-1) = ' ';
        cc = 0;
      }
    } 
    *d++ = '\r'; 
    *d++ = '\n';
    if ((++startr) > 94)
      startr = 0;
  }

  _udp_write_segment( (udp_pcb FAR *)0, dataptr, (rows * 74), 
                      dest_ip, dest_port, local_ip, local_port );
  FAR_FREE(dataptr);

  pktid = pktid;
  return;
}

/* ----------------------------------------------------------------- */

static void __demux_this(udp_pcb FAR *s, const uint8_t FAR *dataptr, uint16_t datalen,
                         uint32_t srcIP, uint16_t srcPort, uint16_t pktid )
{
  pktid = pktid; /* possibly unused */

  if (debug_lev)
    DBGPRINTF(("*UDP-%04x: found receiver PCB:%p\n", pktid, s ));

  {
    /* should we send a sourcequench if low on room?
    */
    uint16_t size = 2+4+2;
    if (size <= (s->rcv_buf.maxlen - s->rcv_buf.usedlen)) 
    {                             /* otherwise no room at all */
      uint8_t FAR *pos;
      size += datalen; 
      if (size > (s->rcv_buf.maxlen - s->rcv_buf.usedlen)) 
      {
        if (debug_lev)
          DBGPRINTF(("*UDP-%04x: PCB:%p packet stor (%u oct - truncated)\n",
                         pktid, s, datalen ));
        size = (uint16_t)(s->rcv_buf.maxlen - s->rcv_buf.usedlen);
      }
      else
      {
        if (debug_lev)
          DBGPRINTF(("*UDP-%04x: PCB:%p packet stor (%u oct)\n",
                         pktid, s, datalen ));
      }
      pos = &(s->rcv_buf.buf[s->rcv_buf.usedlen]);
      s->rcv_buf.usedlen += size;
      /* note: 'size' includes non-data fields*/
      *((uint16_t FAR *)pos) = size;    pos += sizeof(uint16_t);  size -= sizeof(uint16_t);
      *((uint32_t FAR *)pos) = srcIP;  pos += sizeof(uint32_t); size -= sizeof(uint32_t);
      *((uint16_t FAR *)pos) = srcPort; pos += sizeof(uint16_t);  size -= sizeof(uint16_t);
      FAR_MEMCPY( pos, dataptr, size );
    }
  }
  return;
}

/* ----------------------------------------------------------------- */

/*
 * Handler for incoming udp packets.
 */
static void __udp_ip_recv (const in_Header FAR *ippkt, int is_llbcast,
                                                       int raw_handled)
{
  udp_pcb   FAR *s;
  const udp_Header FAR *udp;
  uint32_t  dstIP, srcIP;
  uint16_t  dstPort, srcPort;
  uint16_t  pktid = NTOHS(ippkt->identification);
  uint16_t  udplen; /* size of udp segment (including udp header) */
  uint16_t  datalen; /* size of data payload */
  const uint8_t FAR *dataptr;
  int is_bcast, is_mcast;

  __validate_allsocs(NULL);

  if (ippkt->length < (in_GetHdrLen(ippkt) + sizeof(*udp)))
  {
    if (debug_lev)
      DBGPRINTF(("*UDP-%04x: DROP (bad length)\n",pktid));
    return;
  }
  udp = (const udp_Header FAR *) (((const uint8_t FAR *)ippkt) + in_GetHdrLen (ippkt));
  udplen = NTOHS(udp->length);
  if (udplen < sizeof(*udp))
  {
    if (debug_lev)
      DBGPRINTF(("*UDP-%04x: DROP (short header)\n",pktid));
    return;
  }
  if (udplen > (NTOHS(ippkt->length) - in_GetHdrLen(ippkt)))
  {
    if (debug_lev)
      DBGPRINTF(("*UDP-%04x: DROP (bad length)\n",pktid));
    return;
  }  
  dataptr = ((const uint8_t FAR *)udp) + sizeof(*udp);
  datalen = (uint16_t)(udplen - sizeof(*udp)); /* zero byte data len is valid */

  dstIP    = NTOHL(ippkt->destination);
  srcIP    = NTOHL(ippkt->source);
  dstPort  = NTOHS(udp->dstPort);
  srcPort  = NTOHS(udp->srcPort);
  /* link-layer broadcast or (directed) ip-broadcast */
  is_bcast = is_llbcast || dstIP == ((uint32_t)-1L) || is_broadcast(dstIP);
  is_mcast = is_multicast(dstIP);

  /* this next test works even when my_ip_addr hasn't been determined
   * yet because the is_bcast test will let what we need for 
   * bootp/dhcp through.
  */
  if (!is_bcast && !is_mcast) /* its not broadcast or multicast */
  {
    if (_ip_get_ipaddr() && !is_local_addr(dstIP)) /* its not for me */
    {
      if (debug_lev)
        DBGPRINTF(("*UDP-%04x: DROP (not forme/bcast/mcast)\n",pktid));
      return;
    }      
  }

  if (udp->checksum)
  {
    tcp_PseudoHeader ph;
    ph.src      = ippkt->source;
    ph.dst      = ippkt->destination;
    ph.mbz      = 0;
    ph.protocol = IPPROTO_UDP;
    ph.length   = udp->length;
    ph.checksum = checksum (udp, udplen);
    if (checksum(&ph,sizeof(ph)) != 0xFFFF)
    {
      if (debug_lev)
        DBGPRINTF(("*UDP-%04x: DROP (bad checksum)\n", pktid ));
      return;
    }
  }

  if (debug_lev)
    DBGPRINTF(("+*UDP IN-%04x: %I:%u->%I:%u (bcast?=%d,mcast?=%d)\n",
                    pktid, ippkt->source, srcPort, ippkt->destination, dstPort,
                    is_bcast, is_mcast ));

  /* if its a standard (unicast) port then mock up the answer 
  ** (if one is required) and send now. We're done then.
  */ 
  if (dstPort == 9) /* discard */
  {
    if (debug_lev)
      DBGPRINTF(("*-UDP-%04x: DROP (sink null)\n", pktid ));
    return;
  }
  if (dstPort == 7 /* echo */ || dstPort == 19 /* chargen */)  
  {
    if (_ip_get_ipaddr() && is_local_addr(dstIP))
    {
      if (debug_lev)
        DBGPRINTF(("*-UDP-%04x: auto-respond (echo/chargen)\n",pktid));
      /* note that source==dest and dest==us=>source  */
      if (dstPort == 19 /* chargen */)
        __chargen_reply( srcIP, srcPort, dstIP, dstPort, pktid );
      else
        _udp_write_segment( (udp_pcb FAR *)0, dataptr, datalen, 
                            srcIP, srcPort, dstIP, dstPort );
    }
    else
    {
      if (debug_lev)
        DBGPRINTF(("*-UDP-%04x: DROP (non-unicast to echo/chargen)\n",pktid));
    }
    return; 
  }

  /* deliver multicast or broadcast datagrams to *all* sockets for which
   * the remote and local (*see note below) addresses and ports match 
   * those of the incoming datagram.
   *
   * local address note: This algo diverges from the BSD one inasfar
   * as not needing to check s->myaddr because s->myaddr will always be
   * a local IP address (or zero) and never a mcast/bcast address.
  */
  if (is_mcast || is_bcast || !_ip_get_ipaddr()) 
  {
    int numdone = 0;
    for (s = _udp_allsocs; s; s = s->next)
    {
      if (dstPort == s->myport)
      {
        if (is_mcast) /* "connected" and dstIP is an mcast address */
        {
          if (s->hisaddr == dstIP) /* connected to the same group */
          {
            __demux_this(s, dataptr, datalen, srcIP, srcPort,pktid); 
            numdone++;
          }
        }
        else if (is_bcast) /* hardware bcast or (directed) IP bcast */
        {
          /* hisaddr must be 255.255.255.255 or nnnnnnn...{+1111} */
          if (s->hisaddr == ((uint32_t)-1) || s->hisaddr == dstIP) 
          {
            __demux_this(s, dataptr, datalen, srcIP, srcPort, pktid); 
            numdone++;
          }
        }
        else /* if (!_ip_get_ipaddr()) */
        {
          /* we don't have an IP address yet, so s->hisaddr must be wild */
          if (srcPort == s->hisport && s->hisaddr == ((uint32_t)-1))
          {
            __demux_this(s, dataptr, datalen, srcIP, srcPort, pktid); 
            numdone++;
          }
        }

        if ((s->locflags & (WSO_LF_REUSEPORT|WSO_LF_WATTCOMPAT)) == 0)
        {
          /* don't bother scanning through the rest if we're not 
           * going to find another instance of dstPort == s->myport
           *
           * There implies a hidden assumption that once WSO_LF_REUSEPORT is 
           * enabled, its not going to be disabled. So sez BSD.
          */
          break;
        }
      } /* if (dstPort == s->myport) */
    } /* for (s = _udp_allsocs; s; s = s->next) */

    /* return here even if we found nothing. No need to send an
     * ICMP unreach for bcast/mcast datagrams.
    */
    if (debug_lev)
    {
      if (!numdone)
        DBGPRINTF(("*-UDP IN-%04x DROP (no receiver for this bcast)\n",pktid));
      else
      {
        if (debug_lev >= 2)
          __dbg_dump_header(0, pktid, udp );
        if (debug_lev >= 3)
          DBGDUMP(("*UDP IN", pktid, dataptr, datalen ));
        DBGPRINTF(("*-UDP IN-%04x: done %d (was bcast)\n",pktid, numdone));
      }
    }
    return;
  }

  if (_ip_get_ipaddr() && is_local_addr(dstIP))
  {
    /* its not bcast/mcast, then only deal with it if its 
     * unicast and for us. my_ip_addr _must_ be known by now.
    */
    udp_pcb FAR *connmatch = (udp_pcb FAR *)0;
    udp_pcb FAR *wlismatch = (udp_pcb FAR *)0;
    udp_pcb FAR *blismatch = (udp_pcb FAR *)0;

    /* its one-on-one to us. First locate a pcb thats "connected".
     * If found, it gets the packet. Fini. If there are two sockets
     * "connected" to the same peer, the one with the more explicitely
     * bound (local) address gets the packet.
     *
     * Simultaneously scan for pcbs that are listening and fulfill all
     * other required characteristics. 
     *
     * Theoretically, *all* sockets that for which the remote and local 
     * addresses and ports match should get the packet, but there
     * is that beast called backwards compatibility....
    */
    for (s = _udp_allsocs; s; s = s->next)
    {
      if (s->myport == dstPort)             /* its (halfway) for me and ...*/
      {
        if (s->state == tcp_StateESTAB)      /* we're "connected" and ...*/
        {
          if (s->hisport == srcPort &&           /* its "from him" and ... */
              (s->hisaddr == srcIP || s->hisaddr == ((uint32_t)-1)) && 
              (s->myaddr == dstIP || s->myaddr == 0)) /* its truly for me */
          {
            /* the 's->hisaddr == ((uint32_t)-1)' test above is a bit of a hack
             * to allow DHCP responses to get through when using old-style sox
            */
            if (s->myaddr == dstIP)
              connmatch = s;
            else if (!connmatch)
              connmatch = s;
          }
        }
        else if (s->state == tcp_StateLISTEN)
        {
          if (s->locflags & WSO_LF_WATTCOMPAT) /* wattcompat "listener" */
          {
            if ((!s->hisport || s->hisport==dstPort) &&
                (!s->hisaddr || s->hisaddr==((uint32_t)-1) || s->hisaddr==dstIP))
            {
              wlismatch = s;
            }
          }
          else /* BSD compat "unconnected" */
          {
            if (s->myaddr == dstIP)
              blismatch = s;
            else if (!blismatch)
              blismatch = s;
          }
        } /* s->state test */

        if ((s->locflags & (WSO_LF_REUSEPORT|WSO_LF_WATTCOMPAT)) == 0)
        {
          /* port is in use by this socket and doesn't have WSO_LF_REUSEPORT
           * set and is not (or was not originally) a socket opened with
           * udp_listen_wcompat(). So stop looping - no other socket is
           * going to pass the (s->myport == dstPort) test.
           *
           * There implies a hidden assumption that once WSO_LF_REUSEPORT is 
           * enabled, its not going to be disabled. So sez BSD.
           */
           break;
        }
      } /* if (s->myport  == dstPort) */
    } /* for (s = _udp_allsocs; s; s = s->next) */

    if (connmatch || wlismatch || blismatch)
    {
      if (debug_lev >= 2)
        __dbg_dump_header(0, pktid, udp);
      if (debug_lev >= 3)
        DBGDUMP(("UDP-data", pktid, dataptr, datalen ));
    }

    s = connmatch;
    if (s) /* connected endpoint found! */
    {
      if (s->myaddr == 0) /* find it faster in the future */
        s->myaddr = dstIP;

      __demux_this(s, dataptr, datalen, srcIP, srcPort, pktid ); 
      if (debug_lev)
        DBGPRINTF(("*-UDP-%04x: done\n",pktid));
      return; /* allow only one "connected" socket to pick up */
    }

    s = wlismatch;  
    if (s) /* WATTCOMPAT listening endpoint found! */
    {
      /* passive endpoint now becomes active if the socket was
       * not a hacky pseudo-active "connected to bcast" endpoint.
       * (in which case, its almost identical to a BSD socket).
      */
      if (s->hisaddr != ((uint32_t)-1)) /* its not a wattudp hack socket */
      {
        /* make passive socket active */
        s->hisaddr = srcIP;
        s->hisport = srcPort;
       
        /* take on value of expected destination
         * unless it is broadcast
        */
        if (is_local_addr(dstIP))
          s->myaddr = dstIP;
      }
      __demux_this(s, dataptr, datalen, srcIP, srcPort, pktid ); 
      if (debug_lev)
        DBGPRINTF(("*-UDP-%04x: done\n",pktid));
      return; 
    }

    s = blismatch;
    if (s) /* BSD listening endpoint found! */
    {
      __demux_this(s, dataptr, datalen, srcIP, srcPort, pktid ); 
      if (debug_lev)
        DBGPRINTF(("*-UDP-%04x: done\n",pktid));
      return;  
    }

    if (!raw_handled) /* no demultiplexer found anything */
    { 
      /* we know is_local_addr(dstIP) is true */
      icmp_retour( SOICMPERR_UNREACH_PORT,
                   ippkt, ((const void FAR *)0), debug_lev );
    }
  } /* if (_ip_get_ipaddr() && is_local_addr(dstIP)) */

  if (debug_lev)
    DBGPRINTF(("*-UDP-%04x: DISCARD (no receiver)\n", pktid ));
  return;
}

/* ====================================================================== */
/*                             ICMP callback                              */
/* ====================================================================== */

static void __udp_icmp_recv(const in_Header FAR *ip, int soicmperr_code)
{
  uint32_t dstIP = NTOHL(ip->destination); /* the addr we sent the pkt to */
  uint32_t srcIP = NTOHL(ip->source);      /* the addr we used for sending */

  if ((!_ip_get_ipaddr() || is_local_addr(srcIP)) && 
      !is_broadcast(dstIP) && !is_multicast(dstIP))
  {
    const udp_Header FAR *udp = (const udp_Header FAR *)
                                ((const uint8_t FAR *)ip + in_GetHdrLen(ip));
    uint16_t dstPort = NTOHS(udp->dstPort);
    uint16_t srcPort = NTOHS(udp->srcPort);
    uint16_t pktid   = NTOHS(ip->identification);
    udp_pcb FAR *s;

    if (debug_lev)
      DBGPRINTF(("*+UDP-%04x: ICMP-IN: msg %d, orig %I:%u->%I:%u\n", pktid,
             soicmperr_code, ip->source, srcPort, ip->destination, dstPort ));

    for (s = _udp_allsocs; s; s = s->next)
    {
      /* ICMP messages are only meaningful for "connected" UDP sockets.
      ** we really need some kind of hostcache mechanism for tracking
      ** unreaches.
      */
      if (dstPort == s->hisport && 
          dstIP   == s->hisaddr &&
          srcPort == s->myport)
      {
        switch (soicmperr_code)
        {
          case SOICMPERR_CANTFRAG:
               /* bogus - we don't do DF */
               break;

          case SOICMPERR_TIMXCEED_FRAG:
               /* ignore - such is UDP */
               break;

          case SOICMPERR_SOURCEQUENCH:
               /* we should slow down writes - we're flooding a router */ 
               break; /* ignore - such is UDP */

          case SOICMPERR_PARAMPROB:
               /* should never happen since we don't set IP options */
               break;

          case SOICMPERR_TIMXCEED_TTL:
               if (s->ttl < 255)
               {
                 s->ttl++;
                 break;
               }
               /* FALL-THROUGH */

          case SOICMPERR_UNREACH_HOST: /* includes unreachable net etc */
          case SOICMPERR_UNREACH_PORT:
               __pcb_shutdown(s, 1);
               s->locflags |= WSO_LF_GOT_FIN;  
               s->locflags |= WSO_LF_ASYNC_ERROR;
               s->so_errno = WSO_EHOSTUNREACH;
               break;
        }
        if (debug_lev)
          DBGPRINTF(("*-UDP-%04x: ICMP-IN match PCB:%p\n", pktid, s ));

        return; 
      }
    } /* for (s = _udp_allsocs; s; s = s->next) */

    if (debug_lev)
      DBGPRINTF(("*-UDP-%04x: ICMP-IN no active match found\n", pktid ));

    pktid = pktid; /* unused unless debugging */
  } /* if (!_ip_get_ipaddr() || is_local_addr(srcIP) || ...) */
  return;
}

/* ====================================================================== */
/*                       exposed handler interface                        */
/* ====================================================================== */

/*
 * Handler for incoming udp packets and for ICMP messages.
 */
static void __udp_handler (const in_Header FAR *ippkt, int protohndlr_flags)
{
  if (protohndlr_flags & PROTOHNDLR_FLAGS_ICMPCB)
  {
    __udp_icmp_recv( ippkt, (protohndlr_flags & 0xff) );
  }
  else
  {
    __udp_ip_recv(ippkt, (protohndlr_flags & PROTOHNDLR_FLAGS_LLBCAST),
                         (protohndlr_flags & PROTOHNDLR_FLAGS_RAWDONE) );
  }
  return;
}

/* ====================================================================== */
/*                             udp read                                   */
/* ====================================================================== */

#define RECVQ_PERPKT_OVRHD (2u+4u+2u) /* per packet overhead */

/*
** Read all the data matching a specific key (ipaddr:port) combination 
** Returns -1 if no match, else number of bytes (>= 0) copied to buf.
**
** If 'buf' is !NULL (and maxlen > 0), copy the data from queue to buf.
** If 'peek' is !0, remove the data from queue.
**
** If 'buf' is NULL, and 'peek' is !0, __read_by_key() determines how much
** data there is waiting (upto the limit specified by 'maxlen').
*/
static int __read_by_key( udp_pcb FAR *s, uint32_t match_addr, uint16_t match_port,
                          void *buf, unsigned int maxlen, int peek )
{
  uint8_t FAR *pos = s->rcv_buf.buf;
  unsigned int used = 0, found = 0;
  unsigned int remlen = s->rcv_buf.usedlen;
  
  while (maxlen > 0 && remlen > 0)
  {
    /* pktlen includes size of len field/src ip/src port */
    uint16_t pktlen = *((uint16_t FAR *)&(pos[0]));
    uint32_t pkt_ip = *((uint32_t FAR *)&(pos[0+2]));
    uint16_t pkt_port = *((uint16_t FAR *)&(pos[0+2+4]));

    if (pkt_ip != match_addr || pkt_port != match_port)
    {
      pos += pktlen;
      remlen -= pktlen;
    }
    else if ((pktlen - RECVQ_PERPKT_OVRHD) > maxlen)
    {
      found++;
      if (buf)
      {
        FAR_MEMCPY(buf, pos + RECVQ_PERPKT_OVRHD, maxlen );
        buf = ((uint8_t *)buf) + maxlen;
      }
      if (!peek)
      {
        FAR_MEMMOVE( pos + RECVQ_PERPKT_OVRHD, 
                     pos + RECVQ_PERPKT_OVRHD + maxlen, 
                     (remlen - RECVQ_PERPKT_OVRHD) - maxlen );
        *((uint16_t FAR *)&(pos[0])) = (uint16_t)(pktlen - maxlen);
      }      
      used += maxlen;
      maxlen -= maxlen;
      break;
    }
    else
    {
      found++;
      if (buf)
      {
        FAR_MEMCPY(buf, pos + RECVQ_PERPKT_OVRHD, pktlen - RECVQ_PERPKT_OVRHD );
        buf = ((uint8_t *)buf) + (pktlen - RECVQ_PERPKT_OVRHD);
      }
      if (!peek)
      {
        FAR_MEMMOVE( pos, pos + pktlen, remlen - pktlen );
        s->rcv_buf.usedlen -= pktlen;
      }
      used += pktlen - RECVQ_PERPKT_OVRHD;
      maxlen -= pktlen - RECVQ_PERPKT_OVRHD;
      remlen -= pktlen;
    }     
  }
  if (!used && !found)
    used = ~0u; /* ((unsigned int)-1) */

  return used;
}  

/*
** Determine how much data there is for a specific key (addr:port pair).
** This is equivalent to __read_by_key(s, addr, port, NULL, UINT_MAX, 1);
**
** Returns -1 if no data is present, else >= 0.
*/
static unsigned int __match_key( udp_pcb FAR *s, 
                                 uint32_t key_addr, uint16_t key_port,
                                 unsigned int offset )
{
  uint8_t FAR *pos = s->rcv_buf.buf + offset;
  unsigned int remlen = s->rcv_buf.usedlen - offset;
  unsigned int used = ~0u; /* ((unsigned int)-1); */
  
  if (s->rcv_buf.usedlen < offset)
    remlen = 0;

  while (remlen > 0)
  {
    /* pktlen includes size of len field/src ip/src port */
    uint16_t pktlen = *((uint16_t FAR *)&(pos[0]));
    uint32_t pkt_ip = *((uint32_t FAR *)&(pos[0+2]));
    uint16_t pkt_port = *((uint16_t FAR *)&(pos[0+2+4]));

    if (key_addr == pkt_ip && key_port == pkt_port)
    {
      if (used == ~0u)
        used = 0;
      used += pktlen - RECVQ_PERPKT_OVRHD;
    }

    pos += pktlen;
    remlen -= pktlen;
  }
  return used;  
}

/*
** Find an addr:port pair for which data is to be returned.
** Return addr:port and amount of data used (may be zero).
** or ~0u ((unsigned int)-1) if no data ready.
*/
static unsigned int __find_key( udp_pcb FAR *s, 
                                uint32_t *key_addr, uint16_t *key_port )
{
  uint8_t FAR *pos = s->rcv_buf.buf;
  unsigned int remlen = s->rcv_buf.usedlen;
  unsigned int used = ~0u;
  
  while (remlen > 0)
  {
    /* pktlen includes size of len field/src ip/src port */
    uint16_t pktlen = *((uint16_t FAR *)&(pos[0]));
    uint32_t pkt_ip = *((uint32_t FAR *)&(pos[0+2]));
    uint16_t pkt_port = *((uint16_t FAR *)&(pos[0+2+4]));

    /*
    ** Count how much data matches this ipaddr:port combination 
    */
    unsigned int mused = (unsigned int)__match_key( s, pkt_ip, pkt_port, 
                                       (unsigned int)(pos - s->rcv_buf.buf) );

    /* if its >= a count we had seen before then this is the biggest sender.
    ** Note that 'match_used' may be zero if all the pkts had zero length.
    */
    if (used == ~0u || mused > used)
    {
      used = mused;
      *key_addr = pkt_ip;
      *key_port = pkt_port;  
#if 1 /* #if 1 == use FIFO, #if 0 == most-data-first-out */
      break;
#endif
    }

    pos += pktlen;
    remlen -= pktlen;
  }
  return used;
}


/*
 * udp_recv - read socket data to 'buf', does large buffering
 * Returns -1 if no data available, else length of data read (may be zero).
*/
EXPORTED int udp_recv_from( struct sockio_info *__s, void *buf, unsigned int buflen, 
                            int flags, uint32_t *src_ip, uint16_t *src_port )
{
  udp_pcb FAR *s = __validate_allsocs(__s);
  unsigned int prevdatalen;
  uint32_t timeout = 0;
  int oldbrkstate = -1;
  int total_read = -1;
  uint32_t key_addr = 0; 
  uint16_t key_port = 0;
  int got_key = 0;

  if (!s)
    return -1;

  if ((flags & (WSO_MSG_WAITALL|WSO_MSG_DONTWAIT)) == (WSO_MSG_WAITALL|WSO_MSG_DONTWAIT))
  {
    s->so_errno = WSO_EINVAL;
    return -1;
  }
  else if ((flags & (WSO_MSG_WAITALL|WSO_MSG_DONTWAIT)) == 0)
  {
    if (s->locflags & WSO_LF_BLOCKING)
      flags |= WSO_MSG_WAITALL;
  }

  /* only handle the trivial conditions -
  ** the rest is done in __nb_recv_plain_data
  */
  if (s->state == tcp_StateCLOSED ||
      (s->locflags & WSO_LF_CANTRECVMORE))
  {
    s->so_errno = WSO_ESHUTDOWN;
    return -1;
  }

  if (flags & WSO_MSG_OOB)
  {
    s->so_errno = WSO_EINVAL;
    return -1;
  }

  if (!buf) /* "drain" */
  {
    s->rcv_buf.usedlen = 0;
    s->so_errno = WSO_EINVAL;
    return -1; /* can't return 0 here. */
  }

  if (buflen == ~0u) /* can't return -1 otherwise */
  {
    s->so_errno = WSO_EINVAL;
    return -1; 
  }

  timeout = 0;
  if (buflen != 0 && (flags & WSO_MSG_WAITALL))
  {
    if (s->recv_timeout_s != 0) /* else infinite */
      timeout = timer_sec_count() + s->recv_timeout_s;
  }

  prevdatalen = s->rcv_buf.usedlen + 1;
  for (;;)
  {
    if (prevdatalen != s->rcv_buf.usedlen) /* queue has changed? */
    {
      unsigned int used = 0;

      if (got_key)
        used = __match_key( s, key_addr, key_port, 0 );
      else
        used = __find_key( s, &key_addr, &key_port );

      if (used != ~0u )
      {
        int do_read = 0; /* 0=no, -1=yes+break, +1=yes+wait */

        if (debug_lev)
          DBGPRINTF(("*UDP data from %hI:%u (len=%u) in queue\n",
                                      key_addr, key_port, used ));

        /* got enough to satify request? */ 
        if (used >= buflen || !(flags & WSO_MSG_WAITALL))
          do_read = -1; /* read and then break */
        else if ((flags & WSO_MSG_PEEK) == 0)
          do_read = +1; /* read and fall through */
        else if (s->rcv_buf.usedlen >= (s->rcv_buf.maxlen - RECVQ_PERPKT_OVRHD))
          do_read = -1; /* queue is full - won't be able to peek more */
        else if (s->locflags & WSO_LF_GOT_FIN) /* got "disconnect" */
          do_read = +1; /* read and fall through */

        if (do_read)
        {
          got_key = 1;
      
          /* fulfill the request (or a part therof) */ 
          used = __read_by_key(s, key_addr, key_port,
                               buf, buflen, (flags & WSO_MSG_PEEK) );

          if (src_ip)
            *src_ip = key_addr;
          if (src_port)
            *src_port = key_port;
          if (total_read == -1)
            total_read = 0;
          total_read += used;
          if (buf)
            buf = ((uint8_t *)buf) + used;
          buflen -= used;

          /*
          if (debug_lev)
            DBGPRINTF(("*did read %u\n", used));
          */

          if (do_read < 0 || buflen == 0) 
          {
            if (s->locflags & WSO_LF_GOT_FIN)
              s->so_errno = WSO_ECONNRESET;
            break;
          }
          if (timeout)
            timeout = timer_sec_count() + s->recv_timeout_s;

        } /* if (do_read) */
      } /* if (used != ~0u) ) */

      prevdatalen = s->rcv_buf.usedlen;
    }

    if (s->state == tcp_StateCLOSED || 
        (s->locflags & (WSO_LF_GOT_FIN|WSO_LF_CANTRECVMORE)))
    {                    
      s->so_errno = WSO_ECONNRESET;
      break;
    }
    if (!(flags & WSO_MSG_WAITALL))
    {
      s->so_errno = WSO_EWOULDBLOCK;
      break;
    }
    if (timeout && timer_sec_count() > timeout)
    {
      s->so_errno = WSO_ETIMEDOUT;
      break;
    }
    if (oldbrkstate == -1)
    {
      if ((oldbrkstate = set_cbreakstate(0))!=0)
        oldbrkstate = 1;
    }
    if (blocker_yield(YIELDFLAG_HARD|YIELDFLAG_CHKBRK))
    {
      s->so_errno = WSO_EINTR;
      break;
    }
  }

  if (oldbrkstate != -1)
  {
    set_cbreakstate(oldbrkstate);
  }
  return total_read;
}

/* ======================================================================= */
/*                     condition wait and socket options                   */
/* ======================================================================= */

/* moved to a separate file for easier maintainance.
 * The code is mostly identical for udp and tcp
*/
#include "utcommon.ci"

/* ======================================================================= */
/*                          transport initialization                       */
/* ======================================================================= */

static const struct wso_usrreq_tab FAR *__get_dispatch_tab(void)
{
  /* please do not initialize this table at run-time. Initializing it
  ** at compile time makes it easy to catch errors.
  */
  static const struct wso_usrreq_tab FAR dispatch_tab = {
     IPPROTO_UDP,
     udp_create,
     udp_bind,
     udp_connect,
     udp_listen_bsd,
     udp_accept,
     udp_shutdown,
     udp_destroy,
     udp_recv_from,
     udp_send_to,
     xxx_poll,
     xxx_gssockopt,
     udp_listen_wcompat
  };
  return &dispatch_tab;
}

/* ---------------------------------------------------------------------- */

static void __ctor_pcudp(void)
{
  __init_private_data();

  /* register "upwards" */
  _wsio_register_dispatch(__get_dispatch_tab());

  /* register "downwards" */
  _ip_proto_register(IPPROTO_UDP, __udp_handler, __stack_shutdown, &debug_lev);

  /* register "port-in-use-scanner" */
  localport_register_pius(IPPROTO_UDP, &debug_lev, xxx_port_in_use_scanner );
  /* xxx_port_in_use_scanner is in utcommon.ci */

  return;
}

#define __SETUP_CONSTRUCTOR_FOR__ __ctor_pcudp
#include "ctordtor.h"

