/*
** Functions in the sock_*() suite that are DNS specific.
**
** Written by Cyrus Patel <cyp@fb14.uni-mainz.de>
**
** 1) All functions named sock_xxx() are userland things.
** 2) All functions named anything else are kernel things.
**
*/
#define WANT_STACK_CHECKING

#include "target.h"     /* standard types/includes */
#include "dns.h"        /* _dns_{get|set}_{host|dom}name() */
#include "userland.h"   /* ourselves (all sock_*() prototypes) */

/* ===================================================================== */
/*                            get/set hostname                           */
/* ===================================================================== */

/*
** int sock_gethostname( char *buffer, unsigned int buflength );
**
** Description
**    sock_gethostname() retrieves the host component of the host's
**    canonical Internet name, ie if the host's canonical name
**    is 'foo.bar.com', sock_gethostname() retrieves 'foo'.
**
**    NOTE: Unlike the BSD gethostname() function, the name returned
**    is always '\0' terminated, even when there is insufficient room
**    to store the complete name.
**
**    A buffer of 64 or more characters in length is guaranteed to 
**    succeed.
**
** Returns
**    sock_sethostname() returns -1 on error or 0 on success.
**
**    If there is insufficient space to store the entire domain name,
**    the name is truncated and followed by a null terminator.
*/
int sock_gethostname( char *buffer, unsigned int buflength )
{
  return _dns_get_hostname( buffer, buflength, 0 );
}

/* --------------------------------------------------------------------- */

/*
** int sock_sethostname( const char *name, unsigned int namelen );
**
** Description
**    sock_sethostname() sets the host component of the host's
**    canonical Internet name. Refer to sock_gethostname() for a 
**    description of the host component/canonical form.
**
**    As with all components of a canonical name, the host component ...
**    - must contain only ASCII (7bit) alphanumeric characters 
**      (0-9, A-Z, a-z), or hyphens ('-');
**    - must not begin or end with a hyphen;
**    - must be less than or equal to 63 bytes in length.
**
**    NOTE: Unlike the BSD gethostname() function, if 'name' has more 
**    than just the hostname component of a canonical name, 
**    sock_sethostname() will use only the component preceding the first 
**    dot. The remainder of the name will be be ignored.
**
** Returns
**    sock_sethostname() returns 0 on success, or -1 on error (name too 
**    long, invalid characters in name etc).
*/
int sock_sethostname( const char *name, unsigned int namelen )
{
  return _dns_set_hostname( name, namelen );
}  

/* ===================================================================== */
/*                            get/set domain name                        */
/* ===================================================================== */

/*
** int sock_getdomainname( char *buffer, unsigned int buflength );
**
** Description
**    sock_getdomainname() retrieves the domain component of the host's
**    canonical name. 
**
**    The domain name is what is lexically appended to searched entities 
**    during DNS name lookups.
**
**    For example, if the domain name is "wide.pasture.net", and you
**    wish to resolve "moo" (no trailing '.' and therefore assumed
**    to be partial), then the hosts looked for will be 
**    "moo.wide.pasture.net.", "moo.pasture.net.", "moo.net."
**    and finally "moo".
**
**    A buffer of 256 or more characters in length is guaranteed to 
**    succeed. Refer to the notes on sock_setdomainname() for a discussion
**    of the domainname length limit imposed by this library.
**
**    NOTE: Unlike the BSD getdomainname() function, the name returned
**    is always '\0' terminated, even when there is insufficient room
**    to store the complete name.
**
** Returns
**    sock_getdomainname() returns zero on success and -1 on error.
**    If there is insufficient space to store the entire domain name,
**    the name is truncated and followed by a null terminator.
*/
int sock_getdomainname( char *buffer, unsigned int buflength )
{
  return _dns_get_domname( buffer, buflength, 0 );
}

/* --------------------------------------------------------------------- */

/*
** int sock_setdomainname( const char *name, unsigned int namelen );
**
** Description 
**    sock_setdomainname() sets the domain component of the host's
**    canonical name.
**
**    The domain name is what is lexically appended to searched entities 
**    during DNS name lookups.
**
**    For example, if the domain name is "dorm.rutgers.edu", and you
**    wish to resolve "moo.cis" (no trailing '.' and therefore assumed
**    to be partial), then the hosts looked for will be
**    "moo.cis.dorm.rutgers.edu.", "moo.cis.rutgers.edu.", "moo.cis.edu."
**    and finally "moo.cis."
**
**    sock_setdomainname( NULL ) or sock_setdomainname( "" ) is an 
**    acceptable way to remove any domain name. Subsequent DNS lookups
**    will then not attempt to append a domain name. 
**
**    An efficient method to prevent an append of the domain name for
**    individual lookups is simply to ensure that the name to lookup
**    is terminated with a dot.
**
**    Each sub-domain segment of the domain name ...
**    - must not contain any characters other than hyphens ('-')
**      and ASCII (7bit) alphanumeric (0-9, A-Z, a-z), characters;
**    - must not begin or end with a hyphen;
**    - must not be greater than 63 bytes in length, except for the 
**      Top Level Domain (TLD: ie, ".edu", ".org" etc) and the Second
**      Level Domain (the "isi" in "isi.edu") which *together* may not 
**      be more than 63 bytes in length.
**    - must end with a dot, except for the Top Level Domain where 
**      it is optional.
**
**    The total length of a canonical domain name (ie, including the
**    the hostname portion, dots and the terminating '\0') must be 
**    less than 256. This is in accordance with RFC883, 
**    "Domain Names - Implementation and Specification", Page 30, 
**    which states: "the total length of label octets and label 
**    length octets that make up a domain name is restricted to 
**    255 octets or less."
**
** Returns
**    sock_setdomainname() returns zero on success or -1 if the name 
**    could not be set (name too long, invalid characters in name etc).
*/
int sock_setdomainname( const char *name, unsigned int namelen )
{
  return _dns_set_domname( name, namelen );
}

/* ===================================================================== */
/*                             DNS resolution                            */
/* ===================================================================== */

/*
** int sock_gethostbyname( const char *hname, in_addr_t *addrlist, int numaddrs);
**
** Description 
**    sock_gethostbyname() performs a DNS lookup for 'hostname'. 
**    The results are stored in 'addrlist'. 
**
**    If 'hostname' is simply a dotted-quad then sock_gethostbyname() 
**    will return the result of a sock_aton().
**
** Returns
**    On success, sock_gethostbyname() returns the number of IP addresses
**    that were resolved and stored in 'addrlist', which may be zero
**    if the nameserver(s) determine that the name does not exist.
**
**    On failure, sock_gethostbyname() returns one of the following 
**    result codes:
**    -1  format error
**    -2  server failure or no timely response from server
**    -3  request cancelled (EINTR)
**    -4  network error
**    -5  request refused
*/
int sock_gethostbyname( const char *hostname, in_addr_t *addrlist, int numaddrs)
{
  return _dns_resbyname( hostname, addrlist, numaddrs );
}

/* --------------------------------------------------------------------- */

/*
** in_addr_t sock_old_resolv( const char *hostname, char *ntoa_buf );
**
** Description 
**    sock_old_resolv() performs a DNS lookup for 'hostname'.
**
**    sock_old_resolv() is for backwards compatibility and is 
**    simply a wrapper around sock_gethostbyname() as follows:
**
**    in_addr_t sock_old_resolv( const char *hostname, char *ntoa_buf )
**    {
**      in_addr_t result;
**      if (sock_gethostbyname( hostname, &result, 1 )!= 1)
**        result = 0UL;    
**      else if (ntoa_buf)
**        sock_ntoa( result, ntoa_buf, sizeof("255.255.255.255"));
**      return result;
**    }
**
** Returns
**    sock_old_resolv() returns the first IP address returned by a
**    nameserver, or 0 if resolution failed. If 'hostname' is simply a 
**    dotted-quad then sock_old_resolv() is equivalent to sock_aton();
*/
in_addr_t sock_old_resolv( const char *hostname, char *ntoa_buf )
{
  in_addr_t result;
  if (sock_gethostbyname( hostname, &result, 1 ) < 1)
    result = 0UL;
  else if (ntoa_buf)
    sock_ntoa( result, ntoa_buf, sizeof("255.255.255.255"));
  return result;
}

/* --------------------------------------------------------------------- */

/*
** int sock_gethostbyaddr( in_addr_t addr, char *hnamebuf, unsigned int buflen);
**
** Description 
**    sock_gethostbyaddr() performs a reverse DNS lookup for 'addr',
**    and stores the result in 'hnamebuf'. 
**
** Returns                               
**    On success, sock_gethostbyaddr() returns the number of characters
**    copied to the buffer, not including the terminating null character. 
**    If the supplied destination buffer is too small to hold the 
**    resolved name, the name is truncated and followed by a null character, 
**    and the return value is equal to 'buflen' minus one. 
**
**    If reverse resolution failed, or the search was interrupted,
**    sock_gethostbyaddr() returns one of the following:
**    -1  format error
**    -2  server failure or no timely response from server
**    -3  request cancelled (EINTR)
**    -4  network error
**    -5  request refused
*/
int sock_gethostbyaddr( in_addr_t addr, char *hnamebuf, unsigned int buflen)
{
  return _dns_resbyaddr( addr, hnamebuf, buflen );
}

/* --------------------------------------------------------------------- */
