/*
** Functions in the sock_*() suite that DO NOT do direct socket I/O
** (those are in sock_io.c)
**
** Written by Cyrus Patel <cyp@fb14.uni-mainz.de>
**
** 1) All functions named sock_xxx() are userland things.
** 2) All functions named anything else are kernel things.
**
*/
#define WANT_STACK_CHECKING

#include "target.h"     /* standard types/includes */
#include "inaddr.h"     /* _my_ntoa() */
#include "ipbase.h"     /* _ip_get_ipaddr() */
#include "loadcfg.h"    /* _xxx_no_config, debug_on */ 
#include "userland.h"   /* ourselves (all sock_*() prototypes) */

/* ===================================================================== */
/*                            addr<->ascii conversion                    */
/* ===================================================================== */

/* 
** const char *sock_ntoa(in_addr_t ipaddr, char *buffer, unsigned int buflen);
**
** Description 
**
**    Converts a binary IP address to an ASCII string representation 
**    of that address (Internet Standard Dotted Notation).
**
**    'buffer' is a location to place the dotted quad, and 'buflen'
**    is the size of that buffer in bytes. A sufficient buffer size 
**    would be 16 characters in length (sizeof("255.255.255.255")).
**    If 'buffer' is NULL, an internal static buffer is used.
**
** Returns
**    sock_ntoa returns a pointer to the array containing the 
**    ASCII string representing the address in dotted notation. 
**
**    If there is insufficient space to store the entire address,
**    the address is truncated to "", ie, it is guaranteed to be
**    invalid (ie, sock_aton() would fail).
*/
const char *sock_ntoa(in_addr_t ipaddr, char *buffer, unsigned int buflen)
{
  return _my_ntoa (ipaddr, buffer, buflen);
}

/* --------------------------------------------------------------------- */

/*
** int sock_aton( const char *cp, in_addr_t *ipaddr );
** 
** Description 
**    sock_aton() converts a character string representing an IP address 
**    expressed in the Internet standard dotted notation to its binary
**    value that may then be subsequently used as an Internet address.
**
**    No check is made as to whether the resultant address is a 
**    valid internet host.
**
** Returns
**    sock_aton() returns zero on failure ('cp' is a malformed address),
**    or non-zero if 'cp' could be converted sucessfully.
*/
int sock_aton(const char *cp, in_addr_t *ipaddr)
{
  return _my_aton(cp, ipaddr);
}

/* --------------------------------------------------------------------- */

/*
** in_addr_t sock_addr( const char *cp );
** 
** Description 
**    Converts an IP address from dotted notation to its 
**    binary representation. No check is made as to 
**    whether the resultant address is a valid internet host.
**
**    sock_addr() is a compatibility wrapper around sock_aton()
**    and is not recommended since the value returned for 
**    a failure is (under some circumstances) a valid address.
**
** Returns
**    On failure, sock_addr() returns 0, otherwise it returns 
**    the binary representation of 'cp'.
*/
in_addr_t sock_addr( const char *cp )
{
  in_addr_t ina;
  if (!sock_aton(cp, &ina))
    ina = 0;
  return ina;
}

/* ===================================================================== */
/*                        Wilkinson Kernel Protector :)                  */
/* ===================================================================== */

unsigned long *__sock_dummy_lpval__(unsigned long what_to_preset) 
{ 
  static unsigned long kdummy; 
  kdummy = what_to_preset; return &kdummy; 
}

unsigned short *__sock_dummy_spval__(unsigned short what_to_preset) 
{ 
  unsigned short *p = ((unsigned short *)__sock_dummy_lpval__(0));
  *p = what_to_preset; return p;
}

int *__sock_dummy_ipval__(int what_to_preset) 
{ 
  int *p = ((int *)__sock_dummy_lpval__(0)); 
  *p = what_to_preset; return p;
}

int __sock_dummy_ifunc__(int what_to_return) 
{ return what_to_return; }

/* ===================================================================== */
/*                            get/set host ID                            */
/* ===================================================================== */

/*
** in_addr_t sock_gethostid( void );
**
** Description 
**    sock_gethostid() retrieves a 32-bit identifier for the current
**    host.  The host ID is unique among all directly-reachable hosts
**    and is normally a DARPA Internet address.
**
**    NOTE: The host id is volatile, and may change during the 
**    lifetime of an application, for example due to a reassignment
**    by a DHCP server.
**
** Returns
**    A 32-bit host ID. sock_gethostid() does not fail.
*/
in_addr_t sock_gethostid(void)
{
  /* sock_gethostid() is not allowed to fail.   
   * It _must_ return a valid, unicast, host address.
  */
  in_addr_t ina = _ip_get_ipaddr(); 
  if (!ina)
    ina = 0x7F000001UL; /* localhost. */
  return ina;
}

/* --------------------------------------------------------------------- */

/*
** int sock_sethostid( in_addr_t id )
**
** Description 
**    sock_sethostid() sets the 32-bit identifier for the current
**    host. The host ID is unique among all directly-reachable hosts
**    and is normally a DARPA Internet address.
**
**    An attempt to assign a host identifier after the IP stack has
**    been initialized will fail.
**
**    NOTE: The host id is volatile, and may change during the 
**    lifetime of an application.
**
** Return
**    -1 on error, 0 on success.
*/
int sock_sethostid(in_addr_t id)
{
  return _ip_set_ipaddr(id);
}

/* --------------------------------------------------------------------- */

/*
** unsigned long  sock_htonl(unsigned long val);
** unsigned long  sock_ntohl(unsigned long val);
** unsigned short sock_htons(unsigned short val);
** unsigned short sock_ntohs(unsigned short val);
**
** Description 
**    hton*() converts from host-endian-order to network-endian-order.
**    ntoh*() converts from network-endian-order to host-endian-order.
**       *l() functions convert long (32bit) integers, and
**       *s() functions convert short (16bit) integers.
**
**    Network data byte order is big-endian, that is, it places/expects 
**    the most significant octet at the lower memory address, while 
**    little-endian processors such as the Intel x86 family place the 
**    least significant octet at the lower address.
**
** Returns      
**    On little-endian processors such as the Intel x86 family, 
**    these functions return 'val' converted to the opposite 
**    endian-order, ie, big-endian values are converted to little-endian
**    and vice-versa. 
**
**    On big-endian processors these functions simply return 'val' 
**    unchanged since host-order and network-order have the same endianness.
**
**    Middle endian-order (aka "PDP-endian") conversion is not supported.
*/
unsigned long  sock_htonl(unsigned long val)  { return HTONL(val); }
unsigned long  sock_ntohl(unsigned long val)  { return NTOHL(val); }
unsigned short sock_htons(unsigned short val) { return HTONS(val); }
unsigned short sock_ntohs(unsigned short val) { return NTOHS(val); }

/* --------------------------------------------------------------------- */
