/*
 * sock_ping() userland function.
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: ul_ping.c,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
/*
** sock_ping() userland function.
**
** Written May 2001 by Cyrus Patel <cyp@fb14.uni-mainz.de>
**
** Standalone usage:
**
**    ping [-t] [-a] [-n count] [-l size] [-i TTL] [-v TOS]
**          [-w timeout] destination-list
**
**    Options:
**    -t             Ping the specified host until stopped with ^C.
**    -a             Resolve addresses to hostnames.
**    -n count       Number of echo requests to send
**    -l size        Send buffer size.
**    -i TTL         Time To Live.
**    -v TOS         Type Of Service.
**    -w timeout     Timeout in milliseconds to wait for each reply.
*/
#define __OSILAYER__ 4 /* nawt really! needed for raw access to IP and ICMP */
#define WANT_STACK_CHECKING

#include "target.h"   /* uint32_t, FAR etc */
#include "timer.h"    /* timer_msec_count() */
#include "blocker.h"  /* blocker_critsect_*(), blocker_yield() */
#include "icmp.h"     /* icmp_echo_reply_hook_{add|del} */
#include "sigblk.h"   /* set_breakstate() */
#include "inaddr.h"   /* is_broadcast(), etc */
#include "inchksum.h" /* checksum() */
#include "ipbase.h"   /* _ip_output_2(), _ip_getipaddr(), in_Header */
#include "userland.h" /* ourselves (all sock_*() prototypes) */

/* ------------------------------------------------------------- */

/* WSO_PINGERR_* look like icmp.h SOICMPERR_* constants, but they're not! */
/* The SOICMPERR_* consts are kernel things and subject to change. */
/* The WSO_PINGERR_* consts are userland things and must remain constant. */
enum 
{    
  /* "Unreachable Net" doesn't need a separate code (same as unreachable host) */
  WSO_PINGERR_UNREACH_HOST  = 1, /* also unreachable net/bad tos/eaccess */
  WSO_PINGERR_UNREACH_PROTO = 2, /* not meaningful for ping, but oh well. */
  WSO_PINGERR_UNREACH_PORT  = 3, /* not meaningful for ping, but oh well. */
  WSO_PINGERR_CANTFRAG      = 4, /* Fragmentation needed and DF set */
  WSO_PINGERR_TIMXCEED_FRAG = 5, /* TTL expired during reassembly */
  WSO_PINGERR_TIMXCEED_TTL  = 6, /* TTL expired in transit */
  WSO_PINGERR_SOURCEQUENCH  = 7, /* Source quench received */
  WSO_PINGERR_PARAMPROB     = 8  /* IP parameter problem *//* maybe some day */
};

/* ------------------------------------------------------------- */

#pragma pack(1)
struct icmp_header 
{
  uint8_t  type;
  uint8_t  code;
  uint16_t checksum;
  /* followed by op data */
};
struct icmp_echo_reqrep
{
  struct icmp_header hdr;
  uint16_t  identifier;
  uint16_t  sequence;
  /* -- user data begins here -- */
};
#pragma pack()

/* ------------------------------------------------------------- */

struct ping_session_t
{
  struct
  {
    uint32_t ina;
    uint16_t sequence;
    uint16_t identifier;
  } req;
  struct {
    uint32_t source;
    uint8_t type;
    uint8_t code;
    int ttl;
    int tos;
    unsigned int len;
    int errnum;
    unsigned long time;
  } reply;
};

static struct ping_session_t FAR *ping_pending;

/* ------------------------------------------------------------- */

static void __echo_reply_hook(const void FAR *_ippkt)
{
  const in_Header FAR *ippkt = (const in_Header FAR *)_ippkt;
  if (ping_pending)
  {
    const struct icmp_echo_reqrep FAR *erep;
    erep = (const struct icmp_echo_reqrep FAR *)
             (((const uint8_t FAR *)ippkt)+in_GetHdrLen(ippkt));

    if (erep->hdr.type == 0) /* echo reply */
    {
      if ( (NTOHS(ippkt->length) - in_GetHdrLen(ippkt)) >= sizeof(*erep) &&
           erep->identifier == ping_pending->req.identifier &&
           erep->sequence   == ping_pending->req.sequence)
      {
        ping_pending->reply.type = 0; /* echo reply */
        ping_pending->reply.code = 0;
        ping_pending->reply.source = NTOHL(ippkt->source);
        ping_pending->reply.ttl  = ippkt->ttl;
        ping_pending->reply.tos  = ippkt->tos;
        ping_pending->reply.len  = NTOHS(ippkt->length)-in_GetHdrLen(ippkt);
        ping_pending->reply.time = timer_msec_count();
        ping_pending->reply.errnum = 0;
        ping_pending->req.ina = 0;
      }
    }
    else if (erep->hdr.type == 3  || /* Destination Unreachable */
             erep->hdr.type == 4  || /* Source Quench */
             erep->hdr.type == 11 || /* Time Exceeded for a Datagram */
             erep->hdr.type == 12)   /* Bad IP header option (paramprob) */
    {
      uint32_t source = ippkt->source;
      if ((NTOHS(ippkt->length) - in_GetHdrLen(ippkt)) >= 
          (sizeof(struct icmp_header)+sizeof(in_Header)+8))
      {
        ippkt = (const in_Header FAR *)
                 (((const uint8_t FAR *)ippkt) + in_GetHdrLen(ippkt) + 
                 sizeof(struct icmp_header) + sizeof(uint32_t) );
        if (ippkt->proto == IPPROTO_ICMP &&
            NTOHL(ippkt->destination) == ping_pending->req.ina &&
            NTOHL(ippkt->source) == _ip_get_ipaddr() )
        {
          const struct icmp_echo_reqrep FAR *erep2;
          erep2 = (const struct icmp_echo_reqrep FAR *)
            (((const uint8_t FAR *)ippkt)+in_GetHdrLen(ippkt));

          if (erep2->hdr.type   == 8 && /* echo request */
              erep2->identifier == ping_pending->req.identifier &&
              erep2->sequence   == ping_pending->req.sequence)
          {
            int errnum = erep->hdr.code;
            if (erep->hdr.type == 3) /* ICMP_UNREACH */
            {
              if (errnum == 2) /* proto unreachable */
                errnum = WSO_PINGERR_UNREACH_PROTO;       /* HUH! */
              else if (errnum == 3) /* port unreachable */
                errnum = WSO_PINGERR_UNREACH_PORT;        /* HUH! */
              else if (errnum == 4) /* Fragmentation needed and DF set */
                errnum = WSO_PINGERR_CANTFRAG;
              else if (errnum >= 9) /* various access denied/TOS messages */
                errnum = WSO_PINGERR_UNREACH_HOST;
              else /* host/net unreachable, host/net isolated, host unknown */
                errnum = WSO_PINGERR_UNREACH_HOST;
            }
            else if (erep->hdr.type == 4) /* ICMP_SOURCEQUENCH */
            {
              errnum = WSO_PINGERR_SOURCEQUENCH;
            }
            else if (erep->hdr.type == 11) /* ICMP_TIMXCEED */
            {
              if (errnum == 0)
                errnum = WSO_PINGERR_TIMXCEED_TTL;
              else if (errnum == 1)
                errnum = WSO_PINGERR_TIMXCEED_FRAG;
              else /* huh? */
                errnum = -1;
            }
            else if (erep->hdr.type == 12) /* ICMP_PARAMPROB */
            {
              errnum = WSO_PINGERR_PARAMPROB;
            }
            else
            {
              errnum = -1;
            }
            ping_pending->reply.type = erep->hdr.type;
            ping_pending->reply.code = erep->hdr.code;
            ping_pending->reply.source = NTOHL(source);
            ping_pending->reply.ttl = ippkt->ttl;
            ping_pending->reply.tos = ippkt->tos;
            ping_pending->reply.len = 0;
            ping_pending->reply.errnum = errnum;
            ping_pending->reply.time = timer_msec_count();
            ping_pending->req.ina = 0;
          }
        } /* was our ICMP echo */
      }  /* data after ippkt has right length */
    } /* erep->type */
  } /* if (ping_pending) */
  
  return;
}

/* ------------------------------------------------------------- */

static void __patternize( uint8_t FAR *buf1, unsigned int bufsz1,
                          uint8_t FAR *buf2, unsigned int bufsz2,
                          const void FAR *pattern, unsigned int _pattlen )
{
  /* fill the buffer(s) will 'pattern' (or 0..0xff if no pattern)
  **
  ** Pattern length grows exponentially during copy, so even in the
  ** worst case scenario (a two byte pattern length and odd bufsizes)
  ** a full 64K patternize still takes only takes 18 memcpy()s.
  ** The best case scenario is a 256 byte pattern with bufsizes that
  ** are evenly divisible by 256 (=> 8 memcpy()s).
  */
  if (pattern && _pattlen == 1)
  {
    FAR_MEMSET(buf1, *((const uint8_t FAR *)pattern), bufsz1 );
    if (buf2)
      FAR_MEMSET(buf2, *((const uint8_t FAR *)pattern), bufsz2 );
  }
  else
  {
    uint8_t FAR *tmpbuf;
    unsigned int i, len;
    unsigned int pattlen = _pattlen;
    unsigned int numpats = 0;

    if (!pattern || !pattlen)
    {
      pattlen = 256;
      if (pattlen > bufsz1)
        pattlen = bufsz1;
      tmpbuf = buf1;
      len = pattlen;
      for (i = 0; i < len; i++)
        *tmpbuf++ = (unsigned char)i;
    }
    else
    {
      if (pattlen > bufsz1)
        pattlen = bufsz1;
      FAR_MEMCPY( buf1, pattern, pattlen );
      pattern = (const void FAR *)0;
    }
     
    tmpbuf = buf1 + pattlen;
    len = bufsz1 - pattlen;

    while (len >= pattlen)
    {
      FAR_MEMCPY(tmpbuf, buf1, pattlen);
      tmpbuf += pattlen;
      len -= pattlen;
      pattlen += pattlen;
      numpats++;
    }
    if (len > 0)  
    {
      FAR_MEMCPY(tmpbuf, buf1, len);
      numpats++;
    }
    /* DBGDUMP(("", 0, buf1, bufsz1 )); */

    if (buf2)
    { 
      i = (pattlen - len);
      if (i > bufsz2)
        i = bufsz2;
      FAR_MEMCPY( buf2, buf1 + len, i );
      tmpbuf = buf2 + i;
      len = bufsz2 - i;
      numpats++;
    
      while (len >= pattlen)
      {
        FAR_MEMCPY(tmpbuf, buf1, pattlen);
        tmpbuf += pattlen;
        len -= pattlen;
        numpats++;
      }

      if (len > 0)
      {
        FAR_MEMCPY(tmpbuf, buf1, len);
        numpats++;
      }
      /* DBGDUMP(("", 0, buf2, bufsz2 )); */
    }
    /* DBGPRINTF(("NUM PATTS = %u\n", numpats )); */

  } /* not one byte pattern */

  return;
}

/* ------------------------------------------------------------- */

int sock_ping( in_addr_t *inaP, int *ttlP, int *tosP, unsigned int *bufsizeP, 
               unsigned long *timeoutP, int df, 
               const void *pattern, unsigned int pattlen)
{
  static uint16_t last_seqnum = 0;
  struct ping_session_t ps;
  int ttl = 0xff, tos = 0;
  unsigned int bufsize = 32;
  uint32_t ina;
  unsigned long start_time, timeout = 1000ul;
  struct icmp_echo_reqrep FAR *ereq;
  int rc, so_errno, old_brkstate;
  unsigned int bufsz1, bufsz2;
  void FAR *buf2;

  ina = 0;
  if (inaP)
    ina = *inaP;

  if (!ina || is_broadcast(ina) || is_multicast(ina))
  {
    return WSO_EADDRNOTAVAIL;
  }
  if (ttlP)
  {
    ttl = *ttlP; 
    if (ttl == 0 || ttl > 0xff)
      ttl = 0xff;
    else if (ttl < 1)
      return WSO_EINVAL;
  }
  if (tosP)
  {
    tos = *tosP;
    if (tos > 0xff)
      tos = 0xff;
    else if (tos < 1)
      tos = 0;
  }
  if (timeoutP)
  {
    timeout = *timeoutP;
    if (timeout > 256000UL) /* no point waiting after TTL expires */
      timeout = 256000UL;
  }
  if (bufsizeP)
  {
    bufsize = *bufsizeP;
    if (bufsize > ( (((unsigned int)IP_OFFMASK)<<3) -
                    (sizeof(in_Header)+sizeof(*ereq)) ) )
      return WSO_EMSGSIZE;
  }

  bufsz1 = bufsize + sizeof(*ereq);
  bufsz2 = 0;

  if (bufsz1 > (0xffffu >> 1)) 
  {
    /* warning! don't fiddle with this balance */
    bufsz2 = (bufsz1 >> 1);
    bufsz1-= bufsz2;
  }

  ereq = (struct icmp_echo_reqrep FAR *)FAR_MALLOC(bufsz1);
  buf2 = (bufsz2 ? FAR_MALLOC(bufsz2) : ((void FAR *)0) );

  if (!ereq || (bufsz2 && !buf2))
  {
    if (ereq)
      FAR_FREE(ereq);
    if (buf2)
      FAR_FREE(buf2);
    return WSO_ENOBUFS;
  }

  __patternize( ((uint8_t FAR *)ereq) + sizeof(*ereq), bufsz1 - sizeof(*ereq), 
                ((uint8_t FAR *)buf2), bufsz2, pattern, pattlen );

  ps.req.ina        = ina;
  ps.req.sequence   = HTONS(++last_seqnum);
  ps.req.identifier = (uint16_t)Random(0,0xffff);

  ereq->hdr.type     = 8;
  ereq->hdr.code     = 0;
  ereq->hdr.checksum = 0;
  ereq->sequence     = ps.req.sequence;
  ereq->identifier   = ps.req.identifier;
  ereq->hdr.checksum = ~checksum2( ereq, bufsz1, buf2, bufsz2 );

  blocker_critsect_begin();
  ping_pending = &ps;
  icmp_echo_reply_hook_add(__echo_reply_hook);
  old_brkstate = set_cbreakstate(0);

  start_time = timer_msec_count();

  rc = _ip_output_2( _ip_get_ipaddr(), ina, IPPROTO_ICMP,
                     (uint8_t)ttl, (uint8_t)tos, (df ? IP_DF : 0), 
                     ereq, bufsz1, buf2, bufsz2, &so_errno );

  FAR_FREE(ereq);
  if (buf2)
    FAR_FREE(buf2);

  if (rc != -1)
  {
    so_errno = 0;
    while (ps.req.ina && !so_errno)
    {
      if (blocker_yield(/*YIELDFLAG_HARD|*/YIELDFLAG_CHKBRK))
        so_errno = WSO_EINTR;
      else if ((timer_msec_count() - start_time) > timeout)
        so_errno = WSO_ETIMEDOUT;
    }
    if (!ps.req.ina)
    {
      *inaP = ping_pending->reply.source;
      if (ttlP)
        *ttlP = ps.reply.ttl;
      if (tosP)
        *tosP = ps.reply.tos;
      if (timeoutP)
      {
        *timeoutP = ps.reply.time - start_time;   
        if (*timeoutP == 0)
        {
          /* fudge the timediff ONLY if a zero timediff is pretty
           * darn close (in meaning) to a one msec timediff.
           * In other words, read this if() as ...
           *   if (*timeoutP < tres && *timeoutP >= (tres>>1))
           *      *timeoutP = tres;
          */
          *timeoutP = timer_msec_tres();
          if (*timeoutP != 1) 
            *timeoutP = 0;
        }
      }

      if (ps.reply.type == 0) /* echo reply */
      {
        so_errno = 0;
        if (bufsizeP)
          *bufsizeP = ps.reply.len - sizeof(*ereq);
      }
      else
      {
        so_errno = WSO_EREMOTE;
        if (bufsizeP)
          *bufsizeP = ps.reply.errnum;
      }
    }    
  } 

  set_cbreakstate(old_brkstate);
  icmp_echo_reply_hook_del(__echo_reply_hook);
  ping_pending = (struct ping_session_t FAR *)0;
  blocker_critsect_end();

  return so_errno;
}

/* ------------------------------------------------------------- */

#if defined(STANDALONE)
#include <stdio.h>  /* printf(), fprintf() */
#include <string.h> /* memset(), strlen() */
#include <stdlib.h> /* atol() */
#include <limits.h> /* ULONG_MAX */
#include <userland.h> /* sock_*(), WSO_E* */

static int main(int argc, char *argv[])
{
}
#endif /* STANDALONE */  
