/*
** Functions in the (userland part of the library)
** that directly do socket I/O
**
** Written by Cyrus Patel <cyp@fb14.uni-mainz.de>
**
** 1) All functions named sock_xxx() are userland things.
** 2) All functions named anything else are kernel things.
*/
#define __OSILAYER__ 5
#define __COMPILING_SOCKIO__
#define WANT_STACK_CHECKING

#include "target.h"    /* standard types/prototypes, valid_addr_rw() */
#include "loadcfg.h"   /* _conf_add_confextn() */
#include "timer.h"     /* timer_sec_count() */
#include "blocker.h"   /* blocker_delay() */
#include "ipbase.h"    /* IPPROTO_xxx */
#include "printk.h"    /* _printk() */
#include "utsio.h"     /* transport interface */
#include "userland.h"  /* ourselves (all sock_*() prototypes) */

/* ====================================================================== */
/*             functions that have been macroized in userland.h           */
/* ====================================================================== */

#define DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
/*
** This next set of functions are all macro-ized in userland.h
** However, if you desire binary compatibility, turn
** off DONT_GENERATE_CODE_FOR_MACROIZED_FXNS, and rebuild.
*/
#if !defined(DONT_GENERATE_CODE_FOR_MACROIZED_FXNS)

#ifdef __cplusplus
extern "C" {
#endif

extern int tcp_listen( void *s, in_port_t lport, in_addr_t raddr, in_port_t rport, 
                       ProtoHandler handler, in_port_t timeout );
extern int tcp_open( void *s, in_port_t lport, in_addr_t raddr, in_port_t rport, 
                       ProtoHandler handler );
extern int udp_listen( void *s, in_port_t lport, in_addr_t raddr, in_port_t rport, 
                       ProtoHandler handler );
extern int udp_open( void *s, in_port_t lport, in_addr_t raddr, in_port_t rport, 
                       ProtoHandler handler );

extern void sock_abort( void *s );
extern void sock_close( void *s );

extern int sock_write( void *s, const void *dp, unsigned int len );
extern int sock_fastwrite( void *s, const void *dp, unsigned int len );
extern int sock_enqueue( void *s, const void *dp, unsigned int len );
extern int sock_flush( void *s );
extern int sock_noflush( void *s );
extern int sock_flushnext( void *s );
extern int sock_tbsize( void *s );
extern int sock_tbused( void *s );
extern int sock_tbleft( void *s );

extern int sock_recv( void *s, void *dp, unsigned int len );
extern int sock_recv_init( void *s, void *space, unsigned int len);
extern int sock_rbleft( void *s );
extern int sock_rbsize( void *s );
extern int sock_rbused( void *s );
extern int sock_dataready( void *s );
extern int sock_bytesready( void *s );

extern void udp_SetTTL( void *s, unsigned char ttl );
extern int sock_get_errno( void *s );
extern int sock_get_stateno( void *s );
extern const char * sockerr( void *s );
extern const char * sockstate( void *s );
extern void sock_mode( void *s, int mode );
extern void sock_yield( void *s, void (*fn)(void) );

extern void ip_timer_init( void *s, int secs );
extern int ip_timer_expired( void *s );
extern int tcp_tick( void *s );
extern int tcp_established(void * s);
extern int _ip_delay0 (void *s, int seconds, int (*fn)(void *s), int *statusptr);
extern int _ip_delay1 (void *s, int seconds, int (*fn)(void *s), int *statusptr);
extern int _ip_delay2 (void *s, int seconds, int (*fn)(void *s), int *statusptr);

#ifdef __cplusplus
}
#endif
#endif /* #if !defined(DONT_GENERATE_CODE_FOR_MACROIZED_FXNS) */

/* ====================================================================== */
/*                           sock_io configuration                        */
/* ====================================================================== */

static struct
{
  void (*so_yield)(void); /* default sock_yield function */
} 
FAR so_defaults; 

/* ====================================================================== */
/*                socket function dispatch primitives                     */
/* ====================================================================== */

static const struct wso_usrreq_tab FAR * FAR sio_pxlat_tab[5];

/*
** INTERNAL: __TAB_LOOKUP() is used by all of the following 
** sock I/O primitives.
**
** If 's' is non-NULL, __TAB_LOOKUP() verifies that the socket is
** valid by matching 's->dispatch' to a previously registered 
** dispatch table.
**
** If 's' is NULL, wso_usrreq_tab_locate() locates the 
** dispatch table for 'ip_proto'. Note that 'ip_proto' may not be 
** zero.
*/
#define GETTAB_BY_PROTO(__ip_proto) __TAB_LOOKUP(0,__ip_proto)
#define GETTAB_BY_SOCK(__s)         __TAB_LOOKUP(__s,0)

static const struct wso_usrreq_tab FAR *__TAB_LOOKUP(void *s, int ip_proto)
{
  unsigned int i;
  if (s)
  {
    if (valid_addr_rw(s, sizeof(struct sockio_info)))
    {
      const struct wso_usrreq_tab FAR *ur = ((struct sockio_info *)s)->dispatch;
      for (i = 0; i < (sizeof(sio_pxlat_tab)/sizeof(sio_pxlat_tab[0])); i++)
      {
        if (!sio_pxlat_tab[i])
          break;
        if (sio_pxlat_tab[i] == ur)
          return ur;
      }
    }
  }
  else if (ip_proto) 
  {
    for (i = 0; i < (sizeof(sio_pxlat_tab)/sizeof(sio_pxlat_tab[0])); i++)
    {
      if (!sio_pxlat_tab[i])
        break;
      if (sio_pxlat_tab[i]->ip_proto == ip_proto)
        return sio_pxlat_tab[i];
    }
  }
  return (struct wso_usrreq_tab FAR *)0;
}

/* --------------------------------------------------------------------- */

/*
** INTERNAL: __wso_dispatch_register() is used by the protocols to
** register their dispatch tables. Each (user-use) protcol does this 
** at least once.
*/
int _wsio_register_dispatch(const struct wso_usrreq_tab FAR *ur)
{
  int ip_proto = ur->ip_proto;
  unsigned int i;
  for (i = 0; i < (sizeof(sio_pxlat_tab)/sizeof(sio_pxlat_tab[0])); i++)
  {
    if (!sio_pxlat_tab[i])
      break;
    if (sio_pxlat_tab[i]->ip_proto == ip_proto)
    {
      sio_pxlat_tab[i] = ur;
      return (0);
    }
  }
  for (i = 0; i < (sizeof(sio_pxlat_tab)/sizeof(sio_pxlat_tab[0])); i++)
  {
    if (!sio_pxlat_tab[i])
    {
      sio_pxlat_tab[i] = ur;
      return (0);
    }
  }
  return (-1);
}

/* ====================================================================== */
/*                       open/bind/listen/accept                          */
/* ====================================================================== */

/*
** int sock_create(void *s, int ip_proto );
**
** Description
**    sock_create initializes a previously uninitialized socket 's'
**    for communications.
**
**    'ip_proto' determines which protocol the socket will belong to.
**
**    Sockets initialized with sock_create() are essentially 4.2BSD 
**    compatible. Sockets created with {tcp|udp}_listen() and 
**    {tcp|udp}_open() are not.
**
** Returns
**    sock_create() returns 0 if successful or -1 on error. The cause
**    for the error is stored in 'errno' as a normal ISO-C errno value:
**    ENOMEM  - Not enough memory to initialize the socket.
**    EINVAL  - 's' is NULL or 'ip_proto' is unrecognized.
**    EMFILE  - too many open sockets.
**    EIO     - I/O error (stack could not be initialized)
*/
int sock_create(void *s, int ip_proto)
{
  const struct wso_usrreq_tab FAR *ur = GETTAB_BY_PROTO(ip_proto);
  int rc = -1;
  if (!ur || !s)
    errno = EINVAL;
  else
  {
    _conf_load(); /* make sure config is loaded */
    sock_destroy(s); /* in case it was previously open */
    rc = (*(ur->so_create))(((struct sockio_info *)s));
    if (rc == 0 && so_defaults.so_yield)
      sock_set_yieldcb(s, so_defaults.so_yield );
  }
  return rc;
}

/* --------------------------------------------------------------------- */

/*
** int sock_bind(void *s, const in_addr_t *laddr, const in_port_t *lport);
**
** Description
**    sock_bind gives the socket 's' the local address 'laddr'/'lport'. 
**    Traditionally, this is called assigning a name to a socket. 
**
**    When a socket is initialized with sock_create(), it exists in a 
**    name space (address family) but has no name assigned. 
**
**    sock_bind() is optional when the socket will be "connected" 
**    to a remote peer, but it is normally necessary to assign a 
**    name when a socket will be listening for incoming connections.
**    For connectionless protocols (UDP), once a socket is bound it 
**    is ready to send or receive data.
**
**    'lport' and 'laddr' may not be NULL, but either (or both) may 
**    have a zero value.
**
**    If 'lport' has a zero value, the protocol assigns an unused, 
**    nonprivileged value between 1024 and 5000 (inclusive). Values 1 
**    through 1023 and 5001 through 65535 are reserved for well-known
**    server ports.
**
**    'laddr' can usually be set to a zero value, allowing the protocol
**    to determine the local IP address, depending on the route. An
**    explicit address (in host order) may be set if it belongs to
**    the node. sock_gethostid() may be used to obtain the default 
**    local IP address. sock_getsockname() may be used to obtain
**    the address that a socket is bound to.
**
**    sockets initialized with {tcp|udp}_listen() and and {tcp|udp}_open() 
**    cannot use sock_bind(). Such sockets are already bound when those 
**    functions return.
**
** Returns
**    sock_bind() returns 0 if successful or -1 on error. The cause
**    for the error may be retrieved with sock_get_errno().
*/
int sock_bind(void *s, const in_addr_t *laddr, const in_port_t *lport)
{
  const struct wso_usrreq_tab FAR *ur = GETTAB_BY_SOCK(s);
  if (ur)
    return (*(ur->so_bind))(((struct sockio_info *)s), laddr, lport);
  return -1;
}

/* --------------------------------------------------------------------- */

/*
** int sock_connect(void *s, const in_addr_t *laddr, const in_port_t *lport);
**
** Description
**    sock_connect() establishes a connection to a peer.
**
**    If the socket 's' was created for use with a connection-oriented
**    protocol (TCP), a connection to the remote addr 'laddr':'lport'
**    is initiated by sock_connect(). Generally, connection-oriented 
**    sockets can successfully connect to a peer only once.
**    'laddr' and 'lport' may not be NULL and may not have a zero value
**    for this kind of socket.
**
**    If the socket 's' was created for use with a connectionless
**    protocol (UDP), sock_connect() merely associates the 's'
**    with 'laddr':'lport', and subsequent sock_write()s default to
**    using 'laddr':'lport' as the destination, and only data received
**    from 'laddr':'lport' will be readable. Such a socket may use
**    sock_connect() multiple times to change their association, and
**    may end an association by passing NULL or a zero value in 
**    'laddr':'lport'.
**
**    If sock_connect() was called for a socket that had not been
**    previously bound using sock_bind(), the socket is automatically
**    bound.
**
**    If the socket is in blocking mode, sock_connect() blocks
**    until the the connection has been established successfully
**    or it times out, or an error condition is detected.
**
**    For non-blocking mode, the connection attempt cannot be completed
**    immediately and sock_connect() returns with -1/EWOULDBLOCK.
**    In this case, the application may:
**
**    1. Use sock_poll() to determine the completion of the connection 
**       request by checking for WSO_ASYSEL_CONNECT, or
**    2. Call sock_connect() again (with the same 'raddr' and 'rport').
**       sock_connect() will (always) return -1, and sock_get_errno() 
**       will report:
**       EISCONN     - the connection has been established 
**       EINPROGRESS - connect establishment is still in progress
**       Any other errno (ECONNREFUSED/ExxxUNREACH/ETIMEDOUT) 
**       reflects a failed connection.
**
**    The timeout for a non-blocking connect is the same as that used
**    for sock_recv_from() and may be changed with 
**    sock_set_opt(WSO_LF_RCVTIMEO)
**     
**    An attempt to write to a socket that is still in the process of
**    establishing a connection will block (irrespective of blocking
**    mode or WSO_MSG_* flag) until the connection completes or an 
**    error occurs.
**
**    When connected sockets break (that is, become closed for 
**    whatever reason), they should be discarded and recreated. It 
**    is safest to assume that when things go awry for any reason 
**    on a connected socket, the application must discard and recreate 
**    the needed sockets in order to return to a stable point.
**
** Returns
**    sock_connect() returns 0 if successful or -1 on error. The cause
**    for the error may be retrieved with sock_get_errno().
*/
int sock_connect(void *s, const in_addr_t *raddr, const in_port_t *rport)
{
  const struct wso_usrreq_tab FAR *ur = GETTAB_BY_SOCK(s);
  if (ur)
    return (*(ur->so_connect))(((struct sockio_info *)s), raddr, rport);
  return -1;
}

/* --------------------------------------------------------------------- */

/*
** int sock_accept(void *s, void *news, in_addr_t *raddr, in_port_t *rport);
**
** Description
**     sock_accept() is used with connection-based socket types (TCP).
**     The argument 's' is a socket that has been initialized with the 
**     sock_create() function and bound to an address with the 
**     sock_bind() function. The sock_listen() function prepares the 
**     socket to accept incoming connections; then, the sock_accept()
**     function accepts the connection. The sock_accept() function 
**     extracts the first connection from the queue pending connections, 
**     and copies it to 'news' - 'news' will then have the same properties 
**     as 's', but will be connected rather than listening. 's' remains
**     open and continues to listen for connections.
**
**     The result 'raddr' and 'rport' pointers may be used to obtain
**     the address of the newly accepted connection.
**
**     If no connection is pending an accept(), sock_accept() blocks 
**     if the socket is in blocking mode, or returns immediately 
**     with -1/EWOULDBLOCK if the socket is in non-blocking mode.
**
**     As per the XTI specification, it is legal for 'news' and 's' 
**     to have the same value (ie, be pointers to the same data object).
**     In such a case (and if there is a connect indication pending),
**     's'/'news' stops listening and becomes connected.
**
** Returns
**     sock_accept() returns zero if a new connection was accepted 
**     successfully - 'news' may then be read and written to just 
**     as if the socket had been used to connect to a remote service.
**
**     sock_accept() returns -1 if there are no connections pending, or 
**     the socket was not initialized with sock_create() or was not 
**     bound with sock_bind() or was not placed in listening mode with 
**     sock_listen(), or 's' is not associated with a connection-oriented
**     protocol. The precise cause for the error may be obtained
**     with sock_get_errno().
*/
int sock_accept(void *s, void *news, in_addr_t *raddr, in_port_t *rport)
{
  const struct wso_usrreq_tab FAR *ur = GETTAB_BY_SOCK(s);
  if (ur)
  {
    if (s != news) /* its ok for them to be the same, but if they ... */
      sock_destroy(news); /* ... are not, then 'news' must be closed. */
    return (*(ur->so_accept))(((struct sockio_info *)s), 
                              ((struct sockio_info *)news), raddr, rport);
  }
  return -1;
}  

/* --------------------------------------------------------------------- */

/*
** int sock_listen(void *s, int backlog);
**
** Description 
**    sock_listen() is used to prepares a socket to accept connections.
**    The socket must have been previously initialized with sock_create()
**    and bound to a local address/port with sock_bind().
**
**    Only sockets that are associated with a connection-oriented 
**    protocol (TCP) may be placed in listen mode with sock_listen().
**    Sockets associated with a connectionless protocol (UDP) are
**    "listening" once they are bound with sock_bind() but have not 
**    been "connected" with sock_connect().
**
**    'backlog' defines the maximum length for the queue of pending 
**    connections. If a connect request arrives when the queue is full, 
**    the client receives an error with ECONNREFUSED.
**    The backlog parameter is limited (silently) to a reasonable value 
**    as determined by the underlying service provider. Illegal values 
**    are replaced by the nearest legal value.
**
**    An application can call sock_listen() more than once on the same 
**    socket. This has the effect of updating the current backlog for 
**    the listening socket. Should there be more pending connections than 
**    the new backlog value, the excess pending connections will be reset 
**    and dropped.
**
** Return
**    sock_listen() returns 0 on success, and -1 on error. The reason
**    for error may be retrieved with sock_get_errno().
*/
int sock_listen(void *s, int backlog)
{
  const struct wso_usrreq_tab FAR *ur = GETTAB_BY_SOCK(s);
  if (ur)
    return (*(ur->so_listen))(((struct sockio_info *)s), backlog);
  return -1;
}

/* ====================================================================== */
/*                            destroy/shutdown                            */
/* ====================================================================== */

/* 
** int sock_destroy( void * s );
**
** Description 
**   Close a socket previously created with sock_create() or
**   with {udp|tcp}_{open|listen}().
**
**   Once a socket is destroyed it cannot be read or written to and can
**   no longer accept connections. sock_destroy() is therefore 
**   approximately analogous to the file I/O function close().
**
**   If 's' refers to an active TCP socket, and is still connected to a
**   peer, the peer will be sent a RST (reset). This is an emergency 
**   condition, and although not recommended for normal transactions, 
**   this service is available and is used by all TCP/IP systems.
**   It is strongly advisable to use sock_shutdown() before using 
**   sock_destroy(). If the socket was not closed gracefully with 
**   sock_shutdown(), or insufficient time has elapsed between 
**   sock_shutdown() and sock_destroy() for the peer to acknowledge 
**   the graceful close, the port used by the socket remains open 
**   until a "quiet time" elapses.
**
** Returns
**   0 - the socket has been sucessfully closed.
**  -1 - 's' refers to an invalid socket.   
*/
int sock_destroy(void *s)
{
  const struct wso_usrreq_tab FAR *ur = GETTAB_BY_SOCK(s);
  if (ur)
    return (*(ur->so_destroy))(((struct sockio_info *)s));
  return -1;
}
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
void sock_abort(void *s) { sock_destroy(s); }
#endif

/* --------------------------------------------------------------------- */

/* 
** int sock_shutdown( void * s, int how );
**
** Description 
**   The sock_shutdown() function ends all or part of a full-duplex 
**   connection on the socket associated with 's'. 
**
**   If 'how' is 0 (receive shutdown), further recv calls are not allowed. 
**   If 'how' is 1 (send shutdown), further send calls are not allowed. 
**   If 'how' is 2 (send and receive shutdown), subsequent send and recv 
**   calls are not allowed. 
**
**   A send shutdown ('how' is 1 or 2) function on a TCP socket 
**   issues a graceful close. The socket can still be used to receive data. 
**   A receive shutdown ('how' is 0) call merely discards incoming data.
**
** Returns
**   If sock_shutdown() succeeds, it returns a value of 0. 
**   It returns -1 if it fails.
*/
int sock_shutdown(void *s, int how)
{
  const struct wso_usrreq_tab FAR *ur = GETTAB_BY_SOCK(s);
  if (ur)
    return (*(ur->so_shutdown))(((struct sockio_info *)s), how);
  return -1;
}
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
void sock_close(void *s) { sock_shutdown(s, 1); }
#endif

/* ====================================================================== */
/*                       {tcp|udp}_open() [wattsock compat]               */
/* ====================================================================== */

/*
** int tcp_listen( tcp_Socket *s, in_port_t lport, in_addr_t raddr, in_port_t rport, 
**                 int (*signal_handler), in_port_t timeout );
** int udp_listen( tcp_Socket *s, in_port_t lport, in_addr_t raddr, in_port_t rport, 
**                 int (*signal_handler) );
**
** Description:
**    These functions tell the transport that an incoming session for a 
**    particular port will be accepted. 
**
**    After a call to {tcp|udp}_listen(), the function sock_established()
**    may be used to poll the socket to detect when a client has connected, 
**    and the socket has thus changed from passive (listening) to active 
**    (connected). This condition may also be detected with 
**    sock_wait_established() and/or sock_condwait() but note that
**    that it is possible for a connection to be opened, written to and
**    closed between two (non-wait) calls to sock_condwait(SOCKESTABLISHED),
**    so it would additionally be necessary to check for SOCKDATAREADY.
**
**    Note: Unlike BSD networking, multiple calls to {tcp|udp}_listen() to 
**    the same local port (lport) are acceptable and constitute the 
**    transport's mechanism for supporting multiple incoming connections to 
**    the same local port. Each time another host attempts to open a session 
**    on that particular port, another one of the listens will be consumed 
**    until such time as all listens have become established sessions at 
**    which time subsequent attempts to connect to the listener will receive 
**    either a reset (TCP) or an ICMP "port unreachable" message (UDP).
**
**    {tcp|udp}_listen() are simple compatibility wrappers based on
**    sock_create().
**
** Parameters 
**  s     - Pointer to a socket. 
**  lport - Port to listen on (the local port number). 
**          Not specifying a valid port number will cause {udp|tcp}_listen()
**          to fail.
**  raddr - Address to accept connections from. 0 to accept from any.
**  rport - Port to accept connections from. 0 to accept from any.
**  signal_handler - Currently not implemented. See {udp|tcp}_open().
**          Specifying anything but NULL will cause {udp|tcp}_listen()
**          to fail.
**  timeout - (TCP only) if timeout is non-zero, the socket will be
**          shut down if a connection is not received in that many seconds.
**
** Return value:
**  Unlike all other sock_xxx functions, {udp|tcp}_listen() return
**  a boolean success code:
**    0: Error - lport was already in use, 
**   !0: Success - The socket is now capable of accepting a connection.
**  The precise reason for a failure cannot be determined.
*/
int sock_wattlisten( int ip_proto, void *s, in_port_t lport, in_addr_t raddr, 
                     in_port_t rport, ProtoHandler handler, 
                     unsigned int timeout, ... )
{
  if (handler)   /* doc meaning is incompatible with what its ... */
    return (0); /* ... actually used for (as far back as I can remember) */

  if (lport == 0)
    return (0); /* failed */

  if (sock_create( s, ip_proto) != 0)
    return (0); /* failed */

  if (sock_opt_set(s, WSO_LF_BLOCKING, 0) == 0)
  {
    const struct wso_usrreq_tab FAR *ur = GETTAB_BY_SOCK(s);
    if (ur)
    {
      if ( (*(ur->so_listen_wcompat))(((struct sockio_info *)s), lport, raddr, rport) == 0)
      {
        if (timeout)
          sock_opt_set( s, WSO_LF_PHASETIMEO, timeout );
        return (1); /* success */
      }
    }
  }

  sock_destroy(s);
  return (0); /* failed */
}

#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int tcp_listen(void *s,in_port_t lp,in_addr_t ra,in_port_t rp,ProtoHandler hdlr,in_port_t timeout)
{ return sock_wattlisten(IPPROTO_TCP, s, lp, ra, rp, hdlr, timeout); }
#endif

#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int udp_listen(void *s, in_port_t lp, in_addr_t ra, in_port_t rp, ProtoHandler hdlr) 
{ return sock_wattlisten(IPPROTO_UDP, s, lp, ra, rp, hdlr, 0); }
#endif

/* --------------------------------------------------------------------- */

/*
** int tcp_open( tcp_Socket *s, in_port_t lport, in_addr_t raddr, in_port_t rport, 
**                                                 int (*signal_handler)());
** int udp_open( udp_Socket *s, in_port_t lport, in_addr_t raddr, in_port_t rport, 
**                                                 int (*signal_handler)());
** 
** Description 
**    These functions actively create a session with another machine,
**    that is, they initiate the establishment of a connection to a peer.
**
**    Sockets initialized with {tcp|udp}_open() are initially 
**    non-blocking. They do not wait for connection establishment to 
**    complete before returning.
**
**    UDP note 1: If the remote host or remote port is set to 0, 
**    udp_open() is functionally identical to udp_listen().
**    UDP note 2: If the remote host is set to -1 (broadcast), all packets 
**    destined for 'lport' (irrespective of destination address) will 
**    be then available with the various read statements. Write 
**    statements to this socket will cause broadcasts. This mechanism 
**    is suitable for broadcast information such as RIP packets.
**
**    {tcp|udp}_open() are simple compatibility wrappers based on
**    sock_create().
**
** Parameters 
**    s     - Pointer to a socket. 
**    lport - Our port, zero for the next available 1025-65536. 
**            Very feq applications will require you to use a particular 
**            local port number. Most network applications let you use 
**            almost any port with a certain set of restrictions. 
**            For example, FINGER or TELNET clients can use any local 
**            port value, so pass the value of zero for lport and let 
**            the transport pick one for you. 
**    raddr - Address to connect to ('R'emote's address).
**    rport - Port to connect to. 
**    signal_handler - This function is called if the connection is 
**           either closed or reset. The parameter for signal_handler is 
**           the pointer to a function which will be called when the 
**           socket is either closed or reset. Some details for 
**           implementation of this service have not been finalized, and 
**           it is recommended the user insert a value of NULL for the 
**           present time. 
**
** Return value 
**  Unlike all other sock_xxx functions, {udp|tcp}_listen() return
**  a boolean success code:
**   !0: if successful, or
**    0: if unable to allocate the requested local port # (its
**       already in use); or unable to resolve the remote computer's 
**       hardware address; or (when opening the very first socket)
**       unable to initialize the stack.
**  The precise reason for a failure cannot be determined.
*/

int sock_wattopen(int ip_proto, void *s, in_port_t lport, in_addr_t raddr, in_port_t rport, 
                  ProtoHandler handler, ... )
{
  if (handler)   /* doc meaning is incompatible with what ... */
    return (0);  /* ... its actually used for */

  if (raddr == 0 || rport == 0)
  {
    if (ip_proto == IPPROTO_UDP) /* compatibility - see docs */
      return sock_wattlisten(ip_proto, s, lport, raddr, rport, handler, 0 );
    return (0); /* tcp_open() doesn't allow the hack. failed */
  }

  if (sock_create( s, ip_proto ) == 0)
  {
    in_addr_t laddr = 0; 
    if (sock_bind( s, &laddr, &lport ) == 0)    
    {
      if (sock_opt_set(s, WSO_LF_BLOCKING, 0) == 0)
      {
        if (sock_connect(s, &raddr, &rport) == 0)
          return (1); /* success */
        if (sock_opt_get(s, WSO_LF_ERRNO) == WSO_EINPROGRESS)
          return (1); /* success */
      }
    }
    sock_destroy(s);
  }
  return (0); /* failed */
}

#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int tcp_open(void *s, in_port_t lp, in_addr_t ra, in_port_t rp, Protohandler handler) 
{ return sock_wattopen(IPPROTO_TCP, s, lp, ra, rp, handler); }
#endif

#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int udp_open(void *s, in_port_t lp, in_addr_t ra, in_port_t rp, Protohandler handler) 
{ return sock_wattopen(IPPROTO_UDP, s, lp, ra, rp, handler); }
#endif

/* ====================================================================== */
/*                         wattsock writers                               */
/* ====================================================================== */

/*
** int sock_send_to( void *s, const void *dp, unsigned int len, int flags,
**                   const in_addr_t *peer_addr, const in_port_t *peer_port );
**
** Description:
**    Writes up to 'len' bytes from 'dp' on socket 's'. 
**
**    This function is a primitive upon which all other socket writers
**    are based.
**
**    - If the WSO_MSG_WAITALL bit in 'flags' is set, the socket is 
**      considered to be blocking irrespective of whether the socket
**      had previously been placed in blocking mode or not.
**
**      If the WSO_MSG_DONTWAIT bit in 'flags' is set, the socket is
**      considered to be non-blocking irrespective of whether the socket
**      had previously been placed in blocking mode or not.
**
**      If neither WSO_MSG_WAITALL nor WSO_MSG_DONTWAIT are set, the 
**      function will block or not-block depending on whether the socket
**      had previously been placed in blocking mode or not.
**
**      If both WSO_MSG_WAITALL and WSO_MSG_DONTWAIT are set, the function
**      returns an error (EINVAL).
**
**      If, according to the above mentioned rules, the function may
**      block, then it will not return until all 'len' bytes can
**      be sent (or it times out, or the read is interrupted with a ^C),
**      or enqueued on the socket for asynchronous transmission.
**      Otherwise, this function sends the maximum amount of data 
**      that can be handled without needing to wait.
**
**      * For UDP, a non-blocking send will cause sock_send_to() to 
**        immediately transmit as much as may be sent without local 
**        fragmentation - MTU-(20+8) octets, where MTU is the Maximum 
**        Transmission Unit as determined at runtime from the packet driver;
**        20 is the IP header size and 8 is the UDP header size. For an
**        ethernet (DIX) packet driver, the maximum is 1500-(20+8) => 1472
**        octets, but it may be as low as 576-(20+8) => 548.
**      * For TCP, the new data is queued for sending and this function
**        returns the number of bytes that will be sent. The data may be 
**        transmitted immediately if enough room is in the buffer, or 
**        sufficient time has expired, or the user has explicitly used 
**        sock_flushnext() to indicate this data should be flushed 
**        immediately.  In either case, no guarantee of acceptance at 
**        the other end is possible. 
**      * An attempt to write to a socket that is still in the process of
**        establishing a connection will block (irrespective of blocking
**        mode or WSO_MSG_* flag) until the connection completes or an error 
**        occurs. If the connection completes the write will proceed
**        normally.
**      * An attempt to write to a socket that is closed, and 'peer_addr'
**        and 'peer_port' are valid, implies a need for a blocking
**        tcp_connect() to be executed on behalf of the application.
**
**    - If the WSO_MSG_EOF is set in 'flags' and all 'len' bytes have been 
**      sent or enqueued successfully, the socket queue will be flushed 
**      and a graceful connection shutdown initiated.
**
**    - If 'dp' is NULL, sock_send_to() performs a flush. The 'len'
**      parameter is then ignored.
**
**    - If 'len' is zero (and 'dp' is not NULL), sock_send_to() causes
**      UDP to send a zero length (probe) packet. TCP ignores a 'len' of 
**      zero.
**
**    For UDP 'peer_addr' and 'peer_port' may be used to request that the
**    data be sent to a specific destination. If 'peer_*' are NULL, and the
**    socket was opened with udp_open() the address of the destination will
**    be inherited from those specified at udp_open() time.
**    For TCP 'peer_addr' and 'peer_port' must either be NULL, or match 
**    those used when the socket was created. 
**
** Return value 
**     The return values are identical to those returned for BSD sendto().
**     Number of bytes written (may be 0 for UDP) or -1 on an error. 
*/
int sock_send_to( void *s, const void *buf, unsigned int buflen, 
                  int flags, const in_addr_t *addr, const in_port_t *port )
{
  const struct wso_usrreq_tab FAR *ur = GETTAB_BY_SOCK(s);
  if (ur)
    return (*(ur->so_send))(((struct sockio_info *)s), buf, buflen, flags, addr, port);
  return -1; 
}

/* --------------------------------------------------------------------- */

/*
** int sock_write( void *s, const void *dp, unsigned int len );
**
** Description:
**    Writes up to len bytes from dp on socket s. This function 
**    busywaits until either the buffer is completely written or 
**    a socket error occurs. 
**
**    This function is a wrapper around sock_send_to() [with the 
**    WSO_MSG_WAITALL bit set].
**  
** Return value 
**     Number of bytes written or -1 on an error. 
*/
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int sock_write( void *s, const void *dp, unsigned int len )
{
  return sock_send_to( s, dp, len, WSO_MSG_WAITALL, NULL, NULL );
}
#endif

/* --------------------------------------------------------------------- */

/*
** int sock_fastwrite( void *s, const void *dp, unsigned int len );
**
** Description 
**    Writes up to 'len' octets from 'dp' on socket s. This function 
**    writes as much as possible to the socket and returns that 
**    number of octets. 
**
**    This function is a wrapper around sock_send_to() [with the 
**    WSO_MSG_DONTWAIT bit set]
**
** Return value 
**     Number of bytes written or -1 on an error. 
*/
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int sock_fastwrite( void *s, const void *dp, unsigned int len )
{
  return sock_send_to( s, dp, len, WSO_MSG_DONTWAIT, NULL, NULL );
}
#endif

/* --------------------------------------------------------------------- */

/*
** int sock_enqueue( void *s, const void *dp, unsigned int len );
**
** Description 
**    This function is obsolete and is simply a wrapper around
**    sock_write().
**
**    In its original incarnation, sock_enqueue() enabled the 
**    caller to provide a block of data to be sent, but the
**    user then could not (re-)use the buffer until the send 
**    had been completed. It also caused transport disruption 
**    if there was already data pending (re-)transmission. It
**    was also not possible to block until the send had been
**    completed, and determining if the send had been completed
**    was not completely possible if data had been previously
**    written with one of the other socket write functions.
**
** Return value 
**     Number of bytes written or -1 on an error. 
** 
*/
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int sock_enqueue( void *s, const void *dp, unsigned int len )
{
  return sock_write( s, dp, len );
}
#endif

/* --------------------------------------------------------------------- */

/*
** int sock_flush( void *s );
**
** Description 
**   sock_flush() will flush the unwritten portion of the TCP buffer 
**   to the network. No guarantee is given that the data was actually 
**   delivered. In the case of a UDP socket, no action is taken. 
**
**   sock_flushnext() is recommended over sock_flush(). 
**
** Returns
**   0 if successful
**  -1 if 's' is an invalid socket. 
*/
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int sock_flush( void *s )
{
  /* warning - this "feature" of sock_send_to() is 
   * undocumented, and subject to change.
  */
  return sock_write( s, NULL, 0 ); /* flush NOW */
}
#endif

/* --------------------------------------------------------------------- */

/*
** int sock_flushnext( void *s );
**
** Description 
**    Writing to TCP sockets does not guarantee that the data is
**    actually transmitted or that the remote computer will pass that 
**    data to the other client in a timely fashion. Using a flush 
**    function will guarantee that the TCP transport places the data 
**    onto the network. No guarantee is made that the remote client will 
**    receive that data. 
**
**    sock_flushnext() is the more efficient of the flush functions. 
**    It causes the next function that sends data to the socket to flush, 
**    meaning the data will be transmitted immediately. 
**
** Returns
**   0 if successful
**  -1 if 's' is an invalid socket. 
*/
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int sock_noflush   (void *s) { return sock_opt_set(s, WSO_LF_HOLDLOCAL, 1); }
int sock_flushnext (void *s) { return sock_opt_set(s, WSO_LF_HOLDLOCAL, 0); }
#endif

/* --------------------------------------------------------------------- */

/*
** int sock_tbsize( void *s );
** int sock_tbused( void *s );
** int sock_tbleft( void *s );
**
** Description 
**    sock_tbsize() determines the size of the transmit buffer for the 
**                  specified socket. 
**                  ** sock_tbleft() returns sock_opt_get(WSO_LF_SNDBUFSZIZE)
**    sock_tbused() Gets the number of bytes in use in the transmit buffer 
**                  for the specified socket. 
**                  ** sock_tbused() returns sock_opt_get(WSO_LF_SNDBUFUSED)
**    sock_tbleft() gets the number of bytes left in the transmit buffer.
**                  ** sock_tbleft() simply returns tbsize - tbused.
**   
**    Note that the values returned reflect raw data. TCP/ASCII 
**    translation (if enabled) cannot be compensated for.
**    Note that the value returned by sock_tbleft() in not necessarily a
**    measure of how much space is available for writing without blocking.
**
** Return
**    -1: 's' is not a valid TCP/UDP socket
**    otherwise the number of bytes remaining/total/used
*/
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int sock_tbused( void *s ) { return sock_opt_get(s, WSO_LF_SNDBUFUSED ); }
int sock_tbsize( void *s ) { return sock_opt_get(s, WSO_LF_SNDBUFSIZE ); }
int sock_tbleft( void *s ) { return sock_tbsize(s) - sock_tbused(s); }
#endif

/*
** int sock_putc( void *s, int ch );
**
** Description:
**    sock_putc() is a compatibility wrapper that places a single 
**    character is placed on the output buffer. In 'c' is '\n', the 
**    buffer is flushed, even if 's' is not in ASCII mode.
**
**    sock_putc() is based on sock_write() and thus may block if 
**    the send buffer is full.
**
** Returns
**    sock_putc() returns 'ch' on success or -1 on error.
**    Refer to the documentation on sock_write() for 
**    reasons why sock_write() may fail.
*/
int sock_putc( void *s, int ch )
{
  char c = (char)ch;
  ch &= 0xff;
  if (c == '\n')
    sock_opt_set(s, WSO_LF_HOLDLOCAL, 0); /* _flushnext */
  if (sock_send_to(s, &c, 1, WSO_MSG_WAITALL,0,0) != 1) /* _write */
    ch = -1;
  return ch;
}

/*
** int sock_puts( void *s, const char *text );
**
** Description
**    sock_puts() is a compatibility wrapper that places strlen(text)
**    characters from 'text' in the output buffer. If the socket is
**    in ASCII mode, it then appends an '\n'. The buffer is then flushed.
**
**    sock_puts() is based on sock_write(), and thus may block if the 
**    output buffer is full. See sock_write() for details.
**
** Returns
**    On success, sock_puts() returns the length of string (not 
**    including the terminating '\0' or the appended '\n' if any).
**    On failure, sock_puts() returns -1.
*/
int sock_puts (void *s, const char *text)
{
  int isascii, rc = 0;
  if (sock_opt_getset(s, WSO_LF_DOASCII, NULL, &isascii ) != 0)
    isascii = 0; /* sock is bad or sock doesn't do ASCII */
  if (text && *text)
  {
    char *p;
    rc = strlen(text);
    p = (char *)malloc(rc + 2);
    if (!p)
    {
      sock_opt_set(s, WSO_LF_ERRNO, WSO_ENOBUFS);
      rc = -1;
    }
    else
    {
      memcpy( p, text, rc );
      if (isascii)
      {
        isascii = 1;
        p[rc] = '\n';
      }
      if (sock_send_to(s, p, rc+isascii, WSO_MSG_WAITALL,0,0) < 1)
        rc = -1;
      free(p);
    }
  }
  else if (isascii)
  {  
    char c = '\r';
    if (sock_send_to(s, &c, 1, WSO_MSG_WAITALL,0,0) < 1)
      rc = -1;
  }
  if (rc == 0)
    sock_send_to(s, NULL, 0, WSO_MSG_WAITALL,0,0); /* flush */
  return rc;  
}

/* ====================================================================== */
/*                         wattsock readers                               */
/* ====================================================================== */

/*
** int sock_recv_from( void *s, void *dp, unsigned int len, int flags,
**                     in_addr_t *peer_addr, in_port_t *peer_port );
**
** Description:
**    Reads up to 'len' bytes from socket 's' into 'dp'. 
**
**    This function is a primitive upon which all other socket readers
**    are based.
**
**    - If the WSO_MSG_PEEK bit in 'flags' is set, this function reads the
**      data but does not remove it from the queue.
**
**    - If the WSO_MSG_WAITALL bit in 'flags' is set, the socket is 
**      considered to be blocking irrespective of whether the socket
**      had previously been placed in blocking mode or not.
**
**      If the WSO_MSG_DONTWAIT bit in 'flags' is set, the socket is
**      considered to be non-blocking irrespective of whether the socket
**      had previously been placed in blocking mode or not.
**
**      If neither WSO_MSG_WAITALL nor WSO_MSG_DONTWAIT are set, the function
**      will block or not-block depending on whether the socket
**      had previously been placed in blocking mode or not.
**
**      If both WSO_MSG_WAITALL and WSO_MSG_DONTWAIT are set, the function
**      returns an error (EINVAL).
**
**      If, according to the above mentioned rules, the function may
**      block, then it will not return until all 'len' bytes can
**      be read (or it times out, or the read is interrupted with a ^C).
**      Otherwise, this function reads the maximum amount of data 
**      that can be handled without needing to wait for more.
**
**    - 'peer_addr' and 'peer_port' will identify the origin of the data.
**      For UDP this may change from recv() to recv(), but all data read
**      during one recv() will always have been from a single host.
**      For TCP 'peer_addr' and 'peer_port' will always be the peer 
**      that the endpoint is connected to.
**
** Return value 
**    The return values are identical to those returned for BSD recvfrom().
**
**    sock_recv_from() returns -1 on timeout or error - either no data 
**    was available and the read did not have WSO_MSG_WAITALL (which 
**    implies a zero timeout); or WSO_MSG_WAITALL was set and the read 
**    timed out; or WSO_MSG_WAITALL was set and the read timed out 
**    prematurely because of an interruption (eg ^C).
**    sock_get_errno() may be used to determine the precise cause.
**    
**    sock_recv_from() returns 0 if a zero-length datagram was read
**    (UDP only); or the connection has been closed (TCP only - zero
**    being an EOF).
**
**    Any other value except 0 or -1 reflects the number of bytes
**    copied to 'dp'. This may be less than 'len' even when WSO_MSG_WAITALL
**    was set or implied due to a timeout or premature timeout due to
**    an interruption (eg ^C).
*/
int sock_recv_from( void *s, void *buf, unsigned int buflen, int flags,
                    in_addr_t *peer_addr, in_port_t *peer_port )
{
  const struct wso_usrreq_tab FAR *ur = GETTAB_BY_SOCK(s);
  if (ur)
    return (*(ur->so_recv))(((struct sockio_info *)s), buf, buflen, flags, peer_addr, peer_port);
  return -1; 
}

/* --------------------------------------------------------------------- */

/*
** int sock_recv( void *s, void *dp, unsigned int len );
** int sock_fastread( void *s, void *dp, unsigned int len );
**
** Description
**      sock_recv() and sock_fastread() are identical functions.
**      This function reads up to len bytes from the socket into the 
**      buffer dp. The bytes are then removed from the socket's buffer.
**      These functions do not block (irrespective of the blocking 
**      mode the socket has been placed in)
**
**      sock_recv()/sock_fastread() are simple compatibility wrappers
**      around sock_recv_from(). Note however that unlike sock_recv_from()
**      these functions return 0 if no data is waiting and are therefore 
**      inherantly unsuitable for reading zero-length UDP datagrams.
**      If encountered, zero length UDP datagrams will be discarded.
**
**      Historical note: In previous versions of this library, 
**      sock_recv() could only be used on sockets that had the recv
**      queue changed, and sock_fastread() could only be used on
**      sockets that had not had the recv queue changed.
**
** Returns
**   the number of bytes read, or -1 on error, or 0 if no 
**   data available for read.
*/
int sock_fastread( void *s, void *dp, unsigned int len )
{
  int rc = 0;
  while (rc == 0)
  {
    in_addr_t addr = 0;
    rc = sock_recv_from( s, dp, len, WSO_MSG_DONTWAIT, &addr, NULL );
    if (rc == -1) /* no data available or error */
    {
      /* DBGPRINTF(("sock_recv_from() returned -1\n")); */
      if (sock_opt_get(s, WSO_LF_FIONREAD) < 1)
        rc = 0; /* no data available */
      break;
    }
    if (rc == 0) /* sock closed or zero-length UDP dgram */
    {
      /* DBGPRINTF(("sock_recv_from() returned 0\n")); */
      if (addr == 0) /* sock closed */
        rc = -1;
      /* else zero-length datagram - loop */
    }
  }
  /* DBGPRINTF(("*sock_fastread() => %d\n",rc)); */
  return rc;
}

#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int sock_recv( void *s, void *dp, unsigned int len )
{
  return sock_fastread( s, dp, len );
}
#endif

/* --------------------------------------------------------------------- */

/*
** int sock_preread( void *s, void *dp, unsigned int len );
**
** Description 
**     This function reads up to len bytes from the socket into the 
**     buffer dp. The bytes are not removed from the socket's buffer.
**     The function does not busywait.
**
**     This function is a simple compatibility wrapper for
**     sock_recv_from( WSO_MSG_PEEK ) but unlike sock_recv_from() it 
**     returns 0 if no data is waiting and is therefore inherantly
**     unsuitable for reading zero-length UDP datagrams.
**     If encountered, zero length datagrams will be discarded.
**
** Return value 
**   the number of bytes read, or -1 on error, or 0 if no 
**   data available for read.
*/
int sock_preread( void *s, void *dp, unsigned int len )
{
  int rc = sock_opt_get( s, WSO_LF_FIONREAD );
  if (rc != -1 /* -1 == bad sock */ && rc != 0 /* 0 == no data */ )
    rc = sock_recv_from(s, dp, len, WSO_MSG_DONTWAIT|WSO_MSG_PEEK, NULL, NULL );
  return rc;
}

/* --------------------------------------------------------------------- */

/*
** int sock_read( void *s, void *dp, unsigned int len );
**
** Description 
**    sock_read() will busywait until len bytes are read or a 
**    socket error exists. 
**
**    This function is a simple compatibility wrapper for
**    sock_recv_from( WSO_MSG_WAITALL ) but unlike sock_recv_from() it 
**    is not permitted to return 0 for zero length data, and is
**    therefore inherantly unsuitable for reading zero-length UDP 
**    datagrams. If encountered, zero length datagrams will be discarded.
**
** Returns
**   the number of bytes read (may be zero), or -1 on error.
*/
int sock_read( void *s, void *dp, unsigned int len )
{
  int rc = 0;
  while (rc == 0)
  {
    in_addr_t addr = 0;
    rc = sock_recv_from( s, dp, len, WSO_MSG_WAITALL, &addr, NULL );
    /* -1 == no data available on a blocking read - can only mean EINTR etc */
    if (rc == 0) /* may be conn closed or zero length datagram */
    {
      if (addr == 0) /* no peer, so its conn closed */
        rc = -1;
      /* else its a zero length datagram, loop (discard and wait for next) */
    }
  }
  return rc;
}

/* --------------------------------------------------------------------- */

/*
** int sock_getc( void *s );
** 
** Description 
**    Gets the next character from the socket. NOTE: This function blocks. 
**
**    This function is a simple compatibility wrapper around 
**      sock_recv_from( s, &c, 1, WSO_MSG_WAITALL, NULL, NULL )
**
** Return value 
**    Character read or -1 if error. 
*/
int sock_getc( void *s )
{
  char c;
  if (sock_recv_from( s, &c, 1, WSO_MSG_WAITALL, NULL, NULL ) != 1)
    return -1;
  return ((int)c) & 0xff;
}

/* --------------------------------------------------------------------- */

/*
** int sock_gets( void *s, char *text, unsigned int len );
**
** Description 
**   Reads a string from a socket and replaces the linefeed with a '\0'. 
**   If the string is longer than len, the string is null terminated and 
**   the remaining characters in the string are discarded.
**
**   This function is a simple compatibility wrapper around 
**      sock_bytesready() and sock_recv_from()
**     
**   This function is only meaningful for TCP sockets that had previously
**   been placed in ASCII mode.
**   Note that ASCII mode is deprecated and may not be available in future
**   versions of the library.
**
** Returns
**   0 if buffer is empty, or if no '\r' or '\n' is present in the buffer.
**  !0 is the length of the string. 
*/
int sock_gets( void *s, char *text, unsigned int len )
{
  int rc = 0;
  if (len != 0)
  {
    int llen = sock_opt_get(s, WSO_LF_TXTLINELEN);
    if (llen != -1) /* have a line (may be zero length) */
    {
      if (((int)len) > llen)
      {
        /* read llen + '\n' */ 
        rc = sock_recv_from( s, text, llen+1, 0, NULL, NULL );
      }
      else
      { 
        /* read as much as we can */
        rc = sock_recv_from( s, text, len-1, 0, NULL, NULL );
        /* discard the rest of the line (including '\n') */
        sock_recv_from( s, NULL, (llen+1)-(len-1), 0, NULL, NULL );
      }
      if (rc != -1)
        text[rc] = '\0';
    }
  }
  return rc;
}

/* --------------------------------------------------------------------- */

/*
** int sock_recv_init( void *s, void *space, unsigned int len);
**
** Description
**     sock_recv_init() is no longer meaningful and does nothing.
**
**     In previous incarnations of this library, sock_recv_init() 
**     allowed the user to assign a larger receive buffer than 
**     the default 2K buffer.
**
**     The current architecture permits a new MAXIMUM receive buffer
**     SIZE to be assigned to the socket using sock_opt_get().
**     The queue itself is allocated by the protocol layer, and 
**     free'd when the socket is destroyed.
**
** Returns
**    sock_recv_init() always returns -1 (ENOSYS)
*/
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int sock_recv_init(void *s, void *rx_buf, unsigned int rx_len)
{
  rx_buf = rx_buf; /* unused */
  rx_len = rx_len; /* unused */
  s = s; /* possibly unused */
  #if defined(ENOSYS)
  sock_opt_set(s, WSO_LF_ERRNO, ENOSYS );
  #endif
  return -1;
}
#endif

/* --------------------------------------------------------------------- */

/*
** int sock_rbleft( void *s );
** int sock_rbsize( void *s );
** int sock_rbused( void *s );
**
** Description 
**    sock_rbused() Gets the number of bytes in use in the receive buffer.
**                  ** sock_rbused() returns sock_opt_get(WSO_LF_RCVBUFUSED)
**    sock_rbsize() determines the size of the receive buffer. 
**                  ** sock_rbsize() returns sock_opt_get(WSO_LF_RCVBUFSIZE)
**    sock_rbleft() determines the amount of space available 
**                  in the receive buffer. 
**                  ** sock_rbleft() simply returns rbsize - rbused;
**
**    WARNING!! USE AT YOUR OWN RISK. THESE FUNCTIONS HAVE A LOT OF SCOPE 
**    FOR FALSE INTERPRETATION. USE AT YOUR OWN RISK.
**
**    The values returned in no way reflect the true amount of 
**    space/data that remains/is waiting in the buffer.
**
**    - TCP data may be in the buffer but non-contiguous - missing 
**      segments and all data thereafter would be 'invisible' to 
**      sock_dataready() and the various read functions, but not to 
**      sock_rbused(). 
**    - TCP out-of-band data may be visible to sock_dataready()/recv()
**      but will be invisible to sock_rbused()
**    - TCP/ASCII: The values returned by sock_rbleft() and 
**      sock_rbsize() assume raw data. Translation (if enabled) is
**      not compensated for.
**    - UDP receive buffer data may include data from multiple peers.
**      sock_rbxxx() will see all data, but in no way can be read
**      in one call.
**    - UDP zero length UDP packets are seen with more than zero length, 
**      but are invisible to the read functions and sock_dataready().
**    - UDP receive queue data contains internal management structures
**      for separating the individual packets from the the packets 
**      themselves, as well as maintaining the origin of the data. This
**      is all visible to sock_rbxxx() but will never be readable.
**
** Return
**    -1: 's' is not a valid TCP/UDP socket
**    otherwise the number of bytes remaining/total/used
*/
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int sock_rbsize( void *s ) { return sock_opt_get(s, WSO_LF_RCVBUFSIZE ); }
int sock_rbused( void *s ) { return sock_opt_get(s, WSO_LF_RCVBUFUSED ); }
int sock_rbleft( void *s ) { return sock_rbsize(s) - sock_rbused(s); }
#endif

/* --------------------------------------------------------------------- */

/* 
** int sock_dataready( void *s );
**
** Description 
**    Gets the number of bytes waiting to be read. If in TCP/ASCII mode, 
**    it returns zero if a newline character has not been read. See 
**    sock_bytesready() for a more general function. 
**
**   Note that ASCII mode is deprecated and may not be available in future
**   versions of the library.
**
** Returns 
**  0: No bytes to read;
** !0: Number of bytes ready to read. 
*/
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int sock_dataready( void *s ) { return sock_opt_get(s, WSO_LF_FIONREAD); }
#endif

/* --------------------------------------------------------------------- */

/* 
** int sock_bytesready( void *s );
**
** Description 
**   This function obtains the amount of data available for reading.
**   This function differs slightly from sock_dataready() in what
**   it returns if no data is waiting.
**
** Returns
**   If the socket is not a TCP socket, or is a TCP socket in binary mode, 
**   sock_bytesready returns:
**    -1 : there is no data waiting
**    !0 : the amount of data waiting to be read. 
**
**   In TCP/ASCII mode, sock_bytesready returns:
**     -1 :- there are no bytes waiting to be read or 
**         - the line that is waiting is incomplete (no line terminating 
**           character has been read.) 
**      0 :- the line waiting to be read is of zero length.
**   else :- either the length of the line waiting to be read, or
**         - there is no complete line in the buffer but the buffer is full,
**           or no more data will ever be available (connection closed).
**
**   Note that ASCII mode is deprecated and may not be available in future
**   versions of the library.
*/
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int sock_bytesready( void *s ) { return sock_opt_get(s, WSO_LF_TXTLINELEN); }
#endif

/* ====================================================================== */
/*                             socket options                             */
/* ====================================================================== */

/*
** int sock_opt_getset( void *s, unsigned long opt, 
**                      const int *newval, int *oldval );
**
** Description
**     The sock_opt_getset() function reads or manipulates options 
**     associated with a socket. These options are defined as arguments 
**     for these functions. Options exist at either the socket level or the 
**     underlying protocol level.
**
**     'newval' is the new value to be assigned to the option. The minimum
**     size of 'newval' is an 'int', but some options accept values other
**     that 'int' in which case, the pointer needs to be cast to an int 
**     pointer. Boolean on/off values are interpreted as *newval != 0 for
**     on/true, or *newval == 0 for off/false.
**
**     When setting values the underlying transport may adjust the 
**     value to internal limits. For example, the TTL must be between 0
**     and 255 (inclusive), and other values will be adjusted up or down
**     as needed.
**     
**     A complete list of available options is in one of the include headers.
**     Here are some commonly used ones: (add WSO_LF_ prefix)
**
**     NONAGLE   int/bool TCP only  enable/disable send coalescing.
**     REUSEADDR int/bool TCP/UDP   allow/disallow port sharing.
**     KEEPALIVE  minutes TCP only  keep alive interval (0 = disable)
**     FIONREAD    uint   TCP/UDP   how much data is available for read
**     FIONWRITE   uint   TCP/UDP   how much can be written without blocking
**     TOS         int    TCP only  get/set Terms of Service (0...255)
**     TTL         int    TCP/UDP   get/set Time-To-Live/maxhops (0...255)
**     SNDTIMEO    secs   TCP/UDP   get/set transmit timeout (0=infinite)
**     RCVTIMEO    secs   TCP/UDP   get/set receive  timeout (0=infinite)
**     IPPROTO     int    TCP/UDP   returns 6 for a TCP socket, 17 for UDP.
**     ERRNO       int    TCP/UDP   BSDish errno for the last failed call.
**     STATE       int    TCP/UDP   == sock_get_stateno()
**     RCVBUFSIZE  uint   TCP/UDP   get/set size of receive buffer
**     SNFBUFSIZE  uint   TCP/UDP   get size of send buffer
**     xxxBUFUSED  uint   TCP/UDP   get amount of buffer used.
**     HOLDLOCAL int/bool TCP only  1=no flush, 0=flush next
**     DOASCII   int/bool TCP only  enable/disable ASCII mode
**     TXTLINELEN  uint   TCP only  == sock_bytesready()
**     PEERADDR   uint32  TCP/UDP   \_ == sock_getpeername()
**     PEERPORT   uint16  TCP/UDP   /
**     BOUNDADDR  uint32  TCP/UDP   \_ == sock_getsockname()
**     BOUNDPORT  uint16  TCP/UDP   /
**     LINGERTIMEO tsecs  TCP only  get/set linger timeout (0=no linger)
**
**   If 's' is a BSD-compatible listening socket, all new connections
**   accept()ed on 's' will inherit options set on 's'.
**
**   Options are interpreted as reservations (that is, they are assumed
**   to have an effect only at some point in the future), and the 
**   current state of the socket is generally ignored.
**   Thus, sock_opt_getset() always returns success (0) if the socket 
**   and option are valid. Whether the value retrieved is meaningful 
**   or not or whether a new value could be assigned "successfully"
**   or not must be determined by other means and in light of what
**   is to be accomplished. 
**   For example, 'oldval' will have a value of zero when using 
**   sock_opt_getset(WSO_LF_PEERADDR) on a socket that isn't connected.
**   Similarly, assigning a new value will never be seen to fail 
**   since values are always negotiated even when they are blatantly 
**   out of bounds.
**
** Returns
**     0 if successful.
**    -1 if failed (unrecognized option or bad socket).
*/
int sock_opt_getset(void *s, unsigned long opt, const int *newval, int *oldval)
{
  const struct wso_usrreq_tab FAR *ur = GETTAB_BY_SOCK(s);
  if (ur)
    return (*(ur->so_gssockopt))(((struct sockio_info *)s), opt, newval, oldval);
  return -1;
}

/* --------------------------------------------------------------------- */

/* 
** int sock_opt_get(void *s, unsigned long opt);
** int sock_opt_set(void *s, unsigned long opt, int val);
** 
** Description
**    sock_opt_get() and sock_opt_set() are convenience wrappers around
**    sock_opt_getset().
**
**    sock_opt_get() and sock_opt_set() and are only capable of dealing 
**    with values of sizeof('int'), which is what almost all options are.
**
**    Note also that that a value retrieved with sock_opt_get() may be -1, 
**    which makes such a value indistinguishable from an error return. 
**    sock_opt_get() is therefore inherantly unsuitable for retrieving 
**    options where a -1 return value is possible.
**
** Returns
**    sock_opt_set() returns zero on success, or -1 on error.
**    sock_opt_get() returns the current value of the option 'opt',
**    or -1 on error.
*/
int sock_opt_get(void *s, unsigned long opt) 
{
  int val;
  if (sock_opt_getset(s, opt, NULL, &val ) == 0)
    return val;
  return -1;
}

int sock_opt_set(void *s, unsigned long opt, int val)
{
  return sock_opt_getset(s, opt, &val, NULL );
}

/* --------------------------------------------------------------------- */

#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
void udp_SetTTL (void *s, unsigned char ttl) 
{ 
  sock_opt_set( s, WSO_LF_TTL, ttl ); 
}
#endif

/* --------------------------------------------------------------------- */

/* used by sock_strerror() and sock_strstate() to convert unrecognized 
** codes to names. eg, an unrecognized errno is returned as "Error [-]nnnn", 
** and an unrecognized state is returned as "Unknown state [-]nnnn"
*/
static const char *___num2str( const char *prefix, int num )
{
  static char buf[sizeof("Unknown state -")+(sizeof(int)*3)];
  unsigned int i = sizeof(buf);

  buf[--i] = '\0';
  if (num == 0)
    buf[--i] = '0';
  else
  {
    int neg = 0;
    if (num < 0)
    {
      neg = 1;
      num = -num;
    } 
    while (num)
    {
      buf[--i] = (char)('0' + (num % 10));
      num /= 10;
    }
    if (neg)
      buf[--i] = '-';
  }
  if (prefix)
  {
    unsigned int r = strlen(prefix);
    buf[--i] = ' ';
    if (r > i)
      r = i;
    i -= r;
    memcpy( &buf[i], prefix, r );
  }
  return &buf[i];
}

/* --------------------------------------------------------------------- */

/*
** int sock_get_errno( void *s );
** const char *sock_strerror( int so_errno );
** const char *sockerr( void * s );
**
** Description 
**    sock_get_errno() retrieves the error number for the last error that 
**                     was recorded on a socket (sock_opt_get(WSO_LF_ERRNO))
**    sock_strerror()  maps an error number retrieved with sock_get_errno()
**                     to an ASCII label therof.
**    sockerr()        is compatibility wrapper that returns 
**                     sock_strerror(sock_get_errno(s))
**
** Returns 
**    error number or pointer to error label as appropriate.
*/ 
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int sock_get_errno( void *s )
{
  return sock_opt_get( s, WSO_LF_ERRNO );
}
#endif

const char *sock_strerror( int so_errno )
{
  if (so_errno == -1)
    return "-1/Invalid socket";
  #if ((UINT_MAX > USHRT_MAX) || (!defined(__TINY__) && !defined(__SMALL__)))
  if (so_errno > 0 && so_errno < (*__get_sys_nerr_ptr()) )
    return strerror(so_errno);
  switch (so_errno)  
  {
    #define SOERRNO2STR(__x) case __x: return #__x
    case 0: return "error 0";
    SOERRNO2STR(WSO_EINTR        );
    SOERRNO2STR(WSO_EACCES       );
    SOERRNO2STR(WSO_EINVAL       );
    SOERRNO2STR(WSO_EWOULDBLOCK  );
    SOERRNO2STR(WSO_EINPROGRESS  );
    SOERRNO2STR(WSO_EALREADY     );
    SOERRNO2STR(WSO_EDESTADDRREQ );
    SOERRNO2STR(WSO_EMSGSIZE     );
    SOERRNO2STR(WSO_ENOPROTOOPT  );
    SOERRNO2STR(WSO_EOPNOTSUPP   );
    SOERRNO2STR(WSO_EADDRINUSE   );
    SOERRNO2STR(WSO_EADDRNOTAVAIL);
    SOERRNO2STR(WSO_ENETDOWN     );
    SOERRNO2STR(WSO_ECONNABORTED );
    SOERRNO2STR(WSO_ECONNRESET   );
    SOERRNO2STR(WSO_ENOBUFS      );
    SOERRNO2STR(WSO_ENOTCONN     );
    SOERRNO2STR(WSO_ESHUTDOWN    );
    SOERRNO2STR(WSO_ETIMEDOUT    );
    SOERRNO2STR(WSO_ECONNREFUSED );
    SOERRNO2STR(WSO_EHOSTUNREACH );
    SOERRNO2STR(WSO_EREMOTE      );
    default:
        break;
  }
  #endif
  return ___num2str( "Error", so_errno );
}

#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
const char *sockerr( void * s )
{
  return sock_strerror(sock_get_errno(s));
}
#endif

/* --------------------------------------------------------------------- */

/*
** int sock_get_stateno( void *s );
** const char *sock_strstate( int so_stateno );
** const char *sockstate( void * s );
**
** Description 
**    sock_get_stateno() returns a socket's current state.
**                       sock_get_stateno() returns sock_opt_get(WSO_LF_STATE)
**    sock_strstate()    maps a state number to a state 
**                       name and returns a pointer to that name.
**    sockstate()        is a compatibility wrapper that returns
**                       sock_strstate(sock_get_stateno(s))
**
** Returns 
**   An integer representing the current state of a socket, or
**   an ASCII label representing a state number.
**
**   -1: "Not an active socket" : 's' is not a valid TCP/UDP socket.
**    0: "Listen"     : indicates a passively opened socket that is 
**                      waiting for a connection. 
**    1: "SynSent"    : connection phase intermediate state.       (TCP only)
**    2: "SynRcvd"    : connection phase intermediate state.       (TCP only)
**    3: "Established": states that the connection is complete. 
**    4: "EstClosing" : connection termination intermediate state. (TCP only)
**    5: "FinWait1"   : connection termination intermediate state. (TCP only)
**    6: "FinWait2"   : connection termination intermediate state. (TCP only)
**    7: "CloseWait"  : connection termination intermediate state. (TCP only)
**    8: "Closing"    : connection termination intermediate state. (TCP only)
**    9: "LastAck"    : connection termination intermediate state. (TCP only)
**   10: "TimeWait"   : connection termination intermediate state. (TCP only)
**   --: "CloseMSL"   : connection termination intermediate state. (TCP only)
**   11: "Closed"     : indicates that the connection is completely closed. 
** 
**   All other values return "Unknown state"
*/
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int sock_get_stateno( void *s )
{
  return sock_opt_get(s, WSO_LF_STATE);
}
#endif

const char *sock_strstate(int so_state)
{
  #if ((UINT_MAX > USHRT_MAX) || (!defined(__TINY__) && !defined(__SMALL__)))
  static const char *state_names[] = 
  {
    "LISTEN",   /*  0  - listening for connection */
    "SYNSENT",  /*  1  - SYN sent, active open */
    "SYNREC",   /*  2  - SYN received, ACK+SYN sent. */
    "ESTAB",    /*  3  - established */
    "ESTCLOSE", /*  4  - established, but will FIN */
    "FINWT1",   /*  5  - sent FIN */
    "FINWT2",   /*  6  - sent FIN, received FINACK */
    "CLOSWT",   /*  7  - received FIN waiting for close */
    "CLOSING",  /*  8  - sent FIN, received FIN (waiting for FINACK) */
    "LASTACK",  /*  9  - FIN received, FINACK+FIN sent */
    "TIMEWT",   /* 10  - dally after sending final FINACK */
    "CLOSED"    /* 11  - FINACK received */
  };
  if (so_state < 0)
    return "Not an active socket";
  if (((unsigned int)so_state) < (sizeof(state_names)/sizeof(state_names[0])) )
    return state_names[so_state];
  return ___num2str( "Unknown state", so_state );
  #else
  return ___num2str( "State", so_state );
  #endif
}

#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
const char *sockstate( void * s ) 
{ 
  return sock_strstate(sock_get_stateno(s)); 
}
#endif

/* --------------------------------------------------------------------- */

/*
** void sock_mode( void *s, int mode );
**
** Description 
**
**   ** NOTE ** sock_mode() is maintained for compatibility purposes only.
**   It does not work correctly (and never did) in as far as reverting to
**   the original mode is concerned - Due to #define overlaps (all the
**   default states were/are defined to be zero) sock_mode() cannot revert
**   to the default state since it cannot determine which default 
**   state is to be reverted to. Use sock_opt_getset() instead. 
**
**   ** NOTE** If sock_mode() is called with 'mode' equals zero, ie the
**   mode is to be restored to whatever it was "before", the library 
**   has no choice but to exit() with an error message since it has 
**   no means of notifying the caller of the exception.
**   
**   The original documentation is as follows:
**
**   This function changes some of the basic handling of a socket. The 
**   following macros can be passed as the 2nd parameter (OR'd together 
**   if necessary): 
**
**   TCP_MODE_ASCII 
**   TCP sockets are usually in binary mode which means an arbitrary 
**   stream of bytes is allowed (TCP is treated as a byte stream) 
**   The default is TCP_MODE_BINARY. By changing the mode to 
**   TCP_MODE_ASCII, all read and write functions will automatically
**   translate outgoing "\n"s to "\r\n", and incoming "\r\n"s to "\n".
**   ASCII mode is only available for TCP sockets and is DEPRECATED.
**   If ASCII mode is enabled, and urgent ("Out-Of-Band") or other 
**   non-ASCII data is received on the socket, the connection will 
**   be terminated.
**
**   TCP_MODE_NONAGLE 
**   The Nagle algorithm may substantially reduce network traffic with 
**   little negative effect on a user (In most situations, the Nagle 
**   algorithm improves application performance.) The default is 
**   TCP_MODE_NAGLE. This mode only affects TCP connections. If you are 
**   doing X-Windows or real time data collection, you may switch the 
**   Nagle algorithm off by selecting the TCP_MODE_NONAGLE flag. 
**
**   UDP_MODE_NOCHK 
**   Checksums are required for TCP, but not for UDP. The default is 
**   UDP_MODE_CHK. If you are providing a checksum at a higher level, the 
**   low level checksum may be redundant. The checksum for UDP can be 
**   disabled by selecting the UDP_MODE_NOCHK flag. Note that you cannot 
**   control whether the remote computer will send checksums. If that 
**   computer does checksum its outbound data, the UDP transport will check 
**   the received packet's checksum. 
**
** Returns
**   nothing.
*/
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
void sock_mode( void *s, int mode )
{
  if (mode == 0)
  {
    _printk("PANIC: sock_mode() called with mode == 0\n");
    exit(255);
  }
  if (mode & TCP_MODE_ASCII)        /* TCP_MODE_ASCII==1, _BINARY=0!!! */
    sock_opt_set(s, WSO_LF_DOASCII, 1);
  if (mode & TCP_MODE_NONAGLE)      /* TCP_MODE_NONAGLE==4, _NAGLE=0!!! */
    sock_opt_set(s, WSO_LF_NONAGLE, 1);
  if (mode & UDP_MODE_NOCHK)        /* TCP_MODE_NOCHK==2, _CHK=0!!! */
    sock_opt_set(s, WSO_LF_NOCHECKSUM, 1);
}
#endif

/* --------------------------------------------------------------------- */

/*
** int sock_set_yieldcb( void *s, void (*fn)(void) );
**
** Description 
**   This function allows the association of a yield callback
**   that is called repeatedly by socket primitives when a blocking 
**   function is in progress.
**
**   If 's' is NULL, 'fn' is saved and applied to every subsequently
**   created socket. Existing sockets are not updated. 
**   If 's' is not NULL, 'fn' is associated with 's'.
**
**   If 's' is a BSD-compatible listening socket, all new connections
**   accept()ed on 's' will inherit 'fn'.
**
**   x86-16bit note: If the library was compiled for one of the x86 
**   mixed-memory models (ie, pointers to code and pointers to data 
**   have different sizes), sock_set_yieldcb() will appear to succeed,
**   but the yield function will never be called.
**
** Returns
**  -1 if 's' is not NULL but is not a valid TCP/UDP socket.
**   0 if assignment was successful. 
*/
int sock_set_yieldcb( void *s, void (*fn)(void) )
{
  if (s)
    return sock_opt_getset(s, WSO_LF_YIELDFUNC, (const int *)(fn), NULL );
  so_defaults.so_yield = fn;
  return 0;
}
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
void sock_yield( void *s, void (*fn)()) { sock_set_yieldcb( s, fn ); }
#endif

/* --------------------------------------------------------------------- */

/*
** int sock_getpeername( void *s, in_addr_t *raddr, in_port_t *rport );
** int sock_getsockname( void *s, in_addr_t *laddr, in_port_t *lport );
** 
** Description 
**    sock_getpeername() gets the peer's IP address and port information 
**                       for the connection open on socket 's'. 
**    sock_getsockname() gets the locally bound IP address and port 
**                       information for the specified socket 's'. 
**
**    If the socket is not in a state appropriate for 
**    the function (for example, using sock_getpeername() on an
**    unconnected socket, or sock_getsockname() on an unbound
**    socket), then the function will still succeed, but 
**    the returned address and port will both be zero.
**
**    Note that sock_getsockname() may return zero in 'laddr' if
**    the socket is bound to all local addresses, but will never
**    return zero in 'lport' (once the socket has been bound).
**
** Returns
**  -1: failure: 's' is not a valid UDP/TCP socket.
**   0: success; 
*/
static int __get_peer_sock_name(void *s, 
                                in_addr_t *addr, unsigned long lf_addropt,
                                in_port_t *port, unsigned long lf_portopt )
{
  in_addr_t tmp_addr;
  in_port_t  tmp_port;

  if (sock_opt_getset( s, lf_portopt, NULL, (int *)&tmp_port ) != 0)
    return -1; /* EBADF */
  if (sock_opt_getset( s, lf_addropt, NULL, (int *)&tmp_addr ) != 0)
    return -1; /* hmmmm! */
  if (addr)
    *addr = tmp_addr;
  if (port)
    *port = tmp_port;

  return 0;
}

int sock_getpeername( void *s, in_addr_t *raddr, in_port_t *rport ) 
{
  return __get_peer_sock_name(s, raddr, WSO_LF_PEERADDR, rport, WSO_LF_PEERPORT );
}

int sock_getsockname( void *s, in_addr_t *laddr, in_port_t *lport ) 
{
  return __get_peer_sock_name(s, laddr, WSO_LF_BOUNDADDR, lport, WSO_LF_BOUNDPORT );
}

/* --------------------------------------------------------------------- */

/*
** void ip_timer_init(void *s, int seconds);
** int  ip_timer_expired(void *s);
**
** Description
**    ip_timer_init() and ip_timer_expired() are functions
**    that make a per-socket timer available to user applications.
**    
**    What the application chooses to initialize the timer with
**    or chooses to do with the timer, is left to the discretion
**    of the application developer. The socket library does not
**    itself use this timer.
**
**    ip_timer_init() initializes this per-socket timer.
**    if 'seconds' is zero, the timer is stopped.
**
**    ip_timer_expired() tests if this per-socket timer previously
**    initialized with ip_timer_init() has expired. 
**    The timer is not automatically restarted on expiry.
**
**    It might have been more interesting to be able to associate
**    an arbitrary user-defined value on a per-socket basis. But
**    hindsight is of course always 20:20.
**
** Return:
**    ip_timer_expired() returns 1 if the user timer previously
**    initialized with ip_timer_init() has expired, or returns 0 if
**    the timer has yet to expire, or -1 if 's' is invalid.
**
**    ip_timer_init() returns nothing.
*/
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
void ip_timer_init(void *s, int seconds) 
{ 
  sock_opt_set(s, WSO_LF_USERTIMER, seconds );  
}
#endif

#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int ip_timer_expired( void *s )
{
  return sock_opt_get(s, WSO_LF_USERTIMER);
}
#endif

/* --------------------------------------------------------------------- */

#if 0 /* unused */
/* Keep the socket open inspite of receiving "ICMP Unreachable" */
void sock_sturdy (sock_type *s, int level)
{
  s->tcp.rigid = level;
  if (s->tcp.rigid < s->tcp.stress)
     sock_abort (s);
}
#endif

/* ====================================================================== */
/*                             condition wait/select                      */
/* ====================================================================== */

/*
** int sock_poll( void *s, int events, unsigned long timeout );
**
** Description 
**    sock_poll() is the cornerstone for virtually all socket operations 
**    in this library.
**
**    If 's' is NULL, the 'events' parameter is ignored and sock_poll()
**    simply performs background chores and/or hardware/DPMI yields
**    until 'timeout' elapses ('timeout' is discussed in detail below).
**
**    If 's' is not NULL, the 'events' parameter is interpreted as 
**    a bitmask of events to check/wait for. It may be constructed by 
**    OR'ing any of the following constants:
**
**    Manifest constant           Wait for ... / Check if .... 
**    -----------------    ------------------------------------------
**    WSO_ASYSEL_READ	   normal data is available for read.
**    WSO_ASYSEL_WRITE	   there is room to write without blocking.
**    WSO_ASYSEL_RDOOB     higher priority data is available for read.
**    WSO_ASYSEL_WROOB     higher priority data may be written non-blocking.
**    WSO_ASYSEL_QOS       QoS associated with socket 's'has changed.
**    WSO_ASYSEL_GRPQOS    QoS associated with all (IP) sockets has changed.
**    WSO_ASYSEL_ACCEPT    a new connection may be accepted with sock_accept()
**    WSO_ASYSEL_CONNECT   a connection has been established completely
**    WSO_ASYSEL_HANGUP    a connection has been closed/reset by the peer
**    WSO_ASYSEL_ERROR     an exceptional condition has occured that has
**                         or will soon lead to a disconnect.
**    WSO_ASYSEL_EBADF     's' is not a valid socket.
**
**    - WSO_ASYSEL_HANGUP, WSO_ASYSEL_ERROR and WSO_ASYSEL_EBADF are always 
**      checked, even if not present in the events bitmask. 
**    - WSO_ASYSEL_HANGUP and WSO_ASYSEL_WR* are mutually exclusive and
**      will never be present in the result at the same time.
**    - Unrecognized WSO_* flags are ignored, so it is acceptable to
**      call sock_poll() with 'event_mask' set to -1 (0xffff...), if 
**      you wish to check for all events.
**    - WSO_ASYSSEL_CONNECT is a "special" event - unlike all the other
**      events that may come and go, WSO_ASYSEL_CONNECT can only be 
**      seen once - the event is consumed once it has been seen.
**    - a QoS (Quality of Service) change can be expressed as a change
**      to any combination of network-imposed delay (eg TCP stress change), 
**      jitter (eg RTT change), bandwidth (eg MSS change), and reliability
**      (eg out-of-order data, congestion, packet loss etc).
**    - Only connection-oriented transports (ie, TCP) are capable of 
**      detecting _ACCEPT, _CONNECT and _HANGUP.
**    - Only TCP has the ability to read/write higher-priority data.
**      Support for this is currently unimplemented in this library.
**
**    If 'timeout' is zero, sock_poll() returns without blocking.
**    If 'timeout' is ((unsigned long)-1), sock_poll() blocks 
**    indefinitely.
**    If 'timeout' is neither zero nor -1, it specifies a maximum 
**    interval to wait in milliseconds for any of the above mentioned 
**    events to occur. 
**
**    While sock_poll() is waiting it calls the yield callback that may
**    have previously been assigned to the socket with sock_set_yieldcb().
**
**    sock_poll() may be used to implement higher level 4BSD-compatible
**    select() or SysV/SUS/TLI/XOpen-compatible poll() as follows:
**    
**    sock_poll() flag:    4BSD select()       Unix poll()
**    -----------------    -----------------   -----------------------
**    WSO_ASYSEL_READ	   read                POLLRDNORM (POLLIN)
**    WSO_ASYSEL_WRITE	        write          POLLWRNORM (POLLOUT)
**    WSO_ASYSEL_RDOOB     read       except   POLLRDBAND (POLLPRI)
**    WSO_ASYSEL_WROOB          write          POLLWRBAND
**    WSO_ASYSEL_QOS       --                  --
**    WSO_ASYSEL_GRPQOS    --                  --
**    WSO_ASYSEL_ACCEPT    read                POLLIN
**    WSO_ASYSEL_CONNECT        write          POLLOUT
**    WSO_ASYSEL_HANGUP    read                POLLHUP
**    WSO_ASYSEL_ERROR                except   POLLERR
**    WSO_ASYSEL_EBADF     return -1           POLLNVAL
**                         
** Returns
**    sock_poll() returns a bitmask of the events that occurred.
**    If none of the events occurred and the time limit expires,
**    sock_poll() returns zero.
**
**    If sock_poll() could not complete the request, it returns
**    -1 and sets 'errno'. The only possible reasons of failure
**    are 
**        EINTR - sock_poll() was interrupted by a signal.
**        EINVAL - 's' was NULL and 'timeout' for infinite.
**
** Additional comments:
**    As mentioned above, sock_poll() is the cornerstone for virtually 
**    all socket operations in this library.
**
**    On most other computer systems and other kernels, performing tasks
**    such as processing new packets, checking for timeouts and/or 
**    performing retransmissions are operations that run asynchronously 
**    in the background and scheduled by a task switcher.
**
**    Since DOS is a single task, non-preemptive operating system, no such 
**    background processing can occur and it is up to the application to
**    ensure that routine houskeeping chores get a chance to run.
**   
**    sock_poll() is this library's method for allowing such housekeeping
**    to to be performed.
*/
int sock_poll(void *s, int events, unsigned long timeout)
{
  if (s)
  {
    const struct wso_usrreq_tab FAR *ur = GETTAB_BY_SOCK(s);
    if (ur)
      return (*(ur->so_poll))(((struct sockio_info *)s), events, timeout );
    else
      return WSO_ASYSEL_EBADF;
  }
  if (timeout == ((unsigned long)-1))
  {
    errno = EINVAL;
    return -1;
  }
  do
  {
    unsigned int t = ~0u;
    if (timeout < t)
      t = (unsigned int)timeout;
    if (blocker_delay(t) != 0)
    {
      errno = EINTR;
      return -1; /* EINTR */
    }
    timeout -= t; 
  } while (timeout); 
  return 0;
}
  
/* --------------------------------------------------------------------- */

/*
** int tcp_tick( void *s );
**
** Description 
**    tcp_tick() is a compatibility wrapper that returns
**         (sock_poll(s, 0, 0UL) ? (0) : ( s != NULL ))
**
** Returns
**    0: 's' is NULL, or connection reset or closed by other host.
**   !0: Connection is fine. 
*/
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int tcp_tick( void *s )
{
  return (sock_poll(s, 0, 0UL) ? (0) : ( s != NULL ));
}
#endif

/* --------------------------------------------------------------------- */

/*
** int sock_established( void *s );
**
** Description 
**    TCP connections require a handshaked open to ensure that 
**    both sides recognize a connection. Whether the connection was 
**    initiated with tcp_open() or tcp_listen(), sock_establish() will 
**    continue to return zero until the connection is established, at which 
**    time it will return non-zero. It is not enough to spin on 
**    this after a listen because it is possible for the socket to be 
**    opened, written to and closed between two checks. Use sock_poll()
**    instead.
**
**    UDP is a connectionless protocol, hence will return non-zero 
**    immediately after a successful sock_connect() (or udp_open()) 
**    or after a socket initialized with udp_listen() is converted to
**    a "connected" socket when the first incoming data is seen.
**
**    BSD compatible TCP listeners (ie those that were put in listen mode
**    with sock_listen()) will also return non-zero if there is are one or
**    more connections ready to be sock_accept()'ed.
**
**    sock_established() is a simple compatibility wrapper around 
**    sock_poll() and sock_opt_get()
**
** Returns
**    0: Not established;
**   +1: Established
**   -1: No longer established.
*/
int sock_established( void *s )
{
  int rc = sock_opt_get(s, WSO_LF_STATE);
  switch (rc)
  {
    case tcp_StateLISTEN  :
         break; /* go check backwards compatible meaning of "established" */
    case tcp_StateSYNSENT :
    case tcp_StateSYNREC  :
         return 0;          /* not yet established */
    case tcp_StateESTAB   :
    case tcp_StateESTCL   :
    case tcp_StateFINWT1  :
    case tcp_StateFINWT2  :
    case tcp_StateCLOSWT  :
         return +1;         /* established */
    case tcp_StateCLOSING :
    case tcp_StateLASTACK :
    case tcp_StateTIMEWT  :
    case tcp_StateCLOSED  :
    default:
         return -1;         /* no longer established */
  }    
  /* accept for TCP (UDP listener can never accept) */
  rc = sock_poll( s, WSO_ASYSEL_ACCEPT, 0 );
  if (rc != -1 && rc != 0)
    return +1; /* established */
  return 0; /* not yet established */
}
#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int tcp_established (void *s) { return sock_established(s); }
#endif

/* --------------------------------------------------------------------- */

/*
** int sock_sselect( void *s, int state )
** 
** Description
**    sock_sselect() is a compatibility wrapper based on sock_poll()
**
**    'state' is one of the SOCK* state constants as described in
**    the Return section below, and determines what state is to
**    be checked.
**
** Returns
**    sock_sselect - returns one of several constants indicating
**           SOCKESTABLISHED - tcp connection has been established
**           SOCKDATAREAY    - tcp/udp data ready for reading
**           SOCKCLOSED      - socket has been closed
**    or zero to indicate none of the above.
**
**    NOTE: sock_sselect() will return SOCKESTABLISHED only when the
**    connection is truly established. This is unlike sock_established() 
**    (and therefore the sock_wait_established() macro), which returns a 
**    non-zero value if the connection was _ever_ established.
*/
int sock_sselect (void *s, int waitstate)
{
  int rc, events = 0;

  if (waitstate == SOCKDATAREADY)
  {
    events = WSO_ASYSEL_READ|WSO_ASYSEL_RDOOB;
  }
  else if (waitstate == SOCKESTABLISHED)
  {
    /* use sock_opt_get(WSO_LF_STATE) rather than sock_poll(WSO_ASYSEL_CONNECT) 
    ** because a) WSO_ASYSEL_CONNECT consumes the connect event.
    ** and b) we want to see tcp_StateESTCL and tcp_StateCLOSWT too.
    */
    rc = sock_opt_get(s, WSO_LF_STATE);
    if (rc == -1) /* bad sock */
      return SOCKCLOSED;  
    /* Note that the following is quite different from what 
    ** {tcp|sock}_established() and _ip_delay0() check for - 
    ** there, anything >= tcp_StateESTAB is considered 'established'.
    */
    if (rc == tcp_StateESTAB ||
        rc == tcp_StateESTCL ||
        rc == tcp_StateCLOSWT)
      return SOCKESTABLISHED;
    events = WSO_ASYSEL_ACCEPT; /* backwards compatible meaning of ESTAB */
  }

  rc = sock_poll( s, events, 0 );
  /* evaluation order is important here */
  if (rc & WSO_ASYSEL_EBADF)
    return SOCKCLOSED;
  if (rc & WSO_ASYSEL_HANGUP)
    return SOCKCLOSED;
  if (rc & events)
    return waitstate;
  /* WSO_ASYSEL_ERROR is not tested for - 's' is not SOCKCLOSED yet */
  return (0); 
}

/* --------------------------------------------------------------------- */

/*
** int sock_condwait( void *s, int condition, int waitsecs, 
**                    int (*blocking_hook)(void *), int *status,
**                    int _ip_delayX_compatible );
**
** Description
**     sock_condwait() allows the caller to determine if a 
**     condition has been been met, and optionally waiting until
**     that condition is met. It is a compatibility wrapper for 
**     the sock_wait_*() macros and is based on sock_poll().
**     The direct use of sock_condwait(), and the continued use
**     of the sock_wait_*() macros in not recommended.
**
**     'condition' is one of:
**     SOCKCLOSED      - the connection has been terminated or the
**                       socket is invalid.
**     SOCKDATAREADY   - Data is available for reading.
**     SOCKESTABLISHED - A connection has been established.
**
**     'waitsecs' is the number of seconds to wait before timing out.
**     A value of zero indicates the function should wait indefinitely. 
**     If interrupted, and not executing in compatibility-mode 
**     ('_ip_delayX_compatible' is zero), sock_condwait() times out 
**     immediately.
**
**     'blocking_hook' is an optional user-defined function to be called 
**     repeatedly while waiting. If 'blocking_hook' returns non-zero, the 
**     wait is cancelled, and sock_condwait() returns that value. 
**     The hook is called once per second when '_ip_delayX_compatible' is
**     zero, and continuously when '_ip_delayX_compatible' is non-zero.
**
**     'status' is a pointer to a status word and intended for 
**     compatibility with the sock_wait_*() macros that 'goto'
**     a user-defined label. (refer to the sock_wait_*() macros
**     for more information).
**     
**     '_ip_delayX_compatible' is a true/false flag that influences.
**     sock_condwait() backward-compatible behaviour. '_ip_delayX'
**     is primarily to support the 3 (now-obsolete) functions upon
**     which the sock_wait_*() macros were based.
**     If 'compatible' is set:
**       - sock_condwait() is not interruptable (does no ^C checking).
**       - calls 'blocking_hook' continuously rather than once per second.
**       - for SOCKESTABLISHED returns -1 on disconnect rather than +1
**       - issues a sock_shutdown(s,1) if the return value is -1.
**     sock_condwait() is significantly more efficient when 'compatible' 
**     is zero.
**
** Return
**     The values -1, 0, and +1 have the following reserved meanings:
**      0: the condition has been met.
**     +1: a disconnect was seen, or 's' is invalid.
**     -1: the wait timed out (or EINTR) or other cirumstances as 
**         described in the '_ip_delayX_compatible' clause above.
**     These and any other value may be returned by 'blocking_hook'.
*/
#undef  TRACE_CONDWAIT

int sock_condwait( void *s, int condition, int waitsecs, 
                   int (*blocking_hook)(void *), int *statusptr,
                   int _ip_delayX_compatible )
{
  #ifndef TRACE_CONDWAIT
  #define TRACE_CONDWAIT(x) NO_OPERATION()
  #else
  #undef TRACE_CONDWAIT
  #define TRACE_CONDWAIT(x) DBGPRINTF(x)
  #endif

  /* compat check chart:
   *
   *   Function    waited for    retcode/status    action upon timeout
   * ------------  ------------  ---------------    ------------------
   * _ip_delay0    ESTABLISHED   TO: -1 DISCO: -1   tcp_close() 
   * _ip_delay1    DATAREADY     TO: -1 DISCO: +1   tcp_close() 
   * _ip_delay2    CLOSED        TO: -1 DISCO: +1   tcp_abort() 
  */
  int status = +1; /* assume bad 's' */

  TRACE_CONDWAIT(("%s begin\n", 
           (condition == SOCKDATAREADY ? "DREADY" :
           (condition == SOCKESTABLISHED ? "ESTAB" : "CLOSED" )) ));
  if (s)
  {
    unsigned long timeout = 0ul;
    int poll_to, events = 0;

    if (waitsecs)
      timeout = timer_sec_count() + ((unsigned long)waitsecs);

    /* sock_poll() will do sigchecking only if timeout >= 110ms (2 ticks) */ 
    poll_to = 1000u;  /* once per sec */
    if (_ip_delayX_compatible) 
      poll_to = 100u; /* a little less than every 2 ticks */

    if (condition == SOCKDATAREADY)
    {
      /* no idea why a flush is necessary, but thats what was done before */
      sock_send_to( s, NULL, 0, 0, NULL, NULL ); /* flush */
      events = WSO_ASYSEL_READ|WSO_ASYSEL_RDOOB;
    }
    else if (condition == SOCKESTABLISHED)
    {
      /* previously did a tcp_established() here */
      events = WSO_ASYSEL_WRITE|WSO_ASYSEL_ACCEPT;
      /* also check BSD compatibile listeners for sock_accept()'ability */
    } 

    for (;;)
    {
      int revents = sock_poll(s, events, poll_to );

      if (revents == -1) 
      {
        TRACE_CONDWAIT(("return -1 due to EINTR\n"));
        status = -1;
        break;
      }
      else if (revents & WSO_ASYSEL_EBADF)
      {
        /*
         * treat EBADF as timeout - the caller is expecting
         * the sock to unusable after timeout.
        */
        TRACE_CONDWAIT(("return -1 due to EBADF\n"));
        status = +1;
        _ip_delayX_compatible = 0; /* can't close, so don't even try */
        break;
      }
      else if (revents & events)
      {
        TRACE_CONDWAIT(("return 0 due to (revents & events)!=0\n" ));
        status = 0;
        break;
      }
      else if (revents & (WSO_ASYSEL_HANGUP|WSO_ASYSEL_ERROR))
      {
        if (condition == SOCKESTABLISHED)
        {
          status = +1;   /* consistant with meaning of +1 */
          if (_ip_delayX_compatible)
            status = -1; /* compat with old return value */
          TRACE_CONDWAIT(("return %d due to WSO_ASYSEL_HANGUP/*_ERROR\n", status ));
          break;
        } 
        else if (condition == SOCKDATAREADY)
        {
          TRACE_CONDWAIT(("return +1 due to WSO_ASYSEL_HANGUP/*_ERROR\n" ));
          status = +1;
          break;
        }
        else /* condition == SOCKCLOSED */
        {
          if (sock_opt_get(s, WSO_LF_STATE) == tcp_StateCLOSED)
          {
            TRACE_CONDWAIT(("return 0 due to WSO_ASYSEL_HANGUP/*_ERROR\n" ));
            status = 0;
            break;
          }
          /* otherwise wait for it to reach closed */ 
        }
      }
      else if (waitsecs && timer_sec_count() > timeout)
      {
        TRACE_CONDWAIT(("return -1 due to timeout\n"));
        status = -1;
        break;
      }        
      else if (blocking_hook)
      {
        status = (*blocking_hook)(s);   
        if (status != 0)
        {
          TRACE_CONDWAIT(("return %d due to blocking_hook\n",status));
          break;
        }
      }

    } /* for (;;) */     
  } /* if (s) */

  if (status == -1 /* "timed out" */ && _ip_delayX_compatible)
  {
    /* do what the romans did for SOCKanything + TIMEOUT */
    int err = sock_opt_get(s, WSO_LF_ERRNO);
    sock_shutdown(s, 1); /* _ip_delay2 did a sock_abort() here. */
    sock_opt_set(s, WSO_LF_ERRNO, err);
  }

  if (statusptr)
    *statusptr = status;

  TRACE_CONDWAIT(("%s return %d\n", 
           (condition == SOCKDATAREADY ? "DREADY" :
           (condition == SOCKESTABLISHED ? "ESTAB" : "CLOSED" )),
           status));
  return status;
}

#ifndef DONT_GENERATE_CODE_FOR_MACROIZED_FXNS
int _ip_delay0(void *s, int seconds, int (*fn)(void *s), int *statusptr)
{ return sock_condwait(s,SOCKESTABLISHED,seconds,fn,statusptr,1); }
int _ip_delay1(void *s, int seconds, int (*fn)(void *s), int *statusptr)
{ return sock_condwait(s,SOCKDATAREADY,seconds,fn,statusptr,1); }
int _ip_delay2(void *s, int seconds, int (*fn)(void *s), int *statusptr)
{ return sock_condwait(s,SOCKCLOSED,seconds,fn,statusptr,1); }
#endif

/* ====================================================================== */
