/*
 * Reservation of local ports used by udp and tcp including
 * a) support for the "quiet time" reservation as required by RFC
 * b) support for the the REUSEADDR/REUSEPORT ioctl.
 *
 * The full 0-65535 port range is available for both UDP and TCP.
 * Port ranges are discrete - UDP has its set, and TCP has its set,
 * and ne'er the twain shall meet.
 * 
 * Created May 2001 by Cyrus Patel <cyp@fb14.uni-mainz.de>
 *
 * How REUSEADDR/REUSEPORT affect port allocation:
 * - SO_REUSEADDR allows a socket to bind to an address which is in a
 *   TIME_WAIT state. It does NOT allow more than one socket to bind to
 *   the same address. 
 * - SO_REUSEPORT allows multiple sockets to bind to the same address
 *   provided that all of them use SO_REUSEPORT. 
*/
#define __OSILAYER__ 4

#include "target.h"   /* uint32_t, FAR etc */
#include "timer.h"    /* timer_sec_count() */
#include "ctordtor.h" /* _my_atexit() */
#include "debug.h"    /* DBGPRINTF() */
#include "userland.h" /* WSO_LF_REUSEADDR|WSO_LF_REUSEPORT */
#include "printk.h"   /* _printk() */
#include "ipbase.h"   /* IPPROTO_* */
#include "utports.h"  /* ourselves */

/*
 * Manifest constants from netinet/in.h
 *
 * Ports < IPPORT_RESERVED are reserved for
 * privileged processes (e.g. root).         (IP_PORTRANGE_LOW)
 * Ports > IPPORT_USERRESERVED are reserved
 * for servers, not necessarily privileged.  (IP_PORTRANGE_DEFAULT)
*/
#ifndef IPPORT_RESERVED
#define IPPORT_RESERVED         1024
#define IPPORT_USERRESERVED     5000
#endif
/*
 * btw, BSD's have a setsockopt(IPPROTO_IP,IP_PORTRANGE) option
 * to change this default assignment range, and IP_PORTRANGE_DEFAULT to
 * reset it to default.
 *
 * But for crying out loud don't add all that "functionality" to this lib.
 * Its already way too bloated with "features" that nobody ever uses or needs.
 */

/* ------------------------------------------------------------------ */

struct port_waiter_t
{
  uint16_t myport;
  uint32_t expiry_time; /* time this slot expires (in secs) */
};

struct port_data_t
{
  struct port_waiter_t FAR *waiters;
  unsigned int num_waiters;
  uint16_t last_assigned_wild;
  int FAR *debug_lev_ptr;
  int (*is_port_in_use)(uint16_t port, uint32_t *locflags);
};

static struct 
{
  struct port_data_t tcp_data;
  struct port_data_t udp_data;
} FAR my_statics;

/* ------------------------------------------------------------------ */

static struct port_data_t FAR *__get_pdt( int proto )
{
  struct port_data_t FAR *pdt = (struct port_data_t FAR *)0;

  if (proto == IPPROTO_TCP)
    pdt = &my_statics.tcp_data;
  else if (proto == IPPROTO_UDP)
    pdt = &my_statics.udp_data;
  if (pdt)
  {
    if (!pdt->is_port_in_use)
    {
      /* this is a very, very serious condition since scanner callbacks
       * are registered from constructors. If a scanner isn't registered,
       * it could mean that the constructors weren't called.
       */
      __fatal_runtime_error("PANIC: Protocol failed to register port # scanner!\n" );
    }
  }

  return pdt;
}

/* ------------------------------------------------------------------ */

void localport_register_pius(int proto, int FAR *debug_lev_ptr,
                    int (*scanner)(uint16_t port, uint32_t *locflags) )
{
  /* a typical scanner looks like this:
  ** static int __tcp_is_port_in_use(uint16_t port, uint32_t *locflags)
  ** {
  **   tcp_pcb *s = _tcp_allsocs;
  **   while (s)
  **   {
  **     if (s->myport == port && s->state != tcp_StateCLOSED)
  **     {
  **       *locflags = s->locflags;
  **       return ((s->state >= tcp_StateLASTACK) ? 2 : 1);
  **     }    
  **     s = s->next;
  **   }
  **   return 0;
  ** }
  */
  if (proto == IPPROTO_TCP)
  {
    my_statics.tcp_data.is_port_in_use = scanner;
    my_statics.tcp_data.debug_lev_ptr = debug_lev_ptr;
  }
  else if (proto == IPPROTO_UDP)
  {
    my_statics.udp_data.is_port_in_use = scanner;
    my_statics.udp_data.debug_lev_ptr = debug_lev_ptr;
  }
  return;
}

/* ------------------------------------------------------------------ */

static void __at_exit_handler(void)
{
  unsigned int proto;
  for (proto = 0; proto < 2; proto++)
  {
    struct port_data_t FAR *pdt = __get_pdt(proto?IPPROTO_TCP:IPPROTO_UDP);
    if (pdt && pdt->num_waiters)
    {
      #if !defined(NDEBUG)
      if (*(pdt->debug_lev_ptr) > 1)
      {
        uint32_t now = timer_sec_count();
        unsigned int numw = pdt->num_waiters;
        struct port_waiter_t FAR *w = pdt->waiters;    
        unsigned int total = 0, found = 0;
        const char *protoname = (proto?"TCP":"UDP");

        while (numw)
        {
          total++;
          if (w->myport && now < w->expiry_time)
          {
            if (found == 0)
            {
              DBGPRINTF(("*--------------------------------------------\n"));
              DBGPRINTF(("*   port #   | reservation expires in\n" ));
              DBGPRINTF(("*--------------------------------------------\n"));
            }            
            DBGPRINTF(("* %6u/%s | %lu.0 secs\n", 
                         w->myport, protoname,  w->expiry_time - now ));
            found++;
          }
          w++;
          numw--;
        }
        if (found)
        {
          DBGPRINTF(("*--------------------------------------------\n"));
          DBGPRINTF(("* %u %s port slots open, %u active\n",
                        total, protoname, found ));
          DBGPRINTF(("*--------------------------------------------\n"));
        }
      }
      #endif

      if (pdt->waiters)
        FAR_FREE(pdt->waiters);
      pdt->waiters = (struct port_waiter_t FAR *)0;
      pdt->num_waiters = 0;
    }
  }
  return;
}

/* ------------------------------------------------------------------ */

void localport_release(uint16_t port, int proto, unsigned int waitsecs)
{
  struct port_data_t FAR *pdt = (struct port_data_t FAR *)0;

  if (waitsecs)
    pdt = __get_pdt( proto );

  if (pdt)
  {
    uint32_t now = timer_sec_count();
    uint32_t exp = now + waitsecs + 1;
    struct port_waiter_t FAR *freew = (struct port_waiter_t FAR *)0;

    #define WAITER_GROWSTEP (128u / sizeof(struct port_waiter_t))

    if (!pdt->num_waiters)
    {
      static int set_atexit = 0;
      if (!set_atexit)
        _my_atexit(__at_exit_handler);
      set_atexit = 1;

      freew = (struct port_waiter_t FAR *)FAR_MALLOC(
                          WAITER_GROWSTEP * sizeof(struct port_waiter_t));
      if (freew)
      {
        FAR_MEMSET( freew, 0, WAITER_GROWSTEP * sizeof(struct port_waiter_t));
        pdt->num_waiters = WAITER_GROWSTEP;
        pdt->waiters = freew;
      }
    } 
    else
    {
      struct port_waiter_t FAR *idlest = (struct port_waiter_t FAR *)0;
      struct port_waiter_t FAR *expired = (struct port_waiter_t FAR *)0;
      struct port_waiter_t FAR *w = pdt->waiters;
      unsigned int numw = pdt->num_waiters;
      uint32_t oldest = 0;

      while (numw)
      {
        if (!w->myport)
        {
          if (!freew)
            freew = w;
        }
        else if (w->myport == port)
        {
          if (exp < w->expiry_time)
            exp = w->expiry_time;
          break;  
        }
        else if (now > w->expiry_time)
        {
          if (!expired)
            expired = w;
        }
        else if (!oldest || w->expiry_time < oldest)
        {
          oldest = w->expiry_time;
          idlest = w;
        }
        w++;    
        numw--;
      }  
      if (!freew)
      {
        freew = expired;
        if (!freew)
        {
          w = (struct port_waiter_t FAR *)FAR_REALLOC( pdt->waiters, 
               (pdt->num_waiters + WAITER_GROWSTEP) * 
               sizeof(struct port_waiter_t));
          if (w)
          {
            freew = w + pdt->num_waiters; 
            FAR_MEMSET( freew, 0, WAITER_GROWSTEP * sizeof(struct port_waiter_t));
            pdt->num_waiters += WAITER_GROWSTEP;
            pdt->waiters = w;
          }
          if (!freew)
            freew = idlest;
        }
      }
    }

    if (freew)
    {
      freew->expiry_time = exp;
      freew->myport = port;
      if (*(pdt->debug_lev_ptr) > 1)
        DBGPRINTF(("*Port %u/%sP expires at %lu\n", port, 
                         (proto==IPPROTO_TCP?"TC":"UD"), exp ));
    }
  } /* if (pdt) */

  return;
}

/* ------------------------------------------------------------------ */

static int is_port_in_use(uint16_t port, int proto, uint32_t *locflags)
{
  struct port_data_t FAR *pdt = __get_pdt(proto);
  
  if (!pdt)
    return -1; /* error (bad proto) */ 
  else
  {
    int rc = (*(pdt->is_port_in_use))(port, locflags);
    if (rc != 0)
      return rc;

    /* the waiters walk must be done after the active sockets test */
    if (pdt->num_waiters)
    {
      struct port_waiter_t FAR *w = pdt->waiters;
      unsigned int numw = pdt->num_waiters;

      while (numw)
      {
        if (w->myport == port)
        {
          if (timer_sec_count() < w->expiry_time)
            break;
          *locflags = 0;
          return 2;  
        }
        w++;    
        numw--;
      }  
    }
  }

  return 0;
}

/* ------------------------------------------------------------------ */

uint16_t localport_reserve( uint16_t port, int proto, uint32_t so_locflags )
{
  struct port_data_t FAR *pdt = __get_pdt(proto);

  if (!pdt)
    port = 0;
  else
  {
    uint32_t loc_locflags;
    int rc;

    if (port == 0) /* "pick one for me" */
    {
      /* Ports <  IPPORT_RESERVED (1024) are reserved */
      /* Ports >  IPPORT_USERRESERVED (5000) are reserved */
      for (rc = 0; rc < (IPPORT_USERRESERVED-IPPORT_RESERVED); rc++)
      {        
        port = ++pdt->last_assigned_wild;
        if (port < IPPORT_RESERVED || port > IPPORT_USERRESERVED)
          port = pdt->last_assigned_wild = IPPORT_RESERVED;
        if (!is_port_in_use(port, proto, &loc_locflags))
          break;
        port = 0;
      }
    }
    else
    {
      rc = is_port_in_use(port, proto, &loc_locflags);
      if (rc == 2) /* port is in TIMEWT state */
      {
        if ((loc_locflags & WSO_LF_REUSEADDR) == 0)
          port = 0;
      }
      else if (rc == 1) /* port is actively in use */
      {
        /* sharing only possible if _all_ port reservations are REUSEPORT */
        if (((so_locflags & loc_locflags) & WSO_LF_REUSEPORT) == 0)
          port = 0;
      }  
      else if (rc != 0) /* error */
      {
        port = 0;
      }
    }
  }
  return port;
}

/* ----------------------------------------------------------------------- */
